// This file is part of Speedtest kernel module
//
// Speedtest kernel module is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// Speedtest kernel module is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Speedtest kernel module. If not, see <https://www.gnu.org/licenses/>.

#define ENABLE_DEBUG_LOG 0
#define ENABLE_TRACE_LOG 0

#include "speedtest.h"
#include "logging.h"
#include <linux/version.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/file.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/net.h>
#include <linux/in.h>
#include <net/sock.h>
#include <linux/kthread.h>
#include <linux/fdtable.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/crc32.h>
#include <linux/sched.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0)
#include <linux/sched/types.h>
#endif

#define HASH_NONFATAL_OOM 1
#define uthash_malloc(sz) kmalloc(sz, GFP_KERNEL)
#define uthash_free(ptr,sz) kfree(ptr)
#include "uthash.h"


#define SO_NETFILTER_BYPASS 60


MODULE_LICENSE("GPL");
MODULE_AUTHOR("David Hedbor <david@ookla.com>");
MODULE_AUTHOR("Mitja Horvat <mitja@plume.com>");
MODULE_AUTHOR("Bostjan Meglic <bostjan@plume.com>");
MODULE_DESCRIPTION("Speedtest accelerator module.");
MODULE_VERSION("0");

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 0, 0)
#define PRE_40_KERNEL 1
#elif LINUX_VERSION_CODE < KERNEL_VERSION(4, 7, 0)
#define sock_recvmsg(sock, msg, flags)                          \
    sock_recvmsg((sock), (msg), msg_data_left(msg), (flags))
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0)
#define st_access_ok(perm, ptr, size) access_ok(ptr, size)
#else
#define st_access_ok(perm, ptr, size) access_ok(perm, ptr, size)
#endif

static const char *const DEVICE_NAME = "speedtest";
static const char *const CLASS_NAME = "benchmark";

static unsigned int majorNumber;
struct class *speedtestClass = NULL;
static struct device *speedtestDevice = NULL;

static long speedtestDeviceIoctl(struct file *f, unsigned int cmd, unsigned long _udata);

static int handleSpeedtestReceiveRequest(struct file *fd, struct speedtestKernelRequest *request);

static ssize_t speedtestDeviceRead(struct file *fd, char *user_buffer, size_t number_of_chars_to_be_read, loff_t *offset);

static int speedtestDeviceRelease(struct inode *ind, struct file *fd);

static int speedtestDeviceOpen(struct inode *ind, struct file *fd);

static DEFINE_SPINLOCK(operationsLock);

struct operationDetails {
    struct socket *networkFd;
    struct file *statusFd;
    spinlock_t lock;
    struct task_struct *task;
};

struct socketOperation {
    // Data related to the request
    union {
        struct socketReceiveRequest recv;
    };
    // Data related to the operation
    struct operationDetails op;
    // Response data
    atomic64_t pendingReadBytes;
    // Hash table details
    u32 hashKey;
    UT_hash_handle hh;
};

struct socketOperation *socketOperations = NULL;

unsigned int s_numActive = 0;

static struct file_operations speedtestFileOps = {
        .owner = THIS_MODULE,
        .unlocked_ioctl = &speedtestDeviceIoctl,
        .read = &speedtestDeviceRead,
        .release = &speedtestDeviceRelease,
        .open = &speedtestDeviceOpen,
};

inline u32 generateHashKey(struct file *fd) {
    ptrdiff_t addr = (ptrdiff_t)fd;
    return crc32(SPEEDTEST_IOCTL_SOCKET_RECV, &addr, sizeof(addr));
}

static int __init speedtestModuleInit(void) {
    int deviceMajorRC;
    LOG_INFO("Initializing Device Driver.");

    // Allocate a major number for the character device dynamically.
    deviceMajorRC = register_chrdev(0U, DEVICE_NAME, &speedtestFileOps);
    if(deviceMajorRC < 0) {
        LOG_ALERT("Device major number registration failed.");
        return deviceMajorRC;
    }
    majorNumber = (unsigned int)deviceMajorRC;
    LOG_INFO("Registered character device with major number %d.", majorNumber);

    // Register the device class.
    speedtestClass = class_create(THIS_MODULE, CLASS_NAME);
    if(IS_ERR(speedtestClass)) {
        // Check for error and clean up if there is
        unregister_chrdev(majorNumber, DEVICE_NAME);
        LOG_ALERT("Failed to register device class.");
        return (int)PTR_ERR(speedtestClass);
    }
    LOG_INFO("Device class registered.");

    // Register the device driver for the previously allocated class and major number
    speedtestDevice = device_create(speedtestClass, NULL, MKDEV(majorNumber, 0U), NULL, DEVICE_NAME);
    if(IS_ERR(speedtestDevice)) {
        // Clean up if there is an error
        class_destroy(speedtestClass);
        unregister_chrdev(majorNumber, DEVICE_NAME);
        LOG_ALERT("Failed to create the device.");
        return (int)PTR_ERR(speedtestDevice);
    }
    LOG_INFO("Device class created correctly.");

    return 0;
}

static void __exit speedtestModuleExit(void) {
    device_destroy(speedtestClass, MKDEV(majorNumber, 0U));     // remove the device
    class_unregister(speedtestClass);                          // unregister the device class
    class_destroy(speedtestClass);                             // remove the device class
    unregister_chrdev(majorNumber, DEVICE_NAME);             // unregister the major number
    LOG_INFO("Deregistered Speedtest device driver.");
}

// Handle ioctl's on the device. This is used to configure a new session with the file descriptor,
// send command data and number of bytes to receive etc.
static long speedtestDeviceIoctl(struct file *f, unsigned int cmd, unsigned long _udata) {
    int rc;
    switch(cmd) {
        case SPEEDTEST_IOCTL_SOCKET_RECV: {
            void *udata = (void *) _udata;
            struct speedtestKernelRequest recv;

            if(!st_access_ok(VERIFY_READ, udata, sizeof(struct speedtestKernelRequest))) {
                return -EFAULT;
            }

            rc = copy_from_user(&recv, udata, sizeof(recv));
            if(rc > 0) {
                LOG_ERROR("Failed to copy data from user space.");
                return -EINVAL;
            }
            rc = handleSpeedtestReceiveRequest(f, &recv);
            if(rc < 0) {
                LOG_ERROR("Error initiating socket receive.");
            }
            return rc;
        }
        case SPEEDTEST_IOCTL_QUERY_API_LEVEL:
            LOG_DEBUG("Queried kernel module version");
            return SPEEDTEST_API_LEVEL;
    }
    return -EINVAL;
}

#define SPEEDTEST_RECV_SIZE  ((ssize_t)1024*1024)

int sendCommand(struct socketOperation *socketOp) {
    struct msghdr msg;
    struct iovec iov[1];
    int rc = 0;
    if(socketOp->recv.commandLen > 0) {
        memset(&msg, 0, sizeof(msg));
        memset(iov, 0, sizeof(iov));

        iov[0].iov_base = &socketOp->recv.commandString;
        iov[0].iov_len = strlen(socketOp->recv.commandString);
#ifdef PRE_40_KERNEL
        msg.msg_iov = &iov[0];
        msg.msg_iovlen = 1;
        rc = sock_sendmsg(socketOp->op.networkFd, &msg, socketOp->recv.commandLen);
#else
        iov_iter_init(&msg.msg_iter, WRITE, iov, 1, socketOp->recv.commandLen);
        rc = sock_sendmsg(socketOp->op.networkFd, &msg);
#endif
        if(rc < 0) {
            LOG_ERROR("Send failed: %d", rc);
        } else {
            LOG_TRACE("Total Sent: %d bytes.", rc);
        }
    }
    return rc;
}

int speedtestReceiveRequestThread(void *data) {
    struct socketOperation *socketOp = (struct socketOperation *) data;
    struct msghdr msg;
    struct iovec iov[1];

    struct msghdr statusmsg;
    struct iovec statusiov[1];
    int recvFlags = MSG_TRUNC | socketOp->recv.flags;
    int rc = 0;
    ssize_t rxbytes;

    /* Receive stuff */
    memset(&msg, 0, sizeof(msg));
    memset(iov, 0, sizeof(iov));
    iov[0].iov_len = SPEEDTEST_RECV_SIZE;
#ifdef PRE_40_KERNEL
    msg.msg_iov = &iov[0];
    msg.msg_iovlen = 1;
#else
    iov_iter_init(&msg.msg_iter, READ, iov, 1, iov[0].iov_len);
#endif
    memset(&statusmsg, 0, sizeof(statusmsg));
    memset(iov, 0, sizeof(statusiov));

    if(sendCommand(socketOp) < 0) {
        goto exit;
    }

    do {
        rc = 0;
        rxbytes = socketOp->recv.rxBytes;
        /* Write the commandString first */
        if(sendCommand(socketOp) < 0) {
            break;
        }
        while(rxbytes > 0 && rc >= 0) {
            if(kthread_should_stop()) {
                // We're exiting because the device is being closed - it takes care of cleanup
                return -1;
            }
            iov[0].iov_len = min((size_t)rxbytes, socketOp->recv.blockSize);
#ifdef PRE_40_KERNEL
            msg.msg_iov = &iov[0];
            rc = sock_recvmsg(socketOp->op.networkFd, &msg, iov[0].iov_len, recvFlags);
#else
            iov_iter_init(&msg.msg_iter, READ, iov, 1, iov[0].iov_len);
            rc = sock_recvmsg(socketOp->op.networkFd, &msg, recvFlags);
#endif
            if(rc < 0) {
                LOG_ERROR("Error on recv: %d", rc);
                break;
            }
            LOG_TRACE("Received %d bytes, %ld bytes remaining.", rc, (long)rxbytes);
            rxbytes -= rc;
            atomic64_add(rc, &socketOp->pendingReadBytes);
        }
    } while(rc >= 0 && !kthread_should_stop());

    if(rc >= 0) {
        rc = (int)socketOp->recv.rxBytes;
    }
    exit:
    if(!kthread_should_stop()) {
        LOG_DEBUG("Exiting downloader thread.");
        spin_lock(&socketOp->op.lock);
        socketOp->op.task = NULL;
        spin_unlock(&socketOp->op.lock);
    }
    do_exit(rc);
}

int handleSpeedtestReceiveRequest(struct file *fd, struct speedtestKernelRequest *request) {
    struct socket *sock;
    int sockerr;
    int rc;
    struct sched_param sch_param;
    struct socketOperation *socketOp = kmalloc(sizeof(struct socketOperation), GFP_KERNEL);
    if(!socketOp) {
        LOG_ERROR("Unable to allocate memory for socket operation.");
        return -1;
    }
    memset(socketOp, 0, sizeof(struct socketOperation));
    LOG_DEBUG("Creating socket for fd %d.", request->fd);
    sock = sockfd_lookup(request->fd, &sockerr);
    if(!sock) {
        LOG_ERROR("Invalid socket: %d => %d\n", request->fd, sockerr);
        return sockerr;
    }
    socketOp->hashKey = generateHashKey(fd);
    memcpy(&socketOp->recv, &request->recv, sizeof(request->recv));
    if(!socketOp->recv.blockSize) {
        socketOp->recv.blockSize = SPEEDTEST_RECV_SIZE;
    }
    socketOp->op.networkFd = sock;
    socketOp->op.task = kthread_run(speedtestReceiveRequestThread, socketOp, "speedtest/%u-recv", socketOp->hashKey);
    socketOp->op.statusFd = fd;
    atomic64_set(&socketOp->pendingReadBytes, 0);

    if(request->sched.scheduler >= 0) {
        sch_param.sched_priority = request->sched.priority;
        rc = sched_setscheduler(socketOp->op.task, request->sched.scheduler, &sch_param);
        if(rc != 0) {
            LOG_ERROR("sched_setscheduler failed: %d", rc);
        }
    }
    if(socketOp->op.task != NULL) {
        LOG_INFO("Started new task id %u.", socketOp->hashKey);
        spin_lock(&operationsLock);
        HASH_ADD_INT(socketOperations, hashKey, socketOp);
        spin_unlock(&operationsLock);
    } else {
        LOG_ERROR("Failed to start new task");
        kfree(socketOp);
    }
    return socketOp->op.task != NULL ? 0 : -1;
}


static int speedtestDeviceOpen(struct inode *ind, struct file *fd) {
    spin_lock(&operationsLock);
    s_numActive++;
    spin_unlock(&operationsLock);
    LOG_DEBUG("Opening new file descriptor - %u active.", s_numActive);
    return 0;
}

/// This method is called when read() is called on the device and ios used
/// to ret
static ssize_t speedtestDeviceRead(struct file *fd, char *user_buffer, size_t number_of_chars_to_be_read, loff_t *offset) {
    struct socketOperation *socketOp = NULL;
    ssize_t numRead = -1;
    size_t msgLen = sizeof(numRead);
    u32 key = generateHashKey(fd);
    int rc = (int)msgLen;
    spin_lock(&operationsLock);
    HASH_FIND_INT(socketOperations, &key, socketOp);
    spin_unlock(&operationsLock);
    if(!socketOp) {
        return -EINVAL;
    }
    numRead = atomic64_xchg(&socketOp->pendingReadBytes, 0);
    if(numRead > 0) {
        int copyval;
        copyval = copy_to_user(&user_buffer[0], &numRead, msgLen);
        if(copyval > 0) {
            LOG_ERROR("Failed to copy bytes read data to user space.");
            return -EINVAL;
        }
        if(*offset) {
            *offset += msgLen;
        }
    } else {
        int isComplete;
        spin_lock(&socketOp->op.lock);
        isComplete = socketOp->op.task == NULL;
        spin_unlock(&socketOp->op.lock);
        if(numRead < 0 || isComplete) {
            rc = -1;
        }
    }
    return rc;
}

static int speedtestDeviceRelease(struct inode *ind, struct file *fd) {
    struct socketOperation *iter = NULL;
    struct task_struct *task = NULL;
    u32 key = generateHashKey(fd);
    LOG_DEBUG("Closing device FD.");
    spin_lock(&operationsLock);
    HASH_FIND_INT(socketOperations, &key, iter);
    if(iter) {
        // Store the task pointer, so we can stop it outside lock since it's possible
        // that it would be set to pull between this locked region and the unlocked
        // code below.
        task = iter->op.task;
        iter->op.task = NULL;
        HASH_DEL(socketOperations, iter);
    }
    --s_numActive;
    spin_unlock(&operationsLock);
    if(iter != NULL) {
        // Need to clean up the task here.
        if(task != NULL) {
            LOG_DEBUG("Stopping task id %u", key);
            kthread_stop(task);
        } else {
            LOG_DEBUG("Task id %u already ended.", key);
        }
        sockfd_put(iter->op.networkFd);
        kfree(iter);
    }
    LOG_DEBUG("Transfer Task ended.");
    return 0;
}


module_init(speedtestModuleInit)

module_exit(speedtestModuleExit)
