/*
 *  Copyright (C) 2013 Gabor Juhos <juhosg@openwrt.org>
 *  Copyright (C) 2014 Felix Fietkau <nbd@nbd.name>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License version 2 as published
 *  by the Free Software Foundation.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <linux/byteorder/generic.h>

#include "mtdsplit.h"
#define DKMGT_PTN_NAME_LEN		(32)
#define DKMGT_PTN_MAX_CNT		(64)

//#define MTDSPLIT_DBG
#ifdef MTDSPLIT_DBG
#define DBG_PRINTK(fmt,args...)	printk(fmt, ##args)
#else
#define DBG_PRINTK(fmt,args...)
#endif

struct dkmgt_ptn
{
    char        name[DKMGT_PTN_NAME_LEN];
    uint32_t    base;
    uint32_t    size;
    uint32_t    attr;
};
 
struct dkmgt_ptn_table
{
    uint32_t disk_size;
    uint32_t cnt;
    struct dkmgt_ptn ptns[DKMGT_PTN_MAX_CNT];
};

static struct dkmgt_ptn_table omada_ptn_table;

static int mtdsplit_parse_omada(struct mtd_info *master,
				struct mtd_partition **pparts,
				struct mtd_part_parser_data *data)
{
	struct mtd_partition *parts;
	size_t retlen;
	uint32_t i, cnt;
	uint32_t offset, size;
	uint64_t base0;
	int err;

	err = mtd_read(master, 0, sizeof(omada_ptn_table), &retlen, (void *) &omada_ptn_table);
	if (err)
	{
		DBG_PRINTK("[mtdsplit_parse_omada] MTD read failed.\r\n");
		return err;
	}
	
	if (sizeof(omada_ptn_table) != retlen)
	{
		DBG_PRINTK("[mtdsplit_parse_omada] Read length not match.\r\n");
		return -EIO;
	}

	cnt = be32_to_cpu(omada_ptn_table.cnt);
	if (!cnt)
	{
		return 0;
	}
	
	parts = kzalloc(cnt * sizeof(*parts), GFP_KERNEL);
	if (!parts)
	{
		DBG_PRINTK("[mtdsplit_parse_omada] kzalloc failed.\r\n");
		return -ENOMEM;
	}
	
	offset = omada_ptn_table.ptns[0].base;
	base0 = (uint64_t)be32_to_cpu(offset);
	for (i = 0; i < cnt; i++)
	{
		offset = omada_ptn_table.ptns[i].base;
		size = omada_ptn_table.ptns[i].size;
		parts[i].name = omada_ptn_table.ptns[i].name;
		parts[i].offset = (uint64_t)be32_to_cpu(offset) - base0;
		parts[i].size = (uint64_t)be32_to_cpu(size);
		DBG_PRINTK("================== [mtdsplit_parse_omada] Parse partition#%d: name=%s ==================\r\n"
				"offset=0x%016llx, size=0x%016llx.\r\nBefore be32_to_cpu: offset=0x%08x, size=0x%08x.\r\nAfter be32_to_cpu: offset=0x%08x, size=0x%08x.\r\n",
				i, parts[i].name, parts[i].offset, parts[i].size, offset, size, be32_to_cpu(offset), be32_to_cpu(size));
	}

	*pparts = parts;
	return cnt;
}

static struct mtd_part_parser mtdsplit_omada_parser = {
	.owner = THIS_MODULE,
	.name = "omada-fw",
	.parse_fn = mtdsplit_parse_omada,
	.type = MTD_PARSER_TYPE_FIRMWARE,
};

static int __init mtdsplit_omada_init(void)
{
	register_mtd_parser(&mtdsplit_omada_parser);
	return 0;
}

subsys_initcall(mtdsplit_omada_init);