/** @file      bittorrent_events.h
  * @brief     BitTorrent event definitions.
  * @details   @copydoc doc_bittorrent
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef BITTORRENT_EVENTS_H
#define BITTORRENT_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Group ******************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced BITTORRENT event.
 */
#define PACE2_ADVANCED_BITTORRENT_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/
/**
 * @brief Types for advanced bittorrent events.
 */
typedef enum {

    /** Sent for the BitTorrent handshake (once for each direction of the TCP connection); @c data points to
     * a struct ipd_bittorrent_handshake. */
    PACE2_ADVANCED_BITTORRENT_HANDSHAKE,

    /** Reports a BitTorrent 'choke' message; @c data points to a struct ipd_bittorrent_dir containing
     * the direction of the underlying TCP segment. */
    PACE2_ADVANCED_BITTORRENT_CHOKE,

    /** Reports a BitTorrent 'unchoke' message; @c data points to a struct ipd_bittorrent_dir containing
     * the direction of the underlying TCP segment. */
    PACE2_ADVANCED_BITTORRENT_UNCHOKE,

    /** Reports a BitTorrent 'interested' message; @c data points to a struct ipd_bittorrent_dir containing
     * the direction of the underlying TCP segment. */
    PACE2_ADVANCED_BITTORRENT_INTERESTED,

    /** Reports a BitTorrent 'not interested' message; @c data points to a struct ipd_bittorrent_dir
     * containing the direction of the underlying TCP segment. */
    PACE2_ADVANCED_BITTORRENT_NOTINTERESTED,

    /** Reports a BitTorrent 'have' message; @c data points to a struct ipd_bittorrent_have. */
    PACE2_ADVANCED_BITTORRENT_HAVE,

    /** Reports a BitTorrent 'bitfield' message; @c data points to a struct ipd_bittorrent_bitfield. */
    PACE2_ADVANCED_BITTORRENT_BITFIELD,

    /** Reports a BitTorrent 'request' message; @c data points to a struct ipd_bittorrent_request. */
    PACE2_ADVANCED_BITTORRENT_REQUEST,

    /** Reports a BitTorrent 'piece' message; @c data points to a struct ipd_bittorrent_piece.*/
    PACE2_ADVANCED_BITTORRENT_PIECE,

    /** Reports a BitTorrent 'cancel' message; @c data points to a struct ipd_bittorrent_request.*/
    PACE2_ADVANCED_BITTORRENT_CANCEL,

    /** Sent when decoding the message stream in one direction fails; @c data points to
     * a struct ipd_bittorrent_invalid. */
    PACE2_ADVANCED_BITTORRENT_INVALID,

    /** Signals termination or timeout of the underlying TCP flow. Is not sent for flows where both directions
     * have been flagged as invalid. Does not carry additional data. */
    PACE2_ADVANCED_BITTORRENT_END_OF_FLOW,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_BITTORRENT_EVENTS

} PACE2_advanced_bittorrent_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/** Length in bytes of the data pointed to by the @link ipd_bittorrent_handshake::reserved reserved@endlink
 * field of a BitTorrent handshake. */
#define IPD_BITTORRENT_EXTENSIONS_LEN 8

/** Length in bytes of an SHA-1 hash used in the
 * @link ipd_bittorrent_handshake::info_hash info_hash@endlink
 * and @link ipd_bittorrent_handshake::peer_id peer_id@endlink
 * fields of a BitTorrent handshake. */
#define IPD_SHA1_LEN 20

/** BitTorrent handshake from one direction. */
struct ipd_bittorrent_handshake {

    /** Direction (corresponds to @ref PACE2_packet_descriptor::direction). */
    u8 dir;

    /** ::IPD_BITTORRENT_EXTENSIONS_LEN bytes of reserved/extension data (all zero in plain
     * <a href="http://www.bittorrent.org/beps/bep_0003.html">BEP 3</a>
     * BitTorrent). */
    u8 * reserved;

    /** Identifier of the torrent (SHA-1 hash of the info portion of the .torrent file,
     * length ::IPD_SHA1_LEN bytes). */
    u8 * info_hash;

    /** Identifier of the peer sending the handshake, length ::IPD_SHA1_LEN bytes */
    u8 * peer_id;
};

/** Direction flag for BitTorrent messages without payload. */
struct ipd_bittorrent_dir {
    /** Sender of the message (corresponds to @ref PACE2_packet_descriptor::direction) */
    u8 dir;
};

/** Announcement that a peer has received a piece (ie, update to the bitfield transferred in a previous
 * @link ipd_bittorrent_bitfield 'bitfield'@endlink message). */
struct ipd_bittorrent_have {

    /** Sender of the message (corresponds to @ref PACE2_packet_descriptor::direction). */
    u8 dir;

    /** Index of the newly available piece. */
    u32 index;
};

/** Set of all pieces a peer has available. */
struct ipd_bittorrent_bitfield {

    /** Sender of the message (corresponds to @ref PACE2_packet_descriptor::direction). */
    u8 dir;

    /** Bitfield with each index that the peer has available set to one and the rest set to zero;
     * the piece with index <em>i</em> is represented by bit <em>(7-i)</em> of byte <em>i%%8</em>. */
    struct ipd_string bitfield;
};

/** Request for a byte range within a piece */
struct ipd_bittorrent_request {

    /** Sender of the message (corresponds to @ref PACE2_packet_descriptor::direction). */
    u8 dir;

    /** Index of the piece. */
    u32 index;

    /** Start offset within the piece. */
    u32 offset;

    /** Number of bytes requested. */
    u32 length;
};

/** Transfer of a range of bytes from a piece. */
struct ipd_bittorrent_piece {

    /** Sender of the message (corresponds to @ref PACE2_packet_descriptor::direction). */
    u8 dir;

    /** boolean flag; set if this event does not correspond to actual data, but a missing TCP segment. */
    u8 is_bad_data;

    /** Index of the piece. */
    u32 index;

    /** Start offset within the piece. */
    u32 offset;

    /** Data buffer. */
    struct ipd_string data;
};

/** Decoder error for one direction of a BitTorrent flow. */
struct ipd_bittorrent_invalid {

    /** Direction (corresponds to @ref PACE2_packet_descriptor::direction). */
    u8 dir;

    /** Description of the error. */
    char const * error_message;
};

/**
 * @brief Data structure for advanced BITTORRENT events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_BITTORRENT_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** BITTORRENT event subtype. */
    PACE2_advanced_bittorrent_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_BITTORRENT_HANDSHAKE event. */
        struct ipd_bittorrent_handshake handshake;

        /** Contains data for a @ref PACE2_ADVANCED_BITTORRENT_CHOKE event. */
        struct ipd_bittorrent_dir choke;

        /** Contains data for a @ref PACE2_ADVANCED_BITTORRENT_UNCHOKE event. */
        struct ipd_bittorrent_dir unchoke;

        /** Contains data for a @ref PACE2_ADVANCED_BITTORRENT_INTERESTED event. */
        struct ipd_bittorrent_dir interested;

        /** Contains data for a @ref PACE2_ADVANCED_BITTORRENT_NOTINTERESTED event. */
        struct ipd_bittorrent_dir notinterested;

        /** Contains data for a @ref PACE2_ADVANCED_BITTORRENT_HAVE event. */
        struct ipd_bittorrent_have have;

        /** Contains data for a @ref PACE2_ADVANCED_BITTORRENT_BITFIELD event. */
        struct ipd_bittorrent_bitfield bitfield;

        /** Contains data for a @ref PACE2_ADVANCED_BITTORRENT_REQUEST event. */
        struct ipd_bittorrent_request request;

        /** Contains data for a @ref PACE2_ADVANCED_BITTORRENT_PIECE event. */
        struct ipd_bittorrent_piece piece;

        /** Contains data for a @ref PACE2_ADVANCED_BITTORRENT_CANCEL event. */
        struct ipd_bittorrent_request cancel;

        /** Contains data for a @ref PACE2_ADVANCED_BITTORRENT_INVALID event. */
        struct ipd_bittorrent_invalid invalid;

    } event_data;

} PACE2_advanced_bittorrent_event;

#endif /* PACE_KERNEL_LIB */

#endif /* BITTORRENT_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
