/** @file      chat_events.h
  * @brief     Chat class event definitions.
  * @details   @copydoc doc_class_chat
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef CHAT_EVENTS_H
#define CHAT_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "generic_events.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the chat class event.
 */
#define PACE2_CLASS_CHAT_EVENT_GROUP PACE2_CLASS_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for CHAT class events.
 */
typedef enum {

    /** struct ipd_class_chat_session_struct. */
    PACE2_CLASS_CHAT_SESSION_START,

    /** struct ipd_class_chat_connection_struct. */
    PACE2_CLASS_CHAT_NEW_CONNECTION,

    /** struct ipd_class_chat_session_struct. */
    PACE2_CLASS_CHAT_SESSION_END,

    /** struct ipd_class_chat_user_struct. */
    PACE2_CLASS_CHAT_USER,

    /** struct ipd_class_chat_user_status_struct. */
    PACE2_CLASS_CHAT_USER_STATUS,

    /** struct ipd_class_chat_user_info_struct. */
    PACE2_CLASS_CHAT_USER_INFO,

    /** struct ipd_class_chat_chatroom_struct. */
    PACE2_CLASS_CHAT_CHATROOM_OPEN,

    /** struct ipd_class_chat_chatroom_struct. */
    PACE2_CLASS_CHAT_CHATROOM_CLOSE,

    /** struct ipd_class_chat_join_leave_struct. */
    PACE2_CLASS_CHAT_CHATROOM_USER_JOIN,

    /** struct ipd_class_chat_join_leave_struct. */
    PACE2_CLASS_CHAT_CHATROOM_USER_LEAVE,

    /** struct ipd_class_chat_message_struct. */
    PACE2_CLASS_CHAT_CHATROOM_MESSAGE,

    /** struct ipd_class_chat_transfer_announcement_struct. */
    PACE2_CLASS_CHAT_CHATROOM_TRANSFER_ANNOUNCEMENT,

    /** struct ipd_class_chat_transfer_struct. */
    PACE2_CLASS_CHAT_CHATROOM_TRANSFER,

    /** Announcement of associated VoIP session. */
    PACE2_CLASS_CHAT_CHATROOM_VOIP_ANNOUNCEMENT,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_CLASS_CHAT_EVENT_TYPES

} PACE2_class_CHAT_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/** Enumerator for the type of a chat session. */
enum ipd_chat_class_type {
    IPD_CHAT_CLASS_TYPE_OSCAR,
    IPD_CHAT_CLASS_TYPE_YAHOO,
    IPD_CHAT_CLASS_TYPE_RESERVED_1,
    IPD_CHAT_CLASS_TYPE_JABBER,
    IPD_CHAT_CLASS_TYPE_RESERVED_2,
    IPD_CHAT_CLASS_TYPE_IRC,
    IPD_CHAT_CLASS_TYPE_RESERVED_3,
    IPD_CHAT_CLASS_TYPE_RESERVED_4,
    IPD_CHAT_CLASS_TYPE_ICQ_WEBCHAT,
    IPD_CHAT_CLASS_TYPE_RESERVED_5,
    IPD_CHAT_CLASS_TYPE_RESERVED_6,
};

/** Enumerator for the type of status. */
enum ipd_chat_class_status {
    IPD_CHAT_CLASS_STATUS_OFFLINE,
    IPD_CHAT_CLASS_STATUS_ONLINE,
    IPD_CHAT_CLASS_STATUS_INVISIBLE,
    IPD_CHAT_CLASS_STATUS_AWAY,
    IPD_CHAT_CLASS_STATUS_NOT_AVAILABLE,
    IPD_CHAT_CLASS_STATUS_DO_NOT_DISTURB,
    IPD_CHAT_CLASS_STATUS_BUSY,
    IPD_CHAT_CLASS_STATUS_IDLE,
};

/** Enumerator for the type of user info. */
enum ipd_chat_class_user_info {
    IPD_CHAT_CLASS_USER_INFO_GENERIC,
    IPD_CHAT_CLASS_USER_INFO_FIRST_NAME,
    IPD_CHAT_CLASS_USER_INFO_LAST_NAME,
};


/** Enumerator for the state of a data transfer announcement. */
enum ipd_chat_class_transfer_announcement_state {
    IPD_CHAT_CLASS_TRANSFER_ANNOUNCEMENT_NOT_YET_SEEN,
    IPD_CHAT_CLASS_TRANSFER_ANNOUNCEMENT_ANNOUNCED,
    IPD_CHAT_CLASS_TRANSFER_ANNOUNCEMENT_ACKNOWLEDGED,
    IPD_CHAT_CLASS_TRANSFER_ANNOUNCEMENT_REJECTED,
    IPD_CHAT_CLASS_TRANSFER_ANNOUNCEMENT_NOT_SUPPORTED,
};

/** Enumerator for the state of a single file transfer in a data transfer. */
enum ipd_chat_class_transfer_state {
    IPD_CHAT_CLASS_TRANSFER_STARTED,
    IPD_CHAT_CLASS_TRANSFER_FINISHED,
    IPD_CHAT_CLASS_TRANSFER_CANCELED,
    IPD_CHAT_CLASS_TRANSFER_INCOMPLETE,
};

/** Main chat session structure, each chat session has this event as main event. */
struct ipd_class_chat_session_struct {

    /** ID of the session. */
    struct ipd_class_generic_header generic;

    /** Chat type of the session oscar,yahoo,jabber,... */
    enum ipd_chat_class_type type;

    /** The unique id of the user whom this session belongs to. */
    u64 session_user_id;

    /** The time when this session first appeared. */
    u64 first_ts;

    /** The time when the last update to this session happened. */
    u64 last_ts;
};

/** Structure for events that provide information about connection establishment. */
struct ipd_class_chat_connection_struct {

    /** Unique id for connection event. */
    struct ipd_class_generic_header generic;

    /** ID of the new connection. */
    u64 connection_id;
};

/** Structure for events that provide generic user information. */
struct ipd_class_chat_user_struct {

    /** Unique id for user event. */
    struct ipd_class_generic_header generic;

    /** ID of the chat session the user belongs to. */
    u64 chat_session_id;

    /** String with the name of the user. */
    struct ipd_charset_string name;

    /** String with the nickname of the user. */
    struct ipd_charset_string nickname;
};

/** Structure for events that provide chatroom information. */
struct ipd_class_chat_chatroom_struct {

    /** Unique id for chatroom event. */
    struct ipd_class_generic_header generic;

    /** Id of the chat session the chatroom belongs to. */
    u64 chat_session_id;

    /** String with the name of the chatroom. */
    struct ipd_charset_string name;

    /** String with a description of the chatroom. */
    struct ipd_charset_string description;
};

/** Structure for events that provide information about a user joining or leaving a chatroom. */
struct ipd_class_chat_join_leave_struct {

    /** Unique id for join/leave event. */
    struct ipd_class_generic_header generic;

    /** ID of the chat session the user-chatroom interaction relates to. */
    u64 chat_session_id;

    /** ID of the chatroom the user-chatroom interaction relates to. */
    u64 chat_chatroom_id;

    /** ID of the user the user-chatroom interaction relates tp. */
    u64 chat_user_id;
};

/** Structure for events that provide information about a message sent to a chatroom. */
struct ipd_class_chat_message_struct {

    /** Unique id for message event. */
    struct ipd_class_generic_header generic;

    /** ID of the chatroom the message is sent to. */
    u64 chat_chatroom_id;

    /** ID of the chat session the message belongs to. */
    u64 chat_session_id;

    /** ID of the user who is sening the message. */
    u64 sender_id;

    /** String with the message text. */
    struct ipd_charset_string message;
};

/** Structure for events that provide information about a file transfer announcement. */
struct ipd_class_chat_transfer_announcement_struct {

    /** Unique id for object transfer event. */
    struct ipd_class_generic_header generic;

    /** ID of the chat session the file transfer belongs to. */
    u64 session_id;

    /** ID of the chatroom the file transfer belongs to. */
    u64 chat_chatroom_id;

    /** ID of the user who is sending the file to the chatroom. */
    u64 sender_id;

    /** Array with the sizes of the transferred files in bytes. */
    u64 * complete_file_size_array;

    /** String array with the names of the transferred files. */
    struct ipd_charset_string * file_name_array;

    /** Length of the array with the names and the sizes of the transferred files. */
    u64 file_array_length;

    /** The state of the file transfer. */
    enum ipd_chat_class_transfer_announcement_state state;
};

/** Structure for events that provide information about a file transfer and/or a segment of file transfer data. */
struct ipd_class_chat_transfer_struct {

    /** Unique id for file transfer event. */
    struct ipd_class_generic_header generic;

    /** ID of the correlating transfer announcement. This ID was reported with
     * struct ipd_class_chat_transfer_announcement_struct. */
    u64 transfer_announcement_id;

    /** The size of the transferred file in bytes. */
    u64 complete_file_size;

    /** String with the name of the transferred file. */
    struct ipd_charset_string file_name;

    /** String with the description of the transferred file. */
    struct ipd_charset_string file_description;

    /** Offset of the segment of file transfer data in relation to the whole file. */
    u64 data_offset;

    /** String with the actual segment of file transfer data. */
    struct ipd_charset_string data;

    /** The state of the file transfer. */
    enum ipd_chat_class_transfer_state state;
};

/** Structure for events that provide information about the online status of a user. */
struct ipd_class_chat_user_status_struct {

    /** Unique id for the status update event. */
    struct ipd_class_generic_header generic;

    /** ID of the chat session the user belongs to. */
    u64 session_id;

    /** ID of the user for whom the online status is provided. */
    u64 user_id;

    /** The online status of the related user. */
    enum ipd_chat_class_status status;
};

/** Structure for events that provide user profile information. */
struct ipd_class_chat_user_info_struct {

    /** Unique id for the chat user event. */
    struct ipd_class_generic_header generic;

    /** ID of the chat session the user belongs to. */
    u64 session_id;

    /** ID of the user for whom the information is provided. */
    u64 user_id;

    /** The type of information which is provided. */
    enum ipd_chat_class_user_info info_type;

    /** String with the information provided. */
    struct ipd_charset_string info;
};

/** Structure for events that provide information about a VoIP announcement. */
struct pace2_class_chat_voip_announcement_struct {

    /** Unique id for object transfer event.*/
    struct ipd_class_generic_header generic;

    /** ID of the chat session the file transfer belongs to. */
    u64 session_id;

    /** ID of the chatroom the file transfer belongs to. */
    u64 chat_chatroom_id;

    /** ID of the VoIP session. */
    u64 voip_session_id;
};

/**
 * @brief Data structure for CHAT class events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_CLASS_CHAT_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** CHAT event subtype. */
    PACE2_class_CHAT_event_type meta_data_type;

    PACE2_CLASS_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_CLASS_CHAT_SESSION_START event. */
        struct ipd_class_chat_session_struct session_start;

        /** Contains data for a @ref PACE2_CLASS_CHAT_NEW_CONNECTION event. */
        struct ipd_class_chat_connection_struct new_connection;

        /** Contains data for a @ref PACE2_CLASS_CHAT_SESSION_END event. */
        struct ipd_class_chat_session_struct session_end;

        /** Contains data for a @ref PACE2_CLASS_CHAT_USER event. */
        struct ipd_class_chat_user_struct user;

        /** Contains data for a @ref PACE2_CLASS_CHAT_USER_STATUS event. */
        struct ipd_class_chat_user_status_struct user_status;

        /** Contains data for a @ref PACE2_CLASS_CHAT_USER_INFO event. */
        struct ipd_class_chat_user_info_struct user_info;

        /** Contains data for a @ref PACE2_CLASS_CHAT_CHATROOM_OPEN event. */
        struct ipd_class_chat_chatroom_struct chatroom_open;

        /** Contains data for a @ref PACE2_CLASS_CHAT_CHATROOM_CLOSE event. */
        struct ipd_class_chat_chatroom_struct chatroom_close;

        /** Contains data for a @ref PACE2_CLASS_CHAT_CHATROOM_USER_JOIN event. */
        struct ipd_class_chat_join_leave_struct chatroom_user_join;

        /** Contains data for a @ref PACE2_CLASS_CHAT_CHATROOM_USER_LEAVE event. */
        struct ipd_class_chat_join_leave_struct chatroom_user_leave;

        /** Contains data for a @ref PACE2_CLASS_CHAT_CHATROOM_MESSAGE event. */
        struct ipd_class_chat_message_struct chatroom_message;

        /** Contains data for a @ref PACE2_CLASS_CHAT_CHATROOM_TRANSFER_ANNOUNCEMENT event. */
        struct ipd_class_chat_transfer_announcement_struct chatroom_transfer_announcement;

        /** Contains data for a @ref PACE2_CLASS_CHAT_CHATROOM_TRANSFER event. */
        struct ipd_class_chat_transfer_struct chatroom_transfer;

        /** Contains data for a @ref PACE2_CLASS_CHAT_CHATROOM_VOIP_ANNOUNCEMENT event. */
        struct pace2_class_chat_voip_announcement_struct voip_announcement;

    } event_data;

} PACE2_class_CHAT_event;

#endif

#endif /* CHAT_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
