/** @file      dhcp_events.h
  * @brief     DHCP event definitions.
  * @details   @copydoc doc_dhcp
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef DHCP_EVENTS_H
#define DHCP_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the basic DHCP event.
 */
#define PACE2_BASIC_DHCP_EVENT_GROUP PACE2_POLLING_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned metadata group of the advanced DHCP event.
 */
#define PACE2_ADVANCED_DHCP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief DHCP metadata flags of a basic DHCP event.
 */
enum PACE2_basic_DHCP_event_mask {

    /** Host IP-address flag. */
    PACE2_BASIC_DHCP_YOUR_IP_ADDR = 0,

    /** Client MAC-address flag. */
    PACE2_BASIC_DHCP_CLIENT_MAC_ADDR,

    /** DHCP lease time flag. */
    PACE2_BASIC_DHCP_ADDR_LEASE_TIME,

    /** DHCP message type flag. */
    PACE2_BASIC_DHCP_MESSAGE_TYPE,

    /** Number of implemented basic DHCP flags. */
    PACE2_NUMBER_OF_BASIC_DHCP_FLAGS
};

/**
 * @brief Types of advanced DHCP events.
 */
typedef enum {

    /** Fields extracted from the header of each DHCP message; @c data points to
     * a struct ipd_dhcp_general_packet_information.*/
    PACE2_ADVANCED_DHCP_GENERAL_PACKET_INFORMATION,

    /** Represents one option or sub-option inside a DHCP message; @c data points to a struct ipd_dhcp_option. */
    PACE2_ADVANCED_DHCP_OPTION,

    /** Fields extracted from the header of a DHCPv6 message; data points to a struct @ref ipd_dhcpv6_message. */
    PACE2_ADVANCED_DHCPV6_MESSAGE,

    /** Number of implemented advanced DHCP events. */
    PACE2_NUMBER_OF_ADVANCED_DHCP_EVENT_SUBTYPES

} PACE2_advanced_DHCP_event_type;

/* Event Structures *************************************************************/

/**
 * @brief Data structure of a basic DHCP events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_BASIC_DHCP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** Bitmask, which states the available DHCP metadata. */
    PACE2_bitmask meta_data_mask;

    /** Variable of the @ref PACE2_BASIC_DHCP_YOUR_IP_ADDR flag.
        Contains a IPv4-address in network byte order. */
    u32 your_ip_addr;

    /** Data structure of the @ref PACE2_BASIC_DHCP_CLIENT_MAC_ADDR flag. */
    PACE2_byte_buffer client_mac_addr;

    /** Variable of the @ref PACE2_BASIC_DHCP_ADDR_LEASE_TIME flag. */
    u32 addr_lease_time;

    /** Variable of the @ref PACE2_BASIC_DHCP_MESSAGE_TYPE flag. */
    u8 message_type;

} PACE2_basic_DHCP_event;

#ifndef PACE_KERNEL_LIB

/** Enumeration for DHCP Message Types. */
enum ipd_dhcp_message_type_enum {
    IPD_DHCP_MESSAGE_TYPE_REQUEST = 1,
    IPD_DHCP_MESSAGE_TYPE_REPLY = 2,
};

/** Hardware Types, equates to the OSI-Model Layer 1. */
enum ipd_dhcp_hardware_type_enum {
    IPD_DHCP_HARDWARE_TYPE_ETHERNET = 1,
};

/** Enumeration of option message types. */
enum ipd_dhcp_option_message_type_enum {
    IPD_DHCP_OPTION_MESSAGE_TYPE_DISCOVER = 1,
    IPD_DHCP_OPTION_MESSAGE_TYPE_OFFER = 2,
    IPD_DHCP_OPTION_MESSAGE_TYPE_REQUEST = 3,
    IPD_DHCP_OPTION_MESSAGE_TYPE_DECLINE = 4,
    IPD_DHCP_OPTION_MESSAGE_TYPE_ACK = 5,
    IPD_DHCP_OPTION_MESSAGE_TYPE_NACK = 6,
    IPD_DHCP_OPTION_MESSAGE_TYPE_RELEASE = 7,
    IPD_DHCP_OPTION_MESSAGE_TYPE_INFORM = 8,
};

/** Enumerates the different option types. */
enum ipd_dhcp_option_type_enum {

    /** Subnet Mask. */
    IPD_DHCP_OPTION_TYPE_SUBNET_MASK = 1,

    /** Requested IP Address. */
    IPD_DHCP_OPTION_TYPE_REQ_IP = 50,

    /** IP Address Lease Time. */
    IPD_DHCP_OPTION_TYPE_IP_LEASE_TIME = 51,

    /** DHCP Message Type. */
    IPD_DHCP_OPTION_TYPE_DHCP_MESSAGE_TYPE = 53,

    /** Server Identifier. */
    IPD_DHCP_OPTION_TYPE_DHCP_SERVER_IP = 54,

    /** Parameter Request List. */
    IPD_DHCP_OPTION_TYPE_PARAMETER_REQ_LIST = 55,

    /** Renewal (T1) Time Value. */
    IPD_DHCP_OPTION_TYPE_RENEWAL_TIME_VAL = 58,

    /** Rebinding (T2) Time Value. */
    IPD_DHCP_OPTION_TYPE_REBINDING_TIME_VAL = 59,

    /** Client-identifier. */
    IPD_DHCP_OPTION_TYPE_CLIENT_IDENTIFIER = 61,

    /** Relay Agent Information.
     *  One #PACE2_ADVANCED_DHCP_OPTION event is sent for each sub-option. */
    IPD_DHCP_OPTION_TYPE_RELAY_AGENT = 82,

    /** End of valid information. */
    IPD_DHCP_OPTION_TYPE_END = 255,
};

/** This event contains the information sent with every single DHCP packet. */
struct ipd_dhcp_general_packet_information {

    /** Message op code/message type - request and reply. */
    enum ipd_dhcp_message_type_enum message_type;

    /** Hardware address type. */
    enum ipd_dhcp_hardware_type_enum hardware_type;

    /** Client hardware address. */
    struct ipd_string client_hardware_address;

    /** Transaction ID (in network byte order). */
    u32 transaction_id;

    /** Hardware address length (6 for 10mb ethernet). */
    u8 hardware_address_length;

    /** Client sets to zero, optionally used by relay agents when booting via a relay agent. */
    u8 hops;

    /** Seconds elapsed since client began address acquisition or renewal process. (in network byte order). */
    u16 secs;

    /** First bit BROADCAST flag, otherwise ZERO (in network byte order). */
    u16 flags;

    /** client IP address. */
    struct in_addr  client_ip_address;

    /** 'your' IP address. */
    struct in_addr  your_ip_address;

    /** IP address of the next server to use in bootstrap. */
    struct in_addr  server_ip_address;

    /** Relay agent IP address, used in booting via a relay agent. */
    struct in_addr  agent_ip_address;

    /** Server host name. */
    struct ipd_charset_string server_host_name;

    /** Boot file name. */
    struct ipd_charset_string file;
};

/** Enumeration for the content type of a DHCP option. */
enum ipd_dhcp_content_type {

    /** Signals that the content type of the DHCP option is unknown.
     * The field content.data is set. */
    IPD_DHCP_CONTENT_TYPE_UNKNOWN,

    /** Signals that the content type of the DHCP option is an IP.
     * The field content.ip is set. */
    IPD_DHCP_CONTENT_TYPE_IP,

    /** Signals that the content type of the DHCP option is a message type.
     * The field content.message_type is set. */
    IPD_DHCP_CONTENT_TYPE_MESSAGE_TYPE,

    /** Signals that the content type of the DHCP option is a time value.
     * The field content.ip_time_value is set. */
    IPD_DHCP_CONTENT_TYPE_TIME,

    /** Signals that the content type of the DHCP option is a suboption.
     * The field content.suboption is set. */
    IPD_DHCP_CONTENT_TYPE_SUBOPTION,

    /** Signals that the content type of the DHCP option is uninterpreted.
     * The field content.data is set. */
    IPD_DHCP_CONTENT_TYPE_UNINTERPRETED,

    /** Signals that the DHCP option has no explicit content.
     * No content field is set. */
    IPD_DHCP_CONTENT_TYPE_NONE,
};

/** Data decoded from one option or sub-option in a DHCP message. */
struct ipd_dhcp_option {

    /** Transaction ID (in network byte order), ie. the xid field of the DHCP message. */
    u32 transaction_id;

    /** Option type. */
    enum ipd_dhcp_option_type_enum option_type;

    /** Length of the option or sub-option value in bytes (without the type and length byte). */
    u8 option_length;

    /** Type of the options content. */
    enum ipd_dhcp_content_type content_type;

    union {

        /** Option value: DHCP message type (option type ::IPD_DHCP_OPTION_TYPE_DHCP_MESSAGE_TYPE). */
        enum ipd_dhcp_option_message_type_enum message_type;

        /** Suboption struct with type and data member. */
        struct ipd_dhcp_option_suboption {

            /** Sub-option type (option type ::IPD_DHCP_OPTION_TYPE_RELAY_AGENT) .*/
            u8 type;

            /** Uninterpreted byte array (sub-option value for ::IPD_DHCP_OPTION_TYPE_RELAY_AGENT's sub-options). */
            struct ipd_string data;

        } suboption;

        /** Uninterpreted byte array (option value for
         * ::IPD_DHCP_OPTION_TYPE_PARAMETER_REQ_LIST, ::IPD_DHCP_OPTION_TYPE_CLIENT_IDENTIFIER,
         * and option types not explicitly listed in ::ipd_dhcp_option_type_enum). */
        struct ipd_string data;

        /** Option value: IP address in network byte order
         * (option types ::IPD_DHCP_OPTION_TYPE_SUBNET_MASK, ::IPD_DHCP_OPTION_TYPE_REQ_IP,
         * ::IPD_DHCP_OPTION_TYPE_DHCP_SERVER_IP). */
        struct in_addr ip;

        /** Option value: Time in seconds (option types ::IPD_DHCP_OPTION_TYPE_IP_LEASE_TIME,
         * ::IPD_DHCP_OPTION_TYPE_RENEWAL_TIME_VAL, ::IPD_DHCP_OPTION_TYPE_REBINDING_TIME_VAL). */
        u32 ip_time_value;

    } content;
};

typedef struct ipd_dhcp_option ipd_dhcp_option_int_t;

/** @anchor dhcpv6_msg_types
    @name DHCPv6 message type values.
    As listed in <a href="http://tools.ietf.org/html/rfc3315#section-5.3">RFC 3315, section 5.3</a>
    @see ipd_dhcpv6_message::msg_type
    @{
*/

#define IPD_DHCPV6_MSG_SOLICIT 1
#define IPD_DHCPV6_MSG_ADVERTISE 2
#define IPD_DHCPV6_MSG_REQUEST 3
#define IPD_DHCPV6_MSG_CONFIRM 4
#define IPD_DHCPV6_MSG_RENEW 5
#define IPD_DHCPV6_MSG_REBIND 6
#define IPD_DHCPV6_MSG_REPLY 7
#define IPD_DHCPV6_MSG_RELEASE 8
#define IPD_DHCPV6_MSG_DECLINE 9
#define IPD_DHCPV6_MSG_RECONFIGURE 10
#define IPD_DHCPV6_MSG_INFORMATION_REQUEST 11
#define IPD_DHCPV6_MSG_RELAY_FORW 12
#define IPD_DHCPV6_MSG_RELAY_REPL 13
/** @} */

/** @anchor dhcpv6_status_codes
    @name DHCPv6 status codes
    As listed in <a href="http://tools.ietf.org/html/rfc3315#section-24.4">RFC 3315, section 24.4</a>
    @see ipd_dhcpv6_status::status_code
    @{
*/
#define IPD_DHCPV6_STATUS_SUCCESS 0
#define IPD_DHCPV6_STATUS_UNSPEC_FAIL 1
#define IPD_DHCPV6_STATUS_NO_ADDRS_AVAIL 2
#define IPD_DHCPV6_STATUS_NO_BINDING 3
#define IPD_DHCPV6_STATUS_NOT_ON_LINK 4
#define IPD_DHCPV6_STATUS_USE_MULTICAST 5
/** @} */


/** @brief Representation of a Status Code option
    @details If no Status Code option is present, #status_code is set to IPD_DHCPV6_STATUS_SUCCESS, and
    #status_message is set to { { NULL, 0 }, CHARSET_STRING_CHARSET_UNKNOWN }.
*/
struct ipd_dhcpv6_status {

    /** @brief status code
        @details The status-code field of the option; the values defined in RFC 3315 are available as
        @ref dhcpv6_status_codes "preprocessor defines"
    */
    u8 status_code;

    /** @brief Message.
        @details The status-message field of the option, a UTF-8 encoded text string.
    */
    struct ipd_charset_string status_message;
};

/** @brief Representation of an Option Request. */
struct ipd_dhcpv6_option_request {
    /** A list of Option Requests in the DHCPv6 message. */
    struct PACE2_u16_list option_requests;
};

/** @brief Representation of Authentication Option. */
struct ipd_dhcpv6_option_authentication {

    /** The replay detection information for the RDM. */
    u64 replay_detection;

    /** The authentication information, as specified by the protocol and
        algorithm used in this Authentication Option.
    */
    struct ipd_string authentication_information;

    /** The authentication protocol used in this Authentication Option. */
    u8 protocol;

    /** The algorithm used in the authentication protocol. */
    u8 algorithm;

    /** The replay detection method used in this Authentication Option. */
    u8 rdm;
};

/** @brief Representation of not yet supported Option fields. */
struct ipd_dhcpv6_other_option {

    /** Uninterpreted byte array of option-data
        The data for the option.
    */
    struct ipd_string option_data;

    /** An unsigned integer identifying the specific option type carried in this option. */
    u16 option_code;
};

/** @brief Representation of the Vendor Class Option.
 *  @details This option is used by a client to identify the vendor that
 *           manufactured the hardware on which the client is running.
 */
struct ipd_dhcpv6_option_vendor_class {

    /** The vendor's registered Enterprise Number as maintained by IANA. */
    u32 enterprise_number;

    /** The hardware configuration of the node on which the client is running. */
    struct ipd_string vendor_class_data;
};

/** @brief Representation of the Vendor-specific Information Option.
 *  @details This option is used by clients and servers to exchange
 *           vendor-specific information.
 */
struct ipd_dhcpv6_option_vendor_specific_information {

    /** The vendors registered Enterprise Number as maintained by IANA. */
    u32 enterprise_number;

    /** Vendor options, interpreted by vendor-specific code on the clients and servers. */
    struct ipd_string vendor_option_data;
};

/** @brief Representation of the Identity Association for Prefix Delegation Option. */
struct ipd_dhcpv6_option_identity_association_pa {

    /** The unique identifier for this IA_PD. */
    u32 iaid;

    /** The time interval after which the client should
        contact the server from which the prefixes in the
        IA_PD were obtained to extend the lifetimes of the
        prefixes delegated to the IA_PD.
     */
    u32 t1;

    /** The time interval after which the client should
        contact any available server to extend the lifetimes
        of the prefixes assigned to the IA_PD.
     */
    u32 t2;

    /** Options associated with this IA_PD. */
    struct ipd_string ia_pd_options;
};

/** @brief Representation of the IA Prefix Option. */
struct ipd_dhcpv6_option_ia_prefix {

    /** The preferred lifetime for the prefix in the option, expressed in units of seconds. */
    u32 preferred_lifetime;

    /** The valid lifetime for the prefix in the option, expressed in units of seconds. */
    u32 valid_lifetime;

    /** An IPv6 address. */
    struct in6_addr address;

    /** Options associated with this prefix. */
    struct ipd_string iaprefix_options;

    /** Length of this prefix in bits. */
    u8 prefix_length;
};

/** @brief Representation of an IA Address option. */
struct ipd_dhcpv6_address {

    /** @brief IPv6 address. */
    struct in6_addr address;

    /** @brief Preferred lifetime (in seconds) for the address. */
    u32 preferred_lifetime;

    /** @brief Valid lifetime (in seconds) for the address. */
    u32 valid_lifetime;

    /** @brief Status indication.
        @details contents of the Status Code option contained within the IA Address option.
    */
    struct ipd_dhcpv6_status status;
};

/** @brief Identity association types. */
enum ipd_dhcpv6_ia_type {

    /** Identity association for non-temporary addresses (IA_NA option). */
    IPD_DHCPV6_IA_NA,

    /** Identity association for temporary addresses (IA_TA option). */
    IPD_DHCPV6_IA_TA,
};

/** @brief Representation of an Identity Association (IA_NA or IA_TA) option.
*/
struct ipd_dhcpv6_identity_assoc {

    /** Unique identifier for this IA. */
    u32 ia_id;

    /** Kind of identity association (for non-temporary or for temporary addresses). */
    enum ipd_dhcpv6_ia_type ia_type;

    /** @brief T1 parameter (IA_NA only).
        @details time at which the client sends a Renew message to the server originating this IA_NA (in
        seconds, relative to current time); 0 lets the client choose a time, 0xffffffff stands for "never".
    */
    u32 t1;

    /** @brief T2 parameter (IA_NA only).
        @details time at which the client sends a Rebind message to any available server (in seconds, relative
        to current time); 0 lets the client choose a time, 0xffffffff stands for "never".
    */
    u32 t2;

    /** Pointer to an array containing the IA Address options. */
    const struct ipd_dhcpv6_address * addresses;

    /** Number of elements in #addresses. */
    u64 number_of_addresses;

    /** @brief Status indication.
        @details contents of the Status Code option contained within the IA_NA option.
    */
    struct ipd_dhcpv6_status status;
};

/** @brief Representation of an DHCPv6 message. */
struct ipd_dhcpv6_message {

    /* common header */

    /** Msg-type field of the DHCPv6 message; the values defined in RFC 3315 are available as
        @ref dhcpv6_msg_types "preprocessor defines". */
    u8 msg_type;

    /** @brief Transaction ID. */
    u32 transaction_id;

    /** @brief The DHCP unique identifier (DUID) of the client.
        @details Empty if the message contains no client identifier option. */
    struct ipd_string client_identifier;

    /** @brief The link-layer address of the client.
        @details Extracted from the client DUID (types DUID-LLT or DUID-LL); empty if the client identifier
        is not present or of an other type.
    */
    struct ipd_string client_l2_address;

    /** @brief The DHCP unique identifier (DUID) of the server.
        @details Empty if the message contains no server identifier option.
    */
    struct ipd_string server_identifier;

    /** @brief The link-layer address of the server.
        @details Extracted from the server DUID (types DUID-LLT or DUID-LL); empty if the server identifier
        is not present or of an other type.
    */
    struct ipd_string server_l2_address;

    /** @brief Status indication.
        @details Contents of the message-level status code option.
    */
    struct ipd_dhcpv6_status status;

    /** @brief Rapid commit flag.
        @details Set to 1 if the message contains a Rapid Commit option, 0 otherwise. */
    u8 rapid_commit;

    /** Pointer to an array containing the identity association options. */
    struct ipd_dhcpv6_identity_assoc const * ia_options;

    /** @brief Number of elements in @ref ipd_dhcpv6_message::ia_options "ia_options". */
    u64 number_of_ia_options;

    /** @brief Authentication Option. */
    struct ipd_dhcpv6_option_authentication option_auth;

    /** @brief Pointer to array of not yet supported DHCPv6 options. */
    struct ipd_dhcpv6_other_option const * other_options;

    /** @brief Number of elements in @ref ipd_dhcpv6_message::other_options "other_options". */
    u64 number_of_other_options;

    /** @brief IA Prefix Option. */
    struct ipd_dhcpv6_option_ia_prefix option_ia_prefix;

    /** @brief Identity Association for Prefix Delegation Option. */
    struct ipd_dhcpv6_option_identity_association_pa option_ia_pd;

    /** @brief Vendor Class Option. */
    struct ipd_dhcpv6_option_vendor_class option_vendor_class;

    /** @brief Vendor-specific Information Option. */
    struct ipd_dhcpv6_option_vendor_specific_information option_vendor_opts;

    /** @brief Option Request Option.
        @details Contents of the option request code option.
    */
    struct ipd_dhcpv6_option_request option_request;

    /** @brief Information Refresh Time Option. */
    u32 option_information_refresh_time;

    /** @brief SOL_MAX_RT Option. */
    u32 option_sol_max_rt;

    /** @brief INF_MAX_RT Option. */
    u32 option_inf_max_rt;

    /** @brief Server Unicast Option. */
    struct in6_addr option_server_unicast_address;

    /** @brief Relay Message Option. */
    struct ipd_string option_relay_msg;

    /** @brief User Class Option. */
    struct ipd_string option_user_class;

    /** @brief Interface-ID Option. */
    struct ipd_string option_interface_id;

    /** @brief Elapsed Time Option. */
    u16 option_elapsed_time;

    /** @brief Preference Option. */
    u8 option_preference_value;

    /** @brief Reconfigure Message Option. */
    u8 option_reconfigure_message_type;

    /** @brief Reconfigure Accept flag.
        @details Set to 1 if the message contains a Reconfigure Message option, 0 otherwise. */
    u8 option_reconfigure_accept;
};

typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_DHCP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** Advanced DHCP event subtype. */
    PACE2_advanced_DHCP_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    union {

        /** Contains data for a @ref PACE2_ADVANCED_DHCP_GENERAL_PACKET_INFORMATION event. */
        struct ipd_dhcp_general_packet_information gpi;

        /** Contains data for a @ref PACE2_ADVANCED_DHCP_OPTION event. */
        struct ipd_dhcp_option option;

        /** Contains data for a @ref PACE2_ADVANCED_DHCPV6_MESSAGE event. */
        struct ipd_dhcpv6_message dhcpv6_message;
    };

} PACE2_advanced_DHCP_event;

#endif

#endif /* DHCP_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
