/** @file      dns_events.h
  * @brief     DNS event definitions.
  * @details   @copydoc doc_dns
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef DNS_EVENTS_H
#define DNS_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced DNS event.
 */
#define PACE2_ADVANCED_DNS_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP
#define PACE2_BASIC_DNS_EVENT_GROUP PACE2_BASIC_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced DNS events.
 */
typedef enum {

    /** Provides a struct ipd_dns_query * with the transaction id, name, type and class of a query. */
    PACE2_ADVANCED_DNS_QUERY,

    /** Provides a struct ipd_dns_query * with the transaction id, name, type and class of a query
     * from a server response packet. */
    PACE2_ADVANCED_DNS_QUERY_RESPONSE,

    /** Provides a struct ipd_dns_answer_ipv4 * with the transaction id, name, type, class, time to live
     * and IPv4 address of an IPv4 answer. */
    PACE2_ADVANCED_DNS_ANSWER_IPV4,

    /** Provides a struct ipd_dns_answer_ipv6 * with the transaction id, name, type, class, time to live
     * and IPv6 address of an IPv6 answer. */
    PACE2_ADVANCED_DNS_ANSWER_IPV6,

    /** Provides a struct ipd_dns_answer_canonical_name * with the transaction id, name, type, class,
     * time to live and canonical name of a canonical name answer. */
    PACE2_ADVANCED_DNS_ANSWER_CANONICAL_NAME,

    /** Provides a struct ipd_dns_answer_mail_exchange * with the transaction id, name, type, class,
     * time to live, preference and mail exchange name of a mail exchange answer. */
    PACE2_ADVANCED_DNS_ANSWER_MAIL_EXCHANGE,

    /** Provides a struct ipd_dns_answer_pointer_domain_name * with the transaction id, name, type,
     * class, time to live, and the pointer domain name. */
    PACE2_ADVANCED_DNS_ANSWER_DOMAIN_NAME_POINTER,

    /** Provides a struct ipd_dns_answer_text_string * with the transaction id, name, type, class,
     * time to live, and the text string. */
    PACE2_ADVANCED_DNS_ANSWER_TEXT_STRING,

    /** Provides a struct ipd_dns_answer_server_selection * with the transaction id, name, type, class,
     * time to live, priority, weight, port, and target. */
    PACE2_ADVANCED_DNS_ANSWER_SERVER_SELECTION,

    /** Provides a struct ipd_dns_answer_other * with the transaction id, name, type, class,
     * time to live and generic data of an unimplemented answer. */
    PACE2_ADVANCED_DNS_ANSWER_OTHER,

    /** This event is issued when decoding of an TCP DNS connection fails.
     * It provides a char const * to a message describing the failure reason. */
    PACE2_ADVANCED_DNS_INVALID,

    /** This event is thrown if an answer is completely parsed. */
    PACE2_ADVANCED_DNS_ANSWER_COMPLETE,

    /** This event signals end of flow. */
    PACE2_ADVANCED_DNS_END_OF_FLOW,

    /** Provides a struct ipd_dns_answer_authoritative_name_server * with the transaction id, name, type, class,
     * time to live and authoritative name server of an authoritative name server answer. */
    PACE2_ADVANCED_DNS_ANSWER_AUTHORITATIVE_NAME_SERVER,

    /** Provides a struct ipd_dns_answer_start_of_authority * with the transaction id, name, type, class,
     * time to live and start of authority of a start of authority answer. */
    PACE2_ADVANCED_DNS_ANSWER_START_OF_AUTHORITY,

    /** Provides a struct ipd_dns_answer_https * with the transaction id, name, type, class,
     * time to live and data of an HTTPS answer. */
    PACE2_ADVANCED_DNS_ANSWER_HTTPS,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_DNS_EVENT_TYPES

} PACE2_advanced_DNS_event_type;

/**
 * @brief Types for basic DNS events.
 */
typedef enum {
    PACE2_BASIC_DNS_QUERY,
    PACE2_BASIC_DNS_QUERY_RESPONSE,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_BASIC_DNS_EVENT_TYPES

} PACE2_basic_DNS_event_type;

/** Enumeration for DNS answer types.
 */
typedef enum {

    PACE2_DNS_RESOURCE_RECORD_TYPE_ERROR = 0,

    /** A host address. */
    PACE2_DNS_RESOURCE_RECORD_TYPE_A = 1,

    /** An authoritative name server. */
    PACE2_DNS_RESOURCE_RECORD_TYPE_NS = 2,

    /** The canonical name for an alias. */
    PACE2_DNS_RESOURCE_RECORD_TYPE_CNAME = 5,

    /** Marks the start of a zone of authority. */
    PACE2_DNS_RESOURCE_RECORD_TYPE_SOA = 6,

    /** A well known service description. */
    PACE2_DNS_RESOURCE_RECORD_TYPE_WKS = 11,

    /** A domain name pointer. */
    PACE2_DNS_RESOURCE_RECORD_TYPE_PTR = 12,

    /** Host information. */
    PACE2_DNS_RESOURCE_RECORD_TYPE_HINFO = 13,

    /** Mailbox or mail list information. */
    PACE2_DNS_RESOURCE_RECORD_TYPE_MINFO = 14,

    /** Mail exchange. */
    PACE2_DNS_RESOURCE_RECORD_TYPE_MX = 15,

    /** Text string. */
    PACE2_DNS_RESOURCE_RECORD_TYPE_TXT = 16,

    /** IPv6. */
    PACE2_DNS_RESOURCE_RECORD_TYPE_AAAA = 28,

    /** Server selection. */
    PACE2_DNS_RESOURCE_RECORD_TYPE_SRV = 33,

    /** HTTPS. */
    PACE2_DNS_RESOURCE_RECORD_TYPE_HTTPS = 65,

} PACE2_dns_record_type_enum;

/** Enumeration for DNS class types. */
typedef enum {

    PACE2_DNS_CLASS_TYPE_ERROR = 0,

    /** Internet (IN). */
    PACE2_DNS_CLASS_TYPE_INTERNET = 1,

    /** Chaos (CH). */
    PACE2_DNS_CLASS_TYPE_CHAOS = 3,

    /** Hesiod (HS). */
    PACE2_DNS_CLASS_TYPE_HESIOD = 4,

    /** QCLASS NONE. */
    PACE2_DNS_CLASS_TYPE_QCLASS_NONE = 254,

    /** QCLASS * (ANY). */
    PACE2_DNS_CLASS_TYPE_QCLASS_ANY = 255,

} PACE2_dns_class_type_enum;

/** Enumeration for DNS response codes.
 */
typedef enum {
    PACE2_DNS_RESPONSE_CODE_NO_ERROR = 0,
    PACE2_DNS_RESPONSE_CODE_FORMAT_ERROR = 1,
    PACE2_DNS_RESPONSE_CODE_SERVER_FAILURE = 2,
    PACE2_DNS_RESPONSE_CODE_NON_EXISTENT_DOMAIN = 3,
    PACE2_DNS_RESPONSE_CODE_NOT_IMPLEMENTED = 4,
    PACE2_DNS_RESPONSE_CODE_QUERY_REFUSED = 5,
    PACE2_DNS_RESPONSE_CODE_NAME_EXISTS_WHEN_IT_SHOULD_NOT = 6,
    PACE2_DNS_RESPONSE_CODE_RR_SET_EXISTS_WHEN_IT_SHOULD_NOT = 7,
    PACE2_DNS_RESPONSE_CODE_RR_SET_THAT_SHOULD_EXISTS_DOES_NOT = 8,
    PACE2_DNS_RESPONSE_CODE_NOT_AUTHORIZED = 9,
    PACE2_DNS_RESPONSE_CODE_NAME_NOT_CONTAINED_IN_ZONE = 10,
    PACE2_DNS_RESPONSE_CODE_BAD_OPT_VERSION_OR_TSIG_SIGNATURE = 16,
    PACE2_DNS_RESPONSE_CODE_KEY_NOT_RECOGNIZED = 17,
    PACE2_DNS_RESPONSE_CODE_SIGNATURE_OUT_OF_TIME_WINDOW = 18,
    PACE2_DNS_RESPONSE_CODE_BAD_TKEY_MODE = 19,
    PACE2_DNS_RESPONSE_CODE_DUPLICATE_KEY_NAME = 20,
    PACE2_DNS_RESPONSE_CODE_ALGORITHM_NOT_SUPPORTED = 21,
    PACE2_DNS_RESPONSE_CODE_BAD_TRUNCATION = 22,
    PACE2_DNS_RESPONSE_CODE_BAD_OR_MISSING_SERVER_COOKIE = 23,
} PACE_dns_response_code;

/* Event Structures *************************************************************/

/** This structure provides information after an answer is parsed completely. */
struct pace2_dns_answer_complete {

    /** The transaction ID of the answer. */
    u16 transaction_id;
};

#ifndef PACE_KERNEL_LIB

/** Enumeration for DNS answer types. */
enum ipd_dns_record_type_enum {

    /** A host address. */
    IPD_DNS_RESOURCE_RECORD_TYPE_A = 1,

    /** An authoritative name server. */
    IPD_DNS_RESOURCE_RECORD_TYPE_NS = 2,

    /** The canonical name for an alias. */
    IPD_DNS_RESOURCE_RECORD_TYPE_CNAME = 5,

    /** Marks the start of a zone of authority. */
    IPD_DNS_RESOURCE_RECORD_TYPE_SOA = 6,

    /** A well known service description. */
    IPD_DNS_RESOURCE_RECORD_TYPE_WKS = 11,

    /** A domain name pointer. */
    IPD_DNS_RESOURCE_RECORD_TYPE_PTR = 12,

    /** Host information. */
    IPD_DNS_RESOURCE_RECORD_TYPE_HINFO = 13,

    /** Mailbox or mail list information. */
    IPD_DNS_RESOURCE_RECORD_TYPE_MINFO = 14,

    /** Mail exchange. */
    IPD_DNS_RESOURCE_RECORD_TYPE_MX = 15,

    /** Text string. */
    IPD_DNS_RESOURCE_RECORD_TYPE_TXT = 16,

    /** IPv6. */
    IPD_DNS_RESOURCE_RECORD_TYPE_AAAA = 28,

    /** Server selection. */
    IPD_DNS_RESOURCE_RECORD_TYPE_SRV = 33,

    /** HTTPS. */
    IPD_DNS_RESOURCE_RECORD_TYPE_HTTPS = 65,
};

/** Enumeration for DNS class types. */
enum ipd_dns_class_type_enum {

    /** Internet (IN). */
    IPD_DNS_CLASS_TYPE_INTERNET = 1,

    /** Chaos (CH). */
    IPD_DNS_CLASS_TYPE_CHAOS = 3,

    /** Hesiod (HS). */
    IPD_DNS_CLASS_TYPE_HESIOD = 4,

    /** QCLASS NONE. */
    IPD_DNS_CLASS_TYPE_QCLASS_NONE = 254,

    /** QCLASS * (ANY). */
    IPD_DNS_CLASS_TYPE_QCLASS_ANY = 255,
};

/** Enumeration for DNS response codes. */
enum ipd_dns_response_code {
    IPD_DNS_RESPONSE_CODE_NO_ERROR = 0,
    IPD_DNS_RESPONSE_CODE_FORMAT_ERROR = 1,
    IPD_DNS_RESPONSE_CODE_SERVER_FAILURE = 2,
    IPD_DNS_RESPONSE_CODE_NON_EXISTENT_DOMAIN = 3,
    IPD_DNS_RESPONSE_CODE_NOT_IMPLEMENTED = 4,
    IPD_DNS_RESPONSE_CODE_QUERY_REFUSED = 5,
    IPD_DNS_RESPONSE_CODE_NAME_EXISTS_WHEN_IT_SHOULD_NOT = 6,
    IPD_DNS_RESPONSE_CODE_RR_SET_EXISTS_WHEN_IT_SHOULD_NOT = 7,
    IPD_DNS_RESPONSE_CODE_RR_SET_THAT_SHOULD_EXISTS_DOES_NOT = 8,
    IPD_DNS_RESPONSE_CODE_NOT_AUTHORIZED = 9,
    IPD_DNS_RESPONSE_CODE_NAME_NOT_CONTAINED_IN_ZONE = 10,
    IPD_DNS_RESPONSE_CODE_BAD_OPT_VERSION_OR_TSIG_SIGNATURE = 16,
    IPD_DNS_RESPONSE_CODE_KEY_NOT_RECOGNIZED = 17,
    IPD_DNS_RESPONSE_CODE_SIGNATURE_OUT_OF_TIME_WINDOW = 18,
    IPD_DNS_RESPONSE_CODE_BAD_TKEY_MODE = 19,
    IPD_DNS_RESPONSE_CODE_DUPLICATE_KEY_NAME = 20,
    IPD_DNS_RESPONSE_CODE_ALGORITHM_NOT_SUPPORTED = 21,
    IPD_DNS_RESPONSE_CODE_BAD_TRUNCATION = 22,
    IPD_DNS_RESPONSE_CODE_BAD_OR_MISSING_SERVER_COOKIE = 23,
};

/** Enumeration for the DNS HTTPS SVC (Service) Parameter Keys. */
enum ipd_dns_svc_param_key {
    IPD_DNS_SVC_PARAM_KEY_ALPN = 1,
    IPD_DNS_SVC_PARAM_KEY_NO_DEFAULT_ALPN = 2,
    IPD_DNS_SVC_PARAM_KEY_PORT = 3,
    IPD_DNS_SVC_PARAM_KEY_IPV4HINT = 4,
    IPD_DNS_SVC_PARAM_KEY_ECH = 5,
    IPD_DNS_SVC_PARAM_KEY_IPV6HINT = 6,
    IPD_DNS_SVC_PARAM_KEY_DOHPATH = 7,
};

/** This structure provides information for DNS query events. */
struct ipd_dns_query {

    /** The transaction ID of the query. */
    u16 transaction_id;

    /** The type field of the query, e.g. 1 (A) for IPv4 address query. */
    enum ipd_dns_record_type_enum type;

    /** The class field of the query, always 1 (IN) for internet. */
    enum ipd_dns_class_type_enum qclass;

    /** The flag field of the dns query/response. */
    u16 flag;

    /** The reply code of the response. Only set if it is a response. */
    enum ipd_dns_response_code rcode;

    /** The queried domain name. */
    struct ipd_charset_string name;
};

/** This structure provides information for DNS IPv4 answer events. */
struct ipd_dns_answer_ipv4 {

    /** The transaction ID of the IPv4 answer. */
    u16 transaction_id;

    /** The type field of the resource record, always 1 (A). */
    enum ipd_dns_record_type_enum type;

    /** The class field of the resource record, always 1 (IN) for internet. */
    enum ipd_dns_class_type_enum rr_class;

    /** The time to live of the IPv4 answer. */
    u32 ttl;

    /** The IPv4 address of the IPv4 answer. */
    u32 ipv4_address;

    /** The domain name of the IPv4 answer. */
    struct ipd_charset_string name;
};

/** This structure provides information for DNS IPv6 answer events. */
struct ipd_dns_answer_ipv6 {

    /** The transaction ID of the IPv6 answer. */
    u16 transaction_id;

    /** The type field of the resource record, always 28 (AAAA). */
    enum ipd_dns_record_type_enum type;

    /** The class field of the resource record, always 1 (IN) for internet. */
    enum ipd_dns_class_type_enum rr_class;

    /** The time to live of the IPv6 answer. */
    u32 ttl;

    /** The IPv6 address of the IPv6 answer. */
    struct in6_addr ipv6_address;

    /** The domain name of the IPv6 answer. */
    struct ipd_charset_string name;
};

/** This structure provides information for DNS canonical name answer events. */
struct ipd_dns_answer_canonical_name {

    /** The transaction ID of the canonical name answer. */
    u16 transaction_id;

    /** The type field of the resource record, always 5 (CNAME). */
    enum ipd_dns_record_type_enum type;

    /** The class field of the resource record, always 1 (IN) for internet. */
    enum ipd_dns_class_type_enum rr_class;

    /** The time to live of the canonical name answer. */
    u32 ttl;

    /** The domain name of the canonical name answer. */
    struct ipd_charset_string name;

    /** The canonical name of the canonical name answer. */
    struct ipd_charset_string canonical_name;
};

/** This structure provides information for DNS authoritative name server answer events. */
struct ipd_dns_answer_authoritative_name_server {

    /** The transaction ID of the authoritative name server answer. */
    u16 transaction_id;

    /** The type field of the resource record, always 2 (NS). */
    enum ipd_dns_record_type_enum type;

    /** The class field of the resource record, always 1 (IN) for internet. */
    enum ipd_dns_class_type_enum rr_class;

    /** The time to live of the authoritative name server answer. */
    u32 ttl;

    /** The domain name of the authoritative name server answer. */
    struct ipd_charset_string name;

    /** The authoritative name server of the authoritative name server answer. */
    struct ipd_charset_string authoritative_name_server;
};

/** This structure provides information for DNS start of authority answer events. */
struct ipd_dns_answer_start_of_authority {

    /** The transaction ID of the start of authority answer. */
    u16 transaction_id;

    /** The type field of the resource record, always 6 (SOA). */
    enum ipd_dns_record_type_enum type;

    /** The class field of the resource record, always 1 (IN) for internet. */
    enum ipd_dns_class_type_enum rr_class;

    /** The time to live of the start of authority answer. */
    u32 ttl;

    /** The domain name of the start of authority answer. */
    struct ipd_charset_string name;

    /** The name of the primary server of the start of authority answer. */
    struct ipd_charset_string mname;

    /** The name of the party responsible for the domain of the start of authority answer. */
    struct ipd_charset_string rname;

    /** The serial number of the start of authority answer. */
    u32 serial_number;

    /** The refresh interval of the start of authority answer. */
    u32 refresh_interval;

    /** The retry interval of the start of authority answer. */
    u32 retry_interval;

    /** The expire limit of the start of authority answer. */
    u32 expire_limit;

    /** The minimum ttl of the start of authority answer. */
    u32 minimum_ttl;
};

/** This structure provides information for DNS mail exchange answer events. */
struct ipd_dns_answer_mail_exchange {

    /** The transaction ID of the mail exchange answer. */
    u16 transaction_id;

    /** The type field of the resource record, always 15 (MX). */
    enum ipd_dns_record_type_enum type;

    /** The class field of the resource record, always 1 (IN) for internet. */
    enum ipd_dns_class_type_enum rr_class;

    /** The preference of the mail exchange answer. */
    u16 preference;

    /** The time to live of the mail exchange answer. */
    u32 ttl;

    /** The domain name of the mail exchange answer. */
    struct ipd_charset_string name;

    /** The mail exchange name of the mail exchange answer. */
    struct ipd_charset_string mail_exchange_name;
};

/** This structure provides information for DNS pointer domain name answer events. */
struct ipd_dns_answer_pointer_domain_name {

    /** The transaction ID of the answer. */
    u16 transaction_id;

    /** The type field of the resource record, always 12 (PTR). */
    enum ipd_dns_record_type_enum type;

    /** The class field of the resource record, always 1 (IN) for Internet. */
    enum ipd_dns_class_type_enum rr_class;

    /** The time to live of the answer. */
    u32 ttl;

    /** The answer name. */
    struct ipd_charset_string name;

    /** The obtained pointer domain name. */
    struct ipd_charset_string pointer_domain_name;
};

/** This structure provides information for DNS text string answer events. */
struct ipd_dns_answer_text_string {

    /** The transaction ID of the answer. */
    u16 transaction_id;

    /** The type field of the resource record, always 16 (TXT). */
    enum ipd_dns_record_type_enum type;

    /** The class field of the resource record, always 1 (IN) for Internet. */
    enum ipd_dns_class_type_enum rr_class;

    /** The time to live of the answer. */
    u32 ttl;

    /** The answer name. */
    struct ipd_charset_string name;

    /** The text string. */
    struct ipd_charset_string text_string;
};

/** This structure provides information for DNS server selection answer events. */
struct ipd_dns_answer_server_selection {

    /** The transaction ID of the answer. */
    u16 transaction_id;

    /** The type field of the resource record, always 33 (SRV). */
    enum ipd_dns_record_type_enum type;

    /** The class field of the resource record, always 1 (IN) for Internet. */
    enum ipd_dns_class_type_enum rr_class;

    /** The time to live of the answer. */
    u32 ttl;

    /** The answer name. */
    struct ipd_charset_string name;

    /** The priority of this target host. */
    u16 priority;

    /** A server selection mechanism. The weight field specifies a relative weight
     * for entries with the same priority. */
    u16 weight;

    /** The port on this target host of this service. */
    u16 port;

    /** The domain name of the target host. */
    struct ipd_charset_string target;
};

#define PACE2_MAX_NUMBER_OF_DNS_SVC_COUNT 8

/** This structure provides information for DNS HTTPS answer events. */
struct ipd_dns_answer_https {

    /** The transaction ID of the HTTPS answer. */
    u16 transaction_id;

    /** The type field of the resource record, always 65 (HTTPS). */
    enum ipd_dns_record_type_enum type;

    /** The class field of the resource record, always 1 (IN) for internet. */
    enum ipd_dns_class_type_enum rr_class;

    /** The time to live of the HTTPS answer. */
    u32 ttl;

    /** The domain name of the HTTPS answer. */
    struct ipd_charset_string name;

    /** The priority of this target host. */
    u16 priority;

    /** The domain name of the target host. */
    struct ipd_charset_string target;

    /** Number of optional SVCs. */
    u8 svc_count;

    /** This structure provides information for DNS HTTPS SVC (Service) parameter. */
    struct ipd_dns_answer_https_svc_param {

        /** Number of Application-Layer Protocol Negotiations (ALPNs). */
        u8 alpn_count;

        /** Key of the SVC parameter. */
        enum ipd_dns_svc_param_key param_key;

        /** Length of the SVC parameter. */
        u16 param_len;

        /** The port on this target host of this service. */
        u16 svc_port;

        /** Array of ALPN strings. */
        struct ipd_charset_string alpn_string[PACE2_MAX_NUMBER_OF_DNS_SVC_COUNT];

        /**  IPv4 addresses of the HTTPS answer. */
        struct in_addr * ipv4_addr;

        /** Array of IPv6 addresses of the HTTPS answer. */
        struct in6_addr * ipv6_addr;

    } svc_param[PACE2_MAX_NUMBER_OF_DNS_SVC_COUNT];
};

/** This structure provides information for not implemented / other answer events. */
struct ipd_dns_answer_other {

    /** The transaction ID of a unimplemented answer. */
    u16 transaction_id;

    /** The type field of the resource record, refer to IANA's
     * <a href="http://www.iana.org/assignments/dns-parameters">DNS TYPE registry</a>
     * for a current list of resource record types. */
    enum ipd_dns_record_type_enum type;

    /** The class field of the resource record, always 1 (IN) for internet. */
    enum ipd_dns_class_type_enum rr_class;

    /** The time to live of the unimplemented answer. */
    u32 ttl;

    /** The domain name of the unimplemented answer. */
    struct ipd_charset_string name;

    /** The generic data of the unimplemented answer. */
    struct ipd_charset_string generic_data;
};

/**
 * @brief Data structure for advanced DNS events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_DNS_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** DNS event subtype. */
    PACE2_advanced_DNS_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_DNS_QUERY event. */
        struct ipd_dns_query query;

        /** Contains data for a @ref PACE2_ADVANCED_DNS_QUERY_RESPONSE event. */
        struct ipd_dns_query query_response;

        /** Contains data for a @ref PACE2_ADVANCED_DNS_ANSWER_IPV4 event. */
        struct ipd_dns_answer_ipv4 answer_ipv4;

        /** Contains data for a @ref PACE2_ADVANCED_DNS_ANSWER_IPV6 event. */
        struct ipd_dns_answer_ipv6 answer_ipv6;

        /** Contains data for a @ref PACE2_ADVANCED_DNS_ANSWER_CANONICAL_NAME event. */
        struct ipd_dns_answer_canonical_name answer_canonical_name;

        /** Contains data for a @ref PACE2_ADVANCED_DNS_ANSWER_MAIL_EXCHANGE event. */
        struct ipd_dns_answer_mail_exchange answer_mail_exchange;

        /** Contains data for a @ref PACE2_ADVANCED_DNS_ANSWER_DOMAIN_NAME_POINTER event. */
        struct ipd_dns_answer_pointer_domain_name answer_pointer_domain_name;

        /** Contains data for a @ref PACE2_ADVANCED_DNS_ANSWER_TEXT_STRING event. */
        struct ipd_dns_answer_text_string answer_text_string;

        /** Contains data for a @ref PACE2_ADVANCED_DNS_ANSWER_SERVER_SELECTION event. */
        struct ipd_dns_answer_server_selection answer_server_selection;

        /** Contains data for a @ref PACE2_ADVANCED_DNS_ANSWER_OTHER event. */
        struct ipd_dns_answer_other answer_other;

        /** Contains data for a @ref PACE2_ADVANCED_DNS_ANSWER_COMPLETE event. */
        struct pace2_dns_answer_complete answer_complete;

        /** Contains data for a @ref PACE2_ADVANCED_DNS_ANSWER_AUTHORITATIVE_NAME_SERVER event. */
        struct ipd_dns_answer_authoritative_name_server answer_authoritative_name_server;
        
        /** Contains data for a @ref PACE2_ADVANCED_DNS_ANSWER_START_OF_AUTHORITY event. */
        struct ipd_dns_answer_start_of_authority answer_start_of_authority;

        /** Contains data for a @ref PACE2_ADVANCED_DNS_ANSWER_HTTPS event. */
        struct ipd_dns_answer_https answer_https;

        /** Contains data for a @ref PACE2_ADVANCED_DNS_INVALID event. */
        char const * invalid;

    } event_data;

} PACE2_advanced_DNS_event;

#endif

#define PACE2_MAX_NUMBER_OF_DNS_ANSWER_COUNT 25

/**
 * @brief Types for DNS Answer lines.
 */
typedef enum {

    /** DNS answer type IPv4. */
    PACE2_DNS_ANSWER_TYPE_IPV4 = 0,

    /** DNS answer type IPv6. */
    PACE2_DNS_ANSWER_TYPE_IPV6,

    /** DNS answer type Canonical Name. */
    PACE2_DNS_ANSWER_TYPE_CANONICAL_NAME,

    /** Number of implemented DNS answer types. */
    PACE2_NUMBER_OF_DNS_ANSWER_TYPES

} PACE2_DNS_answer_type;

/**
 * @brief Data structure for basic DNS events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_BASIC_DNS_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** The DNS event type. See @ref PACE2_basic_DNS_event_type for possible mask values. */
    PACE2_bitmask meta_data_mask;

    /** The transaction ID of the query. */
    u16 transaction_id;

    /** The flag field of the dns query/response. */
    u16 flag;

    /** Number of query records. */
    u16 queries_count;

    /** Number of answer research records. */
    u16 answer_rr_count;

    /** Number of authority research records. */
    u16 authority_rr_count;

    /** Number of additional research records. */
    u16 additional_rr_count;

    /** The type field of the resource record, always 5 (CNAME). */
    PACE2_dns_record_type_enum rtype;

    /** The class field of the query, always 1 (IN) for internet. */
    PACE2_dns_class_type_enum class_type;

    /** The reply code of the response. Only set if it is a response. */
    PACE_dns_response_code rcode;

    /** The query name as byte buffer. */
    PACE2_byte_buffer query_name;

    struct dns_answer_line {

        /** The time to live of the answer type. */
        u32 ttl;

        /** The IPv4 address of answer type IPv4 / IPv6. Only set if type is Host Address. */
        PACE2_ip_address ip;

        /** The byte buffer to the content of the answer type. */
        PACE2_byte_buffer content;

        /** The class field of the answer type. */
        PACE2_dns_class_type_enum answer_class;

        /** The type field of the resource record. */
        PACE2_dns_record_type_enum resource_record_type;

    } answer_line[PACE2_NUMBER_OF_DNS_ANSWER_TYPES];

    struct dns_answer_multiple_lines {
        struct dns_answer_line answers[PACE2_NUMBER_OF_DNS_ANSWER_TYPES];
    } answer_multiple_lines[PACE2_MAX_NUMBER_OF_DNS_ANSWER_COUNT];

} PACE2_basic_DNS_event;

#endif /* DNS_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
