/** @file      edonkey_events.h
  * @brief     eDonkey event definitions.
  * @details   @ref doc_edonkey
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef EDONKEY_EVENTS_H
#define EDONKEY_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Group ******************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced EDONKEY event.
 */
#define PACE2_ADVANCED_EDONKEY_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/
/**
 * @brief Types for advanced edonkey events.
 */
typedef enum {

    /** Issued when an eDonkey client connects to a server. Provides a struct ipd_edonkey_login *
     * with information about the connecting client. */
    PACE2_ADVANCED_EDONKEY_LOGIN,

    /** Issued when an eDonkey client connects to another client. Provides a struct ipd_edonkey_hello *
     * with information about the connecting client. */
    PACE2_ADVANCED_EDONKEY_HELLO,

    /** Issued when the connected client answers to a connection request by another client.
     * Provides a struct ipd_edonkey_hello * with information about the client being connected. */
    PACE2_ADVANCED_EDONKEY_HELLO_ANSWER,

    /** Reports the message which is sent from eDonkey servers to their connected clients.
     * Provides a struct ipd_charset_string * with the message text. */
    PACE2_ADVANCED_EDONKEY_SERVER_MESSAGE,

    /** Reports a file search request sent from a client to an eDonkey server.
     * Provides a struct ipd_charset_string * with a text representation of the search request. */
    PACE2_ADVANCED_EDONKEY_FILE_SEARCH,

    /** Reports a single file search result replied from the eDonkey server to the client.
     * Provides a struct ipd_edonkey_file_search_result * with information about the file. */
    PACE2_ADVANCED_EDONKEY_FILE_SEARCH_RESULT,

    /** Issued when an eDonkey client requests a file from another eDonkey client.
     * Provides a struct ipd_edonkey_file_request * with information about the requested file. */
    PACE2_ADVANCED_EDONKEY_FILE_REQUEST,

    /** Issued when an eDonkey client answers a file request from another eDonkey client.
     * Provides a struct ipd_edonkey_file_request * with information about the requested file. */
    PACE2_ADVANCED_EDONKEY_FILE_REQUEST_ANSWER,

    /** Reports that an eDonkey client did not find the file requested by another eDonkey client.
     * Provides a struct ipd_edonkey_file_request * with information about the requested file. */
    PACE2_ADVANCED_EDONKEY_FILE_NOT_FOUND,

    /** Issued when a piece of file data is transferred from one eDonkey client to another.
     * Provides a struct ipd_edonkey_sending_part_piece * with the piece data. */
    PACE2_ADVANCED_EDONKEY_SENDING_PART_PIECE,

    /** Issued when decoding of an eDonkey connection fails.
     * It provides a char const * to a message describing the failure reason. */
    PACE2_ADVANCED_EDONKEY_INVALID,

    /** Issued when an eDonkey connection ends. It does not provide any data. */
    PACE2_ADVANCED_EDONKEY_END_OF_FLOW,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_EDONKEY_EVENTS

} PACE2_advanced_edonkey_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/** The default length of an eDonkey user hash in bytes. */
#define IPD_EDONKEY_USER_HASH_LEN (16)

/** Structure for events that provide information about an eDonkey login. */
struct ipd_edonkey_login {

    /** Direction (corresponds to the @c direction argument of several API functions). */
    u8 direction;

    /** Identifier of the user (hash value with length ::IPD_EDONKEY_USER_HASH_LEN bytes). */
    const u8 * user_hash;

    /** The IP address reported by the client. */
    u32 client_id;

    /** The TCP port reported by the client. */
    u16 tcp_port;
};

/** Structure for events that provide information about an eDonkey hello handshake. */
struct ipd_edonkey_hello {

    /** Direction (corresponds to the @c direction argument of several API functions). */
    u8 direction;

    /** The length of the user hash in bytes. */
    u8 user_hash_size;

    /** Identifier of the user (hash value with length ipd_edonkey_hello::user_hash_size bytes). */
    const u8 * user_hash;

    /** The IP address reported by the client. */
    u32 client_id;

    /** The TCP port reported by the client. */
    u16 tcp_port;

    /** The server IP address reported by the client. */
    u32 server_ip;

    /** The server TCP port reported by the client. */
    u16 server_tcp_port;
};

/** The length of an eDonkey file ID hash in bytes. */
#define IPD_EDONKEY_FILE_ID_LENGTH (16)

/** Structure for events that provide information about an eDonkey file request. */
struct ipd_edonkey_file_request {

    /** Direction (corresponds to the @c direction argument of several API functions). */
    u8 direction;

    /** Identifier of the file (hash value with length ::IPD_EDONKEY_FILE_ID_LENGTH bytes). */
    const u8 * file_id;

    /** The name of the file. */
    struct ipd_charset_string file_name;
};

/** Structure for events that provide information about an eDonkey file search result. */
struct ipd_edonkey_file_search_result {

    /** Identifier of the file (hash value with length ::IPD_EDONKEY_FILE_ID_LENGTH bytes). */
    const u8 * file_id;

    /** The client IP address reported by the server. */
    u32 client_id;

    /** The client port reported by the server. */
    u16 port;

    /** The name of the file. */
    struct ipd_charset_string file_name;

    /** The size of the file in bytes. */
    u64 size;

    /** The availability value reported by the server. */
    u32 availability;

    /** The source count value reported by the server. */
    u32 sources;
};

/** Structure for events that provide information about a piece of eDonkey file transfer data. */
struct ipd_edonkey_sending_part_piece {

    /** Direction (corresponds to the @c direction argument of several API functions). */
    u8 direction;

    /** Identifier of the file (hash value with length ::IPD_EDONKEY_FILE_ID_LENGTH bytes). */
    const u8 * file_id;

    /** The byte offset inside the file where the current part starts. */
    u64 part_start_pos;

    /** The byte offset inside the file where the current part ends. This value is not provided for compressed file transfers and set to 0. */
    u64 part_end_pos;

    /** The byte offset of the transferred data piece inside the current part.
     * The byte offset of the data piece inside the file can be calculated like
     * ipd_edonkey_sending_part_piece::part_start_pos + ipd_edonkey_sending_part_piece::part_piece_offset. */
    u64 part_piece_offset;

    /** The actual file data of the transferred piece. */
    struct ipd_charset_string data;
};

/**
 * @brief Data structure for advanced EDONKEY events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_EDONKEY_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** EDONKEY event subtype. */
    PACE2_advanced_edonkey_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_EDONKEY_LOGIN event. */
        struct ipd_edonkey_login login;

        /** Contains data for a @ref PACE2_ADVANCED_EDONKEY_HELLO event. */
        struct ipd_edonkey_hello hello;

        /** Contains data for a @ref PACE2_ADVANCED_EDONKEY_HELLO_ANSWER event. */
        struct ipd_edonkey_hello hello_answer;

        /** Contains data for a @ref PACE2_ADVANCED_EDONKEY_SERVER_MESSAGE event. */
        struct ipd_charset_string server_message;

        /** Contains data for a @ref PACE2_ADVANCED_EDONKEY_FILE_SEARCH event. */
        struct ipd_charset_string file_search;

        /** Contains data for a @ref PACE2_ADVANCED_EDONKEY_FILE_SEARCH_RESULT event. */
        struct ipd_edonkey_file_search_result file_search_result;

        /** Contains data for a @ref PACE2_ADVANCED_EDONKEY_FILE_REQUEST event. */
        struct ipd_edonkey_file_request file_request;

        /** Contains data for a @ref PACE2_ADVANCED_EDONKEY_FILE_REQUEST_ANSWER event. */
        struct ipd_edonkey_file_request file_request_answer;

        /** Contains data for a @ref PACE2_ADVANCED_EDONKEY_FILE_NOT_FOUND event. */
        struct ipd_edonkey_file_request file_not_found;

        /** Contains data for a @ref PACE2_ADVANCED_EDONKEY_SENDING_PART_PIECE event. */
        struct ipd_edonkey_sending_part_piece sending_part_piece;

        /** Contains data for a @ref PACE2_ADVANCED_EDONKEY_INVALID event. */
        char const * invalid;

    } event_data;

} PACE2_advanced_edonkey_event;

#endif /* PACE_KERNEL_LIB */

#endif /* EDONKEY_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
