/** @file      flute_events.h
  * @brief     FLUTE event definitions.
  * @details   @copydoc doc_flute
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef FLUTE_EVENTS_H
#define FLUTE_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Group ******************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced FLUTE event.
 */
#define PACE2_ADVANCED_FLUTE_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/
/**
 * @brief Types for advanced flute events.
 */
typedef enum {

    /**
     * New FLUTE connection is observed.<br />
     * NULL.
     */
    PACE2_ADVANCED_FLUTE_NEW_CONNECTION,

    /**
     * LCT header information.<br />
     * Provides: struct ipd_flute_lct_header * : LCT header information.
     */
    PACE2_ADVANCED_FLUTE_LCT_HEADER,

    /**
     * FDT header information.<br />
     * Provides: struct ipd_flute_fdt_header * : FDT header information.
     */
    PACE2_ADVANCED_FLUTE_FDT_HEADER,

    /**
     * CENC header information.<br />
     * Provides: struct ipd_flute_cenc_header * : CENC header information.
     */
    PACE2_ADVANCED_FLUTE_CENC_HEADER,

    /**
     * FTI header information.<br />
     * Provides: struct ipd_flute_fti_header * : FTI header information.
     */
    PACE2_ADVANCED_FLUTE_FTI_HEADER,

    /**
     * FEC Payload ID header information.<br />
     * Provides: struct ipd_flute_fec_payload_id_header * : FEC Payload ID header information.
     */
    PACE2_ADVANCED_FLUTE_FEC_PAYLOAD_ID_HEADER,

    /**
     * FLUTE transfer description.<br />
     * Provides: struct ipd_flute_transfer_description * : transfer description.
     */
    PACE2_ADVANCED_FLUTE_TRANSFER_DESCRIPTION,

    /**
     * Data carried by the packet. Data is reported as it is - without any decoding or decompression.<br />
     * Provides: struct ipd_flute_data * : raw binary data.
     */
    PACE2_ADVANCED_FLUTE_DATA,

    /**
     * Decoding of a FLUTE connection fails.<br />
     * Provides: const char * : message describing the failure reason.
     */
    PACE2_ADVANCED_FLUTE_INVALID,

    /**
     * FLUTE flow ends.<br />
     * Provides: NULL.
     */
    PACE2_ADVANCED_FLUTE_END_OF_FLOW,

    /**
     * A dummy event to signal the end of the event list.
     */
    PACE2_NUMBER_OF_ADVANCED_FLUTE_EVENTS

} PACE2_advanced_flute_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * @brief Enumeration for the payload content encodings.
 */
enum ipd_flute_payload_content_encoding_enumeration {
    IPD_FLUTE_PAYLOAD_ENCODING_ALGORITHM_UNKNOWN,                       /**< Unknown. */
    IPD_FLUTE_PAYLOAD_ENCODING_ALGORITHM_NONE,                          /**< None. */
    IPD_FLUTE_PAYLOAD_ENCODING_ALGORITHM_ZLIB,                          /**< Zlib: RFC1950. */
    IPD_FLUTE_PAYLOAD_ENCODING_ALGORITHM_DEFLATE,                       /**< Deflate: RFC1951. */
    IPD_FLUTE_PAYLOAD_ENCODING_ALGORITHM_GZIP                           /**< Gzip: RFC1952. */
};

/**
 * @brief Enumeration for the FEC schemes.
 */
enum ipd_flute_fec_scheme_enumeration {
    IPD_FLUTE_FEC_SCHEME_UNKNOWN,
    IPD_FLUTE_FEC_SCHEME_COMPACT_NO_CODE,                               /**< Compact No-Code FEC Scheme. */
    IPD_FLUTE_FEC_SCHEME_SMALL_BLOCK_LARGE_BLOCK_EXPANDABLE,            /**< Small Block, Large Block, or Expandable FEC Scheme. */
    IPD_FLUTE_FEC_SCHEME_SMALL_BLOCK_SYSTEMATIC,                        /**< Small Block Systematic FEC Scheme. */
    IPD_FLUTE_FEC_SCHEME_COMPACT,                                       /**< Compact FEC Scheme. */
    IPD_FLUTE_FEC_SCHEME_RAPTOR                                         /**< Raptor FEC Scheme. */
};

/**
 * @brief Enumeration for the data types returned by the DATA event.
 */
enum ipd_flute_data_type_enumeration {
    IPD_FLUTE_DATA_TYPE_CENC_ENCODED_OR_CENC_COMPRESSED_FDT_INSTANCE,   /**< The data include an FDT instance, which is compressed or encoded by a method described in the CENC header. */
    IPD_FLUTE_DATA_TYPE_FEC_PAYLOAD_ID_AND_FEC_ENCODED_FDT_INSTANCE,    /**< FEC Encoding Scheme unknown, so we cannot strip headers. The thrown data include the FEC Payload ID header + the FEC encoded FDT instance. */
    IPD_FLUTE_DATA_TYPE_FEC_ENCODED_FDT_INSTANCE,                       /**< The data include only the FEC encoded FDT instance. */
    IPD_FLUTE_DATA_TYPE_FEC_PAYLOAD_ID_AND_FEC_ENCODED_PAYLOAD,         /**< FEC Encoding Scheme unknown, so we cannot strip headers. The thrown data include the FEC Payload ID header + the FEC encoded payload. */
    IPD_FLUTE_DATA_TYPE_FEC_ENCODED_PAYLOAD                             /**< The data include only the FEC encoded payload. */
};

/**
 * @brief Transport Session Identifier.
 */
struct ipd_tsi_bin {
    u8 buffer[6];   /**< Buffer for storing the Transport Session Identifier (Big Endian format - Network Byte Order). */
    u8 length;      /**< Length of the Transport Session Identifier - denotes the number of used bytes in the buffer. */
};

/**
 * @brief Transport Object Identifier.
 */
struct ipd_toi_bin {
    u8 buffer[14];  /**< Buffer for storing the Transport Object Identifier (Big Endian format - Network Byte Order). */
    u8 length;      /**< Length of the Transport Object Identifier - denotes the number of used bytes in the buffer. */
};

/**
 * @brief LCT header.
 */
struct ipd_flute_lct_header {
    u8 version_number;                                  /**< LCT protocol version. */
    u8 close_session_flag;                              /**< Set to 1indicates to a receiver that the sender will immediately stop sending packets for the session. */
    u8 close_object_flag;                               /**< Set to 1indicates to a receiver that the sender will immediately stop sending packets for the object. */
    u8 fec_encoding_id;                                 /**< Determines the FEC Encoding Scheme that must be used to decode the packet. */
    struct ipd_tsi_bin transport_session_identifier;    /**< Transport Session Identifier. */
    struct ipd_toi_bin transport_object_identifier;     /**< Transport Object Identifier. */
};

/**
 * @brief FDT header.
 */
struct ipd_flute_fdt_header {
    u8 version_number;                      /**< FLUTE protocol version. */
    u32 fdt_instance_id;                    /**< Identifier of the FDT instance. */
};

/**
 * @brief CENC header.
 */
struct ipd_flute_cenc_header {
    enum ipd_flute_payload_content_encoding_enumeration payload_content_encoding;   /**< Payload content encoding. */
};

/**
 * @brief FTI header.
 */
struct ipd_flute_fti_header {
    u64 transfer_length;                    /**< Transfer length. */
    u16 fec_instance_id;                    /**< FEC instance id. */
    u16 encoding_symbol_length;             /**< Encoding symbol length. */
    u32 maximum_source_block_length;        /**< Maximum source block length. */
    u16 max_number_of_encoding_symbols;     /**< Maximum number of encoding symbols. */
};

/**
 * @brief FEC Payload ID header.
 */
struct ipd_flute_fec_payload_id_header {
    enum ipd_flute_fec_scheme_enumeration fec_scheme;   /**< FEC scheme. */
    u32 source_block_number;                            /**< Source block number. */
    u32 encoding_symbol_id;                             /**< Encoding symbol id. */
    u16 source_block_length;                            /**< Source block length. */
};

/**
 * @brief FLUTE transfer description.
 */
struct ipd_flute_transfer_description {
    u64 content_length;                                     /**< Content length. */
    struct ipd_charset_string content_location;             /**< Content location. */
    struct ipd_charset_string transport_object_identifier;  /**< Transport Object Identifier. */
    u64 transfer_length;                                    /**< Transfer length. */
    struct ipd_charset_string content_type;                 /**< Content type. */
    struct ipd_charset_string content_encoding;             /**< Content encoding. */
    struct ipd_charset_string content_md5;                  /**< Content MD5. */
    u8 fec_encoding_id;                                     /**< FEC encoding id. */
    u64 fec_instance_id;                                    /**< FEC instance id. */
    u64 maximum_source_block_length;                        /**< Maximum source block length. */
    u64 encoding_symbol_length;                             /**< Encoding symbol length. */
    u64 max_number_of_encoding_symbols;                     /**< Maximum number of encoding symbols. */
    struct ipd_charset_string scheme_specific_info;         /**< Scheme-specific info. */
};

/**
 * @brief FLUTE data.
 */
struct ipd_flute_data {
    enum ipd_flute_data_type_enumeration type;              /**< Type of the thrown data. */
    struct ipd_charset_string data;                         /**< Data. */
};

/**
 * @brief Data structure for advanced FLUTE events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_ADVANCED_FLUTE_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_advanced_flute_event_type meta_data_type;                                                     /**< FLUTE event subtype. */
    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;
    union {
        struct ipd_flute_lct_header lct_header;                                                         /**< Contains data for a @ref PACE2_ADVANCED_FLUTE_LCT_HEADER event. */
        struct ipd_flute_fdt_header fdt_header;                                                         /**< Contains data for a @ref PACE2_ADVANCED_FLUTE_FDT_HEADER event. */
        struct ipd_flute_cenc_header cenc_header;                                                       /**< Contains data for a @ref PACE2_ADVANCED_FLUTE_CENC_HEADER event. */
        struct ipd_flute_fti_header fti_header;                                                         /**< Contains data for a @ref PACE2_ADVANCED_FLUTE_FTI_HEADER event. */
        struct ipd_flute_fec_payload_id_header fec_payload_id_header;                                   /**< Contains data for a @ref PACE2_ADVANCED_FLUTE_FEC_PAYLOAD_ID_HEADER event. */
        struct ipd_flute_transfer_description transfer_description;                                     /**< Contains data for a @ref PACE2_ADVANCED_FLUTE_TRANSFER_DESCRIPTION event. */
        struct ipd_flute_data data;                                                                     /**< Contains data for a @ref PACE2_ADVANCED_FLUTE_DATA event. */
        const char *invalid;                                                                            /**< Contains data for a @ref PACE2_ADVANCED_FLUTE_INVALID event. */
    } event_data;                                                                                       /**< Named union which contains the data of the sub-event types. */
} PACE2_advanced_flute_event;


#endif

#endif /* FLUTE_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
