/** @file      ftp_events.h
  * @brief     FTP event definitions.
  * @details   @copydoc doc_ftp
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef FTP_EVENTS_H
#define FTP_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "generic_events.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced FTP event.
 */
#define PACE2_ADVANCED_FTP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned metadata group of the ftp class event.
 */
#define PACE2_CLASS_FTP_EVENT_GROUP PACE2_CLASS_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced FTP events.
 */
typedef enum {

    /** Signals the start of a new FTP control connection; @c data is a NULL pointer. */
    PACE2_ADVANCED_FTP_NEW_CONTROL_CONNECTION,

    /** Signals the start of a new transfer in a FTP data connection; @c data points to
     * a struct ipd_ftp_data_flow_info. */
    PACE2_ADVANCED_FTP_NEW_DATA_CONNECTION,

    /** This event indicates that the server has signaled successful completion of the transfer (226 reply);
     * @c data points to the u64 transfer ID from the #PACE2_ADVANCED_FTP_TRANSFER_INIT event. */
    PACE2_ADVANCED_FTP_TRANSFER_END,

    /** This event indicates that the server has signaled that the transfer was aborted (4xx or 5xx reply);
     * * @c data points to the u64 transfer ID from the #PACE2_ADVANCED_FTP_TRANSFER_INIT event. */
    PACE2_ADVANCED_FTP_TRANSFER_CANCELED,

    /** Provides a struct ipd_ftp_data * with the session flow id, offset and data for a transfer.
     * Can be the result of list command or a file transfer. */
    PACE2_ADVANCED_FTP_DATA,

    /** Indicates the end of an FTP control flow; @c data is a NULL pointer. */
    PACE2_ADVANCED_FTP_END_OF_CONTROL_FLOW,

    /** Indicates the end of an FTP data flow; @c data points to a struct ipd_ftp_data_flow_info. */
    PACE2_ADVANCED_FTP_END_OF_DATA_FLOW,

    /** This event is issued when decoding of an FTP connection fails. It provides a char const * to
     * a message describing the failure reason. */
    PACE2_ADVANCED_FTP_INVALID,

    /** Provides a struct ipd_ftp_command * with the information about a send command in FTP. */
    PACE2_ADVANCED_FTP_COMMAND,

    /** Provides a struct ipd_ftp_reply * with the information about a send reply in FTP. */
    PACE2_ADVANCED_FTP_REPLY,

    /** Provides a struct ipd_ftp_transfer_info * with the information about a transfer what is initialized.
     * Will be send after a transfer command like STOR, RETR or LIST. */
    PACE2_ADVANCED_FTP_TRANSFER_INIT,

    /** Context was lost while decoding interrupted flows. */
    PACE2_ADVANCED_FTP_CONTEXT_LOST,

    /** Information about data transfer of any sort is announced. */
    PACE2_ADVANCED_FTP_DATA_ANNOUNCEMENT,

    /** A gap in the transferred data due to lost TCP segments. */
    PACE2_ADVANCED_FTP_DATA_MISSING,

    /** Signals the start of a new connection intended for FTP data transfer. */
    PACE2_ADVANCED_FTP_START_NEW_DATA_CONNECTION,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_FTP_EVENT_TYPES

} PACE2_advanced_FTP_event_type;

/**
 * @brief Types for FTP class events.
 */
typedef enum {

    /** Provides a struct ipd_class_ftp_session * with all information about a ftp session,
     * like user_name, password, state, generic. */
    PACE2_CLASS_FTP_SESSION_UPDATE,

    /** Provides a struct ipd_class_ftp_transfer * with all information about a ftp transfer,
     * like type, state, generic and filename. */
    PACE2_CLASS_FTP_TRANSFER_UPDATE,

    /** Provides a struct ipd_class_ftp_transfer_data * with a dump for transferred data. */
    PACE2_CLASS_FTP_TRANSFER_DATA,

    /** FTP command. */
    PACE2_CLASS_FTP_COMMAND,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_CLASS_FTP_EVENT_TYPES

} PACE2_class_FTP_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * Enumeration for the type of a transfer in FTP.
 */
enum ipd_ftp_transfer_enumeration {

    /** Transfer is an upload (STOR, STOU, APPE commands). */
    IPD_TRANSFER_FTP_UPLOAD,

    /** Transfer is a download (RETR command). */
    IPD_TRANSFER_FTP_DOWNLOAD,

    /** Transfer contains a directory listing (LIST, NLST, MLSD commands). */
    IPD_TRANSFER_FTP_LIST,
};

/**
 * Structure for start of down- and uploads in a ftp session or for the result of a list command.
 */
struct ipd_ftp_transfer_info {

    u64 transfer_id;

    /** The type of a transfer. @see ipd_ftp_transfer_enumeration. */
    enum ipd_ftp_transfer_enumeration transfer_type;

    /** Name of the transferred file (only set for types #IPD_TRANSFER_FTP_UPLOAD and #IPD_TRANSFER_FTP_DOWNLOAD). */
    struct ipd_charset_string server_file_name;
};

/**
 * Structure that associates an FTP data connection with the control flow transfer that caused it to be opened.
 */
struct ipd_ftp_data_flow_info {

    /** Flow ID of the control flow. */
    u64 session_flow_id;

    /** The transfer ID from the corresponding FTP_TRANSFER_INIT event on the control connection. */
    u64 transfer_id;
};

/**
 * Structure for data which are send in a transfer.
 * @see ipd_ftp_transfer_info.
 */
struct ipd_ftp_data {

    /** The connection ID of the correlating session. */
    u64 session_flow_id;

    /** The transfer ID from the corresponding FTP_TRANSFER_INIT event on the control connection. */
    u64 transfer_id;

    /** The offset of this data block in the transmitted file (does not start at 0 if the REST command was used). */
    u64 offset;

    /** The data's of this data block. */
    struct ipd_charset_string data;
};

/**
 * Structure for reporting missing data (due to lost TCP segments) in a data connection.
 * @see ipd_ftp_data.
 */
struct ipd_ftp_data_missing {

    /** Flow ID of the FTP control flow where this data connection was announced. */
    u64 session_flow_id;

    /** The transfer ID from the corresponding FTP_TRANSFER_INIT event on the control connection. */
    u64 transfer_id;

    /** The offset of this data block in the transmitted file (does not start at 0 if the REST command was used). */
    u64 offset;

    /** The length of the missing segment. */
    u64 length;
};

/**
 * Structure for information about a ftp data announcement.
 */
struct ipd_ftp_data_announcement_info {

    /** Port for the data connection, in network byte order. */
    u16 port;

    /** Address for the next data connection. */
    union ip_v4_v6_addr ip_address;

    /** Flag whether ip_address describes an ipv6 or ipv4 address. */
    u8 is_ip_v6:1;

    /** Flag whether active ftp mode is used or not
     * active mode:     server connects to client on the given ip_address and port.
     * passive mode:    client connects to server on the given ip_address and port. */
    u8 active_ftp:1;

    /** The ID of the transfer that will happen over this data connection.
     * If the client commands are available, there will be also a corresponding
     * FTP_TRANSFER_INIT event with the same ID. */
    u64 transfer_id;
};

/**
 * Enumeration for the type of a command.
 */
enum ipd_ftp_command_enum {

    /** unknown command */
    IPD_FTP_COMMAND_UNKNOWN,

    /** CDUP command */
    IPD_FTP_COMMAND_CDUP,

    /** RNFR command */
    IPD_FTP_COMMAND_RNFR,

    /** RNTO command */
    IPD_FTP_COMMAND_RNTO,

    /** USER command */
    IPD_FTP_COMMAND_USER,

    /** PASS command */
    IPD_FTP_COMMAND_PASS,

    /** ACCT command */
    IPD_FTP_COMMAND_ACCT,

    /** CWD command */
    IPD_FTP_COMMAND_CWD,

    /** QUIT command */
    IPD_FTP_COMMAND_QUIT,

    /** PORT command */
    IPD_FTP_COMMAND_PORT,

    /** PASV command */
    IPD_FTP_COMMAND_PASV,

    /** RETR command */
    IPD_FTP_COMMAND_RETR,

    /** STOR command */
    IPD_FTP_COMMAND_STOR,

    /** STOU command */
    IPD_FTP_COMMAND_STOU,

    /** APPE command */
    IPD_FTP_COMMAND_APPE,

    /** REST command */
    IPD_FTP_COMMAND_REST,

    /** ABOR command */
    IPD_FTP_COMMAND_ABOR,

    /** DELE command */
    IPD_FTP_COMMAND_DELE,

    /** RMD command */
    IPD_FTP_COMMAND_RMD,

    /** MKD command */
    IPD_FTP_COMMAND_MKD,

    /** PWD command */
    IPD_FTP_COMMAND_PWD,

    /** LIST command */
    IPD_FTP_COMMAND_LIST,

    /** NLST command */
    IPD_FTP_COMMAND_NLST,

    /** MLSD command */
    IPD_FTP_COMMAND_MLSD,

    /** MODE command */
    IPD_FTP_COMMAND_MODE,

    /** NOOP command */
    IPD_FTP_COMMAND_NOOP,

    /** EPRT command */
    IPD_FTP_COMMAND_EPRT,

    /** EPSV command */
    IPD_FTP_COMMAND_EPSV,

    /** FEAT command */
    IPD_FTP_COMMAND_FEAT,

    /** MFMT command */
    IPD_FTP_COMMAND_MFMT,

    /** MLST command */
    IPD_FTP_COMMAND_MLST,

    /** OPTS command */
    IPD_FTP_COMMAND_OPTS,

    /** SITE command */
    IPD_FTP_COMMAND_SITE,

    /** SIZE command */
    IPD_FTP_COMMAND_SIZE,

    /** SYST command */
    IPD_FTP_COMMAND_SYST,

    /** TYPE command */
    IPD_FTP_COMMAND_TYPE,

    /** ADAT command */
    IPD_FTP_COMMAND_ADAT,

    /** ALLO command */
    IPD_FTP_COMMAND_ALLO,

    /** AUTH command */
    IPD_FTP_COMMAND_AUTH,

    /** CCC command */
    IPD_FTP_COMMAND_CCC,

    /** CONF command */
    IPD_FTP_COMMAND_CONF,

    /** ENC command */
    IPD_FTP_COMMAND_ENC,

    /** HELP command */
    IPD_FTP_COMMAND_HELP,

    /** HOST command */
    IPD_FTP_COMMAND_HOST,

    /** LANG command */
    IPD_FTP_COMMAND_LANG,

    /** LPRT command */
    IPD_FTP_COMMAND_LPRT,

    /** LPSV command */
    IPD_FTP_COMMAND_LPSV,

    /** MDTM command */
    IPD_FTP_COMMAND_MDTM,

    /** MFCT command */
    IPD_FTP_COMMAND_MFCT,

    /** MFF command */
    IPD_FTP_COMMAND_MFF,

    /** MIC command */
    IPD_FTP_COMMAND_MIC,

    /** PBSZ command */
    IPD_FTP_COMMAND_PBSZ,

    /** PROT command */
    IPD_FTP_COMMAND_PROT,

    /** REIN command */
    IPD_FTP_COMMAND_REIN,

    /** SMNT command */
    IPD_FTP_COMMAND_SMNT,

    /** SPSV command */
    IPD_FTP_COMMAND_SPSV,

    /** STAT command */
    IPD_FTP_COMMAND_STAT,

    /** STRU command */
    IPD_FTP_COMMAND_STRU,

    /** XCUP command */
    IPD_FTP_COMMAND_XCUP,

    /** XMKD command */
    IPD_FTP_COMMAND_XMKD,

    /** XPWD command */
    IPD_FTP_COMMAND_XPWD,

    /** XRCP command */
    IPD_FTP_COMMAND_XRCP,

    /** XRMD command */
    IPD_FTP_COMMAND_XRMD,

    /** XRSQ command */
    IPD_FTP_COMMAND_XRSQ,

    /** XSEM command */
    IPD_FTP_COMMAND_XSEM,

    /** XSEN command */
    IPD_FTP_COMMAND_XSEN,
};

/**
 * Structure for a reply in FTP.
 */
struct ipd_ftp_reply {

    /**
     * The code of the reply. @ref ipd_ftp_reply::code "code"/100 signals
     * the @link ipd_ftp_reply::code category@endlink of a reply.
     *  - category 1: command successfully done, but one more reply will come later.
     *  - category 2: command successfully done.
     *  - category 3: command successfully done, but one more command is needed.
     *  - category 4: command failed (temporary), maybe a second call will be successfully.
     *  - category 5: command failed (error).
     */
    u16 code;

    /** The message which is send with the reply. */
    struct ipd_charset_string message;

    /** Is true (1) if the reply is a multiline reply and signals, that one more line will follow.
     * The last line is set to false (0). */
    u8 continued:1;
};

/**
 * Structure for a command in FTP.
 */
struct ipd_ftp_command {

    /** The sent command code as enumeration code. @see ipd_ftp_command_enum. */
    enum ipd_ftp_command_enum code;

    /** The command name as string. */
    struct ipd_charset_string name;

    /** The given argument for this command as string.*/
    struct ipd_charset_string argument;

    /** Indicates the direction of the client, sending the command. This is either mapped on to the endpoints given to
     * the packet-descriptor, if @ref pace2_s4_process_packet is used or the index of the stream in the stream
     * descriptor if @ref pace2_s4_process_stream is used for decoding. */
    u8 client_direction:1;
};

typedef struct ipd_ftp_command ipd_ftp_command_int_t;

/** Enumeration for FTP session state. */
enum ipd_class_ftp_session_state {

    /** The control connection is open, but the user is not yet logged in. */
    IPD_CLASS_FTP_SESSION_INIT,

    /** The user has logged in. */
    IPD_CLASS_FTP_SESSION_LOGGEDIN,

    /** The user has logged out, or the control connection has been closed. */
    IPD_CLASS_FTP_SESSION_CLOSE,
};

/** State of an FTP transfer, as observed on the control connection. */
enum ipd_class_ftp_transfer_control_state {

    /** Client has issued a command that opens a data connection (RETR, STOR, LIST etc). */
    IPD_CLASS_FTP_TRANSFER_REQUESTED,

    /** Server has responded with a positive completion reply (2xx). */
    IPD_CLASS_FTP_TRANSFER_COMPLETED,

    /** Server has responded with a negative completion reply (4xx/5xx). */
    IPD_CLASS_FTP_TRANSFER_ABORTED,
};

/** State of an FTP data connection. */
enum ipd_class_ftp_transfer_data_state {

    /** There is no data connection. */
    IPD_CLASS_FTP_DATA_NONE,

    /** The data connection is open. */
    IPD_CLASS_FTP_DATA_OPEN,

    /** The data connection has been closed. */
    IPD_CLASS_FTP_DATA_CLOSED,
};

/** Information about one FTP transfer. */
struct ipd_class_ftp_transfer {

    /** The ID number of this transfer. */
    struct ipd_class_generic_header generic;

    /** The flow ID of the data connection. */
    u64 connection_id;

    /** The session ID (which is also the flow ID of the control connection). */
    u64 session_id;

    /** The state of the transfer command on the control connection. */
    enum ipd_class_ftp_transfer_control_state control_state;

    /** The state of the data connection. */
    enum ipd_class_ftp_transfer_data_state data_state;

    /** The type of the transfer. */
    enum ipd_ftp_transfer_enumeration type;

    /** The filename on server side. */
    struct ipd_charset_string server_file_name;

    /** The working directory for this transfer. */
    struct ipd_charset_string working_directory;
};

/** Structure for a FTP data transfer. */
struct ipd_class_ftp_transfer_data {

    /** The generic. Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The ID of the correlating session. */
    u64 session_id;

    /** The ID of the correlating transfer. */
    u64 transfer_id;

    /** The data of this data block. */
    struct ipd_charset_string data;

    /** The offset for this data block to the complete data transfer. */
    u64 data_offset;
};

/** Structure for a FTP session */
struct ipd_class_ftp_session {

    /** The generic. Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The connection ID of the used connection. */
    u64 connection_id;

    /** The login name of the user for this ftp session. */
    struct ipd_charset_string user_name;

    /** The password of the user for this ftp session. */
    struct ipd_charset_string password;

    /** The state of the session. @see ipd_class_ftp_session_state. */
    enum ipd_class_ftp_session_state state;

    /** The working directory where the session is in. */
    struct ipd_charset_string working_directory;
};

/**
 * @brief Structure for the FTP command.
 */
struct pace2_class_ftp_command {

    /** Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** Session id. */
    u64 session_id;

    /** Command as enumeration code. */
    enum ipd_ftp_command_enum code;

    /** Command name as a string. */
    struct ipd_charset_string name;

    /** Arguments for this command as a string. */
    struct ipd_charset_string arguments;
};

/**
 * @brief Data structure for advanced FTP events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_FTP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** FTP event subtype. */
    PACE2_advanced_FTP_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_FTP_NEW_DATA_CONNECTION event. */
        struct ipd_ftp_data_flow_info new_data_connection;

        /** Contains data for a @ref PACE2_ADVANCED_FTP_START_NEW_DATA_CONNECTION event. */
        u64 start_new_data_connection;

        /** Contains data for a @ref PACE2_ADVANCED_FTP_TRANSFER_END event. */
        u64 transfer_end;

        /** Contains data for a @ref PACE2_ADVANCED_FTP_TRANSFER_CANCELED event. */
        u64 transfer_cancelled;

        /** Contains data for a @ref PACE2_ADVANCED_FTP_DATA event. */
        struct ipd_ftp_data data;

        /** Contains data for a @ref PACE2_ADVANCED_FTP_END_OF_DATA_FLOW event. */
        struct ipd_ftp_data_flow_info end_of_data_flow;

        /** Contains data for a @ref PACE2_ADVANCED_FTP_INVALID event. */
        char const * invalid;

        /** Contains data for a @ref PACE2_ADVANCED_FTP_COMMAND event. */
        struct ipd_ftp_command command;

        /** Contains data for a @ref PACE2_ADVANCED_FTP_REPLY event. */
        struct ipd_ftp_reply reply;

        /** Contains data for a @ref PACE2_ADVANCED_FTP_TRANSFER_INIT event. */
        struct ipd_ftp_transfer_info transfer_init;

        /** Contains data for a @ref PACE2_ADVANCED_FTP_DATA_ANNOUNCEMENT event. */
        struct ipd_ftp_data_announcement_info data_announcement;

        /** Contains data for a @ref PACE2_ADVANCED_FTP_DATA_MISSING event. */
        struct ipd_ftp_data_missing data_missing;

    } event_data;

} PACE2_advanced_FTP_event;

/**
 * @brief Data structure for FTP class events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_CLASS_FTP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** FTP event subtype. */
    PACE2_class_FTP_event_type meta_data_type;

    PACE2_CLASS_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_CLASS_FTP_SESSION_UPDATE event. */
        struct ipd_class_ftp_session session;

        /** Contains data for a @ref PACE2_CLASS_FTP_TRANSFER_UPDATE event. */
        struct ipd_class_ftp_transfer transfer;

        /** Contains data for a @ref PACE2_CLASS_FTP_TRANSFER_DATA event. */
        struct ipd_class_ftp_transfer_data transfer_data;

        /** Contains data for a @ref PACE2_CLASS_FTP_COMMAND event. */
        struct pace2_class_ftp_command command;

    } event_data;

} PACE2_class_FTP_event;

#endif

#endif /* FTP_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
