/** @file      generic_events.h
  * @brief     Generic PACE 2 event definitions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef GENERIC_EVENTS_H
#define GENERIC_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "../pace2_pht_core.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the flow started event.
 */
#define PACE2_FLOW_STARTED_EVENT_GROUP PACE2_START_GROUP

/**
 * @brief Public define for the assigned metadata group of the flow dropped event.
 */
#define PACE2_FLOW_DROPPED_EVENT_GROUP PACE2_DROP_GROUP

/**
 * @brief Public define for the assigned metadata group of the subscriber started event.
 */
#define PACE2_SUBSCRIBER_STARTED_EVENT_GROUP PACE2_START_GROUP

/**
 * @brief Public define for the assigned metadata group of the subscriber dropped event.
 */
#define PACE2_SUBSCRIBER_DROPPED_EVENT_GROUP PACE2_DROP_GROUP

/**
 * @brief Public define for the assigned metadata group of the flow info event.
 */
#define PACE2_FLOW_INFO_EVENT_GROUP PACE2_START_GROUP

/**
 * @brief Public define for the assigned metadata group of the flow process event.
 */
#define PACE2_FLOW_PROCESS_EVENT_GROUP PACE2_NO_GROUP

/**
 * @brief Public define for the assigned metadata group of the subscriber process event.
 */
#define PACE2_SUBSCRIBER_PROCESS_EVENT_GROUP PACE2_NO_GROUP

/**
 * @brief Public define for the assigned metadata group of the dropped events event.
 */
#define PACE2_EVENTS_DROPPED_EVENT_GROUP PACE2_DROP_GROUP

/**
 * @brief Public define for the assigned metadata group of the pht threshold notification event.
 */
#define PACE2_PHT_USAGE_NOTIFICATION_EVENT_GROUP PACE2_NO_GROUP

/* Event Types ******************************************************************/

/**
 * @brief The internal pht reported by @ref PACE2_pht_usage_notification_event
 */
typedef enum pace2_pht_notification_source {
    PACE2_PHT_NOTIFICATION_SOURCE_FLOW_TRACKING = 0,
    PACE2_PHT_NOTIFICATION_SOURCE_SUBSCRIBER_TRACKING
} PACE2_pht_notification_source;

/**
 * @brief Type of the pht usage notification used by @ref PACE2_pht_usage_notification_event
 */
typedef enum pace2_pht_notification_type {
    PACE2_PHT_NOTIFICATION_TYPE_BELOW = 0,
    PACE2_PHT_NOTIFICATION_TYPE_REACHED
} PACE2_pht_threshold_type;

/* Event Structures *************************************************************/

/**
 * @brief Placeholder event structure.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_NO_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
} PACE2_no_event;

/**
 * @brief Data structure of flow start events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_FLOW_STARTED_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    u64 flow_id;                /**< flow ID (assigned by PACE2) of this flow */
} PACE2_flow_started;

/**
 * @brief Data structure of flow drop events. One reason for the drop can be a timeout.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_FLOW_DROPPED_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    u64 flow_id;                                        /**< flow ID (assigned by PACE2) of this flow */
    enum pace2_pht_removal_reason removal_reason;       /**< Reason for the drop of this flow. Values can be @ref PACE2_pht_removal_reason  */
    void *flow_ptr;                               /**< pointer to the flow, can be used to access CLI functions requiring this */
    const void *int_flow_key;                           /**< contains a pointer to the flow key. only available for internal tracking */
    PACE2_timestamp last_ts;                            /**< contains the last updated timestamp for internal tracking */
} PACE2_flow_dropped;

/**
 * @brief Data structure of flow info events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_FLOW_INFO_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    u64 flow_id;                /**< flow ID (assigned by PACE2) of this flow */
    u64 src_id;                 /**< subscriber id of the source of the first packet in this flow.
                                     The ID will also be set for external tracking, if a subscriber is provided. */
    u64 dst_id;                 /**< subscriber id of the destination of the first packet in this flow.
                                     The ID will also be set for external tracking, if a subscriber is provided. */
    u16 src_port;               /**< source port of the source of the first packet in this flow, in network byte order.*/
    u16 dst_port;               /**< destination port of the destination of the first packet in this flow, in network byte order. */
    u8 initial_packet_direction;/**< direction flag of the first packet */
    u8 community_id[31];        /**< Community ID of this flow */
} PACE2_flow_info;

/**
 * @brief Data structure of flow process events.
 * The event is only generated when using internal tracking.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_FLOW_PROCESS_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    u64 flow_id;                    /**< flow ID (assigned by PACE2) of this flow */
    u64 bytes;                      /**< accumulated number of transferred payload data (Layer 7) in bytes */
    u64 total_bytes;                /**< accumulated number of overall transferred data in bytes */
    u64 missing_bytes;              /**< accumulated number of missing bytes in TCP flows */
    PACE2_timestamp start_ts;       /**< timestamp of the first packet in this flow */
    PACE2_timestamp last_packet_ts; /**< timestamp of the last seen packet before flow drop */
} PACE2_flow_process;

/**
 * @brief Data structure of subscriber start events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_SUBSCRIBER_STARTED_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    u8 track_dst;                           /**< Tracking direction 0 or 1. */
    void *subscriber_user_data;             /**< User data of the @ref PACE2_SUBSCRIBER_STARTED_EVENT event. */
    u64 subscriber_id;                      /**< ID of the subscriber. */
    PACE2_ip_address ip;                    /**< The IP address of the subscriber. */
    const void *customer_subscriber_key;    /**< Contains a pointer to the customer subscriber key. */
} PACE2_subscriber_started;

/**
 * @brief Data structure of subscriber drop events. One reason for the drop can be a timeout.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_SUBSCRIBER_DROPPED_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    void *subscriber_user_data;                         /**< User data of the @ref PACE2_SUBSCRIBER_DROPPED_EVENT event. */
    enum pace2_pht_removal_reason removal_reason;       /**< Reason for the drop of this Subscriber entry. Values can be @ref PACE2_pht_removal_reason  */
    u64 subscriber_id;                                  /**< ID of the subscriber. */
    const void *subscriber_ptr;                         /**< Pointer to the subscriber, can be used to access CLI functions requiring this */
    const void *int_subscriber_key;                     /**< Contains a pointer to the subscriber key. only available for internal tracking */
    PACE2_timestamp last_ts;                            /**< Contains the last updated timestamp for internal tracking */
} PACE2_subscriber_dropped;

/**
 * @brief Data structure of subscriber process events.
 * The event is only generated when using internal tracking.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_SUBSCRIBER_PROCESS_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    u64 subscriber_id;          /**< subscriber ID (assigned by PACE2) of this subscriber. */
    u64 bytes_up;               /**< accumulated number of uploaded data in bytes */
    u64 bytes_down;             /**< accumulated number of downloaded data in bytes */
    u64 total_bytes;            /**< accumulated number of transferred data in bytes */
} PACE2_subscriber_process;

/**
 * @brief Data structure for dropped events report.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_EVENTS_DROPPED_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    u32 dropped_events;         /**< amount of dropped events during last event processing loop. */
} PACE2_events_dropped;

/**
 * @brief Data structure for hashtable threshold events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_PHT_USAGE_NOTIFICATION_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    enum pace2_pht_notification_source notification_source; /**< Hash table that reached or went below the threshold
                                                                 (flow hash table or subscriber hash table) */
    enum pace2_pht_notification_type notification_type;     /**< Indicates if number of entries went below or above
                                                                 the threshold */
    u32 used_elements;                                      /**< Number of entries in the hash table */
    u32 thread_id;                                          /**< ID of the thread */
} PACE2_pht_usage_notification_event;

/** Enumeration for possible class event types. */
enum ipd_class_event {
    IPD_CLASS_MAIL,             /**< mail class event, see ipd_class_mail_definitions.h */
    IPD_CLASS_HTTP,             /**< http class event, see ipd_class_http_definitions.h */
    IPD_CLASS_CHAT,             /**< chat class event, see ipd_class_chat_definitions.h */
    IPD_CLASS_FTP,              /**< ftp class event, see ipd_class_ftp_definitions.h */
    IPD_CLASS_VOIP,             /**< VoIP class event, see ipd_class_voip_definitions.h  */
    IPD_CLASS_GENERIC,          /**< Generic class event, see class_generic_events.h */
    IPD_CLASS_SSL,              /**< SSL class event, see ssl_events.h */
};

/**
 * Structure for a generic information for many class events.
 * It's guaranteed that the ID of the event is unique.
 */
struct ipd_class_generic_header {
    /** The unique ID of the event. */
    u64 id;
};

#endif /* GENERIC_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
