/** @file      gtp_events.h
  * @brief     GTP event definitions.
  * @details   @copydoc doc_gtp
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef GTP_EVENTS_H
#define GTP_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced GTP event.
 */
#define PACE2_ADVANCED_GTP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced GTP events.
 */
typedef enum {

    // Standard information

    /** This event is sent when a new flow/connection is detected by PACE and handled by PADE as GTP. */
    PACE2_ADVANCED_GTP_NEW_CONNECTION,

    /** This event signals the end or timeout of a known GTP connection. */
    PACE2_ADVANCED_GTP_CONNECTION_END,

    /** This event signals that this GTP connection can't be decoded. Provides a char * with information
     * about the problems. */
    PACE2_ADVANCED_GTP_INVALID,

    /** This event is sent for every GTP message. Provides a struct ipd_gtp_message * with information
     * about the message. */
    PACE2_ADVANCED_GTP_MESSAGE,

    /** This event is sent for every GTP message extension header. Provides a struct ipd_gtp_message_data *
     * with information about the extension header. */
    PACE2_ADVANCED_GTP_EXTENSION,

    /** This event is sent for every GTP message data part. Provides a struct ipd_gtp_message_data *
     * with information about the data. */
    PACE2_ADVANCED_GTP_DATA,

    /** This event is sent for every GTP message information element. Provides a struct ipd_gtp_information *
     * with information about the information element. */
    PACE2_ADVANCED_GTP_INFORMATION,

    /** This event is sent for every GTP message information element, which groups sub information elements.
     * Provides a struct ipd_gtp_information_group * with information about the group. */
    PACE2_ADVANCED_GTP_INFORMATION_GROUP,

    /** This event signals that for the correlating GTP message no more events will follow.
     * This event is only thrown if no #PACE2_ADVANCED_GTP_INVALID event was before.
     * Provides an u64 * with the message ID of the correlating GTP message. */
    PACE2_ADVANCED_GTP_MESSAGE_END,

    /** This event signals that for the correlating GTP information element group no more events will follow.
     * Provides an u64 * with the group ID of the correlating GTP group. */
    PACE2_ADVANCED_GTP_INFORMATION_GROUP_END,

    // Additional information

    /** This event is sent for every GTP message information element, which contains a IMSI.
     * Provides a struct ipd_gtp_information_string * with information about the IMSI. */
    PACE2_ADVANCED_GTP_INFORMATION_IMSI,

    /** This event is sent for every GTP message information element, which contains a IMEI.
     * Provides a struct ipd_gtp_information_string * with information about the IMEI. */
    PACE2_ADVANCED_GTP_INFORMATION_IMEI,

    /** This event is sent for every GTP message information element, which contains a MSISDN.
     * Provides a struct ipd_gtp_information_string * with information about the MSISDN. */
    PACE2_ADVANCED_GTP_INFORMATION_MSISDN,

    /** This event is sent for every GTP message information element, which contains a APN.
     * Provides a struct ipd_gtp_information_string * with information about the APN. */
    PACE2_ADVANCED_GTP_INFORMATION_APN,

    /** This event is sent for every GTP message information element, which contains a TEID data plane part I.
     * Provides a struct ipd_gtp_information_value * with information about the TEID. */
    PACE2_ADVANCED_GTP_INFORMATION_TEID_DATA_I,

    /** This event is sent for every GTP message information element, which contains a TEID data plane part II.
     * Provides a struct ipd_gtp_information_value * with information about the TEID. */
    PACE2_ADVANCED_GTP_INFORMATION_TEID_DATA_II,

    /** This event is sent for every GTP message information element, which contains a TEID control plane.
     * Provides a struct ipd_gtp_information_value * with information about the TEID. */
    PACE2_ADVANCED_GTP_INFORMATION_TEID_CONTROL,

    /** This event is sent for every GTP message information element, which contains a LAC.
     * Provides a struct ipd_gtp_information_value * with information about the LAC. */
    PACE2_ADVANCED_GTP_INFORMATION_LAC,

    /** This event is sent for every GTP message information element, which contains a RAC.
     * Provides a struct ipd_gtp_information_value * with information about the RAC. */
    PACE2_ADVANCED_GTP_INFORMATION_RAC,

    /** This event is sent for every GTP message information element, which contains a RAT.
     * Provides a struct ipd_gtp_information_value * with information about the RAT.
     *  The contained field struct ipd_gtp_information_value.data can be casted into a enum ipd_gtp_rat_type. */
    PACE2_ADVANCED_GTP_INFORMATION_RAT,

    /** This event is sent for every GTP message information element, which contains a EBI.
     * Provides a struct ipd_gtp_information_value * with information about the bearer id. */
    PACE2_ADVANCED_GTP_INFORMATION_EBI,

    /** This event is sent for every GTP message information element, which contains a Cell identifier.
     * Provides a struct ipd_gtp_information_cell_ident * with information about the CID. */
    PACE2_ADVANCED_GTP_INFORMATION_CID,

    /** This event is sent for every GTP message information element, which contains a IP address.
     * Provides a struct ipd_gtp_information_address * with information about the IP address. */
    PACE2_ADVANCED_GTP_INFORMATION_ADDRESS,

    /** This event is sent for every GTP message information element, which contains a ARP from a QoS profile.
     * Provides a struct ipd_gtp_information_value * with information about the ARP. */
    PACE2_ADVANCED_GTP_INFORMATION_ARP,

    /** This event is sent for every GTP message information element, which contains a QCI from a QoS profile.
     * Provides a struct ipd_gtp_information_value * with information about the QCI. */
    PACE2_ADVANCED_GTP_INFORMATION_QCI,

    /** This event is sent for every GTP message information element, which contains the upload MBR from a QoS profile.
     * Provides a struct ipd_gtp_information_value * with information about the MBR. */
    PACE2_ADVANCED_GTP_INFORMATION_MBR_UP,

    /** This event is sent for every GTP message information element, which contains the download MBR from a QoS
     * profile. Provides a struct ipd_gtp_information_value * with information about the MBR. */
    PACE2_ADVANCED_GTP_INFORMATION_MBR_DOWN,

    /** This event is sent for every GTP message information element, which contains the upload GBR from a QoS profile.
     * Provides a struct ipd_gtp_information_value * with information about the GBR. */
    PACE2_ADVANCED_GTP_INFORMATION_GBR_UP,

    /** This event is sent for every GTP message information element, which contains the download GBR from a QoS
     * profile. Provides a struct ipd_gtp_information_value * with information about the GBR. */
    PACE2_ADVANCED_GTP_INFORMATION_GBR_DOWN,

    /** This event is sent for every GTP message information element, which contains end user address
    information elements. Provides a struct ipd_gtp_information_end_user_address * with information about
    the IP address. */
    PACE2_ADVANCED_GTP_INFORMATION_END_USER_ADDRESS,

    /** This event is sent for every GTP message information element, which contains cell global identifier
     * information elements. Provides a struct ipd_gtp_v2_information_user_location_information_cgi_field * with
     * CAI information. */
    PACE2_ADVANCED_GTP_INFORMATION_CGI,

    /** This event is sent for every GTP message information element, which contains service area identifier
     * information elements. Provides a struct ipd_gtp_v2_information_user_location_information_sai_field * with
     * SAI information. */
    PACE2_ADVANCED_GTP_INFORMATION_SAI,

    /** This event is sent for every GTP message information element, which contains routing area identity
     * information elements. Provides a struct ipd_gtp_v2_information_user_location_information_rai_field * with
     * RAI information. */
    PACE2_ADVANCED_GTP_INFORMATION_RAI,

    /** This event is sent for every GTP message information element, which contains tracking area identity
     * information elements. Provides a struct ipd_gtp_v2_information_user_location_information_tai_field * with
     * TAI information. */
    PACE2_ADVANCED_GTP_INFORMATION_TAI,

    /** This event is sent for every GTP message information element, which contains E-UTRAN cell global identifier
     * information elements. Provides a struct ipd_gtp_v2_information_user_location_information_ecgi_field * with ECGI
     * information. */
    PACE2_ADVANCED_GTP_INFORMATION_ECGI,

    /** This event is sent for every GTP message information element, which contains location area code
     * information elements. Provides a struct ipd_gtp_v2_information_user_location_information_lai_field * with
     * LAI information. */
    PACE2_ADVANCED_GTP_INFORMATION_LAI,

    /** This event is sent for every GTP message information element, which contains address of a GSN
     * information element. Provides a struct ipd_gtp_information_address * with GSN address information. */
    PACE2_ADVANCED_GTP_INFORMATION_GSN_ADDR,

    /** This event is sent for every GTP message information element, which contains user location
     * information elements. Provides a struct ipd_gtp_information_user_location_information * with information
     * about the user location. */
    PACE2_ADVANCED_GTP_INFORMATION_ULI,

    /** This event is sent for every GTP message information element, which contains cause value information
     * element. Provides a struct ipd_gtp_information_value * with information about the cause. */
    PACE2_ADVANCED_GTP_INFORMATION_CAUSE,

    /** This event is sent for every F-TEID message information element, which groups sub information elements.
     * Provides a struct ipd_gtp_information_f_teid * with information about the fully qualified tunnel
     * endpoint identifier. */
    PACE2_ADVANCED_GTP_INFORMATION_F_TEID,

    /** This event is sent for every PDN message information element, which groups sub information elements.
     * Provides a struct ipd_gtp_information_pdn_address_allocation * with information about the PDN
     * address allocation. */
    PACE2_ADVANCED_GTP_INFORMATION_PAA,

    /** This event is sent for every GTPv2 message information element, which contains cause value
     * information element. Provides a struct ipd_gtp_information_value * with information about the cause. */
    PACE2_ADVANCED_GTP_INFORMATION_V2_CAUSE,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_GTP_EVENT_TYPES

} PACE2_advanced_GTP_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * Enumeration for message types of GTPv1 and GTP'. For more details see TS 29060 and TS 32295.
 */
enum ipd_gtp_v1_message_type_enum {
    IPD_GTP_V1_MESSAGE_TYPE_ECHO_REQUEST = 1,
    IPD_GTP_V1_MESSAGE_TYPE_ECHO_RESPONSE = 2,
    IPD_GTP_V1_MESSAGE_TYPE_VERSION_NOT_SUPPORTED = 3,
    IPD_GTP_V1_MESSAGE_TYPE_NODE_ALIVE_REQ = 4,
    IPD_GTP_V1_MESSAGE_TYPE_NODE_ALIVE_RESP = 5,
    IPD_GTP_V1_MESSAGE_TYPE_REDIRECT_REQ = 6,
    IPD_GTP_V1_MESSAGE_TYPE_REDIRECT_RESP = 7,
    IPD_GTP_V1_MESSAGE_TYPE_CREATE_PDU_CONTEXT_REQ = 16,
    IPD_GTP_V1_MESSAGE_TYPE_CREATE_PDU_CONTEXT_RESP = 17,
    IPD_GTP_V1_MESSAGE_TYPE_UPDATE_PDU_CONTEXT_REQ = 18,
    IPD_GTP_V1_MESSAGE_TYPE_UPDATE_PDU_CONTEXT_RESP = 19,
    IPD_GTP_V1_MESSAGE_TYPE_DELETE_PDU_CONTEXT_REQ = 20,
    IPD_GTP_V1_MESSAGE_TYPE_DELETE_PDU_CONTEXT_RESP = 21,
    IPD_GTP_V1_MESSAGE_TYPE_INIT_PDU_CONTEXT_ACT_REQ = 22,
    IPD_GTP_V1_MESSAGE_TYPE_INIT_PDU_CONTEXT_ACT_RESP = 23,
    IPD_GTP_V1_MESSAGE_TYPE_ERROR_INDICATION = 26,
    IPD_GTP_V1_MESSAGE_TYPE_PDU_NOTIF_REQ = 27,
    IPD_GTP_V1_MESSAGE_TYPE_PDU_NOTIF_RESP = 28,
    IPD_GTP_V1_MESSAGE_TYPE_PDU_NOTIF_REJECT_REQ = 29,
    IPD_GTP_V1_MESSAGE_TYPE_PDU_NOTIF_REJECT_RESP = 30,
    IPD_GTP_V1_MESSAGE_TYPE_SUPPORTED_EXT_NOTIF_HEADER = 31,
    IPD_GTP_V1_MESSAGE_TYPE_SEND_ROUTE_INFO_GPRS_REQ = 32,
    IPD_GTP_V1_MESSAGE_TYPE_SEND_ROUTE_INFO_GPRS_RESP = 33,
    IPD_GTP_V1_MESSAGE_TYPE_FAILURE_REPORT_REQ = 34,
    IPD_GTP_V1_MESSAGE_TYPE_FAILURE_REPORT_RESP = 35,
    IPD_GTP_V1_MESSAGE_TYPE_NOTE_MS_GPRS_PRESENT_REQ = 36,
    IPD_GTP_V1_MESSAGE_TYPE_NOTE_MS_GPRS_PRESENT_RESP = 37,
    IPD_GTP_V1_MESSAGE_TYPE_IDENT_REQ = 48,
    IPD_GTP_V1_MESSAGE_TYPE_IDENT_RESP = 49,
    IPD_GTP_V1_MESSAGE_TYPE_SGSN_CONTEXT_REQ = 50,
    IPD_GTP_V1_MESSAGE_TYPE_SGSN_CONTEXT_RESP = 51,
    IPD_GTP_V1_MESSAGE_TYPE_SGSN_CONTEXT_ACK = 52,
    IPD_GTP_V1_MESSAGE_TYPE_FORWARD_RELOC_REQ = 53,
    IPD_GTP_V1_MESSAGE_TYPE_FORWARD_RELOC_RESP = 54,
    IPD_GTP_V1_MESSAGE_TYPE_FORWARD_RELOC_COMPLETE = 55,
    IPD_GTP_V1_MESSAGE_TYPE_RELOC_CANCEL_REQ = 56,
    IPD_GTP_V1_MESSAGE_TYPE_RELOC_CANCEL_RESP = 57,
    IPD_GTP_V1_MESSAGE_TYPE_FORWARD_SRNS_CONTEXT = 58,
    IPD_GTP_V1_MESSAGE_TYPE_FORWARD_RELOC_COMPLETE_ACK = 59,
    IPD_GTP_V1_MESSAGE_TYPE_FORWARD_SRNS_CONTEXT_ACK = 60,
    IPD_GTP_V1_MESSAGE_TYPE_RAN_INFO_RELAY = 70,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_NOTIF_REQ = 96,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_NOTIF_RESP = 97,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_NOTIF_REJECT_REQ = 98,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_NOTIF_REJECT_RESP = 99,
    IPD_GTP_V1_MESSAGE_TYPE_CREATE_MBMS_CONTEXT_REQ = 100,
    IPD_GTP_V1_MESSAGE_TYPE_CREATE_MBMS_CONTEXT_RESP = 101,
    IPD_GTP_V1_MESSAGE_TYPE_UPDATE_MBMS_CONTEXT_REQ = 102,
    IPD_GTP_V1_MESSAGE_TYPE_UPDATE_MBMS_CONTEXT_RESP = 103,
    IPD_GTP_V1_MESSAGE_TYPE_DELETE_MBMS_CONTEXT_REQ = 104,
    IPD_GTP_V1_MESSAGE_TYPE_DELETE_MBMS_CONTEXT_RESP = 105,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_REGISTER_REQ = 112,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_REGISTER_RESP = 113,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_DE_REGISTER_REQ = 114,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_DE_REGISTER_RESP = 115,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_SESSION_START_REQ = 116,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_SESSION_START_RESP = 117,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_SESSION_STOP_REQ = 118,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_SESSION_STOP_RESP = 119,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_SESSION_UPDATE_REQ = 120,
    IPD_GTP_V1_MESSAGE_TYPE_MBMS_SESSION_UPDATE_RESP = 121,
    IPD_GTP_V1_MESSAGE_TYPE_MS_INFO_CHANGE_NOTIF_REQ = 128,
    IPD_GTP_V1_MESSAGE_TYPE_MS_INFO_CHANGE_NOTIF_RESP = 129,
    IPD_GTP_V1_MESSAGE_TYPE_DATA_RECORD_TRANSFER_REQ = 240,
    IPD_GTP_V1_MESSAGE_TYPE_DATA_RECORD_TRANSFER_RESP = 241,
    IPD_GTP_V1_MESSAGE_TYPE_END_MARKER = 254,
    IPD_GTP_V1_MESSAGE_TYPE_G_PDU = 255,
};

/**
 * Enumeration for message types of GTPv2. For more details see TS 29274.
 */
enum ipd_gtp_v2_message_type_enum {
    IPD_GTP_V2_MESSAGE_TYPE_ECHO_REQUEST = 1,
    IPD_GTP_V2_MESSAGE_TYPE_ECHO_RESPONSE = 2,
    IPD_GTP_V2_MESSAGE_TYPE_CREATE_SESSION_REQ = 32,
    IPD_GTP_V2_MESSAGE_TYPE_CREATE_SESSION_RESP = 33,
    IPD_GTP_V2_MESSAGE_TYPE_MODIFY_BEARER_REQ = 34,
    IPD_GTP_V2_MESSAGE_TYPE_MODIFY_BEARER_RESP = 35,
    IPD_GTP_V2_MESSAGE_TYPE_DELETE_SESSION_REQ = 36,
    IPD_GTP_V2_MESSAGE_TYPE_DELETE_SESSION_RESP = 37,
    IPD_GTP_V2_MESSAGE_TYPE_CHANGE_NOTIF_REQ = 38,
    IPD_GTP_V2_MESSAGE_TYPE_CHANGE_NOTIF_RESP = 39,
    IPD_GTP_V2_MESSAGE_TYPE_MODIFY_BEARER_CMD = 64,
    IPD_GTP_V2_MESSAGE_TYPE_MODIFY_BEARER_FAILURE_IND = 65,
    IPD_GTP_V2_MESSAGE_TYPE_DELETE_BEARER_CMD = 66,
    IPD_GTP_V2_MESSAGE_TYPE_DELETE_BEARER_FAILURE_IND = 67,
    IPD_GTP_V2_MESSAGE_TYPE_BEARER_RES_CMD = 68,
    IPD_GTP_V2_MESSAGE_TYPE_BEARER_RES_FAILURE_IND = 69,
    IPD_GTP_V2_MESSAGE_TYPE_DOWNLINK_DATA_NOTIF_FAILURE_IND = 70,
    IPD_GTP_V2_MESSAGE_TYPE_TRACE_SESSION_ACTIVATION = 71,
    IPD_GTP_V2_MESSAGE_TYPE_TRACE_SESSION_DE_ACTIVATION = 72,
    IPD_GTP_V2_MESSAGE_TYPE_STOP_PAGING_IND = 73,
    IPD_GTP_V2_MESSAGE_TYPE_CREATE_BEARER_REQ = 95,
    IPD_GTP_V2_MESSAGE_TYPE_CREATE_BEARER_RESP = 96,
    IPD_GTP_V2_MESSAGE_TYPE_UPDATE_BEARER_REQ = 97,
    IPD_GTP_V2_MESSAGE_TYPE_UPDATE_BEARER_RESP = 98,
    IPD_GTP_V2_MESSAGE_TYPE_DELETE_BEARER_REQ = 99,
    IPD_GTP_V2_MESSAGE_TYPE_DELETE_BEARER_RESP = 100,
    IPD_GTP_V2_MESSAGE_TYPE_DELETE_PDN_CONNECTION_SET_REQ = 101,
    IPD_GTP_V2_MESSAGE_TYPE_DELETE_PDN_CONNECTION_SET_RESP = 102,
    IPD_GTP_V2_MESSAGE_TYPE_PGW_DOWNLINK_TRIGGER_NOTIF = 103,
    IPD_GTP_V2_MESSAGE_TYPE_PGW_DOWNLINK_TRIGGER_ACK = 104,
    IPD_GTP_V2_MESSAGE_TYPE_IDENT_REQ = 128,
    IPD_GTP_V2_MESSAGE_TYPE_IDENT_RESP = 129,
    IPD_GTP_V2_MESSAGE_TYPE_CONTEXT_REQ = 130,
    IPD_GTP_V2_MESSAGE_TYPE_CONTEXT_RESP = 131,
    IPD_GTP_V2_MESSAGE_TYPE_CONTEXT_ACK = 132,
    IPD_GTP_V2_MESSAGE_TYPE_FORWARD_RELOC_REQ = 133,
    IPD_GTP_V2_MESSAGE_TYPE_FORWARD_RELOC_RESP = 134,
    IPD_GTP_V2_MESSAGE_TYPE_FORWARD_RELOC_COMPLETE_NOTIF = 135,
    IPD_GTP_V2_MESSAGE_TYPE_FORWARD_RELOC_COMPLETE_ACK = 136,
    IPD_GTP_V2_MESSAGE_TYPE_FORWARD_ACCESS_CONTEXT_NOTIF = 137,
    IPD_GTP_V2_MESSAGE_TYPE_FORWARD_ACCESS_CONTEXT_ACK = 138,
    IPD_GTP_V2_MESSAGE_TYPE_RELOC_CANCEL_REQ = 139,
    IPD_GTP_V2_MESSAGE_TYPE_RELOC_CANCEL_RESP = 140,
    IPD_GTP_V2_MESSAGE_TYPE_CONFIGURATION_TRANSFER_TUNNEL = 141,
    IPD_GTP_V2_MESSAGE_TYPE_DETACH_NOTIF = 149,
    IPD_GTP_V2_MESSAGE_TYPE_DETACH_ACK = 150,
    IPD_GTP_V2_MESSAGE_TYPE_CS_PAGING_IND = 151,
    IPD_GTP_V2_MESSAGE_TYPE_RAN_INFO_RELAY = 152,
    IPD_GTP_V2_MESSAGE_TYPE_ALERT_MME_NOTIF = 153,
    IPD_GTP_V2_MESSAGE_TYPE_ALERT_MME_ACK = 154,
    IPD_GTP_V2_MESSAGE_TYPE_UE_ACTIVITY_NOTIF = 155,
    IPD_GTP_V2_MESSAGE_TYPE_UE_ACTIVITY_ACK = 156,
    IPD_GTP_V2_MESSAGE_TYPE_ISR_STATUS_IND = 157,
    IPD_GTP_V2_MESSAGE_TYPE_CREATE_FORWARDING_TUNNEL_REQ = 160,
    IPD_GTP_V2_MESSAGE_TYPE_CREATE_FORWARDING_TUNNEL_RESP = 161,
    IPD_GTP_V2_MESSAGE_TYPE_SUSPEND_NOTIF = 162,
    IPD_GTP_V2_MESSAGE_TYPE_SUSPEND_ACK = 163,
    IPD_GTP_V2_MESSAGE_TYPE_RESUME_NOTIF = 164,
    IPD_GTP_V2_MESSAGE_TYPE_RESUME_ACK = 165,
    IPD_GTP_V2_MESSAGE_TYPE_CREATE_INDIRECT_DATA_FORWARDING_TUNNEL_REQ = 166,
    IPD_GTP_V2_MESSAGE_TYPE_CREATE_INDIRECT_DATA_FORWARDING_TUNNEL_RESP = 167,
    IPD_GTP_V2_MESSAGE_TYPE_DELETE_INDIRECT_DATA_FORWARDING_TUNNEL_REQ = 168,
    IPD_GTP_V2_MESSAGE_TYPE_DELETE_INDIRECT_DATA_FORWARDING_TUNNEL_RESP = 169,
    IPD_GTP_V2_MESSAGE_TYPE_RELEASE_ACCESS_BEARER_REQ = 170,
    IPD_GTP_V2_MESSAGE_TYPE_RELEASE_ACCESS_BEARER_RESP = 171,
    IPD_GTP_V2_MESSAGE_TYPE_DOWNLINK_DATA_NOTIF = 176,
    IPD_GTP_V2_MESSAGE_TYPE_DOWNLINK_DATA_NOTIF_ACK = 177,
    IPD_GTP_V2_MESSAGE_TYPE_PGW_RESTART_NOTIF = 179,
    IPD_GTP_V2_MESSAGE_TYPE_PGW_RESTART_NOTIF_ACK = 180,
    IPD_GTP_V2_MESSAGE_TYPE_UPDATE_PDN_CONNECTION_SET_REQ = 200,
    IPD_GTP_V2_MESSAGE_TYPE_UPDATE_PDN_CONNECTION_SET_RESP = 201,
    IPD_GTP_V2_MESSAGE_TYPE_MODIFY_ACCESS_BEARER_REQ = 211,
    IPD_GTP_V2_MESSAGE_TYPE_MODIFY_ACCESS_BEARER_RESP = 212,
    IPD_GTP_V2_MESSAGE_TYPE_MBMS_SESSION_START_REQ = 231,
    IPD_GTP_V2_MESSAGE_TYPE_MBMS_SESSION_START_RESP = 232,
    IPD_GTP_V2_MESSAGE_TYPE_MBMS_SESSION_UPDATE_REQ = 233,
    IPD_GTP_V2_MESSAGE_TYPE_MBMS_SESSION_UPDATE_RESP = 234,
    IPD_GTP_V2_MESSAGE_TYPE_MBMS_SESSION_STOP_REQ = 235,
    IPD_GTP_V2_MESSAGE_TYPE_MBMS_SESSION_STOP_RESP = 236,
};

/**
 * Enumeration for the different protocol versions of GTP.
 */
enum ipd_gtp_protocol_type_enum {

    /** Signals that this event correlates to a GTPv0 packet. */
    IPD_GTP_PROTOCOL_TYPE_V0,

    /** Signals that this event correlates to a GTPv1 packet. */
    IPD_GTP_PROTOCOL_TYPE_V1,

    /** Signals that this event correlates to a GTPv2 packet. */
    IPD_GTP_PROTOCOL_TYPE_V2,

    /** Signals that this event correlates to a GTP' packet. */
    IPD_GTP_PROTOCOL_TYPE_PRIME,
};

/**
 * Structure for a GTP message.
 */
struct ipd_gtp_message {

    /**
     * A PADE internal ID for this message. It can be used to correlate corresponding GTP events to this message.
     */
    u64 message_id;

    /**
     * If a TEID is contained in the message, then this field will contain it. Otherwise it is set to "0".
     */
    u32 teid;

    /**
     * Message sequence number. If it is not in the packet (GTPv0 and GTPv1 if the sequence number flag is not set),
     * it is set to "0".
     */
    u32 sequence_number;

    /**
     * The value of the contained message type. Together with struct @ref ipd_gtp_message::protocol_type
     * "protocol_type", enum @ref ipd_gtp_v1_message_type_enum and enum @ref ipd_gtp_v2_message_type_enum
     * it is possible to get a more detailed description as a enum variable for this message type.
     * A string representation are available over the ipd_gtp_get_message_type_description function.
     */
    u8 message_type;

    /**
     * The protocol type in which this message was encoded.
     */
    enum ipd_gtp_protocol_type_enum protocol_type;

    /**
     * The completely byte stream which forms the GTP message.
     */
    struct ipd_charset_string message;

    /**
     * The part of the byte stream, which forms the header of this GTP message.
     */
    struct ipd_charset_string header;

    /**
     * The part of the byte stream, which follows of the header of the GTP message.
     */
    struct ipd_charset_string data;
};

/**
 * Structure for a subset of a byte stream of a GTP message.
 */
struct ipd_gtp_message_data {

    /**
     * Correlating message id, to which this data/extension header correlate.
     */
    u64 message_id;

    /**
     * The bytes of the data/extension header.
     */
    struct ipd_charset_string data;
};

/**
 * Enumeration for Information Element types of GTPv1 and GTP'. For more details see TS 29060 and TS 32295.
 */
enum ipd_gtp_v1_information_type_enum {
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_CAUSE = 1,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_IMSI = 2,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RAI = 3,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TLLI = 4,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TMSI = 5,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_REORDERING_REQUIRED = 8,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_AUTH_TRIPLET = 9,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MAP_CAUSE = 11,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TMSI_SIG = 12,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MS_VALID = 13,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RECOVERY = 14,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_SELECT_MODE = 15,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TEID_DATA1 = 16,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TEID_CONTROL = 17,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TEID_DATA2 = 18,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TEARDOWN = 19,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_NSAPI = 20,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RANAP_CAUSE = 21,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RAB_CONTEXT = 22,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RADIO_PRIO_SMS = 23,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RADIO_PRIO = 24,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_PACKET_FLOW_ID = 25,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_CHARGING_CHAR = 26,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TRACE_REF = 27,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TRACE_TYPE = 28,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MS_NOT_REACHABLE_REASON = 29,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_PACKET_TRANSFER_CMD = 126,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_CHARGING_ID = 127,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_END_USER_ADDR = 128,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MM_CONTEXT = 129,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_PDP_CONTEXT = 130,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_ACCESS_POINT_NAME = 131,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_PROT_CONFIG = 132,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_GSN_ADDR = 133,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MSISDN = 134,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_QOS_PROFILE = 135,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_AUTH_QUINTUPLET = 136,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TRAFFIC_FLOW_TEMP = 137,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TARGET_IDENT = 138,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_UTRAN_TRANSPORT_CONTAINER = 139,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RAB_SETUP_INFO = 140,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_EXTENSION_HEADER_LIST = 141,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TRIGGER_ID = 142,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_OMC_IDENT = 143,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RAN_TRANSPORT_CONTAINER = 144,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_PDP_CONTEXT_PRIO = 145,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RAB_SETUP_INFO_ADD = 146,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_SGSN = 147,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_COMMON_FLAGS = 148,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_APN_REST = 149,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RADIO_PRIO_LCS = 150,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RAT_TYPE = 151,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_USER_LOC_INFO = 152,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MS_TIME_ZONE = 153,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_IMEI = 154,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_CAMEL_CHARGING_INFO = 155,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MBMS_UE_CONTEXT = 156,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TMGI = 157,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RIM_ROUTE_ADDR = 158,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MBMS_PROT_CONF_OPTION = 159,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MBMS_SERVICE_AREA = 160,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_SRC_RNC_PDCP_CONTEXT = 161,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_TRACE_INFO_ADD = 162,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_HOP_COUNT = 163,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_SELECT_PLMN_ID = 164,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MBMS_SESSION_IDENT = 165,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MBMS_2G_3G_INDICATOR = 166,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_NSAPI_ENH = 167,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MBMS_SESSION_DURATION = 168,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MBMS_TRACE_INFO_ADD = 169,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MBMS_SESSION_REP = 170,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MBMS_TIME_TO_DATA_TRANS = 171,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_BSS_CONTAINER = 173,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_CELL_IDENT = 174,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_PDU_NUMBER = 175,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_BSSGP_CAUSE = 176,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MBMS_REQ_BEARER_CAPA = 177,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RIM_ROUTE_ADDR_DISC = 178,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_PFC_SETUP_LIST = 179,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_PS_HANDOVER_XID_PARA = 180,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MS_INFO_CHANGE_REPORT_ACTION = 181,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_DIRECT_TUNNEL_FLAGS = 182,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_CORRELATION_ID = 183,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_BEARER_CONTROL_MODE = 184,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MBMS_FLOW_IDENT = 185,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MBMS_IP_MCAST_DIST = 186,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MBMS_DIST_ACK = 187,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RAT_HANDOVER_INFO = 188,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RFSP_INDEX = 189,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_FQDN = 190,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_E_ALLOC_RETENTION_PRIO1 = 191,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_E_ALLOC_RETENTION_PRIO2 = 192,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_COMMON_FLAGS_EXT = 193,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_UCI = 194,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_CSG_INFO_REPORT_ACTION = 195,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_CSG_ID = 196,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_CMI = 197,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_AMBR = 198,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_UE_NETWORK_CAPA = 199,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_UE_AMBR = 200,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_APN_AMBR = 201,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_GGSN_BACKOFF_TIME = 202,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_SIGNAL_PRIO_INDICATION = 203,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_SIGNAL_PRIO_INDICATION_NSAPI = 204,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_HIGHER_16MBPS_BITRATE_FLAG = 205,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_MM_CONTEXT_ADD = 207,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_SRVCC_ADD_FLAGS = 208,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_STN_SR = 209,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_C_MSISDN = 210,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_RANAP_CAUSE_EXT = 211,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_ENODEB_ID = 212,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_SELECT_MODE_NSAPI = 213,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_ULI_TIMESTAMP = 214,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_LHN_ID_NSAPI = 215,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_CN_OP_SELECT_ENTITY = 216,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_SEQ_NUM_OF_RELEASED_PACKETS = 249,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_SEQ_NUM_OF_CANCELLED_PACKETS = 250,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_CHARGING_GATEWAY_ADDR = 251,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_DATA_RECORD_PACKET = 252,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_REQUESTS_RESPONDED = 253,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_ADDR_OF_RECOMMENDED_NODE = 254,
    IPD_GTP_V1_INFORMATION_ELEMENT_TYPE_PRIVATE_EXTENSION = 255,
};


/**
 * Enumeration for Information Element types of GTPv2. For more details see TS 29274.
 */
enum ipd_gtp_v2_information_type_enum {
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_IMSI = 1,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_CAUSE = 2,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_RECOVERY = 3,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_STN_SR = 51,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_APN = 71,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_AMBR = 72,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_EBI = 73,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_IP_ADDR = 74,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MEI = 75,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MSISDN = 76,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_INDICATION = 77,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_PCO = 78,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_PAA = 79,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_BEARER_QOS = 80,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_FLOW_QOS = 81,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_RAT_TYPE = 82,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_SERVING_NETWORK = 83,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_BEARER_TFT = 84,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_TAD = 85,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_ULI = 86,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_F_TEID = 87,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_TMSI = 88,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_GLOBAL_CN_ID = 89,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_S103PDF = 90,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_S1UDF = 91,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_DELAY_VALUE = 92,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_BEARER_CONTEXT = 93,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_CHARGING_ID = 94,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_CHARGING_CHAR = 95,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_TRACE_INFO = 96,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_BEARER_FLAGS = 97,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_PDN_TYPE = 99,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_PROCEDURE_TRANSACTION_ID = 100,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MM_CONTEXT_GSM_KEY_T = 103,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MM_CONTEXT_UMTS_KEY_UC_Q = 104,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MM_CONTEXT_GSM_KEY_UC_Q = 105,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MM_CONTEXT_UMTS_KEY_Q = 106,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MM_CONTEXT_EPS_QUAD_Q = 107,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MM_CONTEXT_UMTS_KEY_QUAD_Q = 108,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_PDN_CONNECTION = 109,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_PDN_NUMBERS = 110,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_P_TMSI = 111,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_P_TMSI_SIGNATURE = 112,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_HOP_COUNTER = 113,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_UE_TIME_ZONE = 114,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_TRACE_REF = 115,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_COMPLETE_REQ_MESSAGE = 116,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_GUTI = 117,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_F_CONTAINER = 118,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_F_CAUSE = 119,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_PLMN_ID = 120,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_TARGET_IDENT = 121,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_PACKET_FLOW_ID = 123,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_RAB_CONTEXT = 124,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_SRC_RNC_PDCP_CONTEXT_INFO = 125,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_UDP_SRC_PORT_NUMBER = 126,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_APN_RESTRICTION = 127,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_SELECT_MODE = 128,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_SRC_IDENT = 129,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_CHANGE_REPORT_ACTION = 131,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_FQ_CSID = 132,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_CHANNEL_NEEDED = 133,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_E_MLPP_PRIO = 134,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_NODE_TYPE = 135,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_FQDN = 136,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_TI = 137,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MBMS_SESSION_DURATION = 138,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MBMS_SERVICE_AREA = 139,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MBMS_SESSION_IDENT = 140,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MBMS_FLOW_IDENT = 141,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MBMS_IP_MULTICAT_DIST = 142,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MBMS_DIST_ACK = 143,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_RFSP_INDEX = 144,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_UCI = 145,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_CSG_INFO_REPORT_ACTION = 146,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_CSG_ID = 147,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_CMI = 148,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_SERVICE_INDICATOR = 149,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_DETACH_TYPE = 150,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_LDN = 151,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_NODE_FEATURES = 152,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MBMS_TIME_TO_DATA_TRANSFER = 153,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_THROTTLING = 154,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_ARP = 155,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_EPC_TIMER = 156,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_SIGNAL_PRIO_IND = 157,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_TMGI = 158,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_SRVCC_ADD_MM_CONTEXT = 159,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_SRVCC_ADD_FLAGS = 160,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MDT_CONFIGURATION = 162,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_APCO = 163,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_ABSOLUTE_TIME_OF_MBMS_DATA_TRANSFER = 164,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_HNB_INFO_REPORT = 165,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_IP4CP = 166,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_CHANGE_TO_REPORT_FLAGS = 167,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_ACTION_IND = 168,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_TWAN_IDENT = 169,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_ULI_TIMESTAMP = 170,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_MBMS_FLAGS = 171,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_RAN_NAS_CAUSE = 172,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_CN_OP_SELECT_MODE = 173,
    IPD_GTP_V2_INFORMATION_ELEMENT_TYPE_PRIVATE_EXTENSION = 255,
};

/**
 * Structure for a GTP information Element.
 */
struct ipd_gtp_information {

    /**
     * Correlating message or group id, to which this data/extension header correlate.
     */
    u64 message_id;

    /**
     * The value of the contained information element type. Together with struct @ref ipd_gtp_message::protocol_type
     * "protocol_type", enum @ref ipd_gtp_v1_information_type_enum and enum @ref ipd_gtp_v1_information_type_enum
     * it is possible to get a more detailed description as a enum variable for this information element type.
     * A string representation are available over the ipd_gtp_get_information_element_description function.
     */
    u8 information_type;

    /**
     * The part of the byte stream of a GTP message, which contains the completely information element.
     */
    struct ipd_charset_string information;

    /**
     * The part of the byte stream of a GTP message, which contains the data of the information element.
     */
    struct ipd_charset_string data;

    /**
     * The protocol type in which this message was encoded.
     */
    enum ipd_gtp_protocol_type_enum protocol_type;
};

/**
 * Structure for a parsed GTP information element, which contains a String (TBCD or not) as value.
 */
struct ipd_gtp_information_string {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * The contained string of this information element.
     */
    struct ipd_charset_string data;
};

/**
 * Structure for a parsed GTP information element, which contains a number as value.
 */
struct ipd_gtp_information_value {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * The contained number of this information element.
     */
    u64 data;
};

/**
 * Structure for a parsed GTP information element, which groups sub information elements.
 */
struct ipd_gtp_information_group {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * A PADE internal ID for this group. It can be used to correlate corresponding GTP events to this group.
     */
    u64 group_id;

    /**
     * The part of the byte stream of a GTP message, which contains the data of the information group.
     */
    struct ipd_charset_string data;
};

/**
 * Enumeration for the type of a source cell identifier.
 */
enum ipd_gtp_cell_ident_type {
    IPD_GTP_CELL_IDENT_TYPE_CELL_ID = 0,
    IPD_GTP_CELL_IDENT_TYPE_RNC_ID = 1,
};

/**
 * Structure for a parsed GTP information element, which contains a cell identifier as value.
 */
struct ipd_gtp_information_cell_ident {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * The type of the source.
     */
    enum ipd_gtp_cell_ident_type source_type;

    /**
     * The ID of the source.
     */
    u64 source_id;

    /**
     * The ID of the target.
     */
    u64 target_id;
};

struct ipd_gtp_information_end_user_address {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * The organisation which is responsible for the PDP type number field and the PDP address format.
     */
    u8 pdp_type_organization;

    /**
     * Organisation dependent PDP type number field.
     */
    u8 pdp_type_number;

    /**
     * Flag to signal if IP address information are available.
     */
    u8 found_ip;

    /**
     * Flag to signal if IPv4 or IPv6 address is used.
     */
    u8 is_ip_v6;

    /**
     * The contained IP address(es) of this information element.
     */
    union ip_v4_v6_addr ipv4_v6addr;
};

struct ipd_gtp_v2_information_user_location_information_cgi_field {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * The mobile country code.
     */
    u16 MCC_digit;

    /**
     * The mobile network code.
     */
    u16 MNC_digit;

    /**
     * The location area code (LAC).
     */
    u16 location_area_code;

    /**
     * The cell identity (CI).
     */
    u16 cell_identity;

    /**
     * The home network identity.
     */
    struct ipd_string hni;
};

struct ipd_gtp_v2_information_user_location_information_sai_field {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * The mobile country code.
     */
    u16 MCC_digit;

    /**
     * The mobile network code.
     */
    u16 MNC_digit;

    /**
     * The location area code (LAC).
     */
    u16 location_area_code;

    /**
     * The service area code (SAC).
     */
    u16 service_area_code;

    /**
     * The home network identity.
     */
    struct ipd_string hni;
};

struct ipd_gtp_v2_information_user_location_information_rai_field {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * The mobile country code.
     */
    u16 MCC_digit;

    /**
     * The mobile network code.
     */
    u16 MNC_digit;

    /**
     * The location area code (LAC).
     */
    u16 location_area_code;

    /**
     * The routing area code (RAC).
     */
    u16 routing_area_code;

    /**
     * The home network identity.
     */
    struct ipd_string hni;
};

struct ipd_gtp_v2_information_user_location_information_tai_field {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * The mobile country code.
     */
    u16 MCC_digit;

    /**
     * The mobile network code.
     */
    u16 MNC_digit;

    /**
     * The tracking area code (TAC).
     */
    u16 tracking_area_code;

    /**
     * The home network identity.
     */
    struct ipd_string hni;
};

struct ipd_gtp_v2_information_user_location_information_ecgi_field {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * The mobile country code.
     */
    u16 MCC_digit;

    /**
     * The mobile network code.
     */
    u16 MNC_digit;

    /**
     * The E-UTRAN cell identifier (ECI).
     */
    u32 e_utran_cell_identifier;

    /**
     * The home network identity.
     */
    struct ipd_string hni;
};

struct ipd_gtp_v2_information_user_location_information_lai_field {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * The mobile country code.
     */
    u16 MCC_digit;

    /**
     * The mobile network code.
     */
    u16 MNC_digit;

    /**
     * The location area code (LAC).
     */
    u16 location_area_code;

    /**
     * The home network identity.
     */
    struct ipd_string hni;
};

union geographic_location_field {
    u16 cell_identity;
    u16 service_area_code;
    u16 routing_area_code;
};

enum geographic_location_type_t {
    CELL_GLOBAL_IDENTIFICATION = 0,
    SERVICE_AREA_IDENTITY,
    ROUTING_AREA_IDENTIFICATION
};

struct ipd_gtp_v1_information_user_location_information {

    /**
     * The type of location information.
     */
    enum geographic_location_type_t geographic_location_type;

    /**
     * The mobile country code. Value is not validated, it is set regardless.
     */
    u16 MCC_digit;

    /**
     * The mobile network code. Value is not validated, it is set regardless.
     */
    u16 MNC_digit;

    /**
     * The location area code (LAC).
     */
    u16 location_area_code;

    /**
     * Used to convey the actual geographic information as indicated in geographic_location_type.
     */
    union geographic_location_field geographic_location;

    /**
     * The home network identity. Only generated if MNC and MCC are valid.
     */
    struct ipd_string hni;
};

struct ipd_gtp_v2_information_user_location_information {

    /**
     * Indicates if the corresponding type shall be present in a respective field.
     */
    u8 cgi_flag;
    u8 sai_flag;
    u8 rai_flag;
    u8 tai_flag;
    u8 ecgi_flag;
    u8 lai_flag;
};

union gtp_version_field {
    struct ipd_gtp_v1_information_user_location_information gtp_v1;
    struct ipd_gtp_v2_information_user_location_information gtp_v2;
};

struct ipd_gtp_information_user_location_information {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * The protocol type in which this message was encoded.
     */
    enum ipd_gtp_protocol_type_enum protocol_type;

    /**
     * The corresponding protocol type as in ipd_gtp_protocol_type_enum stated.
     */
    union gtp_version_field gtp_version_type;
};

enum ipd_gtp_FTEID_interfaces {
    S1U_eNodeB_GTPU_INTERFACE = 0,
    S1U_SGW_GTPU_INTERFACE,
    S12_RNC_GTPU_INTERFACE,
    S12_SGW_GTPU_INTERFACE,
    S5S8_SGW_GTPU_INTERFACE,
    S5S8_PGW_GTPU_INTERFACE,
    S5S8_SGW_GTPC_INTERFACE,
    S5S8_PGW_GTPC_INTERFACE,
    S5S8_SGW1_PMIPv6_INTERFACE,
    S5S8_SGW2_PMIPv6_INTERFACE,
    S11_MME_GTPC_INTERFACE,
    S11S4_SGW_GTPC_INTERFACE,
    S10_MME_GTPC_INTERFACE,
    S3_MME_GTPC_INTERFACE,
    S3_SGSN_GTPC_INTERFACE,
    S4_SGSN_GTPU_INTERFACE,
    S4_SGW_GTPU_INTERFACE,
    S4_SGSN_GTPC_INTERFACE,
    S16_SGSN_GTPC_INTERFACE,
    ENODEB_GTPU_INTERFACE_FOR_DL_DATA_FORWARDING,
    ENODEB_GTPU_INTERFACE_FOR_UL_DATA_FORWARDING,
    RNC_GTPU_INTERFACE_FOR_DATA_FORWARDING,
    SGSN_GTPU_INTERFACE_FOR_DATA_FORWARDING,
    SGW_GTPU_INTERFACE_FOR_DATA_FORWARDING,
    SM_MBMS_GW_GTPC_INTERFACE,
    SN_MBMS_GW_GTPC_INTERFACE,
    SM_MME_GTPC_INTERFACE,
    SN_SGSN_GTPC_INTERFACE,
    SGW_GTPU_INTERFACE_FOR_UL_DATA_FORWARDING,
    SN_SGSN_GTPU_INTERFACE,
    S2B_EPDG_GTPC_INTERFACE,
    S2BU_EPDG_GTPC_INTERFACE,
    S2B_PGW_GTPC_INTERFACE,
    S2BU_PGW_GTPU_INTERFACE,
    S2A_TWAN_GTPU_INTERFACE,
    S2A_TWAN_GTPC_INTERFACE,
    S2A_PGW_GTPC_INTERFACE,
    S2A_PGW_GTPU_INTERFACE,
    S11_MME_GTPU_INTERFACE,
    S11_SGW_GTPU_INTERFACE,

    COUNT_FTEID_INTERFACES
};

/**
 * Enumeration for the type of a RAT information element value.
 *
 * The information element itself will be reported as struct ipd_gtp_information_value and
 * it's struct ipd_gtp_information_value.data can be casted to this enumeration.
 */
enum ipd_gtp_rat_type {
    IPD_GTP_RAT_TYPE_UTRAN = 1,
    IPD_GTP_RAT_TYPE_GERAN = 2,
    IPD_GTP_RAT_TYPE_WLAN = 3,
    IPD_GTP_RAT_TYPE_GAN = 4,
    IPD_GTP_RAT_TYPE_HSPA_EVOLUTION = 5,
    IPD_GTP_RAT_TYPE_E_UTRAN = 6,
    IPD_GTP_RAT_TYPE_VIRTUAL = 7,
};

/**
 * Structure for a parsed GTP information element, which contains a IP address (IPv4, IPv6) as value.
 */
struct ipd_gtp_information_address {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * Flag to signal if it is a IPv4 or a IPv6 address.
     */
    u8 is_ip_v6;

    /**
     * The contained IP address of this information element.
     */
    union ip_v4_v6_addr address;
};

/**
 * Structure for a parsed GTP F-TEID element.
 */
struct ipd_gtp_information_f_teid {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * Flag to signal a IPv4 address.
     */
    u8 existing_ip_v4;

    /**
     * Flag to signal a IPv6 address.
     */
    u8 existing_ip_v6;

    /**
     * The contained interface type and endpoint.
     */
    enum ipd_gtp_FTEID_interfaces interface_type;

    /**
     * The field present either a TEID or a GRE key.
     */
    u32 teid_gre_key;

    /**
     * The array contains the address(es) of this information element.
     * Variable existing_ip_v4 indicates a given IPv4 address while
     * existing_ip_v6 represents an IPv6 address.
     * If both addresses are present, the first one will be
     * IPv4 and the second one the IPv6 address.
     * The address_array_len variable is also an indicator
     * for the number of addresses.
     * This field is not included in the serialization.
     */
    union ip_v4_v6_addr * address;

    /**
     * The length of the address array.
     */
    u8 address_array_len;
};

struct ipd_gtp_information_pdn_address_allocation {

    /**
     * Correlating message or group id, to which this information element correlate.
     */
    u64 message_id;

    /**
     * provides information of ip address(es).
     */
    u8 pdn_type;

    /**
     * The array contains the address(es) of this information element.
     * Variable existing_ip_v4 indicates a given IPv4 address while
     * existing_ip_v6 represents an IPv6 address.
     * If both addresses are present, the first one will be
     * IPv4 and the second one the IPv6 address.
     * The address_array_len variable is also an indicator
     * for the number of addresses.
     * This field is not included in the serialization.
     */
    union ip_v4_v6_addr * address;

    /**
     * The length of the address array.
     */
    u8 address_array_len;
};

enum ipd_gtp_v2_cause {
    LOCAL_DETACH = 2,
    COMPLETE_DETACH = 3,
    RAT_CHANGED_FROM_3GPP_TO_NON_3GPP = 4,
    ISR_DEACTIVATION = 5,
    ERROR_INDICATION_RECEIVED_FROM_RNC_eNodeB_S4_SGSN_MME = 6,
    IMSI_DETACH_ONLY = 7,
    REACTIVATION_REQUESTED = 8,
    PDN_RECONNECTION_TO_THIS_APN_DISALLOWED = 9,
    ACCESS_CHANGED_FROM_NON_3GPP_TO_3GPP = 10,
    PDN_CONNECTION_INACTIVITY_TIMER_EXPIRES = 11,
    PGW_NOT_RESPONDING = 12,
    NETWORK_FAILURE = 13,
    QOS_PARAMETERS_MISMATCH = 14,
    REQUEST_ACCEPTED = 16,
    REQUEST_ACCEPTED_PARTIALLY = 17,
    NEW_PDN_TYPE_DUE_TO_NETWORK_PREFERENCE = 18,
    NEW_PDN_TYPE_DUE_TO_SINGLE_ADDRESS_BEARER_ONLY = 19,
    CONTEXT_NOT_FOUND = 64,
    INVALID_MESSAGE_FORMAT = 65,
    VERSION_NOT_SUPPORTED_BY_NEXT_PEER = 66,
    INVALID_LENGTH = 67,
    SERVICE_NOT_SUPPORTED = 68,
    MANDATORY_IE_INCORRECT = 69,
    MANDATORY_IE_MISSING = 70,
    SYSTEM_FAILURE = 72,
    NO_RESOURCES_AVAILABLE = 73,
    SEMANTIC_ERROR_IN_THE_TFT_OPERATION = 74,
    SYNTACTIC_ERROR_IN_THE_TFT_OPERATION = 75,
    SEMANTIC_ERRORS_IN_PACKET_FILTERS = 76,
    SYNTACTIC_ERRORS_IN_PACKET_FILTERS = 77,
    MISSING_OR_UNKNOWN_APN = 78,
    GRE_KEY_NOT_FOUND = 80,
    RELOCATION_FAILURE = 81,
    DENIED_IN_RAT = 82,
    PREFERRED_PDN_TYPE_NOT_SUPPORTED = 83,
    ALL_DYNAMIC_ADDRESSES_ARE_OCCUPIED = 84,
    UE_CONTEXT_WITHOUT_TFT_ALREADY_ACTIVATED = 85,
    PROTOCOL_TYPE_NOT_SUPPORTED = 86,
    UE_NOT_RESPONDING = 87,
    UE_REFUSES = 88,
    SERVICE_DENIED = 89,
    UNABLE_TO_PAGE_UE = 90,
    NO_MEMORY_AVAILABLE = 91,
    USER_AUTHENTICATION_FAILED = 92,
    APN_ACCESS_DENIED__NO_SUBSCRIPTION = 93,
    REQUEST_REJECTED__REASON_NOT_SPECIFIED = 94,
    P_TMSI_SIGNATURE_MISMATCH = 95,
    IMSI_IMEI_NOT_KNOWN = 96,
    SEMANTIC_ERROR_IN_THE_TAD_OPERATION = 97,
    SYNTACTIC_ERROR_IN_THE_TAD_OPERATION = 98,
    REMOTE_PEER_NOT_RESPONDING = 100,
    COLLISION_WITH_NETWORK_INITIATED_REQUEST = 101,
    UNABLE_TO_PAGE_UE_DUE_TO_SUSPENSION = 102,
    CONDITIONAL_IE_MISSING = 103,
    APN_RESTRICTION_TYPE_INCOMPATIBLE_WITH_CURRENTLY_ACTIVE_PDN_CONNECTION = 104,
    INVALID_OVERALL_LENGTH_OF_THE_TRIGGERED_RESPONSE_MESSAGE_AND_A_PIGGYBACKED_INITIAL_MESSAGE = 105,
    DATA_FORWARDING_NOT_SUPPORTED = 106,
    INVALID_REPLY_FROM_REMOTE_PEER = 107,
    FALLBACK_TO_GTPV1 = 108,
    INVALID_PEER = 109,
    TEMPORARILY_REJECTED_DUE_TO_HANDOVER_TAU_RAU_PROCEDURE_IN_PROGRESS = 110,
    MODIFICATIONS_NOT_LIMITED_TO_S1_U_BEARERS = 111,
    REQUEST_REJECTED_FOR_A_PMIPV6_REASON = 112,
    APN_CONGESTION = 113,
    BEARER_HANDLING_NOT_SUPPORTED = 114,
    UE_ALREADY_RE_ATTACHED = 115,
    MULTIPLE_PDN_CONNECTIONS_FOR_A_GIVEN_APN_NOT_ALLOWED = 116,
    TARGET_ACCESS_RESTRICTED_FOR_THE_SUBSCRIBER = 117,
    MME_SGSN_REFUSES_DUE_TO_VPLMN_POLICY = 119,
    GTP_C_ENTITY_CONGESTION = 120,
    LATE_OVERLAPPING_REQUEST = 121,
    TIMED_OUT_REQUEST = 122,
    UE_IS_TEMPORARILY_NOT_REACHABLE_DUE_TO_POWER_SAVING = 123,
    RELOCATION_FAILURE_DUE_TO_NAS_MESSAGE_REDIRECTION = 124,
    UE_NOT_AUTHORISED_BY_OCS_OR_EXTERNAL_AAA_SERVER = 125,
    MULTIPLE_ACCESSES_TO_A_PDN_CONNECTION_NOT_ALLOWED = 126,
    REQUEST_REJECTED_DUE_TO_UE_CAPABILITY = 127
};

/**
 * Function to get a String with a short description of the given message type.
 */
char const * ipd_gtp_get_message_type_description(enum ipd_gtp_protocol_type_enum protocol_type, u8 type);

/**
 * Function to get a String with a short description of the given information element type.
 */
char const * ipd_gtp_get_information_element_description(enum ipd_gtp_protocol_type_enum protocol_type, u8 type);

/**
 * Function to get a String with a short description of the given information element type.
 */
char const * ipd_gtp_get_information_element_cause_description (u8 type);

/**
 * Function to get a String with a short description of the given fully qualified tunnel endpoint identifier.
 */
char const * ipd_gtp_get_information_element_F_TEID_interface (const enum ipd_gtp_FTEID_interfaces type);

/**
 * Function to get a String with a short description of the given gtp v2 cause type.
 */
char const * ipd_gtp_get_information_element_v2_cause_description(u8 type);

/**
 * @brief Data structure for advanced GTP events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_GTP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** GTP event subtype. */
    PACE2_advanced_GTP_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_GTP_MESSAGE event. */
        struct ipd_gtp_message message;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_EXTENSION event. */
        struct ipd_gtp_message_data extension;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_DATA event. */
        struct ipd_gtp_message_data data;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_MESSAGE_END event. */
        u64 message_end;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION event. */
        struct ipd_gtp_information information;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_IMSI event. */
        struct ipd_gtp_information_string imsi;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_IMEI event. */
        struct ipd_gtp_information_string imei;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_MSISDN event. */
        struct ipd_gtp_information_string msisdn;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_APN event. */
        struct ipd_gtp_information_string apn;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_TEID_DATA_I event. */
        struct ipd_gtp_information_value teid_data_1;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_TEID_DATA_II event. */
        struct ipd_gtp_information_value teid_data_2;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_TEID_CONTROL event. */
        struct ipd_gtp_information_value teid_control;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_RAT event. */
        struct ipd_gtp_information_value rat;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_LAC event. */
        struct ipd_gtp_information_value lac;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_RAC event. */
        struct ipd_gtp_information_value rac;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_EBI event. */
        struct ipd_gtp_information_value ebi;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_CID event. */
        struct ipd_gtp_information_cell_ident cid;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_ADDRESS event. */
        struct ipd_gtp_information_address address;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_GROUP. */
        struct ipd_gtp_information_group group;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_ARP event. */
        struct ipd_gtp_information_value arp;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_QCI event. */
        struct ipd_gtp_information_value qci;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_MBR_UP event. */
        struct ipd_gtp_information_value mbr_up;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_MBR_DOWN event. */
        struct ipd_gtp_information_value mbr_down;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_GBR_UP event. */
        struct ipd_gtp_information_value gbr_up;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_GBR_DOWN event. */
        struct ipd_gtp_information_value gbr_down;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_END_USER_ADDRESS event. */
        struct ipd_gtp_information_end_user_address end_user_address;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_ULI event. */
        struct ipd_gtp_information_user_location_information uli;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_CGI event. */
        struct ipd_gtp_v2_information_user_location_information_cgi_field cgi;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_SAI event. */
        struct ipd_gtp_v2_information_user_location_information_sai_field sai;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_RAI event. */
        struct ipd_gtp_v2_information_user_location_information_rai_field rai;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_TAI event. */
        struct ipd_gtp_v2_information_user_location_information_tai_field tai;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_ECGI event. */
        struct ipd_gtp_v2_information_user_location_information_ecgi_field ecgi;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_LAI event. */
        struct ipd_gtp_v2_information_user_location_information_lai_field lai;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_GSN_ADDR event. */
        struct ipd_gtp_information_address gsn;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_CAUSE event. */
        struct ipd_gtp_information_value cause;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_V2_CAUSE event. */
        struct ipd_gtp_information_value gtp_v2_cause;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_F_TEID event. */
        struct ipd_gtp_information_f_teid f_teid;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_PAA event. */
        struct ipd_gtp_information_pdn_address_allocation paa;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INFORMATION_GROUP_END. */
        u64 group_end;

        /** Contains data for a @ref PACE2_ADVANCED_GTP_INVALID event. */
        char const * invalid;

    } event_data;

} PACE2_advanced_GTP_event;

#endif

#endif /* GTP_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
