/** @file      h323_events.h
  * @brief     H.323 event definitions.
  * @details   @copydoc doc_h323
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef H323_EVENTS_H
#define H323_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "rtp_events.h"

/* Event Group ******************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced H.323 event.
 */
#define PACE2_ADVANCED_H323_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/
/**
 * @brief Types for advanced H.323 events.
 */
typedef enum {

    /** Issued for each decoded H.323 message.  @c data points to a struct h323_state_struct. */
    PACE2_ADVANCED_H323_MESSAGE,

    /** Signals connection termination (TCP transport) or flow timeout (UDP, TCP). No data will be passed. */
    PACE2_ADVANCED_H323_END_OF_FLOW,

    /** Issued when decoding a message fails.  An error message is passed as a char * in the @c data argument. */
    PACE2_ADVANCED_H323_INVALID,

    /** Issued for each RTP packet in flows correlated to this H.323 flow.
     * @c data points to a struct ipd_rtp_data. */
    PACE2_ADVANCED_H323_RTP_DATA,

    /** Signals flow timeout of a correlated RTP flow. No data will be passed. */
    PACE2_ADVANCED_H323_RTP_END_OF_FLOW,

    /** Issued when decoding a packet in a correlated RTP flow fails.
     * An error message is passed as a <tt>char *</tt> in the @c data argument. */
    PACE2_ADVANCED_H323_RTP_INVALID,

    /** Issued when a SSRC value is found. @c data points to a struct ipd_rtp_ssrc. */
    PACE2_ADVANCED_H323_RTP_SSRC,

    /** Signals a new H.323 connection. */
    PACE2_ADVANCED_H323_NEW_CONNECTION,

    /** Reports the correlation candidates to open a logical channel. */
    PACE2_ADVANCED_H323_H245_OPEN_LOGICAL_CHANNEL,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_H323_EVENT_TYPES

} PACE2_advanced_h323_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * Enumeration for the H.323 flow states.
 */
enum h323_flow_state {

    /** Signals the idle state of a call (before start). */
    IPD_H323_FLOW_START,

    /** The setup phase of the call (start). */
    IPD_H323_FLOW_SETUP,

    /** The setup command will be transmitted to the called terminal. */
    IPD_H323_FLOW_CALL_PROCEEDING,

    /** The called terminal is ringing. */
    IPD_H323_FLOW_ALERTING,

    /** The call is established. */
    IPD_H323_FLOW_CONNECT,

    /** The call is shut down (busy, cancel, rejected, finished). */
    IPD_H323_FLOW_RELEASE_COMPLETE,

    /** Signals a H.245 tunneling in H.225 without a H.225 message. */
    IPD_H323_FLOW_EMPTY
};

/**
 * Struct for the participants.
 */
struct h323_participant_struct {

    /** Dialed digits (e.g. 0123456789). */
    struct ipd_charset_string dialed_digits;

    /** H.323 ID (e.g. nickname). */
    struct ipd_charset_string h323_id;

    /** The H.323 URL (e.g. name@h323.net). */
    struct ipd_charset_string url_id;

    /** The H.323 email. */
    struct ipd_charset_string email_id;

    /** The IP address of the participant. */
    struct ipd_charset_string ip_address;

    /** The port of the participant. */
    u16 port;

    /** The IP address is IPv6 or not. */
    u8 is_ipv6;
};

/**
 * Parsed fields from a H.323 message and its payload.
 * Equals the cb data and the dialog-state data.
 */
struct h323_state_struct {

    /** The current state of the H.323 call. */
    enum h323_flow_state flow_state;

    /** The call identifier of the H.323 call. This value is guaranteed in all messages except "Empty" messages. */
    struct ipd_charset_string callIdentifier;

    /** The source information of the H.323 call. Only the component H.323-Id of this value is guaranteed in all messages. */
    struct h323_participant_struct source;

    /** The destination information of the H.323 call. Only the component H.323-Id of this value is guaranteed in all messages. */
    struct h323_participant_struct destination;

    /** The direction of the current packet respectively the direction of the call (caller -> sender).
        The point of usage and a leading 1 will show the difference. */
    u8 direction;

    /** Signals whether the H.245 parts are tunneled in the H.225 packets or not. */
    u8 h245_tunneling;

    /** The IP address of the H.245 flow. Should be occurred only once per side in a  H.323 call. */
    struct ipd_charset_string h245_address;

    /** The port of the H.245 flow. Should be occurred only once per side in a  H.323 call. */
    u16 h245_port;

    /** The H.245 is IPv6 or not. */
    u8 h245_is_ipv6;
};

/**
 * Parsed fields from a H.245 open logical channel.
 */
struct h245_open_logical_channel {

    /** The IP address of the rtp connection. */
    struct ipd_string ip;

    /** The UDP port of the rtp connection. */
    u16 port;

    /** if true the follow ip address is the caller (rtp connection to be opened) ip address, otherwise it is
     * the ip address from the callee (other side of the rtp connection to be opened). */
    u8 is_caller;

    /** The H.245 is IPv6 or not. */
    u8 is_ipv6;
};

/**
 * @brief Data structure for advanced H.323 events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_H323_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** H.323 event subtype. */
    PACE2_advanced_h323_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_H323_MESSAGE event. */
        struct h323_state_struct h323_message;

        /** Contains data for a @ref PACE2_ADVANCED_H323_RTP_DATA event. */
        struct ipd_rtp_data rtp_data;

        /** Contains data for a @ref PACE2_ADVANCED_H323_INVALID event. */
        char const * invalid;

        /** Contains data for a @ref PACE2_ADVANCED_H323_RTP_INVALID event. */
        char const * rtp_invalid;

        /** Contains data for a @ref PACE2_ADVANCED_H323_RTP_SSRC event. */
        struct ipd_rtp_ssrc rtp_ssrc;

        /** Contains data for a @ref PACE2_ADVANCED_H323_H245_OPEN_LOGICAL_CHANNEL event. */
        struct h245_open_logical_channel h245_open_logical_channel;

    } event_data;

} PACE2_advanced_h323_event;

#endif

#endif /* H323_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
