/** @file      hi5_events.h
  * @brief     hi5 event definitions.
  * @details   @copydoc doc_hi5
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef HI5_EVENTS_H
#define HI5_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Group ******************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced Hi5 event.
 */
#define PACE2_ADVANCED_HI5_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

#ifndef PACE_KERNEL_LIB

/* Event Types ******************************************************************/
/**
 * @brief Types for advanced hi5 events.
 */
typedef enum {

    /** Signals the login of a user on Hi5 website. Provides a struct @ref ipd_hi5_logon with
        information about the user. */
    PACE2_ADVANCED_HI5_LOGON,

    /** Signals the logout of a user. Provides a struct @ref ipd_hi5_generic with information
        about the user. */
    PACE2_ADVANCED_HI5_LOGOFF,

    /** Signals that the user updated its status. Provides an @ref ipd_hi5_status_update structure with
        the update message. */
    PACE2_ADVANCED_HI5_STATUS_UPDATE,

    /** Provides a struct @ref ipd_hi5_recent_updates that contain recent updates of a user. */
    PACE2_ADVANCED_HI5_RECENT_UPDATES,

    /** Signals the comment of an update. Provides an @ref ipd_hi5_comment structure that provides
        the update message. This is obsolete since 05.2012. */
    PACE2_ADVANCED_HI5_COMMENT_UPDATE,

    /** Signals the comment of an user profile. Provides an @ref ipd_hi5_comment structure that provides
        the comment message. */
    PACE2_ADVANCED_HI5_COMMENT_PROFILE,

    /** Signals the comment of a status message. Provides an @ref ipd_hi5_comment structure that provides
        the comment message. */
    PACE2_ADVANCED_HI5_COMMENT_STATUS,

    /** Signals that a user search for a friend by its name. Provides an @ref ipd_hi5_friend_search_by_name
        structure with the name of the friend. This is obsolete since 05.2012. */
    PACE2_ADVANCED_HI5_FRIEND_SEARCH_BY_NAME,

    /** Signals that a user search for friends by their email addresses. Provides an @ref ipd_hi5_friend_search_by_email
        structure with a list of email addresses. This is obsolete since 05.2012. */
    PACE2_ADVANCED_HI5_FRIEND_SEARCH_BY_EMAIL,

    /** Signals that a user search for a friend. Provides an @ref ipd_hi5_friend_search
        structure with the search phrase. */
    PACE2_ADVANCED_HI5_FRIEND_SEARCH,

    /** Signals a friendship request. Provides an @ref ipd_hi5_friendship_request structure with
        the ID of the friend and the decision of the user (accept/reject). */
    PACE2_ADVANCED_HI5_FRIENDSHIP_REQUEST,

    /** Signals that a user removed a friend of his/her list of friends. Provides an @ref ipd_hi5_friend_remove
        structure with the ID of the friend to remove. */
    PACE2_ADVANCED_HI5_FRIEND_REMOVE,

    /** Provides an @ref ipd_hi5_view_friends structure that provide a list of user's friends
        as shown to the user. */
    PACE2_ADVANCED_HI5_VIEW_FRIENDS,

    /** Provides an @ref ipd_hi5_mailbox structure with information about a mailbox of a user and contains
        a list of messages stored in the mailbox. */
    PACE2_ADVANCED_HI5_MAILBOX,

    /** Signals that the user send a mail. Provides an @ref ipd_hi5_message structure with information
        about the message. */
    PACE2_ADVANCED_HI5_MAIL_SEND,

    /** Provides an @ref ipd_hi5_message structure with information and content of the message which
        is viewed by the user. Obsolete. */
    PACE2_ADVANCED_HI5_MAIL_VIEW,

    /** Provides an @ref ipd_hi5_message structure with information and content of the message which
        is received by the user in a chat. */
    PACE2_ADVANCED_HI5_MAIL_RECEIVE,

    /** Provides an @ref ipd_hi5_message_thread structure with the messages which are viewed by the user. */
    PACE2_ADVANCED_HI5_MAILTHREAD_VIEW,

    /** Signals that the user has created a new group. Provides an @ref ipd_hi5_group_create with
        information about the new group. */
    PACE2_ADVANCED_HI5_GROUP_CREATE,

    /** Signals that a user joined a group. Provides an @ref ipd_hi5_group_join structure with the
        ID of the group. */
    PACE2_ADVANCED_HI5_GROUP_JOIN,

    /** Signals that a user leaved a group. Provides an @ref ipd_hi5_group_leave structure with the
        ID of the group. */
    PACE2_ADVANCED_HI5_GROUP_LEAVE,

    /** Signals that the user add a new thread/topic to a group. Provides an @ref ipd_hi5_thread
        structure with information about the new thread/topic. */
    PACE2_ADVANCED_HI5_GROUP_ADD_TOPIC,

    /** Signals that a user add a message to a thread/topic of a group. Provides an @ref ipd_hi5_thread
        structure with information of the thread and the content of the message. */
    PACE2_ADVANCED_HI5_GROUP_ADD_MESSAGE,

    /** Provides an @ref ipd_hi5_thread_list structure which contains a list of threads/topics of
        a group. */
    PACE2_ADVANCED_HI5_GROUP_LIST_THREADS,

    /** Signals that a user visit a profile, its own or another. Provides an @ref ipd_hi5_visit_profile
        structure with information about the user and the visited profile. */
    PACE2_ADVANCED_HI5_VISIT_PROFILE,

    /** Signals that a user visit a group. Provides an @ref ipd_hi5_group_visit structure with information
        about the user and the group. */
    PACE2_ADVANCED_HI5_GROUP_VISIT,

    /** Signals a user search. Provides an @ref ipd_hi5_browse_search structure with an @ref ipd_hi5_filter
        * structure and the request line. */
    PACE2_ADVANCED_HI5_BROWSE_SEARCH,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_HI5_EVENT_TYPES

} PACE2_advanced_hi5_event_type;

/* Event Structures *************************************************************/

/**
 * Structure describing the generic Hi5 event.
 */
struct ipd_hi5_generic {

    /** The correlating http transfer ID. */
    u64 http_transaction_id;

    /** The correlating http session ID. */
    u64 http_session_id;

    /** The main user of this Hi5 session. */
    struct ipd_hi5_user * user;
};

/**
 * Structure describing a Hi5 user.
 *
 * Only the ID of the Hi5 user is always provided. Full name is only provided if available.
 */
struct ipd_hi5_user  {

    /** The Hi5 ID of the user. */
    u64 id;

    /** The full name of the Hi5 user. */
    struct ipd_charset_string name;

    /** The url of user's profile photo. Only used in the @ref PACE2_ADVANCED_HI5_VIEW_FRIENDS event. */
    struct ipd_charset_string profile_photo_url;

    /** The url of the users profile. */
    struct ipd_charset_string profile_url;
};

/** Structure of a status update event. */
struct ipd_hi5_status_update {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** Text of the status update message. */
    struct ipd_charset_string status_message;
};

/** Structure of update entry comments. */
struct ipd_hi5_recent_update_entry_comment {

    /** The author of the comment. */
    struct ipd_hi5_user author;

    /** The comment text. */
    struct ipd_charset_string comment;
};

/** Structure for an entry for the recent update list. */
struct ipd_hi5_recent_update_entry {

    /** The author of the update. */
    struct ipd_hi5_user author;

    /** The type of the update as used by Hi5 internally. */
    struct ipd_charset_string type;

    /** The text associated with the update, only if the value is given. */
    struct ipd_charset_string content;

    /** The list length of updates. */
    u32 comment_list_length;

    /** The list of comments. */
    struct ipd_hi5_recent_update_entry_comment * comment_list;

    /** The photo count that belongs to the update. */
    u32 photo_list_length;

    /** The photo URLs of the update. */
    struct ipd_charset_string * photo_list;
};

/** Structure of the recent updates messages. */
struct ipd_hi5_recent_updates {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** Owner of the profile. */
    struct ipd_hi5_user owner;

    /** An array that contains the updates. */
    struct ipd_hi5_recent_update_entry * update_list;

    /** The number of entries in the update message list. */
    u32 update_array_length;
};

/** Structure of the visit profile event. */
struct ipd_hi5_visit_profile {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** Owner of the profile. */
    struct ipd_hi5_user owner;

    /** The URL of the visited profile. */
    struct ipd_charset_string profile_url;

    /** Sex of the profile owner. */
    struct ipd_charset_string sex;

    /** The age of the profile owner. */
    struct ipd_charset_string age;

    /** The birthday of the profile owner. */
    struct ipd_charset_string birthday;

    /** The location resp. the country. */
    struct ipd_charset_string location;

    /** The name of the city. */
    struct ipd_charset_string city;

    /** The URL of profile owners's profile photo. */
    struct ipd_charset_string profile_photo_url;
};

/** Structure of a Hi5 comment. */
struct ipd_hi5_comment {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** Owner of the profile. */
    struct ipd_hi5_user profile_owner;

    /** Text of the comment. */
    struct ipd_charset_string comment;
};

/** Structure of a friend search event if the friend was searched by his name. */
struct ipd_hi5_friend_search {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** The name of the friend. */
    struct ipd_charset_string search_text;
};

/** Structure of a friend search event if the friend was searched by his name. Obsolete since 05.2012. */
struct ipd_hi5_friend_search_by_name {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** The name of the friend. */
    struct ipd_charset_string name;
};

/** Structure of a friend search event if the friends was searched by their email addresses.
 *  Obsolete since 05.2012. */
struct ipd_hi5_friend_search_by_email {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** An array the contains all email addresses. */
    struct ipd_charset_string * email_list;

    /** The number of elements in the email list. */
    u32 email_array_length;
};

/** Enumeration of answers that are possible for friendship requests. */
enum ipd_hi5_friendship_request_answer {

    /** Accept the friendship request. */
    IPD_HI5_FRIENDSHIP_REQUEST_ANSWER_ACCEPT,

    /** Reject the request for friendship. */
    IPD_HI5_FRIENDSHIP_REQUEST_ANSWER_REJECT,
};

/** Enumeration of friendship request types. */
enum ipd_hi5_friendship_request_type {

    /** Accept the friendship request. */
    IPD_HI5_FRIENDSHIP_REQUEST_TYPE_REQUEST,

    /** Reject the request for friendship. */
    IPD_HI5_FRIENDSHIP_REQUEST_TYPE_RESPONSE,
};

/** Structure of a Hi5 friendship request. */
struct ipd_hi5_friendship_request {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** The type of friendship request. */
    enum ipd_hi5_friendship_request_type type;

    /** The ID of the new friend. */
    u64 friend_id;

    /** The answer to the friendship request. The request can be accepted or rejected.
        It is only available if type is @ref IPD_HI5_FRIENDSHIP_REQUEST_TYPE_RESPONSE. */
    enum ipd_hi5_friendship_request_answer answer;
};

/** Structure of an event to remove a friend from the list of friends. */
struct ipd_hi5_friend_remove {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** The ID of the friend to remove. */
    u64 friend_id;
};

/** Structure of the list of friends that can be seen if the user browse any friend list. */
struct ipd_hi5_view_friends {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** Owner of the profile. */
    struct ipd_hi5_user owner;

    /** An array containing all users that will be shown. */
    struct ipd_hi5_user * friend_list;

    /** The number of entries in the list of friends. */
    u32 friend_array_length;
};

/** Structure of a Hi5 message thread between two users. */
struct ipd_hi5_message_thread {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** An array containing the messages in the thread as shown to the user. */
    struct ipd_hi5_message * message_list;

    /** The number of elements in the message list. */
    u32 message_array_length;
};

/** Structure of a Hi5 message. */
struct ipd_hi5_message {

    /** The generic Hi5 event description.
     It is not used if the message is part of a message list of a ipd_hi5_mailbox structure. */
    struct ipd_hi5_generic generic;

    /** The sender of the message. */
    struct ipd_hi5_user sender;

    /** The subject of the message. */
    struct ipd_charset_string subject;

    /** An array of all recipients of the message */
    struct ipd_hi5_user * recipients;

    /** The number of elements in the recipients list. */
    u32 recipients_array_length;

    /** The date of the message as shown to the user. */
    struct ipd_charset_string date;

    /** The timestamp as submitted with the data. */
    u64 timestamp;

    /** The message_id as submitted with the data. Obsolete since 05.2012. */
    u64 message_id;

    /** The text of the message. It is only provided if available. */
    struct ipd_charset_string text;
};

/** Enumeration of the mailbox types. */
enum ipd_hi5_mailbox_type {
    IPD_HI5_MAILBOX_TYPE_INBOX,
    IPD_HI5_MAILBOX_TYPE_BULK,
    IPD_HI5_MAILBOX_TYPE_SENT,
};

/** Structure of a Hi5 mailbox. */
struct ipd_hi5_mailbox {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** The type of mailbox. */
    enum ipd_hi5_mailbox_type type;

    /** An array containing the messages in the mailbox as shown to the user. It is only provided if available. */
    struct ipd_hi5_message * message_list;

    /** The number of elements in the message list. */
    u32 message_array_length;
};

/** Structure of a group create event. */
struct ipd_hi5_group_create {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** The name of the new group. */
    struct ipd_charset_string group_name;

    /** The ID of the chosen category. */
    u32 category_id;

    /** The text to describe the topic of the new group. */
    struct ipd_charset_string short_description;

    /** The text to describe the topic of the new group that can contain HTML-tags - obsolete. */
    struct ipd_charset_string html_description;

    /** A shortcut of the language e.g. 'de', 'en', 'it', etc.. */
    struct ipd_charset_string language;

    /** The ID of the country. */
    long long int country;

    /** The name of the city. */
    struct ipd_charset_string city;

    /** The access policy of the group - obsolete. */
    u32 access_policy;

    /** Notify abuse report: 0 no, 1 yes - obsolete. */
    u8 notify_abuse_report;

    /** Notify join requests: 0 no, 1 yes - obsolete. */
    u8 notify_join_request;

    /** The access policy like setting to make the group private: 1 = group is private, 0 = group is public. */
    u8 is_private;
};

/** Structure of a group joint event. */
struct ipd_hi5_group_join {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** The ID of the group to join. */
    u64 group_id;
};

/** Structure of a group leave event. */
struct ipd_hi5_group_leave {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** The ID of the group to leave. **/
    u64 group_id;
};

/** Structure of a group visit event. */
struct ipd_hi5_group_visit {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** The ID of the group. */
    u64 group_id;

    /** The name of the group. */
    struct ipd_charset_string group_name;
};

/** Structure of a message for a thread of a group. */
struct ipd_hi5_thread {

    /** The generic Hi5 event description. Not used if this structure is part of ipd_hi5_thread_list. */
    struct ipd_hi5_generic generic;

    /** The ID of the thread. */
    u64 thread_id;

    /** The ID of the group. */
    u64 group_id;

    /** The subject/title of the message. */
    struct ipd_charset_string subject;

    /** The text of the message. */
    struct ipd_charset_string message;
};

/** Structure of a list of thread messages. */
struct ipd_hi5_thread_list {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** An array of threads that is shown to the user if he/she visit a group. */
    struct ipd_hi5_thread * thread_list;

    /** The number of elements in the thread list. */
    u32 thread_array_length;
};

struct ipd_hi5_logon {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** Email address of the user, obsolete. */
    struct ipd_charset_string email_address;

    /** Logon password, obsolete. */
    struct ipd_charset_string password;
};

/**
 * @brief Enumeration for the gender in the browse filter.
 */
enum hi5_filter_gender {
    HI5_GENDER_FEMALE,
    HI5_GENDER_MALE,
    HI5_GENDER_BOTH
};

/**
 * @brief The ipd_hi5_filter struct represents the filter of a browse search.
 */
struct ipd_hi5_filter {

    /** The selected gender filter settings. Could be Female, Male or Both. */
    enum hi5_filter_gender gender;

    /** The selected minimum age. Between 18 - 99. */
    u8 age_min;

    /** The selected maximum age. Between 18 - 99. */
    u8 age_max;

    /** The selected country code. */
    struct ipd_charset_string country;

    /** The selected city. */
    struct ipd_charset_string city;

    /** The selected search radius from the selected city. */
    u8 proximity;

    /** The selected sexual orientations. The corresponding enumeration to interpret the value
     * is hi5_filter_sexual_orientation. */
    u32 sexual_orientation;

    /** The selected relationship status. The corresponding enumeration to interpret the value
     * is hi5_filter_relationship_status. */
    u32 relationship_status;

    /** The selected ethnicity. The corresponding enumeration to interpret the value is. */
    u32 ethnicity;
};

/**
 * @brief Structure of a people search event with a filter.
 */
struct ipd_hi5_browse_search {

    /** The generic Hi5 event description. */
    struct ipd_hi5_generic generic;

    /** The selected browse filter. */
    struct ipd_hi5_filter filter;

    /** The undecoded request filter line. */
    struct ipd_charset_string filter_request_line;
};

/**
 * @brief Enumeration for the sexual orientation in the browse filter.
 */
enum hi5_filter_sexual_orientation {
    HI5_SEXUAL_ORIENTATION_STRAIGHT  = 1,
    HI5_SEXUAL_ORIENTATION_GAY       = 2,
    HI5_SEXUAL_ORIENTATION_BI        = 4,
    HI5_SEXUAL_ORIENTATION_NO_ANSWER = 8
};

/**
 * @brief Enumeration for the relationship status in the browse filter.
 */
enum hi5_filter_relationship_status {
    HI5_RELATIONSHIP_STATUS_SINGLE            = 1,
    HI5_RELATIONSHIP_STATUS_DATING            = 2,
    HI5_RELATIONSHIP_STATUS_IN_A_RELATIONSHIP = 4,
    HI5_RELATIONSHIP_STATUS_ENGAGED           = 8,
    HI5_RELATIONSHIP_STATUS_MARRIED           = 16,
    HI5_RELATIONSHIP_STATUS_IT_S_COMPLICATED  = 32,
    HI5_RELATIONSHIP_STATUS_OPEN_RELATIONSHIP = 128,
    HI5_RELATIONSHIP_STATUS_SEPARATED         = 256,
    HI5_RELATIONSHIP_STATUS_DIVORCED          = 512,
    HI5_RELATIONSHIP_STATUS_WIDOWED           = 1024
};

/**
 * @brief Enumeration for the ethnicity in the browse filter.
 */
enum hi5_filter_ethnicity {
    HI5_ETHNICITY_BLACK            = 1,
    HI5_ETHNICITY_ASIAN            = 2,
    HI5_ETHNICITY_CAUCASIAN_WHITE  = 4,
    HI5_ETHNICITY_EAST_INDIAN      = 8,
    HI5_ETHNICITY_HISPANIC_LATINO  = 16,
    HI5_ETHNICITY_MIDDLE_EASTERN   = 32,
    HI5_ETHNICITY_NATIVE_AMERICAN  = 64,
    HI5_ETHNICITY_PACIFIC_ISLANDER = 128,
    HI5_ETHNICITY_OTHER            = 256
};

/**
 * @brief Data structure for advanced Hi5 events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_HI5_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** HI5 event subtype. */
    PACE2_advanced_hi5_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_HI5_LOGON event. */
        struct ipd_hi5_logon logon;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_LOGOFF event. */
        struct ipd_hi5_generic logoff;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_STATUS_UPDATE event. */
        struct ipd_hi5_status_update status_update;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_RECENT_UPDATES event. */
        struct ipd_hi5_recent_updates recent_updates;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_COMMENT_UPDATE event. */
        struct ipd_hi5_comment comment_update;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_COMMENT_PROFILE event. */
        struct ipd_hi5_comment comment_profile;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_COMMENT_STATUS event. */
        struct ipd_hi5_comment comment_status;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_FRIEND_SEARCH_BY_NAME event. */
        struct ipd_hi5_friend_search_by_name friend_search_by_name;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_FRIEND_SEARCH_BY_EMAIL event. */
        struct ipd_hi5_friend_search_by_email friend_search_by_email;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_FRIEND_SEARCH event. */
        struct ipd_hi5_friend_search friend_search;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_FRIENDSHIP_REQUEST event. */
        struct ipd_hi5_friendship_request friendship_request;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_FRIEND_REMOVE event. */
        struct ipd_hi5_friend_remove friend_remove;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_VIEW_FRIENDS event. */
        struct ipd_hi5_view_friends view_friends;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_MAILBOX event. */
        struct ipd_hi5_mailbox mailbox;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_MAIL_SEND event. */
        struct ipd_hi5_message mail_send;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_MAIL_VIEW event. */
        struct ipd_hi5_message mail_view;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_MAIL_RECEIVE event. */
        struct ipd_hi5_message mail_receive;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_MAILTHREAD_VIEW event. */
        struct ipd_hi5_message_thread mailthread_view;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_GROUP_CREATE event. */
        struct ipd_hi5_group_create group_create;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_GROUP_JOIN event. */
        struct ipd_hi5_group_join group_join;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_GROUP_LEAVE event. */
        struct ipd_hi5_group_leave group_leave;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_GROUP_ADD_TOPIC event. */
        struct ipd_hi5_thread group_add_topic;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_GROUP_ADD_MESSAGE event. */
        struct ipd_hi5_thread group_add_message;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_GROUP_LIST_THREADS event. */
        struct ipd_hi5_thread_list list;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_VISIT_PROFILE event. */
        struct ipd_hi5_visit_profile visit_profile;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_GROUP_VISIT event. */
        struct ipd_hi5_group_visit group_visit;

        /** Contains data for a @ref PACE2_ADVANCED_HI5_BROWSE_SEARCH event. */
        struct ipd_hi5_browse_search browse_search;

    } event_data;

} PACE2_advanced_hi5_event;

#endif /* PACE_KERNEL_LIB */

#endif /* HI5_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
