/** @file      http2_events.h
  * @brief     HTTP/2 event definitions.
  * @details   @copydoc doc_http2
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef _HTTP2_EVENTS_H_
#define _HTTP2_EVENTS_H_

#include "common_event_defs.h"

/**
 * @brief Public define for the assigned metadata group of the basic HTTP/2 event.
 */
#define PACE2_BASIC_HTTP2_EVENT_GROUP PACE2_BASIC_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned metadata group of the advanced HTTP/2 event.
 */
#define PACE2_ADVANCED_HTTP2_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/**
 * Types of supported metadata in a HTTP/2 HEADERS frame.
 */
typedef enum {
    PACE2_HTTP2_AUTHORITY,
    PACE2_HTTP2_PATH,
    PACE2_HTTP2_USER_AGENT,
    PACE2_HTTP2_REFERER,

    PACE2_NUMBER_OF_HTTP2_HEADER_TYPES
} PACE2_basic_HTTP2_header_field;

/**
 * @brief Types for advanced HTTP/2 events.
 */
typedef enum {
    PACE2_ADVANCED_HTTP2_NEW_CONNECTION, /**< First event of a HTTP/2 flow.
                                              This Event signals, that the flow of the current packet will be
                                              handled by the HTTP/2 decoder. */
    PACE2_ADVANCED_HTTP2_END_OF_FLOW,    /**< Signals, that the processing of the flow of the current packet is done.
                                              No further events will be thrown. It is guaranteed that this event or the
                                              event PACE2_ADVANCED_HTTP2_INVALID is always the last event for a flow. */
    PACE2_ADVANCED_HTTP2_INVALID,        /**< Signals, that this HTTP/2 flow is invalid. */
    PACE2_ADVANCED_HTTP2_OTHER,          /**< Signals that an other header frame was found and does not belong to
                                              the static table definition of Appendix A in RFC 7541
                                              (https://tools.ietf.org/html/rfc7541#appendix-A) */
    PACE2_ADVANCED_HTTP2_DATA,           /**< Signals a DATA frame. */
    PACE2_ADVANCED_HTTP2_SETTINGS,       /**< Signals a SETTINGS frame. */
    PACE2_ADVANCED_HTTP2_WINDOW_UPDATE,  /**< Signals a WINDOW UPDATE frame. */
    PACE2_ADVANCED_HTTP2_PRIORITY,       /**< Signals a PRIORITY frame. */
    PACE2_ADVANCED_HTTP2_RST_STREAM,     /**< Signals a RST_STREAM frame. */
    PACE2_ADVANCED_HTTP2_PUSH_PROMISE,   /**< Signals a PUSH_PROMISE frame. */
    PACE2_ADVANCED_HTTP2_PING,           /**< Signals a PING frame. */
    PACE2_ADVANCED_HTTP2_GOAWAY,         /**< Signals a GOAWAY frame. */
    PACE2_ADVANCED_HTTP2_CONTINUATION,   /**< Signals a CONTINUATION frame. */
    PACE2_ADVANCED_HTTP2_AUTHORITY,      /**< Signals the HEADER name :authority. */
    PACE2_ADVANCED_HTTP2_METHOD,         /**< Signals the HEADER name :method. */
    PACE2_ADVANCED_HTTP2_PATH,           /**< Signals the HEADER name :path. */
    PACE2_ADVANCED_HTTP2_SCHEME,         /**< Signals the HEADER name :scheme. */
    PACE2_ADVANCED_HTTP2_STATUS,         /**< Signals the HEADER name :status. */
    PACE2_ADVANCED_HTTP2_ACCEPT_CHARSET, /**< Signals the HEADER name accept-charset. */
    PACE2_ADVANCED_HTTP2_ACCEPT_ENCODING,/**< Signals the HEADER name accept-encoding. */
    PACE2_ADVANCED_HTTP2_ACCEPT_LANGUAGE,/**< Signals the HEADER name accept-language. */
    PACE2_ADVANCED_HTTP2_ACCEPT_RANGES,  /**< Signals the HEADER name accept-ranges. */
    PACE2_ADVANCED_HTTP2_ACCEPT,         /**< Signals the HEADER name accept. */
    PACE2_ADVANCED_HTTP2_ACCESS_CONTROL_ALLOW_ORIGIN,/**< Signals the HEADER name access-control-allow-origin. */
    PACE2_ADVANCED_HTTP2_AGE,            /**< Signals the HEADER name age. */
    PACE2_ADVANCED_HTTP2_ALLOW,          /**< Signals the HEADER name allow. */
    PACE2_ADVANCED_HTTP2_AUTHORIZATION,  /**< Signals the HEADER name authorization. */
    PACE2_ADVANCED_HTTP2_CACHE_CONTROL,  /**< Signals the HEADER name control. */
    PACE2_ADVANCED_HTTP2_CONTENT_DISPOSITION,/**< Signals the HEADER name content-disposition. */
    PACE2_ADVANCED_HTTP2_CONTENT_ENCODING,/**< Signals the HEADER name content-encoding. */
    PACE2_ADVANCED_HTTP2_CONTENT_LANGUAGE,/**< Signals the HEADER name content-language. */
    PACE2_ADVANCED_HTTP2_CONTENT_LENGTH,/**< Signals the HEADER name content-length. */
    PACE2_ADVANCED_HTTP2_CONTENT_LOCATION,/**< Signals the HEADER name content-location. */
    PACE2_ADVANCED_HTTP2_CONTENT_RANGE,  /**< Signals the HEADER name content-range. */
    PACE2_ADVANCED_HTTP2_CONTENT_TYPE,   /**< Signals the HEADER name content-type. */
    PACE2_ADVANCED_HTTP2_COOKIE,         /**< Signals the HEADER name cookie. */
    PACE2_ADVANCED_HTTP2_DATE,           /**< Signals the HEADER name date. */
    PACE2_ADVANCED_HTTP2_ETAG,           /**< Signals the HEADER name etag. */
    PACE2_ADVANCED_HTTP2_EXPECT,         /**< Signals the HEADER name expect. */
    PACE2_ADVANCED_HTTP2_EXPIRES,        /**< Signals the HEADER name expires. */
    PACE2_ADVANCED_HTTP2_FROM,           /**< Signals the HEADER name from. */
    PACE2_ADVANCED_HTTP2_HOST,           /**< Signals the HEADER name host. */
    PACE2_ADVANCED_HTTP2_IF_MATCH,       /**< Signals the HEADER name if-match. */
    PACE2_ADVANCED_HTTP2_IF_MODIFIED_SINCE,/**< Signals the HEADER name if-modified-since. */
    PACE2_ADVANCED_HTTP2_IF_NONE_MATCH,  /**< Signals the HEADER name if-none-match. */
    PACE2_ADVANCED_HTTP2_IF_RANGE,       /**< Signals the HEADER name if-range. */
    PACE2_ADVANCED_HTTP2_IF_UNMODIFIED_SINCE,/**< Signals the HEADER name if-unmodified-since. */
    PACE2_ADVANCED_HTTP2_LAST_MODIFIED,  /**< Signals the HEADER name last-modified. */
    PACE2_ADVANCED_HTTP2_LINK,           /**< Signals the HEADER name link. */
    PACE2_ADVANCED_HTTP2_LOCATION,       /**< Signals the HEADER name location. */
    PACE2_ADVANCED_HTTP2_MAX_FORWARDS,   /**< Signals the HEADER name max-forwards. */
    PACE2_ADVANCED_HTTP2_PROXY_AUTHENTICATE,/**< Signals the HEADER name proxy-authenticate. */
    PACE2_ADVANCED_HTTP2_PROXY_AUTHORIZATION,/**< Signals the HEADER name proxy-authorization. */
    PACE2_ADVANCED_HTTP2_RANGE,         /**< Signals the HEADER name range. */
    PACE2_ADVANCED_HTTP2_REFERER,       /**< Signals the HEADER name referer. */
    PACE2_ADVANCED_HTTP2_REFRESH,       /**< Signals the HEADEr name refresh. */
    PACE2_ADVANCED_HTTP2_RETRY_AFTER,   /**< Signals the HEADER name retry-after. */
    PACE2_ADVANCED_HTTP2_SERVER,        /**< Signals the HEADER name server. */
    PACE2_ADVANCED_HTTP2_SET_COOKIE,    /**< Signals the HEADER name set-cookie. */
    PACE2_ADVANCED_HTTP2_STRICT_TRANSPORT_SECURITY,/**< Signals the HEADER name strict-transport-security. */
    PACE2_ADVANCED_HTTP2_TRANSFER_ENCODING,/**< Signals the HEADER name transfer-encoding. */
    PACE2_ADVANCED_HTTP2_USER_AGENT,    /**< Signals the HEADER name user-agent. */
    PACE2_ADVANCED_HTTP2_VARY,          /**< Signals the HEADER name vary. */
    PACE2_ADVANCED_HTTP2_VIA,           /**< Signals the HEADER name via. */
    PACE2_ADVANCED_HTTP2_WWW_AUTHENTICATE,/**< Signals the HEADER name www-authenticate. */

    PACE2_NUMBER_OF_ADVANCED_HTTP2_EVENT_TYPES /**< Number of implemented event types. */
} PACE2_advanced_HTTP2_event_type;

/**
 * Data structure describing the basic HTTP/2 event.
 */
typedef struct {
    /**
     * Mandatory event members.
     */
    PACE2_EVENT_COMMON_FIELDS;

    /**
     * Signals which header field values are set for this event.
     */
    PACE2_bitmask meta_data_mask;
    struct {
        /** Header field of type @ref PACE2_basic_HTTP2_header_field.
         *
         * @deprecated Please do not uses this variable any longer. It will be removed in future release of this
         *             library. The header type is implicit in the index of the headers array.*/
        PACE2_DEPRECATED(
            "Please do not uses this variable any longer. It will be removed in future release of this library. The "
            "header type is implicit in the index of the headers array.")
        PACE2_basic_HTTP2_header_field type;
        PACE2_byte_buffer value;                     /**< Header field data. */
    } headers[PACE2_NUMBER_OF_HTTP2_HEADER_TYPES];   /**< Data structure for a single HTTP/2 header.
                                                          Contains content for each available header line of type
                                                          @ref PACE2_basic_HTTP2_header_field*/
} PACE2_basic_HTTP2_event;

#ifndef PACE_KERNEL_LIB

/*
 * The basic HTTP/2 frame format is defined in https://tools.ietf.org/html/rfc7540#page-12
 *  +-----------------------------------------------+
    |                 Length (24)                   |
    +---------------+---------------+---------------+
    |   Type (8)    |   Flags (8)   |
    +-+-------------+---------------+-------------------------------+
    |R|                 Stream Identifier (31)                      |
    +=+=============================================================+
    |                   Frame Payload (0...)                      ...
    +---------------------------------------------------------------+

 */

struct http2_basic_frame_struct {
    u8 type;                          /**< The 8-bit type of the frame. */
    u8 flags;                         /**< An 8-bit field reserved for boolean flags specific to the frame type. */
    u16 length;                       /**< The length of the frame payload. */
    u32 stream_identifier;            /**< A stream identifier. */
    struct ipd_charset_string data;   /**< The data of the frame. */
};

struct http2_header_frame_line_other {
    struct ipd_charset_string name;    /**< The name of the header line. */
    struct ipd_charset_string value;   /**< The value of the header line. */
};

/**
 * @brief Data structure for advanced HTTP/2 events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_ADVANCED_HTTP2_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_advanced_HTTP2_event_type meta_data_type; /**< HTTP/2 event subtype. */
    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;
    union {
        struct ipd_charset_string authority;      /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_AUTHORITY event. */
        struct ipd_charset_string method;         /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_METHOD event. */
        struct ipd_charset_string path;           /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_PATH event. */
        struct ipd_charset_string scheme;         /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_SCHEME event. */
        struct ipd_charset_string status;         /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_STATUS event. */
        struct ipd_charset_string accept_charset; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_ACCEPT_CHARSET
                                                     event. */
        struct ipd_charset_string accept_encoding; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_ACCEPT_ENCODING
                                                      event. */
        struct ipd_charset_string accept_language; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_ACCEPT_LANGUAGE
                                                      event. */
        struct ipd_charset_string
            accept_ranges;                /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_ACCEPT_RANGES event. */
        struct ipd_charset_string accept; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_ACCEPT event. */
        struct ipd_charset_string access_control_allow_origin; /**< Contains data for a @ref
                                                                  PACE2_ADVANCED_HTTP2_ACCESS_CONTROL_ALLOW_ORIGIN
                                                                  event. */
        struct ipd_charset_string age;   /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_AGE event. */
        struct ipd_charset_string allow; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_ALLOW event. */
        struct ipd_charset_string
            authorization; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_AUTHORIZATION event. */
        struct ipd_charset_string
            cache_control; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_CACHE_CONTROL event. */
        struct ipd_charset_string content_disposition; /**< Contains data for a @ref
                                                          PACE2_ADVANCED_HTTP2_CONTENT_DISPOSITION event. */
        struct ipd_charset_string
            content_encoding; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_CONTENT_ENCODING event. */
        struct ipd_charset_string
            content_language; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_CONTENT_LANGUAGE event. */
        struct ipd_charset_string content_length; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_CONTENT_LENGTH
                                                     event. */
        struct ipd_charset_string
            content_location; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_CONTENT_LOCATION event. */
        struct ipd_charset_string
            content_range; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_CONTENT_RANGE event. */
        struct ipd_charset_string content_type; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_CONTENT_TYPE event.
                                                   */
        struct ipd_charset_string cookie;       /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_COOKIE event. */
        struct ipd_charset_string date;         /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_DATE event. */
        struct ipd_charset_string etag;         /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_ETAG event. */
        struct ipd_charset_string expect;       /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_EXPECT event. */
        struct ipd_charset_string expires;      /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_EXPIRES event. */
        struct ipd_charset_string from;         /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_FROM event. */
        struct ipd_charset_string host;         /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_HOST event. */
        struct ipd_charset_string if_match;     /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_IF_MATCH event. */
        struct ipd_charset_string if_modified_since; /**< Contains data for a @ref
                                                        PACE2_ADVANCED_HTTP2_IF_MODIFIED_SINCE event. */
        struct ipd_charset_string
            if_none_match;                  /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_IF_NONE_MATCH event. */
        struct ipd_charset_string if_range; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_IF_RANGE event. */
        struct ipd_charset_string if_unmodified_since; /**< Contains data for a @ref
                                                          PACE2_ADVANCED_HTTP2_IF_UNMODIFIED_SINCE event. */
        struct ipd_charset_string
            last_modified;                  /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_LAST_MODIFIED event. */
        struct ipd_charset_string link;     /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_LINK event. */
        struct ipd_charset_string location; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_LOCATION event. */
        struct ipd_charset_string max_forwards; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_MAX_FORWARDS event.
                                                   */
        struct ipd_charset_string proxy_authenticate;  /**< Contains data for a @ref
                                                          PACE2_ADVANCED_HTTP2_PROXY_AUTHENTICATE event. */
        struct ipd_charset_string proxy_authorization; /**< Contains data for a @ref
                                                          PACE2_ADVANCED_HTTP2_PROXY_AUTHORIZATION event. */
        struct ipd_charset_string range;       /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_RANGE event. */
        struct ipd_charset_string referer;     /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_REFERER event. */
        struct ipd_charset_string refresh;     /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_REFRESH event. */
        struct ipd_charset_string retry_after; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_RETRY_AFTER event. */
        struct ipd_charset_string server;      /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_SERVER event. */
        struct ipd_charset_string set_cookie;  /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_SET_COOKIE event. */
        struct ipd_charset_string strict_transport_security; /**< Contains data for a @ref
                                                                PACE2_ADVANCED_HTTP2_STRICT_TRANSPORT_SECURITY event. */
        struct ipd_charset_string transfer_encoding;         /**< Contains data for a @ref
                                                                PACE2_ADVANCED_HTTP2_TRANSFER_ENCODING event. */
        struct ipd_charset_string user_agent; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_USER_AGENT event. */
        struct ipd_charset_string vary;       /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_VARY event. */
        struct ipd_charset_string via;        /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_VIA event. */
        struct ipd_charset_string
            www_authenticate; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_WWW_AUTHENTICATE event. */
        struct http2_basic_frame_struct settings; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_SETTINGS event. */
        struct http2_basic_frame_struct data;     /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_DATA event. */
        struct http2_basic_frame_struct
            window_update; /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_WINDOW_UPDATE event. */
        struct http2_basic_frame_struct priority;     /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_PRIORITY event. */
        struct http2_basic_frame_struct rst_stream;     /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_RST_STREAM event. */
        struct http2_basic_frame_struct push_promise;     /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_PUSH_PROMISE event. */
        struct http2_basic_frame_struct ping;     /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_PING event. */
        struct http2_basic_frame_struct goaway;     /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_GOAWAY event. */
        struct http2_basic_frame_struct continuation;     /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_CONTINUATION event. */
        struct http2_header_frame_line_other other;     /**< Contains data for a @ref PACE2_ADVANCED_HTTP2_OTHER event. */
        u8 padding[128]; /**< Ensures ABI compatibility for the following fields if any event has to be increased. */
    } event_data; /**< Named union which contains the data of the sub-event types. */
    u32 stream_identifier; /**< Contains the HTTP/2 stream identifier for all sub events. */
} PACE2_advanced_HTTP2_event;

#endif // PACE_KERNEL_LIB
#endif // _HTTP2_EVENTS_H_
