/** @file      http_events.h
  * @brief     HTTP event definitions.
  * @details   @copydoc doc_http
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef HTTP_EVENTS_H
#define HTTP_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Defines **********************************************************************/

/**
 * @brief Maximum amount of non standard header lines in @ref PACE2_basic_HTTP_request_event or @ref
 * PACE2_basic_HTTP_response_event.
 */
#define PACE2_NUMBER_OF_HTTP_EXTRA_LINES 25

/**
 * @brief Represents the output of a parsed header line. e.g: "Accept-Encoding: gzip, deflate". The @ref key value will
 *        be "Accept_Encoding" and @ref content value will be "gzip, deflate".
 * @note  In cases of blank content entries, the @ref content value has a length of zero. Empty key values are not
 *        possible.
 */
typedef struct {

    /** Type identifier. */
    PACE2_byte_buffer key;

    /** Line content.*/
    PACE2_byte_buffer content;

} PACE2_http_extra_line;

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the basic HTTP request event.
 */
#define PACE2_BASIC_HTTP_REQUEST_EVENT_GROUP PACE2_BASIC_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned metadata group of the basic HTTP response event.
 */
#define PACE2_BASIC_HTTP_RESPONSE_EVENT_GROUP PACE2_BASIC_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned metadata group of the advanced HTTP event.
 */
#define PACE2_ADVANCED_HTTP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned metadata group of the class HTTP event.
 */
#define PACE2_CLASS_HTTP_EVENT_GROUP PACE2_CLASS_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned metadata group of the basic HTTP parsed request event.
 */
#define PACE2_BASIC_HTTP_PARSED_REQUEST_EVENT_GROUP PACE2_BASIC_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced HTTP events.
 */
typedef enum {

    /** Provides a struct ipd_5tuple with information of the 5-tuple describing the transport-layer connection. */
    PACE2_ADVANCED_HTTP_5TUPLE,

    /// request data

    /** The entire request header, including the request-line. */
    PACE2_ADVANCED_HTTP_REQUEST_HEADER,

    /** The method string of the HTTP request like "GET" or "POST". */
    PACE2_ADVANCED_HTTP_REQUEST_METHOD,

    /** The path of the requested file on the server like "/stuff/index.html". */
    PACE2_ADVANCED_HTTP_REQUEST_PATH,

    /** The version string of the HTTP request like "HTTP/1.1". */
    PACE2_ADVANCED_HTTP_REQUEST_VERSION,

    /** The host-name of the HTTP server that the request is sent to like "www.notreallythere.net"; @c data points
     * to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_REQUEST_HOST,

    /** The useragent string of the program sending the HTTP request like
     * "Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.2.2pre) Gecko/20100130 Ubuntu/9.10 (karmic) Namoroka/3.6.2pre";
     * @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_REQUEST_USERAGENT,

    /** The referrer string that i.e. is set by web browsers to the http page that the request originated from like
     * "http://www.notreallythere.net/"; @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_REQUEST_REFERER,

    /** The string describing the requested content type like "text/html"; @c data points to
     * a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_REQUEST_CONTENTTYPE,

    /** The cookie(s) sent with the request; @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_REQUEST_COOKIE,

    /** A string describing the requested range if only a part of the requested file should be transferred;
     * @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_REQUEST_RANGE,

    /** The length of the request data as a string; @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_REQUEST_CONTENT_LENGTH,

    /** This event is issued for all HTTP request header lines that no special events exist for.
     * The string contains the complete header line like "Connection: keep-alive". */
    PACE2_ADVANCED_HTTP_REQUEST_UNKNOWN,

    /** This event is issued for each decoded piece of HTTP request payload data. */
    PACE2_ADVANCED_HTTP_REQUEST_PAYLOAD,

    /** This event indicates that the decoding of the HTTP request completed. It does not provide any data. */
    PACE2_ADVANCED_HTTP_REQUEST_DONE,

    /// response data

    /** The entire response header, including the status-line. */
    PACE2_ADVANCED_HTTP_RESPONSE_HEADER,

    /** The version string of the HTTP response like "HTTP/1.1". */
    PACE2_ADVANCED_HTTP_RESPONSE_VERSION,

    /** The HTTP response code as a string like "200". */
    PACE2_ADVANCED_HTTP_RESPONSE_CODE,

    /** The HTTP response reason phrase string like "OK". */
    PACE2_ADVANCED_HTTP_RESPONSE_REASON_PHRASE,

    /** The length of the response data as a string; @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_RESPONSE_CONTENT_LENGTH,

    /** A string describing the range inside the requested document that is actually delivered with the HTTP response;
     * @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_RESPONSE_CONTENT_RANGE,

    /** The content type of the response data like "image/gif"; @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_RESPONSE_CONTENTTYPE,

    /** The encoding of the actually transmitted response data like "gzip";
     * @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_RESPONSE_CONTENT_ENCODING,

    /** The transfer encoding of the actually transmitted response data like "chunked";
     * @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_RESPONSE_TRANSFER_ENCODING,

    /** The connection type sent with the HTTP response like "keep-alive";
     * @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_RESPONSE_CONNECTION,

    /** A string with cookie(s) to be set with the HTTP response;
     * @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_RESPONSE_SET_COOKIE,

    /** The new location of the requested document if it was moved permanently or temporarily (response codes 3xx)
     * like "http://www.notreallythere.net/new/index.html"; @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_RESPONSE_LOCATION,

    /** This event is issued for all HTTP response header lines that no special events exist for.
     * The string contains the complete header line like "Server: Apache-Coyote/1.1". */
    PACE2_ADVANCED_HTTP_RESPONSE_UNKNOWN,

    /** This event is issued for each decoded piece of HTTP response payload data. */
    PACE2_ADVANCED_HTTP_RESPONSE_PAYLOAD,

    /** This event indicates that the decoding of the HTTP response completed.  It does not provide any data. */
    PACE2_ADVANCED_HTTP_RESPONSE_DONE,

    /// end of the flow

    /** This event indicates that the decoding of the HTTP connection cannot be continued.
     * It provides a char const * to a message describing the failure reason. */
    PACE2_ADVANCED_HTTP_INVALID,

    /** This event indicates that the HTTP connection ended. It does not provide any data. */
    PACE2_ADVANCED_HTTP_END_OF_FLOW,

    /** This event is sent to indicate that none of the currently configured HTTP-postprocessing
     * decoders needs this flow. */
    PACE2_ADVANCED_HTTP_NO_POSTPROC,

    /** Context was lost when the decoder tried to decode flow with holes. */
    PACE2_ADVANCED_HTTP_CONTEXT_LOST,

    /** Hole of the given length was observed in the request. */
    PACE2_ADVANCED_HTTP_REQUEST_HOLE,

    /** Hole of the given length was observed in the response. */
    PACE2_ADVANCED_HTTP_RESPONSE_HOLE,
    PACE2_ADVANCED_HTTP_INVALID_VERBOSE,

    /** The transfer encoding of the currently transmitted response data (e.g. "chunked").
     * @c data points to a struct ipd_http_header_field. */
    PACE2_ADVANCED_HTTP_REQUEST_TRANSFER_ENCODING,

    /** This event indicates the presence of HTTP content which does not conform to RFCs. The appropriate config option
     * must be set to enable this event. */
    PACE2_ADVANCED_HTTP_NONCONFORMING_CONTENT,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_HTTP_EVENT_TYPES

} PACE2_advanced_HTTP_event_type;

/**
 * @brief Types for HTTP class events.
 */
typedef enum {

    /** Main HTTP session structure, each HTTP connection session has this as initial event.
     * Returns an ipd_class_http_session_struct *. */
    PACE2_CLASS_HTTP_SESSION,

    /** Start of a transaction. Returns an ipd_class_http_transaction_struct *. */
    PACE2_CLASS_HTTP_TRANSACTION,

    /** Request for data transfer. Returns an ipd_class_http_transfer_payload_struct *. */
    PACE2_CLASS_HTTP_REQUEST_DATA_TRANSFER,

    /** Response for data transfer. Returns an ipd_class_http_transfer_payload_struct *. */
    PACE2_CLASS_HTTP_RESPONSE_DATA_TRANSFER,

    /** End of a transaction. Returns an ipd_class_http_transaction_struct *. */
    PACE2_CLASS_HTTP_TRANSACTION_DONE,

    /** Every HTTP connection session has this as end event. Returns an ipd_class_http_session_struct *. */
    PACE2_CLASS_HTTP_SESSION_DONE,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_CLASS_HTTP_EVENT_TYPES

} PACE2_class_HTTP_event_type;

/**
 * @brief Meta data flags of basic HTTP request events.
 */
typedef enum {

    /** Request method part. */
    PACE2_HTTP_REQUEST_METHOD = 0,

    /** Request URI part. */
    PACE2_HTTP_REQUEST_URI,

    /** Request host line. */
    PACE2_HTTP_HOST,

    /** Request user-agent line. */
    PACE2_HTTP_USER_AGENT,

    /** Request referer line. */
    PACE2_HTTP_REFERER,

    /** Request content type line. */
    PACE2_HTTP_CONTENT_TYPE,

    /** Request content length line. */
    PACE2_HTTP_CONTENT_LENGTH,

    /** Request accept line. */
    PACE2_HTTP_ACCEPT,

    /** Request origin line. */
    PACE2_HTTP_ORIGIN,

    /** Request cookie line. */
    PACE2_HTTP_COOKIE,

    /** Request soap action line. */
    PACE2_HTTP_SOAP_ACTION,

    /** Request x-forwarded-for line. */
    PACE2_HTTP_X_FORWARDED_FOR,

    /** Request x-session-type line. */
    PACE2_HTTP_X_SESSION_TYPE,

    /** Request x-stream-type line. */
    PACE2_HTTP_X_STREAM_TYPE,

    /** Request x-online-host line. */
    PACE2_HTTP_X_ONLINE_HOST,

    /** Request x-requested-with line. */
    PACE2_HTTP_X_REQUESTED_WITH,

    /** Request connection line. */
    PACE2_HTTP_CONNECTION,

    /** Number of implemented HTTP request line flags. */
    PACE2_NUMBER_OF_HTTP_REQUEST_FLAGS

} PACE2_basic_HTTP_request_line;

/**
 * @brief List of HTTP error messages with their corresponding enum values. These messages are thrown by the
 * @ref PACE2_ADVANCED_HTTP_INVALID_VERBOSE event.
 */
#define PACE2_ADVANCED_HTTP_INVALID_STRING_LIST(X) \
    X(0,  PACE2_ADVANCED_HTTP_INVALID_UNKNOWN, \
            "Unknown")\
    X(1,  PACE2_ADVANCED_HTTP_INVALID_UNFINISHED, \
            "Unfinished request but there is not enough data.")\
    X(2,  PACE2_ADVANCED_HTTP_INVALID_TOO_LITTLE_DATA_CLIENT, \
            "too little data in client-side buffer")\
    X(3,  PACE2_ADVANCED_HTTP_INVALID_NO_REQUEST_TYPE, \
            "No request type found - exiting")\
    X(4,  PACE2_ADVANCED_HTTP_INVALID_REQUEST_METHOD_LENGTH, \
            "Invalid request method length")\
    X(5,  PACE2_ADVANCED_HTTP_INVALID_REQUEST_METHOD_CHARACTERS, \
            "Invalid request method characters")\
    X(6,  PACE2_ADVANCED_HTTP_INVALID_NO_REQUEST_HTTP_VERSION, \
            "No request HTTP version found - exiting")\
    X(7,  PACE2_ADVANCED_HTTP_INVALID_REQUEST_LINE, \
            "Invalid request line")\
    X(8,  PACE2_ADVANCED_HTTP_INVALID_REQUEST_LINE_NO_VERSION, \
            "Invalid request line (could not find version)")\
    X(9,  PACE2_ADVANCED_HTTP_INVALID_CHARACTER_REQUEST_CONTENT_LENGTH, \
            "invalid character in request content length")\
    X(10,  PACE2_ADVANCED_HTTP_INVALID_REQUEST_CONTENT_LENGTH_NUMBER_TOO_LONG, \
            "request content length number too long (see debug level 3)")\
    X(11,  PACE2_ADVANCED_HTTP_INVALID_REQUEST_HEADER_LENGTH_LIMIT_EXCEEDED, \
            "Request header length limit exceeded.")\
    X(12,  PACE2_ADVANCED_HTTP_INVALID_NO_CRLF_OR_LF, \
            "No CRLF or LF after header found. Parser probably screwed up.")\
    X(13,  PACE2_ADVANCED_HTTP_INVALID_NO_HEADER_START_ASYMMETRIC_RESPONSE, \
            "no valid response header start in asymmetric HTTP response")\
    X(14,  PACE2_ADVANCED_HTTP_INVALID_TOO_LITTLE_DATA_SERVER, \
            "too little data in server-side buffer")\
    X(15,  PACE2_ADVANCED_HTTP_INVALID_RESPONSE_NO_VERSION, \
            "Invalid response (could not find version)")\
    X(16,  PACE2_ADVANCED_HTTP_INVALID_UNKNOWN_RESPONSE_TYPE, \
            "unknown response type")\
    X(17,  PACE2_ADVANCED_HTTP_INVALID_RESPONSE_CODE_LENGTH, \
            "invalid response code len (see debug level 3)")\
    X(18,  PACE2_ADVANCED_HTTP_INVALID_CHARACTER_RESPONSE_CONTENT_LENGTH, \
            "invalid character in response content length")\
    X(19,  PACE2_ADVANCED_HTTP_INVALID_RESPONSE_CONTENT_LENGTH_NUMBER_TOO_LONG, \
            "response content length number too long (see debug level 3)")\
    X(20,  PACE2_ADVANCED_HTTP_INVALID_RESPONSE_HEADER_LENGTH_LIMIT_EXCEEDED, \
            "Response header length limit exceeded.")\
    X(21,  PACE2_ADVANCED_HTTP_INVALID_RESPONSE_HEADER_LENGTH_LIMIT_REACHED, \
            "Response header length limit reached.")\
    X(22,  PACE2_ADVANCED_HTTP_INVALID_RESPONSE_LENGTH, \
            "invalid response length")\
    X(23,  PACE2_ADVANCED_HTTP_INVALID_CHUNK_NUMBER_TOO_LONG, \
            "chunk number too long (see debug level 3)")\
    X(24,  PACE2_ADVANCED_HTTP_INVALID_NO_CR_OR_SEMICOLON_AFTER_CHUNK_LENGTH, \
            "char behind chunk length no '\\r' or ';' (see debug level 3)")\
    X(25,  PACE2_ADVANCED_HTTP_INVALID_NEXT_STREAM_DIRECTION, \
            "Invalid next_stream_direction after hole.")\
    X(26,  PACE2_ADVANCED_HTTP_INVALID_UNFINISHED_RESPONSE, \
            "Unfinished response but there will be no more server data in this connection!")\
    X(27,  PACE2_ADVANCED_HTTP_INVALID_EXPECTING_REQUEST, \
            "Expecting a request but there will be no more client data in this connection!")\
    X(28,  PACE2_ADVANCED_HTTP_INVALID_UNCOMPRESS_ERROR, \
            "Error occurred while trying to uncompress!")\
    X(29,  PACE2_ADVANCED_HTTP_INVALID_PARTIAL_UNCOMPRESS_ERROR, \
            "Error occurred while trying to uncompress partial data!")\
    X(30,  PACE2_ADVANCED_HTTP_INVALID_FAILED_TO_GET_STREAM, \
            "Failed to get the stream after the hole - marking flow as invalid!")\
    X(31,  PACE2_ADVANCED_HTTP_INVALID_STREAM_DECODER_SYNC, \
            "Cannot synchronize the stream and the decoder - flow direction is not known, marking as invalid flow!")\
    X(32,  PACE2_ADVANCED_HTTP_INVALID_BAD_DATA, \
            "Got bad data, marking as invalid flow!")\

/**
 * @brief enum of HTTP error messages
 */
enum pace2_advanced_http_invalid_string_enums {
    PACE2_ADVANCED_HTTP_INVALID_STRING_LIST(PACE2_ADVANCED_INVALID_VERBOSE_TO_ENUM)

    PACE2_ADVANCED_HTTP_INVALID_STRING_COUNTER
};

/**
 * @brief Meta data flags of basic HTTP response events.
 */
typedef enum {

    /** Server response line. */
    PACE2_HTTP_SERVER_RESPONSE = 0,

    /** Server response content type line. */
    PACE2_HTTP_SERVER_CONTENT_TYPE,

    /** Server response content encoding line. */
    PACE2_HTTP_SERVER_CONTENT_ENCODING,

    /** Server response content length line. */
    PACE2_HTTP_SERVER_CONTENT_LENGTH,

    /** Server response transfer encoding line. */
    PACE2_HTTP_TRANSFER_ENCODING,

    /** Server response server line. */
    PACE2_HTTP_SERVER,

    /** Server response set-cookie line. */
    PACE2_HTTP_SET_COOKIE_LINE,

    /** Server response connection. */
    PACE2_HTTP_SERVER_CONNECTION,

    /** Server response x-tf-access. */
    PACE2_HTTP_SERVER_X_TF_ACCESS,

    /** Number of implemented HTTP response line flags. */
    PACE2_NUMBER_OF_HTTP_RESPONSE_FLAGS

} PACE2_basic_HTTP_response_line;

/* Event Structures *************************************************************/

/**
 * @brief Data structure of a basic HTTP request metadata event.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_BASIC_HTTP_REQUEST_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** Bitmask, which states the available HTTP request metadata. */
    PACE2_bitmask meta_data_mask;

    /** Flag to signal, that the HTTP protocol classification is done. If 0, the request event is
        possibly related to HTTP, but the HTTP classification needs further packets to ascertain the result. */
    u8 http_already_detected:1;

    /** Data structure for all HTTP request header lines, which are not in the above structure @ref line. */
    PACE2_http_extra_line extra_lines[PACE2_NUMBER_OF_HTTP_EXTRA_LINES];

    /** Number of inserted elements in @ref extra_lines array. */
    u8 length_extra_lines;

    /** Data structure for a single HTTP request header line.
        Contains content for each available header line of type @ref PACE2_basic_HTTP_request_line. */
    struct {

        /** Request line of type @ref PACE2_basic_HTTP_request_line.
         *
         *  @deprecated Please do not uses this variable any longer. It will be removed in a future release of this
         *              library. The type of the line is implicit in the index of the line array. */
        PACE2_DEPRECATED(
            "Please do not uses this variable any longer. It will be removed in a future release of this library. The "
            "type of the line is implicit in the index of the line array.")
        PACE2_basic_HTTP_request_line type;

        /** Request line content. */
        PACE2_byte_buffer content;

        /** Flag to signal, that the reported line was complete. */
        u8 line_complete:1;

    } line[PACE2_NUMBER_OF_HTTP_REQUEST_FLAGS];

} PACE2_basic_HTTP_request_event;

/**
 * @brief Data structure of a basic HTTP response metadata event.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_BASIC_HTTP_RESPONSE_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** Bitmask, which states the available HTTP response metadata. */
    PACE2_bitmask meta_data_mask;

    /** Data structure for all HTTP response header lines, which are not in the above structure @ref line. */
    PACE2_http_extra_line extra_lines[PACE2_NUMBER_OF_HTTP_EXTRA_LINES];

    /** Number of inserted elements in @ref extra_lines array. */
    u8 length_extra_lines;

    /** Data structure for a single HTTP response header.
        Contains content for each available header line of type @ref PACE2_basic_HTTP_response_line. */
    struct {

        /** Response line of type @ref PACE2_basic_HTTP_response_line.
         *
         *  @deprecated Please do not uses this variable any longer. It will be removed in a future release of this
         *              library. The type of the line is implicit in the index of the line array. */
        PACE2_DEPRECATED(
            "Please do not uses this variable any longer. It will be removed in a future release of this library. The "
            "type of the line is implicit in the index of the line array.")
        PACE2_basic_HTTP_response_line type;

        /** Response line content. */
        PACE2_byte_buffer content;

        /** Flag to signal, that the reported line was complete. */
        u8 line_complete:1;

    } line[PACE2_NUMBER_OF_HTTP_RESPONSE_FLAGS];

} PACE2_basic_HTTP_response_event;

/**
 * @brief Data structure of a basic HTTP parsed request metadata event.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_BASIC_HTTP_PARSED_REQUEST_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** Parsed data structure. */
    struct pace_http_request_kv_pair_result parsed_data;

} PACE2_basic_HTTP_parsed_request_event;

#ifndef PACE_KERNEL_LIB

/** Structure describing an HTTP header field. */
struct ipd_http_header_field {

    /** The field-value (everything after the first ':', with leading whitespace removed). */
    struct ipd_charset_string value;

    /** Complete header line. */
    struct ipd_charset_string entire_field;
};

/** Enumerator for possible states of a HTTP transaction. */
enum ipd_http_class_transaction_state_enum {

    /** The request part of the transaction is currently processed */
    IPD_HTTP_CLASS_TRANSACTION_STATE_REQUEST,

    /** The response part of the transaction is currently processed */
    IPD_HTTP_CLASS_TRANSACTION_STATE_RESPONSE,
};

/** Main HTTP session structure, each HTTP connection session has this event as initial event. */
struct ipd_class_http_session_struct {

    /** ID of the session. */
    u64 http_session_id;

    /** IP address and port for client and server. */
    struct ipd_5tuple flow_tuple;

    /** Connection ID of the used connection. */
    u64 connection_id;
};

struct pace2_class_http_transaction_stat {
    u64 request;
    u64 response;
};

/**
 * Event for each HTTP request-response pair.
 * This event might be flushed multiple times in the future when HUGE POSTs happens.
 */
struct ipd_class_http_transaction_struct {

    /** ID of the transaction. */
    u64 http_transaction_id;

    /** ID of the ipd_class_http_session the transaction belongs to. */
    u64 http_session_id;

    /** The current state of the transaction. */
    enum ipd_http_class_transaction_state_enum state;

    /** Ip address and port for client and server. */
    struct ipd_5tuple flow_tuple;

    /** String of used method like "GET, POST,...". */
    struct ipd_charset_string request_method;

    /** String of used url like "/index.html". */
    struct ipd_charset_string request_path;

    /** String of request HTTP version like "HTTP/1.0". */
    struct ipd_charset_string request_version;

    /** String of used host like "www.twitter.com". */
    struct ipd_charset_string request_host;

    /** String of used user agent like "Mozilla/5.0 (X11; U; Linux i686; de; rv:1.8.1.14)". */
    struct ipd_charset_string request_useragent;

    /** String of used referrer like "www.facebook.com/dummy.html". */
    struct ipd_charset_string request_referrer;

    /** String of used content type like "application/xml". */
    struct ipd_charset_string request_contenttype;

    /** String of cookie. */
    struct ipd_charset_string request_cookie;

    /** String of the requested range. */
    struct ipd_charset_string request_range;

    /** String of request content length sent to the HTTP server. */
    struct ipd_charset_string request_content_length;

    /** Number of other lines in the request_other array. */
    u64 request_other_lines;

    /** Array of "other" lines like "Accept-Encoding: gzip,deflate". */
    struct ipd_charset_string * request_other;

    /** String of response HTTP version like "HTTP/1.0". */
    struct ipd_charset_string response_version;

    /** String of response code number returned from the HTTP server like "200". */
    struct ipd_charset_string response_code;

    /** Response reason phrase returned from the HTTP server like "OK". */
    struct ipd_charset_string response_reason_phrase;

    /** String of response content length returned from the HTTP server. */
    struct ipd_charset_string response_content_length;

    /** String of the content range returned from the HTTP server. */
    struct ipd_charset_string response_content_range;

    /** Response content type like "application/xml". */
    struct ipd_charset_string response_contenttype;

    /** Length of response content encoding array. */
    u64 response_content_encoding_array_len;

    /** Response content encoding like "gzip". */
    struct ipd_charset_string * response_content_encoding;

    /** Response transfer encoding like "chunked". */
    struct ipd_charset_string response_transfer_encoding;

    /** Response connection like "close". */
    struct ipd_charset_string response_connection;

    u64 response_set_cookie_array_len;

    /** Response Set-Cookie header. */
    struct ipd_charset_string * response_set_cookie;

    /** Response location for a http redirect. */
    struct ipd_charset_string response_location;

    /** Number of other lines in the response_other array. */
    u64 response_other_lines;

    /** Array of "other" lines like "Date: Wed, 07 May 2008 13:21:41 GMT". */
    struct ipd_charset_string * response_other;

    /** Number of packets that carry some data.
     * Overflow occurs (counter provides wrong value) after reaching 65535. */
    struct pace2_class_http_transaction_stat data_packets_count;

    /** Number of packets that do not carry any data (for example, acknowledgments only).
     * Overflow occurs (counter provides wrong value) after reaching 65535. */
    struct pace2_class_http_transaction_stat control_packets_count;

    /** Number of layer-3 bytes including layer-3 header that belong to packets that carry some data.
     * Overflow occurs (counter provides wrong value) after reaching 65535. */
    struct pace2_class_http_transaction_stat l3_data_bytes;

    /** Number of layer-3 bytes including layer-3 header that belong to packets that do not carry any data.
     * Overflow occurs (counter provides wrong value) after reaching 65535. */
    struct pace2_class_http_transaction_stat l3_control_bytes;

    /** Number of layer-7 bytes including layer-7 header.
     * Overflow occurs (counter provides wrong value) after reaching 4294967295. */
    struct pace2_class_http_transaction_stat l7_data_bytes;

    /** Request transfer encoding (e.g. "chunked"). */
    struct ipd_charset_string request_transfer_encoding;
};

/** Events for each HTTP request - response pair. */
struct ipd_class_http_transfer_payload_struct {

    /** ID of the ipd_class_http_transaction it belongs to. */
    u64 http_transaction_id;

    /** ID of the ipd_class_http_session it belongs to. */
    u64 http_session_id;

    /** Offset of the data in the downloaded data. */
    u64 data_offset;

    /** Data with encoding. */
    struct ipd_charset_string data;

    /** Is !=0 in the final segment of a request or response payload. */
    u8 payload_done;
};

typedef struct ipd_class_http_transfer_payload_struct ipd_class_http_transfer_payload_int_t;
typedef struct ipd_class_http_transaction_struct ipd_class_http_transaction_int_t;
typedef struct ipd_class_http_session_struct ipd_class_http_session_int_t;

/**
 * @brief Data structure for advanced HTTP events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_HTTP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** HTTP event subtype. */
    PACE2_advanced_HTTP_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_5TUPLE event. */
        struct ipd_5tuple tuple;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_HEADER event. */
        struct ipd_charset_string request_header;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_METHOD event. */
        struct ipd_charset_string request_method;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_PATH event. */
        struct ipd_charset_string request_path;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_VERSION event. */
        struct ipd_charset_string request_version;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_HOST event. */
        struct ipd_http_header_field request_host;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_USERAGENT event. */
        struct ipd_http_header_field request_useragent;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_REFERER event. */
        struct ipd_http_header_field request_referer;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_CONTENTTYPE event. */
        struct ipd_http_header_field request_contenttype;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_COOKIE event. */
        struct ipd_http_header_field request_cookie;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_RANGE event. */
        struct ipd_http_header_field request_range;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_CONTENT_LENGTH event. */
        struct ipd_http_header_field request_content_length;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_UNKNOWN event. */
        struct ipd_charset_string request_unknown;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_PAYLOAD event. */
        struct ipd_charset_string request_payload;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_HEADER event. */
        struct ipd_charset_string response_header;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_VERSION event. */
        struct ipd_charset_string response_version;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_CODE event. */
        struct ipd_charset_string response_code;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_REASON_PHRASE event. */
        struct ipd_charset_string response_reason_phrase;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_CONTENT_LENGTH event. */
        struct ipd_http_header_field response_content_length;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_CONTENT_RANGE event. */
        struct ipd_http_header_field response_content_range;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_CONTENTTYPE event. */
        struct ipd_http_header_field response_contenttype;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_CONTENT_ENCODING event. */
        struct ipd_http_header_field response_content_encoding;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_TRANSFER_ENCODING event. */
        struct ipd_http_header_field response_transfer_encoding;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_CONNECTION event. */
        struct ipd_http_header_field response_connection;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_SET_COOKIE event. */
        struct ipd_http_header_field response_set_cookie;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_LOCATION event. */
        struct ipd_http_header_field response_location;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_UNKNOWN event. */
        struct ipd_charset_string response_unknown;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_PAYLOAD event. */
        struct ipd_charset_string response_payload;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_INVALID event. */
        char const * invalid;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_REQUEST_HOLE event. */
        u64 request_hole;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_RESPONSE_HOLE event. */
        u64 response_hole;

        /** Contains data for a @ref PACE2_ADVANCED_HTTP_INVALID_VERBOSE event. */
        struct pace2_invalid_message invalid_verbose;

    } event_data;

} PACE2_advanced_HTTP_event;

/**
 * @brief Data structure for HTTP class events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_CLASS_HTTP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** HTTP event subtype. */
    PACE2_class_HTTP_event_type meta_data_type;

    PACE2_CLASS_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_CLASS_HTTP_SESSION event. */
        struct ipd_class_http_session_struct session;

        /** Contains data for a @ref PACE2_CLASS_HTTP_TRANSACTION event. */
        struct ipd_class_http_transaction_struct transaction;

        /** Contains data for a @ref PACE2_CLASS_HTTP_REQUEST_DATA_TRANSFER event. */
        struct ipd_class_http_transfer_payload_struct request_data_transfer;

        /** Contains data for a @ref PACE2_CLASS_HTTP_RESPONSE_DATA_TRANSFER event. */
        struct ipd_class_http_transfer_payload_struct response_data_transfer;

        /** Contains data for a @ref PACE2_CLASS_HTTP_TRANSACTION_DONE event. */
        struct ipd_class_http_transaction_struct transaction_done;

        /** Contains data for a @ref PACE2_CLASS_HTTP_SESSION_DONE event. */
        struct ipd_class_http_session_struct session_done;

    } event_data;

} PACE2_class_HTTP_event;

#endif /* PACE_KERNEL_LIB */

#endif /* HTTP_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
