/** @file      icq_webchat_events.h
  * @brief     ICQ Webchat event definitions.
  * @details   @copydoc doc_icq_webchat
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef ICQ_WEBCHAT_EVENTS_H
#define ICQ_WEBCHAT_EVENTS_H

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced ICQ webchat event.
 */
#define PACE2_ADVANCED_ICQ_WEBCHAT_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced ICQ webchat events.
 */
typedef enum {

    /** This event is issued when an ICQ Webchat connection starts.
     * Provides a struct ipd_icq_webchat_user_info * with information about the connecting user. */
    PACE2_ADVANCED_ICQ_WEBCHAT_SESSION_START,

    /** This event is issued when additional information about the connected user becomes available.
     * It provides a struct ipd_icq_webchat_user_info *. */
    PACE2_ADVANCED_ICQ_WEBCHAT_SELFUSER_INFO,

    /** This event is issued when information about a buddy becomes available.
     * It provides a struct ipd_icq_webchat_user_info *. */
    PACE2_ADVANCED_ICQ_WEBCHAT_BUDDY_INFO,

    /** This event is issued when the status of any user changes.
     * It provides a struct ipd_icq_webchat_user_info *. */
    PACE2_ADVANCED_ICQ_WEBCHAT_USER_STATUS,

    /** This event is issued when the connected user sends a message to another user.
     * It provides a struct ipd_icq_webchat_message *. */
    PACE2_ADVANCED_ICQ_WEBCHAT_MESSAGE_OUTGOING,

    /** This event is issued when the connected user receives a message from another user.
     * It provides a struct ipd_icq_webchat_message *. */
    PACE2_ADVANCED_ICQ_WEBCHAT_MESSAGE_INCOMING,

    /** This event is issued when an ICQ Webchat connection ends. */
    PACE2_ADVANCED_ICQ_WEBCHAT_END_OF_FLOW,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_ICQ_WEBCHAT_EVENT_TYPES

} PACE2_advanced_icq_webchat_event_type;

#ifndef PACE_KERNEL_LIB

/**
 * Enumeration for the type of ICQ Webchat user status.
 */
enum ipd_icq_webchat_user_status_enum {

    /** The user is offline. */
    IPD_ICQ_WEBCHAT_USER_STATUS_OFFLINE,

    /** The user is online. */
    IPD_ICQ_WEBCHAT_USER_STATUS_ONLINE,

    /** The user is away. */
    IPD_ICQ_WEBCHAT_USER_STATUS_AWAY,

    /** The user is invisible. */
    IPD_ICQ_WEBCHAT_USER_STATUS_INVISIBLE,

    /** The user is occupied. */
    IPD_ICQ_WEBCHAT_USER_STATUS_OCCUPIED,
};

/**
 * Structure for ICQ Webchat user information.
 */
struct ipd_icq_webchat_user_info {

    /** The ID of the user owning the session (the connected user). */
    struct ipd_charset_string session_owner;

    /** The aim ID of the user. */
    struct ipd_charset_string aim_id;

    /** The display ID of the user.
     * In the most cases the aim_id or the display_name of the user again.
     * Is a internal value from the icq webchat itself. */
    struct ipd_charset_string display_id;

    /** The display name of the user. */
    struct ipd_charset_string display_name;

    /** The status of the user. @see ipd_icq_webchat_user_status_enum. */
    enum ipd_icq_webchat_user_status_enum status;

    /** The status message of the user. */
    struct ipd_charset_string status_message;

    /** The group the user belongs to (only set for #PACE2_ADVANCED_ICQ_WEBCHAT_BUDDY_INFO events). */
    struct ipd_charset_string group;
};

/**
 * Structure for an ICQ Webchat message.
 */
struct ipd_icq_webchat_message {

    /** Information about the user the message is sent to or the user the message is coming from.
     * @see ipd_icq_webchat_user_info. */
    struct ipd_icq_webchat_user_info * user;

    /** The message string. */
    struct ipd_charset_string message;
};

/**
 * @brief Data structure for advanced ICQ webchat events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_ICQ_WEBCHAT_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** ICQ webchat event subtype. */
    PACE2_advanced_icq_webchat_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    union {

        /** Contains data for a @ref PACE2_ADVANCED_ICQ_WEBCHAT_SESSION_START event. */
        struct ipd_icq_webchat_user_info session_start;

        /** Contains data for a @ref PACE2_ADVANCED_ICQ_WEBCHAT_SELFUSER_INFO event. */
        struct ipd_icq_webchat_user_info selfuser_info;

        /** Contains data for a @ref PACE2_ADVANCED_ICQ_WEBCHAT_BUDDY_INFO event. */
        struct ipd_icq_webchat_user_info buddy_info;

        /** Contains data for a @ref PACE2_ADVANCED_ICQ_WEBCHAT_USER_STATUS event. */
        struct ipd_icq_webchat_user_info user_status;

        /** Contains data for a @ref PACE2_ADVANCED_ICQ_WEBCHAT_MESSAGE_OUTGOING event. */
        struct ipd_icq_webchat_message message_out;

        /** Contains data for a @ref PACE2_ADVANCED_ICQ_WEBCHAT_MESSAGE_INCOMING event. */
        struct ipd_icq_webchat_message message_in;

    } event_data;

} PACE2_advanced_icq_webchat_event;

#endif /* PACE_KERNEL_LIB */

#endif // ICQ_WEBCHAT_EVENTS_H

/*********************************************************************************
 ** EOF
 ********************************************************************************/
