/** @file      imap_events.h
  * @brief     IMAP event definitions.
  * @details   @copydoc doc_imap
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef IMAP_EVENTS_H
#define IMAP_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "imf_events.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced IMAP event.
 */
#define PACE2_ADVANCED_IMAP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced imap events.
 */
typedef enum {
    /** This event is sent when the first payload data has been processed and the client and server roles have been
     * determined. The provided data is a pointer to an int that contains the client direction (0 or 1). */
    PACE2_ADVANCED_IMAP_CLIENT_DIRECTION,

    /** Provides a struct ipd_charset_string * with the login name of the user. */
    PACE2_ADVANCED_IMAP_USER,

    /** Provides a struct ipd_charset_string * with the password of the user, if available. */
    PACE2_ADVANCED_IMAP_PASS,

    /** This event indicates the start of a TLS connection. */
    PACE2_ADVANCED_IMAP_STARTTLS,

    /** The server greeting, @c data points to a struct ipd_imap_greeting. */
    PACE2_ADVANCED_IMAP_GREETING,

    /** This event is sent for each data item in a FETCH response that contains message text (BODY[], RFC822,
     * RFC822.HEADER, RFC822.TEXT). All IMF events in this flow up to the next IMF_DONE belong to this
     * (section of a) message. @c data points to a struct ipd_imap_fetch_body_item. */
    PACE2_ADVANCED_IMAP_FETCH_BODY_ITEM,

    /** Sent at the beginning of each APPEND command issued by the client, ie. before the IMF events of a mail that
     * the client uploads to the server. @c data points to a struct ipd_charset_string that contains
     * the command tag. */
    PACE2_ADVANCED_IMAP_APPEND,

    /** Sent for each tagged server response, @c data points to a struct ipd_imap_response. */
    PACE2_ADVANCED_IMAP_RESPONSE,

    /** Sent when the client issues a LOGOUT command, @c data is a NULL pointer. */
    PACE2_ADVANCED_IMAP_LOGOUT,

    /** Indicates that the server has sent a BYE response, @c data is a NULL pointer. */
    PACE2_ADVANCED_IMAP_BYE,

    /** Value of the From: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_IMAP_IMF_FROM,

    /** Value of the To: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_IMAP_IMF_TO,

    /** Value of the Cc: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_IMAP_IMF_CC,

    /** Value of the Bcc: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_IMAP_IMF_BCC,

    /** Value of the Subject: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_SUBJECT,

    /** Value of the Date: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_DATE,

    /** Value of the Sender: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_IMAP_IMF_SENDER,

    /** Value of the Reply-To: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_IMAP_IMF_REPLY_TO,

    /** Value of the Message-Id: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_MESSAGE_ID,

    /** Value of the In-Reply-To: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_IN_REPLY_TO,

    /** Value of the References: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_REFERENCES,

    /** Value of the Comments: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_COMMENTS,

    /** Value of the Keywords: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_KEYWORDS,

    /** Value of the Trace: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_TRACE,

    /** Value of the Resent-Date: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_RESENT_DATE,

    /** Value of the Resent-From: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_IMAP_IMF_RESENT_FROM,

    /** Value of the Resent-Sender: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_IMAP_IMF_RESENT_SENDER,

    /** Value of the Resent-To: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_IMAP_IMF_RESENT_TO,

    /** Value of the Resent-Cc: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_IMAP_IMF_RESENT_CC,

    /** Value of the Resent-Bcc: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_IMAP_IMF_RESENT_BCC,

    /** Value of the Resent-Msg-Id: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_RESENT_MSG_ID,

    /** Value of the MIME-Version: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_MIME_VERSION,

    /** Complete header field (including field name).
     * Sent for all message or part header fields of a transmitted message not covered by more specific events;
     * @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_IMAP_IMF_OTHER,

    /** The entire header of a mail message; @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_HEADER,

    /** Value of the Content-Type: header field of a message or MIME part;
     * @c data points to a struct ipd_imf_content_type. */
    PACE2_ADVANCED_IMAP_IMF_CONTENT_TYPE,

    /** Value of the Content-Transfer-Encoding: header field of a message or MIME part;
     * @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_IMAP_IMF_CONTENT_TRANSFER_ENCODING,

    /** Value of the Content-Disposition: header field of a MIME part;
     * @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_IMAP_IMF_CONTENT_DISPOSITION,

    /** The entire header of a MIME part; @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_IMAP_IMF_PART_HEADER,

    /** This event indicates the begin of a byte stream (MIME part body, multipart preamble or epilogue);
     * @c data points to a struct ipd_charset_string containing the section number of the part. */
    PACE2_ADVANCED_IMAP_IMF_DATA_BEGIN,

    /** A block of data in a byte stream; @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_IMAP_IMF_DATA,

    /** This event indicates the end of a byte stream;
     * @c data points to a struct ipd_charset_string containing the section number of the part. */
    PACE2_ADVANCED_IMAP_IMF_DATA_END,

    /** This event indicates the end of a continuous stream of IMF events belonging to one IMF message;
     * @c data is a NULL pointer. */
    PACE2_ADVANCED_IMAP_IMF_DONE,

    /** This event is issued when decoding of an IMAP connection fails.
     * It provides a char const * to a message describing the failure reason. */
    PACE2_ADVANCED_IMAP_INVALID,

    /** This event indicates the end of a connection. It does not provide any data. */
    PACE2_ADVANCED_IMAP_CONNECTION_END,

    /** Value of Return-Path: header field of a transmitted mail message; denotes the address where
     * to send error messages; @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_RETURN_PATH,

    /** Value of Received: header field of a transmitted mail message; an array of strings: each hop
     * across the path from sender to receiver adds information about the respective sender.
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_IMAP_IMF_RECEIVED,

    /** Compression is requested for this connection. */
    PACE2_ADVANCED_IMAP_COMPRESSION_REQUESTED,

    /** Compression is started for this connection. */
    PACE2_ADVANCED_IMAP_COMPRESSION_STARTED,

    /** Mailbox was successfully selected. */
    PACE2_ADVANCED_IMAP_SELECT,

    /** This event indicates a missing data length associated with a specific MIME section number. */
    PACE2_ADVANCED_IMAP_IMF_DATA_MISSING,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_IMAP_EVENT_TYPES

} PACE2_advanced_imap_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/** IMAP response states for tagged responses. */
enum ipd_imap_resp_state {
    IPD_IMAP_RESPONSE_OK,
    IPD_IMAP_RESPONSE_NO,
    IPD_IMAP_RESPONSE_BAD,
};

/** A tagged IMAP response. */
struct ipd_imap_response {

    /** Tag string. */
    struct ipd_charset_string tag;

    /** @copydoc ipd_imap_resp_state */
    enum ipd_imap_resp_state state;

    /** Response text. */
    struct ipd_charset_string text;
};

/** Condition codes for an IMAP greeting. */
enum ipd_imap_greeting_state {

    /** Connection without pre-authentication. */
    IPD_IMAP_GREETING_OK,

    /** Pre-authenticated connection. */
    IPD_IMAP_GREETING_PREAUTH,

    /** Rejected connection. */
    IPD_IMAP_GREETING_BYE,
};

/** IMAP server greeting. */
struct ipd_imap_greeting {

    /** @copydoc ipd_imap_greeting_state */
    enum ipd_imap_greeting_state state;

    /** Greeting text. */
    struct ipd_charset_string text;
};

/** Describes which part of an IMF message are transmitted in a FETCH data item (in the following, "message"
 * refers either to a top-level message (when ipd_imap_fetch_body_item::number is empty), or to a MIME part
 * with content-type message/rfc822, ie. a forwarded message stored in an attachment). */
enum ipd_imap_part_spec {

    /** The entire message, or the body of a MIME part. */
    IMAP_SECTION_ALL,

    /** Only the message header. */
    IMAP_SECTION_HEADER,

    /** Just selected message header fields. */
    IMAP_SECTION_HEADER_FIELDS,

    /** All not-selected message header fields. */
    IMAP_SECTION_HEADER_FIELDS_NOT,

    /** The body of a message (omitting the header). */
    IMAP_SECTION_TEXT,

    /** The header of a MIME part. */
    IMAP_SECTION_MIME,
};

/** Description of a FETCH data item that contains message text
 * (parsed representation of BODY data items, cf. RFC 3501). */
struct ipd_imap_fetch_body_item {

    /** Message sequence number (from the response head). */
    u64 sequence_number;

    /** MIME part number, eg. "1.2" (empty for "entire message"). */
    struct ipd_charset_string number;

    /** Section of the message/MIME part that is fetched. */
    enum ipd_imap_part_spec section;

    /** Space-separated list of header fields (only valid if section is @ref IMAP_SECTION_HEADER_FIELDS or
     * @ref IMAP_SECTION_HEADER_FIELDS_NOT. */
    struct ipd_charset_string header_fields;

    /** Byte offset into the message part selected by the number and section fields. */
    u64 origin_octet;

    /** Flag; nonzero if the @ref ipd_imap_fetch_body_item::origin_octet "origin_octet" field is set. */
    u8 origin_octet_set;
};

/**
 * @brief Data structure for advanced imap events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_IMAP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** IMAP event subtype. */
    PACE2_advanced_imap_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_CLIENT_DIRECTION event. */
        int client_direction;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_USER event. */
        struct ipd_charset_string user;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_PASS event. */
        struct ipd_charset_string pass;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_GREETING event. */
        struct ipd_imap_greeting greeting;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_FETCH_BODY_ITEM event. */
        struct ipd_imap_fetch_body_item fetch_body_item;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_APPEND event. */
        struct ipd_charset_string append;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_RESPONSE event. */
        struct ipd_imap_response response;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_FROM event. */
        struct ipd_imf_address imf_from;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_TO event. */
        struct ipd_imf_address imf_to;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_CC event. */
        struct ipd_imf_address imf_cc;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_BCC event. */
        struct ipd_imf_address imf_bcc;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_SUBJECT event. */
        struct ipd_charset_string imf_subject;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_DATE event. */
        struct ipd_charset_string imf_date;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_SENDER event. */
        struct ipd_imf_address imf_sender;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_REPLY_TO event. */
        struct ipd_imf_address imf_reply_to;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_MESSAGE_ID event. */
        struct ipd_charset_string imf_message_id;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_IN_REPLY_TO event. */
        struct ipd_charset_string imf_in_reply_to;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_REFERENCES event. */
        struct ipd_charset_string imf_references;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_COMMENTS event. */
        struct ipd_charset_string imf_comments;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_KEYWORDS event. */
        struct ipd_charset_string imf_keywords;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_TRACE event. */
        struct ipd_charset_string imf_trace;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_RESENT_DATE event. */
        struct ipd_charset_string imf_resent_date;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_RESENT_FROM event. */
        struct ipd_imf_address imf_resent_from;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_RESENT_SENDER event. */
        struct ipd_imf_address imf_resent_sender;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_RESENT_TO event. */
        struct ipd_imf_address imf_resent_to;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_RESENT_CC event. */
        struct ipd_imf_address imf_resent_cc;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_RESENT_BCC event. */
        struct ipd_imf_address imf_resent_bcc;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_RESENT_MSG_ID event. */
        struct ipd_charset_string imf_resent_msg_id;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_MIME_VERSION event. */
        struct ipd_charset_string imf_mime_version;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_OTHER event. */
        struct ipd_imf_part_data imf_other;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_HEADER event. */
        struct ipd_charset_string imf_header;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_CONTENT_TYPE event. */
        struct ipd_imf_content_type imf_content_type;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_CONTENT_TRANSFER_ENCODING event. */
        struct ipd_imf_part_data imf_content_transfer_encoding;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_CONTENT_DISPOSITION event. */
        struct ipd_imf_part_data imf_content_disposition;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_PART_HEADER event. */
        struct ipd_imf_part_data imf_part_header;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_DATA_BEGIN event. */
        struct ipd_charset_string imf_data_begin;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_DATA event. */
        struct ipd_imf_part_data imf_data;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_DATA_END event. */
        struct ipd_charset_string imf_data_end;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_INVALID event. */
        char const * invalid;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_RETURN_PATH event. */
        struct ipd_charset_string imf_return_path;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_RECEIVED event. */
        struct ipd_charset_string imf_received;

        /** Contains data for a @ref PACE2_ADVANCED_IMAP_IMF_DATA_MISSING event. */
        struct ipd_imf_missing_data imf_data_missing;

    } event_data;

} PACE2_advanced_imap_event;

#endif

#endif /* IMAP_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
