/** @file      irc_events.h
  * @brief     IRC event definitions.
  * @details   @copydoc doc_irc
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef IRC_EVENTS_H
#define IRC_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Group ******************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced IRC event.
 */
#define PACE2_ADVANCED_IRC_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * Structure for generic IRC server information which is not covered by other specialised event types
 */
struct pace2_irc_server_info {

    /** Command code. */
    struct ipd_charset_string command_code;

    /** Command arguments. */
    struct ipd_charset_string command_arguments;
};

/**
 * @brief Types for advanced IRC events.
 */
typedef enum {

    /** Provides a struct ipd_irc_nick_data * with nickname information for the currently connected user. */
    PACE2_ADVANCED_IRC_NICK_USER,

    /** Provides a struct ipd_irc_nick_data * with nickname information for another user
     * on the connected IRC server. */
    PACE2_ADVANCED_IRC_NICK_OTHER,

    /** Provides a struct ipd_irc_chat_data * with information about a private message that is transmitted. */
    PACE2_ADVANCED_IRC_PRIMSG,

    /** Provides a struct ipd_charset_string * with the username of the user that is currently connecting
     * to a ICR server. */
    PACE2_ADVANCED_IRC_LOGIN_USER,

    /** Provides a struct ipd_charset_string * with the password of the user that is currently connecting
     * to a ICR server. */
    PACE2_ADVANCED_IRC_PASS,

    /** Provides a struct ipd_charset_string * with the hostname of the client machine from which the user
     * is connecting. */
    PACE2_ADVANCED_IRC_LOGIN_HOST,

    /** Provides a struct ipd_charset_string * with the name of the IRC server which the user is connecting to. */
    PACE2_ADVANCED_IRC_LOGIN_SERVERNAME,

    /** Provides a struct ipd_charset_string * with the real name of the user that is currently connecting to
     * a ICR server. */
    PACE2_ADVANCED_IRC_LOGIN_REALNAME,

    /** This event is issued when the currently connected user quits from the IRC server.
     * It does not provide any data. */
    PACE2_ADVANCED_IRC_QUIT_USER,

    /** Provides a struct ipd_irc_chat_data * with information about another user who is quitting
     * from the ICR server. */
    PACE2_ADVANCED_IRC_QUIT_OTHER,

    /** Provides a struct ipd_irc_chat_data * with information about a user joining a chatroom. */
    PACE2_ADVANCED_IRC_JOIN,

    /** Provides a struct ipd_irc_chat_data * with information about a user leaving a chatroom. */
    PACE2_ADVANCED_IRC_PART,

    /** Provides a struct ipd_irc_chat_data * with information about the topic of a chatroom. */
    PACE2_ADVANCED_IRC_TOPIC,

    /** Provides a struct ipd_irc_ft_data * with information about a new file transfer that is announced. */
    PACE2_ADVANCED_IRC_FT_ANNOUNCE,

    /** Provides a struct ipd_irc_ft_data * with a segment of file transfer data. */
    PACE2_ADVANCED_IRC_FT_DATA,

    /** Provides a struct ipd_irc_command * with the details of the command. */
    PACE2_ADVANCED_IRC_COMMAND,

    /** This event is issued when decoding of a IRC connection fails. It provides a char const * to a message
     * describing the failure reason. */
    PACE2_ADVANCED_IRC_INVALID,

    /** This event is issued when a IRC connection ends. It does not provide any data. */
    PACE2_ADVANCED_IRC_END_OF_FLOW,

    /** Provides a struct pace2_irc_server_info with the details of the server information. */
    PACE2_ADVANCED_IRC_SERVER_INFO,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_IRC_EVENT_TYPES

} PACE2_advanced_irc_event_type;

/* Event Structures *************************************************************/

/** Structure for events that provide information about IRC chatrooms or actual messages. */
struct ipd_irc_chat_data {

    /** The chatroom the event relates to. This can be a the username of the user receiving a message in case of
        a private message or a chatroom name that starts with a # character. */
    struct ipd_charset_string chatroom;

    /** The nickname of the user sending a message. */
    struct ipd_charset_string nick;

    /** The username of the user sending a message. This information is not always available. */
    struct ipd_charset_string user;

    /** A message string the event relates to. Can be empty depending on the actual event. */
    struct ipd_charset_string message;
};

/** Structure for events that provide IRC nickname information. */
struct ipd_irc_nick_data {

    /** The username of the user the event relates to. */
    struct ipd_charset_string user;

    /** The old nickname of the user. */
    struct ipd_charset_string old_nick;

    /** The current nickname of the user. */
    struct ipd_charset_string nick;
};

/** Structure for events that provide IRC file transfer information and data. */
struct ipd_irc_ft_data {

    /** The username of the user that is connected to the IRC server. */
    struct ipd_charset_string user;

    /** The name of the IRC server the user is connected to. */
    struct ipd_charset_string server;

    /** A transfer ID that is unique for the IRC connection. */
    u64 transfer_id;

    /** The user who is the sender of the file. */
    struct ipd_charset_string sender;

    /** The user who is the receiver of the file. */
    struct ipd_charset_string receiver;

    /** The filename of the file being transmitted. */
    struct ipd_charset_string filename;

    /** The file size in bytes of the file being transmitted. */
    u32 file_size;

    /** A segment of actual file transfer data. */
    struct ipd_charset_string data;

    /** Number of bytes of the particular file already transmitted. */
    u64 already_transferred_size;

    u8 last_file_segment;
};

/**
 * Structure for IRC command, which was received.
 */
struct ipd_irc_command {

    /** Complete command. */
    struct ipd_charset_string command_line;

    /** Command code like NICK or JOIN. */
    struct ipd_charset_string command_code;

    /** Command arguments. */
    struct ipd_charset_string command_arguments;
};

/**
 * @brief Data structure for advanced IRC events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_IRC_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** IRC event subtype. */
    PACE2_advanced_irc_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_IRC_NICK_USER event. */
        struct ipd_irc_nick_data nick_user;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_NICK_OTHER event. */
        struct ipd_irc_nick_data nick_other;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_PRIMSG event. */
        struct ipd_irc_chat_data primsg;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_LOGIN_USER event. */
        struct ipd_charset_string login_user;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_PASS event. */
        struct ipd_charset_string pass;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_LOGIN_HOST event. */
        struct ipd_charset_string login_host;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_LOGIN_SERVERNAME event. */
        struct ipd_charset_string login_servername;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_LOGIN_REALNAME event. */
        struct ipd_charset_string login_realname;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_QUIT_OTHER event. */
        struct ipd_irc_chat_data quit_other;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_JOIN event. */
        struct ipd_irc_chat_data join;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_PART event. */
        struct ipd_irc_chat_data part;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_TOPIC event. */
        struct ipd_irc_chat_data topic;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_FT_ANNOUNCE event. */
        struct ipd_irc_ft_data ft_announce;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_FT_DATA event. */
        struct ipd_irc_ft_data ft_data;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_COMMAND event. */
        struct ipd_irc_command command;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_INVALID event. */
        char const * invalid;

        /** Contains data for a @ref PACE2_ADVANCED_IRC_SERVER_INFO event. */
        struct pace2_irc_server_info server_info;

    } event_data;

} PACE2_advanced_irc_event;

#endif /* PACE_KERNEL_LIB */

#endif /* IRC_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
