/** @file      jabber_events.h
  * @brief     Jabber/XMPP event definitions.
  * @details   @copydoc doc_jabber
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef JABBER_EVENTS_H
#define JABBER_EVENTS_H

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced Jabber event.
 */
#define PACE2_ADVANCED_JABBER_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "rtp_events.h"

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced jabber events.
 */
typedef enum {

    /** This event indicates a new JABBER connection. It does not provide any data. */
    PACE2_ADVANCED_JABBER_NEW_CONNECTION,

    /** This event indicates the end of an JABBER connection. It does not provide any data. */
    PACE2_ADVANCED_JABBER_CLOSE_CONNECTION,

    /** Signals the start of a TLS connection. */
    PACE2_ADVANCED_JABBER_STARTTLS,

    /** Provides a struct ipd_charset_string * with the JABBER ID of the connecting user. */
    PACE2_ADVANCED_JABBER_LOG_ON,

    /** Provides a struct ipd_charset_string * with the JABBER ID of the buddy. */
    PACE2_ADVANCED_JABBER_BUDDY_ADD,

    /** Provides a struct ipd_charset_string * with the JABBER ID of the buddy which should be added. */
    PACE2_ADVANCED_JABBER_ADD_NEW_BUDDY,

    /** Provides a struct ipd_charset_string * with the JABBER ID of the buddy which accept the buddy invite. */
    PACE2_ADVANCED_JABBER_NEW_BUDDY_ACCEPT,

    /** Provides a struct ipd_charset_string * with the JABBER ID of the buddy which cancel the buddy invite. */
    PACE2_ADVANCED_JABBER_NEW_BUDDY_CANCEL,

    /** Provides a struct ipd_charset_string * with the JABBER ID of the user who goes to online. */
    PACE2_ADVANCED_JABBER_USER_ON,

    /** Provides a struct ipd_charset_string * with the JABBER ID of the user who goes to offline. */
    PACE2_ADVANCED_JABBER_USER_OFF,

    /** Provides a struct ipd_charset_string * with the JABBER ID of the user who goes to do not disturb. */
    PACE2_ADVANCED_JABBER_USER_DO_NOT_DISTURB,

    /** Signals that the user has send a message.
     * Provides a struct ipd_jabber_message_struct * with the send message. */
    PACE2_ADVANCED_JABBER_MESSAGE_OUT,

    /** Signals that the user has received a message.
     * Provides a struct ipd_jabber_message_struct * with the received message. */
    PACE2_ADVANCED_JABBER_MESSAGE_IN,

    /** Signals that a buddy has change his status.
     * Provides a struct ipd_jabber_message_struct * with from for the buddy and message for the new status. */
    PACE2_ADVANCED_JABBER_BUDDY_CHANGE_STATUS,

    /** Signals that the use has change his status. Provides a struct ipd_charset_string * with the new status. */
    PACE2_ADVANCED_JABBER_CHANGE_STATUS,

    /** Signals the announce of a VOIP.
     * Provides a struct ipd_jabber_voip_struct * with the information about the av_session. */
    PACE2_ADVANCED_JABBER_VOIP_ANNOUNCE,

    /** Signals the accept of a VOIP.
     * Provides a struct ipd_jabber_voip_struct * with the information about the av_session. */
    PACE2_ADVANCED_JABBER_VOIP_ACCEPT,

    /** Signals the reject of a VOIP.
     * Provides a struct ipd_jabber_voip_struct * with the information about the av_session. */
    PACE2_ADVANCED_JABBER_VOIP_REJECT,

    /** Signals the end of a VOIP.
     * Provides a struct ipd_jabber_voip_struct * with the information about the av_session. */
    PACE2_ADVANCED_JABBER_VOIP_END,

    /** Provides a struct ipd_rtp_data * with a RTP packet which is related to an JABBER voice over IP session.
     * This event is actually issued by the RTP decoder. */
    PACE2_ADVANCED_JABBER_RTP_DATA,

    /** This event is issued when decoding of an JABBER connection fails.
     * It provides a char const * to a message describing the failure reason. */
    PACE2_ADVANCED_JABBER_INVALID,

    /** This event is issued when an JABBER stream is compressed.
     * It provides the information message containing compression type used. */
    PACE2_ADVANCED_JABBER_STREAM_COMPRESSED,

    /** This event is issued when an JABBER connection contains compressed data.
     * It provides the compressed data. */
    PACE2_ADVANCED_JABBER_COMPRESSED_DATA,

    /** This event is issued when an JABBER connection ends. It does not provide any data. */
    PACE2_ADVANCED_JABBER_END_OF_FLOW,

    /** This event is issued when a SSRC value is occurred. */
    PACE2_ADVANCED_JABBER_RTP_SSRC,

    /** Signals a file transfer announcement. Provides a PACE2_jabber_file_transfer_info pointer. */
    PACE2_ADVANCED_JABBER_FILE_TRANSFER_ANNOUNCEMENT,

    /** IBB file transfer opening. Issued when the file transferred is opened.
     * Provides the struct ipd_jabber_ibb_open_filetransfer_struct * (SID, maximum block size,
     * and the identifier of the involved parties if possible). */
    PACE2_ADVANCED_JABBER_IBB_OPEN_FILETRANSFER,

    /** IBB file transfer closure. Issued when the file transfer is completed.
     * Provides the struct ipd_jabber_ibb_announcement_struct *
     * (SID and the identifier of the involved parties if possible). */
    PACE2_ADVANCED_JABBER_IBB_CLOSE_FILETRANSFER,

    /** Signals that a file transfer was rejected. Provides a PACE2_jabber_file_transfer_info pointer. */
    PACE2_ADVANCED_JABBER_FILE_TRANSFER_REJECT,

    /** IBB file transfer data. Issued when found BASE-64 encoded data sent by an IBB file transfer.
        * Provides the struct ipd_jabber_ibb_data_struct * (SID, chunk sequence number, pointer to the data,
        * and the identifier of the involved parties if possible). */
    PACE2_ADVANCED_JABBER_IBB_DATA,

    /** RTP end of flow. */
    PACE2_ADVANCED_JABBER_RTP_END_OF_FLOW,

    /** RTP invalid. */
    PACE2_ADVANCED_JABBER_RTP_INVALID,

    /** Signals that a file transfer was opened.
     * Provides a struct PACE2_jabber_generic_file_transfer_open pointer. */
    PACE2_ADVANCED_JABBER_FILE_TRANSFER_OPEN,

    /** Signals that a file transfer was closed.
     * Provides a struct PACE2_jabber_generic_file_transfer_info pointer. */
    PACE2_ADVANCED_JABBER_FILE_TRANSFER_CLOSE,

    /** Signals that data was send within a file transfer.
     * Provides a struct PACE2_jabber_generic_file_transfer_data pointer. */
    PACE2_ADVANCED_JABBER_FILE_TRANSFER_DATA,

    /** Signals that a stream host was sent.
     * Provides a struct PACE2_jabber_generic_file_transfer_stream_host pointer. */
    PACE2_ADVANCED_JABBER_FILE_TRANSFER_STREAM_HOST,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_JABBER_EVENT_TYPES

} PACE2_advanced_Jabber_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * Structure for a JABBER chat message.
 */
struct ipd_jabber_message_struct {

    /** The JABBER ID for the contact from which the message comes. */
    struct ipd_charset_string from_jid;

    /** The JABBER ID for the contact to which the message goes. */
    struct ipd_charset_string to_jid;

    /** The send message. */
    struct ipd_charset_string message;

    /** The ID of the chat room. */
    struct ipd_charset_string room_id;
};

/**
 * Structure for VoIP.
 */
struct ipd_jabber_voip_struct {

    /** The JABBER ID for the contact from which the voip part comes. */
    struct ipd_charset_string from_jid;

    /** The JABBER ID for the contact to which the voip part goes. */
    struct ipd_charset_string to_jid;

    /** The ID of the correlating audio and video session. */
    struct ipd_charset_string av_session_id;
};

/**
 * Structure for IBB announcements, opening, closure, and error messages.
 */
struct ipd_jabber_ibb_announcement_struct {

    /** The ID of the side sending the message/file. Could be empty if XEP-0065 is used or from field is not sent. */
    struct ipd_charset_string from;

    /** The ID of the side receiving the message/file. Could be empty if XEP-0065 is used or to field is not sent. */
    struct ipd_charset_string to;

    /** The identifier of the announced transfer. Could be empty if XEP-0065 is used. */
    struct ipd_charset_string ibb_sid;

    /** The name  of the transferred file. */
    struct ipd_charset_string file_name;

    /** The size of the transferred file. */
    u64 file_size;
};

/**
 * Structure for opening an IBB file transfer.
 */
struct ipd_jabber_ibb_open_filetransfer_struct {

    /** The ID of the side sending the message/file. Could be empty if XEP-0065 is used or from field is not sent. */
    struct ipd_charset_string from;

    /** The ID of the side receiving the message/file. Could be empty if XEP-0065 is used or to field is not sent. */
    struct ipd_charset_string to;

    /** The identifier of the announced transfer. Could be empty if XEP-0065 is used. */
    struct ipd_charset_string ibb_sid;

    /** The maximum block size for the transferred file. Could be 0 if XEP-0065 is used. */
    u32 maximum_block_size;
};

/**
 * Structure for IBB BASE-64 encoded data.
 */
struct ipd_jabber_ibb_data_struct {

    /** The ID of the side sending the message/file. Could be empty if XEP-0065 is used or from field is not sent. */
    struct ipd_charset_string from;

    /** The ID of the side receiving the message/file. Could be empty if XEP-0065 is used or to field is not sent. */
    struct ipd_charset_string to;

    /** The identifier of the announced transfer. Could be empty if XEP-0065 is used. */
    struct ipd_charset_string ibb_sid;

    /** The sequence number of the part of the transferred file. Could be 0 if XEP-0065 is used. */
    u32 sequence_number;

    /** The offset of the data from the beginning of the file. Could be 0 if XEP-0065 is used. */
    u64 offset;

    /** The BASE-64 encoded data of the transferred file. If XEP-0065 is used this field is empty. */
    struct ipd_charset_string data;

    /** The decoded raw data of the transferred file. */
    struct ipd_charset_string data_raw;
};

/**
 * Structure for jabber file transfer information.
 */
typedef struct ipd_jabber_ibb_announcement_struct PACE2_jabber_file_transfer_info;
typedef struct ipd_jabber_ibb_open_filetransfer_struct PACE2_jabber_file_transfer_open;
typedef struct ipd_jabber_ibb_data_struct PACE2_jabber_file_transfer_data;

/**
 * Types of jabber file transfers.
 */
enum PACE2_jabber_file_transfer_type {
    PACE2_JABBER_FILE_TRANSFER_XEP_0065,
    PACE2_NUMBER_OF_JABBER_FILE_TRANSFER_TYPES
};

/**
 * The generic structure of a file transfer in jabber.
 */
struct PACE2_jabber_generic_file_transfer_info {

    /** The file transfer type. */
    enum PACE2_jabber_file_transfer_type type;

    /** The information structure. */
    PACE2_jabber_file_transfer_info file_transfer_info;
};

/**
 * The generic structure of a file transfer open event in jabber.
 */
struct PACE2_jabber_generic_file_transfer_open {

    /** The file transfer type. */
    enum PACE2_jabber_file_transfer_type type;

    /** The file transfer open structure. */
    PACE2_jabber_file_transfer_open file_transfer_open;
};

/**
 * The generic structure of a file transfer data event in jabber.
 */
struct PACE2_jabber_generic_file_transfer_data {

    /** The file transfer type. */
    enum PACE2_jabber_file_transfer_type type;

    /** The file transfer data structure. */
    PACE2_jabber_file_transfer_data file_transfer_data;
};

/**
 * The structure of a stream host.
 */
struct PACE2_jabber_streamhost_struct
{
    /** The streamhost host name. */
    struct ipd_charset_string host;

    /** The jid of the streamhost. */
    struct ipd_charset_string jid;

    /** The port of the streamhost. */
    u32 port;
};

/**
 * The structure of a generic stream host event.
 */
struct PACE2_jabber_generic_file_transfer_stream_host
{
    /** The ID of the side sending the message/file. */
    struct ipd_charset_string from;

    /** The ID of the side receiving the message/file. */
    struct ipd_charset_string to;

    /** The identifier of the announced transfer. */
    struct ipd_charset_string sid;

    /** The found streamhost. */
    struct PACE2_jabber_streamhost_struct stream_host;
};

/**
 * @brief Data structure for advanced jabber events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_UNENCRYPTED_JABBER_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** Jabber event subtype. */
    PACE2_advanced_Jabber_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_LOG_ON event. */
        struct ipd_charset_string log_on;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_BUDDY_ADD event. */
        struct ipd_charset_string buddy_add;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_ADD_NEW_BUDDY event. */
        struct ipd_charset_string add_new_buddy;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_NEW_BUDDY_ACCEPT event. */
        struct ipd_charset_string new_buddy_accept;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_NEW_BUDDY_ACCEPT event. */
        struct ipd_charset_string new_buddy_cancel;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_USER_ON event. */
        struct ipd_charset_string user_on;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_USER_OFF event. */
        struct ipd_charset_string user_off;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_USER_DO_NOT_DISTURB event. */
        struct ipd_charset_string user_do_not_disturb;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_MESSAGE_OUT event. */
        struct ipd_jabber_message_struct message_out;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_MESSAGE_IN event. */
        struct ipd_jabber_message_struct message_in;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_BUDDY_CHANGE_STATUS event. */
        struct ipd_jabber_message_struct buddy_ch_status;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_CHANGE_STATUS event. */
        struct ipd_charset_string change_status;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_VOIP_ANNOUNCE event. */
        struct ipd_jabber_voip_struct voip_announce;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_VOIP_ACCEPT event. */
        struct ipd_jabber_voip_struct voip_accept;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_VOIP_REJECT event. */
        struct ipd_jabber_voip_struct voip_reject;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_VOIP_END event. */
        struct ipd_jabber_voip_struct voip_end;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_RTP_DATA event. */
        struct ipd_rtp_data rtp_data;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_INVALID event. */
        char const* invalid;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_STREAM_COMPRESSED event. */
        struct ipd_charset_string compression_type;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_COMPRESSED_DATA event. */
        struct ipd_charset_string compressed_data;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_RTP_SSRC event. */
        struct ipd_rtp_ssrc rtp_ssrc;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_IBB_OPEN_FILETRANSFER event. */
        struct ipd_jabber_ibb_open_filetransfer_struct ibb_open_filetransfer;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_IBB_CLOSE_FILETRANSFER event. */
        PACE2_jabber_file_transfer_info ibb_close_filetransfer;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_IBB_DATA event. */
        struct ipd_jabber_ibb_data_struct ibb_data;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_FILE_TRANSFER_ANNOUNCEMENT event. */
        PACE2_jabber_file_transfer_info file_transfer_announcement;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_FILE_TRANSFER_REJECT event. */
        PACE2_jabber_file_transfer_info file_transfer_reject;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_FILE_TRANSFER_REJECT event. */
        struct PACE2_jabber_generic_file_transfer_open file_transfer_open;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_FILE_TRANSFER_OPEN event. */
        struct PACE2_jabber_generic_file_transfer_info file_transfer_close;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_FILE_TRANSFER_DATA event. */
        struct PACE2_jabber_generic_file_transfer_data file_transfer_data;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_FILE_TRANSFER_STREAM_HOST event. */
        struct PACE2_jabber_generic_file_transfer_stream_host file_transfer_stream_host;

        /** Contains data for a @ref PACE2_ADVANCED_JABBER_RTP_INVALID event. */
        char const * rtp_invalid;
    } event_data;

} PACE2_advanced_jabber_event;

#endif

#endif /* JABBER_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
