/** @file      mail_events.h
  * @brief     Mail class event definitions.
  * @details   @copydoc doc_class_mail
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef MAIL_EVENTS_H
#define MAIL_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "generic_events.h"
#include "imf_events.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the mail class event.
 */
#define PACE2_CLASS_MAIL_EVENT_GROUP PACE2_CLASS_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for MAIL class events.
 */
typedef enum {

    /** Indicates an update for a mail session or a new mail session and provides
     * a struct ipd_class_mail_session_struct * with the generic, protocol type, status and a ID for
     * the correlating connection. */
    PACE2_CLASS_MAIL_SESSION,

    /** Provides a struct ipd_class_mail_session_string_struct * with the generic, the session ID
     * for the correlating session and the login name of the user. */
    PACE2_CLASS_MAIL_LOGIN,

    /** Provides a struct ipd_class_mail_session_string_struct * with the generic, the session ID
     * for the correlating session and the password of the user. */
    PACE2_CLASS_MAIL_PASS,

    /** Indicates an update for a mail transfer or a new mail transfer and provides
     * a struct ipd_class_mail_transfer_struct * with all information about the transfer. */
    PACE2_CLASS_MAIL_SESSION_TRANSFER,

    /** Provides a struct ipd_class_mail_transfer_body_struct * with the next part of the body by a mail transfer. */
    PACE2_CLASS_MAIL_SESSION_BODY_TRANSFER,

    /** Indicates that the mail transfer with the ID specified in the argument is done. For the webmail, POP3,
     * and SMTP decoders this means that the transfer ID won't be referenced in further events. The IMAP
     * decoder will use this ID again if the same message (identified by its message sequence number in the
     * mailbox) is referenced by another command. Webmail decoders are not guaranteed to emit a TRANSFER_END
     * event for all mail transfers. @c data points to a u64. */
    PACE2_CLASS_MAIL_SESSION_TRANSFER_END,

    /** Provides a struct ipd_class_mail_attachment_info_struct * with information about an attachment.
     * Can be thrown more than one time for a update of some values. */
    PACE2_CLASS_MAIL_ATTACHMENT_INFO,

    /** Provides a struct ipd_class_mail_attachment_content_struct * with one segment of a mail attachment. */
    PACE2_CLASS_MAIL_ATTACHMENT_CONTENT,

    /** Provides data if one segment of a mail attachment including metadata about the attachment. */
    PACE2_CLASS_MAIL_ATTACHMENT_CONTENT_EXT,

    /** Provides a struct ipd_class_mail_list_folder_struct * with content about the mail folder content
     * in a webmail session.*/
    PACE2_CLASS_MAIL_LIST_FOLDER,

    /** An event for AME lists containing events that are lost.
        That can be due to SMTP connection reset or because of having an e-mail sending denied. */
    PACE2_CLASS_MAIL_EVENTS_LOST,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_CLASS_MAIL_EVENT_TYPES

} PACE2_class_MAIL_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/** Enumeration for the mail protocol type. */
enum ipd_mail_class_type {

    /** Indicates a POP3 mail class type. */
    IPD_MAIL_PROTOCOL_CLASS_POP3,

    /** Indicates a SMTP mail class type. */
    IPD_MAIL_PROTOCOL_CLASS_SMTP,

    /** Indicates a IMAP mail class type. */
    IPD_MAIL_PROTOCOL_CLASS_IMAP,

    /** Reserved for future use. */
    IPD_MAIL_PROTOCOL_CLASS_RESERVED_1,

    /** Reserved for future use. */
    IPD_MAIL_PROTOCOL_CLASS_RESERVED_2,

    /** Reserved for future use. */
    IPD_MAIL_PROTOCOL_CLASS_RESERVED_3,

    /** Reserved for future use. */
    IPD_MAIL_PROTOCOL_CLASS_RESERVED_4,

    /** Reserved for future use. */
    IPD_MAIL_PROTOCOL_CLASS_RESERVED_5,

    /** Reserved for future use. */
    IPD_MAIL_PROTOCOL_CLASS_RESERVED_6,

    /** Reserved for future use. */
    IPD_MAIL_PROTOCOL_CLASS_RESERVED_7,
};

/** Enumeration for the mail session status. */
enum ipd_mail_class_session_status {

    /** Indicates that the session is started. */
    IPD_MAIL_CLASS_SESSION_START,

    /** Indicates that the session is ended. */
    IPD_MAIL_CLASS_SESSION_END,

    /** Used for Advanced Metadata Events which do not effect any Class Events.
     * It is active in the correlation feature only. */
    IPD_MAIL_CLASS_SESSION_UPDATE,
};

/**
 * Enumeration for the direction of a mail transfer.
 */
enum ipd_mail_class_direction_enum {
    IPD_MAIL_CLASS_DIRECTION_FETCH,
    IPD_MAIL_CLASS_DIRECTION_SEND,
    IPD_MAIL_CLASS_DIRECTION_STORE
};

/**
 * Structure for a mail session update.
 * Used for: #PACE2_CLASS_MAIL_SESSION.
 */
struct ipd_class_mail_session_struct {

    /** The generic. Contains the unique ID of this event @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The type of this mail session (protocol type) @see ipd_mail_class_type. */
    enum ipd_mail_class_type type;

    /** The status of this mail session @see ipd_mail_class_session_status. */
    enum ipd_mail_class_session_status status;

    /** The unique ID of the corresponding tcp connection of the mail. */
    u64 connection_id;

    /** The list of lost ID lists of the event correlation. */
    pace2_lost_id_lists * lost_id_lists;
};

/**
 * Structure for a information about the user, like login name and password \n
 * Used for: #PACE2_CLASS_MAIL_LOGIN, #PACE2_CLASS_MAIL_PASS.
 */
struct ipd_class_mail_session_string_struct {

    /** The generic. Contains the ID of this event @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The ID of the correlating session. */
    u64 mail_session_id;

    /** The information, which is send. */
    struct ipd_charset_string string;
};

/**
 * Structure for a mail transfer.
 * Used for: #PACE2_CLASS_MAIL_SESSION_TRANSFER.
 */
struct ipd_class_mail_transfer_struct {

    /** The generic. Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The ID of the correlating session. */
    u64 mail_session_id;

    /** The direction of this transfer. @see ipd_mail_class_direction_enum. */
    enum ipd_mail_class_direction_enum direction;

    /** The sender sender provided by the server (SMTP MAIL FROM). */
    struct ipd_charset_string envelope_from;

    /** The length of the envelope_to array. */
    u64 envelope_to_array_len;

    /** The array containing the recipients provided by the server (SMTP RCPT TO). */
    struct ipd_charset_string * envelope_to;

    /** The sender of the mail. */
    struct ipd_imf_address from;

    /** The length of the to array of "to" recipients. */
    u64 to_array_len;

    /** The array containing the "to" recipients of the mail. */
    struct ipd_imf_address * to;

    /** The length of the cc array. */
    u64 cc_array_len;

    /** The array containing the "cc" recipients of the mail. */
    struct ipd_imf_address * cc;

    /** The length of the bcc array. */
    u64 bcc_array_len;

    /** The array containing the "bcc" recipients of the mail. */
    struct ipd_imf_address * bcc;

    /** The length of the imf_received array. */
    u64 imf_received_array_len;

    /** The array containing the "imf received lines" of the mail. */
    struct ipd_charset_string * imf_received;

    /** The return-path contained in the header. */
    struct ipd_charset_string return_path;

    /** The length of the imf_other array. */
    u64 imf_others_array_len;

    /** The array containing the "imf other lines" of the mail. */
    struct ipd_charset_string * imf_others;

    /** The date in the protocol headerline. */
    struct ipd_charset_string date_headerline;

    /** The subject of the mail. */
    struct ipd_charset_string subject;

    /** The sender of the mail. */
    struct ipd_imf_address sender;

    /** The reply_to address of the mail. */
    struct ipd_imf_address reply_to;

    /** The message ID of the mail. */
    struct ipd_charset_string message_id;

    /** The in_reply_to address of the mail. */
    struct ipd_charset_string in_reply_to;

    /** The references header line of the mail. */
    struct ipd_charset_string references;

    /** The comments header line of the mail. */
    struct ipd_charset_string comments;

    /** The keywords header line of the mail. */
    struct ipd_charset_string keywords;

    /** The trace header line of the mail. */
    struct ipd_charset_string trace;

    /** The resent_date header line of the mail. */
    struct ipd_charset_string resent_date;

    /** The resent_from header line of the mail. */
    struct ipd_imf_address resent_from;

    /** The resent_sender header line of the mail. */
    struct ipd_imf_address resent_sender;

    /** The length of the resent_to array. */
    u64 resent_to_array_len;

    /** The array containing the "resent to" header lines of the mail. */
    struct ipd_imf_address * resent_to;

    /** The length of the resent_cc array.*/
    u64 resent_cc_array_len;

    /** The array containing the "resent cc" header lines of the mail. */
    struct ipd_imf_address * resent_cc;

    /** The length of the resent_bcc array. */
    u64 resent_bcc_array_len;

    /** The array containing the "resent bcc" header lines of the mail. */
    struct ipd_imf_address * resent_bcc;

    /** The resent_message_id header line of the mail. */
    struct ipd_charset_string resent_message_id;
};

/**
 * Structure for a body transfer.
 * Used for #PACE2_CLASS_MAIL_SESSION_BODY_TRANSFER.
 */
struct ipd_class_mail_transfer_body_struct {

    /** The generic. Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The ID of the mail session. */
    u64 mail_session_id;

    /** The ID of the correlating mail transfer. */
    u64 mail_transfer_id;

    /** The offset of this part of from the mail body. */
    u64 body_offset;

    /** The part of the body, which is send in this event. */
    struct ipd_charset_string body;

    /** The MIME content-type of this mail part. */
    struct ipd_charset_string content_type;

    /** The content-transfer-encoding of the data delivered in @ref ipd_class_mail_transfer_body_struct::body. */
    struct ipd_charset_string transfer_encoding;

    /** Direction of the mail transfer. */
    enum ipd_mail_class_direction_enum direction;

    /** A Flag that is set when the present event contains the final segment of the body data. */
    u8 body_transfer_finished;
};

/** The possible causes of an #PACE2_CLASS_MAIL_ATTACHMENT_INFO event. */
enum ipd_mail_class_attachment_info_origin {

    /** Information @em about an attachment (meta-information) has been observed. */
    IPD_MAIL_CLASS_ATTACHMENT_INFO_FROM_REFERENCE,

    /** The attachment content itself is transmitted. */
    IPD_MAIL_CLASS_ATTACHMENT_INFO_FROM_TRANSFER
};

/**
 * Structure for information about a mail attachment.
 * Used for: #PACE2_CLASS_MAIL_ATTACHMENT_INFO.
 *
 * Apart from the key (#generic) and the origin (#info_origin), all fields are optional and may be set to
 * the empty string or 0, respectively, if they are not known at the time the event is reported. If at a later time,
 * more information about the attachment becomes known, another #PACE2_CLASS_MAIL_ATTACHMENT_INFO event with
 * the same ID (#ipd_class_generic_header::id) will be sent, where some of the previously unknown fields will be defined.
 */
struct ipd_class_mail_attachment_info_struct {

    /** The generic. Contains the ID of this event @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** @brief The cause of this INFO event
        @details PACE2 generates #PACE2_CLASS_MAIL_ATTACHMENT_INFO events on two different occasions:
          @li When information @em about an attachment is observed (meta-information; for instance, a list of
              attachments in a "view message" window).
          @li When the attachment content itself is transmitted.

          In the former case, this field contains #IPD_MAIL_CLASS_ATTACHMENT_INFO_FROM_REFERENCE, in the
          latter case, #IPD_MAIL_CLASS_ATTACHMENT_INFO_FROM_TRANSFER
    */
    enum ipd_mail_class_attachment_info_origin info_origin;

    /** The ID of the mail session. */
    u64 mail_session_id;

    /** The ID of the correlating mail transfer. */
    u64 mail_transfer_id;

    /** Direction of the mail transfer. */
    enum ipd_mail_class_direction_enum direction;

    /** An unique identifier string. */
    struct ipd_charset_string identifier_string;

    /** A flag that states whether or not the @ref ipd_class_mail_attachment_info_struct::size field
     * of this ipd_class_mail_attachment_info_struct is set. */
    u8 size_is_set:1;

    /** The size of the attachment. Is set when this information is available and not an approximated value. */
    u64 size;

    /** The MIME content-type of the attachment. */
    struct ipd_charset_string content_type;

    /** The content-transfer-encoding of the data delivered in the #PACE2_CLASS_MAIL_ATTACHMENT_CONTENT event. */
    struct ipd_charset_string transfer_encoding;

    /** The file name of the attachment. */
    struct ipd_charset_string file_name;
};

/**
 * Structure for one segment of a mail attachment content transfer.
 * Used for: #PACE2_CLASS_MAIL_ATTACHMENT_CONTENT.
 */
struct ipd_class_mail_attachment_content_struct {

    /** The generic header. Contains the same ID like the correlating #PACE2_CLASS_MAIL_ATTACHMENT_INFO event.
     * See ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The ID of the mail session. */
    u64 mail_session_id;

    /** Direction of the mail transfer. */
    enum ipd_mail_class_direction_enum direction;

    /** Offset (relative to the start of the entire attachment) of the data in
     * @ref ipd_class_mail_attachment_content_struct::content. */
    u64 offset;

    /** One segment of the mail attachment. If the mail transfer encoding is base64 it will decoded. */
    struct ipd_charset_string content;

    /** A flag that is set when the present event contains the final segment of the attachment data. */
    u8 attachment_transfer_finished;
};

/**
 * Structure for a mail list folder entry.
 * Used for: ipd_class_mail_list_folder_struct.
 */
struct ipd_class_mail_list_folder_entry {

    /** The name of the folder. */
    struct ipd_charset_string folder_name;

    /** The subject of the entry. */
    struct ipd_charset_string subject;

    /** The information about from (name). */
    struct ipd_charset_string from_name;

    /** The information about from (mail address). */
    struct ipd_charset_string from_email;

    /** This flag signals, if a mail is unreaded or not. */
    u8 unread;

    /** The size of the mail, is set, if the information is available and not a approximated value. */
    u64 size;

    /** The timestamp of the mail, is set, if the information is available. */
    u64 timestamp;
};

/**
 * Structure for a mail list folder content.
 * Used for: #PACE2_CLASS_MAIL_LIST_FOLDER.
 */
struct ipd_class_mail_list_folder_struct {

    /** The generic. Contains the ID of this event @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The ID of the correlating mail session. */
    u64 mail_session_id;

    /** The name of the folder. */
    struct ipd_charset_string folder_name;

    /** The length of the entry array. */
    u64 entry_array_len;

    /** The array containing the mail summaries of the mails in this folder.
     * @see ipd_class_mail_list_folder_entry. */
    struct ipd_class_mail_list_folder_entry * entry;
};

/**
 * @brief Structure used to correlate all Advanced Metadata Events that are lost due to SMTP reset or other issues.
 */
typedef struct {

    /** The generic header. Contains the ID of this event @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** Current mail session id. */
    u64 mail_session_id;

    /** The list of lost ID lists of the event correlation. */
    pace2_lost_id_lists * lost_id_lists;

} PACE2_class_MAIL_events_lost;

/**
 * @brief Data structure for @ref PACE2_CLASS_MAIL_ATTACHMENT_CONTENT_EXT events.
 *
 * This event is created at the same time @ref PACE2_CLASS_MAIL_ATTACHMENT_CONTENT. It contains the same data
 * as the according @ref PACE2_CLASS_MAIL_ATTACHMENT_CONTENT but additionally contains the data already submitted
 * via the @ref PACE2_CLASS_MAIL_ATTACHMENT_INFO event related to the same attachment.
 *
 * This event will not be created if
 * @ref PACE2_config_s4_decoding.mail_class_enable_extended_attachment_content_events is disabled (default).
 */
typedef struct {

    /** Contains data for the related @ref PACE2_CLASS_MAIL_ATTACHMENT_INFO event. */
    struct ipd_class_mail_attachment_info_struct attachment_info;

    /** Contains the actual attachment content data. */
    struct ipd_class_mail_attachment_content_struct attachment_content;

} PACE2_class_MAIL_attachment_content_ext;

/**
 * @brief Data structure for MAIL class events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_CLASS_MAIL_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** MAIL event subtype. */
    PACE2_class_MAIL_event_type meta_data_type;

    PACE2_CLASS_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_CLASS_MAIL_SESSION event. */
        struct ipd_class_mail_session_struct session;

        /** Contains data for a @ref PACE2_CLASS_MAIL_LOGIN event. */
        struct ipd_class_mail_session_string_struct login;

        /** Contains data for a @ref PACE2_CLASS_MAIL_PASS event. */
        struct ipd_class_mail_session_string_struct password;

        /** Contains data for a @ref PACE2_CLASS_MAIL_SESSION_TRANSFER event. */
        struct ipd_class_mail_transfer_struct transfer;

        /** Contains data for a @ref PACE2_CLASS_MAIL_SESSION_BODY_TRANSFER event. */
        struct ipd_class_mail_transfer_body_struct body;

        /** Contains data for a @ref PACE2_CLASS_MAIL_SESSION_TRANSFER_END event. */
        PACE2_class_MAIL_events_lost transfer_end_ext;

        /** Contains data for a @ref PACE2_CLASS_MAIL_ATTACHMENT_INFO event. */
        struct ipd_class_mail_attachment_info_struct attachment_info;

        /** Contains data for a @ref PACE2_CLASS_MAIL_ATTACHMENT_CONTENT event. */
        struct ipd_class_mail_attachment_content_struct attachment_content;

        /** Contains data for a @ref PACE2_CLASS_MAIL_LIST_FOLDER event. */
        struct ipd_class_mail_list_folder_struct list_folder;

        /** Contains data for a @ref PACE2_CLASS_MAIL_ATTACHMENT_CONTENT_EXT event. */
        PACE2_class_MAIL_attachment_content_ext attachment_content_ext;

        /** Contains data for a @ref PACE2_CLASS_MAIL_EVENTS_LOST event. */
        PACE2_class_MAIL_events_lost events_lost;

    } event_data;

} PACE2_class_MAIL_event;

#endif

#endif /* MAIL_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
