/** @file      mdns_events.h
  * @brief     mDNS event definitions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef MDNS_EVENTS_H
#define MDNS_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Groups *****************************************************************/

#define PACE2_BASIC_MDNS_EVENT_GROUP PACE2_BASIC_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/* Event Structures *************************************************************/

/**
 * @brief Public define for the maximum question section record count.
 */
#define PACE2_MAX_NUMBER_OF_MDNS_QD_COUNT 15

/**
 * @brief Public define for the maximum answer section record count.
 */
#define PACE2_MAX_NUMBER_OF_MDNS_AN_COUNT 15

/**
 * @brief Public define for the maximum authority section record count.
 */
#define PACE2_MAX_NUMBER_OF_MDNS_NS_COUNT 15

/**
 * @brief Public define for the maximum additional section record count.
 */
#define PACE2_MAX_NUMBER_OF_MDNS_AR_COUNT 15

/**
 * @brief Structure for the mDNS packet header.
 */
struct mdns_header {
    u16 transaction_id; /**< Transaction ID for the Name Service transaction. */
    u16 flags;          /**< The flags field of the dns query/response. */
    u16 qdcount;        /**< Number of question section entries. */
    u16 ancount;        /**< Number of answer section entries. */
    u16 nscount;        /**< Number of authority section entries. */
    u16 arcount;        /**< Number of additional section entries. */
};

/**
 * @brief Public define for the maximum number of subdomains.
 */
#define PACE2_MAX_NUMBER_OF_MDNS_SUBDOMAIN_COUNT 6

/**
 * @brief Structure for the mDNS name. The mDNS name consists of label.
 */
struct mdns_name {
    PACE2_byte_buffer label[PACE2_MAX_NUMBER_OF_MDNS_SUBDOMAIN_COUNT];
    u8 lcount;
};

/**
 * @brief Structure for the mDNS question record.
 */
struct mdns_question {
    struct mdns_name qname;    /**< Name of the node. */
    u16 qtype;                 /**< Type of request. */
    u16 qclass;                /**< Class of request. */
};

/**
 * @brief Structure for the mDNS resource record.
 */
struct mdns_rrecord {
    struct mdns_name rrname;    /**< Name of the node. */
    u16 rrtype;                 /**< Resource record type code. */
    u16 rrclass;                /**< Resource record class code. */
    u32 ttl;                    /**< Time To Live of a resource records name. */
    PACE2_byte_buffer rdata;    /**< Resource record information data. Depends on @ref rrtype and @ref rrclass. */
};

/**
 * @brief Data structure for basic mDNS events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_BASIC_MDNS_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** Header field. */
    struct mdns_header header;

    /** Array of question section records. */
    struct mdns_question qdrecords[PACE2_MAX_NUMBER_OF_MDNS_QD_COUNT];

    /** Array of answer section records. */
    struct mdns_rrecord anrecords[PACE2_MAX_NUMBER_OF_MDNS_AN_COUNT];

    /** Array of authority section records. */
    struct mdns_rrecord nsrecords[PACE2_MAX_NUMBER_OF_MDNS_NS_COUNT];

    /** Array of additional section records. */
    struct mdns_rrecord arrecords[PACE2_MAX_NUMBER_OF_MDNS_AR_COUNT];
} PACE2_basic_mdns_event;

#endif /* MDNS_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
