/** @file      ml_events.h
  * @brief     Generic machine learning event definitions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef ML_EVENTS_H
#define ML_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "pace2_ml_class.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the machine learning event.
 */
#define PACE2_ML_CLASSIFICATION_EVENT_GROUP PACE2_NO_GROUP

/**
 * @brief Public define for the assigned metadata group of the machine learning deregister event.
 */
#define PACE2_ML_DEREGISTER_EVENT_GROUP PACE2_NO_GROUP

/* Event Structures *************************************************************/

/**
 * @brief Data structure of the machine learning classification event.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_ML_CLASSIFICATION_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    u32 category_id; /*!< Identifier for the ML Category the event belongs to. Further information about the category
                          can be derived via @ref pace2_ml_get_category_info. */
    u32 class_id;    /*!< Identifier for the ML Class the event belongs to. Further information about the class can be
                          derived via @ref pace2_ml_get_class_info. */
    u32 subclass_id; /*!< Identifier for the ML Subclass the event belongs to. Further information about the subclass
                          can be derived via @ref pace2_ml_get_subclass_info. */

    double probability; /*!< Summarizes the likelihood of an event belonging to the specified subclass */

    /**
     * An indicator of the trustworthiness of the classification result based on the prediction probability of all
     * subclasses.
     */
    double confidence_level;

    /**
     * An indicator of the trustworthiness of the classification result based on the prediction probability of all
     * subclasses, also taking the subclass precision into account.
     */
    double confidence_level_with_precision;

} PACE2_ml_classification_event;

/**
 * @brief Data structure for the machine learning class deregister event.
 *
 * This event is triggered if a machine learning class is deregistered for a flow. This indicates that the current flow
 * will not trigger a @ref PACE2_ml_classification_event for the class that is specified within the event.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_ML_DEREGISTER_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    u32 class_id;    /*!< Identifier for the ML Class the event belongs to. Further information about the class can be
                          derived via @ref pace2_ml_get_class_info. */

} PACE2_ml_deregister_event;

#endif /* ML_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
