/** @file      mmse_events.h
  * @brief     MMSE event definitions.
  * @details   @copydoc doc_mmse
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef MMSE_EVENTS_H
#define MMSE_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Group ******************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced MMSE event.
 */
#define PACE2_ADVANCED_MMSE_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/
/**
 * @brief Types for advanced MMSE events.
 */
typedef enum {

    /** Start of a new flow with MMSE traffic; payload: struct ipd_5tuple. */
    PACE2_ADVANCED_MMSE_NEW_CONNECTION,

    /** The message type of the current message. payload: struct ipd_mmse_message_type_struct. */
    PACE2_ADVANCED_MMSE_PDU_MESSAGE_TYPE,

    /** One header field of the intercepted Mms-Message-Type; payload: struct ipd_mmse_pdu_value. */
    PACE2_ADVANCED_MMSE_PDU_HEADERLINE,

    /** End of header fields; payload: struct ipd_mmse_header_end_struct. */
    PACE2_ADVANCED_MMSE_PDU_HEADER_END,

    /** Start of multiparts with information of multipart_amount;
     * payload: struct ipd_mmse_multipart_start_struct. */
    PACE2_ADVANCED_MMSE_PDU_DATA_MULTIPART_START,

    /** The content type of this multipart; payload: struct ipd_mmse_multipart_content_type_struct. */
    PACE2_ADVANCED_MMSE_PDU_DATA_MULTIPART_PART_CONTENT_TYPE,

    /** One multipart header field of this multipart; payload: struct ipd_mmse_multipart_header_struct. */
    PACE2_ADVANCED_MMSE_PDU_DATA_MULTIPART_PART_HEADERLINE,

    /** The content field of this multipart; payload: struct ipd_mmse_multipart_content_struct. */
    PACE2_ADVANCED_MMSE_PDU_DATA_MULTIPART_PART_CONTENT,

    /** End of this multipart; payload: struct ipd_mmse_multipart_end_struct. */
    PACE2_ADVANCED_MMSE_PDU_DATA_MULTIPART_PART_END,

    /** End of this MMSE message; payload: struct ipd_mmse_message_end_struct. */
    PACE2_ADVANCED_MMSE_PDU_MESSAGE_END,

    /** End of the MMSE flow; no payload. */
    PACE2_ADVANCED_MMSE_END_CONNECTION,

    /** The MMSE flow is set to invalid. The error message is passed as a <tt>char *</tt> in the @c data argument. */
    PACE2_ADVANCED_MMSE_INVALID,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_MMSE_EVENT_TYPES

} PACE2_advanced_mmse_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/** MMSE PDU Header field types. */
enum ipd_mmse_pdu_type_enum {

    /** The BCC value. The value will be a struct ipd_charset_string. */
    IPD_MMSE_PDU_MM_BCC = 0x01,

    /** The CC value. The value will be a struct ipd_charset_string. */
    IPD_MMSE_PDU_MM_CC = 0x02,

    /** The content location. The value will be a struct ipd_mmse_pdu_content_location_struct.
        Only if the field "message_type_is_mbox_delete_conf" is set, the field "status_count_value" is valid. */
    IPD_MMSE_PDU_MM_CONTENT_LOCATION = 0x03,

    /** The content type. The value will be a struct ipd_mmse_pdu_content_type_struct with the given type
     * is placed in a buffer and an enumeration. */
    IPD_MMSE_PDU_MM_CONTENT_TYPE = 0x04,

    /** The date value. This will be a u64 integer with the timestamp to the date. */
    IPD_MMSE_PDU_MM_DATE = 0x05,

    /** The value of the delivery report will be an enum ipd_mmse_status_enum with "yes" or "no". */
    IPD_MMSE_PDU_MM_DELIVERY_REPORT = 0x06,

    /** The value of the delivery time will be a struct ipd_mmse_pdu_delivery_time_struct.
        The enumeration token decides whether in the union "delivery_time" is a date (timestamp) or
        a delta in seconds. */
    IPD_MMSE_PDU_MM_DELIVERY_TIME = 0x07,

    /** The value of the expiry will be a struct ipd_mmse_pdu_expiry_struct.
        The enumeration token decides whether in the union "expiry" is a date (timestamp) or a delta in seconds. */
    IPD_MMSE_PDU_MM_EXPIRY = 0x08,

    /** The from value. This will be a struct ipd_mmse_pdu_from_struct. If the enumeration is set to
     * "insert address", the field "from_value" will be empty. */
    IPD_MMSE_PDU_MM_FROM = 0x09,

    /** The message class value will be a struct ipd_mmse_pdu_message_class_struct. The enumeration corresponds the value.
        If the enumeration is set to token text, the "token_text" field will be set with a string value. */
    IPD_MMSE_PDU_MM_MESSAGE_CLASS = 0x0A,

    /** The message ID will be a struct ipd_string value. */
    IPD_MMSE_PDU_MM_MESSAGE_ID = 0x0B,

    /** The message type value will be a struct ipd_mmse_message_type_struct and will be thrown in a IPD_EVENT_MMSE_PDU_MESSAGE_TYPE event. */
    IPD_MMSE_PDU_MM_MESSAGE_TYPE = 0x0C,

    /** The version will be a struct ipd_mmse_version_struct. There will be a major and a minor number.
        The flag "is_minor_version_set" indicates whether the minor number is present or not. */
    IPD_MMSE_PDU_MM_VERSION = 0x0D,

    /** The message size value will be a u64. */
    IPD_MMSE_PDU_MM_MESSAGE_SIZE = 0x0E,

    /** The priority will be an enumeration ipd_mmse_priority_enum. */
    IPD_MMSE_PDU_MM_PRIORITY = 0x0F,

    /** The value of read repowrt will be an enum ipd_mmse_status_enum with "yes" or "no". */
    IPD_MMSE_PDU_MM_READ_REPORT = 0x10,

    /** The value of report allowed will be an enum ipd_mmse_status_enum with "yes" or "no". */
    IPD_MMSE_PDU_MM_REPORT_ALLOWED = 0x11,

    /** The response status value will be an enum ipd_mmse_response_status_enum. */
    IPD_MMSE_PDU_MM_RESPONSE_STATUS = 0x12,

    /** The response text. The value will be a struct ipd_mmse_pdu_response_text_struct.
        Only if the field "message_type_is_mbox_delete_conf" is set, the field "status_count_value" is valid. */
    IPD_MMSE_PDU_MM_RESPONSE_TEXT = 0x13,

    /** The sender visibility will be an enum ipd_mmse_sender_visibility_enum. */
    IPD_MMSE_PDU_MM_SENDER_VISIBILITY = 0x14,

    /** The status value will be an enum ipd_mmse_message_status_enum. */
    IPD_MMSE_PDU_MM_STATUS = 0x15,

    /** The subject will be a struct ipd_charset_string. */
    IPD_MMSE_PDU_MM_SUBJECT = 0x16,

    /** The to value will be a struct ipd_charset_string. */
    IPD_MMSE_PDU_MM_TO = 0x17,

    /** The transaction ID will be a struct ipd_string. */
    IPD_MMSE_PDU_MM_TRANSACTION_ID = 0x18,

    /** The retrieve status will be an enum ipd_mmse_retrieve_status_enum. */
    IPD_MMSE_PDU_MM_RETRIEVE_STATUS = 0x19,

    /** The retrieve text will be a struct ipd_charset_string. */
    IPD_MMSE_PDU_MM_RETRIEVE_TEXT = 0x1A,

    /** The read status value will be an enum ipd_mmse_read_status_enum. */
    IPD_MMSE_PDU_MM_READ_STATUS = 0x1B,

    /** The reply charging value will be an enum ipd_mmse_reply_charging_enum. */
    IPD_MMSE_PDU_MM_REPLY_CHARGING = 0x1C,

    /** The value of the reply charging deadline will be a struct ipd_mmse_pdu_reply_charging_deadline_struct.
        The enumeration token decides whether in the union "reply_charging_deadline" is a date (timestamp)
        or a delta in seconds. */
    IPD_MMSE_PDU_MM_REPLY_CHARGING_DEADLINE = 0x1D,

    /** The reply charging ID will be a struct ipd_string. */
    IPD_MMSE_PDU_MM_REPLY_CHARGING_ID = 0x1E,

    /** The reply charging size will be a u64. */
    IPD_MMSE_PDU_MM_REPLY_CHARGING_SIZE = 0x1F,

    /** The previously send by value will be a struct ipd_mmse_pdu_previously_sent_by_struct. */
    IPD_MMSE_PDU_MM_PREV_SENT_BY = 0x20,

    /** The previously send date value will be a struct ipd_mmse_pdu_previously_sent_date_struct. */
    IPD_MMSE_PDU_MM_PREV_SENT_DATE = 0x21,

    /** The value of store will be an enum ipd_mmse_status_enum with "yes" or "no". */
    IPD_MMSE_PDU_MM_STORE = 0x22,

    /** The value of state will be an enum ipd_mmse_pdu_state_enum. */
    IPD_MMSE_PDU_MM_STATE = 0x23,

    /** The value of flags will be a struct ipd_mmse_pdu_flags_struct. */
    IPD_MMSE_PDU_MM_FLAGS = 0x24,

    /** The value of store status will be an enum ipd_mmse_store_status_enum. */
    IPD_MMSE_PDU_MM_STORE_STATUS = 0x25,

    /** The store status text will be a struct ipd_charset_string. */
    IPD_MMSE_PDU_MM_STORE_STATUS_TEXT = 0x26,

    /** The value of stored will be an enum ipd_mmse_status_enum with "yes" or "no". */
    IPD_MMSE_PDU_MM_STORED = 0x27,

    /** The attributes will be an enum ipd_mmse_pdu_type_enum. */
    IPD_MMSE_PDU_MM_ATTRIBUTES = 0x28,

    /** The value of totals will be an enum ipd_mmse_status_enum with "yes" or "no". */
    IPD_MMSE_PDU_MM_TOTALS = 0x29,

    /** The mbox totals will be a struct ipd_mmse_pdu_mbox_totals_struct. */
    IPD_MMSE_PDU_MM_MBOX_TOTALS = 0x2A,

    /** The value of quotas will be an enum ipd_mmse_status_enum with "yes" or "no". */
    IPD_MMSE_PDU_MM_QUOTAS = 0x2B,

    /** The mbox quotas will be a struct ipd_mmse_pdu_mbox_quotas_struct. */
    IPD_MMSE_PDU_MM_MBOX_QUOTAS = 0x2C,

    /** The message count will be a u64. */
    IPD_MMSE_PDU_MM_MESSAGE_COUNT = 0x2D,

    /** The content value will not be thrown in any event. */
    IPD_MMSE_PDU_MM_CONTENT = 0x2E,

    /** The start value will be a u64. */
    IPD_MMSE_PDU_MM_START = 0x2F,

    /** The additional headers will not be thrown in any event. */
    IPD_MMSE_PDU_MM_ADDITIONAL_HEADERS = 0x30,

    /** The value of distribution indicator will be an enum ipd_mmse_status_enum with "yes" or "no". */
    IPD_MMSE_PDU_MM_DISTRIBUTION_INDICATOR = 0x31,

    /** The element descriptor will be a struct ipd_mmse_pdu_element_descriptor_struct. */
    IPD_MMSE_PDU_MM_ELEMENT_DESCRIPTOR = 0x32,

    /** The limit will be a u64. */
    IPD_MMSE_PDU_MM_LIMIT = 0x33,

    IPD_NUMBER_OF_MMSE_PDUS,
};

/** MMS message types.*/
enum ipd_mmse_message_type_enum {

    /** Indicates that a MM message is sent. */
    IPD_MMSE_MESSAGE_TYPE_M_SEND_REQ = 0x80,

    /** The acknowledgement of the server that the MM message was sent to the receiver. */
    IPD_MMSE_MESSAGE_TYPE_M_SEND_CONF = 0x81,

    /** The server indicates the client that there are messages on the server. */
    IPD_MMSE_MESSAGE_TYPE_M_NOTIFICATION_IND = 0x82,

    /** The acknowledgement of the client to the server, that the notifications has arrived. */
    IPD_MMSE_MESSAGE_TYPE_M_NOTIFYRESP_IND = 0x83,

    /** The server sends the MM message to the client, who requested the MM message. */
    IPD_MMSE_MESSAGE_TYPE_M_RETRIEVE_CONF = 0x84,

    /** The client confirms the delivery of the MM message. */
    IPD_MMSE_MESSAGE_TYPE_M_ACKNOWLEDGE_IND = 0x85,

    /** The server sends a delivery report to the sending client, if this report was not denied. */
    IPD_MMSE_MESSAGE_TYPE_M_DELIVERY_IND = 0x86,

    /** The receiver of a message sends a read report. */
    IPD_MMSE_MESSAGE_TYPE_M_READ_REC_IND = 0x87,

    /** The originating side of the message sends a read report. */
    IPD_MMSE_MESSAGE_TYPE_M_READ_ORIG_IND = 0x88,

    /** The client requests a forwarding of a MM message. */
    IPD_MMSE_MESSAGE_TYPE_M_FORWARD_REQ = 0x89,

    /** The server indicates the status of the forwarding to the client. */
    IPD_MMSE_MESSAGE_TYPE_M_FORWARD_CONF = 0x8A,

    /** The client requests the storing of a message which was not retrieved yet. */
    IPD_MMSE_MESSAGE_TYPE_M_MBOX_STORE_REQ = 0x8B,

    /** The acknowledgement of the server that the message was stored. */
    IPD_MMSE_MESSAGE_TYPE_M_MBOX_STORE_CONF = 0x8C,

    /** The client requests information about all stored messages in the mbox. */
    IPD_MMSE_MESSAGE_TYPE_M_MBOX_VIEW_REQ = 0x8D,

    /** The server sends the requested information to the client. */
    IPD_MMSE_MESSAGE_TYPE_M_MBOX_VIEW_CONF = 0x8E,

    /** The client uploads a message to the mbox. */
    IPD_MMSE_MESSAGE_TYPE_M_MBOX_UPLOAD_REQ = 0x8F,

    /** The acknowledgement of the server that the message was stored. */
    IPD_MMSE_MESSAGE_TYPE_M_MBOX_UPLOAD_CONF = 0x90,

    /** The client deletes messages from the mbox. */
    IPD_MMSE_MESSAGE_TYPE_M_MBOX_DELETE_REQ = 0x91,

    /** The acknowledgement of the server that the messages were deleted. */
    IPD_MMSE_MESSAGE_TYPE_M_MBOX_DELETE_CONF = 0x92,

    /** This type can be used with a M-Mbox-View.conf and a M-Mbox-Upload.req to describe the messages. */
    IPD_MMSE_MESSAGE_TYPE_M_MBOX_DESCR = 0x93,

    /** Indicates an undefined message type. */
    IPD_MMSE_MESSAGE_TYPE_M_UNDEFINED = 0x00,
};

/**
 * Structure for a MMSE message type event.
 */
struct ipd_mmse_message_type_struct {

    /** The ID of the current message. */
    u64 message_id;

    /** The enumeration of the message type. */
    enum ipd_mmse_message_type_enum message_type;
};

/**
 * Structure for a MMSE header end event.
 */
struct ipd_mmse_header_end_struct {

    /** The ID of the current message. */
    u64 message_id;
};

/**
 * Structure for a MMSE header end event.
 */
struct ipd_mmse_multipart_start_struct {

    /** The ID of the current message. */
    u64 message_id;

    /** The amount of the multiparts in the content of the MMSE-message */
    u32 count_of_multiparts;
};

/**
 * Structure for a MMSE multipart end event.
 */
struct ipd_mmse_multipart_end_struct {

    /** The ID of the current message. */
    u64 message_id;

    /** The ID of the current multipart. */
    u64 multipart_id;
};

/**
 * Structure for a MMSE message end event.
 */
struct ipd_mmse_message_end_struct {

    /** The ID of the current message. */
    u64 message_id;
};

/** Enumeration for a "yes"/"no" choice. */
enum ipd_mmse_status_enum
{
    IPD_MMSE_STATUS_YES = 0x80,
    IPD_MMSE_STATUS_NO = 0x81,
};

/** MMS message class types. */
enum ipd_mmse_message_class_enum
{
    IPD_MMSE_MESSAGE_CLASS_PERSONAL = 0x80,
    IPD_MMSE_MESSAGE_CLASS_ADVERTISEMENT = 0x81,
    IPD_MMSE_MESSAGE_CLASS_INFORMATIONAL = 0x82,
    IPD_MMSE_MESSAGE_CLASS_AUTO = 0x83,
    IPD_MMSE_MESSAGE_CLASS_TOKEN_TEXT = 0xFF,
};

/** MMS priority types. */
enum ipd_mmse_priority_enum
{
    IPD_MMSE_PRIORITY_LOW = 0x80,
    IPD_MMSE_PRIORITY_NORMAL = 0x81,
    IPD_MMSE_PRIORITY_HIGH = 0x82,
};

/** MMS response status types. */
enum ipd_mmse_response_status_enum
{
    IPD_MMSE_RESPONSE_STATUS_OK = 0x80,
    IPD_MMSE_RESPONSE_STATUS_UNSPECIFIED = 0x81,
    IPD_MMSE_RESPONSE_STATUS_SERVICE_DENIED = 0x82,
    IPD_MMSE_RESPONSE_STATUS_MESSAGE_FORMAT_CORRUPT = 0x83,
    IPD_MMSE_RESPONSE_STATUS_SENDING_ADDRESS_UNRESOLVED = 0x84,
    IPD_MMSE_RESPONSE_STATUS_MESSAGE_NOT_FOUND = 0x85,
    IPD_MMSE_RESPONSE_STATUS_NETWORK_PROBLEM = 0x86,
    IPD_MMSE_RESPONSE_STATUS_CONTENT_NOT_ACCEPTED = 0x87,
    IPD_MMSE_RESPONSE_STATUS_UNSUPPORTED_MESSAGE = 0x88,
    IPD_MMSE_RESPONSE_STATUS_TRANSIENT_FAILURE = 0xc0,
    IPD_MMSE_RESPONSE_STATUS_TRANSIENT_SENDING_ADDRESS_UNRESOLVED = 0xc1,
    IPD_MMSE_RESPONSE_STATUS_TRANSIENT_MESSAGE_NOT_FOUND = 0xc2,
    IPD_MMSE_RESPONSE_STATUS_TRANSIENT_NETWORK_PROBLEM = 0xc3,
    IPD_MMSE_RESPONSE_STATUS_TRANSIENT_PARTIAL_SUCCESS = 0xc4,
    IPD_MMSE_RESPONSE_STATUS_PERMANENT_FAILURE = 0xe0,
    IPD_MMSE_RESPONSE_STATUS_PERMANENT_SERVICE_DENIED = 0xe1,
    IPD_MMSE_RESPONSE_STATUS_PERMANENT_MESSAGE_FORMAT_CORRUPT = 0xe2,
    IPD_MMSE_RESPONSE_STATUS_PERMANENT_SENDING_ADDRESS_UNRESOLVED = 0xe3,
    IPD_MMSE_RESPONSE_STATUS_PERMANENT_MESSAGE_NOT_FOUND = 0xe4,
    IPD_MMSE_RESPONSE_STATUS_PERMANENT_CONTENT_NOT_ACCEPTED = 0xe5,
    IPD_MMSE_RESPONSE_STATUS_PERMANENT_REPLY_CHARGING_LIMITATION_NOT_MET = 0xe6,
    IPD_MMSE_RESPONSE_STATUS_PERMANENT_REPLY_CHARGING_REQUEST_NOT_ACCEPTED = 0xe7,
    IPD_MMSE_RESPONSE_STATUS_PERMANENT_REPLY_CHARGING_FORWARDING_DENIED = 0xe8,
    IPD_MMSE_RESPONSE_STATUS_PERMANENT_REPLY_CHARGING_NOT_SUPPORTED = 0xe9,
    IPD_MMSE_RESPONSE_STATUS_PERMANENT_ADDRESS_HIDING_NOT_SUPPORTED = 0xea
};

/** MMS sender visibility types. */
enum ipd_mmse_sender_visibility_enum {
    IPD_MMSE_SENDER_VISIBILITY_HIDE = 0x80,
    IPD_MMSE_SENDER_VISIBILITY_SHOW = 0x81,
};

/** MMS retrieve status types. */
enum ipd_mmse_retrieve_status_enum
{
    IPD_MMSE_RETRIEVE_STATUS_OK = 0x80,
    IPD_MMSE_RETRIEVE_STATUS_TRANSIENT_FAILURE = 0xc0,
    IPD_MMSE_RETRIEVE_STATUS_TRANSIENT_MESSAGE_NOT_FOUND = 0xc1,
    IPD_MMSE_RETRIEVE_STATUS_TRANSIENT_NETWORK_PROBLEM = 0xc2,
    IPD_MMSE_RETRIEVE_STATUS_PERMANENT_FAILURE = 0xe0,
    IPD_MMSE_RETRIEVE_STATUS_PERMANENT_SERVICE_DENIED = 0xe1,
    IPD_MMSE_RETRIEVE_STATUS_PERMANENT_MESSAGE_NOT_FOUND = 0xe2,
    IPD_MMSE_RETRIEVE_STATUS_PERMANENT_CONTENT_UNSUPPORTED = 0xe3
};

/** MMS read status types. */
enum ipd_mmse_read_status_enum
{
    IPD_MMSE_READ_STATUS_READ = 0x80,
    IPD_MMSE_READ_STATUS_DELETED_WITHOUT_BEING_READ = 0x81,
};

/** MMS reply charging types. */
enum ipd_mmse_reply_charging_enum
{
    IPD_MMSE_REPLY_CHARGING_REQUESTED = 0x80,
    IPD_MMSE_REPLY_CHARGING_REQUESTED_TEXT_ONLY = 0x81,
    IPD_MMSE_REPLY_CHARGING_ACCEPTED = 0x82,
    IPD_MMSE_REPLY_CHARGING_ACCEPTED_TEXT_ONLY = 0x83,
};

/** MMS message status types. */
enum ipd_mmse_message_status_enum
{
    IPD_MMSE_MESSAGE_STATUS_EXPIRED = 0x80,
    IPD_MMSE_MESSAGE_STATUS_RETRIEVED = 0x81,
    IPD_MMSE_MESSAGE_STATUS_REJECTED = 0x82,
    IPD_MMSE_MESSAGE_STATUS_DEFERRED = 0x83,
    IPD_MMSE_MESSAGE_STATUS_UNRECOGNIZED = 0x84,
    IPD_MMSE_MESSAGE_STATUS_INDETERMINATE = 0x85,
    IPD_MMSE_MESSAGE_STATUS_FORWARDED = 0x86,
    IPD_MMSE_MESSAGE_STATUS_UNREACHABLE = 0x87,
};

/** MMS Content types defined in WAPWSP 8.4.2.24 April 30 1998 (Appendix A, table 40). */
enum ipd_wap_content_type_enum
{

    /** Indicates all types. */
    IPD_MMSE_MMS_CONTENT_TYPE_ALL = 0x00,

    /** Indicates 'text/'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_ALL = 0x01,

    /** Indicates 'text/html'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_HTML = 0x02,

    /** Indicates 'text/plain'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_PLAIN = 0x03,

    /** Indicates 'text/x-hdml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_X_HDML = 0x04,

    /** Indicates 'text/x-ttml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_X_TTML = 0x05,

    /** Indicates 'text/x-vCalendar'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_X_CALENDAR = 0x06,

    /** Indicates 'text/x-vCard'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_V_CARD = 0x07,

    /** Indicates 'text/vnd.wap.wml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_VND_WAP_WML = 0x08,

    /** Indicates 'text/vnd.wap.wmlscript'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_VND_WAP_WMLSCRIPT = 0x09,

    /** Indicates 'text/vnd.wap.wta-event'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_VND_WAP_EVENT = 0x0A,

    /** Indicates 'multipart/'. */
    IPD_MMSE_MMS_CONTENT_TYPE_MULTIPART_ALL = 0x0B,

    /** Indicates 'multipart/mixed'. */
    IPD_MMSE_MMS_CONTENT_TYPE_MULTIPART_MIXED = 0x0C,

    /** Indicates 'multipart/form-data'. */
    IPD_MMSE_MMS_CONTENT_TYPE_MULTIPART_FORM_DATA = 0x0D,

    /** Indicates 'multipart/byteranges'. */
    IPD_MMSE_MMS_CONTENT_TYPE_MULTIPART_BYTERANGES = 0x0E,

    /** Indicates 'multipart/alternative'. */
    IPD_MMSE_MMS_CONTENT_TYPE_MULTIPART_ALTERNATIVE = 0x0F,

    /** Indicates 'application/'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_ALL = 0x10,

    /** Indicates 'application/java-vm'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_JAVA_VM = 0x11,

    /** Indicates 'application/x-www-form-urlencoded'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_X_WWW_FORM_URLENCODED = 0x12,

    /** Indicates 'application/x-hdmlc'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_X_HDMLC = 0x13,

    /** Indicates 'application/vnd.wap.wmlc'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_WMLC = 0x14,

    /** Indicates 'application/vnd.wap.wmlscriptc'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_WMLSCRIPTC = 0x15,

    /** Indicates 'application/vnd.wap.wta-eventc'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_WTA_EVENTC = 0x16,

    /** Indicates 'application/vnd.wap.uaprof'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_UAPROF = 0x17,

    /** Indicates 'application/vnd.wap.wtls-ca-certificate'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_WTLS_CA_CERTIFICATE = 0x18,

    /** Indicates 'application/vnd.wap.wtls-user-certificate'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_WTLS_USER_CERTIFICATE = 0x19,

    /** Indicates 'application/x-x509-ca-cert'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_X_X509_CA_CERT = 0x1A,

    /** Indicates 'application/x-x509-user-cert'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_X_X509_USER_CERT = 0x1B,

    /** Indicates 'image/'. */
    IPD_MMSE_MMS_CONTENT_TYPE_IMAGE_ALL = 0x1C,

    /** Indicates 'image/gif'. */
    IPD_MMSE_MMS_CONTENT_TYPE_IMAGE_GIF = 0x1D,

    /** Indicates 'image/jpeg'. */
    IPD_MMSE_MMS_CONTENT_TYPE_IMAGE_JPEG = 0x1E,

    /** Indicates 'image/tiff'. */
    IPD_MMSE_MMS_CONTENT_TYPE_IMAGE_TIFF = 0x1F,

    /** Indicates 'image/png'. */
    IPD_MMSE_MMS_CONTENT_TYPE_IMAGE_PNG = 0x20,

    /** Indicates 'image/vnd.wap.wbmp'. */
    IPD_MMSE_MMS_CONTENT_TYPE_IMAGE_VND_WAP_WBMP = 0x21,

    /** Indicates 'application/vnd.wap.multipart.*'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_MULTIPART_ALL = 0x22,

    /** Indicates 'application/vnd.wap.multipart.mixed'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_MULTIPART_MIXED = 0x23,

    /** Indicates 'application/vnd.wap.multipart.form-data'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_MULTIPART_FORM_DATA = 0x24,

    /** Indicates 'application/vnd.wap.multipart.byteranges'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_MULTIPART_BYTERANGES = 0x25,

    /** Indicates 'application/vnd.wap.multipart.alternative'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_MULTIPART_ALTERNATIVE = 0x26,

    /** Indicates 'application/xml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_TEXT_XML = 0x27,

    /** Indicates 'text/xml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_XML = 0x28,

    /** Indicates 'application/vnd.wap.wbxml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_WBXML = 0x29,

    /** Indicates 'application/x-x968-cross-cert'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_X_X968_CROSS_CERT = 0x2A,

    /** Indicates 'application/x-x968-ca-cert'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_X_X968_CA_CERT = 0x2B,

    /** Indicates 'application/x-x968-user-cert'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_X_X968_USER_CERT = 0x2C,

    /** Indicates 'text/vnd.wap.si'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_VND_WAP_SI = 0x2D,

    /** Indicates 'application/vnd.wap.sic'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_SIC = 0x2E,

    /** Indicates 'text/vnd.wap.sl'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_VND_WAP_SL = 0x2F,

    /** Indicates 'application/vnd.wap.slc',. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_SLC = 0x30,

    /** Indicates 'text/vnd.wap.co'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_VND_WAP_CO = 0x31,

    /** Indicates 'application/vnd.wap.coc'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_COC = 0x32,

    /** Indicates 'application/vnd.wap.multipart.related'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_MULTIPART_RELATED = 0x33,

    /** Indicates 'application/vnd.wap.sia'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_SIA = 0x34,

    /** Indicates 'text/vnd.wap.connectivity-xml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_VND_WAP_CONNECTIVITY_XML = 0x35,

    /** Indicates 'application/vnd.wap.connectivity-wbxml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_CONNECTIVITY_WBXML = 0x36,

    /** Indicates 'application/pkcs7-mime'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_PKCS7_MIME = 0x37,

    /** Indicates 'application/vnd.wap.hashed-certificate'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_HASHED_CERTIFICATE = 0x38,

    /** Indicates 'application/vnd.wap.signed-certificate'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_SIGNED_CERTIFICATE = 0x39,

    /** Indicates 'application/vnd.wap.cert-response'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_CERT_RESPONSE = 0x3A,

    /** Indicates 'application/xhtml+xml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_XHTML_XML = 0x3B,

    /** Indicates 'application/wml+xml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_WML_XML = 0x3C,

    /** Indicates 'text/css'. */
    IPD_MMSE_MMS_CONTENT_TYPE_TEXT_CSS = 0x3D,

    /** Indicates 'application/vnd.wap.mms-message'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_MMS_MESSAGE = 0x3E,

    /** Indicates 'application/vnd.wap.rollover-certificate'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_ROLLOVER_CERTIFICATE = 0x3F,

    /** Indicates 'application/vnd.wap.locc+wbxml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_LOCC_WBXML = 0x40,

    /** Indicates 'application/vnd.wap.loc+xml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_LOC_XML = 0x41,

    /** Indicates 'application/vnd.syncml.dm+wbxml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_SYNCML_DM_WBXML = 0x42,

    /** Indicates 'application/vnd.syncml.dm+xml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_SYNCML_DM_XML = 0x43,

    /** Indicates 'application/vnd.syncml.notification'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_SYNCML_NOTIFICATION = 0x44,

    /** Indicates 'application/vnd.wap.xhtml+xml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WAP_XHTML_XML = 0x45,

    /** Indicates 'application/vnd.wv.csp.cir'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WV_CSP_CIR = 0x46,

    /** Indicates 'application/vnd.oma.dd+xml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_OMA_DD_XML = 0x47,

    /** Indicates 'application/vnd.oma.drm.message'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_OMA_DRM_MESSAGE = 0x48,

    /** Indicates 'application/vnd.oma.drm.content'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_OMA_DRM_CONTENT = 0x49,

    /** Indicates 'application/vnd.oma.drm.rights+xml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_OMA_DRM_RIGHT_XML = 0x4A,

    /** Indicates 'application/vnd.oma.drm.rights+wbxml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_OMA_DRM_RIGHT_WBXML = 0x4B,

    /** Indicates 'application/vnd.wv.csp+xml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WV_CSP_XML = 0x4C,

    /** Indicates 'application/vnd.wv.csp+wbxml'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_WV_CSP_WBXML = 0x4D,

    /** Indicates 'application/vnd.syncml.ds.notification'. */
    IPD_MMSE_MMS_CONTENT_TYPE_APPLICATION_VND_SYNCML_DS_NOTIFICATION = 0x4E,

    /** Indicates 'audio/'. */
    IPD_MMSE_MMS_CONTENT_TYPE_AUDIO = 0x4F,

    /** Indicates 'video/'. */
    IPD_MMSE_MMS_CONTENT_TYPE_VIDEO = 0x50,

    /** Should be last.. */
    IPD_MMSE_MMS_CONTENT_TYPE_UNKNOWN = 0xFF,
};

/** Enumeration for the parameter types. */
enum ipd_wap_content_parameters_enum
{
    IPD_MMSE_MMS_PARAMETER_Q = 0x00,
    IPD_MMSE_MMS_PARAMETER_CHARSET = 0x01,
    IPD_MMSE_MMS_PARAMETER_LEVEL = 0x02,
    IPD_MMSE_MMS_PARAMETER_TYPE_DEP = 0x03,
    IPD_MMSE_MMS_PARAMETER_UAPROF = 0x04,
    IPD_MMSE_MMS_PARAMETER_NAME = 0x05,
    IPD_MMSE_MMS_PARAMETER_FILENAME = 0x06,
    IPD_MMSE_MMS_PARAMETER_DIFFERENCES = 0x07,
    IPD_MMSE_MMS_PARAMETER_PADDING = 0x08,
    IPD_MMSE_MMS_PARAMETER_TYPE_TEXT = 0x09,
    IPD_MMSE_MMS_PARAMETER_START = 0xA,
    IPD_MMSE_MMS_PARAMETER_START_INFO = 0xB,
    IPD_MMSE_MMS_PARAMETER_UNKNOWN = 0xFF,
};

/** Enumeration of the WSP header fields in the multiparts. */
enum ipd_wap_wsp_header_field_enum
{
    IPD_MMSE_MMS_WSP_HEADER_CONTENT_ID = 0x40,
    IPD_MMSE_MMS_WSP_HEADER_CONTENT_LOCATION = 0x0E,
    IPD_MMSE_MMS_WSP_HEADER_CONTENT_DISPOSITION_OBSOLETE = 0x2E,
    IPD_MMSE_MMS_WSP_HEADER_CONTENT_DISPOSITION = 0x45,
    IPD_MMSE_MMS_WSP_HEADER_UNKNOWN = 0xFF
};

/** Structure for a MMSE version. */
struct ipd_mmse_version_struct {

    /** The major version number. This has to be between 1 and 7. */
    u8 major_version_number;

    /** The minor version number. This has to be between 0 and 14. */
    u8 minor_version_number;

    /** Indicates whether the minor number is valid or not. */
    u8 is_minor_version_set;
};

/** Structure for a parameter value. */
struct ipd_mmse_pdu_parameter {

    /** The parameter type as a buffer. */
    struct ipd_string parameter_type_buffer;

    /** The corresponding enumeration. */
    enum ipd_wap_content_parameters_enum parameter_type_enum;

    /** Union of the indicatores of some parameter values. */
    union {

        /** Indicates whether the integer_value (0) or the text_value (1) is used with an untyped type. */
        u8 untyped_type;

        /** Indicates whether the version value as a struct (0) or the version_string (1) is used with
         * the version type. */
        u8 version_type;

        /** Indicates whether the integer_value (0) or the text_value (1) is used with the differences type. */
        u8 differences_type;

        /** Indicates whether the integer_value (0) or the text_value (1) is used with the type_text type. */
        u8 type_text_type;

    } ext_value_type;

    /** Union of the parameter values. */
    union {

        /** The text value as a struct ipd_string. */
        struct ipd_string text_value;

        /** The integer value as a u64. */
        u64 integer_value;

        /** The timestamp to a date as a u64. */
        u64 date_value;

        /** The delta in seconds as a u64. */
        u64 delta_seconds_value;

        /** The Q-value as a u32. */
        u32 q_value;

        /** The URI value as a struct ipd_string. */
        struct ipd_string uri_value;

        /** Union of the version value. */
        union {

            /** The version as a struct ipd_mmse_version_struct. */
            struct ipd_mmse_version_struct version;

            /** The string version. */
            struct ipd_string version_string;

        } version_value;

        /** The charset value as an enumeration. */
        enum ipd_charset charset_value;

    } parameter_value;
};

/** Structure for a parameter list. */
struct ipd_mmse_parameter_struct {

    /** The list of all parameters. */
    struct ipd_mmse_pdu_parameter * parameters;

    /** The length of the list. */
    u64 parameter_length;
};

/** Structure for a content type. */
struct ipd_mmse_pdu_content_type_struct {

    /** The content type as a buffer. */
    struct ipd_string content_type_buffer;

    /** The corresponding enumeration. */
    enum ipd_wap_content_type_enum content_type_enum;

    /** The parameter list as a struct ipd_mmse_parameter_struct. */
    struct ipd_mmse_parameter_struct parameter;
};

/** Structure for a content location. */
struct ipd_mmse_pdu_content_location_struct {

    /** Indicates whether the message type is M-Mbox-Delete.conf or not. */
    u8 message_type_is_mbox_delete_conf;

    /** The status count is set only if the flag "message_type_is_mbox_delete_conf" is set to 1. */
    u64 status_count_value;

    /** The content location as a struct ipd_string. */
    struct ipd_string content_location;
};

/** Enumeration for the distinction of absolute or relative values. */
enum ipd_mmse_pdu_token_enum {
    IPD_MMSE_ABSOLUTE_TOKEN = 0x80,
    IPD_MMSE_RELATIVE_TOKEN = 0x81,
};

/** Structure for a delivery time value. */
struct ipd_mmse_pdu_delivery_time_struct {

    /** Indicates whether the date_value (IPD_MMSE_ABSOLUTE_TOKEN) or the delta_seconds_value (IPD_MMSE_RELATIVE_TOKEN) is used. */
    enum ipd_mmse_pdu_token_enum token;

    /** The union for the delivery time. */
    union {

        /** The timestamp to the date as a u64. */
        u64 date_value;

        /** The delta in seconds as a u64. */
        u64 delta_seconds_value;

    } delivery_time;
};

/** Structure for a element descriptor. */
struct ipd_mmse_pdu_element_descriptor_struct {

    /** The content reference value as a struct ipd_string. */
    struct ipd_string content_reference_value;

    /** The parameter list as a struct ipd_mmse_parameter_struct. */
    struct ipd_mmse_parameter_struct parameter;
};

/** Structure for a expiry value. */
struct ipd_mmse_pdu_expiry_struct {

    /** Indicates whether the date_value (IPD_MMSE_ABSOLUTE_TOKEN) or the delta_seconds_value (IPD_MMSE_RELATIVE_TOKEN) is used. */
    enum ipd_mmse_pdu_token_enum token;

    /** The union for the expiry. */
    union {

        /** The timestamp to the date as a u64. */
        u64 date_value;

        /** The delta in seconds as a u64. */
        u64 delta_seconds_value;

    } expiry;
};

/** Enumeration whether a address value is present or has to be inserted. */
enum ipd_mmse_pdu_address_enum {
    IPD_MMSE_ADDRESS_PRESENT_TOKEN = 0x80,
    IPD_MMSE_INSERT_ADDRESS_TOKEN = 0x81,
};

/** Structure of a from value. */
struct ipd_mmse_pdu_from_struct {

    /** Indicates whether the from value is present (IPD_MMSE_ADDRESS_PRESENT_TOKEN) or
     * has to be inserted (IPD_MMSE_INSERT_ADDRESS_TOKEN). */
    enum ipd_mmse_pdu_address_enum address_type;

    /** The from value if IPD_MMSE_ADDRESS_PRESENT_TOKEN is present. */
    struct ipd_charset_string from_value;
};

/** Structure of a previously sent by value. */
struct ipd_mmse_pdu_previously_sent_by_struct {

    /** The forwarded count value as a u64. */
    u64 forwarded_count_value;

    /** The string value as a struct ipd_charset_string. */
    struct ipd_charset_string string;
};

/** Structure of a previously sent date value. */
struct ipd_mmse_pdu_previously_sent_date_struct {

    /** The forwarded count value as a u64. */
    u64 forwarded_count_value;

    /** The timestamp to the date as a u64. */
    u64 date;
};

/** Enumeration for a quotas value. */
enum ipd_mmse_pdu_quota_enum {
    IPD_MMSE_MESSAGE_QUOTA_TOKEN = 0x80,
    IPD_MMSE_SIZE_QUOTA_TOKEN = 0x81
};

/** Structure for a mbox quotas value. */
struct ipd_mmse_pdu_mbox_quotas_struct {

    /** Indicates the type of the quota as an enumeration. */
    enum ipd_mmse_pdu_quota_enum quota;

    /** The integer value as a u64. */
    u64 integer_value;
};

/** Enumeration for a totals value. */
enum ipd_mmse_pdu_total_enum {
    IPD_MMSE_MESSAGE_TOTAL_TOKEN = 0x80,
    IPD_MMSE_SIZE_TOTAL_TOKEN = 0x81
};

/** Structure for a mbox totals value. */
struct ipd_mmse_pdu_mbox_totals_struct {

    /** Indicates the type of the totals as an enumeration. */
    enum ipd_mmse_pdu_total_enum total;

    /** The integer value as a u64. */
    u64 integer_value;
};

/** Structure for a message class value. */
struct ipd_mmse_pdu_message_class_struct {

    /** The message class value as an enumeration. Indicates if the field token_text is set. */
    enum ipd_mmse_message_class_enum message_class;

    /** The value of the message class if the enumeration is set to IPD_MMSE_MESSAGE_CLASS_TOKEN_TEXT. */
    struct ipd_string token_text;
};

/** Enumeration for an action. */
enum ipd_mmse_pdu_token_action_enum {
    IPD_MMSE_ADD_TOKEN = 0x80,
    IPD_MMSE_REMOVE_TOKEN = 0x81,
    IPD_MMSE_FILTER_TOKEN = 0x82,
};

/** Structure for a flag value. */
struct ipd_mmse_pdu_flags_struct {

    /** The action as an enumeration. */
    enum ipd_mmse_pdu_token_action_enum token;

    /** The flags as a struct ipd_charset_string. */
    struct ipd_charset_string flags;
};

/** Enumeration for the state value. */
enum ipd_mmse_pdu_state_enum {
    IPD_MMSE_STATE_DRAFT = 0x80,
    IPD_MMSE_STATE_SENT = 0x81,
    IPD_MMSE_STATE_NEW = 0x82,
    IPD_MMSE_STATE_RETRIEVED = 0x83,
    IPD_MMSE_STATE_FORWARDED = 0x84,
};

/** Structure for a reply charging deadline value. */
struct ipd_mmse_pdu_reply_charging_deadline_struct {

    /** Indicates whether the date_value (IPD_MMSE_ABSOLUTE_TOKEN) or the delta_seconds_value (IPD_MMSE_RELATIVE_TOKEN) is used. */
    enum ipd_mmse_pdu_token_enum token;

    /** The union for the reply charging deadline. */
    union {

        /** The timestamp to the date as a u64. */
        u64 date_value;

        /** The delta in seconds as a u64. */
        u64 delta_seconds_value;

    } reply_charging_deadline;
};

/** Structure for a response text value. */
struct ipd_mmse_pdu_response_text_struct {

    /** Indicates whether the message type is M-Mbox-Delete.conf or not. */
    u8 message_type_is_mbox_delete_conf;

    /** The status count is set only if the flag "message_type_is_mbox_delete_conf" is set to 1. */
    u64 status_count_value;

    /** The content location as a struct ipd_charset_string. */
    struct ipd_charset_string response_text;
};

/** Enumeration of the store status. */
enum ipd_mmse_store_status_enum {
    IPD_MMSE_STORE_STATUS_SUCCESS = 0x80,
    IPD_MMSE_STORE_STATUS_TRANSIENT_FAILURE = 0xC0,
    IPD_MMSE_STORE_STATUS_TRANSIENT_NETWORK_PROBLEM = 0xC1,
    IPD_MMSE_STORE_STATUS_PERMANENT_FAILURE = 0xE0,
    IPD_MMSE_STORE_STATUS_PERMANENT_SERVICE_DENIED = 0xE1,
    IPD_MMSE_STORE_STATUS_PERMANENT_MESSAGE_FORMAT_CORRUPT = 0xE2,
    IPD_MMSE_STORE_STATUS_PERMANENT_MESSAGE_NOT_FOUND = 0xE3,
    IPD_MMSE_STORE_STATUS_PERMANENT_MMBOX_FULL = 0xE4,
};

/** Enumeration of the disposition status. */
enum ipd_mmse_disposition_enum {
    IPD_MMSE_DISPOSITION_FORM_DATA = 0x80,
    IPD_MMSE_DISPOSITION_ATTACHMENT = 0x81,
    IPD_MMSE_DISPOSITION_INLINE = 0x82,
    IPD_MMSE_DISPOSITION_TOKEN_TEXT = 0xFF,
};

/** Structure for a content disposition value. */
struct ipd_mmse_content_disposition_struct {

    /** The disposition value as an enumeration. Indicates whether the field disposition_text is set or not. */
    enum ipd_mmse_disposition_enum disposition;

    /** The disposition value if the enumeration is set to IPD_MMSE_DISPOSITION_TOKEN_TEXT. */
    struct ipd_string disposition_text;

    /** The parameter list as a struct ipd_mmse_parameter_struct. */
    struct ipd_mmse_parameter_struct parameter;
};

/** Structure for a pdu mmse header field value. */
struct ipd_mmse_pdu_value {

    /** The ID of the current message. */
    u64 message_id;

    /** The enumeration of the header field value. */
    enum ipd_mmse_pdu_type_enum pdu_type;

    /** The union of the pdu header values. */
    union {

        /** The attributes as an enumeration. */
        enum ipd_mmse_pdu_type_enum attributes;

        /** The BCC value as a struct ipd_charset_string. */
        struct ipd_charset_string bcc;

        /** The CC value as a struct ipd_charset_string. */
        struct ipd_charset_string cc;

        /** The content location as a struct ipd_mmse_pdu_content_location_struct. */
        struct ipd_mmse_pdu_content_location_struct content_location;

        /** The content type as a struct ipd_mmse_pdu_content_type_struct. */
        struct ipd_mmse_pdu_content_type_struct content_type;

        /** The timstamp to the date as a u64. */
        u64 date;

        /** The delivery report as an enumeration. */
        enum ipd_mmse_status_enum delivery_report;

        /** The delivery time as a struct ipd_mmse_pdu_delivery_time_struct. */
        struct ipd_mmse_pdu_delivery_time_struct delivery_time;

        /** The distribution indicator as an enumeration. */
        enum ipd_mmse_status_enum distribution_indicator;

        /** The element descriptor as a struct ipd_mmse_pdu_element_descriptor_struct. */
        struct ipd_mmse_pdu_element_descriptor_struct element_descriptor;

        /** The expiry as a struct ipd_mmse_pdu_expiry_struct. */
        struct ipd_mmse_pdu_expiry_struct expiry;

        /** The from value as a struct ipd_mmse_pdu_from_struct. */
        struct ipd_mmse_pdu_from_struct from;

        /** The limit as a u64. */
        u64 limit;

        /** The previously sent by value as a struct ipd_mmse_pdu_previously_sent_by_struct. */
        struct ipd_mmse_pdu_previously_sent_by_struct previously_sent_by;

        /** The previously sent date as a struct ipd_mmse_pdu_previously_sent_date_struct. */
        struct ipd_mmse_pdu_previously_sent_date_struct previously_sent_date;

        /** The mbox quotas as a struct ipd_mmse_pdu_mbox_quotas_struct. */
        struct ipd_mmse_pdu_mbox_quotas_struct mbox_quotas;

        /** The mbox totals as a struct ipd_mmse_pdu_mbox_totals_struct. */
        struct ipd_mmse_pdu_mbox_totals_struct mbox_totals;

        /** The message_class as a struct ipd_mmse_pdu_message_class_struct. */
        struct ipd_mmse_pdu_message_class_struct message_class;

        /** The message count as a u64. */
        u64 message_count;

        /** The message ID as a struct ipd_string. */
        struct ipd_string message_id;

        /** The message size as a u64. */
        u64 message_size;

        /** The flags as a struct ipd_mmse_pdu_flags_struct. */
        struct ipd_mmse_pdu_flags_struct flags;

        /** The state value as an enumeration. */
        enum ipd_mmse_pdu_state_enum state;

        /** The version value as a struct ipd_mmse_version_struct. */
        struct ipd_mmse_version_struct version;

        /** The priority as an enumeration. */
        enum ipd_mmse_priority_enum priority;

        /** The quotas as an enumeration. */
        enum ipd_mmse_status_enum quotas;

        /** The read report value as an enumeration. */
        enum ipd_mmse_status_enum read_report;

        /** The read status as an enumeration. */
        enum ipd_mmse_read_status_enum read_status;

        /** The reply charging value as an enumeration. */
        enum ipd_mmse_reply_charging_enum reply_charging;

        /** The reply charging deadline as a struct ipd_mmse_pdu_reply_charging_deadline_struct. */
        struct ipd_mmse_pdu_reply_charging_deadline_struct reply_charging_deadline;

        /** The reply charging ID as a struct ipd_string. */
        struct ipd_string reply_charging_id;

        /** The reply charging size as a u64. */
        u64 reply_charging_size;

        /** The report allowed value as an enumeration. */
        enum ipd_mmse_status_enum report_allowed;

        /** The response status value as an enumeration. */
        enum ipd_mmse_response_status_enum response_status;

        /** The response text as a struct ipd_mmse_pdu_response_text_struct. */
        struct ipd_mmse_pdu_response_text_struct response_text;

        /** The retrieve status as an enumeration. */
        enum ipd_mmse_retrieve_status_enum retrieve_status;

        /** The retrieve text as a struct ipd_charset_string. */
        struct ipd_charset_string retrieve_text;

        /** The sender visibility as an enumeration. */
        enum ipd_mmse_sender_visibility_enum sender_visibility;

        /** The start value as a u64. */
        u64 start;

        /** The status value as an enumeration. */
        enum ipd_mmse_message_status_enum status;

        /** The store value as an enumeration. */
        enum ipd_mmse_status_enum store;

        /** The stored value as an enumeration. */
        enum ipd_mmse_status_enum stored;

        /** The store_status as an enumeration. */
        enum ipd_mmse_store_status_enum store_status;

        /** The store status text as a struct ipd_charset_string. */
        struct ipd_charset_string store_status_text;

        /** The subject as a struct ipd_charset_string. */
        struct ipd_charset_string subject;

        /** The to value as a struct ipd_charset_string. */
        struct ipd_charset_string to;

        /** The totals as an enumeration. */
        enum ipd_mmse_status_enum totals;

        /** The transaction ID as a struct ipd_string. */
        struct ipd_string transaction_id;

    } pdu_value;
};

/**
 * Structure for a MMSE multipart header field value.
 */
struct ipd_mmse_multipart_header_struct {

    /** The ID of the current message. */
    u64 message_id;

    /** The ID of the current multipart. */
    u64 multipart_id;

    /** The multipart header field enumeration. */
    enum ipd_wap_wsp_header_field_enum header_field;

    /** The union of the multipart header value */
    union {

        /** The content ID as a struct ipd_string. */
        struct ipd_string content_id;

        /** The content location as a struct ipd_string. */
        struct ipd_string content_location;

        /** The content disposition as a struct ipd_mmse_content_disposition_struct. */
        struct ipd_mmse_content_disposition_struct content_disposition;

    } header_value;
};

/**
 * Structure for a MMSE multipart content type.
 */
struct ipd_mmse_multipart_content_type_struct {

    /** The ID of the current message. */
    u64 message_id;

    /** The ID of the current multipart. */
    u64 multipart_id;

    /** The content type */
    struct ipd_mmse_pdu_content_type_struct content_type;
};

/**
 * Structure for a MMSE multipart content.
 */
struct ipd_mmse_multipart_content_struct {

    /** The ID of the current message. */
    u64 message_id;

    /** The ID of the current multipart. */
    u64 multipart_id;

    /** The multipart data content. */
    struct ipd_string content;
};


/**
 * @brief Data structure for advanced MMSE events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_MMSE_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** MMSE event subtype. */
    PACE2_advanced_mmse_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_MMSE_NEW_CONNECTION event. */
        struct ipd_5tuple connection_5tupel;

        /** Contains data for a @ref PACE2_ADVANCED_MMSE_PDU_MESSAGE_TYPE event. */
        struct ipd_mmse_message_type_struct message_type;

        /** Contains data for a @ref PACE2_ADVANCED_MMSE_PDU_HEADERLINE event. */
        struct ipd_mmse_pdu_value headerline;

        /** Contains data for a @ref PACE2_ADVANCED_MMSE_PDU_HEADER_END event. */
        struct ipd_mmse_header_end_struct header_end;

        /** Contains data for a @ref PACE2_ADVANCED_MMSE_PDU_DATA_MULTIPART_START event. */
        struct ipd_mmse_multipart_start_struct multipart_start;

        /** Contains data for a @ref PACE2_ADVANCED_MMSE_PDU_DATA_MULTIPART_PART_CONTENT_TYPE event. */
        struct ipd_mmse_multipart_content_type_struct multipart_content_type;

        /** Contains data for a @ref PACE2_ADVANCED_MMSE_PDU_DATA_MULTIPART_PART_HEADERLINE event. */
        struct ipd_mmse_multipart_header_struct multipart_headerline;

        /** Contains data for a @ref PACE2_ADVANCED_MMSE_PDU_DATA_MULTIPART_PART_CONTENT event. */
        struct ipd_mmse_multipart_content_struct multipart_content;

        /** Contains data for a @ref PACE2_ADVANCED_MMSE_PDU_DATA_MULTIPART_PART_END event. */
        struct ipd_mmse_multipart_end_struct multipart_end;

        /** Contains data for a @ref PACE2_ADVANCED_MMSE_PDU_MESSAGE_END event. */
        struct ipd_mmse_message_end_struct message_end;

        /** Contains data for a @ref PACE2_ADVANCED_MMSE_INVALID event. */
        char const * invalid;

    } event_data;

} PACE2_advanced_mmse_event;

#endif /* PACE_KERNEL_LIB */

#endif /* MMSE_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
