/** @file      modbus_events.h
  * @brief     Modbus event definitions.
  * @details   @copydoc doc_modbus
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef MODBUS_EVENTS_H
#define MODBUS_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced Modbus event.
 */
#define PACE2_ADVANCED_MODBUS_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced Modbus events.
 */
typedef enum {
    PACE2_ADVANCED_MODBUS_NEW_CONNECTION,
    PACE2_ADVANCED_MODBUS_CONNECTION_END,
    PACE2_ADVANCED_MODBUS_MESSAGE_START,
    PACE2_ADVANCED_MODBUS_MESSAGE_END,
    PACE2_ADVANCED_MODBUS_READ_COILS_REQUEST,
    PACE2_ADVANCED_MODBUS_READ_COILS_RESPONSE,
    PACE2_ADVANCED_MODBUS_READ_DISCRETE_INPUTS_REQUEST,
    PACE2_ADVANCED_MODBUS_READ_DISCRETE_INPUTS_RESPONSE,
    PACE2_ADVANCED_MODBUS_READ_HOLDING_REGISTERS_REQUEST,
    PACE2_ADVANCED_MODBUS_READ_HOLDING_REGISTERS_RESPONSE,
    PACE2_ADVANCED_MODBUS_READ_INPUT_REGISTERS_REQUEST,
    PACE2_ADVANCED_MODBUS_READ_INPUT_REGISTERS_RESPONSE,
    PACE2_ADVANCED_MODBUS_WRITE_SINGLE_COIL_REQUEST,
    PACE2_ADVANCED_MODBUS_WRITE_SINGLE_COIL_RESPONSE,
    PACE2_ADVANCED_MODBUS_WRITE_SINGLE_REGISTER_REQUEST,
    PACE2_ADVANCED_MODBUS_WRITE_SINGLE_REGISTER_RESPONSE,
    PACE2_ADVANCED_MODBUS_WRITE_MULTIPLE_COILS_REQUEST,
    PACE2_ADVANCED_MODBUS_WRITE_MULTIPLE_COILS_RESPONSE,
    PACE2_ADVANCED_MODBUS_WRITE_MULTIPLE_REGISTERS_REQUEST,
    PACE2_ADVANCED_MODBUS_WRITE_MULTIPLE_REGISTERS_RESPONSE,
    PACE2_ADVANCED_MODBUS_REPORT_SERVER_ID_REQUEST,
    PACE2_ADVANCED_MODBUS_REPORT_SERVER_ID_RESPONSE,
    PACE2_ADVANCED_MODBUS_READ_FILE_RECORD_REQUEST,
    PACE2_ADVANCED_MODBUS_READ_FILE_RECORD_RESPONSE,
    PACE2_ADVANCED_MODBUS_WRITE_FILE_RECORD_REQUEST,
    PACE2_ADVANCED_MODBUS_WRITE_FILE_RECORD_RESPONSE,
    PACE2_ADVANCED_MODBUS_MASK_WRITE_REGISTER_REQUEST,
    PACE2_ADVANCED_MODBUS_MASK_WRITE_REGISTER_RESPONSE,
    PACE2_ADVANCED_MODBUS_READ_WRITE_MULTIPLE_REGISTERS_REQUEST,
    PACE2_ADVANCED_MODBUS_READ_WRITE_MULTIPLE_REGISTERS_RESPONSE,
    PACE2_ADVANCED_MODBUS_READ_FIFO_QUEUE_REQUEST,
    PACE2_ADVANCED_MODBUS_READ_FIFO_QUEUE_RESPONSE,
    PACE2_ADVANCED_MODBUS_EXCEPTION_RESPONSE,
    PACE2_ADVANCED_MODBUS_USER_DEFINED_FUNCTION,
    PACE2_ADVANCED_MODBUS_INVALID,
    PACE2_NUMBER_OF_ADVANCED_MODBUS_EVENT_TYPES
} PACE2_advanced_modbus_event_type;

/**
 * @brief Modbus Commands which the decoder can process.
 */
typedef enum {
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_READ_COILS                    = 0x01,
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_READ_DISCRETE_INPUTS          = 0x02,
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_READ_HOLDING_REGISTERS        = 0x03,
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_READ_INPUT_REGISTERS          = 0x04,
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_WRITE_SINGLE_COIL             = 0x05,
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_WRITE_SINGLE_REGISTER         = 0x06,
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_WRITE_MULTIPLE_COILS          = 0x0F,
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_WRITE_MULTIPLE_REGISTERS      = 0x10,
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_REPORT_SERVER_ID              = 0x11,
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_READ_FILE_RECORD              = 0x14,
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_WRITE_FILE_RECORD             = 0x15,
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_MASK_WRITE_REGISTER           = 0x16,
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_READ_WRITE_MULTIPLE_REGISTERS = 0x17,
    PACE2_ADVANCED_MODBUS_COMMAND_TYPE_READ_FIFO_QUEUE               = 0x18
} PACE2_advanced_modbus_command_type;

/**
 * @brief Modbus exception codes.
 */
typedef enum {
    PACE2_ADVANCED_MODBUS_EXCEPTION_CODE_ILLEGAL_FUNCTION      = 0x01,
    PACE2_ADVANCED_MODBUS_EXCEPTION_CODE_ILLEGAL_DATA_ADDRESS  = 0x02,
    PACE2_ADVANCED_MODBUS_EXCEPTION_CODE_ILLEGAL_DATA_VALUE    = 0x03,
    PACE2_ADVANCED_MODBUS_EXCEPTION_CODE_SERVER_DEVICE_FAILURE = 0x04,
    PACE2_ADVANCED_MODBUS_EXCEPTION_CODE_ACKNOWLEDGE           = 0x05,
    PACE2_ADVANCED_MODBUS_EXCEPTION_CODE_SERVER_DEVICE_BUSY    = 0x06,
    PACE2_ADVANCED_MODBUS_EXCEPTION_CODE_MEMORY_PARITY_ERROR   = 0x08
} PACE2_advanced_modbus_exception_code;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * @brief Event type for a (address,quantity)-tuple.
 */
typedef struct pace2_modbus_address_quantity {
    u16 address;  /**< Starting address */
    u16 quantity; /**< Quantity */
} PACE2_modbus_address_quantity;

/**
 * @brief Event type for a (address,value)-tuple.
 */
typedef struct pace2_modbus_address_value {
    u16 address; /**< Output Address */
    u16 value;   /**< Output Value */
} PACE2_modbus_address_value;

/**
 * @brief Event type for a (address,quantity,value)-triple.
 */
typedef struct pace2_modbus_address_quantity_value {
    u16 address;                     /**< Starting Address */
    u16 quantity;                    /**< Quantity of Outputs */
    struct ipd_charset_string value; /**< Value */
} PACE2_modbus_address_quantity_value;


/**
 * @brief Event type for mask write register request and response
 */
typedef struct pace2_modbus_mask_write_register {
    u16 address;  /**< Starting Address */
    u16 and_mask; /**< AND Mask */
    u16 or_mask;  /**< OR Mask */
} PACE2_modbus_mask_write_register;

/**
 * @brief Modbus/TCP header data
 */
typedef struct pace2_modbus_message_start {
    PACE2_advanced_modbus_command_type command; /**< Modbus command */
    u16 transaction_id;                         /**< Transaction identifier for sync between messages of server & client */
    u16 protocol_id;                            /**< Protocol identifier (zero for modbus/tcp) */
    u16 length;                                 /**< Length, number of remaining bytes in this frame */
    u8 unit_id;                                 /**< Unit identifier, server address (255 if not used) */
} PACE2_modbus_message_start;

/**
 * @brief Event type for a read write multiple registers request
 */
typedef struct pace2_modbus_read_write_multiple_registers_request {
    u16 read_address;                /**< Read Starting Address */
    u16 read_quantity;               /**< Quantity to Read */
    u16 write_address;               /**< Write Starting Address */
    u16 write_quantity;              /**< Quantity to Write */
    struct ipd_charset_string value; /**< Write Registers Value */
} PACE2_modbus_read_write_multiple_registers_request;


/**
 * @brief Data structure for advanced Modbus events.
 */
typedef struct {
    PACE2_EVENT_COMMON_FIELDS;                                                                    /**< Global event type, has to be @ref PACE2_ADVANCED_MODBUS_EVENT for this structure. */
    PACE2_advanced_modbus_event_type meta_data_type;                                              /**< Modbus event subtype. */
    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;
    union {
        PACE2_modbus_message_start message_start;                                                 /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_MESSAGE_START event. */
        PACE2_modbus_address_quantity read_coils_request;                                         /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_COILS_REQUEST event. */
        struct ipd_charset_string read_coils_response;                                            /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_COILS_RESPONSE event. */
        PACE2_modbus_address_quantity read_discrete_inputs_request;                               /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_DISCRETE_INPUTS_REQUEST event. */
        struct ipd_charset_string read_discrete_inputs_response;                                  /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_DISCRETE_INPUTS_RESPONSE event. */
        PACE2_modbus_address_quantity read_holding_registers_request;                             /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_HOLDING_REGISTERS_REQUEST event. */
        struct ipd_charset_string read_holding_registers_response;                                /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_HOLDING_REGISTERS_RESPONSE event. */
        PACE2_modbus_address_quantity read_input_registers_request;                               /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_INPUT_REGISTERS_REQUEST event. */
        struct ipd_charset_string read_input_registers_response;                                  /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_INPUT_REGISTERS_RESPONSE event. */
        PACE2_modbus_address_value write_single_coil_request;                                     /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_WRITE_SINGLE_COIL_REQUEST event. */
        PACE2_modbus_address_value write_single_coil_response;                                    /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_WRITE_SINGLE_COIL_RESPONSE event. */
        PACE2_modbus_address_value write_single_register_request;                                 /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_WRITE_SINGLE_REGISTER_REQUEST event. */
        PACE2_modbus_address_value write_single_register_response;                                /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_WRITE_SINGLE_REGISTER_RESPONSE event. */
        PACE2_modbus_address_quantity_value write_multiple_coils_request;                         /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_WRITE_MULTIPLE_COILS_REQUEST event. */
        PACE2_modbus_address_quantity write_multiple_coils_response;                              /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_WRITE_MULTIPLE_COILS_RESPONSE event. */
        PACE2_modbus_address_quantity_value write_multiple_registers_request;                     /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_WRITE_MULTIPLE_REGISTERS_REQUEST event. */
        PACE2_modbus_address_quantity write_multiple_registers_response;                          /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_WRITE_MULTIPLE_REGISTERS_RESPONSE event. */
        struct ipd_charset_string report_server_id_response;                                      /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_REPORT_SERVER_ID_RESPONSE event */
        struct ipd_charset_string read_file_record_request;                                       /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_FILE_RECORD_REQUEST event. */
        struct ipd_charset_string read_file_record_response;                                      /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_FILE_RECORD_RESPONSE event. */
        struct ipd_charset_string write_file_record_request;                                      /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_WRITE_FILE_RECORD_REQUEST event. */
        struct ipd_charset_string write_file_record_response;                                     /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_WRITE_FILE_RECORD_RESPONSE event. */
        PACE2_modbus_mask_write_register mask_write_register_request;                             /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_MASK_WRITE_REGISTER_REQUEST event. */
        PACE2_modbus_mask_write_register mask_write_register_response;                            /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_MASK_WRITE_REGISTER_RESPONSE event. */
        PACE2_modbus_read_write_multiple_registers_request read_write_multiple_registers_request; /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_WRITE_MULTIPLE_REGISTERS_REQUEST event. */
        struct ipd_charset_string read_write_multiple_registers_response;                         /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_WRITE_MULTIPLE_REGISTERS_RESPONSE event. */
        u16 read_fifo_queue_request_address;                                                      /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_FIFO_QUEUE_REQUEST event. */
        struct ipd_charset_string read_fifo_queue_response;                                       /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_READ_FIFO_QUEUE_RESPONSE event. */
        struct ipd_charset_string user_defined_function_data;                                     /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_USER_DEFINED_FUNCTION event. */
        u8 exception_response_code;                                                               /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_EXCEPTION_RESPONSE event. */
        const char * invalid;                                                                     /**< Contains data for a @ref PACE2_ADVANCED_MODBUS_INVALID event. */
    } event_data;                                                                                 /**< Named union which contains the data of the sub-event types. */
} PACE2_advanced_modbus_event;

#endif

#endif /* MODBUS_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
