/** @file      mpeg_events.h
  * @brief     MPEG event definitions.
  * @details   @copydoc doc_mpeg
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef MPEG_EVENTS_H
#define MPEG_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

#ifndef PACE_KERNEL_LIB
#endif

/* Event Groups *****************************************************************/

/**
 * @brief Public define for the assigned meta data group of the advanced MPEG event.
 */
#define PACE2_ADVANCED_MPEG_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced MPEG events.
 */
typedef enum {

    /** This event is issued for every MPEG-2 TS packet. Provides a struct ipd_mpeg_ts_packet. */
    PACE2_ADVANCED_MPEG_TS_PACKET,

    /** This event is issued when decoding of an MPEG flow fails.
     * It provides a char const * to a message describing the failure reason. */
    PACE2_ADVANCED_MPEG_INVALID,

    /** This event is issued when an MPEG flow ends. It does not provide any data (NULL). */
    PACE2_ADVANCED_MPEG_END_OF_FLOW,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_MPEG_EVENT_TYPES

} PACE2_advanced_mpeg_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

struct ipd_mpeg_ts_packet {

    /** 13-bit PID (packet identifier) from an MPEG TS (transport stream). */
    u16 packet_identifier;

    /** Indication whether the packet actually contains payload. */
    u8 contains_payload;

    /** Continuity counter (4-bit), only valid if packet contains payload. */
    u8 continuity_counter;

    /** Indication whether the packet contains the start of a payload unit. */
    u8 payload_unit_start;

    /** Actual payload of packet. */
    struct ipd_charset_string payload;
};

/**
 * @brief Data structure of advanced MPEG events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_MPEG_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** MPEG event subtype. */
    PACE2_advanced_mpeg_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    union {

        /** Contains data for a @ref PACE2_ADVANCED_MPEG_TS_PACKET event. */
        struct ipd_mpeg_ts_packet mpeg_ts_packet;

        /** Contains data for a @ref PACE2_ADVANCED_MPEG_INVALID event. */
        char const * invalid;

    } event_data;

} PACE2_advanced_mpeg_event;

#endif

#endif /* MPEG_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
