/** @file      mqtt_events.h
  * @brief     MQTT event definitions.
  * @details   @copydoc doc_mqtt
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef MQTT_EVENTS_H
#define MQTT_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Groups *****************************************************************/

/**
 * @brief Public define for the assigned meta data group of the advanced MQTT event.
 */
#define PACE2_ADVANCED_MQTT_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief   Enum of command types.
 * @details These flags are possible message types inside the fixed header part of a valid MQTT message.
*/
typedef enum {
    PACE2_MQTT_CONTROL_COMMAND_TYPE_CONNECT = 1,        /**< Client request to connect to Server **/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_CONNACK = 2,        /**< Connect Acknowledgment**/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_PUBLISH = 3,        /**< Publish message**/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_PUBACK = 4,         /**< Publish Acknowledgment**/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_PUBREC = 5,         /**< Publish Received (assured delivery part 1) **/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_PUBREL = 6,         /**< Publish Release (assured delivery part 2)**/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_PUBCOMP = 7,        /**< Publish Complete (assured delivery part 3) **/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_SUBSCRIBE = 8,      /**< Client Subscribe request**/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_SUBACK = 9,         /**< Subscribe Acknowledgment**/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_UNSUBSCRIBE = 10,   /**< Client Unsubscribe request**/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_UNSUBACK = 11,      /**< Unsubscribe Acknowledgment **/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_PINGREQ = 12,       /**< PING Request**/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_PINGRESP = 13,      /**< PING Response **/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_DISCONNECT = 14,    /**< Client is Disconnecting **/
    PACE2_MQTT_CONTROL_COMMAND_TYPE_AUTH = 15,          /**< Authentication Exchange (MQTT 5.0+) **/
} PACE2_mqtt_control_command_type;

/**
 * @brief Enum of Quality of Service (QoS) type.
 */
typedef enum pace2_mqtt_qos_type{
    PACE2_MQTT_COMMAND_QOS_AT_MOST_ONE_FIRE_AND_FORGET = 0,     /**< At most once, Fire and Forget */
    PACE2_MQTT_COMMAND_QOS_AT_LEAST_ONE_ACK_DELIVERY = 1,       /**< At least once, Acknowledged delivery */
    PACE2_MQTT_COMMAND_QOS_EXACTLY_ONE_ASSURED_DELIVERY = 2,    /**< Exactly once, Assured delivery */
} PACE2_mqtt_command_qos_type;

/**
 * @brief Enum of optional header entry data types.
 */
typedef enum {
    PACE2_MQTT_HEADER_DATA_TYPE_STRING, /**< Signals that the optional header contains a string as value
                                             (see @ref PACE2_mqtt_header_data::data). */
    PACE2_MQTT_HEADER_DATA_TYPE_NUMBER, /**< Signals that the optional header contains a number as value
                                             (see @ref PACE2_mqtt_header_data::data). */
    PACE2_MQTT_HEADER_DATA_TYPE_FLAG,   /**< Signals that the optional header contains a flag as value
                                             (see @ref PACE2_mqtt_header_data::data). */
    PACE2_MQTT_HEADER_DATA_TYPE_QOS     /**< Signals that the optional header contains a Quality of Service -flag as
                                             value (see @ref PACE2_mqtt_header_data::data). */

} PACE2_mqtt_header_data_type;

/**
 * @brief Enum of property entry data types.
 */
typedef enum {
    PACE2_MQTT_PROPERTY_DATA_TYPE_STRING,   /**< Signals that the property contains a string as value
                                                 (see @ref PACE2_mqtt_property_data::data). */
    PACE2_MQTT_PROPERTY_DATA_TYPE_NUMBER,   /**< Signals that the property contains a number as value
                                                 (see @ref PACE2_mqtt_property_data::data). */
    PACE2_MQTT_PROPERTY_DATA_TYPE_FLAG,     /**< Signals that the property contains a flag as value
                                                 (see @ref PACE2_mqtt_property_data::data). */
    PACE2_MQTT_PROPERTY_DATA_TYPE_QOS,      /**< Signals that the property contains a Quality of Service -flag as
                                                 value (see @ref PACE2_mqtt_property_data::data). */
    PACE2_MQTT_PROPERTY_DATA_TYPE_KEYVALUE, /**< Signals that the property contains a userdata name/value pair
                                                 (see @ref PACE2_mqtt_property_data::data). */
} PACE2_mqtt_property_data_type;

/**
 * @brief   Enum of fixed header entry types.
 * @details These are flags for entries of a variable header information of MQTT messages.
 * @note    Most of these types are only present in a specific command header types
 *          (@ref PACE2_mqtt_control_command_type).
 */
typedef enum {
    PACE2_MQTT_OPT_HEADER_TYPE_PROTOCOL_NAME,       /**< CONNECT - protocol name */
    PACE2_MQTT_OPT_HEADER_TYPE_PROTOCOL_VERSION,    /**< CONNECT - protocol version */
    PACE2_MQTT_OPT_HEADER_TYPE_FLAG_USER_NAME,      /**< CONNECT - user name flag */
    PACE2_MQTT_OPT_HEADER_TYPE_FLAG_PASSWORD,       /**< CONNECT - password flag */
    PACE2_MQTT_OPT_HEADER_TYPE_FLAG_WILL_RETAIN,    /**< CONNECT - will RETAIN flag */
    PACE2_MQTT_OPT_HEADER_TYPE_FLAG_WILL_QOS,       /**< CONNECT - will QoS flag */
    PACE2_MQTT_OPT_HEADER_TYPE_FLAG_WILL,           /**< CONNECT - will flag */
    PACE2_MQTT_OPT_HEADER_TYPE_FLAG_CLEAN_SESSION,  /**< CONNECT - clean session flag */
    PACE2_MQTT_OPT_HEADER_TYPE_KEEP_ALIVE,          /**< CONNECT - keep alive timer */
    PACE2_MQTT_OPT_HEADER_TYPE_FLAG_SESSION_PRESENT,/**< CONNACK - session present flag */
    PACE2_MQTT_OPT_HEADER_TYPE_RETURN_CODE,         /**< CONNACK - return code */
    PACE2_MQTT_OPT_HEADER_TYPE_TOPIC_NAME,          /**< PUBLISH - topic name*/
    PACE2_MQTT_OPT_HEADER_TYPE_MESSAGE_ID,          /**< message id */
} PACE2_mqtt_optional_header_type;

/**
 * @brief   Enum of fixed property types.
 * @details These are flags for entries of a header property information of MQTT messages.
 * @note    Most of these types are only present in a specific command header types
 *          (@ref PACE2_mqtt_control_command_type). This a MQTT 5.0 feature.
 */
typedef enum {
    PACE2_MQTT_PROPERTY_PAYLOAD_FORMAT,              /**< PUBLISH / WILL MESSAGE - payload format indicator */
    PACE2_MQTT_PROPERTY_MESSAGE_EXPIRY,              /**< PUBLISH / WILL MESSAGE - message expiry interval */
    PACE2_MQTT_PROPERTY_CONTENT_TYPE,                /**< PUBLISH / WILL MESSAGE - content type */
    PACE2_MQTT_PROPERTY_RESPONSE_TOPIC,              /**< PUBLISH / WILL MESSAGE - response topic */
    PACE2_MQTT_PROPERTY_CORRELATION_DATA,            /**< PUBLISH / WILL MESSAGE - correlation data */
    PACE2_MQTT_PROPERTY_SUBSCRIPTION_ID,             /**< PUBLISH / SUBSCRIBE - subscription id */
    PACE2_MQTT_PROPERTY_SESSION_EXPIRY,              /**< CONNECT / CONNACK / DISCONNECT - session expiry interval */
    PACE2_MQTT_PROPERTY_ASSIGNED_CLIENT_ID,          /**< CONNACK - assigned client id */
    PACE2_MQTT_PROPERTY_SERVER_KEEP_ALIVE,           /**< CONNACK - server keep alive */
    PACE2_MQTT_PROPERTY_AUTH_METHOD,                 /**< CONNECT / CONNACK / AUTH - authentication method */
    PACE2_MQTT_PROPERTY_AUTH_DATA,                   /**< CONNECT / CONNACK / AUTH - authentication data */
    PACE2_MQTT_PROPERTY_FLAG_REQUEST_PROBLEM_INFO,   /**< CONNECT - request problem information flag */
    PACE2_MQTT_PROPERTY_WILL_DELAY,                  /**< WILL MESSAGE - will delay interval */
    PACE2_MQTT_PROPERTY_FLAG_REQUEST_RESPONSE,       /**< CONNECT - request response information flag */
    PACE2_MQTT_PROPERTY_RESPONSE,                    /**< CONNACK - Response Information */
    PACE2_MQTT_PROPERTY_SERVER_REFERENCE,            /**< CONNACK / DISCONNECT - server reference */
    PACE2_MQTT_PROPERTY_REASON,                      /**< reason string */
    PACE2_MQTT_PROPERTY_MAX_RECEIVE,                 /**< CONNECT / CONNACK - receive maximum */
    PACE2_MQTT_PROPERTY_MAX_TOPIC_ALIAS,             /**< CONNECT / CONNACK - topic alias maximum */
    PACE2_MQTT_PROPERTY_TOPIC_ALIAS,                 /**< PUBLISH - topic alias */
    PACE2_MQTT_PROPERTY_MAX_QOS,                     /**< CONNACK - maximum QoS */
    PACE2_MQTT_PROPERTY_FLAG_RETAIN,                 /**< CONNACK - retain available */
    PACE2_MQTT_PROPERTY_USER_PROPERTY,               /**< pair of user properties */
    PACE2_MQTT_PROPERTY_MAX_PACKET_SIZE,             /**< CONNECT / CONNACK - maximum packet size */
    PACE2_MQTT_PROPERTY_FLAG_WILDCARD_SUBSCRIPTIONS, /**< CONNACK - wildcard subscription available */
    PACE2_MQTT_PROPERTY_FLAG_SUBSCRIPTION_IDS,       /**< CONNACK - subscription identifier available */
    PACE2_MQTT_PROPERTY_FLAG_SHARED_SUBSCRIPTIONS,   /**< CONNACK - shared subscription available */
} PACE2_mqtt_property_type;

/**
 * @brief Enum of MQTT payload types.
 * @note  These are flags for entries of the payload information of MQTT messages.
 * @note  These types are only present in a specific command header types
 *        (@ref PACE2_mqtt_control_command_type).
 */
typedef enum {
    PACE2_MQTT_PAYLOAD_TYPE_CLIENT_IDENTIFIER,      /**< CONNECT - client identifier */
    PACE2_MQTT_PAYLOAD_TYPE_TOPIC_NAME,             /**< (UN)SUBSCRIBE - Topic name or CONNECT - will topic */
    PACE2_MQTT_PAYLOAD_TYPE_WILL_MESSAGE,           /**< CONNECT - will message */
    PACE2_MQTT_PAYLOAD_TYPE_USER_NAME,              /**< CONNECT - user name */
    PACE2_MQTT_PAYLOAD_TYPE_PASSWORD,               /**< CONNECT - password */
} PACE2_mqtt_payload_type;

/**
 * @brief   Types for advanced MQTT events.
 * @details The MQTT protocol consists of a fixed header part (COMMAND-event), a variable header part
 *          (multiple OPT_HEADER-events), and payload data (PAYLOAD_DATA or PAYLOAD_INFO).
 *          These main MQTT events have a common "command_identifier".
 *          This identifier will be incremented by each "MQTT-COMMAND" - event. So that all following events could be
 *          assigned to this MQTT-COMMAND.
 * @note    See http://public.dhe.ibm.com/software/dw/webservices/ws-mqtt/mqtt-v3r1.html#msg-format for further
 *          information.
 */
typedef enum {
    PACE2_ADVANCED_MQTT_NEW_CONNECTION, /**< First event of a MQTT flow see @ref PACE2_advanced_mqtt_event.
                                             This Event signals, that the flow of the current packet will be
                                             handled by the MQTT decoder. */
    PACE2_ADVANCED_MQTT_COMMAND,        /**< Signals, that a MQTT fixed header was found.(see @ref PACE2_mqtt_command)
                                              The rest of the command will follow with additional events. */
    PACE2_ADVANCED_MQTT_OPT_HEADER,     /**< Signals, that an entry of a variable header was found.
                                             (see @ref PACE2_mqtt_optional_header struct) */
    PACE2_ADVANCED_MQTT_PAYLOAD_DATA,   /**< Signals, that a payload of a command were arrived. For payload,
                                             which are stretched across multiple packages, each event has an individual
                                             offset. (see @ref PACE2_mqtt_payload_data struct) */
    PACE2_ADVANCED_MQTT_PAYLOAD_INFO,   /**< Signals, that a significant part of the payload was found e.g.: topic name,
                                             username, password etc. (see @ref PACE2_mqtt_payload_info struct) */
    PACE2_ADVANCED_MQTT_INVALID,        /**< Signals, that this MQTT flow is invalid. */
    PACE2_ADVANCED_MQTT_CONNECTION_END, /**< Signals, that the processing of the flow of the current packet is done.
                                             No further events will be thrown. It is guaranteed that this event or the
                                             event PACE2_ADVANCED_MQTT_INVALID is always the last event for a flow. */
    PACE2_ADVANCED_MQTT_PROPERTY,       /**< Signals, that an entry of a header property was found.
                                             (see @ref PACE2_mqtt_property struct) */

    PACE2_NUMBER_OF_ADVANCED_MQTT_EVENT_TYPES /**< Number of implemented event types. */
} PACE2_advanced_mqtt_event_type;


/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * @brief Data structure of fixed MQTT header flags.
 */
typedef struct {
    u8 dup_flag:1;      /**< Duplicate delivery flag */
    u8 qos_flag:2;      /**< Quality of Service: can be mapped to @ref PACE2_mqtt_command_qos_type */
    u8 retain_flag:1;   /**< Retain flag */
} PACE2_mqtt_command_header_flags;

/**
 * @brief Data of a MQTT optional header entry.
 */
typedef struct pace2_mqtt_header_data{
    PACE2_mqtt_header_data_type data_type;  /**< specifies, which entry should be used in the union.*/
    union
    {
        struct ipd_string string;           /**< String entry e.g. for topic name, protocol name etc.*/
        u64 number;                         /**< Number entry e.g. for message id, keep alive timer etc.*/
        u8 flag;                            /**< Boolean flag type for DUP or Retain-flags.*/
        PACE2_mqtt_command_qos_type qos;    /**< Quality of Service flag. */
    } data;                                 /**< Contains the value of the optional header entry. */
} PACE2_mqtt_header_data;

/**
 * @brief Data structure of fixed MQTT header.
 */
typedef struct pace2_mqtt_command{
    u64 command_identifier;                                 /**< Unique pace2 internal identifier for one MQTT command.
                                                                 All events related to this command will use the same
                                                                 command_identifier. */
    PACE2_mqtt_control_command_type control_command_type;   /**< Individual control command type inside the fixed
                                                                 header.*/
    PACE2_mqtt_command_header_flags flags;                  /**< All flags of the fixed header. For some command types
                                                                 not all flags are valid.*/
    u64 length;                                             /**< Length information of the following data (variable
                                                                 header and payload) of the fixed header. */
} PACE2_mqtt_command;

/**
 * @brief Data structure of a MQTT variable header entry.
 */
typedef struct pace2_mqtt_optional_header{
    u64 command_identifier;               /**< Unique pace2 internal identifier for one MQTT command. All events related
                                               to this command will use the same command_identifier. */
    PACE2_mqtt_optional_header_type type; /**< Individual type of the optional header entry. */
    PACE2_mqtt_header_data content;       /**< Content of the optional header entry. */
} PACE2_mqtt_optional_header;

/**
 * @brief A pair of utf-8 encoded strings.
 */
typedef struct pace2_mqtt_user_property {
    struct ipd_charset_string name;  /**< The key of this string pair. */
    struct ipd_charset_string value; /**< The value of this string pair. */
} PACE2_mqtt_user_property;

/**
 * @brief Data of a MQTT property entry.
 */
typedef struct pace2_mqtt_property_data {
    PACE2_mqtt_property_data_type data_type; /**< specifies, which entry should be used in the union. */
    union {
        struct ipd_charset_string string; /**< String entry e.g. for topic name, protocol name, or raw binary data. */
        u64 number;                       /**< Number entry e.g. for message id, keep alive timer etc. */
        u8 flag;                          /**< Boolean flag type for DUP or Retain-flags. */
        PACE2_mqtt_command_qos_type qos;  /**< Quality of Service flag. */
        PACE2_mqtt_user_property user;    /**< A userdata name-value string pair. */

    } data; /**< Contains the value of the Property entry. */
} PACE2_mqtt_property_data;

/**
 * @brief Data structure of a MQTT (5.0) property entry.
 */
typedef struct pace2_mqtt_property {
    u64 command_identifier;           /**< Unique pace2 internal identifier for one MQTT command. All events related
                                           to this command will use the same command_identifier. */
    PACE2_mqtt_property_type type;    /**< Individual type of the property entry. */
    PACE2_mqtt_property_data content; /**< Content of the property entry. */
} PACE2_mqtt_property;

/**
 * @brief Data structure of a MQTT-PUBLISH payload.
 * @details Offset is necessary to represent payload over multiple TCP-packages.
 */
typedef struct pace2_mqtt_payload_data{
    u64 command_identifier;     /**< Unique pace2 internal identifier for one MQTT command. All events related to this
                                     command will use the same command_identifier. */
    u64 offset;                 /**< Multiple data_blobs with the same command_identifier could be merged with this
                                     offset. */
    struct ipd_charset_string data_blob; /**< Specific payload data. */
} PACE2_mqtt_payload_data;

/**
 * @brief Data structure of a MQTT payload.
 */
typedef struct pace2_mqtt_payload_info{
    u64 command_identifier;               /**< Unique pace2 internal identifier for one MQTT command. All events
                                               related to this command will use the same command_identifier. */
    PACE2_mqtt_payload_type type;         /**< Distinguish the different payload types from each other. */
    struct ipd_charset_string data_blob;  /**< Specific payload data. */
} PACE2_mqtt_payload_info;

/**
 * @brief Data structure of advanced MQTT events.
 * @note  Detailed description can be found at https://docs.oasis-open.org/mqtt/mqtt/v5.0/os/mqtt-v5.0-os.html.
 */
typedef struct {
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_advanced_mqtt_event_type meta_data_type;
    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;
    union {
        const char* invalid;                                /**< Contains data for a @ref PACE2_ADVANCED_MQTT_INVALID
                                                                 event. */
        struct pace2_mqtt_command header;                   /**< Contains data for a @ref PACE2_ADVANCED_MQTT_COMMAND
                                                                 event. */
        struct pace2_mqtt_optional_header optional_header;  /**< Contains data for a @ref PACE2_ADVANCED_MQTT_OPT_HEADER
                                                                 event. */
        struct pace2_mqtt_property header_property;         /**< Contains data for a @ref PACE2_ADVANCED_MQTT_PROPERTY
                                                                 event. */
        struct pace2_mqtt_payload_data payload_data;        /**< Contains data for a
                                                                 @ref PACE2_ADVANCED_MQTT_PAYLOAD_DATA event. */
        struct pace2_mqtt_payload_info payload_info;        /**< Contains data for a
                                                                 @ref PACE2_ADVANCED_MQTT_PAYLOAD_INFO event. */
    } event_data;                                           /**< Named union that contains the data of the sub-event
                                                                 types. */
} PACE2_advanced_mqtt_event;

#endif /* PACE_KERNEL_LIB */

#endif /* MQTT_EVENTS_H */
