/** @file      netbios_events.h
  * @brief     NetBIOS event definitions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef NETBIOS_EVENTS_H
#define NETBIOS_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Groups *****************************************************************/

#define PACE2_BASIC_NETBIOS_EVENT_GROUP PACE2_BASIC_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/* Event Structures *************************************************************/

/**
 * @brief Public define for the maximum question section record count.
 */
#define PACE2_MAX_NUMBER_OF_NETBIOS_QD_COUNT 2

/**
 * @brief Public define for the maximum answer section record count.
 */
#define PACE2_MAX_NUMBER_OF_NETBIOS_AN_COUNT 2

/**
 * @brief Public define for the maximum authority section record count.
 */
#define PACE2_MAX_NUMBER_OF_NETBIOS_NS_COUNT 2

/**
 * @brief Public define for the maximum additional section record count.
 */
#define PACE2_MAX_NUMBER_OF_NETBIOS_AR_COUNT 2

/**
 * @brief Structure for the NetBIOS Name Service packet header.
 */
struct netbios_header {
    u16 transaction_id; /**< Transaction ID for the Name Service transaction. */
    u16 nm_flags;       /**< Flags for operation. */
    u8 opcode;          /**< Packet type code. */
    u8 rcode;           /**< Result code of request. */
    u16 qdcount;        /**< Number of question section entries. */
    u16 ancount;        /**< Number of answer section entries. */
    u16 nscount;        /**< Number of authority section entries. */
    u16 arcount;        /**< Number of additional section entries. */
};

/**
 * @brief Public define for the NetBIOS label length. The NetBIOS name consists always of 16 ASCII characters.
 */
#define PACE2_NETBIOS_LABEL_LEN 16

/**
 * @brief Structure for the compressed NetBIOS name representation.
 */
struct netbios_name {
    u8 label[PACE2_NETBIOS_LABEL_LEN]; /**< NetBIOS name label. The last byte contains the NetBIOS suffix. */
    PACE2_byte_buffer scope_id;        /**< NetBIOS Scope ID. */
};

/**
 * @brief Structure for the NetBIOS question record.
 */
struct netbios_question {
    struct netbios_name qname; /**< Compressed NetBIOS name. */
    u16 qtype;                 /**< Type of request. */
    u16 qclass;                /**< Class of request. */
};

/**
 * @brief Structure for the NetBIOS resource record.
 */
struct netbios_rrecord {
    struct netbios_name rrname; /**< Compressed NetBIOS name. */
    u16 rrtype;                 /**< Resource record type code. */
    u16 rrclass;                /**< Resource record class code. */
    u32 ttl;                    /**< Time To Live of a resource records name. */
    PACE2_byte_buffer rdata;    /**< Resource record information data. Depends on @ref rrtype and @ref rrclass. */
};

/**
 * @brief Data structure for basic NetBIOS events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_BASIC_NETBIOS_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** Header field. */
    struct netbios_header header;

    /** Array of question section records. */
    struct netbios_question qdrecords[PACE2_MAX_NUMBER_OF_NETBIOS_QD_COUNT];

    /** Array of answer section records. */
    struct netbios_rrecord anrecords[PACE2_MAX_NUMBER_OF_NETBIOS_AN_COUNT];

    /** Array of authority section records. */
    struct netbios_rrecord nsrecords[PACE2_MAX_NUMBER_OF_NETBIOS_NS_COUNT];

    /** Array of additional section records. */
    struct netbios_rrecord arrecords[PACE2_MAX_NUMBER_OF_NETBIOS_AR_COUNT];
} PACE2_basic_netbios_event;

#endif /* NETBIOS_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
