/** @file      nntp_events.h
  * @brief     NNTP event definitions.
  * @details   @copydoc doc_nntp
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef NNTP_EVENTS_H
#define NNTP_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Group ******************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced NNTP event.
 */
#define PACE2_ADVANCED_NNTP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/
/**
 * @brief Types for advanced nntp events.
 */
typedef enum {

    ////////////////////////////////////////////////
    // General events related to the NNTP flow    //
    ////////////////////////////////////////////////

    /**
     * First payload data has been processed and the client and server roles have been determined.<br />
     * Provides: int * : client -> server direction (0 or 1).
     */
    PACE2_ADVANCED_NNTP_CLIENT_DIRECTION,

    /**
     * The end of the connection.<br />
     * Provides: NULL.
     */
    PACE2_ADVANCED_NNTP_END_OF_FLOW,

    /**
     * Decoding of an NNTP connection fails.<br />
     * Provides: char* : message describing the failure reason.
     */
    PACE2_ADVANCED_NNTP_INVALID,

    /**
     * A new NNTP connection is noticed.<br />
     * Provides: NULL.
     */
    PACE2_ADVANCED_NNTP_NEW_CONNECTION,

    //////////////////////////////////////////////////////////////////
    // Events related to authentication (PLAIN and SASL) and TLS    //
    //////////////////////////////////////////////////////////////////

    /**
     * Plain-text authentication user name sent by the client.<br />
     * Provides: struct ipd_charset_string * : user name.
     */
    PACE2_ADVANCED_NNTP_COMMAND_AUTHENTICATION_USERNAME,

    /**
     * Plain-text authentication password sent by the client.<br />
     * Provides: struct ipd_charset_string * : password.
     */
    PACE2_ADVANCED_NNTP_COMMAND_AUTHENTICATION_PASSWORD,

    /**
     * Client requested a SASL authentication.<br />
     * Provides: struct ipd_nntp_command_authinfo_sasl * : requested authentication mechanism and optionally,
     * the initial response.
     */
    PACE2_ADVANCED_NNTP_COMMAND_AUTHENTICATION_SASL,

    /**
     * SASL authentication was accepted and the server provides the challenge.<br />
     * Provides: struct ipd_nntp_response_auth_sasl * :
     * the same as the general response structure + the server challenge.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_SASL_AUTHENTICATION_ACCEPTED_WITH_SUCCESS_DATA,

    /**
     * Client should continue with SASL exachange.<br />
     * Provides: struct ipd_nntp_response_auth_sasl * :
     * the same as the general response structure + the server challenge.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_SASL_CONTINUE_WITH_SASL_EXCHANGE,

    /**
     * Client sends a BASE64-encoded challenge during the SASL authentication.<br />
     * Provides: struct ipd_charset_string * : client challenge.
     */
    PACE2_ADVANCED_NNTP_SASL_CLIENT_CHALLENGE,

    /////////////////////////////////////////////////////////////////
    // Events related to obtaining articles from the server        //
    /////////////////////////////////////////////////////////////////

    /**
     * Client requests to fetch the specified article from the server.<br />
     * Provides: struct ipd_nntp_article * : article-id or article number.
     */
    PACE2_ADVANCED_NNTP_COMMAND_GET_ARTICLE,

    /**
     * Client requests to fetch the body of the specified article from the server.<br />
     * Provides: struct ipd_nntp_article * : article-id or article number.
     */
    PACE2_ADVANCED_NNTP_COMMAND_GET_ARTICLE_BODY,

    /**
     * Client requests to fetch the header of the specified article from the server.<br />
     * Provides: struct ipd_nntp_article * : article-id or article number.
     */
    PACE2_ADVANCED_NNTP_COMMAND_GET_ARTICLE_HEADER,

    /**
     * Client requests to fetch the status of the specified article on the server.<br />
     * Provides: struct ipd_nntp_article * : article-id or article number.
     */
    PACE2_ADVANCED_NNTP_COMMAND_GET_ARTICLE_STATUS,

    /**
     * Server informs that it will provide the specified article to the client in the next lines.<br />
     * Provides: struct ipd_nntp_article * : article-id or article number.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_FOLLOWS,

    /**
     * Server informs that it will provide the header of the specified article to the client in the next lines.<br />
     * Provides: struct ipd_nntp_article * : article-id or article number.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_HEADERS_FOLLOW,

    /**
     * Server informs that it will provide the body of the specified article to the client in the next lines.<br />
     * Provides: struct ipd_nntp_article * : article-id or article number.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_BODY_FOLLOWS,

    ///////////////////////////////////////////////
    // Events related to posting articles        //
    ///////////////////////////////////////////////

    /**
     * Client checks if the specified article is desired by the server or not.<br />
     * Provides: struct ipd_charset_string * :  article-id to be checked.
     */
    PACE2_ADVANCED_NNTP_COMMAND_CHECK_IF_ARTICLE_DESIRED_BY_SERVER,

    /////////////////////////////////////////////////
    // Events related to selecting articles        //
    /////////////////////////////////////////////////

    /**
     * The indicated article exists and got selected.<br />
     * Provides: struct ipd_nntp_article * : article-id of the selected article.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_EXISTS_AND_SELECTED,

    ////////////////////////////////////////////////////////////////////
    // Events related to transferring articles using streaming        //
    ////////////////////////////////////////////////////////////////////

    /**
     * Client checks if the specified article is desired by the server or not (while using the streaming mode).<br />
     * Provides: struct ipd_charset_string * :  article-id to be checked.
     */
    PACE2_ADVANCED_NNTP_COMMAND_CHECK_IF_ARTICLE_DESIRED_BY_SERVER_IN_STREAMING,

    /**
     * Server wants the client to send the particular article using the streaming mode.<br />
     * Provides: struct ipd_nntp_article * : article-id.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_SEND_ARTICLE_USING_STREAMING_MODE,

    /**
     * Server temporarily defers receiving the particular article using the streaming mode.<br />
     * Provides: struct ipd_nntp_article * : article-id.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_TRANSFER_OF_ARTICLE_USING_STREAMING_TEMPORARY_UNAVAILABLE,

    /**
     * Server rejects receiving the particular article using the streaming mode.<br />
     * Provides: struct ipd_nntp_article * : article-id.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_IN_STREAMING_NOT_WANTED,

    /**
     * Article was successfully transferred using the streaming mode.<br />
     * Provides: struct ipd_nntp_article * : article-id.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_TRANSFERRED_OK_USING_STREAMING_MODE,

    /**
     * Transfer of the particular article failed using the streaming mode.<br />
     * Provides: struct ipd_nntp_article * : article-id.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_TRANSFER_FAILED_USING_STREAMING_MODE,

    /**
     * Client indicates that the next lines will be the lines of the particular article sent to the server.<br />
     * Provides: struct ipd_nntp_article * : article-id.
     */
    PACE2_ADVANCED_NNTP_COMMAND_ARTICLE_IN_STREAMING_FOLLOWS,

    //////////////////////////////////////////////////////////////////////////////
    // Events related to article content processing (including attachments)     //
    //////////////////////////////////////////////////////////////////////////////

    /**
     * Indicates the presence of a header in the article.<br />
     * Provides: struct ipd_nntp_article_header * : type, name, value, and e-mail (if exists) of the header.
     */
    PACE2_ADVANCED_NNTP_ARTICLE_HEADER,

    /**
     * Indicates a chunk of an article. In case of short articles, this will be the only chunk.<br />
     * Provides: struct ipd_nntp_article_textstream_chunk * : content.
     */
    PACE2_ADVANCED_NNTP_ARTICLE_BODY_CHUNK,

    /**
     * Indicates the presence of the starting header for an yencoded attachment.<br />
     * Provides: struct ipd_nntp_header_ybegin * : header.
     */
    PACE2_ADVANCED_NNTP_ARTICLE_ATTACHMENT_YENC_HEADER_BEGIN,

    /**
     * Indicates the presence of the "part" header for an yencoded attachment.<br />
     * Provides: struct ipd_nntp_header_ypart * : header.
     */
    PACE2_ADVANCED_NNTP_ARTICLE_ATTACHMENT_YENC_HEADER_PART,

    /**
     * Indicates a chunk of an attachment. In case of short attachments, this will be the only chunk.<br />
     * Provides: struct ipd_nntp_article_textstream_chunk * : content.
     */
    PACE2_ADVANCED_NNTP_ARTICLE_ATTACHMENT_CHUNK,

    /**
     * Indicates the presence of the ending header for an yencoded attachment.<br />
     * Provides: struct ipd_nntp_header_yend * : header.
     */
    PACE2_ADVANCED_NNTP_ARTICLE_ATTACHMENT_YENC_HEADER_END,

    /////////////////////////////////////////////////////////////
    // Events related to various commands and responses        //
    /////////////////////////////////////////////////////////////

    /**
     * A simple NNTP command (carries no data) was noticed.<br />
     * Provides: ipd_nntp_simple_command_enumeration * : type of the simple command.
     */
    PACE2_ADVANCED_NNTP_COMMAND_SIMPLE,

    /**
     * Various NNTP responses that carry no data.<br />
     * Provides: struct ipd_nntp_response * : general response structure (response code, success state,
     * and message returned by the server).
     */
    PACE2_ADVANCED_NNTP_RESPONSE_INFORMATIVE,

    /**
     * An unknown NNTP command was noticed.<br />
     * Provides: struct ipd_nntp_command_unknown * : general command structure (command, line of arguments).
     */
    PACE2_ADVANCED_NNTP_COMMAND_UNKNOWN,

    /**
     * An unknown NNTP response was noticed.<br />
     * Provides: struct ipd_nntp_response * : general response structure (response code, success state,
     * and message returned by the server).
     */
    PACE2_ADVANCED_NNTP_RESPONSE_UNKNOWN,

    /**
     * Next line of an unknown NNTP response was noticed.<br />
     * Provides: struct ipd_charset_string * : entire response line as text.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_UNKNOWN_NEXT_LINE,

    /**
     * Select the defined group.<br />
     * Provides: struct ipd_charset_string * : name of the group to be selected.
     */
    PACE2_ADVANCED_NNTP_COMMAND_SELECT_GROUP,

    /**
     * List article numbers from a group.<br />
     * Provides: struct ipd_nntp_command_listgroup * : name of the group (if not current) and range of articles.
     */
    PACE2_ADVANCED_NNTP_COMMAND_LIST_GROUP,

    /**
     * List some unknown custom feature specific to the server.<br />
     * Provides: struct ipd_charset_string * : content of the list command.
     */
    PACE2_ADVANCED_NNTP_COMMAND_LIST_CUSTOM,

    /**
     * Listed article number.<br />
     * Provides: u64 * : article number.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_LISTED,

    /**
     * Group successfully selected.<br />
     * Provides: struct ipd_nntp_response_group_selected * : details of the selected group
     * (group name, number of articles, lowest and highest article number).
     */
    PACE2_ADVANCED_NNTP_RESPONSE_GROUP_SELECTED,

    /**
     * Provide a line of a one valid newsgrouop and associated information.<br />
     * Provides: struct ipd_nntp_response_valid_newsgroup_and_associated_information * :
     * group name, low and high article numbers, group status (read-only, writable, moderated).
     */
    PACE2_ADVANCED_NNTP_RESPONSE_VALID_NEWSGROUP_AND_ASSOCIATED_INFORMATION,

    /**
     * Provide a line of a one creator and creating time of a valid newsgroup.<br />
     * Provides: struct ipd_nntp_response_creator_and_creating_time_of_valid_newsgroup * : group name,
     * creator, and creating time.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_CREATOR_AND_CREATING_TIME_OF_VALID_NEWSGROUP,

    /**
     * Provide a line of a one default value for the distribution header in new articles.<br />
     * Provides: struct ipd_charset_string * : default value for the distribution header in new articles.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_DEFAULT_VALUE_FOR_DISTRIBUTION_HEADER_IN_NEW_ARTICLES,

    /**
     * Provide a line of a one group description.<br />
     * Provides: struct ipd_nntp_response_newsgroup_description * : group, description.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_NEWSGROUP_DESCRIPTION,

    /**
     * Provide a line of a one available header field.<br />
     * Provides: struct ipd_charset_string * : available header field.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_AVAILABLE_HEADER_FIELD,

    /**
     * Provide a line of a one field in the overview database.<br />
     * Provides: struct ipd_charset_string * : field in the overview database.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_FIELD_IN_OVERVIEW_DATABASE,

    /**
     * Provide a line of a custom listing command specific to the server.<br />
     * Provides: struct ipd_charset_string * : custom listing line.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_LIST_CUSTOM_LINE,

    /**
     * Requests the contents of all the fields in the database for an article or a range of articles.<br />
     * Provides: struct ipd_nntp_command_over * : requested identifier or range of articles.
     */
    PACE2_ADVANCED_NNTP_COMMAND_OVER,

    /**
     * Provide a line of one overview of an article.<br />
     * Provides: struct ipd_nntp_overview_information_line * : article overview information line.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_OVERVIEW_INFORMATION_LINE,

    /**
     * Requests the server to provide a list of newsgroups created on the server since
     * the specified date and time.<br />
     * Provides: struct ipd_nntp_command_new_news_groups * : date and time since which
     * the created newsgroups will be listed.
     */
    PACE2_ADVANCED_NNTP_COMMAND_GET_NEW_GROUPS,

    /**
     * A single line denoting the individual capability.<br />
     * Provides: struct ipd_charset_string * : capability.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_CAPABILITY,

    /**
     * Server is in wrong mode; indicated capability should be used to change mode.<br />
     * Provides: struct ipd_nntp_response_use_capability * : capability to be used, server message.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_SERVER_IN_WRONG_MODE_USE_INDICATED_CAPABILITY,

    /**
     * Provides the current UTC date and time on the server.<br />
     * Provides: struct ipd_nntp_response_datetime * : date and time.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_SERVER_UTC_DATE_TIME,

    /**
     * Help line returned by the server.<br />
     * Provides: struct ipd_charset_string * : help line.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_HELP_LINE,

    /**
     * Requests a list of new news on the server since the specified date and time.<br />
     * Provides: struct ipd_nntp_command_new_news_groups * : wildmat, date, and time.
     */
    PACE2_ADVANCED_NNTP_COMMAND_GET_NEW_NEWS,

    /**
     * Article-id of a new article.<br />
     * Provides: struct ipd_charset_string * : article-id.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_ID_OF_NEW_ARTICLE,

    /**
     * Requests access to specific fields from an article or a range of articles.<br />
     * Provides: struct ipd_nntp_command_hdr * : article or range of articles.
     */
    PACE2_ADVANCED_NNTP_COMMAND_GET_HEADER_FIELD,

    /**
     * Provides the content of the specified header per-article.<br />
     * Provides: struct ipd_nntp_response_header_content * : article number, content of the specified header.
     */
    PACE2_ADVANCED_NNTP_RESPONSE_HEADER_FIELD_CONTENT,

    ///////////////////////////////////////////////////////////
    // A dummy event to signal the end of the event list     //
    ///////////////////////////////////////////////////////////

    /**
     * A dummy event to signal the end of the event list.
     */
    PACE2_NUMBER_OF_ADVANCED_NNTP_EVENTS,

} PACE2_advanced_nntp_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * @brief Types of responses used for the RESPONSE_INFORMATIVE event (struct ipd_nntp_response.code).
 */
enum ipd_nntp_response_informative_enumeration {

    /**
     * "Service available, posting allowed".<br />
     * Generated by: initial connection, MODE READER.
     */
    NNTP_RESPONSE_SERVICE_AVAILABLE_POSTING_ALLOWED = 200,

    /**
     * "Service available, posting prohibited".<br />
     * Generated by: initial connection, MODE READER.
     */
    NNTP_RESPONSE_SERVICE_AVAILABLE_POSTING_PROHIBITED = 201,

    /**
     * "Streaming permitted".<br />
     * Generated by: MODE STREAM.<br />
     * Returned when the user tests if streaming of articles is possible on this server.
     */
    NNTP_RESPONSE_STREAMING_OF_ARTICLES_PERMITTED = 203,

    /**
     * "Connection closing".<br />
     * Generated by: QUIT.<br />
     * Returned when the user wants to close the connection.
     */
    NNTP_RESPONSE_CONNECTION_CLOSED = 205,

    /**
     * "Article transferred OK".<br />
     * Generated by: IHAVE (second stage).<br />
     * Returned when an article was successfully transferred to the server.
     */
    NNTP_RESPONSE_ARTICLE_TRANSFERRED_OK = 235,

    /**
     * "Article received OK".<br />
     * Generated by: POST (second stage).<br />
     * Returned when an article was successfully received by the server.
     */
    NNTP_RESPONSE_ARTICLE_RECEIVED_OK = 240,

    /**
     * "Authentication accepted".<br />
     * Generated by: AUTHINFO USER, AUTHINFO PASS, AUTHINFO SASL.<br />
     * Returned when the user wants to authenticate to a news server.
     */
    NNTP_RESPONSE_AUTHENTICATION_ACCEPTED = 281,

    /**
     * "Send article to be transferred".<br />
     * Generated by: IHAVE (first stage).<br />
     * Returned when the server accepts the user's request to transfer an article.
     */
    NNTP_RESPONSE_SEND_ARTICLE_TO_BE_TRANSFERRED = 335,

    /**
     * "Send article to be posted".<br />
     * Generated by: POST (first stage).<br />
     * Returned when the server accepts the user's request to post an article.
     */
    NNTP_RESPONSE_SEND_ARTICLE_TO_BE_POSTED = 340,

    /**
     * "Password required".<br />
     * Generated by: AUTHINFO USER.<br />
     * Returned when the user wants to authenticate to a news server.
     */
    NNTP_RESPONSE_PASSWORD_REQUIRED = 381,

    /**
     * "Continue with TLS negotiation".<br />
     * Generated by: STARTTLS.<br />
     * Returned when the user wants to starts a TLS session with the server.
     */
    NNTP_RESPONSE_CONTINUE_WITH_TLS_NEGOTIATION = 382,

    /**
     * "Service not available or no longer available".<br />
     * Generated by: initial connection, generic response.<br />
     * Returned when the user connects to the server, or a general error message.
     */
    NNTP_RESPONSE_SERVICE_NO_LONGER_AVAILABLE = 400,

    /**
     * "Internal fault or problem preventing action being taken".<br />
     * Generic response.
     */
    NNTP_RESPONSE_INTERNAL_SERVER_ERROR = 403,

    /**
     * "No such newsgroup".<br />
     * Generated by: GROUP, LISTGROUP.<br />
     * Returned when the user wants to select a usenet group and the group does not exist.
     */
    NNTP_RESPONSE_NO_SUCH_NEWSGROUP = 411,

    /**
     * "No newsgroup selected".<br />
     * Generated by: ARTICLE, BODY, GROUP, HDR, HEAD, LAST, LISTGROUP, NEXT, OVER, STAT, XOVER.<br />
     * Returned when the user wants to select a usenet group and no group has been specified.
     */
    NNTP_RESPONSE_NO_NEWSGROUP_SELECTED = 412,

    /**
     * "Current article number is invalid".<br />
     * Generated by: ARTICLE, BODY, HDR, HEAD, LAST, NEXT, OVER, STAT, XOVER.<br />
     * Returned when the number of the currently selected article is invalid.
     */
    NNTP_RESPONSE_CURRENT_ARTICLE_NUMBER_INVALID = 420,

    /**
     * "No next article in this group".<br />
     * Generated by: NEXT.<br />
     * Returned when the user wants to select the next article in that newsgroup and the next article does not exist.
     */
    NNTP_RESPONSE_NO_NEXT_ARTICLE_IN_THIS_GROUP = 421,

    /**
     * "No previous article in this group".<br />
     * Generated by: LAST.<br />
     * Returned when the user wants to select the previous article in that newsgroup and
     * the previous article does not exist.
     */
    NNTP_RESPONSE_NO_PREVIOUS_ARTICLE_IN_THIS_GROUP = 422,

    /**
     * "No article with that number or in that range".<br />
     * Generated by: ARTICLE, BODY, HDR, HEAD, OVER, STAT, XOVER.<br />
     * Returned when the user refers to an article with invalid number or there is no article in the defined range.
     */
    NNTP_RESPONSE_NO_ARTICLE_WITH_THAT_NUMBER_OR_IN_THAT_RANGE = 423,

    /**
     * "No article with that article-id".<br />
     * Generated by: ARTICLE, BODY, HDR, HEAD, OVER, STAT.<br />
     * Returned when there is no article with the provided id.
     */
    NNTP_RESPONSE_NO_ARTICLE_WITH_THAT_ID = 430,

    /**
     * "Article not wanted".<br />
     * Generated by: IHAVE (first stage).<br />
     * Returned when the server rejects the user's request to transfer an article.
     */
    NNTP_RESPONSE_ARTICLE_NOT_WANTED = 435,

    /**
     * "Transfer not possible; try again later" (first stage),
     * "Transfer failed; try again later" (second stage).<br />
     * Generated by: IHAVE (either stage).<br />
     * Returned when the server cannot perform the transfer of an article at the given time.
     */
    NNTP_RESPONSE_TRANSFER_NOT_POSSIBLE_OR_FAILED_TRY_AGAIN_LATER = 436,

    /**
     * "Transfer rejected; do not retry".<br />
     * Generated by: IHAVE (second stage).<br />
     * Returned when the server rejects to the transfer an article.
     */
    NNTP_RESPONSE_TRANSFER_REJECTED_DO_NOT_RETRY = 437,

    /**
     * "Posting not permitted".<br />
     * Generated by: POST (first stage).<br />
     * Returned when the server rejects to post an article.
     */
    NNTP_RESPONSE_POSTING_NOT_PERMITTED = 440,

    /**
     * "Posting failed".<br />
     * Generated by: POST (second stage).<br />
     * Returned when posting of an article failed.
     */
    NNTP_RESPONSE_POSTING_FAILED = 441,

    /**
     * "Command unavailable until the client has authenticated itself".<br />
     * Generic response.
     */
    NNTP_RESPONSE_CLIENT_MUST_AUTHENTICATE_TO_USE_THIS_COMMAND = 480,

    /**
     * "Authentication failed/rejected".<br />
     * Generated by: AUTHINFO USER, AUTHINFO PASS, AUTHINFO SASL.<br />
     * Returned when the user wants to authenticate to a news server.
     */
    NNTP_RESPONSE_AUTHENTICATION_FAILED = 481,

    /**
     * "Authentication commands issued out of sequence".<br />
     * Generated by: AUTHINFO USER, AUTHINFO PASS, AUTHINFO SASL.<br />
     * Returned when the user wants to authenticate to a news server.
     */
    NNTP_RESPONSE_AUTHENTICATION_COMMANDS_OUT_OF_SEQUENCE = 482,

    /**
     * "Command unavailable until suitable privacy has been arranged".<br />
     * Generic response.
     */
    NNTP_RESPONSE_PRIVACY_MUST_BE_ARRANGED_TO_USE_THIS_COMMAND = 483,

    /**
     * "Unknown command".<br />
     * Generic response.
     */
    NNTP_RESPONSE_COMMAND_UNKNOWN = 500,

    /**
     * "Syntax error in command".<br />
     * Generic response.
     */
    NNTP_RESPONSE_SYNTAX_ERROR_IN_COMMAND = 501,

    /**
     * "Service permanently unavailable", "command not permitted".<br />
     * Generated by: initial connection, MODE READER, generic response.
     */
    NNTP_RESPONSE_SERVICE_PERMANENTLY_UNAVAILABLE_OR_COMMAND_NOT_PERMITTED = 502,

    /**
     * "Feature not supported".<br />
     * Generic response.<br />
     * Returned when the requested feature is not supported or the server does not contain the requested data.
     */
    NNTP_RESPONSE_FEATURE_NOT_SUPPORTED = 503,

    /**
     * "Error in base64-encoding [RFC4648] of an argument".<br />
     * Generic response.
     */
    NNTP_RESPONSE_ARGUMENT_ENCODED_INCORRECTLY = 504,

    /**
     * "Can not initiate TLS negotiation".<br />
     * Generated by: STARTTLS.<br />
     * Returned when the user wants to starts a TLS session with the server.
     */
    NNTP_RESPONSE_CANNOT_INITIATE_TLS_NEGOTIATION = 580,
};

/**
 * @brief Types of headers used for the ARTICLE_HEADER event (struct ipd_nntp_article_header.type).
 */
enum ipd_nntp_header_enumeration {

    /**
     * Indicates the presence of an unknown header in the article.
     */
    NNTP_ARTICLE_HEADER_UNKNOWN,

    /**
     * Indicates the presence of the "Date:" header in the article.
     */
    NNTP_ARTICLE_HEADER_DATE,

    /**
     * Indicates the presence of the "From:" header in the article.
     */
    NNTP_ARTICLE_HEADER_FROM,

    /**
     * Indicates the presence of the "Message-ID:" header in the article.
     */
    NNTP_ARTICLE_HEADER_MESSAGE_ID,

    /**
     * Indicates the presence of the "Newsgroups:" header in the article.
     */
    NNTP_ARTICLE_HEADER_NEWSGROUPS,

    /**
     * Indicates the presence of the "Path:" header in the article.
     */
    NNTP_ARTICLE_HEADER_PATH,

    /**
     * Indicates the presence of the "Subject:" header in the article.
     */
    NNTP_ARTICLE_HEADER_SUBJECT,

    /**
     * Indicates the presence of the "Lines:" header in the article.
     */
    NNTP_ARTICLE_HEADER_LINES,

    /**
     * Indicates the presence of the Reply-To:" header in the article.
     */
    NNTP_ARTICLE_HEADER_REPLY_TO,

    /**
     * Indicates the presence of the "MIME-Version:" header in the article.
     */
    NNTP_ARTICLE_HEADER_MIME_VERSION,

    /**
     * Indicates the presence of the "Content-Type:" header in the article.
     */
    NNTP_ARTICLE_HEADER_CONTENT_TYPE,

    /**
     * Indicates the presence of the "Content-Transfer-Encoding:" header in the article.
     */
    NNTP_ARTICLE_HEADER_CONTENT_TRANSFER_ENCODING,

    /**
     * Indicates the presence of the "Organization:" header in the article.
     */
    NNTP_ARTICLE_HEADER_ORGANIZATION,

    /**
     * Indicates the presence of the "NNTP-Posting-Host:" header in the article.
     */
    NNTP_ARTICLE_HEADER_NNTP_POSTING_HOST,

    /**
     * Indicates the presence of the "User-Agent:" header in the article.
     */
    NNTP_ARTICLE_HEADER_USER_AGENT,

    /**
     * Indicates the presence of the "NNTP-Posting-Date:" header in the article.
     */
    NNTP_ARTICLE_HEADER_NNTP_POSTING_DATE,

    /**
     * Indicates the presence of the "Approved:" header in the article.
     */
    NNTP_ARTICLE_HEADER_APPROVED,

    /**
     * Indicates the presence of the "Archive:" header in the article.
     */
    NNTP_ARTICLE_HEADER_ARCHIVE,

    /**
     * Indicates the presence of the "Control:" header in the article.
     */
    NNTP_ARTICLE_HEADER_CONTROL,

    /**
     * Indicates the presence of the "Distribution:" header in the article.
     */
    NNTP_ARTICLE_HEADER_DISTRIBUTION,

    /**
     * Indicates the presence of the "Expires:" header in the article.
     */
    NNTP_ARTICLE_HEADER_EXPIRES,

    /**
     * Indicates the presence of the "Followup-To:" header in the article.
     */
    NNTP_ARTICLE_HEADER_FOLLOWUP_TO,

    /**
     * Indicates the presence of the "Injection-Date:" header in the article.
     */
    NNTP_ARTICLE_HEADER_INJECTION_DATE,

    /**
     * Indicates the presence of the "Injection-Info:" header in the article.
     */
    NNTP_ARTICLE_HEADER_INJECTION_INFO,

    /**
     * Indicates the presence of the "References:" header in the article.
     */
    NNTP_ARTICLE_HEADER_REFERENCES,

    /**
     * Indicates the presence of the "Summary:" header in the article.
     */
    NNTP_ARTICLE_HEADER_SUMMARY,

    /**
     * Indicates the presence of the "Supersedes:" header in the article.
     */
    NNTP_ARTICLE_HEADER_SUPERSEDES,

    /**
     * Indicates the presence of the "Xref:" header in the article.
     */
    NNTP_ARTICLE_HEADER_XREF,
};

/**
 * @brief Types of simple commands used for the COMMAND_SIMPLE event (ipd_nntp_simple_command_enumeration).
 */
enum ipd_nntp_simple_command_enumeration {

    /**
     * Client requests to abort the SASL authentication process.
     */
    NNTP_SIMPLE_COMMAND_ABORT_SASL_AUTHENTICATION,

    /**
     * Client requests to establish a TLS session with the server.
     */
    NNTP_SIMPLE_COMMAND_REQUEST_TLS_SESSION,

    /**
     * Client requests posting an article.
     */
    NNTP_SIMPLE_COMMAND_POST,

    /**
     * Select previous article in that newsgroup.
     */
    NNTP_SIMPLE_COMMAND_SELECT_PREVIOUS_ARTICLE,

    /**
     * Select next article in that newsgroup.
     */
    NNTP_SIMPLE_COMMAND_SELECT_NEXT_ARTICLE,

    /**
     * Test if streaming of articles is supported by server.
     */
    NNTP_SIMPLE_COMMAND_TEST_MODE_STREAM,

    /**
     * Set the mode to reader.
     */
    NNTP_SIMPLE_COMMAND_SET_MODE_READER,

    /**
     * Set the mode to poster.
     */
    NNTP_SIMPLE_COMMAND_SET_MODE_POSTER,

    /**
     * Finish NNTP session.
     */
    NNTP_SIMPLE_COMMAND_QUIT,

    /**
     * List valid newsgroups and associated information.
     */
    NNTP_SIMPLE_COMMAND_LIST_VALID_NEWSGROUPS_AND_ASSOCIATED_INFORMATION,

    /**
     * List creators and creating times of valid newsgroups.
     */
    NNTP_SIMPLE_COMMAND_LIST_CREATORS_AND_CREATING_TIMES_OF_VALID_NEWSGROUPS,

    /**
     * List default values for distribution headers in new articles.
     */
    NNTP_SIMPLE_COMMAND_LIST_DEFAULT_VALUES_FOR_DISTRIBUTION_HEADER_IN_NEW_ARTICLES,

    /**
     * List newsgroup descriptions.
     */
    NNTP_SIMPLE_COMMAND_LIST_NEWSGROUP_DESCRIPTIONS,

    /**
     * List available header fields.
     */
    NNTP_SIMPLE_COMMAND_LIST_AVAILABLE_HEADER_FIELDS,

    /**
     * List fields in the overview database.
     */
    NNTP_SIMPLE_COMMAND_LIST_FIELDS_IN_OVERVIEW_DATABASE,

    /**
     * Requests the server to list its capabilities.
     */
    NNTP_SIMPLE_COMMAND_SHOW_CAPABILITIES,

    /**
     * Requests the server to list its UTC date and time.
     */
    NNTP_SIMPLE_COMMAND_SHOW_SERVER_UTC_DATE_TIME,

    /**
     * Requests the server to show help about the available commands.
     */
    NNTP_SIMPLE_COMMAND_SHOW_HELP,
};

/**
 * @brief Unknown NNTP command.
 */
struct ipd_nntp_command_unknown {

    /** Base command. */
    struct ipd_charset_string base_command;

    /** Line of arguments. */
    struct ipd_charset_string arguments;
};

/**
 * @brief "Listgroup" command.
 */
struct ipd_nntp_command_listgroup {

    /** The group from which articles will be listed. */
    struct ipd_charset_string group;

    /** Lowest article number. */
    u64 range_begin;

    /** Highest article number. */
    u64 range_end;
};

/**
 * @brief "Newnews" and "Newgroups" commands.
 */
struct ipd_nntp_command_new_news_groups {

    /** Newsgroups of interest (only for new news). */
    struct ipd_charset_string newsgroups_of_interest;

    /** Year since which the new news will be listed. */
    u16 since_year;

    /** Month since which the new news will be listed. */
    u8 since_month;

    /** Day since which the new news will be listed. */
    u8 since_day;

    /** Hour since which the new news will be listed. */
    u8 since_hour;

    /** Minute since which the new news will be listed. */
    u8 since_minute;

    /** Second since which the new news will be listed. */
    u8 since_second;

    /** If the time is specified in the GMT time zone. */
    u8 is_gmt :1;
};

/**
 * @brief "Over" command.
 */
struct ipd_nntp_command_over {

    /** Article-id. */
    struct ipd_charset_string article_id;

    /** Lowest article number. */
    u64 range_begin;

    /** Highest article number. */
    u64 range_end;
};

/**
 * @brief "Hdr" command.
 */
struct ipd_nntp_command_hdr {

    /** Header we want to list. */
    struct ipd_charset_string header;

    /** Article-id. */
    struct ipd_charset_string article_id;

    /** Lowest article number. */
    u64 range_begin;

    /** Highest article number. */
    u64 range_end;
};

/**
 * @brief "Authinfo SASL" command.
 */
struct ipd_nntp_command_authinfo_sasl {

    /** Desired authentication mechanism. */
    struct ipd_charset_string mechanism;

    /** Initial response. */
    struct ipd_charset_string initial_response;
};

/**
 * @brief Article-fetching commands and responses.
 */
struct ipd_nntp_article {

    /** Article-id. */
    struct ipd_charset_string article_id;

    /** Article number. */
    u64 number;

    /** If it denotes success (1) or failure (0). */
    u8 success;

    /** Message given by the server. */
    struct ipd_charset_string message;
};

/**
 * @brief Article header.
 */
struct ipd_nntp_article_header {

    /** Type of the header. */
    enum ipd_nntp_header_enumeration type;

    /** Header name. */
    struct ipd_charset_string name;

    /** Header content. */
    struct ipd_charset_string value;

    /** E-mail address if it exists. */
    struct ipd_charset_string email;
};

/**
 * @brief General NNTP response.
 */
struct ipd_nntp_response {

    /** Response code that denotes the message type. */
    enum ipd_nntp_response_informative_enumeration code;

    /** If it denotes success (1) or failure (0). */
    u8 success;

    /** Message given by the server. */
    struct ipd_charset_string message;
};

/**
 * @brief "Use capability" response.
 */
struct ipd_nntp_response_use_capability {

    /** If it denotes success (1) or failure (0). */
    u8 success;

    /** Capability to be used by the client. */
    struct ipd_charset_string capability;

    /** Message given by the server. */
    struct ipd_charset_string message;
};

/**
 * @brief "Date time" response.
 */
struct ipd_nntp_response_datetime {

    /** If it denotes success (1) or failure (0). */
    u8 success;

    /** UTC date on the server. */
    struct ipd_charset_string utc_date;

    /** UTC time on the server. */
    struct ipd_charset_string utc_time;

    /** Message given by the server. */
    struct ipd_charset_string message;
};

/**
 * @brief "Group selected" response.
 */
struct ipd_nntp_response_group_selected {

    /** If it denotes success (1) or failure (0). */
    u8 success;

    /** Number of articles in group. */
    u32 number;

    /** Lowest article number. */
    u32 low;

    /** Highest article number. */
    u32 high;

    /** Name of the group. */
    struct ipd_charset_string group;

    /** Message given by the server. */
    struct ipd_charset_string message;
};

/**
 * @brief Article overview information line.
 */
struct ipd_nntp_overview_information_line {

    /** Article number. */
    u64 article_number;

    /** Subject. */
    struct ipd_charset_string subject;

    /** Sender. */
    struct ipd_charset_string from;

    /** Date. */
    struct ipd_charset_string date;

    /** Article-id. */
    struct ipd_charset_string article_id;

    /** References header. */
    struct ipd_charset_string references;

    /** Size in bytes. */
    u64 bytes;

    /** Number of lines. */
    u64 lines;

    /** Number of other headers in the array. */
    u8 other_number;

    /** Other headers. */
    struct ipd_charset_string * other;
};

/**
 * @brief Response for "list active" command.
 */
struct ipd_nntp_response_valid_newsgroup_and_associated_information {

    /** Name of the group. */
    struct ipd_charset_string group;

    /** Highest article number. */
    u64 high;

    /** Lowest article number. */
    u64 low;

    /** Posting status of the group (writable, read-only, moderated). */
    u8 status;
};

/**
 * @brief Response for "list active.times" command.
 */
struct ipd_nntp_response_creator_and_creating_time_of_valid_newsgroup {

    /** Name of the group. */
    struct ipd_charset_string group;

    /** Epoch time when the group was created. */
    u64 epoch_time;

    /** Name and/or e-mail address of the creator. */
    struct ipd_charset_string creator;
};

/**
 * @brief Response for "list newsgroups" command.
 */
struct ipd_nntp_response_newsgroup_description {

    /** Name of the group. */
    struct ipd_charset_string group;

    /** Description of the group. */
    struct ipd_charset_string description;
};

/**
 * @brief Response for "hdr" command.
 */
struct ipd_nntp_response_header_content {

    /** Article number. */
    u64 article_number;

    /** Content of the specified header. */
    struct ipd_charset_string header_content;
};

/**
 * @brief "Auth SASL" responses.
 */
struct ipd_nntp_response_auth_sasl {

    /** If it denotes success (1) or failure (0). */
    u8 success;

    /** Server challenge. */
    struct ipd_charset_string server_challenge;

    /** Message given by the server. */
    struct ipd_charset_string message;
};

/**
 * @brief "ybegin" attachment header.
 */
struct ipd_nntp_header_ybegin {

    /** Part number. */
    u32 part_number;

    /** Total number of parts. */
    u32 total_parts;

    /** Typical length of a standard attachment line in the article. */
    u32 typical_line_length;

    /** Size of the total file (not the part). */
    u64 total_file_size;

    /** Name of the file. */
    struct ipd_charset_string file_name;
};

/**
 * @brief "ypart" attachment header.
 */
struct ipd_nntp_header_ypart {

    /** Indicates the beginning of the block in the original file (in bytes). */
    u64 begin_of_the_block_in_original_file_in_bytes;

    /** Indicates the end of the block in the original file (in bytes). */
    u64 end_of_the_block_in_original_file_in_bytes;
};

/**
 * @brief "yend" attachment header.
 */
struct ipd_nntp_header_yend {

    /** Part number. */
    u32 part_number;

    /** Size of the currently transmitted part. */
    u64 current_size;

    /** CRC-32 checksum of the entire file. */
    struct ipd_charset_string crc32;

    /** CRC-32 checksum of the currently transmitted part. */
    struct ipd_charset_string part_crc32;
};

/**
 * @brief Structure used to cover chunks of text stream.
 */
struct ipd_nntp_article_textstream_chunk {

    /** Data in the chunk. */
    struct ipd_charset_string data;

    /** Offset from the beginning of the original data. */
    u64 offset;

    /** Set to 1 when the data is yencoded. */
    u8 yencoded;

    /** Indicates if this is the last chunk of the text stream (1) or not (0). */
    u8 last_chunk;
};


/**
 * @brief Data structure for advanced NNTP events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_NNTP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** NNTP event subtype. */
    PACE2_advanced_nntp_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_CLIENT_DIRECTION event. */
        int client_direction;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_INVALID event. */
        char const * invalid;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_AUTHENTICATION_USERNAME event. */
        struct ipd_charset_string command_authentication_username;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_AUTHENTICATION_PASSWORD event. */
        struct ipd_charset_string command_authentication_password;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_AUTHENTICATION_SASL event. */
        struct ipd_nntp_command_authinfo_sasl command_authentication_sasl;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_SASL_AUTHENTICATION_ACCEPTED_WITH_SUCCESS_DATA
         * event. */
        struct ipd_nntp_response_auth_sasl response_sasl_authentication_accepted_with_success_data;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_SASL_CONTINUE_WITH_SASL_EXCHANGE event. */
        struct ipd_nntp_response_auth_sasl response_sasl_continue_with_sasl_exchange;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_SASL_CLIENT_CHALLENGE event. */
        struct ipd_charset_string sasl_client_challenge;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_GET_ARTICLE event. */
        struct ipd_nntp_article command_get_article;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_GET_ARTICLE_BODY event. */
        struct ipd_nntp_article command_get_article_body;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_GET_ARTICLE_HEADER event. */
        struct ipd_nntp_article command_get_article_header;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_GET_ARTICLE_STATUS event. */
        struct ipd_nntp_article command_get_article_status;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_FOLLOWS event. */
        struct ipd_nntp_article response_article_follows;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_HEADERS_FOLLOW event. */
        struct ipd_nntp_article response_article_headers_follow;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_BODY_FOLLOWS event. */
        struct ipd_nntp_article response_article_body_follows;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_CHECK_IF_ARTICLE_DESIRED_BY_SERVER event. */
        struct ipd_charset_string command_check_if_article_desired_by_server;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_EXISTS_AND_SELECTED event. */
        struct ipd_nntp_article response_article_exists_and_selected;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_CHECK_IF_ARTICLE_DESIRED_BY_SERVER_IN_STREAMING
         * event. */
        struct ipd_charset_string command_check_if_article_desired_by_server_in_streaming;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_SEND_ARTICLE_USING_STREAMING_MODE event. */
        struct ipd_nntp_article response_send_article_using_streaming_mode;

        /** Contains data for a
         * @ref PACE2_ADVANCED_NNTP_RESPONSE_TRANSFER_OF_ARTICLE_USING_STREAMING_TEMPORARY_UNAVAILABLE event. */
        struct ipd_nntp_article response_transfer_of_article_using_streaming_temporary_unavailable;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_IN_STREAMING_NOT_WANTED event. */
        struct ipd_nntp_article response_article_in_streaming_not_wanted;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_TRANSFERRED_OK_USING_STREAMING_MODE event. */
        struct ipd_nntp_article response_article_transferred_ok_using_streaming_mode;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_TRANSFER_FAILED_USING_STREAMING_MODE
         * event. */
        struct ipd_nntp_article response_article_transfer_failed_using_streaming_mode;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_ARTICLE_IN_STREAMING_FOLLOWS event. */
        struct ipd_nntp_article command_article_in_streaming_follows;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_ARTICLE_HEADER event. */
        struct ipd_nntp_article_header article_header;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_ARTICLE_BODY_CHUNK event. */
        struct ipd_nntp_article_textstream_chunk article_body_chunk;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_ARTICLE_ATTACHMENT_YENC_HEADER_BEGIN event. */
        struct ipd_nntp_header_ybegin article_attachment_yenc_header_begin;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_ARTICLE_ATTACHMENT_YENC_HEADER_PART event. */
        struct ipd_nntp_header_ypart article_attachment_yenc_header_part;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_ARTICLE_ATTACHMENT_CHUNK event. */
        struct ipd_nntp_article_textstream_chunk article_attachment_chunk;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_ARTICLE_ATTACHMENT_YENC_HEADER_END event. */
        struct ipd_nntp_header_yend article_attachment_yenc_header_end;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_SIMPLE event. */
        enum ipd_nntp_simple_command_enumeration command_simple;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_INFORMATIVE event. */
        struct ipd_nntp_response response_informative;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_UNKNOWN event. */
        struct ipd_nntp_command_unknown command_unknown;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_UNKNOWN event. */
        struct ipd_nntp_response response_unknown;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_UNKNOWN_NEXT_LINE event. */
        struct ipd_charset_string response_unknown_next_line;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_SELECT_GROUP event. */
        struct ipd_charset_string command_select_group;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_LIST_GROUP event. */
        struct ipd_nntp_command_listgroup command_list_group;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_LIST_CUSTOM event. */
        struct ipd_charset_string command_list_custom;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_LISTED event. */
        u64 article_listed;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_GROUP_SELECTED event. */
        struct ipd_nntp_response_group_selected response_group_selected;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_VALID_NEWSGROUP_AND_ASSOCIATED_INFORMATION event. */
        struct ipd_nntp_response_valid_newsgroup_and_associated_information
            response_valid_newsgroup_and_associated_information;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_CREATOR_AND_CREATING_TIME_OF_VALID_NEWSGROUP
         * event. */
        struct ipd_nntp_response_creator_and_creating_time_of_valid_newsgroup
            response_creator_and_creating_time_of_valid_newsgroup;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_DEFAULT_VALUE_FOR_DISTRIBUTION_HEADER_IN_NEW_ARTICLES
         * event. */
        struct ipd_charset_string response_default_value_for_distribution_header_in_new_articles;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_NEWSGROUP_DESCRIPTION event. */
        struct ipd_nntp_response_newsgroup_description response_newsgroup_description;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_AVAILABLE_HEADER_FIELD event. */
        struct ipd_charset_string response_available_header_field;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_FIELD_IN_OVERVIEW_DATABASE event. */
        struct ipd_charset_string response_field_in_overview_database;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_LIST_CUSTOM_LINE event. */
        struct ipd_charset_string response_list_custom_line;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_OVER event. */
        struct ipd_nntp_command_over command_over;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_OVERVIEW_INFORMATION_LINE event. */
        struct ipd_nntp_overview_information_line response_overview_information_line;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_GET_NEW_GROUPS event. */
        struct ipd_nntp_command_new_news_groups command_get_new_groups;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_CAPABILITY event. */
        struct ipd_charset_string response_capability;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_SERVER_IN_WRONG_MODE_USE_INDICATED_CAPABILITY
         * event. */
        struct ipd_nntp_response_use_capability response_server_in_wrong_mode_use_indicated_capability;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_SERVER_UTC_DATE_TIME event. */
        struct ipd_nntp_response_datetime response_server_utc_date_time;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_HELP_LINE event. */
        struct ipd_charset_string response_help_line;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_GET_NEW_NEWS event. */
        struct ipd_nntp_command_new_news_groups command_get_new_news;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_ARTICLE_ID_OF_NEW_ARTICLE event. */
        struct ipd_charset_string response_article_id_of_new_article;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_COMMAND_GET_HEADER_FIELD event. */
        struct ipd_nntp_command_hdr command_get_header_field;

        /** Contains data for a @ref PACE2_ADVANCED_NNTP_RESPONSE_HEADER_FIELD_CONTENT event. */
        struct ipd_nntp_response_header_content response_header_field_content;

    } event_data;

} PACE2_advanced_nntp_event;


#endif

#endif /* NNTP_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
