/** @file      ocsp_events.h
  * @brief     OCSP event definitions.
  * @details   @copydoc doc_ocsp
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef OCSP_EVENTS_H
#define OCSP_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced OCSP event.
 */
#define PACE2_ADVANCED_OCSP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced OCSP events.
 */
typedef enum {
    /** This event indicates that the a new OCSP connection was seen. */
    PACE2_ADVANCED_OCSP_NEW_CONNECTION,
    /** This event indicates that the OCSP connection ended. */
    PACE2_ADVANCED_OCSP_CONNECTION_END,
    /** This event indicates that the decoding of the OCSP connection cannot be continued. It provides a message describing the failure reason. */
    PACE2_ADVANCED_OCSP_INVALID,
    /** This event is sent for every issuer name hash in a Certificate ID in OCSP. Provides the content for the issuer name hash. */
    PACE2_ADVANCED_OCSP_CERTIFICATE_ID_ISSUER_NAME_HASH,
    /** This event is sent for every issuer key hash in a Certificate ID in OCSP. Provides the issuer key hash. */
    PACE2_ADVANCED_OCSP_CERTIFICATE_ID_ISSUER_KEY_HASH,
    /** This event is sent for every serial number in a Certificate ID in OCSP. Provides the serial number. */
    PACE2_ADVANCED_OCSP_CERTIFICATE_ID_SERIAL_NUMBER,
    /** This event is sent for every Certificate status in OCSP. Provides the certificate status. */
    PACE2_ADVANCED_OCSP_CERTIFICATE_STATUS,

    PACE2_NUMBER_OF_ADVANCED_OCSP_EVENT_TYPES               /**< Number of implemented event types. */
} PACE2_advanced_OCSP_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * @brief Enumeration for the certificate status.
 */
typedef enum {
    PACE2_OCSP_CERTIFICATE_STATUS_GOOD = 0,             /**< Signals that Certificate Status is good */
    PACE2_OCSP_CERTIFICATE_STATUS_REVOKED = 1,          /**< Signals that Certificate Status is revoked */
    PACE2_OCSP_CERTIFICATE_STATUS_UNKNOWN = 2           /**< Signals that Certificate Status is unknown */
} PACE2_ocsp_certificate_status_enum;

/**
 * @brief Enumeration for the certificate revoke reason.
 */
typedef enum {
    PACE2_OCSP_REVOKE_REASON_UNSPECIFIED = 0,               /**< Signals a revoke by the certificate revocation list with an unspecified reason. */
    PACE2_OCSP_REVOKE_REASON_KEY_COMPROMISE = 1,            /**< Signals a revoke by the certificate revocation list with a compromised key. */
    PACE2_OCSP_REVOKE_REASON_CA_COMPROMISE = 2,             /**< Signals a revoke by the certificate revocation list with a compromised CA certificate. */
    PACE2_OCSP_REVOKE_REASON_AFFILATION_CHANGED= 3,         /**< Signals a revoke by the certificate revocation list with a change of its affinity. */
    PACE2_OCSP_REVOKE_REASON_SUPERSEDED = 4,                /**< Signals a revoke by the certificate revocation list with a certificate replacement. */
    PACE2_OCSP_REVOKE_REASON_CESSATION_OF_OPERATION = 5,    /**< Signals a revoke by the certificate revocation list with a operation termination. */
    PACE2_OCSP_REVOKE_REASON_CERTIFICATE_HOLD= 6,           /**< Signals a revoke by the certificate revocation list with the certificate is on hold. */
    PACE2_OCSP_REVOKE_REASON_REMOVE_FROM_CRL = 8,           /**< Signals a revoke that it was removed from the certificate revocation list. */
    PACE2_OCSP_REVOKE_REASON_PRIVILEGE_WITHDRAWN = 9,       /**< Signals a revoke by the certificate revocation list with the certificate privilege withdrawn. */
    PACE2_OCSP_REVOKE_REASON_AA_COMPROMISE = 10             /**< Signals a revoke by the certificate revocation list with a compromised authority attribute. */
} PACE2_ocsp_revoke_reason;

/**
 * @brief Structure for the certificate status.
 */
typedef struct PACE2_ocsp_certificate_status_struct {
    PACE2_ocsp_certificate_status_enum status;              /**< Status value of the certificate */
    struct {
        struct ipd_charset_string revocation_time;          /**< in case of status == @ref PACE2_OCSP_CERTIFICATE_STATUS_REVOKED the revocation time */
        PACE2_ocsp_revoke_reason revocation_reason;         /**< in case of status == @ref PACE2_OCSP_CERTIFICATE_STATUS_REVOKED the revocation reason */
    } revoked_info;
} PACE2_ocsp_certificate_status;

/**
 * @brief Data structure for advanced OCSP events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_ADVANCED_OCSP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_advanced_OCSP_event_type meta_data_type;                          /**< OCSP event subtype. */
    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;
    union {
        struct ipd_charset_string issuer_name_hash;                         /**< Contains data of a @ref PACE2_ADVANCED_OCSP_CERTIFICATE_ID_ISSUER_NAME_HASH event. */
        struct ipd_charset_string issuer_key_hash;                          /**< Contains data of a @ref PACE2_ADVANCED_OCSP_CERTIFICATE_ID_ISSUER_KEY_HASH event. */
        struct ipd_charset_string serial_number;                            /**< Contains data of a @ref PACE2_ADVANCED_OCSP_CERTIFICATE_ID_SERIAL_NUMBER event. */
        struct PACE2_ocsp_certificate_status_struct certificate_status;     /**< Contains data of a @ref PACE2_ADVANCED_OCSP_CERTIFICATE_STATUS event. */
        const char *invalid;                                                /**< Contains data of a @ref PACE2_ADVANCED_OCSP_INVALID event. */
    } event_data;                                                           /**< Named union which contains the data of the sub-event types. */
} PACE2_advanced_OCSP_event;

#endif

#endif /* OCSP_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
