/** @file      oscar_events.h
  * @brief     OSCAR event definitions.
  * @details   @copydoc doc_oscar
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef OSCAR_EVENTS_H
#define OSCAR_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "rtp_events.h"
#include "sip_events.h"

/* Event Group ******************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced OSCAR event.
 */
#define PACE2_ADVANCED_OSCAR_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/
/**
 * @brief Types for advanced Oscar events.
 */
typedef enum {

    /** This event indicates a new OSCAR connection. It does not provide any data. */
    PACE2_ADVANCED_OSCAR_NEW_CONNECTION,

    /** This event indicates the end of an OSCAR connection. It does not provide any data. */
    PACE2_ADVANCED_OSCAR_CLOSE_CONNECTION,

    /** This event indicates the receive of a SNAC message. */
    PACE2_ADVANCED_OSCAR_SNAC_DATA,

    /** Provides a struct ipd_oscar_self_info * with information about the connecting user. */
    PACE2_ADVANCED_OSCAR_SELF_INFO,

    /** Provides a struct ipd_charset_string * with the login name of the connecting user. */
    PACE2_ADVANCED_OSCAR_SIGN_IN,

    /** Provides a struct ipd_charset_string * with the login name of the connecting user. */
    PACE2_ADVANCED_OSCAR_LOG_ON,

    /** Provides a struct ipd_charset_string * with the login name of a buddy on the user's buddy list. */
    PACE2_ADVANCED_OSCAR_BUDDY_ADD,

    /** Provides a struct ipd_oscar_nickname * with nickname information about a buddy. */
    PACE2_ADVANCED_OSCAR_BUDDY_NICKNAME,

    /** Provides a struct ipd_charset_string * with the login name of a user who is coming online. */
    PACE2_ADVANCED_OSCAR_USER_ON,

    /** Provides a struct ipd_charset_string * with the login name of a user who is going offline. */
    PACE2_ADVANCED_OSCAR_USER_OFF,

    /** Provides a struct ipd_charset_string * with the login name of a user who is becoming 'not available'. */
    PACE2_ADVANCED_OSCAR_USER_NOT_AVAILABLE,

    /** Provides a struct ipd_charset_string * with the login name of a user who is becoming 'busy'. */
    PACE2_ADVANCED_OSCAR_USER_BUSY,

    /** Provides a struct ipd_charset_string * with the login name of a user who is going 'away'. */
    PACE2_ADVANCED_OSCAR_USER_AWAY,

    /** Provides a struct ipd_charset_string * with the login name of a user who is going 'invisible'. */
    PACE2_ADVANCED_OSCAR_USER_INVISIBLE,

    /** Provides a struct ipd_oscar_message_data * with information about an outgoing private message. */
    PACE2_ADVANCED_OSCAR_MESSAGE_OUT,

    /** Provides a struct ipd_oscar_message_data * with information about an incoming private message. */
    PACE2_ADVANCED_OSCAR_MESSAGE_IN,

    /** Provides a struct ipd_oscar_message_data * with information about an outgoing chatroom message. */
    PACE2_ADVANCED_OSCAR_CHAT_MESSAGE_OUT,

    /** Provides a struct ipd_oscar_message_data * with information about an incoming chatroom message. */
    PACE2_ADVANCED_OSCAR_CHAT_MESSAGE_IN,

    /** Provides a struct ipd_oscar_chat_data * with information about a chatroom which the user is joining. */
    PACE2_ADVANCED_OSCAR_CHAT_JOIN,

    /** Provides a struct ipd_oscar_chat_data * with information about a chatroom which the user is leaving. */
    PACE2_ADVANCED_OSCAR_CHAT_LEAVE,

    /** Provides a struct ipd_oscar_ft_data * with information about a file transfer which is being announced. */
    PACE2_ADVANCED_OSCAR_FT_ANNOUNCE,

    /** Provides a struct ipd_oscar_ft_data * with information about a file transfer which is being accepted. */
    PACE2_ADVANCED_OSCAR_FT_ACCEPT,

    /** Provides a struct ipd_oscar_ft_data * with information about a file transfer which is being rejected. */
    PACE2_ADVANCED_OSCAR_FT_REJECT,

    /** Provides a struct ipd_oscar_ft_data * with information about a file transfer which is being canceled. */
    PACE2_ADVANCED_OSCAR_FT_CANCEL,

    /** Provides a struct ipd_oscar_ft_data * with information about a file transfer which is staring. */
    PACE2_ADVANCED_OSCAR_FT_START,

    /** Provides a struct ipd_oscar_ft_data * with information about a file transfer and a segment
     * of file transfer data. */
    PACE2_ADVANCED_OSCAR_FT_DATA,

    /** Provides a struct ipd_charset_string * with a message that the user is setting as information
     * about him/herself. */
    PACE2_ADVANCED_OSCAR_USER_INFO_SENT,

    /** Provides a struct ipd_oscar_voip_data * with information about a voice over ip session
     * which is being announced. */
    PACE2_ADVANCED_OSCAR_VOIP_ANNOUNCE,

    /** Provides a struct ipd_oscar_voip_data * with information about a voice over ip session
     * which is being rejected. */
    PACE2_ADVANCED_OSCAR_VOIP_REJECT,

    /** Provides a struct ipd_oscar_voip_data * with information about a voice over ip session
     * which is being canceled. */
    PACE2_ADVANCED_OSCAR_VOIP_CANCEL,

    /** Provides a struct ipd_sip_message * with a SIP message which is related to an OSCAR voice over ip session.
     *  This event is actually issued by the SIP decoder. */
    PACE2_ADVANCED_OSCAR_SIP_MESSAGE,

    /** Provides a struct ipd_rtp_data * with a RTP packet which is related to an OSCAR voice over ip session.
      * This event is actually issued by the RTP decoder. */
    PACE2_ADVANCED_OSCAR_RTP_DATA,

    /** This event is issued when decoding of an OSCAR connection fails. It provides a const char * to a message
     * describing the failure reason. */
    PACE2_ADVANCED_OSCAR_INVALID,

    /** This event is issued when an OSCAR connection ends. It does not provide any data. */
    PACE2_ADVANCED_OSCAR_END_OF_FLOW,

    /** This event is issued when a SSRC value is found. Provides a struct ipd_sip_message * with the current SSRC. */
    PACE2_ADVANCED_OSCAR_RTP_SSRC,

    /** RTP end of flow. */
    PACE2_ADVANCED_OSCAR_RTP_END_OF_FLOW,

    /** RTP invalid. */
    PACE2_ADVANCED_OSCAR_RTP_INVALID,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_OSCAR_EVENT_TYPES

} PACE2_advanced_oscar_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/** Structure for events that provide information about an OSCAR instant message. */
struct ipd_oscar_message_data {

    /** Depending on the event type the user or chatroom the message is from. */
    struct ipd_charset_string from;

    /** Depending on the event type the user or chatroom the message is sent to. */
    struct ipd_charset_string to;

    /** The actual message text. */
    struct ipd_charset_string message;
};

/** Structure for events that provide information about an OSCAR chatroom. */
struct ipd_oscar_chat_data {

    /** A generated chatroom name. */
    struct ipd_charset_string chatroom;

    /** The chatroom ID found in the OSCAR protocol */
    struct ipd_charset_string id;
};

/** Structure for events that provide information about an OSCAR file transfer. */
struct ipd_oscar_ft_data {

    /** A unique cookie identifying the file transfer. */
    struct ipd_charset_string cookie;

    /** The connected user of the main OSCAR connection. */
    struct ipd_charset_string user;

    /** The user sending the file. */
    struct ipd_charset_string from;

    /** The user receiving the file. */
    struct ipd_charset_string to;

    /** The name of the actual transferred file. */
    struct ipd_charset_string filename;

    /** The size of the transfer file part in bytes. */
    u32 size;

    /** The count of parts, which will be send. */
    u16 part_count;

    /** The number of the actual part. */
    u16 part_number;

    /** The number of files, which will be send. */
    u16 file_count;

    /** The number of the actual file. */
    u16 file_number;

    /** A segment of actual file transfer data. */
    struct ipd_charset_string data;
};

/** Structure for events that provide information about an OSCAR voice over ip session. */
struct ipd_oscar_voip_data {

    /** The initiating user of the voice over ip session. */
    struct ipd_charset_string from;

    /** The user being 'called' in the voice over ip session. */
    struct ipd_charset_string to;

    /** A unique session ID describing the voice over ip session. */
    struct ipd_charset_string av_session_id;
};

/** Structure for events that provide information about OSCAR buddy nicknames. */
struct ipd_oscar_nickname {

    /** The user the nickname belongs to. */
    struct ipd_charset_string user;

    /** The nickname. */
    struct ipd_charset_string nickname;
};


/** Structure for the IPD_EVENT_OSCAR_SELF_INFO event providing information about the connecting user. */
struct ipd_oscar_self_info {

    /** The login name of the user. */
    struct ipd_charset_string user;

    /** A flag indicating if the login name is entirely composed of numeric characters. */
    u8 user_name_is_numeric;
};

/** Enumeration with main subset of the known SNAC ID's (also known as Family ID's) for the oscar protocol. */
enum ipd_oscar_snac_family_id_value {
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_GENERIC = 0x0001,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_LOCATION = 0x0002,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_BUDDY_LIST = 0x0003,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_ICBM = 0x0004,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_ADVERTISEMENT = 0x0005,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_INVITATION = 0x0006,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_ADMINISTRATION = 0x0007,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_POPUP = 0x0008,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_PRIVACY = 0x0009,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_USER_LOOKUP = 0x000a,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_USER_STATS = 0x000b,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_TRANSLATION = 0x000c,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_CHATROOM = 0x000d,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_CHAT = 0x000e,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_DIRECTORY = 0x000f,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_BUDDY_ICON = 0x0010,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_SSI = 0x0013,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_EXTENSION = 0x0015,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_AUTHORIZATION = 0x0017,
    IPD_OSCAR_SNAC_FAMILY_ID_ENUM_BROADCAST = 0x0085,
};

/**
 * Returns a short name for the given SNAC ID. INFO: Not every ID is known, for members,
 * which are not contained in @c enum ::ipd_oscar_snac_family_id_value the result will
 * be "unknown".
 */
char const * ipd_oscar_snac_family_id_to_name(u16 id);

/** Structure for oscar snac data. */
struct ipd_oscar_snac_data {

    /** The SNAC ID (also known as family ID). */
    u16 family_id;

    /** The subtype ID of the SNAC ID. */
    u16 subtype_id;

    /** The contained SNAC data. */
    struct ipd_charset_string data;
};

/**
 * @brief Data structure for advanced OSCAR events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_OSCAR_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** OSCAR event subtype. */
    PACE2_advanced_oscar_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_SNAC_DATA event. */
        struct ipd_oscar_snac_data snac_data;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_SELF_INFO event. */
        struct ipd_oscar_self_info self_info;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_SIGN_IN event. */
        struct ipd_charset_string sign_in;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_LOG_ON event. */
        struct ipd_charset_string log_on;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_BUDDY_ADD event. */
        struct ipd_charset_string buddy_add;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_BUDDY_NICKNAME event. */
        struct ipd_oscar_nickname buddy_nickname;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_USER_ON event. */
        struct ipd_charset_string user_on;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_USER_OFF event. */
        struct ipd_charset_string user_off;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_USER_NOT_AVAILABLE event. */
        struct ipd_charset_string user_not_available;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_USER_BUSY event. */
        struct ipd_charset_string user_busy;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_USER_AWAY event. */
        struct ipd_charset_string user_away;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_USER_INVISIBLE event. */
        struct ipd_charset_string user_invisible;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_MESSAGE_OUT event. */
        struct ipd_oscar_message_data message_out;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_MESSAGE_IN event. */
        struct ipd_oscar_message_data message_in;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_CHAT_MESSAGE_OUT event. */
        struct ipd_oscar_message_data chat_message_out;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_CHAT_MESSAGE_IN event. */
        struct ipd_oscar_message_data chat_message_in;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_CHAT_JOIN event. */
        struct ipd_oscar_chat_data chat_join;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_CHAT_LEAVE event. */
        struct ipd_oscar_chat_data chat_leave;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_FT_ANNOUNCE event. */
        struct ipd_oscar_ft_data ft_announce;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_FT_ACCEPT event. */
        struct ipd_oscar_ft_data ft_accept;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_FT_REJECT event. */
        struct ipd_oscar_ft_data ft_reject;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_FT_CANCEL event. */
        struct ipd_oscar_ft_data ft_cancel;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_FT_START event. */
        struct ipd_oscar_ft_data ft_start;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_FT_DATA event. */
        struct ipd_oscar_ft_data ft_data;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_USER_INFO_SENT event. */
        struct ipd_charset_string user_info_send;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_VOIP_ANNOUNCE event. */
        struct ipd_oscar_voip_data voip_announce;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_VOIP_REJECT event. */
        struct ipd_oscar_voip_data voip_reject;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_VOIP_CANCEL event. */
        struct ipd_oscar_voip_data voip_cancel;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_SIP_MESSAGE event. */
        struct ipd_sip_message sip_message;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_RTP_DATA event. */
        struct ipd_rtp_data rtp_data;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_INVALID event. */
        char const * invalid;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_RTP_INVALID event. */
        char const * rtp_invalid;

        /** Contains data for a @ref PACE2_ADVANCED_OSCAR_RTP_SSRC event. */
        struct ipd_rtp_ssrc rtp_ssrc;

    } event_data;

} PACE2_advanced_oscar_event;

#endif /* PACE_KERNEL_LIB */

#endif /* OSCAR_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
