/** @file      pop3_events.h
  * @brief     POP3 event definitions.
  * @details   @copydoc doc_pop3
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef POP3_EVENTS_H
#define POP3_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "imf_events.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced POP3 event.
 */
#define PACE2_ADVANCED_POP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced pop3 events.
 */
typedef enum {

    /** This event is sent when the first payload data has been processed and the client and server roles have been
     * determined. The provided data is a pointer to an int that contains the client direction (0 or 1). */
    PACE2_ADVANCED_POP3_CLIENT_DIRECTION,

    /** This event provides the text of the initial server greeting as a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_SERVER_GREETING,

    /** Provides a struct ipd_charset_string * with the login name of the user. */
    PACE2_ADVANCED_POP3_USER,

    /** Provides a struct ipd_charset_string * with the password of the user, if available. */
    PACE2_ADVANCED_POP3_PASS,

    /** MD5 digest sent by the user during APOP authentication; @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_APOP_DIGEST,

    /** Indicates that the client issued a RETR command; @c data is a NULL pointer. */
    PACE2_ADVANCED_POP3_RETR,

    /** Indicates that the client issued a TOP command; @c data is a NULL pointer.  */
    PACE2_ADVANCED_POP3_TOP,

    /** Indicates that the client issued a QUIT command; @c data is a NULL pointer.  */
    PACE2_ADVANCED_POP3_QUIT,

    /** Indicates that the server answers to a LIST command; @c data is a ipd_pop3_list_entry.  */
    PACE2_ADVANCED_POP3_LIST_ENTRY,

    /** Indicates that the server answers to a UIDL command; @c data is a ipd_pop3_list_entry.  */
    PACE2_ADVANCED_POP3_UIDL_ENTRY,

    /** Indicates that the server answers to a LIST command and the end was reached; @c data is a NULL pointer.  */
    PACE2_ADVANCED_POP3_LIST_END,

    /** Indicates that the server answers to a UIDL command and the end was reached; @c data is a NULL pointer.  */
    PACE2_ADVANCED_POP3_UIDL_END,

    /** Value of the From: header field of a transmitted mail message; @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_POP3_IMF_FROM,

    /** Value of the To: header field of a transmitted mail message; @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_POP3_IMF_TO,

    /** Value of the Cc: header field of a transmitted mail message; @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_POP3_IMF_CC,

    /** Value of the Bcc: header field of a transmitted mail message; @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_POP3_IMF_BCC,

    /** Value of the Subject: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_SUBJECT,

    /** Value of the Date: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_DATE,

    /** Value of the Sender: header field of a transmitted mail message; @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_POP3_IMF_SENDER,

    /** Value of the Reply-To: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_POP3_IMF_REPLY_TO,

    /** Value of the Message-Id: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_MESSAGE_ID,

    /** Value of the In-Reply-To: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_IN_REPLY_TO,

    /** Value of the References: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_REFERENCES,

    /** Value of the Comments: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_COMMENTS,

    /** Value of the Keywords: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_KEYWORDS,

    /** Value of the Trace: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_TRACE,

    /** Value of the Resent-Date: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_RESENT_DATE,

    /** Value of the Resent-From: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_POP3_IMF_RESENT_FROM,

    /** Value of the Resent-Sender: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_POP3_IMF_RESENT_SENDER,

    /** Value of the Resent-To: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_POP3_IMF_RESENT_TO,

    /** Value of the Resent-Cc: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_POP3_IMF_RESENT_CC,

    /** Value of the Resent-Bcc: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_POP3_IMF_RESENT_BCC,

    /** Value of the Resent-Msg-Id: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_RESENT_MSG_ID,

    /** Value of the MIME-Version: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_MIME_VERSION,

    /** Complete header field (including field name).
     * Sent for all message or part header fields of a transmitted message not  covered by more specific events;
     * @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_POP3_IMF_OTHER,

    /** The entire header of a mail message; @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_HEADER,

    /** Value of the Content-Type: header field of a message or MIME part;
     * @c data points to a struct ipd_imf_content_type. */
    PACE2_ADVANCED_POP3_IMF_CONTENT_TYPE,

    /** Value of the Content-Transfer-Encoding: header field of a message or MIME part;
     * @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_POP3_IMF_CONTENT_TRANSFER_ENCODING,

    /** Value of the Content-Disposition: header field of a MIME part;
     * @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_POP3_IMF_CONTENT_DISPOSITION,

    /** The entire header of a MIME part; @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_POP3_IMF_PART_HEADER,

    /** This event indicates the begin of a byte stream (MIME part body, multipart preamble or epilogue);
     * @c data points to a struct ipd_charset_string containing the section number of the part. */
    PACE2_ADVANCED_POP3_IMF_DATA_BEGIN,

    /** A block of data in a byte stream; @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_POP3_IMF_DATA,

    /** This event indicates the end of a byte stream;
     * @c data points to a struct ipd_charset_string containing the section number of the part. */
    PACE2_ADVANCED_POP3_IMF_DATA_END,

    /** This event indicates the end of a mail message; @c data is a NULL pointer */
    PACE2_ADVANCED_POP3_IMF_DONE,

    /** This event indicates the start of a TLS connection. */
    PACE2_ADVANCED_POP3_STARTTLS,

    /** This event is issued when decoding of an POP3 connection fails.
     * It provides a char const * to a message describing the failure reason. */
    PACE2_ADVANCED_POP3_INVALID,

    /** This event indicates the end of a connection. It does not provide any data. */
    PACE2_ADVANCED_POP3_CONNECTION_END,

    /** The server's response to the most recent command. This event is only emitted if the most recent POP3
     * command was reported in a (non-IMF) PADE event. @c data points to a struct ipd_pop3_response. */
    PACE2_ADVANCED_POP3_RESPONSE,

    /** This event indicates the receive of the given POP3 command and provides with
     * the struct ipd_pop3_command * structure all details about the command. */
    PACE2_ADVANCED_POP3_COMMAND,

    /** Value of Return-Path: header field of a transmitted mail message; denotes the address where
     * to send error messages; @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_RETURN_PATH,

    /** Value of Received: header field of a transmitted mail message; an array of strings: each hop
     * across the path from sender to receiver adds information about the respective sender.
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_POP3_IMF_RECEIVED,

    /** This event indicates a missing data length associated with a specific MIME section number. */
    PACE2_ADVANCED_POP3_IMF_DATA_MISSING,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_POP3_EVENT_TYPES

} PACE2_advanced_pop3_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/** POP3 response line. */
struct ipd_pop3_response {

    /** Complete text of the line sent by the server. */
    struct ipd_charset_string response_line;

    /** Set to 1 if the status indicator is +OK, 0 otherwise. */
    u8 success;
};

/**
 * Structure for POP3 command, which was received.
 */
struct ipd_pop3_command {

    /** Complete command. */
    struct ipd_charset_string command_line;

    /** Command code like AUTH or RETR. */
    struct ipd_charset_string command_code;

    /** Command arguments like 5 by RETR. */
    struct ipd_charset_string command_arguments;

    /** Contained message number or "0" if no message number is contained. */
    u32 message_number;
};

/**
 * Structure for a POP3 LIST or POP3 UIDL answer from the server.
 */
struct ipd_pop3_list_entry {

    /** The additional property from list (size in bytes) or uidl (unique ID of the message) for this entry. */
    struct ipd_charset_string property;

    /** The message number of the entry. */
    u32 message_number;
};

/**
 * @brief Data structure for advanced pop3 events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_POP3_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** POP3 event subtype. */
    PACE2_advanced_pop3_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_POP3_CLIENT_DIRECTION event. */
        int  client_direction;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_SERVER_GREETING event. */
        struct ipd_charset_string server_greeting;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_USER event. */
        struct ipd_charset_string user;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_PASS event. */
        struct ipd_charset_string pass;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_APOP_DIGEST event. */
        struct ipd_charset_string apop_digest;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_LIST_ENTRY event. */
        struct ipd_pop3_list_entry list_entry;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_UIDL_ENTRY event. */
        struct ipd_pop3_list_entry uidl_entry;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_FROM event. */
        struct ipd_imf_address imf_from;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_TO event. */
        struct ipd_imf_address imf_to;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_CC event. */
        struct ipd_imf_address imf_cc;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_BCC event. */
        struct ipd_imf_address imf_bcc;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_SUBJECT event. */
        struct ipd_charset_string imf_subject;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_DATE event. */
        struct ipd_charset_string imf_date;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_SENDER event. */
        struct ipd_imf_address imf_sender;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_REPLY_TO event. */
        struct ipd_imf_address imf_reply_to;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_MESSAGE_ID event. */
        struct ipd_charset_string imf_message_id;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_IN_REPLY_TO event. */
        struct ipd_charset_string imf_in_reply_to;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_REFERENCES event. */
        struct ipd_charset_string imf_references;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_COMMENTS event. */
        struct ipd_charset_string imf_comments;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_KEYWORDS event. */
        struct ipd_charset_string imf_keywords;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_TRACE event. */
        struct ipd_charset_string imf_trace;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_RESENT_DATE event. */
        struct ipd_charset_string imf_resent_date;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_RESENT_FROM event. */
        struct ipd_imf_address imf_resent_from;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_RESENT_SENDER event. */
        struct ipd_imf_address imf_resent_sender;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_RESENT_TO event. */
        struct ipd_imf_address imf_resent_to;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_RESENT_CC event. */
        struct ipd_imf_address imf_resent_cc;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_RESENT_BCC event. */
        struct ipd_imf_address imf_resent_bcc;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_RESENT_MSG_ID event. */
        struct ipd_charset_string imf_resent_msg_id;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_MIME_VERSION event. */
        struct ipd_charset_string imf_mime_version;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_OTHER event. */
        struct ipd_imf_part_data  imf_other;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_HEADER event. */
        struct ipd_charset_string imf_header;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_CONTENT_TYPE event. */
        struct ipd_imf_content_type imf_content_type;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_CONTENT_TRANSFER_ENCODING event. */
        struct ipd_imf_part_data imf_content_transfer_encoding;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_CONTENT_DISPOSITION event. */
        struct ipd_imf_part_data imf_content_disposition;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_PART_HEADER event. */
        struct ipd_imf_part_data imf_part_header;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_DATA_BEGIN event. */
        struct ipd_charset_string imf_data_begin;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_DATA event. */
        struct ipd_imf_part_data imf_data;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_DATA_END event. */
        struct ipd_charset_string imf_data_end;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_INVALID event. */
        char const * invalid;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_RESPONSE event. */
        struct ipd_pop3_response response;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_COMMAND event. */
        struct ipd_pop3_command command;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_RETURN_PATH event. */
        struct ipd_charset_string imf_return_path;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_RECEIVED event. */
        struct ipd_charset_string imf_received;

        /** Contains data for a @ref PACE2_ADVANCED_POP3_IMF_DATA_MISSING event. */
        struct ipd_imf_missing_data imf_data_missing;

    } event_data;

} PACE2_advanced_pop3_event;

#endif

#endif /* POP3_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
