/** @file      q931_events.h
  * @brief     Q.931 event definitions.
  * @details   @copydoc doc_q931
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef Q931_EVENTS_H
#define Q931_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Group ******************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced Q.931 event.
 */
#define PACE2_ADVANCED_Q931_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/
/**
 * @brief Types for advanced Q.931 events.
 */
typedef enum {

    /**
     * Q.931 message header.
     *
     * Provides: PACE2_q931_message_header *.
     */
    PACE2_ADVANCED_Q931_MESSAGE_HEADER,

    /**
     * Reserved value.
     *
     * Provides: u8 *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_RESERVED,

    /**
     * Indicates to the destination user that another USER INFORMATION message will follow.
     *
     * Provides: NULL.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_MORE_DATA,

    /**
     * Indicates completion of called party number.
     *
     * Provides: NULL.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_SENDING_COMPLETE,

    /**
     * Indicates how repeated information elements should be interpreted.
     *
     * Provides: PACE2_q931_repeat_indication *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_REPEAT_INDICATOR,

    /**
     * Describes the congestion status of the call.
     *
     * Provides: PACE2_q931_congestion_level *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_CONGESTION_LEVEL,

    /**
     * Information element, which data is not decoded, but provided in a raw binary form.
     *
     * Provides: PACE2_q931_information_element_general *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_GENERAL,

    /**
     * Indicates a shift to a different codeset.
     *
     * Provides: PACE2_q931_information_element_shift *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_SHIFT,

    /**
     * User-user information element.
     *
     * Provides: PACE2_q931_information_element_user_user *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_USER_USER,

    /**
     * Supplies display information that may be displayed to the user.
     *
     * Provides: struct ipd_charset_string *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_DISPLAY,

    /**
     * Indicates a requested I.231 bearer service to be provided by the network.
     *
     * Provides: PACE2_q931_information_element_bearer_capability *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_BEARER_CAPABILITY,

    /**
     * Optional information for the requested I.231 bearer service to be provided by the network.
     *
     * Provides: PACE2_q931_information_element_bearer_capability_optional_extension *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_BEARER_CAPABILITY_OPTIONAL_EXTENSION,

    /**
     * Identifies the called party of a call.
     *
     * Provides: PACE2_q931_information_element_called_party_number *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_CALLED_PARTY_NUMBER,

    /**
     * Identifies the origin of a call.
     *
     * Provides: PACE2_q931_information_element_calling_party_number *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_CALLING_PARTY_NUMBER,

    /**
     * Describes an event which occurred during the life of a call.
     *
     * Provides: PACE2_q931_information_element_progress_indicator *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_PROGRESS_INDICATOR,

    /**
     * Information for the user.
     *
     * Provides: PACE2_q931_information_element_cause *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_CAUSE,

    /**
     * Indicates which number is connected to a call.
     *
     * Provides: PACE2_q931_information_element_calling_party_number *.
     */
    PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_CONNECTED_NUMBER,

    /**
     * Indicates that the header is invalid.
     *
     * Provides: char const *.
     */
    PACE2_ADVANCED_Q931_INVALID,

    /**
     * Number of implemented event types.
     */
    PACE2_NUMBER_OF_ADVANCED_Q931_EVENT_TYPES

} PACE2_advanced_q931_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * @brief Enumeration for the message direction.
 */
typedef enum {
    PACE2_Q931_MESSAGE_DIRECTION_FROM_SIDE_ORIGINATING_CALL_REFERENCE = 0,  /**< The message is sent from the side that originates the call reference. */
    PACE2_Q931_MESSAGE_DIRECTION_TO_SIDE_ORIGINATING_CALL_REFERENCE = 1,    /**< The message is sent to the side that originates the call reference. */
    PACE2_Q931_MESSAGE_DIRECTION_UNKNOWN                                    /**< The message direction is unknown. */
} PACE2_q931_message_direction;

/**
 * @brief Enumeration for the message class.
 */
typedef enum {
    PACE2_Q931_MESSAGE_CLASS_UNKNOWN,                           /**< Unknown message class. */
    PACE2_Q931_MESSAGE_CLASS_NATIONALLY_SPECIFIC_MESSAGE_TYPE,  /**< Escape to nationally specific message type. */
    PACE2_Q931_MESSAGE_CLASS_CALL_ESTABLISHMENT_MESSAGE,        /**< Call establishment message. */
    PACE2_Q931_MESSAGE_CLASS_CALL_INFORMATION_PHASE_MESSAGE,    /**< Call information phase message. */
    PACE2_Q931_MESSAGE_CLASS_CALL_CLEARING_MESSAGE,             /**< Call clearing message. */
    PACE2_Q931_MESSAGE_CLASS_MISCELLANEOUS_MESSAGE              /**< Miscellaneous message. */
} PACE2_q931_message_class;

/**
 * @brief Enumeration for the message type.
 */
typedef enum {
    PACE2_Q931_MESSAGE_TYPE_ALERTING = 1,                   /**< ALERTING. */
    PACE2_Q931_MESSAGE_TYPE_CALL_PROCEEDING = 2,            /**< CALL PROCEEDING. */
    PACE2_Q931_MESSAGE_TYPE_PROGRESS = 3,                   /**< PROGRESS. */
    PACE2_Q931_MESSAGE_TYPE_SETUP = 5,                      /**< SETUP. */
    PACE2_Q931_MESSAGE_TYPE_GROUP_SERVICE = 6,              /**< GROUP SERVICE. */
    PACE2_Q931_MESSAGE_TYPE_CONNECT = 7,                    /**< CONNECT. */
    PACE2_Q931_MESSAGE_TYPE_RESYNC_REQUEST = 8,             /**< RESYNC REQUEST. */
    PACE2_Q931_MESSAGE_TYPE_RESYNC_RESPONSE = 9,            /**< RESYNC RESPONSE. */
    PACE2_Q931_MESSAGE_TYPE_VERSION = 10,                   /**< VERSION. */
    PACE2_Q931_MESSAGE_TYPE_GROUP_SERVICE_ACKNOWLEDE = 11,  /**< GROUP SERVICE ACKNOWLEDE. */
    PACE2_Q931_MESSAGE_TYPE_SETUP_ACKNOWLEDGE = 13,         /**< SETUP ACKNOWLEDGE. */
    PACE2_Q931_MESSAGE_TYPE_CONNECT_ACKNOWLEDGE = 15,       /**< CONNECT ACKNOWLEDGE. */
    PACE2_Q931_MESSAGE_TYPE_USER_INFORMATION = 32,          /**< USER INFORMATION. */
    PACE2_Q931_MESSAGE_TYPE_SUSPEND_REJECT = 33,            /**< SUSPEND REJECT. */
    PACE2_Q931_MESSAGE_TYPE_RESUME_REJECT = 34,             /**< RESUME REJECT. */
    PACE2_Q931_MESSAGE_TYPE_RESUME_HOLD = 36,               /**< RESUME HOLD. */
    PACE2_Q931_MESSAGE_TYPE_SUSPEND = 37,                   /**< SUSPEND. */
    PACE2_Q931_MESSAGE_TYPE_RESUME = 38,                    /**< RESUME. */
    PACE2_Q931_MESSAGE_TYPE_HOLD_ACKNOWLEDGE = 40,          /**< HOLD ACKNOWLEDGE. */
    PACE2_Q931_MESSAGE_TYPE_SUSPEND_ACKNOWLEDGE = 45,       /**< SUSPEND ACKNOWLEDGE. */
    PACE2_Q931_MESSAGE_TYPE_RESUME_ACKNOWLEDGE = 46,        /**< RESUME ACKNOWLEDGE. */
    PACE2_Q931_MESSAGE_TYPE_HOLD_REJECT = 48,               /**< HOLD REJECT. */
    PACE2_Q931_MESSAGE_TYPE_RETRIEVE = 49,                  /**< RETRIEVE. */
    PACE2_Q931_MESSAGE_TYPE_RETRIEVE_ACKNOWLEDGE = 51,      /**< RETRIEVE ACKNOWLEDGE. */
    PACE2_Q931_MESSAGE_TYPE_RETRIEVE_REJECT = 55,           /**< RETRIEVE REJECT. */
    PACE2_Q931_MESSAGE_TYPE_DETACH = 64,                    /**< DETACH. */
    PACE2_Q931_MESSAGE_TYPE_DISCONNECT = 69,                /**< DISCONNECT. */
    PACE2_Q931_MESSAGE_TYPE_RESTART = 70,                   /**< RESTART. */
    PACE2_Q931_MESSAGE_TYPE_DETACH_ACKNOWLEDGE = 72,        /**< DETACH ACKNOWLEDGE. */
    PACE2_Q931_MESSAGE_TYPE_RELEASE = 77,                   /**< RELEASE. */
    PACE2_Q931_MESSAGE_TYPE_RESTART_ACKNOWLEDGE = 78,       /**< RESTART ACKNOWLEDGE. */
    PACE2_Q931_MESSAGE_TYPE_RELEASE_COMPLETE = 90,          /**< RELEASE COMPLETE. */
    PACE2_Q931_MESSAGE_TYPE_SEGMENT = 96,                   /**< SEGMENT. */
    PACE2_Q931_MESSAGE_TYPE_FACILITY = 98,                  /**< FACILITY. */
    PACE2_Q931_MESSAGE_TYPE_REGISTER = 100,                 /**< REGISTER. */
    PACE2_Q931_MESSAGE_TYPE_FACILITY_ACKNOWLEDGE = 106,     /**< FACILITY ACKNOWLEDGE. */
    PACE2_Q931_MESSAGE_TYPE_NOTIFY = 110,                   /**< NOTIFY. */
    PACE2_Q931_MESSAGE_TYPE_FACILITY_REJECT = 114,          /**< FACILITY REJECT. */
    PACE2_Q931_MESSAGE_TYPE_STATUS_ENQUIRY = 117,           /**< STATUS ENQUIRY. */
    PACE2_Q931_MESSAGE_TYPE_CONGESTION_CONTROL = 121,       /**< CONGESTION CONTROL. */
    PACE2_Q931_MESSAGE_TYPE_INFORMATION = 123,              /**< INFORMATION. */
    PACE2_Q931_MESSAGE_TYPE_STATUS = 125                    /**< STATUS. */
} PACE2_q931_message_type;

/**
 * @brief Enumeration for the information element type.
 */
typedef enum {
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_SEGMENTED_MESSAGE = 0,                          /**< Segmented message. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_BEARER_CAPABILITY = 4,                          /**< Bearer capability. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_CAUSE = 8,                                      /**< Cause. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_CONNECTED_ADDRESS = 12,                         /**< Connected address. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_EXTENDED_FACILITY = 13,                         /**< Extended facility. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_CALL_IDENTITY = 16,                             /**< Call identity. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_CALL_STATE = 20,                                /**< Call state. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_CHANNEL_IDENTIFICATION = 24,                    /**< Channel identification. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_DATA_LINK_CONNECTION_IDENTIFIER = 25,           /**< Data link connection identifier. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_CODING_DECODING_PROCESSING = 27,                /**< Coding decoding processing. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_FACILITY = 28,                                  /**< Facility. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_PROGRESS_INDICATOR = 30,                        /**< Progress indicator. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_NETWORK_SPECIFIC_FACILITIES = 32,               /**< Network-specific facilities. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_TERMINAL_CAPABILITIES = 36,                     /**< Terminal capabilities. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_NOTIFICATION_INDICATOR = 39,                    /**< Notification indicator. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_DISPLAY = 40,                                   /**< Display. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_DATE_TIME = 41,                                 /**< Date/time. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_KEYPAD_FACILITY = 44,                           /**< Keypad facility. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_KEYPAD_ECHO = 48,                               /**< Keypad echo. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_INFORMATION_REQUEST = 50,                       /**< Information request. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_SIGNAL = 52,                                    /**< Signal. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_SWITCHHOOK = 54,                                /**< Switchhook. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_FEATURE_ACTIVATION = 56,                        /**< Feature activation. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_FEATURE_INDICATION = 57,                        /**< Feature indication. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_SERVICE_PROFILE_IDENTIFICATION = 58,            /**< Service profile identification. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_ENDPOINT_IDENTIFIER = 59,                       /**< Endpoint identifier. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_INFORMATION_RATE = 64,                          /**< Information rate. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_PRECEDENCE_LEVEL = 65,                          /**< Precedence level. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_END_TO_END_TRANSIT_DELAY = 66,                  /**< End-to-end transit delay. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_TRANSIT_DELAY_SELECTION_AND_INDICATION = 67,    /**< Transit delay selection and indication. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_PACKET_LAYER_BINARY_PARAMETERS = 68,            /**< Packet layer binary parameters. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_PACKET_LAYER_WINDOW_SIZE = 69,                  /**< Packet layer window size. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_PACKET_SIZE = 70,                               /**< Packet size. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_CLOSED_USER_GROUP = 71,                         /**< Closed user group. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_LINK_LAYER_CORE_PARAMETERS = 72,                /**< Link layer core parameters. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_LINK_LAYER_PROTOCOL_PARAMETERS = 73,            /**< Link layer protocol parameters. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_REVERSE_CHARGING_INDICATION = 74,               /**< Reverse charging indication. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_CONNECTED_NUMBER = 76,                          /**< Connected number. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_CONNECTED_SUBADDRESS = 77,                      /**< Connected subaddress. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_X_213_PRIORITY = 80,                            /**< X.213 priority. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_REPORT_TYPE = 81,                               /**< Report type. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_LINK_INTEGRITY_VERIFICATION = 83,               /**< Link integrity verification. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_PVC_STATUS = 87,                                /**< PVC status. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_CALLING_PARTY_NUMBER = 108,                     /**< Calling party number. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_CALLING_PARTY_SUBADDRESS = 109,                 /**< Calling party subaddress. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_CALLED_PARTY_NUMBER = 112,                      /**< Called party number. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_CALLED_PARTY_SUBADDRESS = 113,                  /**< Called party subaddress. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_REDIRECTING_NUMBER = 116,                       /**< Redirecting number. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_REDIRECTION_NUMBER = 118,                       /**< Redirection number. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_TRANSIT_NETWORK_SELECTION = 120,                /**< Transit network selection. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_RESTART_INDICATOR = 121,                        /**< Restart indicator. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_LOW_LAYER_COMPATIBILITY = 124,                  /**< Low layer compatibility. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_HIGH_LAYER_COMPATIBILITY = 125,                 /**< High layer compatibility. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_USER_USER = 126,                                /**< User-user. */
    PACE2_Q931_INFORMATION_ELEMENT_TYPE_ESCAPE_FOR_EXTENSION = 127                      /**< Escape for extension. */
} PACE2_q931_information_element_type;

/**
 * @brief Enumeration for the protocol discriminator.
 */
typedef enum {
    PACE2_Q931_PROTOCOL_DISCRIMINATOR_USER_SPECIFIC_PROTOCOL = 0,                               /**< User-specific protocol. */
    PACE2_Q931_PROTOCOL_DISCRIMINATOR_OSI_HIGH_LAYER_PROTOCOLS = 1,                             /**< OSI high layer protocols. */
    PACE2_Q931_PROTOCOL_DISCRIMINATOR_RECOMMENDATION_X_244 = 2,                                 /**< Recommendation X.244. */
    PACE2_Q931_PROTOCOL_DISCRIMINATOR_RESERVED_FOR_SYSTEM_MANAGEMENT_CONVERGENCE_FUNCTION = 3,  /**< Reserved for system management convergence function. */
    PACE2_Q931_PROTOCOL_DISCRIMINATOR_IA5_CHARACTERS = 4,                                       /**< IA5 characters. */
    PACE2_Q931_PROTOCOL_DISCRIMINATOR_X_208_AND_X_209_CODED_USER_INFORMATION = 5,               /**< X.208 and X.209 coded user information. */
    PACE2_Q931_PROTOCOL_DISCRIMINATOR_RECOMMENDATION_V_120_RATE_ADAPTION = 7,                   /**< Recommendation V.120 rate adaption. */
    PACE2_Q931_PROTOCOL_DISCRIMINATOR_Q_931_I_451_USER_NETWORK_CALL_CONTROL_MESSAGES = 8        /**< Q.931/I.451 user-network call control messages. */
} PACE2_q931_protocol_discriminator;

/**
 * @brief Enumeration for the repeat indication.
 */
typedef enum pace2_q931_repeat_indication {
    PACE2_Q931_REPEAT_INDICATION_PRIORITIZED_LIST_FOR_SELECTING_ONE_POSSIBILITY = 2 /**<  Prioritized list for selecting one possibility. */
} PACE2_q931_repeat_indication;

/**
 * @brief Enumeration for the congestion level.
 */
typedef enum pace2_q931_congestion_level {
    PACE2_Q931_CONGESTION_LEVEL_RECEIVER_READY = 0,     /**< Receiver ready. */
    PACE2_Q931_CONGESTION_LEVEL_RECEIVER_NOT_READY = 15 /**< Receiver not ready. */
} PACE2_q931_congestion_level;

/**
 * @brief Enumeration for the shift type.
 */
typedef enum {
    PACE2_Q931_SHIFT_TYPE_LOCKING = 0,      /**< Locking shift. */
    PACE2_Q931_SHIFT_TYPE_NON_LOCKING = 1   /**< Non-locking shift. */
} PACE2_q931_shift_type;

/**
 * @brief Enumeration for the codeset.
 */
typedef enum {
    PACE2_Q931_CODESET_0_Q931_INFORMATION_ELEMENTS = 0,                         /**< Codeset 0 (initially active): Q.931 information elements. */
    PACE2_Q931_CODESET_1_RESERVED = 1,                                          /**< Reserved. */
    PACE2_Q931_CODESET_2_RESERVED = 2,                                          /**< Reserved. */
    PACE2_Q931_CODESET_3_RESERVED = 3,                                          /**< Reserved. */
    PACE2_Q931_CODESET_4_INFORMATION_ELEMENTS_FOR_ISO_IEC_USE = 4,              /**< Codeset 4: information elements for ISO/IEC use. */
    PACE2_Q931_CODESET_5_INFORMATION_ELEMENTS_FOR_NATIONAL_USE = 5,             /**< Codeset 5: information elements for national use. */
    PACE2_Q931_CODESET_6_INFORMATION_ELEMENTS_SPEICIFC_TO_LOCAL_NETWORK = 6,    /**< Codeset 6: information elements specific to the local network (either public or private). */
    PACE2_Q931_CODESET_7_USER_SPECIFIC_INFORMATION_ELEMENTS = 7                 /**< Codeset 7: user-specific information elements. */
} PACE2_q931_codeset;

/**
 * @brief Enumeration for the coding standard.
 */
typedef enum {
    PACE2_Q931_CODING_STANDARD_ITU_T_STANDARDIZED_CODING = 0,   /**< ITU-T standardized coding. */
    PACE2_Q931_CODING_STANDARD_ISO_IEC_STANDARD = 1,            /**< ISO/IEC Standard. */
    PACE2_Q931_CODING_STANDARD_NATIONAL_STANDARD = 2,           /**< National standard. */
    PACE2_Q931_CODING_STANDARD_DEFINED_FOR_NETWORK = 3          /**< Standard defined for the network (either public or private) present on the network side of the interface. */
} PACE2_q931_coding_standard;

/**
 * @brief Enumeration for the information transfer capability.
 */
typedef enum {
    PACE2_Q931_INFORMATION_TRANSFER_CAPABILITY_SPEECH = 0,                                                          /**< Speech. */
    PACE2_Q931_INFORMATION_TRANSFER_CAPABILITY_UNRESTRICTED_DIGITAL_INFORMATION = 8,                                /**< Unrestricted digital information. */
    PACE2_Q931_INFORMATION_TRANSFER_CAPABILITY_RESTRICTED_DIGITAL_INFORMATION = 9,                                  /**< Restricted digital information. */
    PACE2_Q931_INFORMATION_TRANSFER_CAPABILITY_3_1_KHZ_AUDIO = 16,                                                  /**< 3.1 kHz audio. */
    PACE2_Q931_INFORMATION_TRANSFER_CAPABILITY_UNRESTRICTED_DIGITAL_INFORMATION_WITH_TONES_OR_ANNOUNCEMENTS = 17,   /**< Unrestricted digital information with tones/announcements. */
    PACE2_Q931_INFORMATION_TRANSFER_CAPABILITY_VIDEO = 24                                                           /**< Video. */
} PACE2_q931_information_transfer_capability;

/**
 * @brief Enumeration for the transfer mode.
 */
typedef enum {
    PACE2_Q931_TRANSFER_MODE_CIRCUIT_MODE = 0,  /**< Circuit mode. */
    PACE2_Q931_TRANSFER_MODE_PACKET_MODE = 2    /**< Packet mode. */
} PACE2_q931_transfer_mode;

/**
 * @brief Enumeration for the information transfer rate.
 */
typedef enum {
    PACE2_Q931_INFORMATION_TRANSFER_RATE_NONE = 0,                          /**< This code shall be used for packet-mode calls. */
    PACE2_Q931_INFORMATION_TRANSFER_RATE_64_KBIT_PER_SECOND = 16,           /**< 64 kbit/s. */
    PACE2_Q931_INFORMATION_TRANSFER_RATE_2_TIMES_64_KBIT_PER_SECOND = 17,   /**< 2 × 64 kbit/s. */
    PACE2_Q931_INFORMATION_TRANSFER_RATE_384_KBIT_PER_SECOND = 19,          /**< 384 kbit/s. */
    PACE2_Q931_INFORMATION_TRANSFER_RATE_1536_KBIT_PER_SECOND = 21,         /**< 1536 kbit/s. */
    PACE2_Q931_INFORMATION_TRANSFER_RATE_1920_KBIT_PER_SECOND = 23,         /**< 1920 kbit/s. */
    PACE2_Q931_INFORMATION_TRANSFER_RATE_MULTIRATE = 24,                    /**< Multirate (64 kbit/s base rate). */
} PACE2_q931_information_transfer_rate;

/**
 * @brief Enumeration for the user information layer-1 protocol.
 */
typedef enum pace2_q931_user_information_layer1_protocol {
    PACE2_Q931_USER_INFORMATION_LAYER1_PROTOCOL_ITU_T_STANDARDIZED_RATE_ADAPTION_V_110_I_460_AND_X_30 = 1,              /**< ITU-T standardized rate adaption V.110, I.460 and X.30. */
    PACE2_Q931_USER_INFORMATION_LAYER1_PROTOCOL_RECOMMENDATION_G_711_U_LAW = 2,                                         /**< Recommendation G.711 u-law. */
    PACE2_Q931_USER_INFORMATION_LAYER1_PROTOCOL_RECOMMENDATION_G_711_A_LAW = 3,                                         /**< Recommendation G.711 A-law. */
    PACE2_Q931_USER_INFORMATION_LAYER1_PROTOCOL_RECOMMENDATION_G_721_32_KBIT_PER_S_ADPCM_AND_RECOMMENDATION_I_460 = 4,  /**< Recommendation G.721 32 kbit/s ADPCM and Recommendation I.460. */
    PACE2_Q931_USER_INFORMATION_LAYER1_PROTOCOL_RECOMMENDATIONS_H_221_AND_H_242 = 5,                                    /**< Recommendations H.221 and H.242. */
    PACE2_Q931_USER_INFORMATION_LAYER1_PROTOCOL_RECOMMENDATIONS_H_223_AND_H_245 = 6,                                    /**< Recommendations H.223 and H.245. */
    PACE2_Q931_USER_INFORMATION_LAYER1_PROTOCOL_NON_ITU_T_STANDARDIZED_RATE_ADAPTION = 7,                               /**< Non-ITU-T standardized rate adaption. */
    PACE2_Q931_USER_INFORMATION_LAYER1_PROTOCOL_ITU_T_STANDARDIZED_RATE_ADAPTION_V_120 = 8,                             /**< ITU-T standardized rate adaption V.120. */
    PACE2_Q931_USER_INFORMATION_LAYER1_PROTOCOL_ITU_T_STANDARDIZED_RATE_ADAPTION_X_31_HDLC_FLAG_STUFFING = 9,           /**< ITU-T standardized rate adaption X.31 HDLC flag stuffing. */
    PACE2_Q931_USER_INFORMATION_LAYER1_PROTOCOL_RECOMMENDATION_G_728_LD_CELP = 10,                                      /**< Recommendation G.728 LD-CELP. */
    PACE2_Q931_USER_INFORMATION_LAYER1_PROTOCOL_RECOMMENDATION_G_729_CS_ACELP = 11                                      /**< Recommendation G.729 CS-ACELP. */
} PACE2_q931_user_information_layer1_protocol;

/**
 * @brief Enumeration for the synchronization mode.
 */
typedef enum {
    PACE2_Q931_SYNCHRONIZATION_MODE_SYNCHRONOUS = 0,    /**< Synchronous data. */
    PACE2_Q931_SYNCHRONIZATION_MODE_ASYNCHRONOUS = 1    /**< Asynchronous data. */
} PACE2_q931_synchronization_mode;

/**
 * @brief Enumeration for the in-band negotiation.
 */
typedef enum {
    PACE2_Q931_IN_BAND_NEGOTIATION_NOT_POSSIBLE = 0,    /**< In-band negotiation not possible. */
    PACE2_Q931_IN_BAND_NEGOTIATION_POSSIBLE = 1         /**< In-band negotiation possible. */
} PACE2_q931_in_band_negotiation;

/**
 * @brief Enumeration for the user rate.
 */
typedef enum {
    PACE2_Q931_USER_RATE_UNSPECIFIED = 0,                                   /**< Unspecified. */
    PACE2_Q931_USER_RATE_0_6_KBIT_PER_S_RECOMMENDATION_X_1 = 1,             /**< 0.6 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_1_2_KBIT_PER_S = 2,                                /**< 1.2 kbit/s. */
    PACE2_Q931_USER_RATE_2_4_KBIT_PER_S_RECOMMENDATION_X_1 = 3,             /**< 2.4 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_3_6_KBIT_PER_S = 4,                                /**< 3.6 kbit/s. */
    PACE2_Q931_USER_RATE_4_8_KBIT_PER_S_RECOMMENDATION_X_1 = 5,             /**< 4.8 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_7_2_KBIT_PER_S = 6,                                /**< 7.2 kbit/s. */
    PACE2_Q931_USER_RATE_8_KBIT_PER_S_RECOMMENDATION_I_460 = 7,             /**< 8 kbit/s Recommendation I.460. */
    PACE2_Q931_USER_RATE_9_6_KBIT_PER_S_RECOMMENDATION_X_1 = 8,             /**< 9.6 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_14_4_KBIT_PER_S = 9,                               /**< 14.4 kbit/s. */
    PACE2_Q931_USER_RATE_16_KBIT_PER_S_RECOMMENDATION_I_460 = 10,           /**< 16 kbit/s Recommendation I.460. */
    PACE2_Q931_USER_RATE_19_2_KBIT_PER_S = 11,                              /**< 19.2 kbit/s. */
    PACE2_Q931_USER_RATE_32_KBIT_PER_S_RECOMMENDATION_I_460 = 12,           /**< 32 kbit/s Recommendation I.460. */
    PACE2_Q931_USER_RATE_38_4_KBIT_PER_S_RECOMMENDATION_V_110 = 13,         /**< 38.4 kbit/s Recommendation V.110. */
    PACE2_Q931_USER_RATE_48_KBIT_PER_S_RECOMMENDATIONS_X_1 = 14,            /**< 48 kbit/s Recommendations X.1. */
    PACE2_Q931_USER_RATE_56_KBIT_PER_S = 15,                                /**< 56 kbit/s. */
    PACE2_Q931_USER_RATE_64_KBIT_PER_S_RECOMMENDATIONS_X_1 = 16,            /**< 64 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_57_6_KBIT_PER_S_RECOMMENDATION_V_14_EXTENDED = 18, /**< 57.6 kbit/s Recommendation V.14 extended. */
    PACE2_Q931_USER_RATE_28_8_KBIT_PER_S_RECOMMENDATION_V_110 = 19,         /**< 28.8 kbit/s Recommendation V.110. */
    PACE2_Q931_USER_RATE_24_KBIT_PER_S_RECOMMENDATION_V_110 = 20,           /**< 24 kbit/s Recommendation V.110. */
    PACE2_Q931_USER_RATE_0_1345_KBIT_PER_S_RECOMMENDATION_X_1 = 21,         /**< 0.1345 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_0_100_KBIT_PER_S_RECOMMENDATION_X_1 = 22,          /**< 0.100 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_0_075_PER_1_2_KBIT_PER_S_RECOMMENDATION_X_1 = 23,  /**< 0.075/1.2 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_1_2_PER_0_075_KBIT_PER_S_RECOMMENDATION_X_1 = 24,  /**< 1.2/0.075 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_0_050_KBIT_PER_S_RECOMMENDATION_X_1 = 25,          /**< 0.050 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_0_075_KBIT_PER_S_RECOMMENDATION_X_1 = 26,          /**< 0.075 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_0_110_KBIT_PER_S_RECOMMENDATION_X_1 = 27,          /**< 0.110 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_0_150_KBIT_PER_S_RECOMMENDATION_X_1 = 28,          /**< 0.150 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_0_200_KBIT_PER_S_RECOMMENDATION_X_1 = 29,          /**< 0.200 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_0_300_KBIT_PER_S_RECOMMENDATION_X_1 = 30,          /**< 0.300 kbit/s Recommendation X.1. */
    PACE2_Q931_USER_RATE_12_KBIT_PER_S = 31                                 /**< 12 kbit/s. */
} PACE2_q931_user_rate;

/**
 * @brief Enumeration for the intermediate rate.
 */
typedef enum {
    PACE2_Q931_INTERMEDIATE_RATE_NOT_USED = 0,      /**< Not used. */
    PACE2_Q931_INTERMEDIATE_RATE_8_KBIT_PER_S = 1,  /**< 8 kbit/s. */
    PACE2_Q931_INTERMEDIATE_RATE_16_KBIT_PER_S = 2, /**< 16 kbit/s. */
    PACE2_Q931_INTERMEDIATE_RATE_32_KBIT_PER_S = 3  /**< 32 kbit/s. */
} PACE2_q931_intermediate_rate;

/**
 * @brief Enumeration for the network independent clock on transmission.
 */
typedef enum {
    PACE2_Q931_NETWORK_INDEPENDENT_CLOCK_ON_TRANSMISSION_NOT_REQUIRED = 0,  /**< Not required to send data with network independent clock. */
    PACE2_Q931_NETWORK_INDEPENDENT_CLOCK_ON_TRANSMISSION_REQUIRED = 1       /**< Required to send data with network independent clock. */
} PACE2_q931_network_independent_clock_on_transmission;

/**
 * @brief Enumeration for the network independent clock on reception.
 */
typedef enum {
    PACE2_Q931_NETWORK_INDEPENDENT_CLOCK_ON_RECEPTION_NOT_ACCEPTED = 0,  /**< Cannot accept data with network independent clock (i.e. sender does not support this optional procedure). */
    PACE2_Q931_NETWORK_INDEPENDENT_CLOCK_ON_RECEPTION_ACCEPTED = 1       /**< Can accept data with network independent clock (i.e. sender does support this optional procedure). */
} PACE2_q931_network_independent_clock_on_reception;

/**
 * @brief Enumeration for the flow control on transmission.
 */
typedef enum {
    PACE2_Q931_FLOW_CONTROL_ON_TRANSMISSION_NOT_REQUIRED = 0,   /**< Not required to send data with flow control mechanism. */
    PACE2_Q931_FLOW_CONTROL_ON_TRANSMISSION_REQUIRED = 1        /**< Required to send data with flow control mechanism. */
} PACE2_q931_flow_control_on_transmission;

/**
 * @brief Enumeration for the flow control on reception.
 */
typedef enum {
    PACE2_Q931_FLOW_CONTROL_ON_RECEPTION_NOT_ACCEPTED = 0,  /**< Cannot accept data with flow control mechanism. */
    PACE2_Q931_FLOW_CONTROL_ON_RECEPTION_ACCEPTED = 1       /**< Can accept data with flow control mechanism. */
} PACE2_q931_flow_control_on_reception;

/**
 * @brief Enumeration for the rate adaption header.
 */
typedef enum {
    PACE2_Q931_RATE_ADAPTION_HEADER_NOT_INCLUDED = 0,  /**< Rate adaption header not included. */
    PACE2_Q931_RATE_ADAPTION_HEADER_INCLUDED = 1       /**< Rate adaption header included. */
} PACE2_q931_rate_adaption_header;

/**
 * @brief Enumeration for the multiple frame establishment in data link.
 */
typedef enum {
    PACE2_Q931_MULTIPLE_FRAME_ESTABLISHMENT_IN_DATA_LINK_NOT_SUPPORTED = 0, /**< Multiple frame establishment not supported. Only UI frames allowed. */
    PACE2_Q931_MULTIPLE_FRAME_ESTABLISHMENT_IN_DATA_LINK_SUPPORTED = 1      /**< Multiple frame establishment supported. */
} PACE2_q931_multiple_frame_establishment_in_data_link;

/**
 * @brief Enumeration for the mode of operation.
 */
typedef enum {
    PACE2_Q931_MODE_OF_OPERATION_BIT_TRANSPARENT = 0,   /**< Bit transparent mode of operation. */
    PACE2_Q931_MODE_OF_OPERATION_PROTOCOL_SENSITIVE = 1 /**< Protocol sensitive mode of operation. */
} PACE2_q931_mode_of_operation;

/**
 * @brief Enumeration for the logical link identifier negotiation.
 */
typedef enum {
    PACE2_Q931_LOGICAL_LINK_IDENTIFIER_NEGOTIATION_DEFAULT = 0, /**< Default, LLI = 256 only. */
    PACE2_Q931_LOGICAL_LINK_IDENTIFIER_NEGOTIATION_FULL = 1     /**< Full protocol negotiation. */
} PACE2_q931_logical_link_identifier_negotiation;

/**
 * @brief Enumeration for the message originator.
 */
typedef enum {
    PACE2_Q931_MESSAGE_ORIGINATOR_DEFAULT_ASSIGNEE = 0, /**< Message originator is "Default assignee". */
    PACE2_Q931_MESSAGE_ORIGINATOR_ASSIGNOR_ONLY = 1     /**< Message originator is "Assignor only". */
} PACE2_q931_message_originator;

/**
 * @brief Enumeration for the negotiation type.
 */
typedef enum {
    PACE2_Q931_NEGOTATION_TYPE_OUT_BAND = 0,    /**< Negotiation is done with USER INFORMATION messages on a temporary signalling connection. */
    PACE2_Q931_NEGOTATION_TYPE_IN_BAND = 1      /**< Negotiation is done in-band using logical link zero. */
} PACE2_q931_negotiation_type;

/**
 * @brief Enumeration for the number of stop bits.
 */
typedef enum {
    PACE2_Q931_NUMBER_OF_STOP_BITS_NOT_USED = 0,    /**< Not used. */
    PACE2_Q931_NUMBER_OF_STOP_BITS_1 = 1,           /**< 1 bit. */
    PACE2_Q931_NUMBER_OF_STOP_BITS_1_5 = 2,         /**< 1.5 bits. */
    PACE2_Q931_NUMBER_OF_STOP_BITS_2 = 3            /**< 2 bits. */
} PACE2_q931_number_of_stop_bits;

/**
 * @brief Enumeration for the number of data bits excluding parity bit.
 */
typedef enum {
    PACE2_Q931_NUMBER_OF_DATA_BITS_EXCLUDING_PARITY_BIT_NOT_USED = 0,   /**< Not used. */
    PACE2_Q931_NUMBER_OF_DATA_BITS_EXCLUDING_PARITY_BIT_5 = 1,          /**< 5 bit. */
    PACE2_Q931_NUMBER_OF_DATA_BITS_EXCLUDING_PARITY_BIT_7 = 2,          /**< 7 bits. */
    PACE2_Q931_NUMBER_OF_DATA_BITS_EXCLUDING_PARITY_BIT_8 = 3           /**< 8 bits. */
} PACE2_q931_number_of_data_bits_excluding_parity_bit;

/**
 * @brief Enumeration for the parity information.
 */
typedef enum {
    PACE2_Q931_PARITY_INFORMATION_ODD = 0,          /**< Odd. */
    PACE2_Q931_PARITY_INFORMATION_EVEN = 2,         /**< Even. */
    PACE2_Q931_PARITY_INFORMATION_NONE = 3,         /**< None. */
    PACE2_Q931_PARITY_INFORMATION_FORCED_TO_0 = 4,  /**< Forced to 0. */
    PACE2_Q931_PARITY_INFORMATION_FORCED_TO_1 = 5   /**< Forced to 1. */
} PACE2_q931_parity_information;

/**
 * @brief Enumeration for the duplex mode.
 */
typedef enum {
    PACE2_Q931_DUPLEX_MODE_HALF = 0,    /**< Half duplex. */
    PACE2_Q931_DUPLEX_MODE_FULL = 1     /**< Full duplex. */
} PACE2_q931_duplex_mode;

/**
 * @brief Enumeration for the modem type.
 */
typedef enum {
    PACE2_Q931_MODEM_TYPE_RECOMMENDATION_V_21 = 17,     /**< Recommendation V.21. */
    PACE2_Q931_MODEM_TYPE_RECOMMENDATION_V_22 = 18,     /**< Recommendation V.22. */
    PACE2_Q931_MODEM_TYPE_RECOMMENDATION_V_22_BIS = 19, /**< Recommendation V.22 bis. */
    PACE2_Q931_MODEM_TYPE_RECOMMENDATION_V_23 = 20,     /**< Recommendation V.23. */
    PACE2_Q931_MODEM_TYPE_RECOMMENDATION_V_26 = 21,     /**< Recommendation V.26. */
    PACE2_Q931_MODEM_TYPE_RECOMMENDATION_V_26_BIS = 22, /**< Recommendation V.26 bis. */
    PACE2_Q931_MODEM_TYPE_RECOMMENDATION_V_26_TER = 23, /**< Recommendation V.26 ter. */
    PACE2_Q931_MODEM_TYPE_RECOMMENDATION_V_27 = 24,     /**< Recommendation V.27. */
    PACE2_Q931_MODEM_TYPE_RECOMMENDATION_V_27_BIS = 25, /**< Recommendation V.27 bis. */
    PACE2_Q931_MODEM_TYPE_RECOMMENDATION_V_27_TER = 26, /**< Recommendation V.27 ter. */
    PACE2_Q931_MODEM_TYPE_RECOMMENDATION_V_29 = 27,     /**< Recommendation V.29. */
    PACE2_Q931_MODEM_TYPE_RECOMMENDATION_V_32 = 28,     /**< Recommendation V.32. */
    PACE2_Q931_MODEM_TYPE_RECOMMENDATION_V_34 = 30      /**< Recommendation V.34. */
} PACE2_q931_modem_type;

/**
 * @brief Enumeration for the user information layer-2 protocol.
 */
typedef enum pace2_q931_user_information_layer2_protocol {
    PACE2_Q931_USER_INFORMATION_LAYER2_PROTOCOL_RECOMMENDATION_Q_921_I_441 = 2,                 /**< Recommendation Q.921/I.441. */
    PACE2_Q931_USER_INFORMATION_LAYER2_PROTOCOL_RECOMMENDATION_X_25_LINK_LAYER = 6,             /**< Recommendation X.25, link layer. */
    PACE2_Q931_USER_INFORMATION_LAYER2_PROTOCOL_LAN_LOGICAL_LINK_CONTROL_ISO_IEC_8802_2 = 12    /**< LAN logical link control (ISO/IEC 8802-2). */
} PACE2_q931_user_information_layer2_protocol;

/**
 * @brief Enumeration for the user information layer-3 protocol.
 */
typedef enum pace2_q931_user_information_layer3_protocol {
    PACE2_Q931_USER_INFORMATION_LAYER3_PROTOCOL_RECOMMENDATION_Q_931 = 2,                                           /**< Recommendation Q.931. */
    PACE2_Q931_USER_INFORMATION_LAYER3_PROTOCOL_RECOMMENDATION_X_25_PACKET_LAYER = 6,                               /**< Recommendation X.25, packet layer. */
    PACE2_Q931_USER_INFORMATION_LAYER3_PROTOCOL_ISO_IEC_TR_9577_PROTOCOL_IDENTIFICATION_IN_THE_NETWORK_LAYER = 11   /**< ISO/IEC TR 9577 (Protocol identification in the network layer). */
} PACE2_q931_user_information_layer3_protocol;

/**
 * @brief Enumeration for the additional layer-3 protocol identification.
 */
typedef enum pace2_q931_additional_layer3_protocol_identification {
    PACE2_Q931_ADDITIONAL_LAYER3_PROTOCOL_IDENTIFICATION_INTERNET_PROTOCOL_RFC_791_ISO_IEC_TR_9577 = 204,   /**< Internet Protocol (RFC 791) (ISO/IEC TR 9577). */
    PACE2_Q931_ADDITIONAL_LAYER3_PROTOCOL_IDENTIFICATION_POINT_TO_POINT_PROTOCOL_RFC_1548 = 207             /**< Point-to-point Protocol (RFC 1548). */
} PACE2_q931_additional_layer3_protocol_identification;

/**
 * @brief Enumeration for the type of number.
 */
typedef enum {
    PACE2_Q931_TYPE_OF_NUMBER_UNKNOWN = 0,                  /**< Unknown. */
    PACE2_Q931_TYPE_OF_NUMBER_INTERNATIONAL_NUMBER = 1,     /**< International number. */
    PACE2_Q931_TYPE_OF_NUMBER_NATIONAL_NUMBER = 2,          /**< National number. */
    PACE2_Q931_TYPE_OF_NUMBER_NETWORK_SPECIFIC_NUMBER = 3,  /**< Network specific number. */
    PACE2_Q931_TYPE_OF_NUMBER_SUBSCRIBER_NUMBER = 4,        /**< Subscriber number. */
    PACE2_Q931_TYPE_OF_NUMBER_ABBREVIATED_NUMBER = 6,       /**< Abbreviated number. */
    PACE2_Q931_TYPE_OF_NUMBER_RESERVED_FOR_EXTENSION = 7    /**< Reserved for extension. */
} PACE2_q931_type_of_number;

/**
 * @brief Enumeration for the numbering plan.
 */
typedef enum {
    PACE2_Q931_NUMBERING_PLAN_UNKNOWN = 0,                                                  /**< Unknown. */
    PACE2_Q931_NUMBERING_PLAN_ISDN_OR_TELEPHONY_NUMBERING_PLAN_RECOMMENDATION_E_164 = 1,    /**< ISDN/telephony numbering plan (Recommendation E.164). */
    PACE2_Q931_NUMBERING_PLAN_DATA_NUMBERING_PLAN_RECOMMENDATION_X_121 = 3,                 /**< Data numbering plan (Recommendation X.121). */
    PACE2_Q931_NUMBERING_PLAN_TELEX_NUMBERING_PLAN_RECOMMENDATION_F_69 = 4,                 /**< Telex numbering plan (Recommendation F.69). */
    PACE2_Q931_NUMBERING_PLAN_NATIONAL_STANDARD_NUMBERING_PLAN = 8,                         /**< National standard numbering plan. */
    PACE2_Q931_NUMBERING_PLAN_PRIVATE_NUMBERING_PLAN = 9,                                   /**< Private numbering plan. */
    PACE2_Q931_NUMBERING_PLAN_RESERVED_FOR_EXTENSION = 15                                   /**< Reserved for extension. */
} PACE2_q931_numbering_plan;

/**
 * @brief Enumeration for the presentation indicator.
 */
typedef enum {
    PACE2_Q931_PRESENTATION_INDICATOR_PRESENTATION_ALLOWED = 0,                     /**< Presentation allowed. */
    PACE2_Q931_PRESENTATION_INDICATOR_PRESENTATION_RESTRICTED = 1,                  /**< Presentation restricted. */
    PACE2_Q931_PRESENTATION_INDICATOR_NUMBER_NOT_AVAILABLE_DUE_TO_INTERWORKING = 2, /**< Number not available due to interworking. */
    PACE2_Q931_PRESENTATION_INDICATOR_RESERVED = 3                                  /**< Reserved. */
} PACE2_q931_presentation_indicator;

/**
 * @brief Enumeration for the screening indicator.
 */
typedef enum {
    PACE2_Q931_SCREENING_INDICATOR_USER_PROVIDED_NOT_SCREENED = 0,          /**< User-provided, not screened. */
    PACE2_Q931_SCREENING_INDICATOR_USER_PROVIDED_VERIFIED_AND_PASSED = 1,   /**< User-provided, verified and passed. */
    PACE2_Q931_SCREENING_INDICATOR_USER_PROVIDED_VERIFIED_AND_FAILED = 2,   /**< User-provided, verified and failed. */
    PACE2_Q931_SCREENING_INDICATOR_NETWORK_PROVIDED = 3                     /**< Network provided. */
} PACE2_q931_screening_indicator;

/**
 * @brief Enumeration for the location.
 */
typedef enum {
    PACE2_Q931_LOCATION_USER = 0,                                       /**< User. */
    PACE2_Q931_LOCATION_PRIVATE_NETWORK_SERVING_THE_LOCAL_USER = 1,     /**< Private network serving the local user. */
    PACE2_Q931_LOCATION_PUBLIC_NETWORK_SERVING_THE_LOCAL_USER = 2,      /**< Public network serving the local user. */
    PACE2_Q931_LOCATION_TRANSIT_NETWORK = 3,                            /**< Transit network. */
    PACE2_Q931_LOCATION_PUBLIC_NETWORK_SERVING_THE_REMOTE_USER = 4,     /**< Public network serving the remote user. */
    PACE2_Q931_LOCATION_PRIVATE_NETWORK_SERVING_THE_REMOTE_USER = 5,    /**< Private network serving the remote user. */
    PACE2_Q931_LOCATION_INTERNATIONAL_NETWORK = 7,                      /**< International network. */
    PACE2_Q931_LOCATION_NETWORK_BEYOND_THE_INTERWORKING_POINT = 10      /**< Network beyond the interworking point. */
} PACE2_q931_location;

/**
 * @brief Enumeration for the progress description.
 */
typedef enum {
    PACE2_Q931_PROGRESS_DESCRIPTION_CALL_IS_NOT_END_TO_END_ISDN = 1,                                            /**< Call is not end-to-end ISDN; further call progress information may be available in-band. */
    PACE2_Q931_PROGRESS_DESCRIPTION_DESTINATION_ADDRESS_IS_NON_ISDN = 2,                                        /**< Destination address is non-ISDN. */
    PACE2_Q931_PROGRESS_DESCRIPTION_ORIGINATION_ADDRESS_IS_NON_ISDN = 3,                                        /**< Origination address is non-ISDN. */
    PACE2_Q931_PROGRESS_DESCRIPTION_CALL_RETURNED_TO_THE_ISDN = 4,                                              /**< Call has returned to the ISDN. */
    PACE2_Q931_PROGRESS_DESCRIPTION_INTERWORKING_OCCURRED_AND_RESULTED_IN_TELECOMMUNICATION_SERVICE_CHANGE = 5, /**< Interworking has occurred and has resulted in a telecommunication service change. */
    PACE2_Q931_PROGRESS_DESCRIPTION_IN_BAND_INFORMATION_OR_AN_APPROPRIATE_PATTERN_AVAILABLE = 8                 /**< In-band information or an appropriate pattern is now available. */
} PACE2_q931_progress_description;

/**
 * @brief Enumeration for the cause class.
 */
typedef enum {
    PACE2_Q931_CAUSE_CLASS_NORMAL_EVENT_0 = 0,                      /**< Normal event. */
    PACE2_Q931_CAUSE_CLASS_NORMAL_EVENT_1 = 1,                      /**< Normal event. */
    PACE2_Q931_CAUSE_CLASS_RESOURCE_UNAVAILABLE = 2,                /**< Resource unavailable. */
    PACE2_Q931_CAUSE_CLASS_SERVICE_OR_OPTION_NOT_AVAILABLE = 3,     /**< Service or option not available. */
    PACE2_Q931_CAUSE_CLASS_SERVICE_OR_OPTION_NOT_IMPLEMENTED = 4,   /**< Service or option not implemented. */
    PACE2_Q931_CAUSE_CLASS_INVALID_MESSAGE = 5,                     /**< Invalid message (e.g. parameter out of range). */
    PACE2_Q931_CAUSE_CLASS_PROTOCOL_ERROR = 6,                      /**< Protocol error (e.g. unknown message). */
    PACE2_Q931_CAUSE_CLASS_INTERWORKING = 7                         /**< Interworking. */
} PACE2_q931_cause_class;

/**
 * @brief Enumeration for the cause value.
 */
typedef enum {
    PACE2_Q931_CAUSE_VALUE_UNALLOCATED_NUMBER = 1,                                                                      /**< Unallocated (unassigned) number. */
    PACE2_Q931_CAUSE_VALUE_NO_ROUTE_TO_SPECIFIED_TRANSIT_NETWORK = 2,                                                   /**< No route to specified transit network. */
    PACE2_Q931_CAUSE_VALUE_NO_ROUTE_TO_DESTINATION = 3,                                                                 /**< No route to destination. */
    PACE2_Q931_CAUSE_VALUE_SEND_SPECIAL_INFORMATION_TONE = 4,                                                           /**< Send special information tone. */
    PACE2_Q931_CAUSE_VALUE_MISDIALLED_TRUNK_PREFIX = 5,                                                                 /**< Misdialled trunk prefix. */
    PACE2_Q931_CAUSE_VALUE_CHANNEL_UNACCEPTABLE = 6,                                                                    /**< Channel unacceptable. */
    PACE2_Q931_CAUSE_VALUE_CALL_AWARDED_AND_BEING_DELIVERED_IN_AN_ESTABLISHED_CHANNEL = 7,                              /**< Call awarded and being delivered in an established channel. */
    PACE2_Q931_CAUSE_VALUE_PREEMPTION = 8,                                                                              /**< Preemption. */
    PACE2_Q931_CAUSE_VALUE_PREEMPTION_CIRCUIT_RESERVED_FOR_REUSE = 9,                                                   /**< Preemption – circuit reserved for reuse. */
    PACE2_Q931_CAUSE_VALUE_PREFIX_1_NOT_DIALED_BUT_REQUIRED = 10,                                                       /**< Prefix 1 not dialed but required. */
    PACE2_Q931_CAUSE_VALUE_MORE_DIGITS_RECEIVED_THAN_ALLOWED_CALL_IS_PROCEEDING = 11,                                   /**< More digits received than allowed, call is proceeding. */
    PACE2_Q931_CAUSE_VALUE_NORMAL_CALL_CLEARING = 16,                                                                   /**< Normal call clearing. */
    PACE2_Q931_CAUSE_VALUE_USER_BUSY = 17,                                                                              /**< User busy. */
    PACE2_Q931_CAUSE_VALUE_NO_USER_RESPONDING = 18,                                                                     /**< No user responding. */
    PACE2_Q931_CAUSE_VALUE_NO_ANSWER_FROM_USER_USER_ALERTED = 19,                                                       /**< No answer from user (user alerted). */
    PACE2_Q931_CAUSE_VALUE_SUBSCRIBER_ABSENT = 20,                                                                      /**< Subscriber absent. */
    PACE2_Q931_CAUSE_VALUE_CALL_REJECTED = 21,                                                                          /**< Call rejected. */
    PACE2_Q931_CAUSE_VALUE_NUMBER_CHANGED = 22,                                                                         /**< Number changed. */
    PACE2_Q931_CAUSE_VALUE_REDIRECTION_TO_NEW_DESTINATION = 23,                                                         /**< Redirection to new destination. */
    PACE2_Q931_CAUSE_VALUE_EXCHANGE_ROUTING_ERROR = 25,                                                                 /**< Exchange routing error. */
    PACE2_Q931_CAUSE_VALUE_NON_SELECTED_USER_CLEARING = 26,                                                             /**< Non-selected user clearing. */
    PACE2_Q931_CAUSE_VALUE_DESTINATION_OUT_OF_ORDER = 27,                                                               /**< Destination out of order. */
    PACE2_Q931_CAUSE_VALUE_INVALID_NUMBER_FORMAT_ADDRESS_INCOMPLETE = 28,                                               /**< Invalid number format (address incomplete). */
    PACE2_Q931_CAUSE_VALUE_FACILITY_REJECTED = 29,                                                                      /**< Facility rejected. */
    PACE2_Q931_CAUSE_VALUE_RESPONSE_TO_STATUS_ENQUIRY = 30,                                                             /**< Response to STATUS ENQUIRY. */
    PACE2_Q931_CAUSE_VALUE_NORMAL = 31,                                                                                 /**< Normal, unspecified. */
    PACE2_Q931_CAUSE_VALUE_CIRCUIT_OUT_OF_ORDER = 33,                                                                   /**< Circuit out of order. */
    PACE2_Q931_CAUSE_VALUE_NO_CIRCUIT_OR_CHANNEL_AVAILABLE = 34,                                                        /**< No circuit/channel available. */
    PACE2_Q931_CAUSE_VALUE_DESTINATION_UNATTAINABLE = 35,                                                               /**< Destination unattainable. */
    PACE2_Q931_CAUSE_VALUE_OUT_OF_ORDER = 36,                                                                           /**< Out of order. */
    PACE2_Q931_CAUSE_VALUE_DEGRADED_SERVICE = 37,                                                                       /**< Degraded service. */
    PACE2_Q931_CAUSE_VALUE_NETWORK_OUT_OF_ORDER = 38,                                                                   /**< Network out of order. */
    PACE2_Q931_CAUSE_VALUE_PERMANENT_FRAME_MODE_CONNECTION_OUT_OF_SERVICE = 39,                                         /**< Permanent frame mode connection out of service. */
    PACE2_Q931_CAUSE_VALUE_PERMANENT_FRAME_MODE_CONNECTION_OPERATIONAL = 40,                                            /**< Permanent frame mode connection operational. */
    PACE2_Q931_CAUSE_VALUE_TEMPORARY_FAILURE = 41,                                                                      /**< Temporary failure. */
    PACE2_Q931_CAUSE_VALUE_SWITCHING_EQUIPMENT_CONGESTION = 42,                                                         /**< Switching equipment congestion. */
    PACE2_Q931_CAUSE_VALUE_ACCESS_INFORMATION_DISCARDED = 43,                                                           /**< Access information discarded. */
    PACE2_Q931_CAUSE_VALUE_REQUESTED_CIRCUIT_OR_CHANNEL_NOT_AVAILABLE = 44,                                             /**< Requested circuit/channel not available. */
    PACE2_Q931_CAUSE_VALUE_PREEMPTED = 45,                                                                              /**< Preempted. */
    PACE2_Q931_CAUSE_VALUE_PRECEDENCE_CALL_BLOCKED = 46,                                                                /**< Precedence call blocked. */
    PACE2_Q931_CAUSE_VALUE_RESOURCE_UNAVAILABLE = 47,                                                                   /**< Resource unavailable, unspecified. */
    PACE2_Q931_CAUSE_VALUE_QUALITY_OF_SERVICE_NOT_AVAILABLE = 49,                                                       /**< Quality of service not available. */
    PACE2_Q931_CAUSE_VALUE_REQUESTED_FACILITY_NOT_SUBSCRIBED = 50,                                                      /**< Requested facility not subscribed. */
    PACE2_Q931_CAUSE_VALUE_REVERSE_CHARGING_NOT_ALLOWED = 51,                                                           /**< Reverse charging not allowed. */
    PACE2_Q931_CAUSE_VALUE_OUTGOING_CALLS_BARRED = 52,                                                                  /**< Outgoing calls barred. */
    PACE2_Q931_CAUSE_VALUE_OUTGOING_CALLS_BARRED_WITHIN_CUG = 53,                                                       /**< Outgoing calls barred within CUG. */
    PACE2_Q931_CAUSE_VALUE_INCOMING_CALLS_BARRED = 54,                                                                  /**< Incoming calls barred. */
    PACE2_Q931_CAUSE_VALUE_INCOMING_CALLS_BARRED_WITHIN_CUG = 55,                                                       /**< Incoming calls barred within CUG. */
    PACE2_Q931_CAUSE_VALUE_CALL_WAITING_NOT_SUBSCRIBED = 56,                                                            /**< Call waiting not subscribed. */
    PACE2_Q931_CAUSE_VALUE_BEARER_CAPABILITY_NOT_AUTHORIZED = 57,                                                       /**< Bearer capability not authorized. */
    PACE2_Q931_CAUSE_VALUE_BEARER_CAPABILITY_NOT_PRESENTLY_AVAILABLE = 58,                                              /**< Bearer capability not presently available. */
    PACE2_Q931_CAUSE_VALUE_INCONSISTENCY_IN_DESIGNATED_OUTGOING_ACCESS_INFORMATION_AND_SUBSCRIBER_CLASS = 62,           /**< Inconsistency in designated outgoing access information and subscriber class. */
    PACE2_Q931_CAUSE_VALUE_SERVICE_OR_OPTION_NOT_AVAILABLE = 63,                                                        /**< Service or option not available, unspecified. */
    PACE2_Q931_CAUSE_VALUE_BEARER_CAPABILITY_NOT_IMPLEMENTED = 65,                                                      /**< Bearer capability not implemented. */
    PACE2_Q931_CAUSE_VALUE_CHANNEL_TYPE_NOT_IMPLEMENTED = 66,                                                           /**< Channel type not implemented. */
    PACE2_Q931_CAUSE_VALUE_TRANSIT_NETWORK_SELECTION_NOT_IMPLEMENTED = 67,                                              /**< Transit network selection not implemented. */
    PACE2_Q931_CAUSE_VALUE_MESSAGE_NOT_IMPLEMENTED = 68,                                                                /**< Message not implemented. */
    PACE2_Q931_CAUSE_VALUE_REQUESTED_FACILITY_NOT_IMPLEMENTED = 69,                                                     /**< Requested facility not implemented. */
    PACE2_Q931_CAUSE_VALUE_ONLY_RESTRICTED_DIGITAL_INFORMATION_BEARER_CAPABILITY_IS_AVAILABLE = 70,                     /**< Only restricted digital information bearer capability is available. */
    PACE2_Q931_CAUSE_VALUE_SERVICE_OR_OPTION_NOT_IMPLEMENTED = 79,                                                      /**< Service or option not implemented, unspecified. */
    PACE2_Q931_CAUSE_VALUE_INVALID_CALL_REFERENCE_VALUE = 81,                                                           /**< Invalid call reference value. */
    PACE2_Q931_CAUSE_VALUE_IDENTIFIED_CHANNEL_DOES_NOT_EXIST = 82,                                                      /**< Identified channel does not exist. */
    PACE2_Q931_CAUSE_VALUE_A_SUSPENDED_CALL_EXISTS_BUT_THIS_CALL_IDENTITY_DOES_NOT = 83,                                /**< A suspended call exists, but this call identity does not. */
    PACE2_Q931_CAUSE_VALUE_CALL_IDENTITY_IN_USE = 84,                                                                   /**< Call identity in use. */
    PACE2_Q931_CAUSE_VALUE_NO_CALL_SUSPENDED = 85,                                                                      /**< No call suspended. */
    PACE2_Q931_CAUSE_VALUE_CALL_HAVING_THE_REQUESTED_CALL_IDENTITY_HAS_BEEN_CLEARED = 86,                               /**< Call having the requested call identity has been cleared. */
    PACE2_Q931_CAUSE_VALUE_USER_NOT_MEMBER_OF_CUG = 87,                                                                 /**< User not member of CUG. */
    PACE2_Q931_CAUSE_VALUE_INCOMPATIBLE_DESTINATION = 88,                                                               /**< Incompatible destination. */
    PACE2_Q931_CAUSE_VALUE_NON_EXISTENT_ABBREVIATED_ADDRESS_ENTRY = 89,                                                 /**< Non-existent abbreviated address entry. */
    PACE2_Q931_CAUSE_VALUE_NON_EXISTENT_CUG = 90,                                                                       /**< Non-existent CUG. */
    PACE2_Q931_CAUSE_VALUE_INVALID_TRANSIT_NETWORK_SELECTION = 91,                                                      /**< Invalid transit network selection. */
    PACE2_Q931_CAUSE_VALUE_INVALID_FACILITY_PARAMETER = 92,                                                             /**< Invalid facility parameter. */
    PACE2_Q931_CAUSE_VALUE_MESSAGE_TYPE_NON_EXISTENT_OR_NOT_IMPLEMENTED_1 = 93,                                         /**< Message type non-existent or not implemented. */
    PACE2_Q931_CAUSE_VALUE_INVALID_MESSAGE = 95,                                                                        /**< Invalid message, unspecified. */
    PACE2_Q931_CAUSE_VALUE_MANDATORY_INFORMATION_ELEMENT_IS_MISSING = 96,                                               /**< Mandatory information element is missing. */
    PACE2_Q931_CAUSE_VALUE_MESSAGE_TYPE_NON_EXISTENT_OR_NOT_IMPLEMENTED_2 = 97,                                         /**< Message type non-existent or not implemented. */
    PACE2_Q931_CAUSE_VALUE_MESSAGE_NOT_COMPATIBLE_WITH_CALL_STATE_OR_MESSAGE_TYPE_NON_EXISTENT_OR_NOT_IMPLEMENTED = 98, /**< Message not compatible with call state or message type non-existent or not implemented. */
    PACE2_Q931_CAUSE_VALUE_INFORMATION_ELEMENT_OR_PARAMETER_NON_EXISTENT_OR_NOT_IMPLEMENTED = 99,                       /**< Information element /parameter non-existent or not implemented. */
    PACE2_Q931_CAUSE_VALUE_INVALID_INFORMATION_ELEMENT_CONTENTS = 100,                                                  /**< Invalid information element contents. */
    PACE2_Q931_CAUSE_VALUE_MESSAGE_NOT_COMPATIBLE_WITH_CALL_STATE = 101,                                                /**< Message not compatible with call state. */
    PACE2_Q931_CAUSE_VALUE_RECOVERY_ON_TIMER_EXPIRY = 102,                                                              /**< Recovery on timer expiry. */
    PACE2_Q931_CAUSE_VALUE_PARAMETER_NON_EXISTENT_OR_NOT_IMPLEMENTED_PASSED_ON = 103,                                   /**< Parameter non-existent or not implemented, passed on. */
    PACE2_Q931_CAUSE_VALUE_MESSAGE_WITH_UNRECOGNIZED_PARAMETER_DISCARDED = 110,                                         /**< Message with unrecognized parameter, discarded. */
    PACE2_Q931_CAUSE_VALUE_PROTOCOL_ERROR = 111,                                                                        /**< Protocol error, unspecified. */
    PACE2_Q931_CAUSE_VALUE_INTERWORKING = 127                                                                           /**< Interworking, unspecified. */
} PACE2_q931_cause_value;

/**
 * @brief Enumeration for the bearer compatibility optional extensions.
 */
typedef enum {
    PACE2_Q931_BEARER_COMPATIBILITY_OPTIONAL_EXTENSION_TYPE_RATE_MULTIPLIER,                            /**< Rate multiplier. */
    PACE2_Q931_BEARER_COMPATIBILITY_OPTIONAL_EXTENSION_TYPE_USER_INFORMATION_LAYER1_PROTOCOL,           /**< Layer-1 protocol information. */
    PACE2_Q931_BEARER_COMPATIBILITY_OPTIONAL_EXTENSION_TYPE_USER_INFORMATION_LAYER2_PROTOCOL,           /**< Layer-2 protocol information. */
    PACE2_Q931_BEARER_COMPATIBILITY_OPTIONAL_EXTENSION_TYPE_USER_INFORMATION_LAYER3_PROTOCOL,           /**< Layer-3 protocol information. */
    PACE2_Q931_BEARER_COMPATIBILITY_OPTIONAL_EXTENSION_TYPE_ADDITIONAL_LAYER3_PROTOCOL_IDENTIFICATION,  /**< Layer-3 additional protocol identification information. */
    PACE2_Q931_BEARER_COMPATIBILITY_OPTIONAL_EXTENSION_TYPE_SET1,                                       /**< Synchronization mode, in-band negotiation state, user rate. */
    PACE2_Q931_BEARER_COMPATIBILITY_OPTIONAL_EXTENSION_TYPE_SET2,                                       /**< Intermediate rate, Network Independent Clock (NIC) on transmission and reception, flow control on transmission and reception. */
    PACE2_Q931_BEARER_COMPATIBILITY_OPTIONAL_EXTENSION_TYPE_SET3,                                       /**< Rate adaption header, multiple frame establishment in data link, mode of operation, logical link identifier negotiation, message originator, negotiation type. */
    PACE2_Q931_BEARER_COMPATIBILITY_OPTIONAL_EXTENSION_TYPE_SET4,                                       /**< Number of stop bits, number of data bits excluding parity bit, parity information. */
    PACE2_Q931_BEARER_COMPATIBILITY_OPTIONAL_EXTENSION_TYPE_SET5                                        /**< Duplex mode, modem type. */
} PACE2_q931_bearer_capability_optional_extension_type;

/**
 * @brief Data structure for Q.931 message header.
 */
typedef struct pace2_q931_message_header {
    u16 call_reference_value;                       /**< Call reference value. */
    PACE2_q931_message_direction message_direction; /**< Message direction. */
    PACE2_q931_message_class message_class;         /**< Message class. */
    PACE2_q931_message_type message_type;           /**< Message type. */
} PACE2_q931_message_header;

/**
 * @brief Data structure for a general (non-decoded) Q.931 information element.
 */
typedef struct pace2_q931_information_element_general {
    PACE2_q931_information_element_type type;   /**< Element type. */
    struct ipd_charset_string data;             /**< Element data. */
} PACE2_q931_information_element_general;

/**
 * @brief Data structure for the "Shift" Q.931 information element.
 */
typedef struct pace2_q931_information_element_shift {
    PACE2_q931_shift_type shift_type;   /**< Shift type. */
    PACE2_q931_codeset new_codeset;     /**< New codeset identification. */
} PACE2_q931_information_element_shift;

/**
 * @brief Data structure for the "User-User" Q.931 information element.
 */
typedef struct pace2_q931_information_element_user_user {
    PACE2_q931_protocol_discriminator protocol_discriminator;   /**< Protocol discriminator. */
    struct ipd_charset_string data;                             /**< Element data. */
} PACE2_q931_information_element_user_user;

/**
 * @brief Data structure for the "Bearer Capability" Q.931 information element.
 */
typedef struct pace2_q931_information_element_bearer_capability {
    PACE2_q931_coding_standard coding_standard;                                 /**< Coding standard. */
    PACE2_q931_information_transfer_capability information_transfer_capability; /**< Information transfer capability. */
    PACE2_q931_transfer_mode transfer_mode;                                     /**< Transfer mode. */
    PACE2_q931_information_transfer_rate information_transfer_rate;             /**< Information transfer rate. */
} PACE2_q931_information_element_bearer_capability;

/**
 * @brief Data structure for optional set of information (1) of the "Bearer Capability" Q.931 information element.
 */
typedef struct pace2_q931_information_element_bearer_capability_optional_extension_set1 {
    PACE2_q931_synchronization_mode synchronization_mode;                   /**< Synchronization mode. */
    PACE2_q931_in_band_negotiation in_band_negotiation;                     /**< In-band negotiation state. */
    PACE2_q931_user_rate user_rate;                                         /**< User rate. */
} PACE2_q931_information_element_bearer_capability_optional_extension_set1;

/**
 * @brief Data structure for optional set of information (2) of the "Bearer Capability" Q.931 information element.
 */
typedef struct pace2_q931_information_element_bearer_capability_optional_extension_set2 {
    PACE2_q931_intermediate_rate intermediate_rate;                                                 /**< Intermediate rate. */
    PACE2_q931_network_independent_clock_on_transmission network_independent_clock_on_transmission; /**< Network Independent Clock (NIC) on transmission (Tx). */
    PACE2_q931_network_independent_clock_on_reception network_independent_clock_on_reception;       /**< Network Independent Clock (NIC) on reception (Rx). */
    PACE2_q931_flow_control_on_transmission flow_control_on_transmission;                           /**< Flow control on transmission (Tx). */
    PACE2_q931_flow_control_on_reception flow_control_on_reception;                                 /**< Flow control on reception (Rx). */
} PACE2_q931_information_element_bearer_capability_optional_extension_set2;

/**
 * @brief Data structure for optional set of information (3) of the "Bearer Capability" Q.931 information element.
 */
typedef struct pace2_q931_information_element_bearer_capability_optional_extension_set3 {
    PACE2_q931_rate_adaption_header rate_adaption_header;                                           /**< Rate adaption header. */
    PACE2_q931_multiple_frame_establishment_in_data_link multiple_frame_establishment_in_data_link; /**< Multiple frame establishment in data link. */
    PACE2_q931_mode_of_operation mode_of_operation;                                                 /**< Mode of operation. */
    PACE2_q931_logical_link_identifier_negotiation logical_link_identifier_negotiation;             /**< Logical link identifier negotiation. */
    PACE2_q931_message_originator message_originator;                                               /**< Message originator. */
    PACE2_q931_negotiation_type negotiation_type;                                                   /**< Negotiation type. */
} PACE2_q931_information_element_bearer_capability_optional_extension_set3;

/**
 * @brief Data structure for optional set of information (4) of the "Bearer Capability" Q.931 information element.
 */
typedef struct pace2_q931_information_element_bearer_capability_optional_extension_set4 {
    PACE2_q931_number_of_stop_bits number_of_stop_bits;                                             /**< Number of stop bits. */
    PACE2_q931_number_of_data_bits_excluding_parity_bit number_of_data_bits_excluding_parity_bit;   /**< Number of data bits excluding parity bit. */
    PACE2_q931_parity_information parity_information;                                               /**< Parity information. */
} PACE2_q931_information_element_bearer_capability_optional_extension_set4;

/**
 * @brief Data structure for optional set of information (5) of the "Bearer Capability" Q.931 information element.
 */
typedef struct pace2_q931_information_element_bearer_capability_optional_extension_set5 {
    PACE2_q931_duplex_mode duplex_mode;                                     /**< Duplex mode. */
    PACE2_q931_modem_type modem_type;                                       /**< Modem type. */
} PACE2_q931_information_element_bearer_capability_optional_extension_set5;

/**
 * @brief Data structure for optional extensions of the "Bearer Capability" Q.931 information element.
 */
typedef struct pace2_q931_information_element_bearer_capability_optional_extension {
    PACE2_q931_bearer_capability_optional_extension_type optional_extension_type;
    union {
        u8 rate_multiplier;                                                                                 /**< Rate multiplier (valid only for multirate). */
        PACE2_q931_user_information_layer1_protocol user_information_layer1_protocol;                       /**< User information layer-1 protocol. */
        PACE2_q931_user_information_layer2_protocol user_information_layer2_protocol;                       /**< User information layer-2 protocol. */
        PACE2_q931_user_information_layer3_protocol user_information_layer3_protocol;                       /**< User information layer-3 protocol. */
        PACE2_q931_additional_layer3_protocol_identification additional_layer3_protocol_identification;     /**< Additional layer-3 protocol identification. */
        PACE2_q931_information_element_bearer_capability_optional_extension_set1 optional_extension_set1;   /**< Optional set of information (1). */
        PACE2_q931_information_element_bearer_capability_optional_extension_set2 optional_extension_set2;   /**< Optional set of information (2). */
        PACE2_q931_information_element_bearer_capability_optional_extension_set3 optional_extension_set3;   /**< Optional set of information (3). */
        PACE2_q931_information_element_bearer_capability_optional_extension_set4 optional_extension_set4;   /**< Optional set of information (4). */
        PACE2_q931_information_element_bearer_capability_optional_extension_set5 optional_extension_set5;   /**< Optional set of information (5). */
    } optional_extension_data;
} PACE2_q931_information_element_bearer_capability_optional_extension;

/**
 * @brief Data structure for the "Called party number" Q.931 information element.
 */
typedef struct pace2_q931_information_element_called_party_number {
    PACE2_q931_type_of_number type_of_number;           /**< Type of number. */
    PACE2_q931_numbering_plan numbering_plan;           /**< Numbering plan. */
    struct ipd_charset_string number;                   /**< Number. */
} PACE2_q931_information_element_called_party_number;

/**
 * @brief Data structure for the "Calling party number" Q.931 information element.
 */
typedef struct pace2_q931_information_element_calling_party_number {
    PACE2_q931_type_of_number type_of_number;                       /**< Type of number. */
    PACE2_q931_numbering_plan numbering_plan;                       /**< Numbering plan. */
    struct ipd_charset_string number;                               /**< Number. */

    u8 contains_extension;                                          /**< Informs in an extension is present. */
    struct {
        PACE2_q931_presentation_indicator presentation_indicator;   /**< Presentation indicator. */
        PACE2_q931_screening_indicator screening_indicator;         /**< Screening indicator. */
    } extension;

} PACE2_q931_information_element_calling_party_number;

/**
 * @brief Data structure for the "Progress indicator" Q.931 information element.
 */
typedef struct pace2_q931_information_element_progress_indicator {
    PACE2_q931_coding_standard coding_standard;             /**< Coding standard. */
    PACE2_q931_location location;                           /**< Location. */
    PACE2_q931_progress_description progress_description;   /**< Progress description. */
} PACE2_q931_information_element_progress_indicator;

/**
 * @brief Data structure for the "Cause" Q.931 information element.
 */
typedef struct pace2_q931_information_element_cause {
    PACE2_q931_coding_standard coding_standard; /**< Coding standard. */
    PACE2_q931_location location;               /**< Location. */
    PACE2_q931_cause_class cause_class;         /**< Cause class. */
    PACE2_q931_cause_value cause_value;         /**< Cause value. */
    u8 contains_diagnostic_information;         /**< Informs if the "diagnostic_information" element is present. */
    u8 diagnostic_information;                  /**< Diagnostic information. */
} PACE2_q931_information_element_cause;

/**
 * @brief Data structure for advanced Q.931 events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_ADVANCED_Q931_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_advanced_q931_event_type meta_data_type;                                                                                      /**< Q931 event subtype. */
    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;
    union {
        PACE2_q931_message_header message_header;                                                                                       /**< Contains data for a @ref PACE2_ADVANCED_Q931_MESSAGE_HEADER event. */
        u8 information_element_reserved;                                                                                                /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_RESERVED event. */
        PACE2_q931_repeat_indication information_element_repeat_indicator;                                                              /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_REPEAT_INDICATOR event. */
        PACE2_q931_congestion_level information_element_congestion_level;                                                               /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_CONGESTION_LEVEL event. */
        PACE2_q931_information_element_general information_element_general;                                                             /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_GENERAL event. */
        PACE2_q931_information_element_shift information_element_shift;                                                                 /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_SHIFT event. */
        PACE2_q931_information_element_user_user information_element_user_user;                                                         /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_USER_USER event. */
        struct ipd_charset_string information_element_display;                                                                          /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_DISPLAY event. */
        PACE2_q931_information_element_bearer_capability information_element_bearer_capability;                                         /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_BEARER_CAPABILITY event. */
        PACE2_q931_information_element_bearer_capability_optional_extension information_element_bearer_capability_optional_extension;   /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_BEARER_CAPABILITY_OPTIONAL_EXTENSION event. */
        PACE2_q931_information_element_called_party_number information_element_called_party_number;                                     /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_CALLED_PARTY_NUMBER event. */
        PACE2_q931_information_element_calling_party_number information_element_calling_party_number;                                   /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_CALLING_PARTY_NUMBER event. */
        PACE2_q931_information_element_progress_indicator information_element_progress_indicator;                                       /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_PROGRESS_INDICATOR event. */
        PACE2_q931_information_element_cause information_element_cause;                                                                 /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_CAUSE event. */
        PACE2_q931_information_element_calling_party_number information_element_connected_number;                                       /**< Contains data for a @ref PACE2_ADVANCED_Q931_INFORMATION_ELEMENT_CONNECTED_NUMBER event. */
        char const * invalid;                                                                                                           /**< Contains data for a @ref PACE2_ADVANCED_Q931_INVALID event. */
    } event_data;                                                                                                                       /**< Named union which contains the data of the sub-event types. */
} PACE2_advanced_q931_event;


#endif

#endif /* Q931_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
