/** @file      rtp_events.h
  * @brief     RTP event definitions.
  * @details   @copydoc doc_rtp
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef RTP_EVENTS_H
#define RTP_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the basic RTP event.
 */
#define PACE2_BASIC_RTP_EVENT_GROUP PACE2_POLLING_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned metadata group of the advanced RTP event.
 */
#define PACE2_ADVANCED_RTP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief PACE 2 type definition for a RTP flow statistic data structure.
 *
 * @deprecated This structure is deprecated and will be removed completely in the future. The information provided by
 *             this structure is moved to RTP dissector result 'dissector_result.rtp.statistics' in the
 *             'PACE2_dissector_metadata_event'.
 */
typedef struct ipoque_rtp_flow_stats_struct PACE2_RTP_flow_stats;

/**
 * @brief RTP performance report data flags.
 */
enum PACE2_RTP_data_flags {

    /** Flow statistics for a RTP flow. */
    PACE2_RTP_FLOW_STATS = 0,

    /** Number of implemented RTP data flags. */
    PACE2_NUMBER_OF_RTP_DATA_FLAGS
};

/**
 * @brief RTP metadata flags.
 */
typedef enum {

    /** Provides a struct ipd_dtmf_tone * with the digit, duration and volume of a event. */
    PACE2_ADVANCED_RTP_DTMF_SYMBOL = 0,

    /** Issued for each RTP packet in flows, which is not correlated to another flow.
     * Points to a struct ipd_rtp_data. */
    PACE2_ADVANCED_RTP_DATA,

    /** Issued for each SSRC in RTP packets in flows, which are not correlated to another flow.
     * Points to a struct ipd_rtp_ssrc. */
    PACE2_ADVANCED_RTP_SSRC,

    /** Issued when the AV session ID of a RTP flow changed. For example when the correlation with SIP occurred late.
     * Points to a struct PACE2_rtp_av_session_id_change. */
    PACE2_ADVANCED_RTP_AV_SESSION_ID_CHANGE,

    /** Signals the flow timeout of a RTP flow, which is not correlated to another flow. */
    PACE2_ADVANCED_RTP_END_OF_FLOW,

    /** Issued when decoding a packet in a RTP flow, which is not correlated to another flow, fails.
     * An error message is passed as a char pointer. */
    PACE2_ADVANCED_RTP_INVALID,

    /** Number of implemented RTP events. */
    PACE2_NUMBER_OF_ADVANCED_RTP_EVENTS

} PACE2_advanced_rtp_event_type;

/* Event Structures *************************************************************/

/**
 * @brief Event structure of a RTP performance report.
 *
 * @deprecated This event is deprecated and will be removed completely in the future. The information provided by this
 *             event is moved to RTP dissector result 'dissector_result.rtp.statistics' in the
 *             'PACE2_dissector_metadata_event'.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_BASIC_RTP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** Bitmask, which states the available RTP performance reports. */
    PACE2_bitmask meta_data_mask;

    /** Data structure of the @ref PACE2_RTP_FLOW_STATS flag.*/
    PACE2_RTP_flow_stats flow_stats;

} PACE2_RTP_event;

#ifndef PACE_KERNEL_LIB

struct pace2_rtp_header_extension_element {

    /** Element id. */
    u8 element_id;

    /** Element data. */
    struct ipd_charset_string element_data;
};

/** Structure wrapping one RTP packet.
 *
 * The meaning of all ID fields depends on the protocol/application that negotiated the addresses and
 * parameters of the RTP flow.
 *
 * See
 * - @ref doc_sip (for VoIP calls)
 * - @ref doc_yahoo_msgr
 * - @ref doc_oscar
 *
 * The #encoding_name, #clock_rate and #channels fields are derived from the payload type of in the packet,
 * using @c rtpmap: lines in the SDP session description used to negotiate this RTP flow. If the packet's
 * payload type was not defined in a session description, these fields are empty or zero.
 *
 * If SDP was used to announce this RTP flow, #media_type is taken from the @c m= line. Otherwise, this field
 * is an empty string.
 */
struct ipd_rtp_data {

    /** Session ID. */
    struct ipd_charset_string session_id;

    /** A/V session ID. */
    struct ipd_charset_string av_session_id;

    /** ID of the sender of the packet. */
    struct ipd_charset_string sender_id;

    /** ID of the receiver of the packet. */
    struct ipd_charset_string receiver_id;

    /** RTP data of the packet (does not include the RTP header). */
    struct ipd_charset_string data;

    /** Name of encoding. */
    struct ipd_charset_string encoding_name;

    /** Clock rate.  */
    u32 clock_rate;

    /** Number of channels. */
    u16 channels;

    /** The Sequence number which is sent. */
    u16 sequence_number;

    /** The Timestamp which is sent. */
    u32 timestamp;

    /** Number of RTP header extension elements. */
    u16 extension_elements_len;

    /** Array of RTP header extension elements. */
    struct pace2_rtp_header_extension_element * extension_elements;

    /** The RTP header of the RTP packet. */
    struct ipd_charset_string header;

    /** RTP payload type value. */
    u8 payload_type;

    /** Media type (eg, "audio" or "video"); empty string if not known. */
    struct ipd_charset_string media_type;
};

/**
 * Structure for the SSRC Event that provides the current SSRC and the current direction.
 */
struct ipd_rtp_ssrc {

    /** Session ID. */
    struct ipd_charset_string session_id;

    /** A/V session ID. */
    struct ipd_charset_string av_session_id;

    /** ID of the sender of the packet. */
    struct ipd_charset_string sender_id;

    /** ID of the receiver of the packet. */
    struct ipd_charset_string receiver_id;

    /** The SSRC which is sent. */
    u32 ssrc;

    /** The direction of the packet with the SSRC in it. */
    u8 direction;
};

/**
 * Structure for the DTMF event that provides information about the pushed digit.
 */
struct ipd_dtmf_tone {

    /** Pushed digit as ascii character, see ETSI ES 201 235-1. */
    u8 digit;

    /** Event duration of pushing the digit. */
    u16 duration;

    /** If not null, volume of the pushed digit. */
    u16 volume;
};

/**
 * @brief Data structure for RTP session ID changes. Used in @ref PACE2_ADVANCED_RTP_AV_SESSION_ID_CHANGE event.
 */
struct PACE2_rtp_av_session_id_change {

    /** The old A/V session ID. */
    struct ipd_charset_string old_av_session_id;

    /** The new A/V session ID. */
    struct ipd_charset_string new_av_session_id;
};

/**
 * @brief Data structure of advanced RTP events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_RTP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** RTP event subtype. */
    PACE2_advanced_rtp_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    union {

        /** Contains data for a @ref PACE2_ADVANCED_RTP_DTMF_SYMBOL event. */
        struct ipd_dtmf_tone dtmf_tone;

        /** Contains data for a @ref PACE2_ADVANCED_RTP_DATA event. */
        struct ipd_rtp_data rtp_data;

        /** Contains data for a @ref PACE2_ADVANCED_RTP_INVALID event. */
        char const * rtp_invalid;

        /** Contains data for a @ref PACE2_ADVANCED_RTP_SSRC event. */
        struct ipd_rtp_ssrc rtp_ssrc;

        /** Contains data for a @ref PACE2_ADVANCED_RTP_AV_SESSION_ID_CHANGE event. */
        struct PACE2_rtp_av_session_id_change av_session_id_change;

    } event_data;

} PACE2_advanced_rtp_event;

#endif

#endif /* RTP_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
