/** @file      rtsp_events.h
  * @brief     RTSP event definitions.
  * @details   @copydoc doc_rtsp
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef RTSP_EVENTS_H_
#define RTSP_EVENTS_H_

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "rtp_events.h"

/* Event Group ******************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced RTSP event.
 */
#define PACE2_ADVANCED_RTSP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/
/**
 * @brief Types for advanced RTSP events.
 */
typedef enum {

    /**
     * Signals that a new connection was noticed.<br />
     * Provides: NULL.
     */
    PACE2_ADVANCED_RTSP_NEW_CONNECTION,

    /**
     * Issued for each decoded RTSP message.<br />
     * Provides: struct pace2_rtsp_message * : RTSP message.
     */
    PACE2_ADVANCED_RTSP_MESSAGE,

    /**
     * Signals connection termination (TCP transport) or flow timeout (UDP, TCP).<br />
     * Provides: NULL.
     */
    PACE2_ADVANCED_RTSP_END_OF_FLOW,

    /**
     * Issued when decoding a message fails.<br />
     * Provides: const char * : message describing the failure reason.
     */
    PACE2_ADVANCED_RTSP_INVALID,

    /**
     * Issued for each RTP packet in flows correlated to this flow.<br />
     * Provides: struct ipd_rtp_data * : RTP data.
     */
    PACE2_ADVANCED_RTSP_RTP_DATA,

    /**
     * Issued for each RTP SSRC in flows correlated to this flow.<br />
     * Provides: struct ipd_rtp_ssrc * : RTP SSRC.
     */
    PACE2_ADVANCED_RTSP_RTP_SSRC,

    /**
     * Signals flow timeout of a correlated RTP flow.<br />
     * Provides: NULL.
     */
    PACE2_ADVANCED_RTSP_RTP_END_OF_FLOW,

    /**
     * Issued when decoding a packet in a correlated RTP flow fails.<br />
     * Provides: const char * : message describing the failure reason.
     */
    PACE2_ADVANCED_RTSP_RTP_INVALID,

    /**
     * A dummy event to signal the end of the event list.
     */
    PACE2_NUMBER_OF_ADVANCED_RTSP_EVENTS,

} PACE2_advanced_rtsp_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * @brief RTSP message type.
 */
typedef enum {
    PACE2_RTSP_REQUEST,                         /**< RTSP request. */
    PACE2_RTSP_RESPONSE,                        /**< RTSP response. */
} pace2_rtsp_message_type;

/**
 * @brief RTSP request-line.
 */
struct pace2_rtsp_request_line {
    struct ipd_charset_string method;           /**< Request method (eg, DESCRIBE, SETUP, PLAY). */
    struct ipd_charset_string request_uri;      /**< Request-URI (eg, "rtsp://foo/twister"). */
    struct ipd_charset_string rtsp_version;     /**< RTSP version ("RTSP/1.0"). */
};

/**
 * @brief RTSP status-line.
 */
struct pace2_rtsp_status_line {
    struct ipd_charset_string rtsp_version;     /**< RTSP version ("RTSP/1.0"). */
    struct ipd_charset_string status_code;      /**< Status code (eg, "200"). */
    struct ipd_charset_string reason_phrase;    /**< Reason phrase (eg., "OK"). */
};

/**
 * @brief RTSP message.
 */
struct pace2_rtsp_message {
    union ip_v4_v6_addr source_address;               /**< Sender IP address. */
    u8 source_address_is_ipv6;                        /**< Flag that is set if @ref pace2_rtsp_message::source_address contains an IPv6 address. */

    pace2_rtsp_message_type type;                     /**< Message type (request or response). */

    union {
        struct pace2_rtsp_request_line request_line;  /**< Entire request line. */
        struct pace2_rtsp_status_line status_line;    /**< Entire status line. */
    };

    u64 cseq;                                         /**< CSeq: header field (if set). */
    struct ipd_charset_string session;                /**< Session: header field (if set). */
    struct ipd_charset_string user_agent;             /**< User-Agent: header field (if set). */
    struct ipd_charset_string user_network;           /**< User-Network: header field (if set). */
    struct ipd_charset_string server;                 /**< Server: header field (if set). */
    struct ipd_charset_string content_type;           /**< Content-Type: header field (if set). */
    struct ipd_charset_string content_length;         /**< Content-Length: header field (if set). */
    struct ipd_charset_string content_base;           /**< Content-Base: header field (if set). */
    struct ipd_charset_string content_location;       /**< Content-Location: header field (if set). */
    struct ipd_charset_string transport;              /**< Transport: header field (if set). */
};

/**
 * @brief Data structure for advanced RTSP events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_ADVANCED_RTSP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_advanced_rtsp_event_type meta_data_type;                                                      /**< RTSP event subtype. */
    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;
    union {
        struct pace2_rtsp_message message;                                                              /**< Contains data for a @ref PACE2_ADVANCED_RTSP_MESSAGE event. */
        struct ipd_rtp_data rtp_data;                                                                   /**< Contains data for a @ref PACE2_ADVANCED_RTSP_RTP_DATA event. */
        struct ipd_rtp_ssrc rtp_ssrc;                                                                   /**< Contains data for a @ref PACE2_ADVANCED_RTSP_RTP_SSRC event. */
        const char *rtp_invalid;                                                                        /**< Contains data for a @ref PACE2_ADVANCED_RTSP_RTP_INVALID event. */
        const char *invalid;                                                                            /**< Contains data for a @ref PACE2_ADVANCED_RTSP_INVALID event. */
    } event_data;                                                                                       /**< Named union which contains the data of the sub-event types. */
} PACE2_advanced_rtsp_event;

#endif /* PACE_KERNEL_LIB */

#endif /* RTSP_EVENTS_H_ */
