/** @file      scada_events.h
  * @brief     SCADA event definitions.
  * @details   @copydoc doc_scada
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef SCADA_EVENTS_H
#define SCADA_EVENTS_H
/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced scada event.
 */
#define PACE2_ADVANCED_SCADA_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced scada events.
 */
typedef enum {

    /** Provides a structure containing information about the fields of IEC-60870-104 APCI. */
    PACE2_ADVANCED_SCADA_APCI,

    /** Provides a structure containing information about the fields of IEC-60870-101 ASDU. */
    PACE2_ADVANCED_SCADA_ASDU,

    /** Provides information about error in data stream. */
    PACE2_ADVANCED_SCADA_INVALID,

    /** Provides a structure containing information about the information object contained
     * in an ASDU. */
    PACE2_ADVANCED_SCADA_INFORMATION_OBJECT,

    /** Provides a structure containing information about an information element contained
     * in an information object. */
    PACE2_ADVANCED_SCADA_INFORMATION_ELEMENT,

    /** Provides information about a new SCADA TCP connection. */
    PACE2_ADVANCED_SCADA_NEW_CONNECTION,

    /** Provides information about SCADA TCP Connection termination. */
    PACE2_ADVANCED_SCADA_CONNECTION_TERMINATION,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_SCADA_EVENT_TYPES

} PACE2_advanced_Scada_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/** Enum of SCADA apci formats. */
enum ipd_scada_apci_formats {

    /** I type APCI. Contains Sender and receiver sequence number. */
    IPD_SCADA_APCI_I_FORMAT = 0,

    /** S type APCI.Contains only receivers sequence number. */
    IPD_SCADA_APCI_S_FORMAT = 1,

    /** U type APCI.Contains options STARTDT,STOPDT and TESTFR. */
    IPD_SCADA_APCI_U_FORMAT = 3,

    /** Unknown APCI format. */
    IPD_SCADA_APCI_UNKNOWN_FORMAT = 4

};
/** Enum of SCADA asdu cause of transmission P/N field values. */
enum ipd_scada_asdu_cot_confirmation {

    /** Cause of transmission is positive/Negative (P/N) flag is 0. */
    IPD_SCADA_ASDU_COT_POSITIVE = 0,

    /** Cause of transmission positive/Negative (P/N) flag is 1. */
    IPD_SCADA_ASDU_COT_NEGATIVE = 1,
};

/** Enum of SCADA asdu cause of transmission test field values. */
enum ipd_scada_asdu_cot_test {

    /** Cause of transmission test bit is 0. */
    IPD_SCADA_ASDU_COT_NO_TEST = 0,

    /** Cause of transmission test bit is 1. */
    IPD_SCADA_ASDU_COT_TEST = 1,
};

/** Enum of SCADA True and false values. */
enum ipd_scada_true_false {

    /** Flag value is false. */
    IPD_SCADA_FALSE = 0,

    /** Flag value is True. */
    IPD_SCADA_TRUE = 1,
};

/** Enum of SCADA Information element QCC freeze values. */
enum ipd_scada_counter_intrg_freeze_qcc {

    /** Counter has been read (no freeze or reset). */
    IPD_SCADA_QCC_READ = 0,

    /** counter freeze without reset (value frozen represents integrated total). */
    IPD_SCADA_QCC_FREEZE_NO_RESET = 1,

    /** Counter freeze with reset (value frozen represents incremental information). */
    IPD_SCADA_QCC_FREEZE_RESET = 2,

    /** Counter has been reset. */
    IPD_SCADA_QCC_RESET = 3,
};

/** Enum of SCADA Information element COI cause values. */
enum ipd_scada_cause_init {

    /** Local power switch on. */
    IPD_SCADA_LOCAL_PWR_ON = 0,

    /** Local manual reset. */
    IPD_SCADA_LOCAL_MAN_RESET = 1,

    /** Remote reset. */
    IPD_SCADA_REMOTE_RESET = 2,

    /** Reserved values between 3-127. */
    IPD_SCADA_RESERVED3 = 3,
    IPD_SCADA_RESERVED127 = 127
};

/** Enum of SCADA Information element RCO command state values. */
enum ipd_scada_io_regulating {

    /** Not permitted. */
    IPD_SCADA_IO_REG_NOT_PERMITTED0 = 0,

    /** Next step LOW. */
    IPD_SCADA_IO_REG_STEP_LOW = 1,

    /** Next step HIGH. */
    IPD_SCADA_IO_REG_STEP_HIGH = 2,

    /** Not permitted. */
    IPD_SCADA_IO_REG_NOT_PERMITTED3 = 3,
};

/** Enum of SCADA Information element DCO command state values. */
enum ipd_scada_io_dc {

    /** Not permitted. */
    IPD_SCADA_IO_DC_NOT_PERMITTED0 = 0,

    /** OFF value. */
    IPD_SCADA_IO_DC_OFF = 1,

    /** ON value. */
    IPD_SCADA_IO_DC_ON = 2,

    /** Not permitted. */
    IPD_SCADA_IO_DC_NOT_PERMITTED3 = 3,
};

/** Enum of SCADA apci U formats fields. */
enum ipd_scada_apci_u_type_flags {

    /** TESTFR confirmation request or its acknowledgment. */
    IPD_SCADA_APCI_U_TYPE_TESTFR = 1,

    /** STOPDT confirmation request or its acknowledgment. */
    IPD_SCADA_APCI_U_TYPE_STOPDT = 2,

    /** STARTDT confirmation request or its acknowledgment. */
    IPD_SCADA_APCI_U_TYPE_STARTDT = 3,

    /** Unknown or error. */
    IPD_SCADA_APCI_U_TYPE_UNKNOWN = 4
};

/** Enum of SCADA Information element types. */
enum ipd_scada_asdu_info_elements_type {

    /** Quality descriptor for events of protection equipment. QDP. */
    IPD_SCADA_INFO_ELEM_QDP = 1,

    /** Double-point information. DIQ. */
    IPD_SCADA_INFO_ELEM_DIQ = 2,

    /** Quality descriptor. QDS. */
    IPD_SCADA_INFO_ELEM_QDS = 3,

    /** Value with transient state indication. VTI. */
    IPD_SCADA_INFO_ELEM_VTI = 4,

    /** Normalized value. NVA.  */
    IPD_SCADA_INFO_ELEM_NVA = 5,

    /** Scaled value. SVA. */
    IPD_SCADA_INFO_ELEM_SVA = 6,

    /** Short floating point number. R32-IEEE STD 754. */
    IPD_SCADA_INFO_ELEM_R32_754 = 7,

    /** Binary counter reading. BCR. */
    IPD_SCADA_INFO_ELEM_BCR = 8,

    /** Single event of protection equipment. SEP. */
    IPD_SCADA_INFO_ELEM_SEP = 9,

    /** Start events of protection equipment. SPE. */
    IPD_SCADA_INFO_ELEM_SPE = 10,

    /** Output circuit information of protection equipment. OCI. */
    IPD_SCADA_INFO_ELEM_OCI = 11,

    /** Binary state information 32 bit. BSI. */
    IPD_SCADA_INFO_ELEM_BSI = 12,

    /** Single command. SCO*/
    IPD_SCADA_INFO_ELEM_SCO = 13,

    /** Single-point information. SIQ. */
    IPD_SCADA_INFO_ELEM_SIQ = 14,

    /** Double command. DCO*/
    IPD_SCADA_INFO_ELEM_DCO = 15,

    /** Regulating step command. RCO. */
    IPD_SCADA_INFO_ELEM_RCO = 16,

    /** Seven octet binary time. CP56Time2a. */
    IPD_SCADA_INFO_ELEM_CP56TM = 17,

    /** Two octet binary time. CP16Time2a. */
    IPD_SCADA_INFO_ELEM_CP16TM = 18,

    /** Cause of initialization. COI. */
    IPD_SCADA_INFO_ELEM_COI = 19,

    /** Qualifier of interrogation. QOI. */
    IPD_SCADA_INFO_ELEM_QOI = 20,

    /** Qualifier of counter interrogation command. QCC. */
    IPD_SCADA_INFO_ELEM_QCC = 21,

    /** Qualifier of parameter of measured values. QPM. */
    IPD_SCADA_INFO_ELEM_QPM = 22,

    /** Qualifier of parameter activation. QPA. */
    IPD_SCADA_INFO_ELEM_QPA = 23,

    /** Qualifier of reset process command. QRP. */
    IPD_SCADA_INFO_ELEM_QRP = 24,

    /** File ready qualifier. FRQ. */
    IPD_SCADA_INFO_ELEM_FRQ = 25,

    /** Section ready qualifier. SRQ. */
    IPD_SCADA_INFO_ELEM_SRQ = 26,

    /** Select and call qualifier. SCQ. */
    IPD_SCADA_INFO_ELEM_SCQ = 27,

    /** Last section or segment qualifier. LSQ. */
    IPD_SCADA_INFO_ELEM_LSQ = 28,

    /** Acknowledge file or section qualifier. AFQ. */
    IPD_SCADA_INFO_ELEM_AFQ = 29,

    /** Name of file. NOF. */
    IPD_SCADA_INFO_ELEM_NOF = 30,

    /** Name of section. NOS. */
    IPD_SCADA_INFO_ELEM_NOS = 31,

    /** Length of file or section. LOF. */
    IPD_SCADA_INFO_ELEM_LOF = 32,

    /** Length of segment. LOS. */
    IPD_SCADA_INFO_ELEM_LOS = 33,

    /** Checksum. CHS. */
    IPD_SCADA_INFO_ELEM_CHS = 34,

    /** Status of file. SOF. */
    IPD_SCADA_INFO_ELEM_SOF = 35,

    /** Qualifier of set-point command. QOS. */
    IPD_SCADA_INFO_ELEM_QOS = 36,

    /** Status and status change detection. SCD. */
    IPD_SCADA_INFO_ELEM_SCD = 37,

    /** 16 bit test sequence counter. TSC. */
    IPD_SCADA_INFO_ELEM_TSC = 38,

    /** File segment block. FSEG. */
    IPD_SCADA_INFO_ELEM_FSEG = 39,

    /** Unknown type*/
    IPD_SCADA_INFO_ELEM_UNKNOWN = 255
};

/** Enum of supported SCADA ASDU types. */
enum ipd_scada_asdu_types {

    /** Single-point information. */
    IPD_SCADA_ASDU_M_SP_NA_1 = 1,

    /** Double-point information. */
    IPD_SCADA_ASDU_M_DP_NA_1 = 3,

    /** Step position information. */
    IPD_SCADA_ASDU_M_ST_NA_1 = 5,

    /** Bitstring of 32 bits. */
    IPD_SCADA_ASDU_M_BO_NA_1 = 7,

    /** Measured value, normalized value. */
    IPD_SCADA_ASDU_M_ME_NA_1 = 9,

    /** Measured value, scaled value. */
    IPD_SCADA_ASDU_M_ME_NB_1 = 11,

    /** Measured value, short floating point number. */
    IPD_SCADA_ASDU_M_ME_NC_1 = 13,

    /** Integrated totals. */
    IPD_SCADA_ASDU_M_IT_NA_1 = 15,

    /** Packed single-point information with status change detection. */
    IPD_SCADA_ASDU_M_PS_NA_1 = 20,

    /** Measured value, normalized value without quality descriptor. */
    IPD_SCADA_ASDU_M_ME_ND_1 = 21,

    /** Single-point information with time tag CP56Time2a. */
    IPD_SCADA_ASDU_M_SP_TB_1 = 30,

    /** Double-point information with time tag CP56Time2a. */
    IPD_SCADA_ASDU_M_DP_TB_1 = 31,

    /** Step position information with time tag CP56Time2a. */
    IPD_SCADA_ASDU_M_ST_TB_1 = 32,

    /** Bitstring of 32 bit with time tag CP56Time2a. */
    IPD_SCADA_ASDU_M_BO_TB_1 = 33,

    /** Measured value, normalized value with time tag CP56Time2a. */
    IPD_SCADA_ASDU_M_ME_TD_1 = 34,

    /** Measured value, scaled value with time tag CP56Time2a. */
    IPD_SCADA_ASDU_M_ME_TE_1 = 35,

    /** Measured value, short floating point number with time tag CP56Time2a. */
    IPD_SCADA_ASDU_M_ME_TF_1 = 36,

    /** Integrated totals with time tag CP56Time2a. */
    IPD_SCADA_ASDU_M_IT_TB_1 = 37,

    /** Event of protection equipment with time tag CP56Time2a. */
    IPD_SCADA_ASDU_M_EP_TD_1 = 38,

    /** Packed start events of protection equipment with time tag CP56Time2a. */
    IPD_SCADA_ASDU_M_EP_TE_1 = 39,

    /** Packed output circuit information of protection equipment with time tag CP56Time2a. */
    IPD_SCADA_ASDU_M_EP_TF_1 = 40,

    /** Single command. */
    IPD_SCADA_ASDU_C_SC_NA_1 = 45,

    /** Double command. */
    IPD_SCADA_ASDU_C_DC_NA_1 = 46,

    /** Regulating step command. */
    IPD_SCADA_ASDU_C_RC_NA_1 = 47,

    /** Set point command, normalized value. */
    IPD_SCADA_ASDU_C_SE_NA_1 = 48,

    /** Set point command, scaled value. */
    IPD_SCADA_ASDU_C_SE_NB_1 = 49,

    /** Set point command, short floating point number. */
    IPD_SCADA_ASDU_C_SE_NC_1 = 50,

    /** Bitstring of 32 bits. */
    IPD_SCADA_ASDU_C_BO_NA_1 = 51,

    /** Single command with time tag CP56Time2a. */
    IPD_SCADA_ASDU_C_SC_TA_1 = 58,

    /** Double command with time tag CP56Time2a. */
    IPD_SCADA_ASDU_C_DC_TA_1 = 59,

    /** Regulating step command with time tag CP56Time2a. */
    IPD_SCADA_ASDU_C_RC_TA_1 = 60,

    /** Set point command, normalized value with time tag CP56Time2a. */
    IPD_SCADA_ASDU_C_SE_TA_1 = 61,

    /** Set point command, scaled value with time tag CP56Time2a. */
    IPD_SCADA_ASDU_C_SE_TB_1 = 62,

    /** Set point command, short floating-point number with time tag CP56Time2a. */
    IPD_SCADA_ASDU_C_SE_TC_1 = 63,

    /** Bitstring of 32 bits with time tag CP56Time2a. */
    IPD_SCADA_ASDU_C_BO_TA_1 = 64,

    /** End of initialization. */
    IPD_SCADA_ASDU_M_EI_NA_1 = 70,

    /** Interrogation command. */
    IPD_SCADA_ASDU_C_IC_NA_1 = 100,

    /** Counter interrogation command. */
    IPD_SCADA_ASDU_C_CI_NA_1 = 101,

    /** Read command. */
    IPD_SCADA_ASDU_C_RD_NA_1 = 102,

    /** Clock synchronization command. */
    IPD_SCADA_ASDU_C_CS_NA_1 = 103,

    /** Test command. */
    IPD_SCADA_ASDU_C_TS_NA_1 = 104,

    /** Reset process command. */
    IPD_SCADA_ASDU_C_RP_NA_1 = 105,

    /** Test command with time tag CP56Time2a. */
    IPD_SCADA_ASDU_C_TS_TA_1 = 107,

    /** Parameter of measured value, normalized value. */
    IPD_SCADA_ASDU_P_ME_NA_1 = 110,

    /** Parameter of measured value, scaled value. */
    IPD_SCADA_ASDU_P_ME_NB_1 = 111,

    /** Parameter of measured value, short floating-point number. */
    IPD_SCADA_ASDU_P_ME_NC_1 = 112,

    /** Parameter activation. */
    IPD_SCADA_ASDU_P_AC_NA_1 = 113,

    /** File ready. */
    IPD_SCADA_ASDU_F_FR_NA_1 = 120,

    /** Section ready. */
    IPD_SCADA_ASDU_F_SR_NA_1 = 121,

    /** Call directory, select file, call file, call section. */
    IPD_SCADA_ASDU_F_SC_NA_1 = 122,

    /** Last section, last segment. */
    IPD_SCADA_ASDU_F_LS_NA_1 = 123,

    /** Ack file, ack section. */
    IPD_SCADA_ASDU_F_AF_NA_1 = 124,

    /** Segment. */
    IPD_SCADA_ASDU_F_SG_NA_1 = 125,

    /** Directory. */
    IPD_SCADA_ASDU_F_DR_TA_1 = 126,

    /** Query Log - Request archive file. */
    IPD_SCADA_ASDU_F_SC_NB_1 = 127,

    /** Unknown type of ASDU. */
    IPD_SCADA_ASDU_UNKNOWN = 256
};

/** Contains the Information about SCADA apci block. */
struct ipd_scada_apci {

    /** Message direction, 0 or 1. can be either from server or client.Direction is provided to
     * correlate request and response events in case of out of order event delivery. */
    u8 message_direction;

    /** Length of APDU (APCI + ASDU). */
    u16 length;

    /** Unique ID which is same for APCI, ASDU and Information objects.
     * This can be used to correlate different events belonging to same scada message. */
    u64 scada_message_id;

    /** Type of APIC, Can be I,S or U type defined by enum ipd_scada_apci_formats. */
    enum ipd_scada_apci_formats apci_type;

    /** Union containing structures of all three types of apci's. */
    union {

        /** I type apci information. */
        struct {

            /** Send sequence number. */
            u16 send_sequence_number;

            /** Receive sequence number. */
            u16 recv_sequence_number;
        } type_i;

        /** S type apci information. */
        struct {
            /** Receive sequence number. */
            u16 recv_sequence_number;
        } type_s;

        /** U type apci information. */
        struct {

            /** U type ACPI function. */
            enum ipd_scada_apci_u_type_flags function;

            /** Flag confirm is true. */
            u8 flag_con;

            /** Flag confirm ack is true. */
            u8 flag_act;
        } type_u;

    } control_information;
};

/** Contains the Information about SCADA asdu block. */
struct ipd_scada_asdu {

    /** Message direction, 0 or 1. can be either from server or client.Direction is
     * provided to correlate request and response events in case of out of order event delivery. */
    u8 message_direction;

    /** Type identifier of SCADA ASDU. This can be a value from enum ipd_scada_asdu_types. */
    enum ipd_scada_asdu_types type_identifier;

    /** Number of information objects in the SCADA ASDU. */
    u8 information_object_count;

    /** The cause of transmission of a specific SCADA ASDU. */
    u8 transmission_cause;

    /** The cause of transmission positive or negative confirmation of a SCADA ASDU. */
    enum ipd_scada_asdu_cot_confirmation cot_confirmation;

    /** The cause of transmission test bit in a specific SCADA ASDU. */
    enum ipd_scada_asdu_cot_test cot_test;

    /** The flag indicating presence of originator address. */
    enum ipd_scada_true_false origin_addr_present;

    /** The address of the requesting entity in both request(control) and response
     * (monitor) directions. This can also be considered the server address. */
    u8 orignator_address;

    /** The address of the responding entity in response (monitor) directions.This address represent
     * the complete ASDU. This either 0xFF or 0xFFFF when sent by a controlling station in control direction
     * when broadcast is intended. */
    u16 asdu_address;

    /** Unique ID which is same for APCI, ASDU and Information objects.
     * This can be used to correlate different events belonging to same SCADA message. */
    u64 scada_message_id;
};

/** Contains the Information sent when a connection ends. */
struct ipd_scada_connection_end {

    /** Total SCADA messages received */
    u64 total_messages;
};

/** Contains the Information sent when a an information object is received. */
struct ipd_scada_info_object {

    /** Message direction, 0 or 1. can be either from server or client.
     * Direction is provided to co-relate request and response events. */
    u8 message_direction;

    /** Unique ID which is same for APCI, ASDU and Information objects.
      * This can be used to correlate different information element events with their parent
      * information object message. */
    u64 scada_message_id;

    /** Unique ID which is same for Information objects and information elements.
      * This can be used to correlate different events belonging to same SCADA message. */
     u64 info_obj_id;

    /** Information object address which is present with every information object in an ASDU. */
    u32 info_obj_addr;

    /** Count of information elements present in one information object. */
    u8 info_element_count;

    /** Address to container ASDU. */
    u16 asdu_address;
};

/** Structure to represent simple command command qualifier. */
struct ipd_scada_command_qualifier {

    /** Can have values from 0-31. */
     u8 qualifier_type;

     /** If 1 then is select, 0 stands from execute. */
     enum ipd_scada_true_false is_select;
};

/** Structure to represent single point information. */
struct ipd_scada_single_point_info_siq {

    /** SPI is ON or OFF */
    enum ipd_scada_true_false spi_state;

    /** True if blocked. */
    enum ipd_scada_true_false blocked;

    /** True if substituted. */
    enum ipd_scada_true_false substituted;

    /** True if not topical. */
    enum ipd_scada_true_false not_topical;

    /** True if invalid. */
    enum ipd_scada_true_false is_invalid;
};

/** Structure to represent double point information. */
struct ipd_scada_double_point_info_diq {

    /** Can have values from enum ipd_scada_io_off_on. */
    enum ipd_scada_true_false dpi_state;

    /** True if blocked. */
    enum ipd_scada_true_false blocked;

    /** True if substituted. */
    enum ipd_scada_true_false substituted;

    /** True if not topical. */
    enum ipd_scada_true_false not_topical;

    /** True if invalid. */
    enum ipd_scada_true_false is_invalid;
};

/** Structure to represent quality descriptor. */
struct ipd_scada_quality_descriptor_qds {

    /** True if overflow. */
    enum ipd_scada_true_false overflow;

    /** True if blocked. */
    enum ipd_scada_true_false blocked;

    /** True if substituted. */
    enum ipd_scada_true_false substituted;

    /** True if not topical. */
    enum ipd_scada_true_false not_topical;

    /** True if invalid. */
    enum ipd_scada_true_false is_invalid;
};

/** Structure to represent protection equipment quality descriptor. */
struct ipd_scada_protection_equip_q_des_qdp {

    /** True if elapsed time is invalid. */
    enum ipd_scada_true_false invalid_elapsed_time;

    /** True if blocked. */
    enum ipd_scada_true_false blocked;

    /** True if substituted. */
    enum ipd_scada_true_false substituted;

    /** True if not topical. */
    enum ipd_scada_true_false not_topical;

    /** True if invalid. */
    enum ipd_scada_true_false is_invalid;
};

/** Structure to represent transient value state indicator. */
struct ipd_scada_value_transient_state_indicator_vti {

    /** Unresolved value is provided.Actual value should be between -64  to +63. */
    short value;

    /** True if equipment is in transient state. */
    enum ipd_scada_true_false is_transient;
};

/** Structure to represent short binary counter reading. */
struct ipd_scada_short_bin_cntr_reading_bcr {

    /** Counter value. */
    int counter_value;

    /** Sequence number between 0 and 31. */
    u8 sequence_nr;

    /** True if counter overflow occurred. */
    enum ipd_scada_true_false counter_carry_flag;

    /** True if counter has been adjusted. */
    enum ipd_scada_true_false counter_adjusted;

    /** True if reading is invalid. */
    enum ipd_scada_true_false is_invalid;
};

/** Structure to represent protection equipment single event. */
struct ipd_scada_protection_equip_single_event_sep {

    /** Event state can have values from enum ipd_scada_io_off_on. */
    u8 event_state;

    /** True if elapsed time is invalid. */
    enum ipd_scada_true_false invalid_elapsed_time;

    /** True if blocked. */
    enum ipd_scada_true_false blocked;

    /** True if substituted. */
    enum ipd_scada_true_false substituted;

    /** True if not topical. */
    enum ipd_scada_true_false not_topical;

    /** True if invalid. */
    enum ipd_scada_true_false is_invalid;
};

/** Structure to represent protection equipment start of events. */
struct ipd_scada_protection_equip_start_events_spe {

    /**  True if general start of operations. */
    enum ipd_scada_true_false general_start;

    /**  True if start of L1 operation. */
    enum ipd_scada_true_false phase_l1;

    /**  True if start of L2 operation. */
    enum ipd_scada_true_false phase_l2;

    /**  True if start of L3 operation. */
    enum ipd_scada_true_false phase_l3;

    /**  True if start of IE operation. */
    enum ipd_scada_true_false start_ie;

    /**  True if start of reverse direction operation. */
    enum ipd_scada_true_false start_reverse_dir;
};

/** Structure to represent protection equipment output circuit information. */
struct ipd_scada_protection_equip_oc_info_oci {

    /**  True if general command to output circuit. */
    enum ipd_scada_true_false general_command;

    /**  True if command to output circuit L1. */
    enum ipd_scada_true_false command_phase_l1;

    /**  True if command to output circuit L2. */
    enum ipd_scada_true_false command_phase_l2;

    /**  True if command to output circuit L3. */
    u8 command_phase_l3;
};

/** Structure to represent Single command information. */
struct ipd_scada_single_command_sco {

    /** True if single command state is ON. */
    enum ipd_scada_true_false  command_state_is_on;

    /** Command qualifier. */
    struct ipd_scada_command_qualifier qoc;
};

/** Structure to represent double command information. */
struct ipd_scada_double_command_dco {

    /** Can be a value from enum ipd_scada_io_dc. */
    enum ipd_scada_io_dc command_state;

    /** Command qualifier. */
    struct ipd_scada_command_qualifier qoc;
};

/** Structure to represent regulating command information. */
struct ipd_scada_regulating_command_rco {

    /** Can be a value from enum ipd_scada_io_regulating. */
    enum ipd_scada_io_regulating command_state;

    /** Command qualifier. */
    struct ipd_scada_command_qualifier qoc;
};

/** Structure to represent 56 bit time information. */
struct ipd_scada_56_bit_time {

    u16 milliseconds;
    u16 seconds;
    u8 minutes;
    u8 hour;

    /** Date, day.can be from 1-31. */
    u8 month_day;

    /** Date, month.1(January) - 12(December). */
    u8 month;

    /** Date, year. the value of year is always between 1970-2070. */
    u16 year;

    /** True if time is substituted. otherwise time is genuine. */
    enum ipd_scada_true_false is_substituted;
};

/** Structure to represent 16 bit time information. */
struct ipd_scada_16_bit_time {

    /** Elapsed time in milli-seconds. */
    u16 elapsed_ms;
};

/** Structure to represent cause of initialization information. */
struct ipd_scada_initialization_cause_coi {

    /** Cause can be either enum ipd_scada_cause_init or one of reserved values . */
    enum ipd_scada_cause_init cause;

    /** True if initialization performed after change in local parameters. */
    enum ipd_scada_true_false with_changed_local_params;
};

/** Structure to interrogation counter qualifier. */
struct ipd_scada_counter_intrg_qualifier_qcc {

    /** Request type. */
    u8 request;

    /** Value specified by ipd_scada_counter_intrg_qual_freeze. */
    enum ipd_scada_counter_intrg_freeze_qcc freeze;
};

/** Structure to represent parameter value qualifier. */
struct ipd_scada_val_param_qualifier_qpm {

    /** Kind of parameter. */
    u8 parameter_kind;

    /** True if local parameter has changed. */
    enum ipd_scada_true_false local_param_change;

    /** True if parameter is not operational. */
    enum ipd_scada_true_false param_not_operational;
};

/** Structure to represent file ready qualifier. */
struct ipd_scada_file_ready_qualifier_frq {

    /** Value of file ready qualifier. */
    u8 value;

    /** True if confirmation is negative. */
    enum ipd_scada_true_false is_confirmation_negative;
};

/** Structure to represent file section ready qualifier. */
struct ipd_scada_section_ready_qualifier_srq {

    /** Value of section ready qualifier. */
    u8 value;

    /** True if section is not ready to load. */
    enum ipd_scada_true_false section_not_ready;
};

/** Structure to represent file status information. */
struct ipd_scada_file_status_sof {

    /** value of file status which has been returned. */
    u8 status;

    /** True if file is the last file in the directory. */
    enum ipd_scada_true_false is_lfd;

    /** True if the entry is a sub-directory . */
    enum ipd_scada_true_false is_subdir;

    /** True if the file transfer is active. */
    enum ipd_scada_true_false transfer_active;
};

/** Structure to represent set point command qualifier. */
struct ipd_scada_set_point_cmd_qualifier_qos {

    /** Value of set point qualifier. */
    u8 qulaifier;

    /** True if select otherwise execute. */
    enum ipd_scada_true_false is_select;
};

/** Structure to represent status and change detection information. */
struct ipd_scada_status_change_detection_scd {

    /** 16 bit bit-mask where each bit represent ON(1)/OFF(0) value. */
    u16 status_bmsk;

    /** 16 bit bit-mask where each bit represent (change detected(1)/no change detected(0)) value. */
    u16 change_detected;
};

/** Structure to represent an information element. */
struct ipd_scada_info_element {

    /** Message direction, 0 or 1. can be either from server or client.
     * Direction is provided to correlate request and response events. */
    u8 message_direction;

    /** Unique ID which is same for Information objects and information elements.
      * This can be used to correlate different information element events with their parent
      * information object message. */
     u64 info_obj_id;

    /** Unique ID which is same for APCI, ASDU and Information objects.
     * This can be used to correlate different events belonging to same SCADA message. */
    u64 scada_message_id;

    /** Information object address which is present with every information object in an ASDU. */
    u32 info_obj_addr;

    /** Unique ID to represent each information element event separately. */
    u64 info_elem_id;

    /** Selector of type for data union.can have one of the values from enum ipd_scada_asdu_info_elements_type. */
    enum ipd_scada_asdu_info_elements_type info_elem_type;

    union {

        /** @see ipd_scada_single_point_info_siq. */
        struct ipd_scada_single_point_info_siq siq;

        /** @see ipd_scada_double_point_info_diq. */
        struct ipd_scada_double_point_info_diq diq;

        /** @see ipd_scada_quality_descriptor_qds. */
        struct ipd_scada_quality_descriptor_qds qds;

        /** @see ipd_scada_protection_equip_q_des_qdp. */
        struct ipd_scada_protection_equip_q_des_qdp qdp;

        /** @see ipd_scada_value_transient_state_indicator_vti. */
        struct ipd_scada_value_transient_state_indicator_vti vti;

        /** Unresolved normalized value contained in the 16 bit field. */
        u16 nva_normalized_value;

        /** Unresolved scaled value contained in the 16 bit field. */
        u16 sva_scaled_value;

        /** IEEE 754 floating point number 4 bytes. */
        float ieee_754_fp;

        /** @see ipd_scada_short_bin_cntr_reading_bcr. */
        struct ipd_scada_short_bin_cntr_reading_bcr bcr;

        /** @see ipd_scada_protection_equip_single_event_sep. */
        struct ipd_scada_protection_equip_single_event_sep sep;

        /** @see ipd_scada_protection_equip_start_events_spe. */
        struct ipd_scada_protection_equip_start_events_spe spe;

        /** @see ipd_scada_protection_equip_oc_info_oci. */
        struct ipd_scada_protection_equip_oc_info_oci oci;

        /** 32 bit bitmap. */
        u32 bsi;

        /** @see ipd_scada_single_command_sco. */
        struct ipd_scada_single_command_sco sco;

        /** @see ipd_scada_double_command_dco. */
        struct ipd_scada_double_command_dco dco;

        /** @see ipd_scada_regulating_command_rco. */
        struct ipd_scada_regulating_command_rco rco;

        /** @see ipd_scada_56_bit_time. */
        struct ipd_scada_56_bit_time cp56_time;

        /** @see ipd_scada_16_bit_time. */
        struct ipd_scada_16_bit_time cp_16_time;

        /** @see ipd_scada_initialization_cause_coi. */
        struct ipd_scada_initialization_cause_coi coi;

        /** Value of interrogation qualifier. */
        u8 qoi_interrg_qual;

        /** @see ipd_scada_counter_intrg_qualifier_qcc. */
        struct ipd_scada_counter_intrg_qualifier_qcc qcc;

        /** @see ipd_scada_val_param_qualifier_qpm. */
        struct ipd_scada_val_param_qualifier_qpm qpm;

        /** Value of parameter activation qualifier. */
        u8 qpa_param_activation_qual;

        /** Value of reset process command. */
        u8 qrp_reset_process_cmd;

        /** @see ipd_scada_file_ready_qualifier_frq. */
        struct ipd_scada_file_ready_qualifier_frq frq;

        /** @see ipd_scada_section_ready_qualifier_srq. */
        struct ipd_scada_section_ready_qualifier_srq srq;

        /** Value of select and call qualifier. */
        u8 scq;

        /** Value of last segment or section qualifier*/
        u8 lsq;

        /** Value of acknowledge file or section qualifier. */
        u8 afq;

        /** 1-65535 is file name. */
        u16 nof;

        /** 1-255 is section name. */
        u8 nos;

        /** 24 bit file or section length. */
        u32 lof;

        /** 1-255 segment length. */
        u8 los;

        /** 0-255 checksum value. */
        u8 chs;

        /** @see ipd_scada_file_status_sof. */
        struct ipd_scada_file_status_sof sof;

        /** @see ipd_scada_set_point_cmd_qualifier_qos. */
        struct ipd_scada_set_point_cmd_qualifier_qos qos;

        /** @see ipd_scada_status_change_detection_scd. */
        struct ipd_scada_status_change_detection_scd scd;

        /** 16-bit test sequence counter. */
        u16 tsc;

        /** @see ipd_scada_single_point_info. */
        struct ipd_string file_seg;

    } data;
};

/**
 * @brief Data structure for advanced scada events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_SCADA_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** Scada event subtype. */
    PACE2_advanced_Scada_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_SCADA_APCI event. */
        struct ipd_scada_apci acpi;

        /** Contains data for a @ref PACE2_ADVANCED_SCADA_ASDU event. */
        struct ipd_scada_asdu asdu;

        /** Contains data for a @ref PACE2_ADVANCED_SCADA_INVALID event. */
        char const * invalid;

        /** Contains data for a @ref PACE2_ADVANCED_SCADA_INFORMATION_OBJECT event. */
        struct ipd_scada_info_object info_obj;

        /** Contains data for a @ref PACE2_ADVANCED_SCADA_INFORMATION_ELEMENT event. */
        struct ipd_scada_info_element info_elem;

        /** Contains data for a @ref PACE2_ADVANCED_SCADA_CONNECTION_TERMINATION event. */
        struct ipd_scada_connection_end conn_end;

    } event_data;

} PACE2_advanced_Scada_event;

#endif /* PACE_KERNEL_LIB */

#endif /* SCADA_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
