/** @file      sip_events.h
  * @brief     SIP event definitions.
  * @details   @copydoc doc_sip
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef SIP_EVENTS_H
#define SIP_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "rtp_events.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the basic SIP event.
 */
#define PACE2_BASIC_SIP_EVENT_GROUP PACE2_BASIC_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned metadata group of the advanced SIP event.
 */
#define PACE2_ADVANCED_SIP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced SIP events.
 */
typedef enum {

    /** Issued for each decoded SIP message. @c data points to a struct ipd_sip_message. */
    PACE2_ADVANCED_SIP_MESSAGE,

    /** Signals connection termination (TCP transport) or flow timeout (UDP, TCP). */
    PACE2_ADVANCED_SIP_END_OF_FLOW,

    /** Issued when decoding a message fails. An error message is passed as a
        <tt>char *</tt> in the @c data argument. */
    PACE2_ADVANCED_SIP_INVALID,

    /** Issued for each RTP packet in flows correlated to this SIP flow. @c data
        points to a struct ipd_rtp_data. */
    PACE2_ADVANCED_SIP_RTP_DATA,

    /** Signals flow timeout of a correlated RTP flow. */
    PACE2_ADVANCED_SIP_RTP_END_OF_FLOW,

    /** Issued when decoding a packet in a correlated RTP flow fails. An error
        message is passed as a <tt>char *</tt> in the @c data argument. */
    PACE2_ADVANCED_SIP_RTP_INVALID,

    /** Issued for each SSRC in RTP packets in flows correlated to this SIP flow. @c data
        points to a struct ipd_rtp_ssrc. */
    PACE2_ADVANCED_SIP_RTP_SSRC,

    /** Context was lost when the decoder tried to decode flow with holes. */
    PACE2_ADVANCED_SIP_CONTEXT_LOST,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_SIP_EVENT_TYPES

} PACE2_advanced_SIP_event_type;

/* Event Structures *************************************************************/

/**
 * @brief SIP metadata flags.
 */
enum PACE2_SIP_data_flags {

    /** SIP call state flag. */
    PACE2_SIP_CALL_STATE = 0,

    /** SIP call attribute flag. */
    PACE2_SIP_CALL_ATTR,

    /** Number of implemented SIP data flags. */
    PACE2_NUMBER_OF_SIP_DATA_FLAGS
};

/**
 * @brief SIP call states.
 */
typedef enum {

    /** Unknown SIP call state. */
    PACE2_SIP_CALL_UNKNOWN = 0,

    /** Setup of a SIP call. */
    PACE2_SIP_CALL_SETUP,

    /** Start of a SIP call. */
    PACE2_SIP_CALL_START,

    /** End of a SIP call. */
    PACE2_SIP_CALL_END,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_SIP_CALL_STATES

} PACE2_SIP_call_state;

/**
 * @brief Data structure for SIP events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_BASIC_SIP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** Bitmask, which states the available SIP metadata */
    PACE2_bitmask meta_data_mask;

    /** Variable of the @ref PACE2_SIP_CALL_STATE flag. */
    PACE2_SIP_call_state call_state;

    /** Data structure of the @ref PACE2_SIP_CALL_ATTR flag. */
    struct {

        /** SIP caller attribute. */
        PACE2_byte_buffer from;

        /** SIP callee attribute. */
        PACE2_byte_buffer to;

        /** SIP call ID. */
        PACE2_byte_buffer id;

    } call_attr;

} PACE2_basic_SIP_event;

#ifndef PACE_KERNEL_LIB

/** Fields of a SIP From: oder To: header. */
struct ipd_sip_contact_param {

    /** Entire value of the header line, eg. "Bob <sip:bob@biloxi.com>;tag=a6c85cf". */
    struct ipd_charset_string header_value;

    /** Display name, eg. "Bob". */
    struct ipd_charset_string display_name;

    /** URI, eg. "sip:bob@biloxi.com". */
    struct ipd_charset_string uri;

    /** Tag, eg. "a6c85cf". */
    struct ipd_charset_string tag;
};

/** Fields of a SIP request-line. */
struct ipd_sip_request_line {

    /** Request method (eg, REGISTER, INVITE, BYE). */
    struct ipd_charset_string method;

    /** Request-URI (eg, "sip:user@example.com"). */
    struct ipd_charset_string request_uri;

    /** SIP version ("SIP/2.0"). */
    struct ipd_charset_string sip_version;
};

/** Fields of a SIP status-line. */
struct ipd_sip_status_line {

    /** SIP version. */
    struct ipd_charset_string sip_version;

    /** Status code (eg, "200"). */
    struct ipd_charset_string status_code;

    /** Reason phrase (eg., "OK"). */
    struct ipd_charset_string reason_phrase;
};

/** Fields of a SIP CSeq header field. */
struct ipd_sip_cseq {

    /** Sequence number. */
    u32 number;

    /** Request method. */
    struct ipd_charset_string method;
};

/** @brief Elements of a Remote-Party-ID header field value.

    @details Refer to
    <a href="http://tools.ietf.org/html/draft-ietf-sip-privacy-04#section-6.1">Remote-Party-ID</a>
    for detailed information about this header field.
*/
struct ipd_sip_remote_party_id {

    /** @brief entire header field value, eg.
     * <tt>"Mary Doe" <sip:mdoe@foo.com>;party=called;ID-type=subscriber;screen=yes</tt>. */
    struct ipd_charset_string field_value;

    /** @brief display name, eg. "Mary Doe". */
    struct ipd_charset_string display_name;

    /** @brief URI, eg. "sip:mdoe@foo.com". */
    struct ipd_charset_string uri;

    /** @brief Pointer to an array of parameters, eg. "party=called", "ID-type=subscriber", "screen=yes". */
    const struct ipd_charset_string * parameters;

    /** @brief Number of parameters in #parameters. */
    u64 number_of_parameters;
};

/** @brief Representation of a P-Asserted-Identity header field value.

    @details Refer to
    <a href="http://tools.ietf.org/html/rfc3325#section-9.1">P-Asserted-Identity</a>
    for detailed information about this header field.
*/
struct ipd_sip_asserted_identity {

    /** @brief Entire header field value, eg.
     * <tt>"Cullen Jennings" <sip:fluffy@cisco.com></tt>.
     */
    struct ipd_charset_string field_value;

    /** @brief Display name, eg. "Cullen Jennings". */
    struct ipd_charset_string display_name;

    /** @brief URI, eg. "sip:fluffy@cisco.com". */
    struct ipd_charset_string uri;
};

/** @brief Representation of a SDP origin (o=) line. */
struct ipd_sdp_origin {

    /** @brief Entire field value (everything after the 'o='). */
    struct ipd_charset_string field_value;

    /** @brief User's login name on the originating host, or "-". */
    struct ipd_charset_string username;

    /** @brief Sess-id field. */
    struct ipd_charset_string sess_id;

    /** @brief Sess-version field. */
    struct ipd_charset_string sess_version;

    /** @brief Address type of the origin address (0 for IP4, 1 for IP6). */
    u8 address_is_ipv6;

    /** @brief Address of the machine from which this session was created.
        @details If the origin address in the session description is not given as an IPv4 or IPv6 address
        literal, but as a FQDN, this structure field is left at all-zeros. */
    union ip_v4_v6_addr address;
};

/** @brief Representation of a SDP media (m=) line. */
struct ipd_sdp_media {

    /** @brief the entire field value (everything after the 'm='). */
    struct ipd_charset_string field_value;

    /** @brief media type - "audio", "video" etc. */

    struct ipd_charset_string media_type;
    /** @brief protocol - usually "RTP/AVP". */
    struct ipd_charset_string proto;

    /** @brief First port (in host byte order). */
    u16 port_base;

    /** @brief Number of ports used for this media description. */
    u16 port_count;

    /** @brief Increment between port numbers.
        @details Usually 2 for RTP/AVP - even ports for RTP, odd for RTCP. */
    u16 port_step;

    /** @brief Pointer to an array of RTP Payload Type Strings */
    struct ipd_charset_string * media_formats;

    /** @brief Number of elements in media_formats. */
    u64 number_of_media_formats;
};

/** @brief Representation of a SDP connection data (c=) line. */
struct ipd_sdp_connection_data {

    /** @brief Entire field value (everything after the 'c='). */
    struct ipd_charset_string field_value;

    /** @brief Address type of the connection address (0 for IP4, 1 for IP6). */
    u8 address_is_ipv6;

    /** @brief Connection address.
        @details If the connection address in the session description is not given as an IPv4 or IPv6 address
        literal, but as a FQDN, this structure field is left at all-zeros. */
    union ip_v4_v6_addr address;

    /** @brief Number of addresses (assigned contiguously, starting from #address). */
    int address_count;

    /** @brief Index of the media description (in ipd_sip_message::sdp_media_lines) that this connection data
        line belongs to, or -1 for session-level connection data. */
    int media_index;
};

/** @brief Representation of a SDP attribute data (a=) line. */
struct ipd_sdp_attribute_data {

    /** @brief Entire field value (everything after the 'a='). */
    struct ipd_charset_string field_value;

    /** @brief Name of the attribute. */

    struct ipd_charset_string attribute;
    /** @brief Value of the attribute. */
    struct ipd_charset_string value;

    /** @brief Index of the media description (in ipd_sip_message::sdp_media_lines) that this attribute data
        line belongs to, or -1 for session-level attribute data. */
    int media_index;
};

/** Enumeration whether a SIP message is a request or a response. */
typedef enum {

    /** Signals a SIP request. */
    IPD_SIP_REQUEST,

    /** Signals a SIP response. */
    IPD_SIP_RESPONSE,

} ipd_sip_message_type;

/** Parsed header fields from a SIP message and its payload. */
struct ipd_sip_message {

    /** Sender IP address. */
    union ip_v4_v6_addr source_address;

    /** Flag that is set if @ref ipd_sip_message::source_address "source_address" contains an IPv6 address. */
    u8 source_address_is_ipv6;

    /** Message type (request or response). */
    ipd_sip_message_type type;

    union {
        struct ipd_sip_request_line request;
        struct ipd_sip_status_line response;
    };

    /* Header fields. */

    /** To: header field. */
    struct ipd_sip_contact_param to;

    /** From: header field. */
    struct ipd_sip_contact_param from;

    /** CSeq: header field. */
    struct ipd_sip_cseq cseq;

    /** Call-ID: header field. */
    struct ipd_charset_string call_id;

    /** User-Agent: header field (if set). */
    struct ipd_charset_string user_agent;

    /** Content-Type: header field. */
    struct ipd_charset_string content_type;

    /** @brief An array of Remote-Party-ID header fields, each represented by a struct ipd_sip_remote_party_id. */
    const struct ipd_sip_remote_party_id * remote_party_ids;

    /** @brief Number of elements in #remote_party_ids. */
    u64 number_of_remote_party_ids;

    /** @brief An array of P-Asserted-Identity header fields, each represented by a struct ipd_sip_asserted_identity. */
    const struct ipd_sip_asserted_identity * asserted_identities;

    /** @brief Number of elements in #asserted_identities. */
    u64 number_of_asserted_identities;

    /** The number of Via: header lines; in requests, a count larger than 1 indicates that the request passed
     * through a SIP proxy (cf. sections 8.1.1.7 and 16.6 of RFC 3261). */
    u64 via_count;

    /** Header of the SIP message. */
    struct ipd_charset_string header;

    /** Body of the SIP message. */
    struct ipd_charset_string body;

    /** @name SDP fields.
     * Fields that are only set when the content type is application/sdp. */
    /**@{*/

    /** @brief SDP origin ("o=") field, represented by a struct ipd_sdp_origin. */
    struct ipd_sdp_origin sdp_origin;

    /** @brief Array of SDP media ("m=") fields, each represented by a struct ipd_sdp_media. */
    const struct ipd_sdp_media * sdp_media_lines;

    /** @brief Number of elements in #sdp_media_lines. */
    u64 number_of_sdp_media_lines;

    /** @brief Array of SDP connection data ("c=") lines, each represented by a struct ipd_sdp_connection_data. */
    const struct ipd_sdp_connection_data * sdp_connection_data_lines;

    /** @brief Number of elements in #sdp_connection_data_lines. */
    u64 number_of_sdp_connection_data_lines;

    /** @brief Array of SDP attribute data ("a=") lines, each represented by a struct ipd_sdp_attribute_data. */
    const struct ipd_sdp_attribute_data * sdp_attribute_data_lines;

    /** @brief Number of elements in #sdp_attribute_data_lines. */
    u64 number_of_sdp_attribute_data_lines;

    /** Array of candidate addresses for RTP flows. */
    struct ipd_address_port * candidates_array;

    /** Number of candidate addresses. */
    u64 candidates_array_len;

    /**@}*/
};

/**
 * @brief Data structure for advanced SIP events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_SIP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** SIP event subtype. */
    PACE2_advanced_SIP_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_SIP_MESSAGE event. */
        struct ipd_sip_message sip_message;

        /** Contains data for a @ref PACE2_ADVANCED_SIP_RTP_DATA event. */
        struct ipd_rtp_data rtp_data;

        /** Contains data for a @ref PACE2_ADVANCED_SIP_INVALID event. */
        char const * invalid;

        /** Contains data for a @ref PACE2_ADVANCED_SIP_RTP_INVALID event. */
        char const * rtp_invalid;

        /** Contains data for a @ref PACE2_ADVANCED_SIP_RTP_SSRC event. */
        struct ipd_rtp_ssrc rtp_ssrc;

    } event_data;

} PACE2_advanced_SIP_event;

#endif

#endif /* SIP_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
