/** @file      smb_events.h
  * @brief     SMB event definitions.
  * @details   @copydoc doc_smb
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef SMB_EVENTS_H
#define SMB_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced SMB event.
 */
#define PACE2_ADVANCED_SMB_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced SMB events.
 */
typedef enum {

    /**
     * New connection is observed.<br />
     * Provides: NULL.
     */
    PACE2_ADVANCED_SMB_NEW_CONNECTION,

    /**
     * SMB1 header is observed.<br />
     * Provides a struct @ref PACE2_SMB1_header.
     */
    PACE2_ADVANCED_SMB_VERSION1_HEADER,

    /**
     * SMB2 header is observed.<br />
     * Provides a struct @ref PACE2_SMB2_header.
     */
    PACE2_ADVANCED_SMB_VERSION2_HEADER,

    /**
     * Parameter part of a SMB message is not decoded and thrown uninterpreted.<br />
     * Provides a struct @ref PACE2_SMB_undecoded_data.
     */
    PACE2_ADVANCED_SMB_UNDECODED_PARAMETER,

    /**
     * Data part of a SMB message is not decoded and thrown uninterpreted.<br />
     * Provides a struct @ref PACE2_SMB_undecoded_data.
     */
    PACE2_ADVANCED_SMB_UNDECODED_DATA,

    /**
     * Decoding of a connection fails.<br />
     * Provides: const char * : message describing the failure reason.
     */
    PACE2_ADVANCED_SMB_INVALID,

    /**
     * Flow ends.<br />
     * Provides: NULL.
     */
    PACE2_ADVANCED_SMB_END_OF_FLOW,

    /**
     * Data part of a SMB message is not decoded and thrown uninterpreted.<br />
     * Provides a struct @ref PACE2_smb_encrypted_data.
     */
    PACE2_ADVANCED_SMB_ENCRYPTED_DATA,

    /**
     * SMB2 transform_header is observed.<br />
     * Provides a struct @ref PACE2_smb2_transform_header.
     */
    PACE2_ADVANCED_SMB_VERSION2_TRANSFORM_HEADER,

    /**
     * Decoding of the SMB data part.<br />
     * Requested File name is given as event.
     * Provides a struct @ref PACE2_smb_request_transaction2.
     */
    PACE2_ADVANCED_SMB_REQUEST_TRANSACTION2_QUERY_PATH,

    /**
     * Decoding of the SMB data part.<br />
     * File name, creation dates are decoded and given as event.
     * Provides a struct @ref PACE2_smb_response_transaction2_query_path.
     */
    PACE2_ADVANCED_SMB_RESPONSE_TRANSACTION2_QUERY_PATH,

    /**
     * Decoding of the SMB data part.<br />
     * Name of Directory is given as event.
     * Provides a struct @ref PACE2_smb_request_transaction2.
     */
    PACE2_ADVANCED_SMB_REQUEST_TRANSACTION2_FIND_FIRST2,

    /**
     * Decoding of the SMB data part.<br />
     * Content of directory at its creation dates are decoded and given as event.
     * Provides a struct @ref PACE2_smb_response_transaction2_find_first2.
     */
    PACE2_ADVANCED_SMB_RESPONSE_TRANSACTION2_FIND_FIRST2,

    /**
     * Decoding of the SMB data part.<br />
     * File name is decoded and given as event.
     */
    PACE2_ADVANCED_SMB_REQUEST_NT_CREATE_ANDX,

    /**
     * Decoding of the SMB parameter part.<br />
     * FileID is decoded and given as event.
     * Provides a struct @ref PACE2_smb_file_id.
     */
    PACE2_ADVANCED_SMB_RESPONSE_NT_CREATE_ANDX_PARAM,

    /**
     * Decoding of the SMB data part.<br />
     * Written file content is decoded and given as event.
     */
    PACE2_ADVANCED_SMB_REQUEST_WRITE_ANDX,

    /**
     * Decoding of the SMB parameter part.<br />
     * FileID is decoded and given as event.
     * Provides a struct @ref PACE2_smb_file_id.
     */
    PACE2_ADVANCED_SMB_REQUEST_WRITE_ANDX_PARAM,

    /**
     * Decoding of the SMB data part.<br />
     * Event is empty and indicates the Answer of PACE2_ADVANCED_SMB_REQUEST_WRITE_ANDX.
     */
    PACE2_ADVANCED_SMB_RESPONSE_WRITE_ANDX,

    /**
     * Decoding of the SMB data part.<br />
     * Domain name, host name and user name in terms on Microsoft LAN Manager is decoded and given as event.
     * Provides a struct @ref PACE2_smb_authentication.
     */
    PACE2_ADVANCED_SMB_AUTHENTICATION,

    /**
     * Decoding of the SMB data part.<br />
     * A selection of interpretable data like read file content is decoded and given as event.
     */
    PACE2_ADVANCED_SMB_RESPONSE_READ_ANDX,

    /**
     * Decoding of the SMB parameter part.<br />
     * FileID is decoded and given as event.
     * Provides a struct @ref PACE2_smb_file_id.
     */
    PACE2_ADVANCED_SMB_REQUEST_READ_ANDX_PARAM,

    /**
     * Decoding of the SMB data part.<br />
     * File name to delete is decoded and given as event.
     */
    PACE2_ADVANCED_SMB_DELETE_FILE,

    /**
     * Decoding of the SMB data part.<br />
     * File name and its disposition is decoded and given as event.
     * Provides a struct @ref PACE2_smb_request_create.
     */
    PACE2_ADVANCED_SMB_REQUEST_CREATE,

    /**
     * Decoding of the SMB data part.<br />
     * File properties is decoded and given as event.
     * Provides a struct @ref PACE2_smb_response_create.
     */
    PACE2_ADVANCED_SMB_RESPONSE_CREATE,

    /**
     * Decoding of the SMB data part.<br />
     * Length and offset of data block and file_id decoded and given as event.
     * Provides a struct @ref PACE2_smb_request_read.
     */
    PACE2_ADVANCED_SMB_REQUEST_READ,

    /**
     * Decoding of the SMB data part.<br />
     * Length of data block and the data read is decoded and given as event.
     */
    PACE2_ADVANCED_SMB_RESPONSE_READ,

    /**
     * Decoding of the SMB data part.<br />
     * The data is decoded and given as event.
     * Provides a struct @ref PACE2_smb_request_write.
     */
    PACE2_ADVANCED_SMB_REQUEST_WRITE,

    /**
     * Decoding of the SMB data part.<br />
     * The data is decoded and given as event.
     * Provides a struct @ref PACE2_smb_request_tree_connect_andx.
     */
    PACE2_ADVANCED_SMB_REQUEST_TREE_CONNECT_ANDX,

    /**
     * Decoding of the SMB data part.<br />
     * The data is decoded and given as event.
     */
    PACE2_ADVANCED_SMB_REQUEST_TREE_CONNECT,

    /**
     * Decoding of the SMB data part.<br />
     * The tree_id for SMB1 is given.
     * Provides a struct @ref PACE2_smb_tree_id.
     */
    PACE2_ADVANCED_SMB_RESPONSE_TREE_CONNECT_ANDX,

    /**
     * Decoding of the SMB data part.<br />
     * The tree_id for SMB2 is given.
     * Provides a struct @ref PACE2_smb_tree_id.
     */
    PACE2_ADVANCED_SMB_RESPONSE_TREE_CONNECT,

    /**
     * A dummy event to signal the end of the event list.
     */
    PACE2_NUMBER_OF_ADVANCED_SMB_EVENTS

} PACE2_advanced_SMB_event_type;

/**
 * @brief Enumeration for SMB1 commands.
 */
enum PACE2_SMB1_command {
    SMB1_COMMAND_CREATE_DIRECTORY        = 0x00,
    SMB1_COMMAND_DELETE_DIRECTORY        = 0x01,
    SMB1_COMMAND_OPEN                    = 0x02,
    SMB1_COMMAND_CREATE                  = 0x03,
    SMB1_COMMAND_CLOSE                   = 0x04,
    SMB1_COMMAND_FLUSH                   = 0x05,
    SMB1_COMMAND_DELETE                  = 0x06,
    SMB1_COMMAND_RENAME                  = 0x07,
    SMB1_COMMAND_QUERY_INFORMATION       = 0x08,
    SMB1_COMMAND_SET_INFORMATION         = 0x09,
    SMB1_COMMAND_READ                    = 0x0A,
    SMB1_COMMAND_WRITE                   = 0x0B,
    SMB1_COMMAND_LOCK_BYTE_RANGE         = 0x0C,
    SMB1_COMMAND_UNLOCK_BYTE_RANGE       = 0x0D,
    SMB1_COMMAND_CREATE_TEMPORARY        = 0x0E,
    SMB1_COMMAND_CREATE_NEW              = 0x0F,
    SMB1_COMMAND_CHECK_DIRECTORY         = 0x10,
    SMB1_COMMAND_PROCESS_EXIT            = 0x11,
    SMB1_COMMAND_SEEK                    = 0x12,
    SMB1_COMMAND_LOCK_AND_READ           = 0x13,
    SMB1_COMMAND_WRITE_AND_UNLOCK        = 0x14,
    SMB1_COMMAND_READ_RAW                = 0x1A,
    SMB1_COMMAND_READ_MPX                = 0x1B,
    SMB1_COMMAND_READ_MPX_SECONDARY      = 0x1C,
    SMB1_COMMAND_WRITE_RAW               = 0x1D,
    SMB1_COMMAND_WRITE_MPX               = 0x1E,
    SMB1_COMMAND_WRITE_COMPLETE          = 0x20,
    SMB1_COMMAND_QUERY_SERVER            = 0x21,
    SMB1_COMMAND_SET_INFORMATION2        = 0x22,
    SMB1_COMMAND_QUERY_INFORMATION2      = 0x23,
    SMB1_COMMAND_LOCKING_ANDX            = 0x24,
    SMB1_COMMAND_TRANSACTION             = 0x25,
    SMB1_COMMAND_TRANSACTION_SECONDARY   = 0x26,
    SMB1_COMMAND_IOCTL                   = 0x27,
    SMB1_COMMAND_IOCTL_SECONDARY         = 0x28,
    SMB1_COMMAND_COPY                    = 0x29,
    SMB1_COMMAND_MOVE                    = 0x2A,
    SMB1_COMMAND_ECHO                    = 0x2B,
    SMB1_COMMAND_WRITE_AND_CLOSE         = 0x2C,
    SMB1_COMMAND_OPEN_ANDX               = 0x2D,
    SMB1_COMMAND_READ_ANDX               = 0x2E,
    SMB1_COMMAND_WRITE_ANDX              = 0x2F,
    SMB1_COMMAND_CLOSE_AND_TREE_DISC     = 0x31,
    SMB1_COMMAND_TRANSACTION2            = 0x32,
    SMB1_COMMAND_TRANSACTION2_SECONDARY  = 0x33,
    SMB1_COMMAND_FIND_CLOSE2             = 0x34,
    SMB1_COMMAND_FIND_NOTIFY_CLOSE       = 0x35,
    SMB1_COMMAND_TREE_CONNECT            = 0x70,
    SMB1_COMMAND_TREE_DISCONNECT         = 0x71,
    SMB1_COMMAND_NEGOTIATE               = 0x72,
    SMB1_COMMAND_SESSION_SETUP_ANDX      = 0x73,
    SMB1_COMMAND_LOGOFF_ANDX             = 0x74,
    SMB1_COMMAND_TREE_CONNECT_ANDX       = 0x75,
    SMB1_COMMAND_SECURITY_PACKAGE_ANDX   = 0x7E,
    SMB1_COMMAND_QUERY_INFORMATION_DISK  = 0x80,
    SMB1_COMMAND_SEARCH                  = 0x81,
    SMB1_COMMAND_FIND                    = 0x82,
    SMB1_COMMAND_FIND_UNIQUE             = 0x83,
    SMB1_COMMAND_FIND_CLOSE              = 0x84,
    SMB1_COMMAND_NT_TRANSACT             = 0xA0,
    SMB1_COMMAND_NT_TRANSACT_SECONDARY   = 0xA1,
    SMB1_COMMAND_NT_CREATE_ANDX          = 0xA2,
    SMB1_COMMAND_NT_CANCEL               = 0xA4,
    SMB1_COMMAND_NT_RENAME               = 0xA5,
    SMB1_COMMAND_OPEN_PRINT_FILE         = 0xC0,
    SMB1_COMMAND_WRITE_PRINT_FILE        = 0xC1,
    SMB1_COMMAND_CLOSE_PRINT_FILE        = 0xC2,
    SMB1_COMMAND_GET_PRINT_QUEUE         = 0xC3,
    SMB1_COMMAND_READ_BULK               = 0xD8,
    SMB1_COMMAND_WRITE_BULK              = 0xD9,
    SMB1_COMMAND_WRITE_BULK_DATA         = 0xDA,
    SMB1_COMMAND_INVALID                 = 0xFE,
    SMB1_COMMAND_NO_ANDX_COMMAND         = 0xFF
};

/**
 * @brief Enumeration for SMB2 commands.
 */
enum PACE2_SMB2_command {
    SMB2_COMMAND_NEGOTIATE            = 0x0000,
    SMB2_COMMAND_SESSION_SETUP        = 0x0001,
    SMB2_COMMAND_LOGOFF               = 0x0002,
    SMB2_COMMAND_TREE_CONNECT         = 0x0003,
    SMB2_COMMAND_TREE_DISCONNECT      = 0x0004,
    SMB2_COMMAND_CREATE               = 0x0005,
    SMB2_COMMAND_CLOSE                = 0x0006,
    SMB2_COMMAND_FLUSH                = 0x0007,
    SMB2_COMMAND_READ                 = 0x0008,
    SMB2_COMMAND_WRITE                = 0x0009,
    SMB2_COMMAND_LOCK                 = 0x000A,
    SMB2_COMMAND_IOCTL                = 0x000B,
    SMB2_COMMAND_CANCEL               = 0x000C,
    SMB2_COMMAND_ECHO                 = 0x000D,
    SMB2_COMMAND_QUERY_DIRECTORY      = 0x000E,
    SMB2_COMMAND_CHANGE_NOTIFY        = 0x000F,
    SMB2_COMMAND_QUERY_INFO           = 0x0010,
    SMB2_COMMAND_SET_INFO             = 0x0011,
    SMB2_COMMAND_OPLOCK_BREAK         = 0x0012
};

/**
 * @brief Enumeration for SMB message type.
 */
enum PACE2_SMB_message_type {
    SMB_MESSAGE_TYPE_REQUEST,
    SMB_MESSAGE_TYPE_RESPONSE
};

/**
 * @brief Enumeration for SMB error status.
 */
enum PACE2_SMB_error_status {
    SMB_ERROR_STATUS_UNKNOWN_STATUS,
    SMB_ERROR_STATUS_OK,
    SMB_ERROR_STATUS_INVALID_FUNCTION,
    SMB_ERROR_STATUS_FILE_NOT_FOUND,
    SMB_ERROR_STATUS_COMPONENT_IN_THE_PATH_PREFIX_IS_NOT_A_DIRECTORY,
    SMB_ERROR_STATUS_TOO_MANY_OPEN_FILES_NO_FIDS_AVAILABLE,
    SMB_ERROR_STATUS_ACCESS_DENIED,
    SMB_ERROR_STATUS_INVALID_FID,
    SMB_ERROR_STATUS_MEMORY_CONTROL_BLOCKS_WERE_DESTROYED,
    SMB_ERROR_STATUS_ISUFFICIENT_SERVER_MEMORY_TO_PERFORM_THE_REQUESTED_OPERATION,
    SMB_ERROR_STATUS_SERVER_PERFORMED_INVALID_MEMORY_ACCESS,
    SMB_ERROR_STATUS_INVALID_ENVIRONMENT,
    SMB_ERROR_STATUS_INVALID_FORMAT,
    SMB_ERROR_STATUS_INVALID_OPEN_MODE,
    SMB_ERROR_STATUS_BAD_DATA,
    SMB_ERROR_STATUS_INVALID_DRIVE_SPECIFIED,
    SMB_ERROR_STATUS_REMOVING_DIRECTORY_FAILED_BECAUSE_IT_WAS_NOT_EMPTY,
    SMB_ERROR_STATUS_FILE_SYSTEM_OPERATION_ACROSS_TWO_DEVICES_WAS_ATTEMPTED,
    SMB_ERROR_STATUS_NO_MORE_FILES_FOUND_FOLLOWING_A_FILE_SEARCH_COMMAND,
    SMB_ERROR_STATUS_GENERAL_ERROR,
    SMB_ERROR_STATUS_SHARING_VIOLATION,
    SMB_ERROR_STATUS_LOCK_REQUEST_SPECIFIED_INVALID_LOCKING_MODE_OR_CONFLICTED_WITH_EXISTING_FILE_LOCK,
    SMB_ERROR_STATUS_ATTMPTED_TO_READ_BEYND_THE_END_OF_FILE,
    SMB_ERROR_STATUS_COMMAND_NOT_SUPPORTED_BY_SERVER,
    SMB_ERROR_STATUS_CANNOT_CREATE_FILE_OR_DIRECTORY_AS_ALREADY_EXISTS,
    SMB_ERROR_STATUS_INVALID_PARAMETER,
    SMB_ERROR_STATUS_INVALID_INFORMATION_LEVEL,
    SMB_ERROR_STATUS_NEGATIVE_ABSOLUTE_SEEK_OFFSET_PROVIDED,
    SMB_ERROR_STATUS_BYTE_RANGE_SPECIFIED_IN_THE_UNLOCK_REQUEST_WAS_NOT_LOCKED,
    SMB_ERROR_STATUS_MAXIMUM_NUMBER_OF_SEARCHES_EXHAUSTED,
    SMB_ERROR_STATUS_NO_LOCK_REQUEST_WAS_OUTSTANDING_FOR_THE_SUPPLIED_CANCEL_REGION,
    SMB_ERROR_STATUS_FILE_SYSTEM_DOES_NOT_SUPPORT_ATOMIC_CHANGES_TO_THE_LOCK_TYPE,
    SMB_ERROR_STATUS_INVALID_NAMED_PIPE,
    SMB_ERROR_STATUS_COPY_FUNCTIONS_CANNOT_BE_USED,
    SMB_ERROR_STATUS_ALL_INSTANCES_OF_THE_DESIGNATED_NAMED_PIPE_BUSY,
    SMB_ERROR_STATUS_DESIGNATED_NAMED_PIPE_IS_IN_THE_PROCESS_OF_BEING_CLOSED,
    SMB_ERROR_STATUS_NO_SERVER_PROCESS_LISTENS_TO_THE_PIPE,
    SMB_ERROR_STATUS_MORE_DATA_AVAILABLE_TO_READ_ON_THE_DESIGNATED_NAMED_PIPE,
    SMB_ERROR_STATUS_INCONSISTENT_EXTENDED_ATTRIBUTE_LIST,
    SMB_ERROR_STATUS_NO_EXTENDED_ATTRIBUTES_OR_THEY_DO_NOT_FIT_INTO_RESPONSE,
    SMB_ERROR_STATUS_EXTENDED_ATTRIBUTES_NOT_SUPPORTED,
    SMB_ERROR_STATUS_ACCESS_TO_EXTENDED_ATTRIBUTE_DENIED,
    SMB_ERROR_STATUS_MORE_CHANGES_OCCURED_THAN_SPACE_IN_BUFFER,
    SMB_ERROR_STATUS_UNSPECIFIED_SERVER_ERROR,
    SMB_ERROR_STATUS_INVALID_PASSWORD,
    SMB_ERROR_STATUS_DFS_PATHNAME_NOT_ON_LOCAL_SERVER,
    SMB_ERROR_STATUS_INVALID_NETWORK_PATH_IDENTIFIER,
    SMB_ERROR_STATUS_INVALID_SERVER_NAME_IN_TREE_CONNECT,
    SMB_ERROR_STATUS_PRINTER_REQUEST_MADE_TO_NONPRINTER_DEVICE_OR_OPPOSITE,
    SMB_ERROR_STATUS_INVALID_CONNECTION_ID,
    SMB_ERROR_STATUS_COMMAND_WITH_MATCHING_MID_OR_SEQUENCE_NUMBER_CURRENTLY_BEING_PROCESSED,
    SMB_ERROR_STATUS_INVALID_NETBIOS_CALLED_NAME,
    SMB_ERROR_STATUS_UNKNOWN_SMB_COMMAND_CODE_RECEIVED,
    SMB_ERROR_STATUS_PRINT_QUEUE_FULL_TOO_MANY_ITEMS,
    SMB_ERROR_STATUS_PRINT_QUEUE_FULL_NO_SPOOL_SPACE,
    SMB_ERROR_STATUS_END_OF_FILE_ON_PRINT_QUEUE_DUMP,
    SMB_ERROR_STATUS_INVALID_FID_FOR_PRINT_FILE,
    SMB_ERROR_STATUS_UNRECOGNIZED_SMB_COMMAND_CODE,
    SMB_ERROR_STATUS_INTERNAL_SERVER_ERROR,
    SMB_ERROR_STATUS_FID_AND_PATHNAME_CONTAIN_INCOMPATIBLE_VALUES,
    SMB_ERROR_STATUS_INVALID_COMBINATION_OF_ACCESS_PERMISSIONS,
    SMB_ERROR_STATUS_ATTRIBUTE_MODE_INVALID,
    SMB_ERROR_STATUS_OPERATION_TIMED_OUT,
    SMB_ERROR_STATUS_NO_RESOURCES_AVAILABLE,
    SMB_ERROR_STATUS_TOO_MANY_UIDS_ACTIVE_FOR_THIS_SMB_CONNECTION,
    SMB_ERROR_STATUS_INVALID_UID_ON_THIS_SERVER_SESSION,
    SMB_ERROR_STATUS_WRITE_TO_A_NAMED_PIPE_WITHOUT_READER,
    SMB_ERROR_STATUS_TEMPORARILY_UNABLE_TO_SUPPORT_RAW_MODE_TRANSFERS,
    SMB_ERROR_STATUS_TEMPORARILY_UNABLE_TO_SUPPORT_RAW_OR_MPX_MODE_TRANSFERS,
    SMB_ERROR_STATUS_CONTINUE_IN_MPX_MODE,
    SMB_ERROR_STATUS_USER_ACCOUNT_ON_TARGET_MACHINE_DISABLED_OR_EXPIRED,
    SMB_ERROR_STATUS_NO_PERMISSIONS_TO_ACCESS_THIS_SERVER,
    SMB_ERROR_STATUS_ACCESS_TO_THE_SERVER_NOT_PERMITTED_AT_THIS_TIME,
    SMB_ERROR_STATUS_USER_PASSWORD_EXPIRED,
    SMB_ERROR_STATUS_FUNCTIN_NOT_SUPPORTED_BY_SERVER,
    SMB_ERROR_STATUS_ATTEMPT_TO_MODIFY_READ_ONLY_FILE,
    SMB_ERROR_STATUS_UNKNOWN_UNIT,
    SMB_ERROR_STATUS_DRIVE_NOT_READY,
    SMB_ERROR_STATUS_UNKNOWN_COMMAND,
    SMB_ERROR_STATUS_DATA_ERROR_INCORRECT_CRC,
    SMB_ERROR_STATUS_BAD_REQUEST_STRUCTURE_LENGTH,
    SMB_ERROR_STATUS_SEEK_ERROR,
    SMB_ERROR_STATUS_UNKNOWN_MEDIA_TYPE,
    SMB_ERROR_STATUS_SECTOR_NOT_FOUND,
    SMB_ERROR_STATUS_PRINTER_OUT_OF_PAPER,
    SMB_ERROR_STATUS_WRITE_FAULT,
    SMB_ERROR_STATUS_READ_FAULT,
    SMB_ERROR_STATUS_GENERAL_HARDWARE_FAILURE,
    SMB_ERROR_STATUS_ATTEMPTED_OPEN_OPERATION_CONFLICTS_WITH_AN_EXISTING_OPEN,
    SMB_ERROR_STATUS_WRONG_DISK_FOUND,
    SMB_ERROR_STATUS_NO_SERVER_SIDE_FILE_CONTROL_BLOCKS_AVAILABLE,
    SMB_ERROR_STATUS_SHARING_BUFFER_EXCEEDED,
    SMB_ERROR_STATUS_NO_SPACE_ON_FILE_SYSTEM
};

/**
 * @brief Enumeration for smb1 transaction2 subcommand.
 */
enum pace2_smb1_transaction_subcommand {
    SMB1_TRANS2_OPEN2                    = 0x0000,
    SMB1_TRANS2_FIND_FIRST2              = 0x0001,
    SMB1_TRANS2_FIND_NEXT2               = 0x0002,
    SMB1_TRANS2_QUERY_FS_INFORMATION     = 0x0003,
    SMB1_TRANS2_SET_FS_INFORMATION       = 0x0004,
    SMB1_TRANS2_QUERY_PATH_INFORMATION   = 0x0005,
    SMB1_TRANS2_SET_PATH_INFORMATION     = 0x0006,
    SMB1_TRANS2_QUERY_FILE_INFORMATION   = 0x0007,
    SMB1_TRANS2_SET_FILE_INFORMATION     = 0x0008,
    SMB1_TRANS2_FSCTL                    = 0x0009,
    SMB1_TRANS2_IOCTL2                   = 0x000A,
    SMB1_TRANS2_FIND_NOTIFY_FIRST        = 0x000B,
    SMB1_TRANS2_FIND_NOTIFY_NEXT         = 0x000C,
    SMB1_TRANS2_CREATE_DIRECTORY         = 0x000D,
    SMB1_TRANS2_SESSION_SETUP            = 0x000E,
    SMB1_TRANS2_GET_DFS_REFERRAL         = 0x0010,
    SMB1_TRANS2_REPORT_DFS_INCONSISTENCY = 0x0011
};

/**
 * @brief Enumeration for SMB1 QUERY Information Level Codes.
 */
enum pace2_smb1_query_path_information_level {
    SMB_QUERY_INFO_STANDARD             = 0x0001,  // _QUERY_ is inserted to separate the enum QUERY and enum FIND
    SMB_QUERY_INFO_QUERY_EA_SIZE        = 0x0002,  // _QUERY_ is inserted to separate the enum QUERY and enum FIND
    SMB_QUERY_INFO_QUERY_EAS_FROM_LIST  = 0x0003,  // _QUERY_ is inserted to separate the enum QUERY and enum FIND
    SMB_QUERY_INFO_QUERY_ALL_EAS        = 0x0004,
    SMB_QUERY_INFO_IS_NAME_VALID        = 0x0006,
    SMB_QUERY_FILE_BASIC_INFO           = 0x0101,
    SMB_QUERY_FILE_STANDARD_INFO        = 0x0102,
    SMB_QUERY_FILE_EA_INFO              = 0x0103,
    SMB_QUERY_FILE_NAME_INFO            = 0x0104,
    SMB_QUERY_FILE_ALL_INFO             = 0x0107,
    SMB_QUERY_FILE_ALT_NAME_INFO        = 0x0108,
    SMB_QUERY_FILE_STREAM_INFO          = 0x0109,
    SMB_QUERY_FILE_COMPRESSION_INFO     = 0x010B
};

/**
 * @brief Enumeration for SMB1 FIND Information Level Codes.
 */
enum pace2_smb1_find_first2_information_level {
    SMB_FIND_INFO_STANDARD             = 0x0001,  // _FIND_ is inserted to separate the enum QUERY and enum FIND
    SMB_FIND_INFO_QUERY_EA_SIZE        = 0x0002,  // _FIND_ is inserted to separate the enum QUERY and enum FIND
    SMB_FIND_INFO_QUERY_EAS_FROM_LIST  = 0x0003,  // _FIND_ is inserted to separate the enum QUERY and enum FIND
    SMB_FIND_FILE_DIRECTORY_INFO       = 0x0101,
    SMB_FIND_FILE_FULL_DIRECTORY_INFO  = 0x0102,
    SMB_FIND_FILE_NAMES_INFO           = 0x0103,
    SMB_FIND_FILE_BOTH_DIRECTORY_INFO  = 0x0104
};

/**
 * @brief Enumeration for SMB2 CreateDispostion Types.
 */
enum pace2_smb2_create_disposition {
    SMB2_FILE_SUPERSEDE             = 0x00000000,
    SMB2_FILE_OPEN                  = 0x00000001,
    SMB2_FILE_CREATE                = 0x00000002,
    SMB2_FILE_OPEN_IF               = 0x00000003,
    SMB2_FILE_OVERWRITE             = 0x00000004,
    SMB2_FILE_OVERWRITE_IF          = 0x00000005
};


/**
 * @brief Enumeration for SMB2 CreateAction Types.
 */
enum pace2_smb2_create_action {
    SMB2_FILE_SUPERSEDED            = 0x00000000,
    SMB2_FILE_OPENED                = 0x00000001,
    SMB2_FILE_CREATED               = 0x00000002,
    SMB2_FILE_OVERWRITTEN           = 0x00000003
};


/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * @brief Header of an SMB1 message.
 */
struct PACE2_SMB1_header {

    enum PACE2_SMB_message_type message_type;       /**< Informs if this message is a command or a response. */
    enum PACE2_SMB1_command command;                /**< Command for which the message was sent. */
    enum PACE2_SMB_error_status error_status;       /**< Error status. */

    u8 direction;                                   /**< The direction is defined identical to the
                                                         @ref PACE2_packet_descriptor.direction member. */

    u8 oplock_requested_or_granted;                 /**< If set, an oplock was requested or granted. */
    u8 path_names_canonicalized;                    /**< If set, all pathnames are canonicalized. */
    u8 path_names_caseless;                         /**< If set, all pathnames are case-less. */
    u8 receive_buffer_posted;                       /**< If set, the receive buffer has been posted. */
    u8 lock_read_supported;                         /**< If set, Lock&Read and Write&Unlock are supported. */

    u8 unicode_strings;                             /**< If set, strings are unicode. */
    u8 execute_only_reads_allowed;                  /**< If set, reads are allowed in the execute-only mode. */
    u8 resolve_path_names_with_dfs;                 /**< If set, the path names should be resolved with Distributed File Systems. */
    u8 extended_security_negotiation_supported;     /**< If set, the client is aware of Extended Security negotiation. */
    u8 reparse_path_used;                           /**< If set, a reparse path is used. */
    u8 long_names_used;                             /**< If set, any path name in the request is a long name. */
    u8 security_signatures_required;                /**< If set, security signatures are required. */
    u8 compression_is_requested;                    /**< If set, compression is requested. */
    u8 security_signatures_supported;               /**< If set, security signatures are supported. */
    u8 extended_attributes_supported;               /**< If set, extended attributes are supported. */
    u8 long_names_allowed;                          /**< If set, long file names are allowed in the response. */

    u8 security_signature[8];                       /**< Security signature. */

    u16 tree_identifier;                            /**< Represents an instance of an authenticated connection to a server resource. */
    u32 process_id;                                 /**< Identifies to the server the "process" that opened a file or that owns a byte range lock. */
    u16 user_id;                                    /**< Assigned by the server after a user authenticates to it, and that it will associate with that user until the client requests the association be broken. */
    u16 multiplex_id;                               /**< Used to allow multiplexing the single client and server connection among the client's multiple processes, threads, and requests per thread. */
};

/**
 * @brief Header of an SMB2 message.
 */
struct PACE2_SMB2_header {

    enum PACE2_SMB_message_type message_type;       /**< Informs if this message is a command or a response. */
    enum PACE2_SMB2_command command;                /**< Command for which the message was sent. */
    enum PACE2_SMB_error_status error_status;       /**< Error status. */

    u8 direction;                                   /**< The direction is defined identical to the
                                                    @ref PACE2_packet_descriptor.direction member. */

    u16 credit_charge;                              /**< The credit charge. */
    u16 credit_granted;                             /**< The granted credit. */
    u8 async_command;                               /**< If set, command is async. */
    u8 chained;                                     /**< If set, command is chained. */
    u8 signed_header;                               /**< If set, the header is signed. */
    u8 dfs_operation;                               /**< If set, this is a Distributed File System operation. */
    u8 replay_operation;                            /**< If set, this command is a replay operation. */
    u8 signature[16];                               /**< The signature. */
    u64 message_identifier;                         /**< A value that identifies a message request and response uniquely across all messages that are sent on the same SMB 2 Protocol transport connection. */
    u64 async_identifier;                           /**< If command is async this field is the async identifier. Otherwise this field is set to 0. */
    u32 tree_identifier;                            /**< If command is not async this field uniquely identifies the tree connect for the command. Otherwise this field is set to 0. */
    u64 session_identifier;                         /**< Uniquely identifies the established session for the command. */
};

/**
 * @brief Header of an SMB2 message.
 */
struct PACE2_smb2_transform_header {
    u8 direction;                                   /**< The direction is defined identical to the
                                                         @ref PACE2_packet_descriptor.direction member.*/
    struct ipd_string signature;                    /**< The signature. */
    struct ipd_string  nonce;                       /**< The nonce. */
    u32 original_message_size;                      /**< The size of the encrypted data */
    struct ipd_string  flags_encryption_algorithm;  /**< The encryption algorithm */
    struct ipd_string  session_id;                  /**< The Session_id */
};

/**
 * @brief Structure for undecoded events.
 */
struct PACE2_SMB_undecoded_data {
    struct ipd_charset_string data;                 /**< The data represented as a pointer to a buffer and a length field. */
    u8 direction;                                   /**< The direction is defined identical to the
                                                         @ref PACE2_packet_descriptor.direction member. */
};

/**
 * @brief Structure for undecoded events.
 */
struct PACE2_smb_encrypted_data {
    struct ipd_charset_string data;                 /**< The data represented as a pointer to a buffer and a length field. */
    u8 direction;                                   /**< The direction is defined identical to the
                                                        @ref PACE2_packet_descriptor.direction member. */
};

/**
 * @brief Structure for REQUEST_TRANSACTION2_QUERY_PATH and REQUEST_TRANSACTION2_FIND_FIRST2 events.
 */
struct PACE2_smb_request_transaction2 {
    struct ipd_string file_name;                    /**< The data represented as a pointer to a buffer and
                                                         a length field containing the SMB file or path name. */
};

/**
 * @brief Structure for RESPONSE_TRANSACTION2_QUERY_PATH events.
 */
struct PACE2_smb_response_transaction2_query_path {
    struct ipd_string name;                         /**< The data represented as a pointer to a buffer and
                                                         a length field containing the SMB file or path name. */
    u64 creation_time;                              /**< The file name Creation Date */
    u64 last_access_time;                           /**< The file name Last Access Date */
    u64 last_write_time;                            /**< The file name Last Write Date */
    u64 last_change_time;                           /**< The file name Last Change Date */
    u8 is_directory;                                /**< The file name is a directory. */
};

/**
 * @brief Structure for RESPONSE_TRANSACTION2_FIND_FIRST2 events.
 */
struct PACE2_smb_response_transaction2_find_first2 {
    struct ipd_string name;                         /**< The data represented as a pointer to a buffer and a length field containing the SMB file or path name. */
    u64 creation_time;                              /**< The file name Creation Date */
    u64 last_access_time;                           /**< The file name Last Access Date */
    u64 last_write_time;                            /**< The file name Last Write Date */
    u64 last_change_time;                           /**< The file name Last Change Date */
};

/** @brief Structure for SMB2 request create data events.
*/
struct PACE2_smb_request_create {
    enum pace2_smb2_create_disposition create_disposition; /**< Disposition of the action Create. */
    struct ipd_string file_name;                     /**< Contains the SMB file or path name to delete. */
};

/** @brief Structure for SMB2 response create data events.
*/
struct PACE2_smb_response_create {
    struct ipd_string file_id;                      /**< The file id, which can be used to references the accessed file */
    enum pace2_smb2_create_action create_action;    /**< The value describes the result of the previously sent SMB2 request*/
    u64 creation_time;                              /**< The file name Creation Date */
    u64 last_access_time;                           /**< The file name Last Access Date */
    u64 last_write_time;                            /**< The file name Last Write Date */
    u64 last_change_time;                           /**< The file name Last Change Date */
};

/** @brief Structure for SMB2 request read data events.
*/
struct PACE2_smb_request_read {
    u32 read_length;                                /**< The length of the data to be read from file or named pipe */
    u64 file_offset;                                /**< The offset of the data to be read from file or named pipe */
    struct ipd_string file_id;                      /**< The data represented as a pointer to a buffer and a length field containing the SMB file or path name to delete. */
};

/** @brief Structure for SMB2 request write data events.
*/
struct PACE2_smb_request_write {
    struct ipd_string content;                      /**< The content written to file or named pipe */
    struct ipd_string file_id;                      /**< The file id, which can be used to references the accessed file. */
};

/**
 * @brief Structure for session_setup_andx events.
 */
struct PACE2_smb_authentication {
    struct ipd_string domainname;                   /**< Containing NT Domain name. */
    struct ipd_string username;                     /**< Containing User name. */
    struct ipd_string hostname;                     /**< Containing Host name. */
};

/**
 * @brief Structure for SMB2 request tree_connect_andx events.
 */
struct PACE2_smb_request_tree_connect_andx {
    struct ipd_string path;                         /**< The path of the accessed resource.  */
    struct ipd_string service;                      /**< The service of the accessed resource ("A:", "LPT1:", "IPC", "COMM", "????"). */
    struct ipd_string password;                     /**< The password in plain text or authentication response or 0x00 */
};

/**
 * @brief Structure for SMB response tree_connect_andx (SMB) and tree_connect (SMB2) events.
 */
struct PACE2_smb_tree_id {
    u32 smb_tree_id;                                /**< The new given tree_id to the requested resource. */
};

/** @brief Structure for SMB response nt_create andx param events
 *               and for SMB request read andx param events.
 */
struct PACE2_smb_file_id {
    u16 smb_file_id;                                 /**< The File ID of the created or read file. */
};

/**
 * @brief Data structure for advanced SMB events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_ADVANCED_SMB_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_advanced_SMB_event_type meta_data_type;                   /**< SMB event subtype. */
    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;
    union {
        struct PACE2_SMB1_header smb1_header;                       /**< Contains data for a @ref PACE2_ADVANCED_SMB_VERSION1_HEADER event. */
        struct PACE2_SMB2_header smb2_header;                       /**< Contains data for a @ref PACE2_ADVANCED_SMB_VERSION2_HEADER event. */
        char const * invalid;                                       /**< Contains data for a @ref PACE2_ADVANCED_SMB_INVALID event. */
        struct PACE2_SMB_undecoded_data smb_undecoded_parameter;    /**< Contains data for a @ref PACE2_ADVANCED_SMB_UNDECODED_PARAMETER event. */
        struct PACE2_SMB_undecoded_data smb_undecoded_data;         /**< Contains data for a @ref PACE2_ADVANCED_SMB_UNDECODED_DATA event. */
        struct PACE2_smb2_transform_header smb_transform_header;    /**< Contains data for a @ref PACE2_ADVANCED_SMB_VERSION2_TRANSFORM_HEADER event. */
        struct PACE2_smb_encrypted_data smb_encrypted_data;         /**< Contains data for a @ref PACE2_ADVANCED_SMB_ENCRYPTED_DATA event. */

        struct PACE2_smb_request_transaction2 smb_request_transaction2;     /**< Contains data for a @ref PACE2_ADVANCED_SMB_REQUEST_TRANSACTION2_QUERY_PATH and
                                                                                 PACE2_ADVANCED_SMB_REQUEST_TRANSACTION2_FIND_FIRST2 event. */

        struct PACE2_smb_response_transaction2_query_path smb_response_transaction2_query_path;   /**< Contains data for a @ref PACE2_ADVANCED_SMB_RESPONSE_TRANSACTION2_QUERY_PATH event. */

        struct PACE2_smb_response_transaction2_find_first2 smb_response_transaction2_find_first2;  /**< Contains data for a PACE2_ADVANCED_SMB_RESPONSE_TRANSACTION2_FIND_FIRST2 event. */

        struct ipd_string smb_nt_create_andx;                        /**< Contains data for a @ref PACE2_ADVANCED_SMB_REQUEST_NT_CREATE_ANDX event */

        struct PACE2_smb_file_id smb_file_id;                        /**< Contains data for a @ref PACE2_ADVANCED_SMB_RESPONSE_NT_CREATE_ANDX_PARAM event
                                                                                          and @ref PACE2_ADVANCED_SMB_REQUEST_READ_ANDX_PARAM
                                                                                          and @ref PACE2_ADVANCED_SMB_REQUEST_WRITE_ANDX_PARAM event*/

        struct ipd_string smb_request_write_andx;                    /**< Contains data for a @ref PACE2_ADVANCED_SMB_REQUEST_WRITE_ANDX event */

        struct ipd_string smb_response_write_andx;                   /**< Contains data for a @ref PACE2_ADVANCED_SMB_RESPONSE_WRITE_ANDX event */

        struct PACE2_smb_authentication smb_authentication;          /**< Contains data for a @ref PACE2_ADVANCED_SMB_AUTHENTICATION event. */

        struct ipd_string smb_response_read_andx;                    /**< Contains data for a @ref PACE2_ADVANCED_SMB_RESPONSE_READ_ANDX event. */

        struct ipd_string smb_delete_file;                           /**< Contains data for a @ref PACE2_ADVANCED_SMB_DELETE_FILE event. */

        struct PACE2_smb_request_create smb_request_create;          /**< Contains data for a @ref PACE2_ADVANCED_SMB_REQUEST_CREATE event. */

        struct PACE2_smb_response_create smb_response_create;        /**< Contains data for a @ref PACE2_ADVANCED_SMB_RESPONSE_CREATE event. */

        struct PACE2_smb_request_read smb_request_read;              /**< Contains data for a @ref PACE2_ADVANCED_SMB_REQUEST_READ event. */

        struct ipd_string smb_response_read;                         /**< Contains data for a @ref PACE2_ADVANCED_SMB_RESPONSE_READ event. */

        struct PACE2_smb_request_write smb_request_write;            /**< Contains data for a @ref PACE2_ADVANCED_SMB_REQUEST_WRITE event. */

        struct ipd_string smb_request_tree_connect;                  /**< Contains data for a @ref PACE2_ADVANCED_SMB_REQUEST_TREE_CONNECT event. */

        struct PACE2_smb_request_tree_connect_andx smb_request_tree_connect_andx; /**< Contains data for a @ref PACE2_ADVANCED_SMB_REQUEST_TREE_CONNECT_ANDX event. */

        struct PACE2_smb_tree_id smb_response_tree_connect;          /**< Contains data for a @ref PACE2_ADVANCED_SMB_RESPONSE_TREE_CONNECT event. */

    } event_data;                                                    /**< Named union which contains the data of the sub-event types. */
} PACE2_advanced_smb_event;

#endif

#endif /* SMB_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
