/** @file      smtp_events.h
  * @brief     SMTP event definitions.
  * @details   @copydoc doc_smtp
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef SMTP_EVENTS_H
#define SMTP_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "imf_events.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced SMTP event.
 */
#define PACE2_ADVANCED_SMTP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced smtp events.
 */
typedef enum {

    /** This event is sent when the first payload data has been processed and the client and server roles have been
     * determined. The provided data is a pointer to an int that contains the client direction (0 or 1). */
    PACE2_ADVANCED_SMTP_CLIENT_DIRECTION,

    /** This event provides the text of the initial server greeting as a struct ipd_charset_string. It may be
     * generated more than once if the server greeting is a multiline reply. */
    PACE2_ADVANCED_SMTP_SERVER_GREETING,

    /** This event indicates the end of a connection. It does not provide any data. */
    PACE2_ADVANCED_SMTP_CONNECTION_END,

    /** This event is issued when decoding of an SMTP connection fails. It provides a char const * to a message describing the failure reason. */
    PACE2_ADVANCED_SMTP_INVALID,

    /** This event is sent when the client issues an EHLO or HELO command. The argument of the command is provided as a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_EHLO,

    /** This event indicates the start of a TLS connection. */
    PACE2_ADVANCED_SMTP_STARTTLS,

    /** User name (extracted from an AUTH PLAIN or AUTH LOGIN command), given as a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_USER,

    /** Password (extracted from an AUTH PLAIN or AUTH LOGIN command), given as a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_PASS,

    /** Provides a struct ipd_charset_string * with the FROM tag provided by the smtp protocol. */
    PACE2_ADVANCED_SMTP_MAIL_FROM,

    /** Provides a u64 * with the SIZE argument provided by the FROM tag by smtp protocol. */
    PACE2_ADVANCED_SMTP_MAIL_FROM_SIZE,

    /** Provides a struct ipd_charset_string * with the TO tag provided by the smtp protocol. */
    PACE2_ADVANCED_SMTP_RCPT_TO,

    /** This event indicates a SMTP RSET command. */
    PACE2_ADVANCED_SMTP_RESET,

    /** This event is sent when the client issues a QUIT command. */
    PACE2_ADVANCED_SMTP_QUIT,

    /** The server's reply to the most recent command. This event is only emitted if the most recent SMTP
     * command was reported in a (non-IMF) PADE event, and the reply is a completion (ie, 2xx, 4xx, or 5xx)
     * reply. Only the last line of a multiline reply is reported. @c data points to a struct ipd_smtp_reply. */
    PACE2_ADVANCED_SMTP_SERVER_REPLY,

    /** This event indicates that a stream of IMF events for one mail message will follow. */
    PACE2_ADVANCED_SMTP_DATA,

    /** Value of the From: header field of a transmitted mail message; @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_SMTP_IMF_FROM,

    /** Value of the To: header field of a transmitted mail message; @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_SMTP_IMF_TO,

    /** Value of the Cc: header field of a transmitted mail message; @c data points to astruct ipd_imf_address. */
    PACE2_ADVANCED_SMTP_IMF_CC,

    /** Value of the Bcc: header field of a transmitted mail message; @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_SMTP_IMF_BCC,

    /** Value of the Subject: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_SUBJECT,

    /** Value of the Date: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_DATE,

    /** Value of the Sender: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_SMTP_IMF_SENDER,

    /** Value of the Reply-To: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_SMTP_IMF_REPLY_TO,

    /** Value of the Message-Id: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_MESSAGE_ID,

    /** Value of the In-Reply-To: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_IN_REPLY_TO,

    /** Value of the References: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_REFERENCES,

    /** Value of the Comments: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_COMMENTS,

    /** Value of the Keywords: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_KEYWORDS,

    /** Value of the Trace: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_TRACE,

    /** Value of the Resent-Date: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_RESENT_DATE,

    /** Value of the Resent-From: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_SMTP_IMF_RESENT_FROM,

    /** Value of the Resent-Sender: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_SMTP_IMF_RESENT_SENDER,

    /** Value of the Resent-To: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_SMTP_IMF_RESENT_TO,

    /** Value of the Resent-Cc: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_SMTP_IMF_RESENT_CC,

    /** Value of the Resent-Bcc: header field of a transmitted mail message;
     * @c data points to a struct ipd_imf_address. */
    PACE2_ADVANCED_SMTP_IMF_RESENT_BCC,

    /** Value of the Resent-Msg-Id: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_RESENT_MSG_ID,

    /** Value of the MIME-Version: header field of a transmitted mail message;
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_MIME_VERSION,

    /** Complete header field (including field name).
     * Sent for all message or part header fields of a transmitted message not covered by more specific events;
     * @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_SMTP_IMF_OTHER,

    /** The entire header of a mail message; @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_HEADER,

    /** Value of the Content-Type: header field of a message or MIME part;
     * @c data points to a struct ipd_imf_content_type. */
    PACE2_ADVANCED_SMTP_IMF_CONTENT_TYPE,

    /** Value of the Content-Transfer-Encoding: header field of a message or MIME part;
     * @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_SMTP_IMF_CONTENT_TRANSFER_ENCODING,

    /** Value of the Content-Disposition: header field of a MIME part;
     * @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_SMTP_IMF_CONTENT_DISPOSITION,

    /** The entire header of a MIME part; @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_SMTP_IMF_PART_HEADER,

    /** This event indicates the begin of a byte stream (MIME part body, multipart preamble or epilogue);
     * @c data points to a struct ipd_charset_string containing the section number of the part. */
    PACE2_ADVANCED_SMTP_IMF_DATA_BEGIN,

    /** A block of data in a byte stream; @c data points to a struct ipd_imf_part_data. */
    PACE2_ADVANCED_SMTP_IMF_DATA,

    /** This event indicates the end of a byte stream;
     * @c data points to a struct ipd_charset_string containing the section number of the part. */
    PACE2_ADVANCED_SMTP_IMF_DATA_END,

    /** This event indicates the end of a mail message; @c data is a NULL pointer. */
    PACE2_ADVANCED_SMTP_IMF_DONE,

    /** Value of Return-Path: header field of a transmitted mail message; denotes the address where
     * to send error messages; @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_RETURN_PATH,

    /** Value of Received: header field of a transmitted mail message; an array of strings: each hop
     * across the path from sender to receiver adds information about the respective sender.
     * @c data points to a struct ipd_charset_string. */
    PACE2_ADVANCED_SMTP_IMF_RECEIVED,

    /** This event indicates a missing data length associated with a specific MIME section number. */
    PACE2_ADVANCED_SMTP_IMF_DATA_MISSING,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_SMTP_EVENT_TYPES

} PACE2_advanced_smtp_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

struct ipd_smtp_reply {

    /** Complete text of the line sent by the server. */
    struct ipd_charset_string reply_line;

    /** Set to 1 if the reply code is 2xx, 0 otherwise. */
    u8 success;

    /** Set to 1 if this is a single-line reply or the last line of a multi-line response, 0 otherwise. */
    u8 last_line;
};

/**
 * @brief Data structure for advanced SMTP events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_SMTP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** SMTP event subtype. */
    PACE2_advanced_smtp_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_CLIENT_DIRECTION event. */
        int client_direction;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_SERVER_GREETING event. */
        struct ipd_charset_string server_greeting;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_INVALID event. */
        char const * invalid;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_EHLO event. */
        struct ipd_charset_string ehlo;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_USER event. */
        struct ipd_charset_string user;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_PASS event. */
        struct ipd_charset_string pass;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_MAIL_FROM event. */
        struct ipd_charset_string mail_from;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_MAIL_FROM_SIZE event. */
        u64 mail_from_size;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_RCPT_TO event. */
        struct ipd_charset_string rcpt_to;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_SERVER_REPLY event. */
        struct ipd_smtp_reply server_reply;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_FROM event. */
        struct ipd_imf_address imf_from;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_TO event. */
        struct ipd_imf_address imf_to;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_CC event. */
        struct ipd_imf_address imf_cc;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_BCC event. */
        struct ipd_imf_address imf_bcc;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_SUBJECT event. */
        struct ipd_charset_string imf_subject;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_DATE event. */
        struct ipd_charset_string imf_date;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_SENDER event. */
        struct ipd_imf_address imf_sender;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_REPLY_TO event. */
        struct ipd_imf_address imf_reply_to;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_MESSAGE_ID event. */
        struct ipd_charset_string imf_message_id;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_IN_REPLY_TO event. */
        struct ipd_charset_string imf_in_reply_to;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_REFERENCES event. */
        struct ipd_charset_string imf_references;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_COMMENTS event. */
        struct ipd_charset_string imf_comments;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_KEYWORDS event. */
        struct ipd_charset_string imf_keywords;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_TRACE event. */
        struct ipd_charset_string imf_trace;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_RESENT_DATE event. */
        struct ipd_charset_string imf_resent_date;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_RESENT_FROM event. */
        struct ipd_imf_address imf_resent_from;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_RESENT_SENDER event. */
        struct ipd_imf_address imf_resent_sender;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_RESENT_TO event. */
        struct ipd_imf_address imf_resent_to;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_RESENT_CC event. */
        struct ipd_imf_address imf_resent_cc;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_RESENT_BCC event. */
        struct ipd_imf_address imf_resent_bcc;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_RESENT_MSG_ID event. */
        struct ipd_charset_string imf_resent_msg_id;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_MIME_VERSION event. */
        struct ipd_charset_string imf_mime_version;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_OTHER event. */
        struct ipd_imf_part_data imf_other;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_HEADER event. */
        struct ipd_charset_string imf_header;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_CONTENT_TYPE event. */
        struct ipd_imf_content_type imf_content_type;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_CONTENT_TRANSFER_ENCODING event. */
        struct ipd_imf_part_data imf_content_transfer_encoding;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_CONTENT_DISPOSITION event. */
        struct ipd_imf_part_data imf_content_disposition;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_PART_HEADER event. */
        struct ipd_imf_part_data imf_part_header;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_DATA_BEGIN event. */
        struct ipd_charset_string imf_data_begin;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_DATA event. */
        struct ipd_imf_part_data  imf_data;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_DATA_END event. */
        struct ipd_charset_string imf_data_end;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_RETURN_PATH event. */
        struct ipd_charset_string imf_return_path;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_RECEIVED event. */
        struct ipd_charset_string imf_received;

        /** Contains data for a @ref PACE2_ADVANCED_SMTP_IMF_DATA_MISSING event. */
        struct ipd_imf_missing_data imf_data_missing;

    } event_data;

} PACE2_advanced_smtp_event;

#endif

#endif /* SMTP_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
