/** @file      socks_events.h
  * @brief     SOCKS event definitions.
  * @details   @copydoc doc_socks
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef SOCKS_EVENTS_H
#define SOCKS_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Group ******************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced SOCKS event.
 */
#define PACE2_ADVANCED_SOCKS_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/
/**
 * @brief Types for advanced socks events.
 */
typedef enum {

    /**
     * New SOCKS connection is observed.
     * NULL.
     */
    PACE2_ADVANCED_SOCKS_NEW_CONNECTION,

    /**
     * SOCKS v4, v4a, or v5 request.
     * struct ipd_socks_request * : client request.
     */
    PACE2_ADVANCED_SOCKS_REQUEST,

    /**
     * SOCKS v4, v4a, or v5 response.
     * struct ipd_socks_response * : server response.
     */
    PACE2_ADVANCED_SOCKS_RESPONSE,

    /**
     * Supported authentication method announced by client.
     * struct ipd_socks_v5_authentication_method * : supported authentication method.
     */
    PACE2_ADVANCED_SOCKS_V5_CLIENT_ANNOUNCEMENT,

    /**
     * Authentication method selected by server.
     * struct ipd_socks_v5_authentication_method * : selected authentication method.
     */
    PACE2_ADVANCED_SOCKS_V5_SERVER_ANNOUNCEMENT_RESPONSE,

    /**
     * SOCKS v5 username authentication request.
     * struct ipd_socks_v5_username_authentication_request * : client request.
     */
    PACE2_ADVANCED_SOCKS_V5_USERNAME_AUTHENTICATION_REQUEST,

    /**
     * SOCKS v5 username authentication response.
     * struct ipd_socks_v5_username_authentication_response * : server response.
     */
    PACE2_ADVANCED_SOCKS_V5_USERNAME_AUTHENTICATION_RESPONSE,

    /**
     * SOCKS connection ends.
     * NULL.
     */
    PACE2_ADVANCED_SOCKS_END_OF_FLOW,

    /**
     * Decoding of a SOCKS connection fails.
     * char const * : message describing the failure reason.
     */
    PACE2_ADVANCED_SOCKS_INVALID,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_SOCKS_EVENT_TYPES

} PACE2_advanced_socks_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/**
 * SOCKS versions.
 */
enum ipd_socks_version_type {
    IPD_SOCKS_VERSION_4,
    IPD_SOCKS_VERSION_4A,
    IPD_SOCKS_VERSION_5,
};

/**
 * Address types.
 */
enum ipd_socks_address_type {
    IPD_SOCKS_ADDRESS_TYPE_IPv4,
    IPD_SOCKS_ADDRESS_TYPE_IPv6,
    IPD_SOCKS_ADDRESS_TYPE_DOMAIN,
};

/**
 * Types of SOCKS request messages.
 */
enum ipd_socks_request_message_type {
    IPD_SOCKS_REQUEST_MESSAGE_ESTABLISH_TCP_IP_STREAM_CONNECTION,
    IPD_SOCKS_REQUEST_MESSAGE_ESTABLISH_TCP_IP_PORT_BINDING,
    IPD_SOCKS_REQUEST_MESSAGE_ASSOCIATE_UDP_PORT,
};

/**
 * Types of SOCKS response messages.
 */
enum ipd_socks_response_message_type {
    IPD_SOCKS_RESPONSE_MESSAGE_REQUEST_GRANTED,
    IPD_SOCKS_RESPONSE_MESSAGE_REQUEST_REJECTED_OR_FAILED,
    IPD_SOCKS_RESPONSE_MESSAGE_REQUEST_FAILED_BECAUSE_CLIENT_IS_NOT_RUNNING_IDENTD_OR_NOT_REACHABLE,
    IPD_SOCKS_RESPONSE_MESSAGE_REQUEST_FAILED_BECAUSE_CLIENT_IDENTD_COULD_NOT_CONFIRM_USER_ID_STRING_IN_REQUEST,
    IPD_SOCKS_RESPONSE_MESSAGE_GENERAL_FAILURE,
    IPD_SOCKS_RESPONSE_MESSAGE_CONNECTION_NOT_ALLOWED_BY_RULESET,
    IPD_SOCKS_RESPONSE_MESSAGE_NETWORK_UNREACHABLE,
    IPD_SOCKS_RESPONSE_MESSAGE_HOST_UNREACHABLE,
    IPD_SOCKS_RESPONSE_MESSAGE_CONNECTION_REFUSED_BY_DESTINATION_HOST,
    IPD_SOCKS_RESPONSE_MESSAGE_TTL_EXPIRED,
    IPD_SOCKS_RESPONSE_MESSAGE_COMMAND_NOT_SUPPORTED_OR_PROTOCOL_ERROR,
    IPD_SOCKS_RESPONSE_MESSAGE_ADDRESS_TYPE_NOT_SUPPORTED,
};

/**
 * Types of SOCKS authentication mechanisms.
 */
enum ipd_socks_authentication_type {
    IPD_SOCKS_AUTHENTICATION_NO_AUTHENTICATION,
    IPD_SOCKS_AUTHENTICATION_GSSAPI,
    IPD_SOCKS_AUTHENTICATION_USERNAME_PASSWORD,
    IPD_SOCKS_AUTHENTICATION_METHOD_ASSIGNED_BY_IANA,
    IPD_SOCKS_AUTHENTICATION_METHOD_RESERVED_FOR_PRIVATE_USE,
};


/**
 * SOCKS v4 and v5 requests.
 */
struct ipd_socks_request {

    /** SOCKS protocol version. */
    enum ipd_socks_version_type socks_version;

    /** Type of the SOCKS request. */
    enum ipd_socks_request_message_type request_type;

    union {
        struct {
            /** Port to which the user wants to connect. */
            u16 port;

            /** IPv4 address to which the user wants to connect. */
            struct in_addr ipv4_address;

            /** ID specified by the user connecting to the SOCKS proxy. */
            struct ipd_charset_string user_id;
        } v4;

        struct {
            /** Port to which the user wants to connect. */
            u16 port;

            /** ID specified by the user connecting to the SOCKS proxy. */
            struct ipd_charset_string user_id;

            /** Domain name to which the user wants to connect. */
            struct ipd_charset_string domain_name;
        } v4a;

        struct {
            /** Port to which the user wants to connect. */
            u16 port;

            /** Type of the address. */
            enum ipd_socks_address_type address_type;

            union {
                /** IPv4 address to which the user wants to connect. */
                struct in_addr ipv4_address;

                /** Domain name to which the user wants to connect. */
                struct ipd_charset_string domain_name;

                /** IPv6 address to which the user wants to connect. */
                struct in6_addr ipv6_address;
            } address;
        } v5;
    };
};

/**
 * SOCKS v4 and v5 responses.
 */
struct ipd_socks_response {

    /** SOCKS protocol version. */
    enum ipd_socks_version_type socks_version;

    /** Type of the SOCKS response. */
    enum ipd_socks_response_message_type response_type;

    union {
        struct {
            /** Bind port. */
            u16 port;

            /** Bind IPv4 address. */
            struct in_addr ipv4_address;
        } v4a;

        struct {
            /** Bind port. */
            u16 port;

            /** Type of the address. */
            enum ipd_socks_address_type address_type;

            union {
                /** Bind IPv4 address. */
                struct in_addr ipv4_address;

                /** Bind IPv6 address. */
                struct in6_addr ipv6_address;

                /** Bind domain name. */
                struct ipd_charset_string domain_name;
            } address;
        } v5;
    };
};

/**
 * Thrown SOCKS v5 authentication method.
 */
struct ipd_socks_v5_authentication_method {

    /** Authentication method. */
    enum ipd_socks_authentication_type authentication_method;
};

/**
 * SOCKS v5 user name authentication request.
 */
struct ipd_socks_v5_username_authentication_request {

    /** Request version number. */
    u8 request_version_number;

    /** User name. */
    struct ipd_charset_string username;

    /** Password. */
    struct ipd_charset_string password;
};

/**
 * SOCKS v5 user name authentication response.
 */
struct ipd_socks_v5_username_authentication_response {

    /** Response version number. */
    u8 response_version_number;

    /** Type of the SOCKS response. */
    enum ipd_socks_response_message_type response_type;
};

/**
 * @brief Data structure for advanced SOCKS events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_SOCKS_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** SOCKS event subtype. */
    PACE2_advanced_socks_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_SOCKS_REQUEST event. */
        struct ipd_socks_request socks_request;

        /** Contains data for a @ref PACE2_ADVANCED_SOCKS_RESPONSE event. */
        struct ipd_socks_response ipd_socks_response;

        /** Contains data for a @ref PACE2_ADVANCED_SOCKS_V5_CLIENT_ANNOUNCEMENT event. */
        struct ipd_socks_v5_authentication_method v5_authentication_method_announcement;

        /** Contains data for a @ref PACE2_ADVANCED_SOCKS_V5_SERVER_ANNOUNCEMENT_RESPONSE event. */
        struct ipd_socks_v5_authentication_method v5_authentication_method_response;

        /** Contains data for a @ref PACE2_ADVANCED_SOCKS_V5_USERNAME_AUTHENTICATION_REQUEST event. */
        struct ipd_socks_v5_username_authentication_request v5_username_authentication_request;

        /** Contains data for a @ref PACE2_ADVANCED_SOCKS_V5_USERNAME_AUTHENTICATION_RESPONSE event. */
        struct ipd_socks_v5_username_authentication_response v5_username_authentication_response;

        /** Contains data for a @ref PACE2_ADVANCED_SOCKS_INVALID event. */
        char const * invalid;

    } event_data;

} PACE2_advanced_socks_event;


#endif

#endif /* SOCKS_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
