/** @file      ssh_events.h
  * @brief     SSH event definitions.
  * @details   @copydoc doc_ssh
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef SSH_EVENTS_H
#define SSH_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced SSH event.
 */
#define PACE2_ADVANCED_SSH_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced ssh events.
 */
typedef enum {
    PACE2_ADVANCED_SSH_VERSION,                             /**< SSH Version, including additional comments as plain text SSH-2.0-OpenSSH_4.6p1 Debian-5ubuntu0.1. */
    PACE2_ADVANCED_SSH_INVALID,                             /**< This event is issued when decoding of an SSH connection fails. It provides a const char * to a message describing the failure reason. */
    PACE2_ADVANCED_SSH_SEND_ENCRYPTED_DATA,                 /**< Send octets of encrypted packet data in hex format with spaces. */
    PACE2_ADVANCED_SSH_KEXINIT_MESSAGE_NUMBER,              /**< Returns the message number of this type of transaction. */
    PACE2_ADVANCED_SSH_KEXINIT_KEX_COOKIE,                  /**< Key Exchange (kex) cookie in hex format with spaces. */
    PACE2_ADVANCED_SSH_KEXINIT_KEX_ALGORITHM,               /**< List of offered kex algorithms as plain text. */
    PACE2_ADVANCED_SSH_KEXINIT_SERVER_HOST_KEY_ALGORITHM,                      /**< List of server host key algorithm as plain text. */
    PACE2_ADVANCED_SSH_KEXINIT_ENCRYPTION_ALGORITHM_CLIENT_TO_SERVER,          /**< List of encryption algorithm client to server as plain text. */
    PACE2_ADVANCED_SSH_KEXINIT_ENCRYPTION_ALGORITHM_SERVER_TO_CLIENT,          /**< List of encryption algorithm server to client as plain text. */
    PACE2_ADVANCED_SSH_KEXINIT_MAC_ALGORITHM_CLIENT_TO_SERVER,                 /**< List of Message Authentication Code (mac) algorithm client to server as plain text. */
    PACE2_ADVANCED_SSH_KEXINIT_MAC_ALGORITHM_SERVER_TO_CLIENT,                 /**< List of Message Authentication Code (mac) algorithm server to client as plain text. */
    PACE2_ADVANCED_SSH_KEXINIT_COMPRESSION_ALGORITHM_CLIENT_TO_SERVER,         /**< List of compression algorithm client to_server as plain text. */
    PACE2_ADVANCED_SSH_KEXINIT_COMPRESSION_ALGORITHM_SERVER_TO_CLIENT,         /**< List of compression algorithm server client as plain text. */
    PACE2_ADVANCED_SSH_KEXINIT_LANGUAGES_CLIENT_TO_SERVER,  /**< List of languages client to server as plain text. */
    PACE2_ADVANCED_SSH_KEXINIT_LANGUAGES_SERVER_TO_CLIENT,  /**< List of languages client to server as plain text. */
    PACE2_ADVANCED_SSH_KEXINIT_KEX_FIRST_PACKET_FOLLOWS,    /**< Flag named KEX_First_Packet_Follows as hex, indicates if a guessed key exchange packet follows. */
    PACE2_ADVANCED_SSH_KEXINIT_RESERVED_4_OCTETS,           /**< Four reserved octets. */
    PACE2_ADVANCED_SSH_DH_KEXINIT_MESSAGE_NUMBER,           /**< Message number of key exchange init as plain text. */
    PACE2_ADVANCED_SSH_DH_KEXINIT_RFC4253_VALUE_OF_E,       /**< RFC4253_value_of e in hex. */
    PACE2_ADVANCED_SSH_DH_KEX_REPLY_MESSAGE_NUMBER,         /**< Message number of DH_KEX_REPLY_MESSAGE_NUMBER packet.  */
    PACE2_ADVANCED_SSH_DH_KEX_REPLY_RFC4253_SERVER_PUBLIC_HOST_KEY_AND_CERTIFICATES_K_S, /**< RFC4253 p 22. server public host key and certificates (K_S) as hex. */
    PACE2_ADVANCED_SSH_DH_KEX_REPLY_RFC4253_VALUE_OF_F,     /**< RFC4253 p. 22, value of f. */
    PACE2_ADVANCED_SSH_DH_KEX_REPLY_RFC4253_VALUE_OF_H,     /**< RFC4253 p. 22, value of h. */
    PACE2_ADVANCED_SSH_NEWKEYS_MESSAGE_NUMBER,              /**< Message number New Keys. */
    PACE2_ADVANCED_SSH_NEWKEYS_SSH_NEGOTIATION_ENDS_WITH_NEW_KEYS,             /**< End of negotiation of the partners in plain text. */
    PACE2_ADVANCED_SSH_KEX_DH_GEX_REQUEST_MESSAGE_NUMBER,   /**< Message number Diffie-Hellman Group exchange request. */
    PACE2_ADVANCED_SSH_KEX_DH_GEX_REQUEST_RFC4419_MIN_MINIMAL_SIZE_IN_BITS_OF_AN_ACCEPTABLE_GROUP, /**< Diffie-Hellman Group and Key Exchange Request, min size of an acceptable group. */
    PACE2_ADVANCED_SSH_KEX_DH_GEX_REQUEST_RFC4419_N_PREFERRED_SIZE_IN_BITS_OF_THE_GROUP_THE_SERVER_WILL_SEND, /**< Diffie-Hellman Group and Key Exchange Request, preferred size of a group  the server will send. */
    PACE2_ADVANCED_SSH_KEX_DH_GEX_REQUEST_RFC4419_MAX_MAXIMAL_SIZE_IN_BITS_OF_AN_ACCEPTABLE_GROUP,/**< Diffie-Hellman Group and Key Exchange Request, max size of an acceptable group. */
    PACE2_ADVANCED_SSH_DH_GEX_GROUP_MESSAGE_NUMBER,         /**< Message number Diffie-Hellman Group/ Key exchange. */
    PACE2_ADVANCED_SSH_DH_GEX_GROUP_RFC4419_P_SAFE_PRIME,   /**< Diffie-Hellman Group exchange safe prime number. */
    PACE2_ADVANCED_SSH_GEX_GROUP_RFC4419_G_GENERATOR_FOR_SUBGROUP_IN_GF_P,      /**< Diffie-Hellman Group/ Key exchange generator g number. */
    PACE2_ADVANCED_SSH_KEX_DH_GEX_INIT_MESSAGE_NUMBER,      /**< Message number DH Group/ Key Exchange Init. */
    PACE2_ADVANCED_SSH_KEX_DH_GEX_INIT_RFC4419_VALUE_OF_E,  /**< RFC4419, p.4 value of e. */
    PACE2_ADVANCED_SSH_KEX_DH_GEX_REPLY_MESSAGE_NUMBER,     /**< Message number DH Group/ Key exchange Reply. */
    PACE2_ADVANCED_SSH_KEX_DH_GEX_REPLY_RFC4419_DH_HOST_KEY,       /**< RFC4419, p.4, DH Group/ Key exchange Reply Host Key. */
    PACE2_ADVANCED_SSH_KEX_DH_GEX_REPLY_RFC4419_VALUE_OF_F,        /**< RFC4419, p.4, value of f. */
    PACE2_ADVANCED_SSH_KEX_DH_GEX_REPLY_RFC4419_SIGNATURE_OF_H,    /**< RFC4419, p.4, signature of H. */
    PACE2_ADVANCED_SSH_CONNECTION_END,                     /**< This is the connection end event. */
    PACE2_ADVANCED_SSH_NEW_CONNECTION,                     /**< This is the new connection event. */
    PACE2_ADVANCED_SSH_MSG_IGNORE_MESSAGE_NUMBER,          /**< MSG Ignore Message Number. */
    PACE2_ADVANCED_SSH_MSG_IGNORE_DATA,                    /**< MSG Ignore data (optional, means mostly empty). */
    PACE2_ADVANCED_SSH_KEX_DH_GEX_REQUEST_OLD,             /**< RFC4419, chapter 5,  for backward compatibility, KEX_DH_GEX_REQUEST_OLD. */
    PACE2_ADVANCED_SSH_SEND_ENCRYPTED_DATA_MISSING,        /**< Informs about the length and direction of a gap in the transferred data due to lost TCP segments. For completeness and
                                                            *   reliability, this event will also be thrown if data is missing during the initial key exchange */

    PACE2_NUMBER_OF_ADVANCED_SSH_EVENT_TYPES               /**< Number of implemented event types. */
} PACE2_advanced_ssh_event_type;

/* Event Structures *************************************************************/


#ifndef PACE_KERNEL_LIB

/**
 * @brief Data structures for ssh event named PACE2_ADVANCED_SSH_SEND_ENCRYPTED_DATA
 */
typedef struct pace2_ssh_encrypted {
    struct ipd_string data; /**< The encrypted data */
    u8 direction;           /**< The direction of the data (0 or 1) */
} PACE2_ssh_encrypted;

/**
 * @brief Data structures for ssh event named @ref PACE2_ADVANCED_SSH_SEND_ENCRYPTED_DATA_MISSING
 */
typedef struct pace2_ssh_missing_data {
    u64 length;     /**< The length of the missing data */
    u8 direction;   /**< The direction of the data (0 or 1) */
} PACE2_ssh_missing_data;

/**
 * @brief Data structure for ssh event named PACE2_ADVANCED_SSH_VERSION
 */
typedef struct pace2_ssh_version_message {
    struct ipd_string data; /**< The content of the version message */
    u8 direction;           /**< The direction of the data (0 or 1) */
    u8 is_client_to_server; /**< Information whether the message is from client to server or vice versa.
                                 ("0": server to client; "1": client to server) */
} PACE2_ssh_version_message;

/**
 * @brief Data structure for advanced ssh events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_ADVANCED_SSH_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_advanced_ssh_event_type meta_data_type;             /**< SSH event subtype. */
    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;
    union {
        PACE2_ssh_version_message version;                    /**< Contains data for a @ref PACE2_ADVANCED_SSH_VERSION event. */
        struct ipd_string invalid;                            /**< Contains data for a @ref PACE2_ADVANCED_SSH_INVALID event. */
        PACE2_ssh_encrypted send_encrypted_data;              /**< Contains data for a @ref PACE2_ADVANCED_SSH_SEND_ENCRYPTED_DATA event. */
        struct ipd_string kexinit_message_number;             /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_MESSAGE_NUMBER event. */
        struct ipd_string kexinit_kex_cookie;                 /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_KEX_COOKIE event. */
        struct ipd_string kexinit_kex_algorithm;              /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_KEX_ALGORITHM event. */
        struct ipd_string kexinit_server_host_key_algorithm;  /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_SERVER_HOST_KEY_ALGORITHM event. */
        struct ipd_string kexinit_encryption_algorithm_client_to_server;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_ENCRYPTION_ALGORITHM_CLIENT_TO_SERVER event. */
        struct ipd_string kexinit_encryption_algorithm_server_to_client;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_ENCRYPTION_ALGORITHM_SERVER_TO_CLIENT event. */
        struct ipd_string kexinit_mac_algorithm_client_to_server;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_MAC_ALGORITHM_CLIENT_TO_SERVER EVENT. */
        struct ipd_string kexinit_mac_algorithm_server_to_client;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_MAC_ALGORITHM_SERVER_TO_CLIENT event. */
        struct ipd_string kexinit_compression_algorithm_client_to_server;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_COMPRESSION_ALGORITHM_CLIENT_TO_SERVER event. */
        struct ipd_string kexinit_compression_algorithm_server_to_client;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_COMPRESSION_ALGORITHM_SERVER_TO_CLIENT event. */
        struct ipd_string kexinit_languages_client_to_server;  /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_LANGUAGES_CLIENT_TO_SERVER event. */
        struct ipd_string kexinit_languages_server_to_client; /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_LANGUAGES_SERVER_TO_CLIENT event. */
        struct ipd_string kexinit_kex_first_packet_follows;    /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_KEX_FIRST_PACKET_FOLLOWS event. */
        struct ipd_string kexinit_reserved_4_octets;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEXINIT_RESERVED_4_OCTETS event. */
        struct ipd_string dh_kexinit_message_number;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_DH_KEXINIT_MESSAGE_NUMBER event. */
        struct ipd_string dh_kexinit_rfc4253_value_of_e; /**< Contains data for a @ref PACE2_ADVANCED_SSH_DH_KEXINIT_RFC4253_VALUE_OF_E event. */
        struct ipd_string dh_kex_reply_message_number;   /**< Contains data for a @ref PACE2_ADVANCED_SSH_DH_KEX_REPLY_MESSAGE_NUMBER event. */
        struct ipd_string dh_kex_reply_rfc4253_server_public_host_key_and_certificates_k_s;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_DH_KEX_REPLY_RFC4253_SERVER_PUBLIC_HOST_KEY_AND_CERTIFICATES_K_S event. */
        struct ipd_string dh_kex_reply_rfc4253_value_of_f;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_DH_KEX_REPLY_RFC4253_VALUE_OF_F event. */
        struct ipd_string dh_kex_reply_rfc4253_value_of_h;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_DH_KEX_REPLY_RFC4253_VALUE_OF_H event. */
        struct ipd_string newkeys_message_number;       /**< Contains data for a @ref PACE2_ADVANCED_SSH_NEWKEYS_MESSAGE_NUMBER event. */
        struct ipd_string newkeys_ssh_negotiation_ends_with_new_keys;         /**< Contains data for a @ref PACE2_ADVANCED_SSH_NEWKEYS_SSH_NEGOTIATION_ENDS_WITH_NEW_KEYS event. */
        struct ipd_string kex_dh_gex_request_message_number;   /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEX_DH_GEX_REQUEST_MESSAGE_NUMBER event. */
        struct ipd_string kex_dh_gex_request_rfc4419_min_minimal_size_in_bits_of_an_acceptable_group;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEX_DH_GEX_REQUEST_RFC4419_MIN_MINIMAL_SIZE_IN_BITS_OF_AN_ACCEPTABLE_GROUP event. */
        struct ipd_string kex_dh_gex_request_rfc4419_n_preferred_size_in_bits_of_the_group_the_server_will_send;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEX_DH_GEX_REQUEST_RFC4419_N_PREFERRED_SIZE_IN_BITS_OF_THE_GROUP_THE_SERVER_WILL_SEND event. */
        struct ipd_string kex_dh_gex_request_rfc4419_max_maximal_size_in_bits_of_an_acceptable_group;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEX_DH_GEX_REQUEST_RFC4419_MAX_MAXIMAL_SIZE_IN_BITS_OF_AN_ACCEPTABLE_GROUP event. */
        struct ipd_string dh_gex_group_message_number;       /**< Contains data for a @ref PACE2_ADVANCED_SSH_DH_GEX_GROUP_MESSAGE_NUMBER event. */
        struct ipd_string dh_gex_group_rfc4419_p_safe_prime; /**< Contains data for a @ref PACE2_ADVANCED_SSH_DH_GEX_GROUP_RFC4419_P_SAFE_PRIME event. */
        struct ipd_string gex_group_rfc4419_g_generator_for_subgroup_in_gf_p;     /**< Contains data for a @ref PACE2_ADVANCED_SSH_GEX_GROUP_RFC4419_G_GENERATOR_FOR_SUBGROUP_IN_GF_P event. */
        struct ipd_string kex_dh_gex_init_message_number;    /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEX_DH_GEX_INIT_MESSAGE_NUMBER event. */
        struct ipd_string kex_dh_gex_init_rfc4419_value_of_e;/**< Contains data for a @ref PACE2_ADVANCED_SSH_KEX_DH_GEX_INIT_RFC4419_VALUE_OF_E event. */
        struct ipd_string kex_dh_gex_reply_message_number;   /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEX_DH_GEX_REPLY_MESSAGE_NUMBER event. */
        struct ipd_string kex_dh_gex_reply_rfc4419_dh_host_key; /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEX_DH_GEX_REPLY_RFC4419_DH_HOST_KEY event. */
        struct ipd_string kex_dh_gex_reply_rfc4419_value_of_f;  /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEX_DH_GEX_REPLY_RFC4419_VALUE_OF_F event. */
        struct ipd_string kex_dh_gex_reply_rfc4419_signature_of_h; /**< Contains data for a @ref PACE2_ADVANCED_SSH_KEX_DH_GEX_REPLY_RFC4419_SIGNATURE_OF_H event. */
        struct ipd_string connection_end;      /**< Contains data for a @ref PACE2_ADVANCED_SSH_CONNECTION_END event. */
        struct ipd_string new_connection;      /**< Contains data for a @ref PACE2_ADVANCED_SSH_NEW_CONNECTION event. */
        struct ipd_string msg_ignore_message_number;  /**< Contains data for MSG Ignore Message Number. */
        struct ipd_string msg_ignore_data;     /**<  Contains data for MSG Ignore data (optional). */
        struct ipd_string kex_dh_gex_request_old;  /**< Contains data for RFC4419, chapter 5,  for backward compatibility, KEX_DH_GEX_REQUEST_OLD. */
        PACE2_ssh_missing_data encrypted_data_missing; /**< Contains the data for a @ref PACE2_ADVANCED_SSH_SEND_ENCRYPTED_DATA_MISSING event. */
    } event_data;                                             /**< Named union that contains the data of the sub-event types. */
} PACE2_advanced_ssh_event;

#endif

#endif /* SSH_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
