/** @file      ssl_events.h
  * @brief     SSL event definitions.
  * @details   @copydoc doc_ssl
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef SSL_EVENTS_H
#define SSL_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

#ifndef PACE_KERNEL_LIB
#include <time.h>
#endif

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the basic SSL client hello event.
 */
#define PACE2_BASIC_SSL_CLIENT_HELLO_EVENT_GROUP PACE2_BASIC_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned metadata group of the basic SSL server hello event.
 */
#define PACE2_BASIC_SSL_SERVER_HELLO_EVENT_GROUP PACE2_BASIC_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned metadata group of the basic SSL dns alt names event.
 */
#define PACE2_BASIC_SSL_DNS_ALT_NAMES_EVENT_GROUP PACE2_BASIC_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned metadata group of the basic SSL session ticket event.
 */
#define PACE2_BASIC_SSL_SESSION_TICKET_EVENT_GROUP PACE2_BASIC_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned meta data group of the advanced SSL event.
 */
#define PACE2_ADVANCED_SSL_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned meta data group of the class SSL event.
 */
#define PACE2_CLASS_SSL_EVENT_GROUP PACE2_CLASS_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Meta data flags of SSL client hello events.
 */
typedef enum {
    PACE2_SSL_SERVER_NAME = 0,              /**< Server name line. */
    PACE2_SSL_CLIENT_SESSION_ID,            /**< Session ID line. */
    PACE2_SSL_CLIENT_VERSION,               /**< Client version line. */
    PACE2_SSL_CLIENT_CIPHER_SUITES,         /**< Cipher suites line. */
    PACE2_SSL_CLIENT_SERVER_NAME_TYPE,      /**< Server name type line. */
    PACE2_SSL_CLIENT_COMPRESSION_METHODS,   /**< Compression line. */
    PACE2_SSL_CLIENT_SIGNATURE_ALGORITHMS,  /**< Signature algorithms line. */
    PACE2_SSL_CLIENT_ELLIPTIC_CURVES,       /**< Elliptic curves line. */
    PACE2_SSL_CLIENT_EC_POINT_FORMATS,      /**< Elliptic curves point formats. */
    PACE2_SSL_CLIENT_SUPPORTED_VERSIONS,    /**< Supported versions list */

    PACE2_NUMBER_OF_CLIENT_HELLO_LINES  /**< Number of implemented client hello lines. */
} PACE2_SSL_client_hello_line;

/**
 * @brief Meta data flags of SSL server hello events.
 */
typedef enum {
    PACE2_SSL_SERIAL_NUMBER = 0,            /**< Serial number line. */
    PACE2_SSL_VALIDITY_NOT_BEFORE,          /**< Validity-not-before line. */
    PACE2_SSL_VALIDITY_NOT_AFTER,           /**< Validity-not-after line. */
    PACE2_SSL_COUNTRY_NAME,                 /**< Country name line. */
    PACE2_SSL_POSTAL_CODE,                  /**< Postal code line. */
    PACE2_SSL_STATE_OR_PROVINCE_NAME,       /**< State name line. */
    PACE2_SSL_LOCALITY_NAME,                /**< Locality name line. */
    PACE2_SSL_STREET_ADDRESS,               /**< Street address line. */
    PACE2_SSL_ORGANIZATION_NAME,            /**< Organization name line. */
    PACE2_SSL_ORGANIZATIONAL_UNIT_NAME,     /**< Organizational unit name line. */
    PACE2_SSL_COMMON_NAME,                  /**< Common name line. */
    PACE2_SSL_SERVER_SESSION_ID,            /**< Session ID line. */
    PACE2_SSL_CIPHER_SUITE,                 /**< SSL Cipher Suite. */
    PACE2_SSL_VERSION,                      /**< SSL Version line. */
    PACE2_SSL_COMPRESSION,                  /**< SSL Compression line. */
    PACE2_SSL_SKE_CURVE_TYPE,               /**< SSL Server Key Exchange Curve Type line. */
    PACE2_SSL_SKE_NAMED_CURVE,              /**< SSL Server Key Exchange Named Curve line. */
    PACE2_SSL_SKE_PUBLIC_KEY_LENTH,         /**< SSL Server Key Exchange Public Key Length line. */
    PACE2_SSL_SKE_SIG_HASH_ALGORITHM_HASH,  /**< SSL Server Key Exchange Signature Hash Algorithm Hash line. */
    PACE2_SSL_SKE_SIG_HASH_ALGORITHM_SIG,   /**< SSL Server Key Exchange Signature Hash Algorithm Signature line. */
    PACE2_SSL_SKE_SIG_KEY_LENGTH,           /**< SSL Server Key Exchange Signature Key length. */
    PACE2_SSL_CERT_VERSION,                 /**< SSL Certificate Version line. */
    PACE2_SSL_ISSUER_COMMON_NAME,           /**< SSL issuer common name. */
    PACE2_SSL_ISSUER_COUNTRY_NAME,          /**< SSL issuer country name. */
    PACE2_SSL_ISSUER_ORGANIZATION_NAME,     /**< SSL issuer organization name. */
    PACE2_SSL_ISSUER_ORGANIZATIONAL_UNIT_NAME,  /**< SSL issuer organizational unit name. */
    PACE2_SSL_ISSUER_STATE_OR_PROVINCE_NAME,/**< SSL issuer state or province name. */
    PACE2_SSL_ISSUER_LOCALITY_NAME,         /**< SSL issuer locality name. */
    PACE2_SSL_EXT_SUPPORTED_VERSIONS,       /**< SSL supported versions extension */

    PACE2_NUMBER_OF_SERVER_HELLO_LINES  /**< Number of implemented server hello lines. */
} PACE2_SSL_server_hello_line;

/**
 * @brief Meta data flags of SSL session ticket events.
 */
typedef enum {
    PACE2_SSL_SESSION_TICKET_IN_CLIENT_HELLO = 0,   /**< Session ticket within client hello. */
    PACE2_SSL_SESSION_TICKET_IN_SERVER_HELLO,       /**< Session ticket within server hello. */
    PACE2_SSL_NEW_SESSION_TICKET,                   /**< New session ticket. */

    PACE2_NUMBER_OF_SESSION_TICKET_SOURCES          /**< Number of implemented session ticket sources. */
} PACE2_SSL_session_ticket_source;

#ifndef PACE_KERNEL_LIB

/**
 * @brief Types of class SSL events.
 */
typedef enum {
    PACE2_CLASS_SSL_CLIENT_HELLO,                   /**< SSL handshake type "client hello". */
    PACE2_CLASS_SSL_SERVER_HELLO,                   /**< SSL handshake type "server hello". */
    PACE2_CLASS_SSL_CERTIFICATE,                    /**< SSL handshake type "certificate". */
    PACE2_CLASS_SSL_SERVER_HELLO_DONE,              /**< SSL handshake type "server hello done". @note The event is currently unused. */
    PACE2_CLASS_SSL_SERVER_KEY_EXCHANGE,            /**< SSL handshake type "server key exchange (Elliptic Curve Diffie-Hellman)". */
    PACE2_CLASS_SSL_REQUEST_ENCRYPTED_DATA,         /**< Request encrypted data. */
    PACE2_CLASS_SSL_RESPONSE_ENCRYPTED_DATA,        /**< Response encrypted data. */
    PACE2_CLASS_SSL_SERVER_KEY_EXCHANGE_RSA_DSS,    /**< SSL handshake type "server key exchange (RSA, DSS)". */
    PACE2_CLASS_SSL_SERVER_KEY_EXCHANGE_ANON,       /**< SSL handshake type "server key exchange (ANON)". */

    PACE2_NUMBER_OF_CLASS_SSL_EVENT_TYPES           /**< Number of implemented SSL class events. */
} PACE2_class_SSL_event_type;

/**
 * @brief Source of the TLS Certificate Message
 */
typedef enum {
    PACE2_SSL_UNKNOWN_CERTIFICATE_SOURCE,
    PACE2_SSL_CLIENT_CERTIFICATE,
    PACE2_SSL_SERVER_CERTIFICATE,

    PACE2_NUMBER_OF_SSL_CERTIFICATE_SOURCES
} PACE2_SSL_certificate_source;

/**
 * @brief Types of the certificates in the certificate chain
 */
typedef enum {
    PACE2_SSL_CHAIN_START,                          /**< Type of the first certificate in the chain". */
    PACE2_SSL_CHAIN_INTERMEDIATE,                   /**< Type of any intermediate certificate in the chain". */
    PACE2_SSL_CHAIN_END,                            /**< Type of the last certificate in the chain". */

    PACE2_NUMBER_OF_SSL_CERTIFICATE_TYPES
} PACE2_SSL_certificate_type;

/**
 * @brief Structure to hold the type of each certificate and its position in the chain.
 */
struct PACE2_SSL_certificate_chain {
    u8 position;
    PACE2_SSL_certificate_type certificate_type;    /**< Holds the type of certificate in the chain". */
    PACE2_SSL_certificate_source chain_source;      /**< Source of the certificate chain (client or server)". */
};

/**
 * @brief Type of advanced SSL events.
 */
typedef enum {
    PACE2_ADVANCED_SSL_INVALID,                         /**< Decoding of a connection fails. */
    PACE2_ADVANCED_SSL_NEW_CONNECTION,                  /**< Information about connection opened. */
    PACE2_ADVANCED_SSL_CONNECTION_END,                  /**< Information about connection closed. */
    PACE2_ADVANCED_SSL_CERTIFICATE,                     /**< SSL handshake type "certificate". */
    PACE2_ADVANCED_SSL_NEW_CERTIFICATE,                 /**< Start event of certificate decoding. */
    PACE2_ADVANCED_SSL_CERT_RAW,                        /**< Raw certificate event. */
    PACE2_ADVANCED_SSL_CERT_VERSION,                    /**< Version number of the X.509 certificate. */
    PACE2_ADVANCED_SSL_CERT_SERIAL_NUMBER,              /**< Certificate serial number. */
    PACE2_ADVANCED_SSL_CERT_SIGNATURE_ALGORITHM,        /**< Signature algorithm identifier. */
    PACE2_ADVANCED_SSL_CERT_SIGNATURE,                  /**< Signature of the certificate . */
    PACE2_ADVANCED_SSL_CERT_PUBLIC_KEY_ALGORITHM,       /**< Public key algorithm identifier. */
    PACE2_ADVANCED_SSL_CERT_PUBLIC_KEY,                 /**< Public key of the certificate. */
    PACE2_ADVANCED_SSL_CERT_VALIDITY_NOT_BEFORE,        /**< Validity starting point. */
    PACE2_ADVANCED_SSL_CERT_VALIDITY_NOT_AFTER,         /**< Validity end point. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_COMMON_NAME,        /**< Subject common name. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_ORGANIZATIONAL_UNIT_NAME,   /**< Subject organizational unit name. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_ORGANIZATION_NAME,  /**< Subject organization name. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_COUNTRY_NAME,       /**< Subject country name. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_COMMON_NAME,         /**< Issuer common name. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_ORGANIZATIONAL_UNIT_NAME,    /**< Issuer organizational unit name. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_ORGANIZATION_NAME,   /**< Issuer organization name. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_COUNTRY_NAME,        /**< Issuer country name. */
    PACE2_ADVANCED_SSL_CERTIFICATE_END,                 /**< End event of certificate decoding. */
    PACE2_ADVANCED_SSL_CERTIFICATE_STATUS,              /**< SSL handshake type "certificate status". */
    PACE2_ADVANCED_SSL_CLIENT_HELLO,                    /**< SSL handshake type "client hello". */
    PACE2_ADVANCED_SSL_SERVER_HELLO,                    /**< SSL handshake type "server hello". */
    PACE2_ADVANCED_SSL_SERVER_HELLO_DONE,               /**< SSL handshake type "server hello done". */
    PACE2_ADVANCED_SSL_CERT_SUBJECT,                    /**< Subject's distinguished name. */
    PACE2_ADVANCED_SSL_CERT_ISSUER,                     /**< Issuer's distinguished name. */
    PACE2_ADVANCED_SSL_CLIENT_SIGNATURE_ALGORITHMS,     /**< Signature algorithm extension of client hello. */
    PACE2_ADVANCED_SSL_CLIENT_ELLIPTIC_CURVES,          /**< Elliptic curves extension of client hello. */
    PACE2_ADVANCED_SSL_CLIENT_EC_POINT_FORMATS,         /**< Elliptic curve point formats extension of client hello. */
    PACE2_ADVANCED_SSL_CLIENT_HELLO_END,                /**< Indicates the end of the "client hello"
                                                         * message that started at the previous
                                                         * PACE2_ADVANCED_SSL_CLIENT_HELLO event. */
    PACE2_ADVANCED_SSL_SERVER_KEY_EXCHANGE,             /**< SSL handshake type "server key exchange". */
    PACE2_ADVANCED_SSL_SKE_CURVE_TYPE,                  /**< SSL server key exchange curve type. */
    PACE2_ADVANCED_SSL_SKE_NAMED_CURVE,                 /**< SSL server key exchange named curve. */
    PACE2_ADVANCED_SSL_SKE_PUBLIC_KEY_LENGTH,           /**< SSL server key exchange public key length. */
    PACE2_ADVANCED_SSL_SKE_SIGNATURE_ALGORITHM,         /**< SSL server key exchange signature hash algorithm value pair. */
    PACE2_ADVANCED_SSL_SKE_SIGNATURE_LENGTH,            /**< SSL server key exchange signature length. */
    PACE2_ADVANCED_SSL_SERVER_KEY_EXCHANGE_END,         /**< End event of server key exchange decoding. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_LOCALITY_NAME,      /**< Subject locality name. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_STATE_OR_PROVINCE_NAME, /**< Subject state or province name. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_LOCALITY_NAME,       /**< Issuer locality name. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_STATE_OR_PROVINCE_NAME,  /**< Issuer state or province name. */
    PACE2_ADVANCED_SSL_SERVER_VERSION,                  /**< SSL version chosen by the server. */
    PACE2_ADVANCED_SSL_SERVER_COMPRESSION_METHOD,       /**< The server returns one byte which represents the chosen compression method. */
    PACE2_ADVANCED_SSL_SERVER_CIPHER_SUITE,             /**< The server returns one byte with its chosen cipher suite. */
    PACE2_ADVANCED_SSL_SERVER_SESSION_ID,               /**< SSL server session id. */
    PACE2_ADVANCED_SSL_CLIENT_SESSION_ID,               /**< SSL client session id. */
    PACE2_ADVANCED_SSL_CLIENT_CIPHER_SUITES,            /**< A list with all cipher suites supported by the client. */
    PACE2_ADVANCED_SSL_CLIENT_COMPRESSION_METHODS,      /**< A bitmask with all compression methods supported by the client. */
    PACE2_ADVANCED_SSL_CLIENT_VERSION,                  /**< Client version extension. */
    PACE2_ADVANCED_SSL_CLIENT_SERVER_NAME,              /**< The SSL server name as the client expects it. */
    PACE2_ADVANCED_SSL_CLIENT_SUPPORTED_VERSIONS,       /**< Supported versions list. */
    PACE2_ADVANCED_SSL_REQUEST_ENCRYPTED_DATA,          /**< Request encrypted data. */
    PACE2_ADVANCED_SSL_RESPONSE_ENCRYPTED_DATA,         /**< Response encrypted data. */
    PACE2_ADVANCED_SSL_REQUEST_HOLE,                    /**< Hole of the given length was observed in the request encrypted data. */
    PACE2_ADVANCED_SSL_RESPONSE_HOLE,                   /**< Hole of the given length was observed in the response encrypted data. */
    PACE2_ADVANCED_SSL_CLIENT_HELLO_INCOMPLETE,         /**< SSL client hello is incomplete due to missing data.
                                                         * @note The event is currently unused. */
    PACE2_ADVANCED_SSL_SERVER_HELLO_INCOMPLETE,         /**< SSL server hello is incomplete due to missing data. */
    PACE2_ADVANCED_SSL_ALERT,                           /**< SSL alert type */
    PACE2_ADVANCED_SSL_ALERT_ENCRYPTED,                 /**< SSL alert type for encrypted alerts*/
    PACE2_ADVANCED_SSL_SERVER_EC_POINT_FORMATS,         /**< Elliptic curve point formats extension of server hello. */
    PACE2_ADVANCED_SSL_CERT_EXT_SUBJECT_ALT_NAME_DNS,   /**< Certificate extension Subject alternative name DNSName. */
    PACE2_ADVANCED_SSL_CERT_HASH_SHA1,                  /**< SHA1 fingerprint of raw certificate as reported by @ref PACE2_ADVANCED_SSL_CERT_RAW.  */
    PACE2_ADVANCED_SSL_SERVER_SUPPORTED_VERSION,        /**< Supported version of TLS (valid for >= TLS 1.3). */
    PACE2_ADVANCED_SSL_CLIENT_HELLO_EXTENSION_GENERIC,  /**< Generic extension of client hello. */
    PACE2_ADVANCED_SSL_SERVER_HELLO_EXTENSION_GENERIC,  /**< Generic extension of server hello. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_SURNAME,            /**< Certificate subject surname. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_SERIALNUMBER,       /**< Certificate subject serial number. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_STREETADDRESS,      /**< Certificate subject street address. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_TITLE,              /**< Certificate subject title. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_GIVENNAME,          /**< Certificate subject given name. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_INITIALS,           /**< Certificate subject initials. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_GENERATIONQUALIFIER,/**< Certificate subject generation qualifier. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_PSEUDONYM,          /**< Certificate subject pseudonym. */
    PACE2_ADVANCED_SSL_CERT_SUBJECT_EMAILADDRESS,       /**< Certificate subject e-mail address. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_SURNAME,             /**< Certificate issuer surname. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_SERIALNUMBER,        /**< Certificate issuer serial number. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_STREETADDRESS,       /**< Certificate issuer street address. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_TITLE,               /**< Certificate issuer title. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_GIVENNAME,           /**< Certificate issuer given name. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_INITIALS,            /**< Certificate issuer initials. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_GENERATIONQUALIFIER, /**< Certificate issuer generation qualifier. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_PSEUDONYM,           /**< Certificate issuer pseudonym. */
    PACE2_ADVANCED_SSL_CERT_ISSUER_EMAILADDRESS,        /**< Certificate issuer e-mail address. */
    PACE2_ADVANCED_SSL_SKE_DH_SERVER_PARAMS,            /**< SSL server key exchange Diffie-Hellman Server Params */
    PACE2_ADVANCED_SSL_SERVER_HELLO_END,                /**< Indicates the end of the "server hello"
                                                         * message that started at the previous
                                                         * PACE2_ADVANCED_SSL_SERVER_HELLO event. */

    PACE2_NUMBER_OF_ADVANCED_SSL_EVENT_TYPES            /**< Number of implemented event types. */
} PACE2_advanced_SSL_event_type;

/**
 * @brief Hash algorithms as defined in RFC 5246, section 7.4.1.4.1.
 */
enum PACE2_SSL_hash_algorithm {
    PACE2_SSL_HASH_ALGORITHM_NONE = 0,                  /**< Hash algorithm not defined. */
    PACE2_SSL_HASH_ALGORITHM_MD5,                       /**< MD5 algorithm. */
    PACE2_SSL_HASH_ALGORITHM_SHA1,                      /**< SHA1 algorithm. */
    PACE2_SSL_HASH_ALGORITHM_SHA224,                    /**< SHA224 algorithm. */
    PACE2_SSL_HASH_ALGORITHM_SHA256,                    /**< SHA256 algorithm. */
    PACE2_SSL_HASH_ALGORITHM_SHA384,                    /**< SHA384 algorithm. */
    PACE2_SSL_HASH_ALGORITHM_SHA512,                    /**< SHA512 algorithm. */

    PACE2_NUMBER_OF_SSL_HASH_ALGORITHMS                 /**< Number of defined hash algorithms. */
};

/**
 * @brief Signature algorithms as defined in RFC 5246, section 7.4.1.4.1.
 */
enum PACE2_SSL_signature_algorithm {
    PACE2_SSL_SIGNATURE_ALGORITHM_ANONYMOUS = 0,        /**< Anonymous signature algorithm. */
    PACE2_SSL_SIGNATURE_ALGORITHM_RSA,                  /**< RSA algorithm. */
    PACE2_SSL_SIGNATURE_ALGORITHM_DSA,                  /**< DSA algorithm. */
    PACE2_SSL_SIGNATURE_ALGORITHM_ECDSA,                /**< ECDSA algorithm. */

    PACE2_NUMBER_OF_SSL_SIGNATURE_ALGORITHMS            /**< Number of defined signature algorithms. */
};

/**
 * @brief Elliptic curve types as defined in RFC 4492, section 5.4.
 */
enum PACE2_SSL_ecc_type {
    PACE2_SSL_ECC_TYPE_EXPLICIT_PRIME = 1,              /**< The underlying finite field is a prime field. */
    PACE2_SSL_ECC_TYPE_EXPLICIT_CHAR2,                  /**< The underlying finite field is a characteristic-2 field. */
    PACE2_SSL_ECC_TYPE_NAMED_CURVE,                     /**< Indicates that a named curve is used. */

    PACE2_NUMBER_OF_SSL_ECC_TYPES
};

/**
 * @brief Elliptic curve named curves as defined in
 * @arg RFC 4492, section 5.1.1
 * @arg RFC 7027, section 3
 * @arg RFC 7919, section 7
 * @arg RFC 8422, section 5.1.1
 *
 */
enum PACE2_SSL_ecc_named {
    PACE2_SSL_ECC_NAMED_SECT163K1 = 1,                                          /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECT163R1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECT163R2,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECT193R1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECT193R2,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECT233K1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECT233R1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECT239K1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECT283K1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECT283R1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECT409K1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECT409R1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECT571K1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECT571R1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECP160K1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECP160R1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECP160R2,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECP192K1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECP192R1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECP224K1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECP224R1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECP256K1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECP256R1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECP384R1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_SECP521R1,                                              /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_BRAINPOOLP256R1,                                        /**< added with RFC 7027 */
    PACE2_SSL_ECC_NAMED_BRAINPOOLP384R1,                                        /**< added with RFC 7027 */
    PACE2_SSL_ECC_NAMED_BRAINPOOLP512R1,                                        /**< added with RFC 7027 */
    PACE2_SSL_ECC_NAMED_X25519,                                                 /**< added with RFC 8422 */
    PACE2_SSL_ECC_NAMED_X448,                                                   /**< added with RFC 8422 */
    PACE2_SSL_ECC_NAMED_FFDHE2048 = 256,                                        /**< added with RFC 7919 */
    PACE2_SSL_ECC_NAMED_FFDHE3072,                                              /**< added with RFC 7919 */
    PACE2_SSL_ECC_NAMED_FFDHE4096,                                              /**< added with RFC 7919 */
    PACE2_SSL_ECC_NAMED_FFDHE6144,                                              /**< added with RFC 7919 */
    PACE2_SSL_ECC_NAMED_FFDHE8192,                                              /**< added with RFC 7919 */
    PACE2_SSL_ECC_NAMED_ARBITRARY_EXPLICIT_PRIME_CURVES = 0xff01,               /**< added with RFC 4492 */
    PACE2_SSL_ECC_NAMED_ARBITRARY_EXPLICIT_CHAR2_CURVES = 0xff02,               /**< added with RFC 4492 */

    PACE2_NUMBER_OF_SSL_ECC_NAMED
};

/**
 * @brief Elliptic curve point formats as defined in RFC 4492, section 5.1.2.
 */
enum PACE2_SSL_ec_point_format {
    PACE2_SSL_EC_POINT_FORMAT_UNCOMPRESSED = 0,
    PACE2_SSL_EC_POINT_FORMAT_ANSI_X962_COMPRESSED_PRIME,
    PACE2_SSL_EC_POINT_FORMAT_ANSI_X962_COMPRESSED_CHAR2,

    PACE2_NUMBER_OF_SSL_EC_POINT_FORMATS
};

/**
 * @brief SSL alert level as defined in
 * @arg RFC 5246, section 7.2
 */
enum PACE2_SSL_alert_level_type {
    PACE2_SSL_ALERT_LEVEL_TYPE_WARNING = 1,
    PACE2_SSL_ALERT_LEVEL_TYPE_FATAL = 2
};

/**
 * @brief SSL alert description as defined in
 * @arg RFC 2246, section 7.2
 * @arg RFC 4346, section 7.2
 * @arg RFC 5246, section 7.2
 * @arg RFC 6066, section 5
 * @arg RFC 7301, section 3.2
 * @arg RFC 8446, section 6
 * @arg draft-ietf-tls-esni-04, section 8.2
 */
enum PACE2_SSL_alert_description_type {
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_CLOSE_NOTIFY = 0,                          /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_UNEXPECTED_MESSAGE = 10,                   /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_BAD_RECORD_MAC = 20,                       /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_DECRYPTION_FAILED_RESERVED = 21,           /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_RECORD_OVERFLOW = 22,                      /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_DECOMPRESSION_FAILURE = 30,                /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_HANDSHAKE_FAILURE = 40,                    /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_NO_CERTIFICATE_RESERVED = 41,              /**< added with RFC 4346 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_BAD_CERTIFICATE = 42,                      /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_UNSUPPORTED_CERTIFICATE = 43,              /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_CERTIFICATE_REVOKED = 44,                  /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_CERTIFICATE_EXPIRED = 45,                  /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_CERTIFICATE_UNKNOWN = 46,                  /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_ILLEGAL_PARAMETER = 47,                    /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_UNKNOWN_CA = 48,                           /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_ACCESS_DENIED = 49,                        /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_DECODE_ERROR = 50,                         /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_DECRYPT_ERROR = 51,                        /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_EXPORT_RESTRICTION_RESERVED = 60,          /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_PROTOCOL_VERSION = 70,                     /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_INSUFFICIENT_SECURITY = 71,                /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_INTERNAL_ERROR = 80,                       /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_INAPPROPRIATE_FALLBACK = 86,               /**< added with RFC 8446 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_USER_CANCELED = 90,                        /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_NO_RENEGOTIATION = 100,                    /**< added with RFC 2246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_MISSING_EXTENSION = 109,                   /**< added with RFC 8446 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_UNSUPPORTED_EXTENSION = 110,               /**< added with RFC 5246 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_CERTIFICATE_UNOBTAINABLE = 111,            /**< added with RFC 6066 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_UNRECOGNIZED_NAME = 112,                   /**< added with RFC 8446 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_BAD_CERTIFICATE_STATUS_RESPONSE = 113,     /**< added with RFC 8446 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_BAD_CERTIFICATE_HASH_VALUE = 114,          /**< added with RFC 8446 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_UNKNOWN_PSK_IDENTITY = 115,                /**< added with RFC 8446 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_CERTIFICATE_REQUIRED = 116,                /**< added with RFC 8446 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_NO_APPLICATION_PROTOCOL = 120,             /**< added with RFC 7301 */
    PACE2_SSL_ALERT_DESCRIPTION_TYPE_ESNI_REQUIRED = 121                        /**< added with draft-ietf-tls-esni-04 */
};

/**
 * @brief Type of the server key exchange message.
 */
enum PACE2_SSL_ske_type {
    PACE2_SSL_SKE_TYPE_UNKNOWN = 0,
    PACE2_SSL_SKE_TYPE_ECDH,
    PACE2_SSL_SKE_TYPE_DHE_DSS,
    PACE2_SSL_SKE_TYPE_DHE_RSA,
    PACE2_SSL_SKE_TYPE_DH_ANON
};

#endif /* PACE_KERNEL_LIB */

/* Event Structures *************************************************************/

/**
 * @brief Data structure of a SSL client hello metadata event.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_BASIC_SSL_CLIENT_HELLO_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_bitmask meta_data_mask;       /**< Bitmask, which states the available SSL client hello metadata */
    struct {
        /** Client hello line of type @ref PACE2_SSL_client_hello_line.
         *
         *  @deprecated Please do not uses this variable any longer. It will be removed in a future release of this
         *              library. The type of the line is implicit in the index of the line array. */
        PACE2_DEPRECATED(
            "Please do not uses this variable any longer. It will be removed in a future release of this library. The "
            "type of the line is implicit in the index of the line array.")
        PACE2_SSL_client_hello_line type;
        PACE2_byte_buffer content;              /**< Client hello line content. */
        u8 line_complete:1;                     /**< Flag to signal, that the reported line was complete. */
    } line[PACE2_NUMBER_OF_CLIENT_HELLO_LINES]; /**< Data structure for a event of type @ref PACE2_BASIC_SSL_CLIENT_HELLO_EVENT.
                                                     Contains content for each available line of type @ref PACE2_SSL_client_hello_line */
} PACE2_SSL_client_hello_event;

/**
 * @brief Data structure of a SSL server hello metadata event.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_BASIC_SSL_SERVER_HELLO_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_bitmask meta_data_mask;       /**< Bitmask, which states the available SSL server hello metadata */
    struct {
        /** Server hello line of type @ref PACE2_SSL_server_hello_line .
         *
         *  @deprecated Please do not uses this variable any longer. It will be removed in a future release of this
         *              library. The type of the line is implicit in the index of the line array. */
        PACE2_DEPRECATED(
            "Please do not uses this variable any longer. It will be removed in a future release of this library. The "
            "type of the line is implicit in the index of the line array.")
        PACE2_SSL_server_hello_line type;
        PACE2_byte_buffer content;              /**< Server hello line content. */
        u8 line_complete:1;                     /**< Flag to signal, that the reported line was complete. */
    } line[PACE2_NUMBER_OF_SERVER_HELLO_LINES]; /**< Data structure for a event of type @ref PACE2_BASIC_SSL_SERVER_HELLO_EVENT.
                                                     Contains the content for each available line of type @ref PACE2_SSL_server_hello_line */
} PACE2_SSL_server_hello_event;

/**
 * @brief Data structure of a SSL DNS alternative names metadata event.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_BASIC_SSL_DNS_ALT_NAMES_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_byte_buffer entries[IPOQUE_MAX_PARSE_SSL_DNS_ALT_NAMES];
    u16 number_of_entries;
} PACE2_SSL_dns_alt_names_event;

/**
 * @brief Data structure of a SSL session ticket metadata event.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_BASIC_SSL_SESSION_TICKET_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_SSL_session_ticket_source source;         /**< Session ticket source of type @ref PACE2_SSL_session_ticket_source .*/
    PACE2_bitmask meta_data_mask;                   /**< Bitmask, which states the available SSL session ticket metadata */
    u32 lifetime;                                   /**< Session ticket lifetime. */
    PACE2_byte_buffer content;                      /**< Session ticket content. */
} PACE2_SSL_session_ticket_event;

#ifndef PACE_KERNEL_LIB

/**
 * @brief Structure of an object identifier (OID) value.
 */
struct PACE2_ssl_oid {
    u32 * oid_part;                 /**< Pointer to the list of OID parts. */
    u8 oid_part_count;              /**< Number of OID parts in @ref oid_part. */
};

/**
 * @brief Structure of a signature and hash algorithm value pair.
 */
struct PACE2_SSL_sh_algorithm_entry {
    u8 hash;                        /**< Hash algorithm value of type @ref PACE2_SSL_hash_algorithm.*/
    u8 signature;                   /**< Signature algorithm value of type @ref PACE2_SSL_signature_algorithm. */
};

/**
 * @brief Generic Diffie Hellman parameters structure. For more details see RFC 5246 section 7.4.3
 */
struct PACE2_ssl_ske_dh_server_params {
    struct ipd_string dh_g;  /**< The generator used for the Diffie-Hellman operation. */
    struct ipd_string dh_p;  /**< The prime modulus used for the Diffie-Hellman operation. */
    struct ipd_string dh_Ys; /**< The server's Diffie-Hellman public value (g^X mod p). In wireshark also called pubkey.
                              */
};

/**
 * @brief This structure represents a single item of the @ref PACE2_server_name_list
 */
struct PACE2_server_name_list_item {
    u8 type;                        /**< From rfc3546: name_type is usually 0 for it being a dns host name. */
    charset_string name;            /**< From the name of that type above (just one entry per type allowed). */
};

/**
 * @brief List structure to hold server names for the PACE2_ADVANCED_SSL_CLIENT_SERVER_NAME event
 *
 * This list provides the names in a more uniform way with its items having a charset_string representation
 */
struct PACE2_server_name_list {
    u16 length;                     /**< length of the list (contains 'length' elements). */
    struct PACE2_server_name_list_item * item; /**< Array list of which this pointer refers to the first item. */
};

/**
 * @brief List structure to store information for the PACE2_ADVANCED_SSL_ALERT event
 *
 * This list provides the SSL alert level and the SSL alert description.
 */
struct PACE2_ssl_alert {
    enum PACE2_SSL_alert_level_type level;                  /**< SSL alert level */
    enum PACE2_SSL_alert_description_type description;      /**< SSL alert description */
};

/**
 * @brief Structure for a PACE2_ADVANCED_SSL_ALERT_ENCRYPTED event.
 */
struct PACE2_ssl_alert_encrypted {
    charset_string encrypted_data;              /**< contains the payload of the encrypted alert as blob. */
};

/**
 * @brief Structure for a PACE2_ADVANCED_SSL_ALERT_ENCRYPTED event.
 */
struct PACE2_ssl_extension {
    u16 extension_type;                         /**< From https://www.iana.org/assignments/tls-extensiontype-values */
    struct ipd_string extension_value;          /**< payload and length of the extension (might be 0). */
};

/**
 * @brief Data structure of advanced SSL events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_ADVANCED_SSL_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_advanced_SSL_event_type meta_data_type;  /**< Event type, which represents the SSL handshake type. */
    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;
    union {
        u8 cert_version;                                     /**< Data variable of @ref PACE2_ADVANCED_SSL_CERT_VERSION. */
        charset_string cert_serial_number;                   /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SERIAL_NUMBER. */
        struct PACE2_ssl_oid cert_signature_algorithm;       /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SIGNATURE_ALGORITHM. */
        charset_string cert_signature;                       /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SIGNATURE. */
        struct PACE2_ssl_oid cert_public_key_algorithm;      /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_PUBLIC_KEY_ALGORITHM. */
        charset_string cert_public_key;                      /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_PUBLIC_KEY. */
        charset_string cert_validity_not_before;             /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_VALIDITY_NOT_BEFORE. */
        charset_string cert_validity_not_after;              /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_VALIDITY_NOT_AFTER. */
        charset_string cert_subject_common_name;             /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_COMMON_NAME. */
        charset_string cert_subject_organizational_unit_name;/**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_ORGANIZATIONAL_UNIT_NAME. */
        charset_string cert_subject_organization_name;       /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_ORGANIZATION_NAME. */
        charset_string cert_subject_country_name;            /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_COUNTRY_NAME. */
        charset_string cert_subject_locality_name;           /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_LOCALITY_NAME. */
        charset_string cert_subject_state_or_province_name;  /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_STATE_OR_PROVINCE_NAME. */
        charset_string cert_subject_surname;                 /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_SURNAME. */
        charset_string cert_subject_serialnumber;            /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_SERIALNUMBER. */
        charset_string cert_subject_streetaddress;           /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_STREETADDRESS. */
        charset_string cert_subject_title;                   /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_TITLE. */
        charset_string cert_subject_givenname;               /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_GIVENNAME. */
        charset_string cert_subject_initials;                /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_INITIALS. */
        charset_string cert_subject_generationqualifier;     /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_GENERATIONQUALIFIER. */
        charset_string cert_subject_pseudonym;               /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_PSEUDONYM. */
        charset_string cert_subject_emailaddress;            /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_EMAILADDRESS. */
        charset_string cert_issuer_common_name;              /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_COMMON_NAME. */
        charset_string cert_issuer_organizational_unit_name; /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_ORGANIZATIONAL_UNIT_NAME. */
        charset_string cert_issuer_organization_name;        /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_ORGANIZATION_NAME. */
        charset_string cert_issuer_country_name;             /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_COUNTRY_NAME. */
        charset_string cert_issuer_locality_name;            /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_LOCALITY_NAME. */
        charset_string cert_issuer_state_or_province_name;   /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_STATE_OR_PROVINCE_NAME. */
        charset_string cert_issuer_surname;                  /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_SURNAME. */
        charset_string cert_issuer_serialnumber;             /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_SERIALNUMBER. */
        charset_string cert_issuer_streetaddress;            /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_STREETADDRESS. */
        charset_string cert_issuer_title;                    /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_TITLE. */
        charset_string cert_issuer_givenname;                /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_GIVENNAME. */
        charset_string cert_issuer_initials;                 /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_INITIALS. */
        charset_string cert_issuer_generationqualifier;      /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_GENERATIONQUALIFIER. */
        charset_string cert_issuer_pseudonym;                /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_PSEUDONYM. */
        charset_string cert_issuer_emailaddress;             /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_EMAILADDRESS. */
        charset_string cert_raw;                             /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_RAW.
                                                                  Contains the raw certificate payload bytes. */
        charset_string cert_issuer;                          /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER.
                                                                  Contains the raw issuer payload bytes. */
        charset_string cert_subject;                         /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT.
                                                                  Contains the raw subject payload bytes. */
        const char *invalid;                                 /**< Data variable for a @ref PACE2_ADVANCED_SSL_INVALID event. */
        struct PACE2_u16_list client_hello_signature_algorithms;    /**< Data structure for a @ref PACE2_ADVANCED_SSL_CLIENT_SIGNATURE_ALGORITHMS event.
                                                                 The list contains data in network byte order. */
        struct PACE2_u16_list client_hello_elliptic_curves; /**< Data structure for a @ref PACE2_ADVANCED_SSL_CLIENT_ELLIPTIC_CURVES event.
                                                                 The list contains data in network byte order. */
        struct PACE2_u8_list client_hello_ec_point_formats; /**< Data structure for a @ref PACE2_ADVANCED_SSL_CLIENT_EC_POINT_FORMATS event. */
        enum PACE2_SSL_ske_type server_key_exchange;        /**< Data structure for a @ref PACE2_ADVANCED_SSL_SERVER_KEY_EXCHANGE event. */
        enum PACE2_SSL_ecc_type ske_curve_type;             /**< Data variable for a @ref PACE2_ADVANCED_SSL_SKE_CURVE_TYPE event. */
        enum PACE2_SSL_ecc_named ske_named_curve;           /**< Data variable for a @ref PACE2_ADVANCED_SSL_SKE_NAMED_CURVE event. */
        u16 ske_public_key_length;                          /**< Data variable for a @ref PACE2_ADVANCED_SSL_SKE_PUBLIC_KEY_LENGTH event. */
        struct PACE2_SSL_sh_algorithm_entry ske_signature_algorithm;    /**< Data variable for a @ref PACE2_ADVANCED_SSL_SKE_SIGNATURE_ALGORITHM event. */
        struct PACE2_ssl_ske_dh_server_params ske_dh_server_params;     /**< Data variable for a @ref PACE2_ADVANCED_SSL_SKE_DH_SERVER_PARAMS event. */
        u16 ske_signature_length;                           /**< Data variable for a @ref PACE2_ADVANCED_SSL_SKE_SIGNATURE_LENGTH event. */
        charset_string server_session_id;                   /**< Data structure of @ref PACE2_ADVANCED_SSL_SERVER_SESSION_ID. */
        charset_string client_session_id;                   /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_SESSION_ID. */
        u16 cipher_suite;                                   /**< Data structure of @ref PACE2_ADVANCED_SSL_SERVER_CIPHER_SUITE.
                                                                 The field contains data in network byte order. */
        u8 compression_method;                              /**< Data structure of @ref PACE2_ADVANCED_SSL_SERVER_COMPRESSION_METHOD. */
        u16 client_version;                                 /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_VERSION.
                                                                 The field contains data in network byte order. */
        u16 server_version;                                 /**< Data structure of @ref PACE2_ADVANCED_SSL_SERVER_VERSION.
                                                                 The field contains data in network byte order. */
        u16 server_supported_version;                       /**< Data structure of @ref PACE2_ADVANCED_SSL_SERVER_SUPPORTED_VERSION. */
        struct PACE2_u16_list cipher_suites;                /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_CIPHER_SUITES.
                                                                 The list contains data in network byte order. */
        struct PACE2_u8_list compression_methods;           /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_COMPRESSION_METHODS. */
        struct PACE2_server_name_list server_name_list;     /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_SERVER_NAME. */
        struct PACE2_u16_list client_hello_supported_versions; /**< Data structure for a @ref PACE2_ADVANCED_SSL_CLIENT_SUPPORTED_VERSIONS event.
                                                                    The list contains data in network byte order. */
        charset_string request_encrypted_data;              /**< Data variable for the @ref PACE2_ADVANCED_SSL_REQUEST_ENCRYPTED_DATA event. */
        charset_string response_encrypted_data;             /**< Data variable for the @ref PACE2_ADVANCED_SSL_RESPONSE_ENCRYPTED_DATA event. */
        u64 request_hole;                                   /**< Contains data for the @ref PACE2_ADVANCED_SSL_REQUEST_HOLE event. */
        u64 response_hole;                                  /**< Contains data for the @ref PACE2_ADVANCED_SSL_RESPONSE_HOLE event. */
        struct PACE2_ssl_alert alert;                       /**< Contains data for the @ref PACE2_ADVANCED_SSL_ALERT event. */
        struct PACE2_ssl_alert_encrypted alert_encrypted;   /**< Contains data for the @ref PACE2_ADVANCED_SSL_ALERT_ENCRYPTED event. */
        struct PACE2_u8_list server_hello_ec_point_formats; /**< Data structure for a @ref PACE2_ADVANCED_SSL_SERVER_EC_POINT_FORMATS event. */
        charset_string cert_ext_subject_alt_name_dns;       /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_EXT_SUBJECT_ALT_NAME_DNS. */
        charset_string cert_hash_sha1;                      /**< Data structure of @ref PACE2_ADVANCED_SSL_CERT_HASH_SHA1.
                                                                 Contains the binary SHA1 hash of the raw certificate as reported by @ref PACE2_ADVANCED_SSL_CERT_RAW. */
        struct PACE2_ssl_extension client_extension_generic;    /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_HELLO_EXTENSION_GENERIC. */
        struct PACE2_ssl_extension server_extension_generic;    /**< Data structure of @ref PACE2_ADVANCED_SSL_SERVER_HELLO_EXTENSION_GENERIC. */
        struct PACE2_SSL_certificate_chain certificate_chain;   /**< Data structure used by the @ref PACE2_ADVANCED_SSL_NEW_CERTIFICATE. */
    } event_data;                                           /**< Named union which contains the data of the event types. */
} PACE2_advanced_SSL_event;

/**
 * @brief Data structure of the event type @ref PACE2_CLASS_SSL_CERTIFICATE.
 */
struct PACE2_class_ssl_certificate_data {

    /** Flow ID to which this event is related to. */
    u64 flow_id;

    /** Data variable of @ref PACE2_ADVANCED_SSL_CERT_VERSION. */
    u8 version;

    /** Data structure of @ref PACE2_ADVANCED_SSL_CERT_SERIAL_NUMBER. */
    charset_string serial_number;

    /** Data structure of @ref PACE2_ADVANCED_SSL_CERT_SIGNATURE_ALGORITHM. */
    struct PACE2_ssl_oid signature_algorithm;

    /** Data structure of @ref PACE2_ADVANCED_SSL_CERT_SIGNATURE. */
    charset_string signature;

    /** Data structure of @ref PACE2_ADVANCED_SSL_CERT_PUBLIC_KEY_ALGORITHM. */
    struct PACE2_ssl_oid public_key_algorithm;

    /** Data structure of @ref PACE2_ADVANCED_SSL_CERT_PUBLIC_KEY. */
    charset_string public_key;

    /** Data structure of @ref PACE2_ADVANCED_SSL_CERT_VALIDITY_NOT_BEFORE. */
    charset_string validity_not_before;

    /** Data structure of @ref PACE2_ADVANCED_SSL_CERT_VALIDITY_NOT_AFTER. */
    charset_string validity_not_after;

    /** Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_COUNTRY_NAME. */
    charset_string subject_country_name;

    /** Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_COUNTRY_NAME. */
    charset_string issuer_country_name;

    /**
     * Data structure of @ref PACE2_ADVANCED_SSL_CERT_RAW.
     * Contains the raw certificate payload bytes.
     */
    charset_string raw;

    /**
     * Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER.
     * Contains the raw issuer payload bytes.
     */
    charset_string issuer;

    /**
     * Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT.
     * Contains the raw subject payload bytes.
     */
    charset_string subject;

    /** Interpreted data of @ref PACE2_ADVANCED_SSL_CERT_VALIDITY_NOT_BEFORE. */
    struct tm validity_not_before_tm;

    /** Interpreted data of @ref PACE2_ADVANCED_SSL_CERT_VALIDITY_NOT_AFTER. */
    struct tm validity_not_after_tm;

    /** Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_LOCALITY_NAME. */
    charset_string subject_locality_name;

    /** Data structure of @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_STATE_OR_PROVINCE_NAME. */
    charset_string subject_state_or_province_name;

    /** Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_LOCALITY_NAME. */
    charset_string issuer_locality_name;

    /** Data structure of @ref PACE2_ADVANCED_SSL_CERT_ISSUER_STATE_OR_PROVINCE_NAME. */
    charset_string issuer_state_or_province_name;

    /**
     * Number of the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_COMMON_NAME.
     */
    u16 subject_common_name_count;

    /**
     * Pointer to the array containing the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_COMMON_NAME.
     */
    struct ipd_charset_string * subject_common_name_array;

    /**
     * Number of the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_ORGANIZATIONAL_UNIT_NAME.
     */
    u16 subject_organizational_unit_name_count;

    /**
     * Pointer to the array containing the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_ORGANIZATIONAL_UNIT_NAME.
     */
    struct ipd_charset_string * subject_organizational_unit_name_array;

    /**
     * Number of the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_ORGANIZATION_NAME.
     */
    u16 subject_organization_name_count;

    /**
     * Pointer to the array containing the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_SUBJECT_ORGANIZATION_NAME.
     */
    struct ipd_charset_string * subject_organization_name_array;

    /**
     * Number of the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_ISSUER_COMMON_NAME.
     */
    u16 issuer_common_name_count;

    /**
     * Pointer to the array containing the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_ISSUER_COMMON_NAME.
     */
    struct ipd_charset_string * issuer_common_name_array;

    /**
     * Number of the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_ISSUER_ORGANIZATIONAL_UNIT_NAME.
     */
    u16 issuer_organizational_unit_name_count;

    /**
     * Pointer to the array containing the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_ISSUER_ORGANIZATIONAL_UNIT_NAME.
     */
    struct ipd_charset_string * issuer_organizational_unit_name_array;

    /**
     * Number of the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_ISSUER_ORGANIZATION_NAME.
     */
    u16 issuer_organization_name_count;

    /**
     * Pointer to the array containing the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_ISSUER_ORGANIZATION_NAME.
     */
    struct ipd_charset_string * issuer_organization_name_array;

    /**
     * Number of the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_EXT_SUBJECT_ALT_NAME_DNS.
     */
    u16 ext_subject_alt_name_dns_count;

    /**
     * Pointer to the array containing the data structures of
     * @ref PACE2_ADVANCED_SSL_CERT_EXT_SUBJECT_ALT_NAME_DNS.
     */
    struct ipd_charset_string * ext_subject_alt_name_dns_array;

    /**
     * Indicates whether the message was sent by the server or the client
     */
    PACE2_SSL_certificate_source source;
};

/**
 * @brief Data structure of the event type @ref PACE2_CLASS_SSL_SERVER_KEY_EXCHANGE. Contains information about a ECDH
 *        key exchange. For more details see RFC 4492.
 */
struct PACE2_class_SSL_server_key_exchange_data {
    u64 flow_id;                          /**< Flow ID to which this event is related to. */
    enum PACE2_SSL_ecc_type curve_type;   /**< Data variable of @ref PACE2_ADVANCED_SSL_SKE_CURVE_TYPE. */
    enum PACE2_SSL_ecc_named named_curve; /**< Data variable of @ref PACE2_ADVANCED_SSL_SKE_NAMED_CURVE. */
    u16 public_key_length;                /**< Data variable of @ref PACE2_ADVANCED_SSL_SKE_PUBLIC_KEY_LENGTH. */
    struct PACE2_SSL_sh_algorithm_entry signature_algorithm; /**< Data variable of @ref
                                                                PACE2_ADVANCED_SSL_SKE_SIGNATURE_ALGORITHM. */
    u16 signature_length; /**< Data variable of @ref PACE2_ADVANCED_SSL_SKE_SIGNATURE_LENGTH. */
};

/**
 * @brief Data structure of the event type @ref PACE2_CLASS_SSL_SERVER_KEY_EXCHANGE_RSA_DSS. Contains information about
 *        a DHE_RSA or DHE_DSS key exchange. For more details see RFC 5246.
 */
struct PACE2_class_SSL_server_key_exchange_data_rsa_dss {
    u64 flow_id;                                         /**< Flow ID to which this event is related to. */
    struct PACE2_ssl_ske_dh_server_params server_params; /**< The basic parameter for the key exchange. */
    u16 signature_length; /**< Data variable of @ref PACE2_ADVANCED_SSL_SKE_SIGNATURE_LENGTH. */
};

/**
 * @brief Data structure of the event type @ref PACE2_CLASS_SSL_SERVER_KEY_EXCHANGE_ANON. Contains information about
 *        a DH_ANON key exchange. For more details see RFC 5246.
 */
struct PACE2_class_SSL_server_key_exchange_data_anon {
    u64 flow_id;                                         /**< Flow ID to which this event is related to. */
    struct PACE2_ssl_ske_dh_server_params server_params; /**< The basic parameter for the key exchange. */
};

/**
 * @brief Data structure of the event type @ref PACE2_CLASS_SSL_CLIENT_HELLO.
 */
struct PACE2_class_SSL_client_hello_data {
    u64 flow_id;                                    /**< Flow ID to which this event is related to. */
    struct PACE2_u16_list signature_algorithms;     /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_SIGNATURE_ALGORITHMS.
                                                         The list contains data in network byte order. */
    struct PACE2_u16_list elliptic_curves;          /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_ELLIPTIC_CURVES.
                                                         The list contains data in network byte order. */
    struct PACE2_u8_list ec_point_formats;          /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_EC_POINT_FORMATS. */
    struct PACE2_u16_list cipher_suites;            /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_CIPHER_SUITES.
                                                         The list contains data in network byte order. */

    struct PACE2_u8_list compression_methods;       /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_COMPRESSION_METHODS. */

    struct PACE2_server_name_list server_name_list; /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_SERVER_NAME. */

    charset_string session_id;                      /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_SESSION_ID. */

    u16 version;                                    /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_VERSION. */

    struct PACE2_u16_list supported_versions;       /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_SUPPORTED_VERSIONS.
                                                         The list contains data in network byte order. */
};

/**
 * @brief Data structure of the event type @ref PACE2_CLASS_SSL_SERVER_HELLO.
 */
struct PACE2_class_SSL_server_hello_data {
    u64 flow_id;                                    /**< Flow ID to which this event is related to.
                                                         The field contains data in network byte order. */

    charset_string session_id;                      /**< Data structure of @ref PACE2_ADVANCED_SSL_SERVER_SESSION_ID. */

    u16 cipher_suite;                               /**< Data structure of @ref PACE2_ADVANCED_SSL_SERVER_CIPHER_SUITE.
                                                         The field contains data in network byte order. */

    u8 compression_method;                          /**< Data structure of @ref PACE2_ADVANCED_SSL_SERVER_COMPRESSION_METHOD. */

    u16 version;                                    /**< Data structure of @ref PACE2_ADVANCED_SSL_SERVER_VERSION.
                                                         The field contains data in network byte order. */

    struct PACE2_u8_list ec_point_formats;          /**< Data structure of @ref PACE2_ADVANCED_SSL_CLIENT_EC_POINT_FORMATS. */

    u16 server_supported_version;                   /**< Data structure of @ref PACE2_ADVANCED_SSL_SERVER_SUPPORTED_VERSION.
                                                         The field contains data in network byte order. */
};

/**
 * @brief Data structure of the event type @ref PACE2_CLASS_SSL_REQUEST_ENCRYPTED_DATA and type @ref PACE2_CLASS_SSL_RESPONSE_ENCRYPTED_DATA.
 */
struct PACE2_class_SSL_encrypted_data {
    u64 flow_id;                                    /**< Flow ID to which this event is related to. */

    u64 offset;                                     /**< Offset from the beginning of the SSL encrypted data. */

    charset_string data;                            /**< Encrypted SSL data. */

    u8 last_chunk;                                  /**< Informs if it is the last chunk of data. */
};

/**
 * @brief Data structure of the class SSL events.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_CLASS_SSL_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_class_SSL_event_type meta_data_type; /**< Event type, which represents the SSL handshake type. */
    PACE2_CLASS_META_DATA_EVENT_EXTRA_FIELDS;
    union {
        /**
         * Data structure which contains the aggregated certificate meta data. @ref meta_data_type should be set to
         * @ref PACE2_CLASS_SSL_CERTIFICATE.
         */
        struct PACE2_class_ssl_certificate_data certificate;
        /**
         * Data structure which contains the aggregated server key exchange data. @ref meta_data_type should be set to
         * @ref PACE2_CLASS_SSL_SERVER_KEY_EXCHANGE.
         */
        struct PACE2_class_SSL_server_key_exchange_data server_key_exchange;
        /**
         * Data structure which contains the aggregated server key exchange data. @ref meta_data_type should be set to
         * @ref PACE2_CLASS_SSL_SERVER_KEY_EXCHANGE_RSA_DSS.
         */
        struct PACE2_class_SSL_server_key_exchange_data_rsa_dss server_key_exchange_rsa_dss;
        /**
         * Data structure which contains the aggregated server key exchange data. @ref meta_data_type should be set to
         * @ref PACE2_CLASS_SSL_SERVER_KEY_EXCHANGE_ANON.
         */
        struct PACE2_class_SSL_server_key_exchange_data_anon server_key_exchange_anon;
        /**
         * Data structure which contains the aggregated client hello meta data. @ref meta_data_type should be set to
         * @ref PACE2_CLASS_SSL_CLIENT_HELLO .
         */
        struct PACE2_class_SSL_client_hello_data client_hello;
        /**
         * Data structure which contains the aggregated server hello meta data @ref meta_data_type should be set to
         * @ref PACE2_CLASS_SSL_SERVER_HELLO .
         */
        struct PACE2_class_SSL_server_hello_data server_hello;
        /**
         * Data structure that contains the encrypted request SSL data.
         */
        struct PACE2_class_SSL_encrypted_data request_encrypted_data;
        /**
         * Data structure that contains the encrypted response SSL data.
         */
        struct PACE2_class_SSL_encrypted_data response_encrypted_data;
    } event_data; /**< Named union which contains the data of the event types. */
} PACE2_class_SSL_event;

#endif /* PACE_KERNEL_LIB */

#endif /* SSL_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
