/** @file      tcp_events.h
  * @brief     TCP metadata event definitions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef TCP_EVENTS_H
#define TCP_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the TCP event.
 */
#define PACE2_TCP_EVENT_GROUP PACE2_POLLING_PROTOCOL_META_DATA_GROUP

/**
 * @brief Public define for the assigned metadata group of the TCP started event.
 */
#define PACE2_TCP_STARTED_EVENT_GROUP PACE2_START_GROUP

/**
 * @brief Public define for the assigned metadata group of the TCP dropped event.
 */
#define PACE2_TCP_CLOSED_EVENT_GROUP PACE2_DROP_GROUP

/**
 * @brief Public define for the assigned metadata group of the advanced TCP event.
 */
#define PACE2_ADVANCED_TCP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief TCP statistic information flags.
 */
enum PACE2_TCP_data_flags {
    PACE2_TCP_LATENCY_STATS = 0,        /**< Latency statistics of a TCP flow. */
    PACE2_TCP_OUT_OF_ORDER_STATS,       /**< Out-of-order statistics of a TCP flow. */
    PACE2_TCP_RETRANSMISSION_STATS,     /**< Global TCP retransmission statistics. */

    PACE2_NUMBER_OF_TCP_DATA_FLAGS      /**< Number of implemented TCP data flags. */
};

/**
 * @brief The reason for a TCP close event
 */
typedef enum pace2_tcp_close_reason {
    TCP_FIN,        /*!< The TCP connection was finished. */
    TCP_RST,        /*!< The TCP connection was aborted. */
    TCP_TIMEOUT     /*!< The TCP connection timed out. */
} PACE2_tcp_close_reason;

/**
 * @brief A flag to signal whether a SYN packet or a SYN/ACK packet was found.
 */
typedef enum pace2_tcp_start_flag {
    TCP_SYN,        /*!< The TCP connection was started with a SYN packet. */
    TCP_SYN_ACK     /*!< The TCP connection was started with a SYN/ACK packet. */
} PACE2_tcp_start_flag;

/**
 * @brief Types for advanced TCP events.
 */
enum pace2_advanced_tcp_event_type {
    PACE2_ADVANCED_TCP_LOST_SEGMENT = 0,    /*!< Event type for a lost TCP segment. */
    PACE2_ADVANCED_TCP_SKIPPED_SEGMENT,     /*!< Event type for skipped TCP data. */

    PACE2_NUMBER_OF_ADVANCED_TCP_EVENT_TYPES
};

/**
 * @brief Reason for segment skipping.
 */
enum pace2_advanced_tcp_skip_reason {
    TCP_SKIPPED_RESYNC = 0,                 /*!< Segment is skipped to re-sync decoder. */
    TCP_SKIPPED_INCOMPLETE                  /*!< Segment is incomplete. */
};

/* Event Structures *************************************************************/

/**
 * @brief Event structure of a TCP statistics information event.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_TCP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_bitmask meta_data_mask;       /**< Bitmask, which states the available TCP statistic information. */
    flow_latency_struct_t latency_stats;                /**< Data structure of the @ref PACE2_TCP_LATENCY_STATS flag. */
    ipoque_flow_ooo_count_struct_t out_of_order_stats;  /**< Data structure of the  @ref PACE2_TCP_OUT_OF_ORDER_STATS flag. */
    u64 retransmission_stats;                           /**< Variable of the @ref PACE2_TCP_RETRANSMISSION_STATS flag */
} PACE2_TCP_event;

/**
 * @brief Event structure of a TCP started event.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_TCP_STARTED_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    u64 flow_id;                        /**< flow ID (assigned by PACE2) of this flow */
    PACE2_tcp_start_flag start_flag;    /**< Indicates the start flag of this TCP connection. Values can be @ref PACE2_tcp_start_flag. */
} PACE2_TCP_started_event;

/**
 * @brief Event structure of a TCP dropped event.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_TCP_CLOSED_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    u64 flow_id;                                  /**< flow ID (assigned by PACE2) of this flow */
    enum pace2_tcp_close_reason close_reason;     /**< Reason for the close of this TCP connection. Values can be @ref PACE2_tcp_close_reason  */
} PACE2_TCP_closed_event;

/**
 * @brief Event data of @ref PACE2_ADVANCED_TCP_LOST_SEGMENT.
 */
struct PACE2_advanced_TCP_lost_segment_data {
    u64 length;                                     /*!< Contains the length of the lost segment. */
};

/**
 * @brief Event data of @ref PACE2_ADVANCED_TCP_SKIPPED_SEGMENT.
 */
struct PACE2_advanced_TCP_skipped_segment_data {
    PACE2_byte_buffer data;                         /*!< Contains the skipped payload data. */
    enum pace2_advanced_tcp_skip_reason reason;     /*!< Reason why the segment was skipped. */
};

/**
 * @brief Event structure of a advanced TCP event.
 */
typedef struct {
    /** Global event type, has to be @ref PACE2_ADVANCED_TCP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;
    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;
    enum pace2_advanced_tcp_event_type event_type;  /*!< Sub-type of the advanced TCP event. */
    union {
        struct PACE2_advanced_TCP_lost_segment_data lost_segment;       /*!< Structure to hold data of event type @ref PACE2_ADVANCED_TCP_LOST_SEGMENT. */
        struct PACE2_advanced_TCP_skipped_segment_data skipped_segment; /*!< Structure to hold data of event type @ref PACE2_ADVANCED_TCP_SKIPPED_SEGMENT. */
    } event_data;                                   /*!< Union to store event information. */
} PACE2_advanced_TCP_event;

#endif /* TCP_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
