/** @file      telnet_events.h
  * @brief     Telnet event definitions.
  * @details   @copydoc doc_telnet
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef TELNET_EVENTS_H
#define TELNET_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

/* Event Group ******************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced Telnet events.
 */
#define PACE2_ADVANCED_TELNET_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Flags ******************************************************************/

/**
 * @brief Telnet metadata flags.
 */
typedef enum {

    /** This event indicates a new Telnet connection.
     * It is just a marker, data is a NULL pointer. */
    PACE2_ADVANCED_TELNET_NEW_CONNECTION = 0,

    /** This event indicates the end of the Telnet connection and flow.
     * It is just a marker, data is a NULL pointer. */
    PACE2_ADVANCED_TELNET_END_OF_FLOW,

    /** This event is issued when decoding of an Telnet connection fails.
     * It provides a char const * to a message describing the failure reason. */
    PACE2_ADVANCED_TELNET_INVALID,

    /** Provides a struct ipd_telnet_command * with the connection ID, direction and the data for the command. */
    PACE2_ADVANCED_TELNET_COMMAND,

    /** Provides a struct ipd_telnet_marked_command * with the command and the type. */
    PACE2_ADVANCED_TELNET_MARKED_COMMAND,

    /** Provides a struct ipd_telnet_data * with the cconnection ID, direction and the data. */
    PACE2_ADVANCED_TELNET_DATA,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_TELNET_EVENTS

} PACE2_advanced_telnet_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/** Enumeration for the type of marked commands. */
enum ipd_telnet_marked_command_type_enum {

    /** Signals that the set term ansi was send. */
    IPD_TELNET_MARKED_COMMAND_SET_TERM_ANSI,

    /** Signals that the set term vt52 was send. */
    IPD_TELNET_MARKED_COMMAND_SET_TERM_VT52,

    /** Signals that the set term vtnt was send. */
    IPD_TELNET_MARKED_COMMAND_SET_TERM_VTNT,

    /** Signals that the set term vt100 was send. */
    IPD_TELNET_MARKED_COMMAND_SET_TERM_VT100,

    /** Signals that the set term vt220 was send. */
    IPD_TELNET_MARKED_COMMAND_SET_TERM_VT220,

    /** Signals that the set term vt320 was send. */
    IPD_TELNET_MARKED_COMMAND_SET_TERM_VT320,

    /** Signals that the set term vt420 was send. */
    IPD_TELNET_MARKED_COMMAND_SET_TERM_VT420,

    /** Signals that the set term xterm was send. */
    IPD_TELNET_MARKED_COMMAND_SET_TERM_XTERM,

    /** Define the max of marked commands */
    IPD_TELNET_MARKED_COMMAND_COUNT,
};

/** Structure for a telnet command. */
struct ipd_telnet_command {

    /** The data which are part of the command. */
    struct ipd_charset_string command;

    /** The direction of the command. */
    u8 direction;
};

/** Structure of a marked telnet command.*/
struct ipd_telnet_marked_command {

    /** The telnet command as struct ipd_telnet_command * @see ipd_telnet_command. */
    struct ipd_telnet_command * command;

    /** The type of the marked command. @see ipd_telnet_marked_command_type_enum. */
    enum ipd_telnet_marked_command_type_enum type;
};

/** Structure for telnet data. */
struct ipd_telnet_data {

    /** The direction of the command. */
    u8 direction;

    /** The data which are send. */
    struct ipd_charset_string data;
};

/**
 * @brief Data structure for advanced Telnet events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_TELNET_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** Telnet event subtype. */
    PACE2_advanced_telnet_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    union {

        /** Contains data for a @ref PACE2_ADVANCED_TELNET_INVALID event. */
        char const * invalid;

        /** Contains data for a @ref PACE2_ADVANCED_TELNET_COMMAND event. */
        struct ipd_telnet_command command;

        /** Contains data for a @ref PACE2_ADVANCED_TELNET_MARKED_COMMAND event. */
        struct ipd_telnet_marked_command marked_command;

        /** Contains data for a @ref PACE2_ADVANCED_TELNET_DATA event. */
        struct ipd_telnet_data data;

    } event_data;

} PACE2_advanced_telnet_event;

#endif

#endif // TELNET_EVENTS_H

/*********************************************************************************
 ** EOF
 ********************************************************************************/
