/** @file      voip_events.h
  * @brief     VoIP event definitions.
  * @details   @copydoc doc_class_voip
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef VOIP_EVENTS_H
#define VOIP_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "generic_events.h"

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the VoIP class event.
 */
#define PACE2_CLASS_VOIP_EVENT_GROUP PACE2_CLASS_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for class VOIP events.
 */
typedef enum {

    /** VoIP session starts: struct ipd_class_voip_session_struct. */
    PACE2_CLASS_VOIP_SESSION,

    /** VoIP call is in buildup or clearing: struct ipd_class_voip_call_struct. */
    PACE2_CLASS_VOIP_CALL,

    /** VoIP call contains RTP Data: mstruct ipd_voip_class_rtp_data. */
    PACE2_CLASS_VOIP_RTP_DATA,

    /** VoIP session ends: struct ipd_class_voip_session_end_struct . */
    PACE2_CLASS_VOIP_SESSION_END,

    /** VoIP RTP session starts: struct ipd_voip_class_rtp_session_struct. */
    PACE2_CLASS_VOIP_RTP_SESSION,

    /** VoIP RTP session ends: struct ipd_voip_class_rtp_session_end_struct. */
    PACE2_CLASS_VOIP_RTP_SESSION_END,

    /** RTP flow of VoIP got an invalid packet: struct ipd_voip_class_rtp_invalid_struct. */
    PACE2_CLASS_VOIP_INVALID,

    /** VoIP flow got an invalid packet: struct ipd_class_voip_invalid_struct. */
    PACE2_CLASS_VOIP_RTP_INVALID,

    /** VoIP flow got a SSRC change: struct ipd_voip_class_ssrc_change_struct. */
    PACE2_CLASS_VOIP_RTP_SSRC_CHANGE,

    /** Session update event for AMEs which do not effect any CE. Is active in the correlation feature only.
     * Points to a struct ipd_class_voip_session_update_struct. */
    PACE2_CLASS_VOIP_SESSION_UPDATE,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_CLASS_VOIP_EVENT_TYPES

} PACE2_class_VOIP_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/** Enumeration for the VoIP protocol type. */
enum ipd_voip_class_type {

    /** Indicates a SIP VoIP class type. */
    IPD_VOIP_PROTOCOL_CLASS_SIP,

    /** Indicates a H.323 VoIP class type. */
    IPD_VOIP_PROTOCOL_CLASS_H323,

    /** Indicates a JABBER VoIP class type. */
    IPD_VOIP_PROTOCOL_CLASS_JABBER,

    /** Indicates a RTP only VoIP class type. */
    IPD_VOIP_PROTOCOL_CLASS_RTP
};

/**
 * Enumeration for the states of a VoIP call (Dialog).
 *
 * When changing this enum, don't forget to update the conversion table to chat AV states above
 * chat_sip_state_cb().
 */
enum ipd_voip_class_call_status {

    /** Will be used after a SIP-(Call-)Request is sent (INVITE). */
    IPD_VOIP_CLASS_CALL_ANNOUNCED,

    /** Will be used after the INVITE is rejected either by the user or a busy signal (INVITE FAIL). */
    IPD_VOIP_CLASS_CALL_REJECTED,

    /** Will be used when TRYING is send (RINGING). */
    IPD_VOIP_CLASS_CALL_ACKNOWLEDGED,

    /** Will be used when the caller cancels the call. */
    IPD_VOIP_CLASS_CALL_CANCELLED,

    /** Will be used when the buildup of the call is complete. */
    IPD_VOIP_CLASS_CALL_ESTABLISHED,

    /** Will be used when the clearing of the call is complete. */
    IPD_VOIP_CLASS_CALL_ENDED,
};

#define VOIP_CALL_STATUS_NAMES \
    "announced",               \
    "rejected",                \
    "acknowledged",            \
    "cancelled",               \
    "established",             \
    "ended"

/**
 * Structure for the VoIP session.
 */
struct ipd_class_voip_session_struct {

    /** The generic. Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The connection ID of the SIP flow. */
    u64 connection_id;

    /** The Class VoIP type of this session. */
    enum ipd_voip_class_type voip_type;
};

/**
 * Structure for the end of a VoIP session.
 */
struct ipd_class_voip_session_end_struct {

    /** The generic. Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The connection ID of the SIP flow. */
    u64 connection_id;

    /** The ID of the VoIP session of this flow. */
    u64 voip_session_id;

    /** The list of lost ID lists of the event correlation. */
    pace2_lost_id_lists * lost_id_lists;
};

/**
 * Structure for an invalid packet in the VoIP session.
 */
struct ipd_class_voip_invalid_struct {

    /** The generic. Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The connection ID of the SIP flow. */
    u64 connection_id;

    /** The ID of the VoIP session of this flow. */
    u64 voip_session_id;
};

/**
 * Structure for a VoIP call event.
 */
struct ipd_class_voip_call_struct {

    /** The generic. Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The ID of the VoIP session. */
    u64 voip_session_id;

    /** The current state of the call @see ipd_voip_class_call_status. */
    enum ipd_voip_class_call_status status;

    /** The display name of the caller. */
    charset_string from_name;

    /** The URI of the caller. */
    charset_string from_uri;

    /** The display name of the receiver. */
    charset_string to_name;

    /** The URI of the receiver. */
    charset_string to_uri;

    /** The call ID (Call-ID: header field). */
    charset_string call_id;

    /** The User-Agent: header field of the SIP message triggering the call state
     * change; it is never set in messages from the server, and is optional in
     * client messages; the call update to status==IPD_VOIP_CLASS_CALL_ANNOUNCED
     * is guaranteed to be triggered by a client request (an INVITE). */
    charset_string user_agent;

    /** The time between ESTABLISHED and END. */
    u64 duration;

    /** A flag, which indicates a caller behind a proxy. (set if INVITE request had more than one Via: header). */
    u8 caller_is_behind_proxy;
};

/**
 * Structure for the RTP session in VoIP.
 */
struct ipd_voip_class_rtp_session_struct {

    /** The generic. Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The connection ID of the RTP flow. */
    u64 rtp_connection_id;

    /** The ID of the VoIP session. */
    u64 voip_session_id;

    /** The VoIP call ID. */
    u64 voip_call_id;
};

/**
 * Structure for the end of a RTP session in VoIP.
 */
struct ipd_voip_class_rtp_session_end_struct {

    /** The generic. Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The connection ID of the RTP flow. */
    u64 rtp_connection_id;

    /** The ID of the RTP session. */
    u64 rtp_session_id;

    /** The list of lost ID lists of the event correlation. */
    pace2_lost_id_lists * lost_id_lists;
};

/**
 * Structure for an invalid packet in the RTP session in VoIP.
 */
struct ipd_voip_class_rtp_invalid_struct {

    /** The generic. Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The connection ID of the RTP flow. */
    u64 rtp_connection_id;

    /** The ID of the RTP session. */
    u64 rtp_session_id;
};

/**
 * Structure that contains a single RTP header extension element: its description and raw data.
 */
struct pace2_voip_class_rtp_header_extension_element {

    /** Element description. */
    struct ipd_charset_string element_description;

    /** Element data. */
    struct ipd_charset_string element_data;
};

/**
 * Enumeration that indicates the direction of the camera used for this video stream.
 * It can be used by the MTSI client in receiver to e.g. display the received video differently
 * depending on the source camera.
 */
enum pace2_voip_class_cvo_camera_direction {

    /** Front-facing camera, facing the user. If camera direction is unknown by the sending MTSI client
     * in the terminal then this is the default value used. */
    PACE2_CLASS_EVENT_VOIP_CAMERA_DIRECTION_FRONT_FACING,

    /** Back-facing camera, facing away from the user. */
    PACE2_CLASS_EVENT_VOIP_CAMERA_DIRECTION_BACK_FACING
};

/**
 * Enumeration that indicates a horizontal (left-right flip) mirror operation on the video as sent on the link.
 */
enum pace2_voip_class_cvo_horizontal_flip {

    /** No flip operation. If the sending MTSI client in terminal does not know if a horizontal
     * mirror operation is necessary, then this is the default value used. */
    PACE2_CLASS_EVENT_VOIP_HORIZONTAL_FLIP_NONE,

    /** Horizontal flip operation. */
    PACE2_CLASS_EVENT_VOIP_HORIZONTAL_FLIP_PRESENT
};

/**
 * Structure that contains the information about the coordination of video orientation.
 */
struct pace2_voip_class_coordination_of_video_orientation {

    /** Camera direction. */
    enum pace2_voip_class_cvo_camera_direction camera_direction;

    /** Horizontal flip. */
    enum pace2_voip_class_cvo_horizontal_flip horizontal_flip;

    /** Indicates the rotation of the video as transmitted on the link.
     * The receiver should rotate the video to compensate that rotation.
     * E.g. a 90 degrees Counter Clockwise rotation should be compensated by the receiver with
     * a 90 degrees Clockwise rotation prior to displaying.*/
    double counter_clockwise_video_rotation;
};

/**
 * Structure for all Information of one RTP Packet.
 */
struct ipd_voip_class_rtp_data {

    /** The generic. Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The ID of the VoIP session. */
    u64 voip_session_id;

    /** The VoIP call ID. */
    u64 voip_call_id;

    /** The RTP connection ID (flow ID). */
    u64 rtp_connection_id;

    /** The ID of the RTP session. */
    u64 rtp_session_id;

    /** The payload type of the RTP packet. */
    charset_string encoding_name;

    /** The clock rate of the RTP packet. */
    u32 clock_rate;

    /** The used channels of the RTP packet. */
    u16 channels;

    /** The direction of the packet; 0: from caller, 1: to caller. */
    int direction;

    /** The data payload of the RTP packet (does not include the RTP header). */
    charset_string data;

    /** The SSRC of the RTP-packet. */
    u32 ssrc;

    /** The Sequence number of the RTP packet. */
    u16 sequence_number;

    /** The Timestamp of the RTP packet. */
    u32 timestamp;

    /** Number of RTP header extension elements. */
    u16 extension_elements_len;

    /** Array of RTP header extension elements. */
    struct pace2_voip_class_rtp_header_extension_element * extension_elements;

    /** Defines if the coordination of video orientation information is present. */
    u8 coordination_of_video_orientation_present;

    /** Information about the coordination of video orientation. */
    struct pace2_voip_class_coordination_of_video_orientation coordination_of_video_orientation;

    /** The RTP header of the RTP packet. */
    charset_string header;

    /** RTP payload type value. */
    u8 payload_type;
};

/**
 * Structure for a SSRC change event.
 */
struct ipd_voip_class_ssrc_change_struct {

    /** The generic. Contains the ID of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The ID of the VoIP session. */
    u64 voip_session_id;

    /** The VoIP call ID. */
    u64 voip_call_id;

    /** The RTP connection ID (flow ID). */
    u64 rtp_connection_id;

    /** The ID of the RTP session. */
    u64 rtp_session_id;

    /** The old SSRC pair. */
    u32 old_ssrc[2];

    /** The new SSRC which was sent. */
    u32 new_ssrc;
};

/**
 * @brief Structure for a VoIP session update event.
 */
struct ipd_class_voip_session_update_struct {

    /** The generic. Contains the id of this event. @see ipd_class_generic_header. */
    struct ipd_class_generic_header generic;

    /** The connection id of the SIP flow. */
    u64 connection_id;

    /** The id of the VoIP session of this flow. */
    u64 voip_session_id;
};

/**
 * @brief Data structure for class VOIP events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_CLASS_VOIP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** VOIP event type. */
    PACE2_class_VOIP_event_type meta_data_type;

    PACE2_CLASS_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_CLASS_VOIP_SESSION event. */
        struct ipd_class_voip_session_struct session;

        /** Contains data for a @ref PACE2_CLASS_VOIP_CALL event. */
        struct ipd_class_voip_call_struct call;

        /** Contains data for a @ref PACE2_CLASS_VOIP_RTP_DATA event. */
        struct ipd_voip_class_rtp_data rtp_data;

        /** Contains data for a @ref PACE2_CLASS_VOIP_SESSION_END event. */
        struct ipd_class_voip_session_end_struct session_end;

        /** Contains data for a @ref PACE2_CLASS_VOIP_RTP_SESSION event. */
        struct ipd_voip_class_rtp_session_struct rtp_session;

        /** Contains data for a @ref PACE2_CLASS_VOIP_RTP_SESSION_END event. */
        struct ipd_voip_class_rtp_session_end_struct rtp_session_end;

        /** Contains data for a @ref PACE2_CLASS_VOIP_INVALID event. */
        struct ipd_class_voip_invalid_struct invalid;

        /** Contains data for a @ref PACE2_CLASS_VOIP_RTP_INVALID event. */
        struct ipd_voip_class_rtp_invalid_struct rtp_invalid;

        /** Contains data for a @ref PACE2_CLASS_VOIP_RTP_SSRC_CHANGE event. */
        struct ipd_voip_class_ssrc_change_struct ssrc_change;

        /** Contains data for a @ref PACE2_CLASS_VOIP_SESSION_UPDATE event. */
        struct ipd_class_voip_session_update_struct session_update;

    } event_data;

} PACE2_class_VOIP_event;

#endif

#endif /* VOIP_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
