/** @file      wsp_events.h
  * @brief     WSP event definitions.
  * @details   @copydoc doc_wsp
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef WSP_EVENTS_H
#define WSP_EVENTS_H

/* Includes *********************************************************************/
#include "common_event_defs.h"

#ifndef PACE_KERNEL_LIB
#include <time.h>
#endif

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced WSP event.
 */
#define PACE2_ADVANCED_WSP_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced WSP events.
 */
typedef enum {

    /** Start of a new flow with WSP traffic; payload: struct ipd_5tuple. */
    PACE2_ADVANCED_WAP_WSP_NEW_CONNECTION ,

    /** Header of a PDU (all fields up to Data); payload: struct ipd_wsp_pdu_header. */
    PACE2_ADVANCED_WAP_WSP_PDU_HEADER,

    /** Data field of a PDU; payload: struct ipd_wsp_pdu_data. */
    PACE2_ADVANCED_WAP_WSP_PDU_DATA,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_ADVANCED_WSP_EVENT_TYPES

} PACE2_advanced_Wsp_event_type;

/* Event Structures *************************************************************/

#ifndef PACE_KERNEL_LIB

/** @anchor pdu_types
    @name PDU type values.
    As defined in WSPSPEC, table 34 (other values in the range 0x01..0x7f are valid).
    @see ipd_wsp_pdu_header::pdu_type
    @{
*/
#define IPD_WSP_PDU_TYPE_CONNECT 0x01
#define IPD_WSP_PDU_TYPE_CONNECT_REPLY  0x02
#define IPD_WSP_PDU_TYPE_REDIRECT 0x03
#define IPD_WSP_PDU_TYPE_REPLY 0x04
#define IPD_WSP_PDU_TYPE_DISCONNECT 0x05
#define IPD_WSP_PDU_TYPE_PUSH 0x06
#define IPD_WSP_PDU_TYPE_CONFIRMED_PUSH 0x07
#define IPD_WSP_PDU_TYPE_SUSPEND 0x08
#define IPD_WSP_PDU_TYPE_RESUME 0x09

// Get PDUs
#define IPD_WSP_PDU_TYPE_GET 0x40
#define IPD_WSP_PDU_TYPE_OPTIONS 0x41
#define IPD_WSP_PDU_TYPE_HEAD 0x42
#define IPD_WSP_PDU_TYPE_DELETE 0x43
#define IPD_WSP_PDU_TYPE_TRACE 0x44

// Post PDUs
#define IPD_WSP_PDU_TYPE_POST 0x60
#define IPD_WSP_PDU_TYPE_PUT 0x61

/** @} */

/** @brief Representation of the non-Data fields of a WSP protocol data unit.
    @details Covers only those PDU fields that are used by the Method Invocation facility (WSPSPEC,
    section 8.2.3), except for acknowledgment headers.
*/
struct ipd_wsp_pdu_header {

    /** @brief Transaction ID.
        @details TID field of the PDU (in connectionless WSP) or WTP TID (in connection-mode WSP).
    */
    u16 tid;

    /** @brief PDU type.
        @details Type field of the PDU (the assigned values from the WSP spec are available as a @ref pdu_types
        "set of defines").
    */
    u8 pdu_type;

    /** @brief Status code.
        @details Status field of the PDU (only present in Reply PDUs, set to 0 if not present). Values
        correspond to HTTP status codes; the exact mapping is defined in WSPSPEC, table 36. */
    u8 reply_status;

    /** @brief Request URI.
        @details URI field of the PDU (only present in Get and Post PDUs, set to a zero-length string if not present).
    */
    struct ipd_string uri;

    /** @brief Type/subtype from the ContentType field of the PDU. */
    struct ipd_string content_type;

    /** @brief Value of the charset parameter (if present) from the ContentType field of the PDU. */
    struct ipd_string content_type_charset;

    /** @brief Pointer to an array of header field descriptions. */
    const struct ipd_wsp_header_field * headers;

    /** @brief Number of elements in @ref ipd_wsp_pdu_header::headers. */
    u64 number_of_headers;

    /** @brief Pointer to an array containing the values of the Content-Encoding header fields,
     * in order of appearance. */
    const struct ipd_string * content_encodings;

    /** @brief Number of elements in @ref ipd_wsp_pdu_header::content_encodings. */
    u64 number_of_content_encodings;

    /** @brief Value of the Content-Length header field.
       @details If this header field is not present, the value ULONG_MAX is used.
    */
    u64 content_length;

    /** @brief First-byte-pos value from the Content-Range header field.
       @details If this header field is not present, the value 0 is used.
    */
    u64 first_byte_pos;

    /** @brief Entity-length value from the Content-Range header field.
        @details If this header field is not present, or does not specify the entity length,
        the value ULONG_MAX is used.
    */
    u64 entity_length;

    /** Value of the host header field. */
    struct ipd_string host;

    /** Value of the location header field. */
    struct ipd_string location;

    /** Value of the referer header field. */
    struct ipd_string referer;

    /** Value of the Transfer-Encoding header field. */
    struct ipd_string transfer_encoding;

    /** Value of the User-Agent header field. */
    struct ipd_string user_agent;

    /** @brief Pointer to an array of cookies.
        @details In server-originated messages, this array represents the Set-Cookie headers; in
        client-originated messages, it represents the Cookie headers.
    */
    const struct ipd_wsp_cookie * cookies;

    /** @brief Number of elements in @ref ipd_wsp_pdu_header::cookies "cookies". */
    u64 number_of_cookies;

    /** @brief Flag that is set if the PDU contains a Data field.
        @details The Data PDU field will be reported in the IPD_WSP_PDU_DATA events
        that follow.
    */
    u8 has_data;

    /** @brief Flag that is set if this PDU is being transmitted from the server to the client. */
    u8 from_server;
};

/** @brief Representation of one segment of the Data field of a WSP protocol data unit. */
struct ipd_wsp_pdu_data {

    /** @copydoc ipd_wsp_pdu_header::tid */
    u16 tid;

    /** @copydoc ipd_wsp_pdu_header::pdu_type */
    u8 pdu_type;

    /** @brief A segment of the Data field of the PDU.
        @details The segments are reported in-order, ie. the Data field of the WSP message with a given
        transaction ID is the concatenation of the @ref ipd_wsp_pdu_data::content "content" fields
        of all IPD_WSP_PDU_DATA events with that @ref ipd_wsp_pdu_data::tid "tid".
    */
    struct ipd_string content;

    /** @brief Flag that is set if more IPD_WSP_PDU_DATA events for this PDU will follow. */
    u8 more_data;
};

/** @brief Representation of a Cookie or Set-Cookie header field. */

struct ipd_wsp_cookie {

    /** @brief Cookie name. */
    struct ipd_string name;

    /** @brief Cookie value. */
    struct ipd_string value;

    /** @brief Domain attribute. */
    struct ipd_string domain;

    /** @brief Path attribute. */
    struct ipd_string path;

    /** @brief Cookie lifetime in seconds (Max-Age attribute, Set-Cookie only).
        @details If the Max-Age attribute is not present, the value ULONG_MAX is used.
    */
    u64 max_age;

    /** @brief Flag that is set when the Secure attribute is present in a Set-Cookie header. */
    u8 secure;
};

/** @brief List of possible representations of a header field value
 * (selects a branch of the ipd_wsp_header_value union). */
enum ipd_wsp_header_value_type {

    /** @brief String in @ref ipd_wsp_header_value::string "ipd_wsp_header_value.string". */
    IPD_WSP_STRING_VALUE,

    /** @brief String that may contain escaped characters in @ref ipd_wsp_header_value::quoted_string
     * "ipd_wsp_header_value.quoted_string".
        @details Refer to the quoted-string production in RFC 2616 for details.
    */
    IPD_WSP_QUOTED_STRING_VALUE,

    /** @brief Unix timestamp in @ref ipd_wsp_header_value::time "ipd_wsp_header_value.time". */
    IPD_WSP_TIME_VALUE,

    /** @brief Unsigned integer in @ref ipd_wsp_header_value::uint "ipd_wsp_header_value.uint". */
    IPD_WSP_UINT_VALUE,

    /** @brief No "cooked" representation of the field value is available; the raw bytes can be found in
        @ref ipd_wsp_header_field::raw_value "ipd_wsp_header_value.raw_value".
    */
    IPD_WSP_UNTYPED_VALUE
};

/** @brief Union of all supported representations of a header field value (an accompanying @ref
    ipd_wsp_header_value_type tag tells which branch is used).
*/
union ipd_wsp_header_value {

    /** @brief A string. */
    struct ipd_string string;

    /** @brief A string that may contain escaped characters.
     * @details Refer to the quoted-string production in RFC 2616 for details. */
    struct ipd_string quoted_string;

    /** @brief Unix timestamp. */
    time_t time;

    /** @brief Unsigned integer. */
    u64 uint;
};

/** @brief Representation of one header field from a WSP Message. */
struct ipd_wsp_header_field {

    /** @brief Numeric code for the header field name.
        @details Refer to WSPSPEC, table 39 for a list of code assignments for "well-known field names".
        The special value 0xFF is used by PADE to indicate that the field name was not numerically coded, but
        sent as a string (which can be found in @ref ipd_wsp_header_field::name_as_string "name_as_string").
    */
    u8 name_as_code;

    /** @brief Header field name.
        @details This string will always be defined. For well-known headers (where the field name is not
        transmitted as a string over the network), PADE provides a value by looking up
        @ref ipd_wsp_header_field::name_as_code "name_as_code" in an internal table.
    */
    struct ipd_string name_as_string;

    /** @brief Header field value as an uninterpreted byte string. */
    struct ipd_string raw_value;

    /** @brief Type tag that indicates which branch of the @ref ipd_wsp_header_field::value field is used. */
    enum ipd_wsp_header_value_type value_type;

    /** @brief "Cooked" (i.e. typed) representation of the header field value. */
    union ipd_wsp_header_value value;
};

/**
 * @brief Data structure for advanced WSP events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_WAP_WSP_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    /** WSP event subtype. */
    PACE2_advanced_Wsp_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    /** Named union which contains the data of the sub-event types. */
    union {

        /** Contains data for a @ref PACE2_ADVANCED_WAP_WSP_NEW_CONNECTION event. */
        struct ipd_5tuple  new_connection;

        /** Contains data for a @ref PACE2_ADVANCED_WAP_WSP_PDU_HEADER event. */
        struct ipd_wsp_pdu_header pdu_header;

        /** Contains data for a @ref PACE2_ADVANCED_WAP_WSP_PDU_DATA event. */
        struct ipd_wsp_pdu_data pdu_data;

    } event_data;

} PACE2_advanced_wsp_event;

#endif

#endif /* WSP_EVENTS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
