/** @file      yahoo_msgr_events.h
  * @brief     Yahoo Messenger event definitions.
  * @details   @copydoc doc_yahoo_msgr
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef YAHOO_MSGR_EVENTS_H
#define YAHOO_MSGR_EVENTS_H

/* Event Groups *****************************************************************/
/**
 * @brief Public define for the assigned metadata group of the advanced Yahoo Messenger event.
 */
#define PACE2_ADVANCED_YAHOO_MSGR_EVENT_GROUP PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP

/* Includes *********************************************************************/
#include "common_event_defs.h"
#include "rtp_events.h"
#include "sip_events.h"

/* Event Types ******************************************************************/

/**
 * @brief Types for advanced Yahoo Messenger events.
 */
typedef enum {

    /** Provides a struct ipd_charset_string * with the login name of the connecting user. */
    PACE2_ADVANCED_YAHOO_MSGR_AUTH,

    /** Provides a struct ipd_charset_string * with the login name of a buddy on the user's buddy list. */
    PACE2_ADVANCED_YAHOO_MSGR_BUDDY,

    /** Provides a struct ipd_charset_string * with the login name of a buddy who is coming online. */
    PACE2_ADVANCED_YAHOO_MSGR_BUDDY_ON,

    /** Provides a struct ipd_charset_string * with the login name of a buddy who is going offline. */
    PACE2_ADVANCED_YAHOO_MSGR_BUDDY_OFF,

    /** Provides a struct ipd_charset_string * with the login name of a buddy who is becoming 'busy'. */
    PACE2_ADVANCED_YAHOO_MSGR_BUDDY_BUSY,

    /** Signals, that the main user is coming online. */
    PACE2_ADVANCED_YAHOO_MSGR_USER_ON,

    /** Signals, that the main user is becoming busy. */
    PACE2_ADVANCED_YAHOO_MSGR_USER_BUSY,

    /** Signals, that the main user is becoming away. */
    PACE2_ADVANCED_YAHOO_MSGR_USER_AWAY,

    /** Signals, that the main user is becoming invisible for other chat user. */
    PACE2_ADVANCED_YAHOO_MSGR_USER_INVISIBLE,

    /** Provides a struct ipd_yahoo_msgr_simple_message * with information about a private message. */
    PACE2_ADVANCED_YAHOO_MSGR_SIMPLE_MESSAGE,

    /** Provides a struct ipd_yahoo_msgr_chatroom_message * with information about a chatroom being created. */
    PACE2_ADVANCED_YAHOO_MSGR_CHATROOM_NEW,

    /** Provides a struct ipd_yahoo_msgr_chatroom_message * with information the chatroom being joined by the user. */
    PACE2_ADVANCED_YAHOO_MSGR_CHATROOM_JOIN,

    /** Provides a struct ipd_yahoo_msgr_chatroom_message * with information the chatroom being left by the user. */
    PACE2_ADVANCED_YAHOO_MSGR_CHATROOM_EXIT,

    /** Provides a struct ipd_yahoo_msgr_chatroom_message * with information about a message sent to a chatroom. */
    PACE2_ADVANCED_YAHOO_MSGR_CHATROOM_MESSAGE,

    /** Provides a struct ipd_yahoo_msgr_ft_info * with information about a file transfer which is being announced. */
    PACE2_ADVANCED_YAHOO_MSGR_FT_ANNOUNCE,

    /** Provides a struct ipd_yahoo_msgr_ft_info * with information about a file transfer which is being accepted. */
    PACE2_ADVANCED_YAHOO_MSGR_FT_ACCEPT,

    /** Provides a struct ipd_yahoo_msgr_ft_info * with information about a file transfer which is being rejected. */
    PACE2_ADVANCED_YAHOO_MSGR_FT_REJECT,

    /** Provides a struct ipd_yahoo_msgr_ft_info * with information about a file transfer which is being canceled. */
    PACE2_ADVANCED_YAHOO_MSGR_FT_CANCEL,

    /** Provides a struct ipd_yahoo_msgr_ft_info * with information about a file transfer which is staring. */
    PACE2_ADVANCED_YAHOO_MSGR_FT_START,

    /** Provides a struct ipd_charset_string * with a segment of file transfer data for the current file transfer
     * connection. */
    PACE2_ADVANCED_YAHOO_MSGR_FT_PAYLOAD,

    /** Provides a struct ipd_sip_message * with a SIP message which is related to a Yahoo Messenger
     * voice over ip session. This event is actually issued by the SIP decoder. */
    PACE2_ADVANCED_YAHOO_MSGR_SIP_MESSAGE,

    /** Provides a struct ipd_rtp_data * with a RTP packet which is related to a Yahoo Messenger voice over ip session.
      * This event is actually issued by the RTP decoder. */
    PACE2_ADVANCED_YAHOO_MSGR_RTP_DATA,

    /** This event is issued when a Yahoo Messenger connection ends. It does not provide any data. */
    PACE2_ADVANCED_YAHOO_MSGR_END_OF_FLOW,

    /** This event is issued when decoding of a Yahoo Messenger connection fails. It provides a char const * to
     * a message describing the failure reason. */
    PACE2_ADVANCED_YAHOO_MSGR_INVALID,

    /** This event is issued when an SSRC value is found. */
    PACE2_ADVANCED_YAHOO_MSGR_RTP_SSRC,

    /** RTP end of flow. */
    PACE2_ADVANCED_YAHOO_MSGR_RTP_END_OF_FLOW,

    /** RTP invalid. */
    PACE2_ADVANCED_YAHOO_MSGR_RTP_INVALID,

    /** Number of implemented event types. */
    PACE2_NUMBER_OF_YAHOO_MSGR_EVENTS

} PACE2_advanced_yahoo_msgr_event_type;

#ifndef PACE_KERNEL_LIB

/** Structure for events that provide information about a Yahoo Messenger instant message. */
struct ipd_yahoo_msgr_simple_message {

    /** The actual message text. */
    struct ipd_charset_string message;

    /** The user the message is from. */
    struct ipd_charset_string sender;

    /** The user the message is sent to. */
    struct ipd_charset_string receiver;
};

/** Structure for events that provide information about an interaction between
 * a Yahoo Messenger user and a chatroom. */
struct ipd_yahoo_msgr_chatroom_event {

    /** The name of the chatroom */
    struct ipd_charset_string chatroom;

    /** The user interacting with a chatroom. */
    struct ipd_charset_string user;
};

/** Structure for events that provide information about a message sent in a Yahoo Messenger chatroom. */
struct ipd_yahoo_msgr_chatroom_message {

    /** The name of the chatroom */
    struct ipd_charset_string chatroom;

    /** The user sending the message to the chatroom. */
    struct ipd_charset_string sender;

    /** The actual message text sent to the chatroom. */
    struct ipd_charset_string message;
};

/** Structure for events that provide information about a Yahoo Messenger file transfer. */
struct ipd_yahoo_msgr_ft_info {

    /** A unique ID describing the file transfer. */
    struct ipd_charset_string transfer_id;

    /** The connected user of the main Yahoo Messenger connection. */
    struct ipd_charset_string user;

    /** The name of the chatroom the file transfer belongs to. */
    struct ipd_charset_string chatroom;

    /** The user sending the file. */
    struct ipd_charset_string sender;

    /** The user receiving the file. */
    struct ipd_charset_string receiver;

    /** The name of the transferred file. */
    struct ipd_charset_string filename;

    /** The size of the transferred file in bytes. */
    u32 file_size;
};

/**
 * @brief Data structure for advanced Yahoo Messenger events.
 */
typedef struct {

    /** Global event type, has to be @ref PACE2_ADVANCED_YAHOO_MSGR_EVENT for this structure. */
    PACE2_EVENT_COMMON_FIELDS;

    PACE2_advanced_yahoo_msgr_event_type meta_data_type;

    PACE2_ADVANCED_META_DATA_EVENT_EXTRA_FIELDS;

    union {

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_AUTH event. */
        struct ipd_charset_string auth;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_BUDDY event. */
        struct ipd_charset_string buddy;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_BUDDY_ON event. */
        struct ipd_charset_string buddy_on;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_BUDDY_OFF event. */
        struct ipd_charset_string buddy_off;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_BUDDY_BUSY event. */
        struct ipd_charset_string buddy_busy;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_SIMPLE_MESSAGE event. */
        struct ipd_yahoo_msgr_simple_message simple_message;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_CHATROOM_NEW event. */
        struct ipd_yahoo_msgr_chatroom_message chatroom_new;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_CHATROOM_JOIN event. */
        struct ipd_yahoo_msgr_chatroom_event chatroom_join;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_CHATROOM_EXIT event. */
        struct ipd_yahoo_msgr_chatroom_event chatroom_exit;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_CHATROOM_MESSAGE event. */
        struct ipd_yahoo_msgr_chatroom_message chatroom_message;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_FT_ANNOUNCE event. */
        struct ipd_yahoo_msgr_ft_info ft_announce;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_FT_ACCEPT event. */
        struct ipd_yahoo_msgr_ft_info ft_accept;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_FT_REJECT event. */
        struct ipd_yahoo_msgr_ft_info ft_reject;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_FT_CANCEL event. */
        struct ipd_yahoo_msgr_ft_info ft_cancel;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_FT_START event. */
        struct ipd_yahoo_msgr_ft_info ft_start;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_FT_PAYLOAD event. */
        struct ipd_charset_string ft_payload;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_SIP_MESSAGE event. */
        struct ipd_sip_message sip_message;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_RTP_DATA event. */
        struct ipd_rtp_data rtp_data;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_INVALID event. */
        char const * error_message;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_RTP_INVALID event. */
        char const *rtp_invalid;

        /** Contains data for a @ref PACE2_ADVANCED_YAHOO_MSGR_RTP_SSRC event. */
        struct ipd_rtp_ssrc rtp_ssrc;

    } event_data;

} PACE2_advanced_yahoo_msgr_event;

#endif

#endif // YAHOO_MSGR_EVENTS_H

/*********************************************************************************
 ** EOF
 ********************************************************************************/
