/** @file      ipq_public_functions.h
  * @brief     PACE public functions definition.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef __IPOQUE_API_INCLUDE_FILE__
#error CANNOT INCLUDE THIS .H FILE, INCLUDE IPQ_API.H
#endif

#ifndef __IPQ_PUBLIC_FUNCTIONS_H__
#define __IPQ_PUBLIC_FUNCTIONS_H__

#ifdef __cplusplus
extern "C" {
#endif

/* get flow and id size */

/* PUBLIC FUNCTIONS, API DOCUMENTATION IN MANPAGES ONLY !! */

	/**
	 * struct for a unique ipv4 flow address
	 */
	typedef struct ipoque_unique_flow_ipv4_address_struct {
		/**
		 * lower ip
		 */
		u32 lower_ip;
		/**
		 * upper ip
		 */
		u32 upper_ip;
		/* we need 3 dummies to fill up to ipv6 address size */
		/**
		 * padding to have same size as the unique ipv6 struct
		 */
		u64 dummy[3];
	} ipoque_unique_flow_ipv4_address_struct_t;

	/**
	 * struct for a unique ipv6 flow address
	 */
	typedef struct ipoque_unique_flow_ipv6_address_struct {
		/**
		 * lower ip
		 */
		u64 lower_ip[2];
		/**
		 * upper ip
		 */
		u64 upper_ip[2];
	} ipoque_unique_flow_ipv6_address_struct_t;

	/**
	 * struct for a unique ipv4 and ipv6 5-tuple (ip,ip,port,port,protocol)
	 */
	typedef struct ipoque_unique_flow_ipv4_and_6_struct {
		/* only ip addresses are different, to minimize compare operations for hash tables, store ipv4 or ipv6 always in the first bit */
		/**
		 * saves if it is a ipv6, if it false so it is a ipv4
		 */
		u16 is_ip_v6;
		/**
		 * the protocol, 16 bit wide for alignment reasons
		 */
		u16 protocol;			/* for alignment reason, protocol is 16 bit, not 8 bit */
		/**
		 * the port of the lower ip address
		 */
		u16 lower_port;
		/**
		 * the port of the upper ip address
		 */
		u16 upper_port;
		union {
			/**
			 * the ipv4 flow address struct. use the same memory area like ipv6 (union)
			 */
			struct ipoque_unique_flow_ipv4_address_struct ipv4;
			/**
			 * the ipv6 flow address struct. use the same memory area like ipv4 (union)
			 */
			struct ipoque_unique_flow_ipv6_address_struct ipv6;
		} ip;
	} ipoque_unique_flow_ipv4_and_6_struct_t;

#define MEDIA_LIST(M) M(NOT_SET,                        The HTTP content type is not set.) \
                      M(UNKNOWN,                        The HTTP content type is unknown.) \
                      M(MPEG,                           The HTTP content type is MPEG.) \
                      M(FLASH,                          The HTTP content type is FLASH.) \
                      M(QUICKTIME,                      The HTTP content type is QUICKTIME.) \
                      M(REALMEDIA,                      The HTTP content type is REALMEDIA.) \
                      M(WINDOWSMEDIA,                   The HTTP content type is WINDOWSMEDIA.) \
                      M(CITRIX,                         The HTTP content type is Citrix.) \
                      M(MICROSOFT_MEDIA_STREAMING,      The HTTP content type is Microsoft Media Streaming (MMS).) \
                      M(MULTIMEDIA_MESSAGE,             The HTTP content type is Multimedia Messaging Service.) \
                      M(OFF,                            The HTTP content type is OFF.) \
                      M(OGG,                            The HTTP content type is OGG.) \
                      M(MP3,                            The HTTP content type is MP3.) \
                      M(AVI,                            The HTTP content type is AVI.) \
                      M(MOVE,                           The HTTP content type is MOVE.) \
                      M(MATROSKA,                       The HTTP content type is Matroska.) \
                      M(AUDIO_INTERCHANGE_FILE,         The HTTP content type is Audio Interchange File.) \
                      M(THREE_GPP,                      The HTTP content type is 3GPP.) \
                      M(WEBM,                           The HTTP content type is WebM.) \
                      M(SOAP,                           The HTTP content type is SOAP.)

#define LIST_TO_MEDIA_ENUM(NAME, ...) PACE_CT_ ## NAME,
#define LIST_TO_MEDIA_STR(NAME, ...) #NAME,

    enum pace_http_content_type {
        MEDIA_LIST(LIST_TO_MEDIA_ENUM)

        NUMBER_OF_MEDIA_TYPES
    };

	/**
	 * This is the struct for storing information about the latency of a flow.
	 */
	typedef struct flow_latency_struct {
		/**
		 * diff_syn_synack_possible a flag that shows if a syn command and the current synack was seen
		 */
		u32 diff_syn_synack_possible:1;
		/**
		 * diff_synack_ack_possible a flag that shows if a synack command and the current ack was seen
		 */
		u32 diff_synack_ack_possible:1;
		/**
		 * diff_syn_synack time difference between syn and synack command
		 */
		IPOQUE_TIMESTAMP_COUNTER_SIZE diff_syn_synack;
		/**
		 * diff_synack_ack time difference between synack and ack command
		 */
		IPOQUE_TIMESTAMP_COUNTER_SIZE diff_synack_ack;
        /**
         * @brief time difference between syn and synack command with higher resolution
         */
        u64 diff_syn_synack_hres;
        /**
         * @brief difference between synack and ack command with higher resolution
         */
        u64 diff_synack_ack_hres;
	} flow_latency_struct_t;

	/**
	 * Struct containing counter for tcp retransmission and ooo-packets of a flow
	 */
	typedef struct ipoque_flow_ooo_counter_struct {
		/**
		 * flow_ooo_counter number of out of order packets with packet->packet_payload_len > 0
		 */
		u16 flow_ooo_counter;

		/**
		 * flow_retransmission_counter number of retransmission packets with packet->packet_payload_len > 0
		 */
        u16 flow_retransmission_counter;  //DEPRECATED
	} ipoque_flow_ooo_count_struct_t;

	enum ipoque_asymmetric_connection_status_enum {
		IPOQUE_CONNECTION_STATUS_SYMMETRIC = 0,
		IPOQUE_CONNECTION_STATUS_ASYMMETRIC = 1,
		IPOQUE_CONNECTION_STATUS_UNKNOWN = 2,
	};

    /**
     * This structure is deprecated and will be removed completely in the future. The information provided by this
     * structure is moved to RTP dissector result 'dissector_result.rtp.statistics' in the
     * 'PACE2_dissector_metadata_event'.
     */
    struct ipoque_rtp_flow_stats_struct {
        u32 packets_sent[2];    /**< number of packets sent, for each direction */
        u32 packets_lost[2];    /**< number of packets lost, for each direction */
        u32 current_jitter[2];      /**< jitter, for each direction */
        u32 current_frequency[2];   /**< frequency used for jitter calculation, for each direction */
        u8 payload_type;        /**< payload type of most recent packet */
        u32 packets_overhead[2]; /**< number of overhead packets for each direction */
        u32 current_jitter_ms[2]; /**< RTP timestamp in millisamples. ( [ms] = [mS] / Hz ) */
    };

	/** structure containing the data from a reception report block in an RTCP sender or receiver report
     * packet (see RFC3550, section 6.4, for the meaning of the fields)
     *
     * all fields use host by order */
    struct ipoque_rtcp_reception_report_struct {
        u32 direction;          /**< direction of the SR/RR packet */
        u8 fraction_lost;       /**< fraction of RTP packets lost since last SR or RR (scaled to 0..255) */
        u32 packets_lost;       /**< number of RTP packets since beginning of reception */
        u32 jitter;             /**< interarrival jitter */
        u32 lsr;                /**< last sender report timestamp */
        u32 dlsr;               /**< delay since (reception of) last sender report */
    };

#define IPOQUE_MAX_PARSE_SSL_CLIENT_HELLO_ALPN 10
#define IPOQUE_MAX_PARSE_SSL_DNS_ALT_NAMES 50

	enum ipoqe_defrag_checksum_enum {
		IPOQUE_DEFRAG_CHECKSUM_IGNORE_WRONG_IP_CHECKSUM,
		IPOQUE_DEFRAG_CHECKSUM_DISCARD_WRONG_IP_CHECKSUM,
		IPOQUE_DEFRAG_CHECKSUM_MARK_WRONG_IP_CHECKSUM_AS_WRONG,
	};

	/************************************************
	 * API and library version                      *
	 ************************************************/

	typedef struct ipoque_pace_api_version {
		/**
		 * API version, increased whenever the API changes
		 */
		u32 api_version;
	} ipoque_pace_api_version_t;

	/**
	 * query API version. The API version will be increased if the API
	 * has been changed, i.e., some public functions or structures have
	 * been changed.
	 *
	 * @return API version structure
	 */
	ipoque_pace_api_version_t ipoque_pace_get_api_version(void);

	typedef struct ipoque_pace_version {
		/**
		 * the major PACE version
		 */
		u16 major_version;

		/**
		 * the minor PACE version
		 */
		u16 minor_version;

		/**
		 * an optional patch version
		 */
		u16 patch_version;

		/**
		 * a string representation "MAJOR.MINOR.PATCH"
		 */
		char version_string[32];

		/**
		 * same as version_string but a build ID may be appended
		 */
		char build_string[32];
	} ipoque_pace_version_t;

	/**
	 * the statistic structure for the Session ID tracker
	 */
	struct ipoque_pace_sit_stats {
		/**
		 * number of session ID inserts
		 */
		u64 id_inserts;

		/**
		 * number of successful ID lookup
		 */
		u64 id_lookup_successful;

		/**
		 * number of failed ID lookups
		 */
		u64 id_lookup_failed;

		/**
		 * this is the maximum number of session IDs that can be stored
		 */
		unsigned long maximum_number_of_elements;

		/**
		 * this is the number of currently stored session IDs
		 */
		unsigned long currently_used_elements;

		/**
		 * this is the timestamp of the oldest entry in the session ID
		 * table. It gives a hint about which period of time is
		 * covered by the session ID tracker.
		 */
		IPOQUE_TIMESTAMP_COUNTER_SIZE ts_of_oldest_entry;
	};

	enum ipoque_client_server_indication_packet_type {
		IPOQUE_DIRECTION_UNKNOWN = 0,       /**< packet direction could not be detected */
		IPOQUE_DIRECTION_NOT_YET_DECIDED,   /**< packet direction is not yet decided */
		IPOQUE_CLIENT_TO_SERVER,            /**< packet goes from client to server */
		IPOQUE_SERVER_TO_CLIENT,            /**< packet goes from server to client */
		IPOQUE_CLIENT_TO_CLIENT,            /**< packet is between two clients */
		IPOQUE_SERVER_TO_SERVER             /**< packet is between two servers */
	};

	enum ipoque_client_server_indication_host_type {
		IPOQUE_HOST_TYPE_UNKNOWN = 0,       /**< it is not known whether host is client or server */
		IPOQUE_HOST_IS_SERVER,              /**< the host is mainly used as a server */
		IPOQUE_HOST_IS_CLIENT,              /**< the host is a client */
	};

	struct ipoque_pace_client_server_indication_host_status {
		enum ipoque_client_server_indication_host_type host_type;     /**< host type enum */
		int percentage_of_client_connections;                         /**< percentage of client connections, 0 means no client connection or value not known,
																		 anything between 1 and 100 gives the rate of client connections over server connections.
																		 The value will change over time. */
	};

    enum ipoque_fastpath_state {
        IPOQUE_FP_NO_PACKET_NEEDED = 0,                           /**< no packet from this flow is needed anymore */
        IPOQUE_FP_NEXT_PACKET_NEEDED = 1,                         /**< the next packet of this flow is needed */
        IPOQUE_FP_NEXT_PACKET_WITH_CONSTRAINT = 2,                /**< the next packet only with additional constraint (direction and/or time) */
    };

    enum ipoque_fastpath_direction_constraint {
        IPOQUE_FP_ANY_DIRECTION = 0,                              /**< no direction limit applies */
        IPOQUE_FP_NEXT_PAYLOAD_PACKET_SAME_DIRECTION = 1,         /**< the next packet with payload for the current direction is needed */
        IPOQUE_FP_NEXT_PAYLOAD_PACKET_OTHER_DIRECTION = 2,        /**< the next packet with payload for the other direction is needed */
    };

	struct ipoque_fastpath_information
	{
		enum ipoque_fastpath_state state;                         /**< the state of the fastpath stating whether additional packets are required */
		u8 skip_time_set;                                         /**< 1 if a time is set until all packets can be skipped, 0 if no time limit applies */
		u32 skip_until_time;                                      /**< time until all packets can be skipped for the state IPOQUE_FP_NEXT_PACKET_WITH_CONSTRAINT */
		enum ipoque_fastpath_direction_constraint direction;      /**< only packets with given direction */
	};

    typedef struct ipoque_pace_license_information {
        /**
         * @brief init_error_code internal initialisation error code
         */
        u32 init_error_code;
        /**
         * @brief init_error_reason verbose initialisation error string
         */
        const char *init_error_reason;
        /**
         * @brief load_error_code internal license loading error code
         */
        u32 load_error_code;
        /**
         * @brief load_error_reason verbose license loading error string
         */
        const char *load_error_reason;
        /**
         * @brief validation_error_code internal validation error code
         */
        u32 validation_error_code;
        /**
         * @brief validation_error_reason verbose validation error string
         */
        const char *validation_error_reason;
        /**
         * @brief limitation_error_code internal limitation error code
         */
        u32 limitation_error_code;
        /**
         * @brief limitation_error_reason verbose limitation error string
         */
        const char *limitation_error_reason;
        /**
         * @brief no_of_mac_addresses_found number of valid mac addresses on the system
         */
        u32 no_of_mac_addresses_found;
        /**
         * @brief current_percentage_bandwidth_limit_usage percentage value which shows the current bandwidth in relation to the bandwidth limit
         */
        u32 current_percentage_bandwidth_limit_usage;

        /**
        * @brief no_of_sha_addresses_found number of valid sha addresses on the system
        */
        u32 no_of_sha_addresses_found;

    } ipoque_pace_license_information_t;

    #define PACE_MAX_NUMBER_OF_HTTP_REQUEST_KEY_VALUE_PAIRS 30

    struct pace_byte_buffer {
        const u8 * ptr;
        u32 len;
    };

    struct pace_http_request_kv_pair_result {
        struct {
            struct pace_byte_buffer key;                  /**< Key string */
            struct pace_byte_buffer value;                /**< Value string */
        } kv_pair[PACE_MAX_NUMBER_OF_HTTP_REQUEST_KEY_VALUE_PAIRS];

        struct pace_byte_buffer *search_key;              /**< Search key string, may be NULL */
        struct pace_byte_buffer *search_key_value;        /**< Search key value, may be NULL */

        u8 number_of_kv_pairs;                            /**< Number of parsed key/value pairs */
    };


#ifdef __cplusplus
}
#endif
#endif /* __IPQ_PUBLIC_FUNCTIONS_H__ */
