/** @file      pace2_application_categories.h
  * @brief     PACE 2 application category definitions
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_APPLICATION_CATEGORIES_H
#define PACE2_APPLICATION_CATEGORIES_H

/* Includes *********************************************************************/
#include "pace2_basic_def.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Types *****************************************************************/

/**
 * @brief Enum defining all available application attributes.
**/
typedef enum pace2_application_category {
    PACE2_APPLICATION_CATEGORY_NONE = 0,
    PACE2_APPLICATION_CATEGORY_PROVIDER,
    PACE2_APPLICATION_CATEGORY_B2C_APPLICATION,
    PACE2_APPLICATION_CATEGORY_B2B_SERVICE,
    PACE2_APPLICATION_CATEGORY_PROVIDER_B2C_APP,
    PACE2_APPLICATION_CATEGORY_B2C_APP_B2B_SERVICE,
    PACE2_APPLICATION_CATEGORY_PROVIDER_B2B_SERVICE,
    PACE2_APPLICATION_CATEGORY_PROVIDER_B2C_APP_B2B_SERVICE,

    PACE2_APPLICATION_CATEGORIES_COUNT

} PACE2_application_categories;


/**
 * @brief Struct representing all available application categories as a bitmask.
**/
typedef struct {
#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
    u64 none:1;                          //!< none
    u64 provider:1;                      //!< provider
    u64 b2c_application:1;               //!< b2c_application
    u64 b2b_service:1;                   //!< b2b_service
    u64 provider_b2c_app:1;              //!< provider_b2c_app
    u64 b2c_app_b2b_service:1;           //!< b2c_app_b2b_service
    u64 provider_b2b_service:1;          //!< provider_b2b_service
    u64 provider_b2c_app_b2b_service:1;  //!< provider_b2c_app_b2b_service
#elif __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
    u64 provider_b2c_app_b2b_service:1;  //!< provider_b2c_app_b2b_service
    u64 provider_b2b_service:1;          //!< provider_b2b_service
    u64 b2c_app_b2b_service:1;           //!< b2c_app_b2b_service
    u64 provider_b2c_app:1;              //!< provider_b2c_app
    u64 b2b_service:1;                   //!< b2b_service
    u64 b2c_application:1;               //!< b2c_application
    u64 provider:1;                      //!< provider
    u64 none:1;                          //!< none
#endif

} PACE2_application_categories_struct;

/**
 * List of strings, each identifying the appropriate application category.
 *
 * Items do not change between releases. Only for new category strings are added.
 */
#define PACE2_APPLICATION_CATEGORY_STRINGS \
    "none",\
    "provider",\
    "b2c-application",\
    "b2b-service",\
    "provider-b2c-app",\
    "b2c-app-b2b-service",\
    "provider-b2b-service",\
    "provider-b2c-app-b2b-service"

/**
 * List of strings, each describing the appropriate application attribute.
 *
 * Items may be changed between releases.
 */
#define PACE2_APPLICATION_CATEGORY_DESCRIPTION \
    "Application has no category.",\
    "Companies that publish B2C applications. This category covers common patterns of some or all applications published by the same provider.",\
    "Applications that end users can download to their devices, such as mobile phones, desktop computers or game consoles.",\
    "Services which are provided to B2C applications. The traffic produced while using some B2C application can include traffic from B2B services. Provided services include CDNs, clouds, advertisement, analytics, and user management.",\
    "Providers that publish both B2C applications and applications without a dedicated ID. B2C applications produce only some traffic that is classified as the provider, whereas the traffic produced by applications without a dedicated ID is classified completely as provider traffic.",\
    "B2C applications that are simultaneously B2B services.",\
    "Providers that are simultaneously B2B services.",\
    "Providers that provide B2B services and also publish both B2C applications and applications without a dedicated ID. B2C applications produce only some traffic that is classified as the provider, whereas the traffic produced by applications without a dedicated ID is classified completely as provider traffic."

/* Public Prototypes ************************************************************/

#include "pace2_applications.h"

/**
 * @brief This function provides the string for the given application category.
 *
 * The function supports the dynamic upgrade feature. If the library is upgraded
 * to a newer version the function implicitly calls it's upgrade variant.
 *
 * @param[in]   pace2_module            Constant pointer to the PACE 2 module
 *                                      Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID               The ID of the thread currently processing.
 *                                      Only used in case of dynamic upgrade otherwise it can be 0.
 * @param[in]   category_id             The application category ID.
 * @returns     A char pointer to the application category string or NULL when one of the input
 *              parameters is invalid.
 */
const char * pace2_get_application_category_string(PACE2_module * const pace2_module, int const thread_ID,
                                                   PACE2_application_categories const category_id);


/**
 * @brief This function provides the description for the given application category.
 *
 * The function supports the dynamic upgrade feature. If the library is upgraded
 * to a newer version the function implicitly calls it's upgrade variant.
 *
 * @param[in]   pace2_module            Constant pointer to the PACE 2 module.
 *                                      Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID               The ID of the thread currently processing.
 *                                      Only used in case of dynamic upgrade otherwise it can be 0.
 * @param[in]   category_id             The application category ID.
 * @returns     The application category description or NULL if one of the input parameters is invalid.
 */
const char * pace2_get_application_category_description(PACE2_module * const pace2_module, int const thread_ID,
                                                        PACE2_application_categories const category_id);

/**
 * @brief This function provides the category by the given application.
 *
 * The function supports the dynamic upgrade feature. If the library is upgraded
 * to a newer version the function implicitly calls it's upgrade variant.
 *
 * @param[in]   pace2_module            Constant pointer to the PACE 2 module.
 *                                      Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID               The ID of the thread currently processing.
 *                                      Only used in case of dynamic upgrade otherwise it can be 0.
 * @param[in]   application_id          The application ID.
 * @param[out]  category                Pointer to category of type
 *                                      @ref PACE2_application_categories "PACE2_application_categories".
 * @retval      PACE2_SUCCESS           @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_FAILURE           @copydoc PACE2_return_state::PACE2_FAILURE
 *                                      The given application id does not exist.
 */
PACE2_return_state pace2_get_category_of_application(PACE2_module * const pace2_module,
                                                     int const thread_ID,
                                                     PACE2_applications const application_id,
                                                     PACE2_application_categories * category);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_APPLICATION_CATEGORIES_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
