/** @file      pace2_attributes_bitmask.h
  * @brief     PACE 2 attributes bitmask definitions and operations.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_ATTRIBUTES_BITMASK_H
#define PACE2_ATTRIBUTES_BITMASK_H

/* Includes *********************************************************************/
#include "pace2_basic_def.h"
#include "pace2_attributes.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Types *****************************************************************/
#define PACE2_APPLICATION_ATTRIBUTES_BITMASK_LENGTH 1
#define PACE2_PROTOCOL_ATTRIBUTES_BITMASK_LENGTH 2

/**
 * @brief Structure contains the bitmask representation of PACE 2 application attributes
 *
 * @deprecated This structure is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 */
typedef struct {
    u64 bitmask[PACE2_APPLICATION_ATTRIBUTES_BITMASK_LENGTH]; /**!< Array to hold the application attribute bitmask. */
} PACE2_application_attributes_bitmask;

/**
 * @brief Structure contains the bitmask representation of PACE 2 protocol attributes
 *
 * @deprecated This structure is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 */
typedef struct {
    u64 bitmask[PACE2_PROTOCOL_ATTRIBUTES_BITMASK_LENGTH]; /**!< Array to hold the protocol attribute bitmask. */
} PACE2_protocol_attributes_bitmask;

/**
 * @brief Union combining the application attribute bitmask and the structural representation of application attributes.
 *
 * @deprecated This union is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 */
typedef union {
    PACE2_application_attributes_struct attributes;   /**!< Structural representation of application attributes. */
    PACE2_application_attributes_bitmask bitmask;     /**!< Bitmask representing applications attributes. */
} PACE2_application_attributes_mask;

/**
 * @brief Union combining the protocol attribute bitmask and the structural representation of protocol attributes.
 *
 * @deprecated This union is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 */
typedef union {
    PACE2_protocol_attributes_struct attributes;   /**!< Structural representation of protocol attributes. */
    PACE2_protocol_attributes_bitmask bitmask;     /**!< Bitmask representing protocol attributes. */
} PACE2_protocol_attributes_mask;

/* Public Macros ****************************************************************/

/**
 * @brief Converts a given value to a bitmask.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[out]  bmask   Destination bitmask which is used to store the value.
 *                      Has to a variable of type @ref PACE2_application_attributes_bitmask
 * @param[in]   value   Value which will be converted to a bitmask.
 */
#define PACE2_APPLICATION_ATTRIBUTES_SAVE_AS_BITMASK(bmask, value)                                                     \
    ((bmask).bitmask[(value) >> 6] = (((u64)1) << ((value)&0x3F)))

/**
 * @brief Converts a given value to a bitmask.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[out]  bmask   Destination bitmask which is used to store the value.
 *                      Has to a variable of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 * @param[in]   value   Value which will be converted to a bitmask.
 */
#define PACE2_PROTOCOL_ATTRIBUTES_SAVE_AS_BITMASK(bmask, value)                                                        \
    do {                                                                                                               \
        (bmask).bitmask[0] = 0;                                                                                        \
        (bmask).bitmask[(value) >> 6] = (((u64)1) << ((value) & 0x3F));                                                \
    } while (0)

/**
 * @brief Compares to bitmasks and returns 1 if they suit each other.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[in]   a   First bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 * @param[in]   b   Second bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 */
#define PACE2_APPLICATION_ATTRIBUTES_BITMASK_MATCH(a,b) (((a).bitmask[0]) == ((b).bitmask[0]))

/**
 * @brief Compares to bitmasks and returns 1 if they suit each other.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[in]   a   First bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 * @param[in]   b   Second bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 */
#define PACE2_PROTOCOL_ATTRIBUTES_BITMASK_MATCH(a,b) (((a).bitmask[0]) == ((b).bitmask[0]) && \
                                                      ((a).bitmask[1]) == ((b).bitmask[1]))

/**
 * @brief Checks if the bitmask a contains the bitmask b.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[in]   a   The bitmask which is checked whether it contains another bitmask.
 *                  Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 * @param[in]   b   The bitmask which is used to check whether another bitmask contains it.
 *                  Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 */
#define PACE2_APPLICATION_ATTRIBUTES_BITMASK_CONTAINS_BITMASK(a,b) (((a).bitmask[0] & (b).bitmask[0]) == (b).bitmask[0])

/**
 * @brief Checks if the bitmask a contains the bitmask b.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[in]   a   The bitmask which is checked whether it contains another bitmask.
 *                  Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 * @param[in]   b   The bitmask which is used to check whether another bitmask contains it.
 *                  Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 */
#define PACE2_PROTOCOL_ATTRIBUTES_BITMASK_CONTAINS_BITMASK(a,b) (((a).bitmask[0] & (b).bitmask[0]) == (b).bitmask[0] && \
                                                                 ((a).bitmask[1] & (b).bitmask[1]) == (b).bitmask[1])

/**
 * @brief Checks if the bitmask a does not contain the bitmask b.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[in]   a   The bitmask which is checked whether it does not contain another bitmask.
 *                  Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 * @param[in]   b   The bitmask which is used to check whether another bitmask does not contain it.
 *                  Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 */
#define PACE2_APPLICATION_ATTRIBUTES_BITMASK_NOT_CONTAINS_BITMASK(a,b)  ((~((a).bitmask[0])) & ((b).bitmask[0]))

/**
 * @brief Checks if the bitmask a does not contain the bitmask b.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[in]   a   The bitmask which is checked whether it does not contain another bitmask.
 *                  Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 * @param[in]   b   The bitmask which is used to check whether another bitmask does not contain it.
 *                  Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 */
#define PACE2_PROTOCOL_ATTRIBUTES_BITMASK_NOT_CONTAINS_BITMASK(a,b)  ((~((a).bitmask[0])) & ((b).bitmask[0]) && \
                                                                      (~((a).bitmask[1])) & ((b).bitmask[1]))

/**
 * @brief Adds a given bitmask b to the bitmask a. Uses a logical or to merge the bitmasks.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 * @param[in]   b   Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 */
#define PACE2_APPLICATION_ATTRIBUTES_BITMASK_ADD(a, b) ((a).bitmask[0] |= (b).bitmask[0])

/**
 * @brief Adds a given bitmask b to the bitmask a. Uses a logical or to merge the bitmasks.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 * @param[in]   b   Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 */
#define PACE2_PROTOCOL_ATTRIBUTES_BITMASK_ADD(a, b)                                                                    \
    do {                                                                                                               \
        (a).bitmask[0] |= (b).bitmask[0];                                                                              \
        (a).bitmask[1] |= (b).bitmask[1];                                                                              \
    } while (0)

/**
 * @brief Uses a logical and to merge the bitmasks a and b. The result is stored in a.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 * @param[in]   b   Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 */
#define PACE2_APPLICATION_ATTRIBUTES_BITMASK_AND(a, b) ((a).bitmask[0] &= (b).bitmask[0])

/**
 * @brief Uses a logical and to merge the bitmasks a and b. The result is stored in a.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 * @param[in]   b   Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 */
#define PACE2_PROTOCOL_ATTRIBUTES_BITMASK_AND(a, b)                                                                    \
    do {                                                                                                               \
        (a).bitmask[0] &= (b).bitmask[0];                                                                              \
        (a).bitmask[1] &= (b).bitmask[1];                                                                              \
    } while (0)

/**
 * @brief Deletes a given bitmask b from the bitmask a.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 * @param[in]   b   Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 */
#define PACE2_APPLICATION_ATTRIBUTES_BITMASK_DEL(a, b) ((a).bitmask[0] &= ~((b).bitmask[0]))

/**
 * @brief Deletes a given bitmask b from the bitmask a.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 * @param[in]   b   Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 */
#define PACE2_PROTOCOL_ATTRIBUTES_BITMASK_DEL(a, b)                                                                    \
    do {                                                                                                               \
        (a).bitmask[0] &= ~((b).bitmask[0]);                                                                           \
        (a).bitmask[1] &= ~((b).bitmask[1]);                                                                           \
    } while (0)

/**
 * @brief Copies the value of bitmask b to bitmask a.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 * @param[in]   b   Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 */
#define PACE2_APPLICATION_ATTRIBUTES_BITMASK_SET(a, b) ((a).bitmask[0] = (b).bitmask[0])

/**
 * @brief Copies the value of bitmask b to bitmask a.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 * @param[in]   b   Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 */
#define PACE2_PROTOCOL_ATTRIBUTES_BITMASK_SET(a, b)                                                                    \
    do {                                                                                                               \
        (a).bitmask[0] = (b).bitmask[0];                                                                               \
        (a).bitmask[1] = (b).bitmask[1];                                                                               \
    } while (0)

/**
 * @brief Resets the value bitmask a to zero.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 */
#define PACE2_APPLICATION_ATTRIBUTES_BITMASK_RESET(a) (((a).bitmask[0]) = 0)

/**
 * @brief Resets the value bitmask a to zero.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 */
#define PACE2_PROTOCOL_ATTRIBUTES_BITMASK_RESET(a)                                                                     \
    do {                                                                                                               \
        ((a).bitmask[0]) = 0;                                                                                          \
        ((a).bitmask[1]) = 0;                                                                                          \
    } while (0)

/**
 * @brief Sets all bits of bitmask a to 1.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 */
#define PACE2_APPLICATION_ATTRIBUTES_BITMASK_SET_ALL(a) (((a).bitmask[0]) = 0xFFFFFFFFFFFFFFFFULL)

/**
 * @brief Sets all bits of bitmask a to 1.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 */
#define PACE2_PROTOCOL_ATTRIBUTES_BITMASK_SET_ALL(a)                                                                   \
    do {                                                                                                               \
        ((a).bitmask[0]) = 0xFFFFFFFFFFFFFFFFULL;                                                                      \
        ((a).bitmask[1]) = 0xFFFFFFFFFFFFFFFFULL;                                                                      \
    } while (0)

/**
 * @brief Adds a given value to a bitmask.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[out]  bmask   Destination bitmask to which the value has to be added.
 *                      Has to a variable of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 * @param[in]   value   Value which will be added to the bitmask.
 */
#define PACE2_ADD_APPLICATION_ATTRIBUTE_TO_BITMASK(bmask, value)                                                       \
    ((bmask).bitmask[(value) >> 6] |= (((u64)1) << ((value) & 0x3F)))

/**
 * @brief Adds a given value to a bitmask.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[out]  bmask   Destination bitmask to which the value has to be added.
 *                      Has to a variable of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 * @param[in]   value   Value which will be added to the bitmask.
 */
#define PACE2_ADD_PROTOCOL_ATTRIBUTE_TO_BITMASK(bmask, value)                                                          \
    ((bmask).bitmask[(value) >> 6] |= (((u64)1) << ((value) & 0x3F)))

/**
 * @brief Deletes a given value from a bitmask.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[out]  bmask   Destination bitmask from which the value has to be deleted.
 *                      Has to a variable of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 * @param[in]   value   Value which will be deleted from the bitmask.
 */
#define PACE2_DEL_APPLICATION_ATTRIBUTE_FROM_BITMASK(bmask, value)                                                     \
    ((bmask).bitmask[(value) >> 6] &= ~(((u64)1) << ((value) & 0x3F)))

/**
 * @brief Deletes a given value from a bitmask.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[out]  bmask   Destination bitmask from which the value has to be deleted.
 *                      Has to a variable of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 * @param[in]   value   Value which will be deleted from the bitmask.
 */
#define PACE2_DEL_PROTOCOL_ATTRIBUTE_FROM_BITMASK(bmask, value)                                                        \
    ((bmask).bitmask[(value) >> 6] &= ~(((u64)1) << ((value) & 0x3F)))

/**
 * @brief Checks whether the bitmask contains the given value.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[out]  bmask   The bitmask which is checked whether it contains the value.
 *                      Has to a variable of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 * @param[in]   value   Value which is used to check whether the bitmask contains it.
 */
#define PACE2_COMPARE_APPLICATION_ATTRIBUTE_TO_BITMASK(bmask,value) ((bmask).bitmask[(value) >> 6] & (((u64)1)<<((value) & 0x3F)))

/**
 * @brief Checks whether the bitmask contains the given value.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[out]  bmask   The bitmask which is checked whether it contains the value.
 *                      Has to a variable of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 * @param[in]   value   Value which is used to check whether the bitmask contains it.
 */
#define PACE2_COMPARE_PROTOCOL_ATTRIBUTE_TO_BITMASK(bmask,value) ((bmask).bitmask[(value) >> 6] & (((u64)1)<<((value) & 0x3F)))

/**
 * @brief Checks whether all bits of the bitmask are set to zero.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 */
#define PACE2_APPLICATION_ATTRIBUTES_BITMASK_IS_ZERO(a) ((a).bitmask[0] == 0)

/**
 * @brief Checks whether all bits of the bitmask are set to zero.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 */
#define PACE2_PROTOCOL_ATTRIBUTES_BITMASK_IS_ZERO(a) ((a).bitmask[0] == 0 &&  \
                                                      (a).bitmask[1] == 0)

/**
 * @brief Checks whether the bitmask a contains the complement of bitmask b.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_application_attributes_of_application and
 *             @ref pace2_get_number_of_application_attributes_of_protocol and
 *             @ref pace2_get_application_attribute_of_application_by_index and
 *             @ref pace2_get_application_attribute_of_protocol_by_index.
 *
 * @param[in]   a   The bitmask which is checked whether it contains the complement.
 *                  Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 * @param[in]   b   The bitmask which is used to check whether another bitmask contains its complement.
 *                  Bitmask of type @ref PACE2_application_attributes_bitmask "PACE2_application_attributes_bitmask".
 */
#define PACE2_APPLICATION_ATTRIBUTES_BITMASK_CONTAINS_NEGATED_BITMASK(a,b) ((((a).bitmask[0] & ~(b).bitmask[0]) == ~(b).bitmask[0]))

/**
 * @brief Checks whether the bitmask a contains the complement of bitmask b.
 *
 * @deprecated This macro is deprecated. Please use the functionality provided by
 *             @ref pace2_get_number_of_protocol_attributes_of_protocol and
 *             @ref pace2_get_protocol_attribute_of_protocol_by_index.
 *
 * @param[in]   a   The bitmask which is checked whether it contains the complement.
 *                  Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 * @param[in]   b   The bitmask which is used to check whether another bitmask contains its complement.
 *                  Bitmask of type @ref PACE2_protocol_attributes_bitmask "PACE2_protocol_attributes_bitmask".
 */
#define PACE2_PROTOCOL_ATTRIBUTES_BITMASK_CONTAINS_NEGATED_BITMASK(a,b) ((((a).bitmask[0] & ~(b).bitmask[0]) == ~(b).bitmask[0]) &&   \
                                                                         (((a).bitmask[1] & ~(b).bitmask[1]) == ~(b).bitmask[1]))

#ifdef __cplusplus
}
#endif
#endif /* PACE2_ATTRIBUTES_BITMASK_H */
