/** @file      pace2_basic_def.h
  * @brief     PACE 2 basic type definitions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_BASIC_DEF_H
#define PACE2_BASIC_DEF_H

/* Includes *********************************************************************/

#include <stdarg.h>

#if defined(__KERNEL__)
# include <linux/in.h>
#elif defined(__linux__)
# include <netinet/ip.h>
# include <netinet/ip6.h>
#elif defined(_WIN32)
# include <winsock2.h>
# include <in6addr.h>
#elif !defined(__OCTEON__)
# include <netinet/in.h>
#endif // __KERNEL__

#include <ipq_api.h>
#include "pace2_stail_queue.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Macros ****************************************************************/

#define PACE2_ALL_THREADS (-1)                      /*!< Thread ID value which references all threads. */
#define PACE2_CURRENT_THREAD (-2)                   /*!< Thread ID value of the currently running thread. */

/**
 * @brief Macro for getting the size of an array.
 */
#define PACE2_STATIC_ARRAY_SIZE(array) (sizeof(array) / sizeof(array[0]))

/**
 * @brief Macro to add a flag to the bitmask.
 *
 * @param   mask    Bitmask to which the flag should be added.
 * @param   flag    Flag which should be added.
 */
#define PACE2_MASK_ADD_FLAG(mask, flag)             ( (mask) |= 1ULL << (flag) )

/**
 * @brief Macro to removes a flag from a bitmask.
 *
 * @param   mask    Bitmask from which the flag should be removed.
 * @param   flag    Flag which should be removed.
 */
#define PACE2_MASK_REMOVE_FLAG(mask, flag)          ( (mask) &= ~(1ULL << (flag) ))

/**
 * @brief Macro to check a set flag inside a bitmask.
 *
 * The check returns 1 in cases the checked flag is set.
 * @param   mask    Bitmask which includes all flags.
 * @param   flag    Flag which should be checked.
 */
#define PACE2_MASK_INCLUDES_FLAG(mask, flag)        (( (mask) & (1ULL << (flag) )) != 0)

/**
 * @brief Macro to add a flag to an array bitmask.
 *
 * @param   mask    Bitmask array to which the flag should be added.
 * @param   flag    Flag which should be added.
 */
#define PACE2_ARRAY_MASK_ADD_FLAG(mask, flag)       PACE2_MASK_ADD_FLAG( (mask)[(flag >> 6)], ( (flag) & 0x3F))

/**
 * @brief Macro to removes a flag from an array bitmask.
 *
 * @param   mask    Bitmask array from which the flag should be removed.
 * @param   flag    Flag which should be removed.
 */
#define PACE2_ARRAY_MASK_REMOVE_FLAG(mask, flag)    PACE2_MASK_REMOVE_FLAG( (mask)[(flag >> 6)],( (flag) & 0x3F))

/**
 * @brief Macro to check a set flag inside an array bitmask.
 *
 * The check returns 1 in cases the checked flag is set.
 * @param   mask    Bitmask array which includes all flags.
 * @param   flag    Flag which should be checked.
 */
#define PACE2_ARRAY_MASK_INCLUDES_FLAG(mask, flag)  PACE2_MASK_INCLUDES_FLAG( (mask)[(flag >> 6)], ( (flag) & 0x3F))

/**
 * @brief Macro for declaring a specific part of the API as deprecated
 *
 * @param   msg     Message displayed by the compiler if a deprecated entity is used.
 */
#if defined(clang) || (__GNUC__ > 4 || (__GNUC__ == 4 && (__GNUC_MINOR__ > 4)))
#define PACE2_DEPRECATED(msg) __attribute__ ((deprecated(msg)))
#else
#define PACE2_DEPRECATED(msg) __attribute__ ((deprecated()))
#endif

/* Public Types *****************************************************************/

typedef struct pace2_module PACE2_module;           /*!< Public type definition of a PACE 2 module. */
typedef struct pace2_flow_struct PACE2_flow_data;   /*!< Public type definition of PACE 2 flow data. */
typedef void PACE2_subscriber_data;                 /*!< Public type definition of PACE 2 subscriber data. */
struct PACE2_global_config;                         /*!< Public definition of PACE 2 config data. */

typedef u64 PACE2_bitmask;                          /*!< Unified type definition for a PACE 2 bitmask. */
typedef u64 PACE2_timestamp;                        /*!< Unified type definition for a PACE 2 timestamp. */

/**
 * @brief Unified type definition for a PACE 2 byte buffer.
 *
 * This type has to be used for a byte buffer or character string with a given length.
 */
typedef struct {
    const char *ptr;                                /*!< Pointer to the first byte of the buffer. */
    u32 len;                                        /*!< Length of the buffer. */
} PACE2_byte_buffer;

/**
 * @brief Generic Structure to hold a dynamic list of unsigned 32bit values.
 */
struct PACE2_u32_list {
    u32 const * u32_list_element;                   /*!< Pointer to the first element of the list. */
    u32 number_of_u32_elements;                     /*!< Number of list elements. */
};

/**
 * @brief Generic Structure to hold a dynamic list of unsigned 16bit values.
 */
struct PACE2_u16_list {
    u16 const * u16_list_element;                   /*!< Pointer to the first element of the list. */
    u32 number_of_u16_elements;                     /*!< Number of list elements. */
};

/**
 * @brief Generic Structure to hold a dynamic list of unsigned 8bit values.
 */
struct PACE2_u8_list {
    u8 const * u8_list_element;                     /*!< Pointer to the first element of the list. */
    u32 number_of_u8_elements;                      /*!< Number of list elements. */
};

/**
 * @brief Unified type definition for a PACE 2 boolean value.
 */
typedef enum {
    IPQ_FALSE = 0,                                  /*!< Represents false. */
    IPQ_TRUE = 1                                    /*!< Represents true */
} PACE2_boolean;

/**
 * @brief Type definition for a unique subscriber key structure.
 */
typedef struct {
    u8 buffer[16];             /*!< Buffer to store the unique key. Size of a IPv6 address. */
} PACE2_subscriber_key;

/**
 * @brief Type definition for a IP address. Covers IPv4 and IPv6.
 */
typedef struct {
    u8 is_ip_v6;                    /**< Signals that ip is an IPv4 or an IPv6 address. */
    union {
        struct in_addr ipv4;        /**< The IPv4 address. Only use this, if is_ipv6 == "0". */
        struct in6_addr ipv6;       /**< The IPv6 address. Only use this, if is_ipv6 == "1". */
    } address;                      /**< The IP address. */
} PACE2_ip_address;

/**
 * @brief Definition of the PACE 2 packet descriptor, which joins all necessary information
 * required during packet processing.
 *
 * Please refer to the manual for a more detailed description of the possible usage of each field.
 */
typedef struct pace2_packet_descriptor {
    /* pointer variables */
    struct pace2_packet_stack *framing;             /*!< Pointer to the describing packet frame stack. */
    void *flow_data;                                /*!< Pointer to the related flow of the packet. */
    void *flow_user_data;                           /*!< Pointer to the related flow user data of the packet. */
    void *packet_user_data;                         /*!< Pointer to the user data of the packet. */
    void *src_stage3;                               /*!< Pointer to the classification source subscriber data. */
    void *src_stage3_user_data;                     /*!< Pointer to the classification source subscriber user data. */
    void *dst_stage3;                               /*!< Pointer to the classification destination subscriber data. */
    void *dst_stage3_user_data;                     /*!< Pointer to the classification destination subscriber user data. */
    void *flow_custom_key;                          /*!< Pointer to the flow key area used for user provided lookup keys */
    void *src_custom_key;                           /*!< Pointer to the source subscriber key area used for user provided lookup keys */
    void *dst_custom_key;                           /*!< Pointer to the destination subscriber key area used for user provided lookup keys */

    /* 64 bit variables */
    PACE2_timestamp packet_ts;                      /*!< Packet timestamp. The unit is defined by the config setting
                                                         @ref PACE2_config_general::io_timestamp_resolution. */
    u64 packet_id;                                  /*!< Externally assigned packet ID. */
    u64 flow_id;                                    /*!< ID of the related flow of the packet. Is calculated internally. */

    /* 32 bit variables */
    enum ipoque_asymmetric_connection_status_enum asymmetric_connection_status; /*!< Current status of asymmetry detection. */

    /* 16 bit variables */
    u16 packet_user_data_len;                       /*!< Length of the user data, which is available through @ref packet_user_data. */

    /* 8 bit variables */
    u8 direction;                                   /*!< Packet direction. */

    /* bit field variables */
    u32 copied_by_paro:1;                           /*!< Flag will be set if the packet got copied by paro. */
    u32 new_flow:1;                                 /*!< Flag will be set if the packet triggers a new flow. */
    u32 fast_path_used_in_s3:1;                     /*!< Flag will be set if the classification fastpath was used.
                                                         Will be set only when using internal tracking. */
    u32 s3_return_require_subscriber:1;             /*!< Flag to signal, that the run for pace2_s3_process_packet
                                                         has ended for this packet with the return code
                                                         PACE2_S3_REQUIRE_SUBSCRIBER. If this flag is set and the
                                                         function pace2_s3_process_packet will be called a second
                                                         time, then s3 will be done without any subscriber tracking.
                                                         Only used if external tracking is configured.*/
    u8 retransmission:1;                            /*!< Flag will be set if a retransmission is detected. This will
                                                         only be set if the slowpath is used or the TCP dissector is
                                                         active. */
    u8 out_of_order:1;                              /*!< Flag will be set if an out of order packet is detected. This will
                                                         only be set if the slowpath is used or the TCP dissector is
                                                         active.*/
    u8 esp_mode:2;                                  /*!< Indicates the Encapsulating Security Payload (ESP) mode
                                                         (encrypted/plain).*/
    void * internal_state;
    u8 reserved[16];                                /*!< Reserved bytes for future extensions. */
} PACE2_packet_descriptor;

/**
 * @brief Contains data generated in stage 3 which is required in stage 4
 */
typedef struct pace2_classifier_token {
    void *token_content;
    u16 token_length;
} PACE2_classifier_token;

/**
 * @brief Reassembly stream hole info structure (used for hole tracking)
 */
typedef struct pace2_reassembly_stream_hole_info {
    u32 offset;               /*!< Offset of the hole relative to the start of the previous hole
                                   or, only for the first hole, to the stream start.
                                   Also holes must not overlap, so the offset of the current hole
                                   must be greater than the len of the previous hole,
                                   to have at least 1 byte of valid data in between the holes. */
    u32 len;                  /*!< Length of the hole inside the stream, must not be 0. */
    u8 next_stream_direction; /*!< Direction of the next stream chunk (might be own or other side). */
} PACE2_reassembly_stream_hole_info;

/**
 * @brief Reassembly hole info structure (used for hole tracking)
 */
typedef struct pace2_reassembly_hole_info {
    /** Used for hole list maintenance, points to next list element. */
    PACE2_STAILQ_ENTRY(pace2_reassembly_hole_info) entry;
    PACE2_reassembly_stream_hole_info info; //!< actual hole info
} PACE2_reassembly_hole_info;

/**
 * @brief Reassembly stream list of holes head declaration
 */
typedef PACE2_STAILQ_HEAD(pace2_reassembly_holes_list, pace2_reassembly_hole_info) PACE2_reassembly_holes_list;

/**
 * @brief hole information list for each direction
 */
typedef struct pace2_reassembly_hole_data {
    u32 count[2];                        //!< number of elements in the list of holes in TCP stream per direction
    PACE2_reassembly_holes_list list[2]; //!< list of holes head per direction
} PACE2_reassembly_hole_data;


/**
 * @brief Definition of the PACE 2 stream descriptor, which joins all necessary information
 * required during stream processing.
 *
 * PACE 2 currently offers processing of streams only in stage 4. A data stream here is
 * the a reassembled representation of the payload of all packets belonging to the same flow, where each direction
 * is treated separately.
 *
 * Please refer to the manual for a more detailed description of the possible usage of each field.
 */
typedef struct pace2_stream_descriptor {
    void *flow_data;                                            /*!< Pointer to the related flow. */
    void *flow_user_data;                                       /*!< Pointer to the related flow user data of the stream. */

    /** Timestamp of most current data fragment. The unit is millisecond by default, can be changed up to nanoseconds
     * by using the config setting @ref PACE2_config_general::io_timestamp_resolution */
    PACE2_timestamp ts;

    const ipoque_unique_flow_ipv4_and_6_struct_t *flow_tuple;   /*!< Pointer to a structure identifying the flow. */

    u64 packet_id;                                              /*!< Externally assigned ID of the latest packet of the flow. */
    u64 flow_id;                                                /*!< ID of the related flow. Is set from packet descriptor. */

    u8 flow_is_asym:1;                                          /*!< Is set if flow is asymmetric. */
    u8 asym_dir:1;                                              /*!< Contains the side of the flow that contains data,
                                                                     if flow is asymmetric. */

    const void *stream[2];                                      /*!< Array containing one pointer per direction of the flow,
                                                                     pointing to the reassembled data of that direction. */
    u64 stream_length[2];                                       /*!< Contains the number of bytes of data available in the
                                                                     @ref stream pointers per direction. */

    u64 stream_remove[2];                                       /*!< Contains the number of bytes to remove from each direction
                                                                     of the input stream buffer. */

    PACE2_reassembly_hole_data * holes;                         /*!< Pointer to the hole information data (one hole
                                                                     information list for each direction). */
    u8 no_further_data[2];                                      /*!< is set to != 0 if no further data is expected per
                                                                     direction */

    u8 reserved[16];                                            /*!< Reserved bytes for future extensions. */
} PACE2_stream_descriptor;


/**
 * @brief Defines the constant values for the level of a process_information.
 */
typedef enum {
    PACE2_PROCESS_INFORMATION_LEVEL_DEBUG,      /*!< Signals that the call of the process information callback is a debug information. */
    PACE2_PROCESS_INFORMATION_LEVEL_WARNING,    /*!< Signals that the call of the process information callback is a warning information. */
    PACE2_PROCESS_INFORMATION_LEVEL_ERROR,      /*!< Signals that the call of the process information callback is a error information. */

    NUMBER_OF_PACE2_PROCESS_INFORMATION_LEVEL
} PACE2_process_information_level;

/**
 * @brief Defines the constant values for the level of a log message.
 */
typedef enum {
    /**
     * Fatal errors that should not happen at all. Null pointer, memory corruption, ...
     */
    PACE2_LOG_LEVEL_BUG,
    /**
     * A component cannot run or does not respond. Could be sourced by misconfiguration.
     */
    PACE2_LOG_LEVEL_ALERT,
    /**
     * An error occurred that could lead to instability of the whole system.
     */
    PACE2_LOG_LEVEL_CRITICAL,
    /**
     * An error occurred but the system was able to recover from it. Data could be lost or remain unprocessed.
     */
    PACE2_LOG_LEVEL_ERROR,
    /**
     * Unexpected behavior happened. It was not an error, meaning it has no impact on other parts, but should be
     * further investigated by the user.
     */
    PACE2_LOG_LEVEL_WARNING,
    /**
     * Default log level for normal operation on high level perspective. Starting instances and core mechanics.
     * NOT to be used for warnings or even errors.
     */
    PACE2_LOG_LEVEL_NOTICE,
    /**
     * More detailed messages, "showing measurable units of positive work passing through under normal functioning
     * of the system. Should not be continuous “I am alive!” messages.
     */
    PACE2_LOG_LEVEL_INFO,
    /**
     * Debug messages are a tool for developers that help to isolate a problem in a running system.
     * Should show the executing code and its context.
     */
    PACE2_LOG_LEVEL_DEBUG,

    NUMBER_OF_PACE2_LOG_LEVEL
} PACE2_log_level;


/**
 * @brief Typedef for a function, which can be used as user provided logging callback in the PACE2_config_general and
 *        PACE2_plugin_configuration.
 *
 * An exemplary implementation of this function interface is @ref pace2_default_logging_callback_impl and can be used
 * for basic logging purposes.
 *
 * @param[in]   thread_ID   The thread ID of the thread, from which this callback was called.
 * @param[in]   level       The kind of information for which this callback was called.
 *                          See also @ref PACE2_log_level.
 * @param[in]   user_ptr    Is a pointer which can be defined from the user over
 *                          @ref PACE2_config_general::logging_callback_user_ptr and
 *                          @ref PACE2_plugin_configuration::logging_user_ptr, and is available in every call of this
 *                          callback function. This is useful to get access to own data structures or keys, which are
 *                          needed to handle this callback.
 * @param[in]   format      Is a format string like in printf, which describes the logging information.
 * @param[in]   args        The set of variable arguments, which are passed to this function. Contains the values
 *                          for the format string like for printf.
 */
typedef void (*pace2_logging_callback)(u32 thread_ID, PACE2_log_level level,
                                       void * user_ptr, const char *format, va_list args);

/**
 * @brief Declares the data type of the provided PLUG-IN configuration pointer. Default: PACE2_PLUGIN_CONFIG_TYPE_STRUCT
 *
 * See @ref PACE2_plugin_configuration
 */
typedef enum {
    PACE2_PLUGIN_CONFIG_TYPE_STRUCT = 0, /*!< Type is a PLUG-IN specific configuration structure */
    PACE2_PLUGIN_CONFIG_TYPE_TEXT,       /*!< Type is a textual representation, like JSON, XML, INI, TOML, ... */
    PACE2_PLUGIN_CONFIG_TYPE_BLOB        /*!< Type is a binary object */
} PACE2_plugin_config_type;

/**
 * @brief Defines the generated ID.
 */
typedef u64 multipart_id_t;

/**
 * @brief Typedef for a function, with which the IDs can be generated.
 *
 * @param[in]  pace2                 The PACE2 module
 * @param[in]  pace2_instance_id     Current PACE 2 instance ID, used for ID generation.
 *                                   Default: 0
 * @param[in]  thread_id             PACE 2 thread ID, used for ID generation.
 *                                   Default: 0
 * @param[in]  object_id             Current thread local successive iterator.
 * @param[in]  usr_data              User data provided by the configuration.
 * @param[out] target                Resulting ID
 *
 * @retval -EINVAL                   If no target was handed over or any other ID exceeded its maximum value.
 * @retval 0                         If ID generation was successful.
 */
typedef int (*PACE2_generate_specific_id_fn)(PACE2_module * pace2,
                                             u32 const pace2_instance_id,
                                             u32 const thread_id,
                                             u64 const object_id,
                                             void * const usr_data,
                                             multipart_id_t * const target);

/**
 * @brief Contains configuration for user defined ID generation.
 */
typedef struct {
    void * id_generation_usr_data;                         /*!< User data which is provided to the specific ID
                                                                generation callbacks.
                                                                Default: NULL */

    PACE2_generate_specific_id_fn generate_flow_id;        /*!< Function pointer for flow ID generation. */
    PACE2_generate_specific_id_fn generate_event_id;       /*!< Function pointer for event ID generation. Currently used
                                                                for generating any IDs used within any advanced or class
                                                                metadata events. */
    PACE2_generate_specific_id_fn generate_subscriber_id;  /*!< Function pointer for subscriber ID generation. */
} PACE2_generate_specific_id_callbacks;

#define GENERATE_STRING(_enum_, _string_, ...)   #_string_,
#define GENERATE_ENUM(_enum_, ...) _enum_,

/**
 * @brief Return state of generic PACE2 functions.
 */
typedef enum {
    PACE2_SUCCESS,                         /**< Everything worked fine. */
    PACE2_PARAM_MISSING,                   /**< Some parameter is missing. */
    PACE2_TIMEOUT_REACHED,                 /**< Timeout has been reached. */
    PACE2_POINTER_NULL,                    /**< Usage of an invalid pointer. */
    PACE2_OUT_OF_MEMORY,                   /**< Memory allocation failed. */
    PACE2_TRACKING_FLOW_TIMEOUT_TOO_SMALL, /**< The tracking flow timeout is smaller than reordering packet timeout. */
    PACE2_INVALID_CONFIGURATION_POINTER,   /**< Usage of an invalid configuration pointer. */
    PACE2_INVALID_MODULE_POINTER,          /**< Usage of an invalid module pointer. */
    PACE2_S2_REORDERING_PACKET_BUFFER_SIZE_TOO_SMALL,  /**< The s2 reordering packet buffer size is smaller than the
                                                          minimum value. */
    PACE2_NO_DYNAMIC_UPGRADE,                          /**< Dynamic upgrade is not available in this build. */
    PACE2_INVALID_GENERAL_ALLOC_FUNCTION_PTR,          /**< The config general pace2_alloc function ptr is invalid. */
    PACE2_INVALID_GENERAL_FREE_FUNCTION_PTR,           /**< The config general pace2_free function ptr is invalid. */
    PACE2_INVALID_GENERAL_REALLOC_FUNCTION_PTR,        /**< The config general pace2_realloc function ptr is invalid. */
    PACE2_INVALID_DYNAMIC_UPGRADE_DLOPEN_FUNCTION_PTR, /**< The config dynamic upgrade dlopen function ptr is invalid.
                                                        */
    PACE2_INVALID_DYNAMIC_UPGRADE_DLSYM_FUNCTION_PTR,  /**< The config dynamic upgrade dlsym function ptr is invalid. */
    PACE2_INVALID_DYNAMIC_UPGRADE_DLCLOSE_FUNCTION_PTR, /**< The config dynamic upgrade dlclose function ptr is
                                                           invalid. */
    PACE2_DYNAMIC_UPGRADE_RELEASE_MEMORY_AGE_TOO_SMALL, /**< The dynamic upgrade memory age is smaller than flow or
                                                           subscriber timeout. */
    PACE2_NO_FLOW_MEMORY_AGE_SET,       /**< The config tracking flow type is EXTERNAL and no memory age is set. */
    PACE2_NO_SUBSCRIBER_MEMORY_AGE_SET, /**< The config tracking subscriber type is EXTERNAL and no memory age is set.
                                         */
    PACE2_EVENT_DECLINED,               /**< Allocation of event slot is rejected. */
    PACE2_EVENT_QUEUE_EMPTY,            /**< The event queue is empty. */
    PACE2_CDD_INVALID_FUNCTION_CALL,    /**< Invalid function call of CDD event allocation. */
    PACE2_CDD_DECODER_DISABLED,         /**< CDD are configured but the decoding stage is disabled. */
    PACE2_CDD_CONFIG_ARRAY_NOT_SET,     /**< The pointer to the CDD configuration is missing. */
    PACE2_CDD_CALLBACK_NOT_SET,         /**< The callback of a CDD is not set. */
    PACE2_CDD_EVENT_SIZE_NOT_SET,       /**< The event size of a CDD is not set. */
    PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS,   /**< The index for array access is outside the array bounds. */
    PACE2_INVALID_TAG,             /**< The tag is not valid because it is outside of the maximal amount of tags. */
    PACE2_FAILURE,                 /**< An error occurred. */
    PACE2_THREAD_INIT_INCOMPLETE,  /**< The thread instance of the given ID is not initialized. */
    PACE2_THREAD_INIT_DISABLED,    /**< The discrete thread initialization is not enabled. See
                                      PACE2_config_general::discrete_thread_init. */
    PACE2_THREAD_ID_INVALID,       /**< The given thread is invalid. */
    PACE2_DYNAMIC_UPGRADE_FAILURE, /**< An error with the dynamic upgrade feature occurred. */
    PACE2_FLOW_TRACKING_TABLE_INITIALIZATION_FAILED, /**< The initialization of the flow tracking table failed. */
    PACE2_STAGE1_FAILURE,                            /**< The initialization of stage1 failed. */
    PACE2_STAGE2_FAILURE,                            /**< The initialization of stage2 failed. */
    PACE2_STAGE4_FAILURE,                            /**< The initialization of stage4 failed. */
    PACE2_INSTANCE_ID_STORAGE_LEN_OUT_OF_RANGE,      /**< The instance ID storage length is out of range. */
    PACE2_THREAD_ID_STORAGE_LEN_OUT_OF_RANGE,        /**< The thread ID storage length is out of range . */
    PACE2_INSTANCE_ID_OUT_OF_RANGE,                  /**< The instance ID is out of range. */
    PACE2_THREAD_COUNT_OUT_OF_RANGE,                 /**< The thread count is out of range. */
    PACE2_FORBIDDEN,                                 /**< The requested operation is forbidden. */
    PACE2_INVALID,                                   /**< The requested operation is invalid. */
    PACE2_STAGE3_ERROR_CSI_DISABLED,                 /**< CSI module is disabled. */
    PACE2_STAGE3_ERROR_OS_DISABLED,                  /**< OS detection module is disabled. */
    PACE2_STAGE3_ERROR_ASYM_DET_DISABLED,            /**< Asymmetric detection is disabled. */
    PACE2_STAGE3_ERROR_PARAM_MISSING,                /**< Function parameter is missing. */
    PACE2_STAGE3_ERROR_MEM_FAILURE,                  /**< Memory allocation failed for a module initialization */
    PACE2_STAGE3_ERROR_CC_REGIST_FAILURE,            /**< Registration for Custom Classification failed. */
    PACE2_STAGE3_ERROR_DISSECTOR_MISSING_DEPENDENCY, /**< a dissector depends on another dissector which is not
                                                        enabled. */
    PACE2_STAGE3_ERROR_DISSECTOR_DISABLED,           /**< Metadata dissector disabled. */
    PACE2_STAGE3_ERROR_MIDSTREAM_DISABLED,           /**< Midstream detection is disabled. */
    PACE2_STAGE3_ERROR_AGGRESSIVE_DISABLED,          /**< Aggressive detection is disabled. */
    PACE2_STAGE3_ERROR_METADATA_DISSECTOR_FAILURE,   /**< Initiation of metadata dissectors failed. */
    PACE2_STAGE3_ERROR_SET_NAT_DETECTION_MODE_ERROR, /**< Setting of the nat_detection_mode failed. */
    PACE2_STAGE3_ERROR_SET_BITTORRENT_CONFIG,        /**< Setting of the bittorrent configuration failed. */
    PACE2_STAGE3_ERROR_SET_EDONKEY_CONFIG,           /**< Setting of the edonkey configuration failed. */
    PACE2_STAGE3_ERROR_DYNAMIC_UPGRADE_FAILURE,      /**< Failure initializing the dynamic upgrade. */
    PACE2_PLUGIN_DLWRAPPER_NOT_SET, /**< Wrapper function for either dlopen or dlsym or dlclose is not set. */
    PACE2_PLUGIN_DLOPEN_ERROR,      /**< An error occurred during dlopen. */
    PACE2_PLUGIN_DLSYM_ERROR,       /**< An error occurred during dlsym. */
    PACE2_PLUGIN_BAD_FILENAME,      /**< PLUG-IN filename is too long or malformed. */
    PACE2_PLUGIN_REGISTER_FUNCTION_NOT_FOUND, /**< The register function of a PLUG-IN was not found. */
    PACE2_PLUGIN_REGISTER_FAILED,             /**< The registration function of a PLUG-IN failed. */
    PACE2_PLUGIN_INIT_THREAD_FAILED,          /**< The thread initialization function of a PLUG-IN failed. */
    PACE2_PLUGIN_COUNT_EXCEEDED,              /**< The maximum PLUG-IN count has been exceeded. */
    PACE2_PLUGIN_CALLBACK_ALREADY_REGISTERED, /**< The given callback is already registered for the access point. */
    PACE2_PLUGIN_OBSERVER_NOT_FOUND,          /**< No matching observer was found. */
    PACE2_PLUGIN_CANNOT_USE_AFTER_INIT,       /**< Function cannot be used after the global initialization is done. */
    PACE2_PLUGIN_CANNOT_USE_IN_GLOBAL_INIT,   /**< Function cannot be used during the global initialization. */
    PACE2_PLUGIN_INCOMPATIBLE_ABI_VERSION,    /**< The loaded PLUG-IN has an incompatible ABI version. */
    PACE2_PLUGIN_FORBIDDEN_PLUGIN_ID,         /**< The used PLUG-IN ID is forbidden. */
    PACE2_PLUGIN_ALREADY_USED_PLUGIN_ID,      /**< The used PLUG-IN ID is already used by another PLUG-IN. */
    PACE2_CONFIG_DICT_TYPE_MISMATCH,          /**< Type does not match expected type. */
    PACE2_CONFIG_DICT_UNKOWN_ENTRY,           /**< The given config entry name is unknown. */
    PACE2_FEATURE_NOT_AVAILABLE,              /**< The requested feature is not available in this release. */
    PACE2_PLUGIN_END_OBSERVING,               /**< The PLUG-IN does not want to process observer data for this flow. */
    PACE2_PLUGIN_EVENT_SIZE_NOT_SET,          /**< The PLUG-IN did not register the event size of an event it used. */
    PACE2_FEATURE_DISABLED,                   /**< The requested feature is not enabled. */
    PACE2_INVALID_CONFIGURATION,              /**< The configuration passed to @ref pace2_init does not meet certain
                                                   invariance criteria. This can have any of the following reasons:
                                                       - The ML module is enabled and the CSI feature or
                                                         subscriber tracking is disabled. */

    PACE2_ML_FLOW_DATA_REGISTER_ERROR,        /**< The flow data could not be properly registered. */
    PACE2_ML_FLOW_DATA_NOT_FOUND,             /**< The flow data for the current flow was not found. */
    PACE2_ML_SUBSCRIBER_DATA_REGISTER_ERROR,  /**< The subscriber data could not be properly registered. */
    PACE2_ML_SUBSCRIBER_DATA_NOT_FOUND,       /**< The subscriber data for the current flow was not found. */
    PACE2_ML_CO_FLOW_ERROR,                   /**< The co-flow feature did not behave as expected. */
    PACE2_ML_MODEL_ERROR,                     /**< A machine learning model behaved unexpectedly. */

    PACE2_STAGE3_ERROR_BROWSER_DET_DISABLED,  /**<! Browser classification module is disabled. */

    PACE2_NUMBER_OF_RETURN_CODES
} PACE2_return_state;

/**
 * @brief The structure of the PLUG-IN IDs.
 */
typedef union pace2_plugin_id_struct {
    u32 full_id;                    /**< The combination of prefix and ID as whole u32. Used for monitoring flag */
    struct {
#if defined(__BYTE_ORDER__) && __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
        u32 id:24;                  /**< The actual ID. */
        u32 prefix:8;               /**< A 8 bit prefix. */
#elif defined(__BYTE_ORDER__) && __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
        u32 prefix:8;               /**< A 8 bit prefix. */
        u32 id:24;                  /**< The actual ID. */
#else
# error "__BYTE_ORDER__ must be defined"
#endif /* __BYTE_ORDER__ */
    };
} PACE2_plugin_id_struct;

/**
 * @brief Return state of load license function.
 */
typedef enum {
    PACE2_LICENSING_LOAD_SUCCESS = 0,       /*!< License has been successfully loaded. */
    PACE2_LICENSING_UNSUPPORTED_FORMAT,     /*!< Format of the input file is not valid. */
    PACE2_LICENSING_LOAD_FAILED,            /*!< Loading of input file failed. */
    PACE2_LICENSING_NOT_AVAILABLE,          /*!< Licensing feature is not enabled. */
    PACE2_LICENSING_LOAD_DEMO_SUCCESS       /*!< Standard demo limitations loaded. */
} PACE2_licensing_state;

/**
 * @brief Enumeration defining the indices into the protocol stack for specific layers.
 */
typedef enum {
    PACE2_RESULT_PROTOCOL_LAYER_3 = 0,
    PACE2_RESULT_PROTOCOL_LAYER_4 = 1,
    PACE2_RESULT_PROTOCOL_LAYER_7 = 2,
    PACE2_RESULT_TRANSFERRED_CONTENT = 3
} PACE2_result_protocol_layer_indexes;

/**
 * @brief Macro for generating enumeration type for PACE 2 scope IDs and dedicated descriptions. Every new
 *        entry has to be added at the end and have to assigned a consecutive value.
 *        If a value is added in the middle all assignments have to be changed accordingly.
 */
#define PACE2_SCOPE_DESCRIPTIONS(X)                                                                         \
    X(PACE2_SCOPE_UNDEFINED,            scope_undefined,          Unspecified scope)                        \
    X(PACE2_SCOPE_MODULE,               scope_module,             PACE module scope)                        \
    X(PACE2_SCOPE_STAGE_2,              scope_stage_2,            Stage 2 scope)                            \
    X(PACE2_SCOPE_STAGE_2_FLOW_DATA,    scope_stage_2_flow,       Stage 2 flow data scope)                  \
    X(PACE2_SCOPE_STAGE_3,              scope_stage_3,            Stage 3 scope)                            \
    X(PACE2_SCOPE_STAGE_4,              scope_stage_4,            Stage 4 scope)                            \
    X(PACE2_SCOPE_STAGE_4_FLOW_DATA,    scope_stage_4_flow,       Stage 4 flow_data scope)                  \
    X(PACE2_SCOPE_STAGE_4_DECODER,      scope_stage_4_decoder,    Stage 4 decoder)                          \
    X(PACE2_SCOPE_STAGE_4_CLASS,        scope_stage_4_class,      Stage 4 class decoder)                    \
    X(PACE2_SCOPE_EVENT_CONTROLLER,     scope_event_controller,   Events related scope)                     \
    X(PACE2_SCOPE_TCPR,                 scope_tcpr,               TCPR scope)                               \
    X(PACE2_SCOPE_SCTPR,                scope_sctpr,              SCTPR scope)                              \
    X(PACE2_SCOPE_UTILS,                scope_utils,              Scope utils)                              \
    X(PACE2_SCOPE_IPC,                  scope_ipc,                Scope IPC)                                \
    X(PACE2_SCOPE_PLUGIN_FRAMEWORK,     scope_plugin_framework,   Scope of Pace2 PLUG-IN Framework)         \
    X(PACE2_SCOPE_PLUGINS,              scope_plugins,            Scope of PLUG-IN memory)                  \
    X(PACE2_SCOPE_EXTERNAL,             scope_external,           External resources scope)                 \
    X(PACE2_SCOPE_ML,                   scope_ml,                 Machine learning scope)

/**
 * @brief Enumeration of Scope IDs used in memory functions (alloc, realloc, free).
 */
typedef enum {
    PACE2_SCOPE_DESCRIPTIONS(GENERATE_ENUM)
    PACE2_NUMBER_OF_SCOPE_IDENTIFIERS
} PACE2_scope_id;

#ifdef __cplusplus
}
#endif
#endif /* PACE2_BASIC_DEF_H */
