/** @file      pace2_classification_results.h
  * @brief     PACE 2 classification results definitions
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_CLASSIFICATION_RESULTS_H
#define PACE2_CLASSIFICATION_RESULTS_H

/* Includes *********************************************************************/
#include "pace2_basic_def.h"
#include "pace2_applications.h"
#include "pace2_attributes.h"
#include "pace2_protocols.h"
#include "pace2_protocols_bitmask.h"
#include "event_types/os_events.h"
#include "event_types/browser_classification_events.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Macros ****************************************************************/

#define PACE2_PROTOCOL_STACK_MAX_DEPTH 4                /**< Depth of the classification protocol stack inclusive
                                                             the field of the transferred content.*/
#define PACE2_APPLICATION_ATTRIBUTE_LIST_MAX_SIZE 6     /**< Maximum number of set application attributes. */
#define PACE2_PROTOCOL_ATTRIBUTE_LIST_MAX_SIZE 6        /**< Maximum number of set protocol attributes. */

/* Public Types *****************************************************************/

/**
 * @brief Definition of the classification protocol stack.
 */
typedef struct {
    PACE2_protocols entry[PACE2_PROTOCOL_STACK_MAX_DEPTH];  /**< Protocol stack entries. Use  the defined enumerations
                                                                 of @ref PACE2_result_protocol_layer_indexes to access
                                                                 the specific layers. */
    u8 length;                                              /**< Number of set protocol stack entries. */
} PACE2_protocol_stack;

/**
 * @brief Definition of the protocol attribute list.
 */
typedef struct {
    PACE2_protocol_attributes list[PACE2_PROTOCOL_ATTRIBUTE_LIST_MAX_SIZE]; /**< List of set protocol attributes. */
    u8 length;                                              /**< Number of set protocol attributes. */
} PACE2_protocol_attribute_list;

/**
 * @brief Definition of the application attribute list.
 */
typedef struct {
    PACE2_application_attributes list[PACE2_APPLICATION_ATTRIBUTE_LIST_MAX_SIZE];   /**< List of set application attributes. */
    u8 length;                                              /**< Number of set application attributes. */
} PACE2_application_attribute_list;

/* Public Prototypes ************************************************************/

/**
 * @brief This function checks, if a given application attribute is set.
 *
 * @param[in]   attr_list   Pointer to an application attribute list which should get tested.
 * @param[in]   attr        The application attribute to check if it is set.
 * @returns 1 if attribute is set, 0 otherwise.
 */
int pace2_classification_is_application_attribute_set(PACE2_application_attribute_list const * const attr_list,
                                                      PACE2_application_attributes const attr);

/**
 * @brief This function checks, if a given protocol attribute is set.
 *
 * @param[in]   attr_list   Pointer to a protocol attribute list which should get tested.
 * @param[in]   attr        The protocol attribute to check if it is set.
 * @returns 1 if attribute is set, 0 otherwise.
 */
int pace2_classification_is_protocol_attribute_set(PACE2_protocol_attribute_list const * const attr_list,
                                                   PACE2_protocol_attributes const attr);

/**
 * @brief Returns a string representing the given OS type.
 *
 * The function supports the dynamic upgrade feature. If the library is upgraded
 * to a newer version the function implicitly calls it's upgrade variant.
 *
 * @param[in]   pace2_module         Constant pointer to the PACE 2 module
 *                                   Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID            The ID of the thread currently processing.
 *                                   Only used in case of dynamic upgrade otherwise it can be 0.
 * @param[in]   os_type              ID of the OS to retrieve the string for.
 * @retval      !=NULL               Zero-terminated string representing the given protocol.
 * @retval      NULL                 Given protocol ID is unknown.
 */
char const * pace2_get_os_string(PACE2_module const * const  pace2_module, int const thread_ID, PACE2_OS_type const os_type);

/**
 * @brief Returns a string representing the given OS type.
 *
 * The function is equivalent to @ref pace2_get_os_string, but supports the dynamic upgrade feature.
 * If the library is upgraded to a newer version the function implicitly calls it's upgrade variant.
 *
 * @deprecated Deprecated due to duplicate function @ref pace2_get_os_string.
 *
 * @param[in]   pace2_module         Constant pointer to the PACE 2 module
 *                                   Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID            The ID of the thread currently processing.
 *                                   Only used in case of dynamic upgrade otherwise it can be 0.
 * @param[in]   os_type              ID of the OS to retrieve the string for.
 * @retval      !=NULL               Zero-terminated string representing the given protocol.
 * @retval      NULL                 Given protocol ID is unknown.
 */
PACE2_DEPRECATED("Please use 'pace2_get_os_string' instead.")
char const * pace2_get_os_str(PACE2_module const * const pace2_module, int const thread_ID, PACE2_OS_type const os_type);

/**
 * @brief Returns a string representing the given browser type.
 *
 * @param[in]   pace2_module         Constant pointer to the PACE 2 module
 *                                   Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID            The ID of the thread currently processing.
 *                                   Only used in case of dynamic upgrade otherwise it can be 0.
 * @param[in]   browser_type         ID of the browser to retrieve the string for.
 * @retval      !=NULL               Zero-terminated string representing the given browser.
 * @retval      NULL                 Given browser ID is unknown.
 */
char const * pace2_get_browser_string(PACE2_module const * const pace2_module, int const thread_ID,
                                      PACE2_browser_type const browser_type);

/**
 * @brief This function returns the number of supported applications.
 *
 * The function supports the dynamic upgrade feature. If the library is upgraded
 * to a newer version the function implicitly calls it's upgrade variant.
 *
 * @param[in]   pace2_module         Constant pointer to the PACE 2 module
 *                                   Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID            The ID of the thread currently processing.
 *                                   Only used in case of dynamic upgrade otherwise it can be 0.
 * @retval      u16                  The number of supported applications.
 */
u16 pace2_get_number_of_applications(PACE2_module * const pace2_module, int const thread_ID);

/**
 * @brief This function returns the number of supported protocols.
 *
 * The function supports the dynamic upgrade feature. If the library is upgraded
 * to a newer version the function implicitly calls it's upgrade variant.
 *
 * @param[in]   pace2_module         Constant pointer to the PACE 2 module
 *                                   Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID            The ID of the thread currently processing.
 *                                   Only used in case of dynamic upgrade otherwise it can be 0.
 * @retval      u16                  The number of protocols.
 */
u16 pace2_get_number_of_protocols(PACE2_module * const pace2_module, int const thread_ID);

/**
 * @brief This function returns the number of supported application attributes.
 *
 * The function supports the dynamic upgrade feature. If the library is upgraded
 * to a newer version the function implicitly calls it's upgrade variant.
 *
 * @param[in]   pace2_module         Constant pointer to the PACE 2 module
 *                                   Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID            The ID of the thread currently processing.
 *                                   Only used in case of dynamic upgrade otherwise it can be 0.
 * @retval      u16                  The number of supported application attributes.
 */
u16 pace2_get_number_of_application_attributes(PACE2_module * const pace2_module, int const thread_ID);

/**
 * @brief This function returns the number of supported protocol attributes.
 *
 * The function supports the dynamic upgrade feature. If the library is upgraded
 * to a newer version the function implicitly calls it's upgrade variant.
 *
 * @param[in]   pace2_module         Constant pointer to the PACE 2 module
 *                                   Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID            The ID of the thread currently processing.
 *                                   Only used in case of dynamic upgrade otherwise it can be 0.
 * @retval      u16                  The number of supported protocol attributes.
 */
u16 pace2_get_number_of_protocol_attributes(PACE2_module * const pace2_module, int const thread_ID);

/**
 * @brief This function returns the number of supported OS types.
 *
 * The function supports the dynamic upgrade feature. If the library is upgraded
 * to a newer version the function implicitly calls it's upgrade variant.
 *
 * @param[in]   pace2_module         Constant pointer to the PACE 2 module
 *                                   Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID            The ID of the thread currently processing.
 *                                   Only used in case of dynamic upgrade otherwise it can be 0.
 * @retval      u16                  The number of supported OS types.
 */
u16 pace2_get_number_of_os_types(PACE2_module * const pace2_module, int const thread_ID);

/**
 * @brief This function returns the number of supported browser types.
 *
 * The function supports the dynamic upgrade feature.
 *
 * @param[in]   pace2_module         Constant pointer to the PACE 2 module
 *                                   Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID            The ID of the thread currently processing.
 *                                   Only used in case of dynamic upgrade otherwise it can be 0.
 * @retval      u16                  The number of supported browser types.
 */
u16 pace2_get_number_of_browser_types(PACE2_module * const pace2_module, int const thread_ID);

/**
 * @brief This function returns the number of supported internal group tag types.
 *
 * The function supports the dynamic upgrade feature. If the library is upgraded
 * to a newer version the function implicitly calls it's upgrade variant.
 *
 * @param[in]   pace2_module         Constant pointer to the PACE 2 module
 *                                   Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID            The ID of the thread currently processing.
 *                                   Only used in case of dynamic upgrade otherwise it can be 0.
 * @retval      u16                  The number of supported group tag types.
 */
u16 pace2_get_number_of_internal_group_tag_types(PACE2_module * const pace2_module, int const thread_ID);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_CLASSIFICATION_RESULTS_H */
