/** @file      pace2_config.h
  * @brief     PACE 2 configuration definitions
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_CONFIG_H
#define PACE2_CONFIG_H

/* Includes *********************************************************************/
#include "pace2_basic_def.h"
#include "pace2_protocols.h"
#include "pace2_protocols_bitmask.h"
#include "pace2_s3_cdc.h"
#include "pace2_s4_cdd.h"
#include "pace2_plugins.h"
#include "pace2_event_policy.h"
#include "pace2_frame_buffer.h"
#include "pace2_ipc.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Types *****************************************************************/

/**
 * @brief Amount of timeout queues to use for internal tracking. +1 for the default queue.
 */
#define PACE2_INT_FLOW_DEFAULT_TIMEOUT_QUEUES ( 32 + 1 )

/**
 * @brief Definition to distinguish tracking type.
 */
typedef enum {
    INTERNAL,                                       /*!< Usage of internal tracking. */
    EXTERNAL                                        /*!< Usage of external tracking. */
} PACE2_tracking_type;

/**
 * @brief Definition to distinguish safe and aggressive detection modes.
 */
typedef enum {
    SAFE_CLASSIFICATION = 0,                        /*!< Usage of safe mode. */
    AGGRESSIVE_CLASSIFICATION                       /*!< Usage of aggressive mode. */
} PACE2_classification_mode;

/**
 * @brief Definition to distinguish the tracked subscriber direction.
 *
 * The initiating direction will be direction 0, whereas the other direction
 * has the index 1.
 *
 */
typedef enum {
    ALL,                                            /*!< Track all directions. */
    SRC_DIR_0,                                      /*!< Track directions 0 only. */
    SRC_DIR_1,                                      /*!< Track directions 1 only. */
    NONE,                                           /*!< Track no direction. */
} PACE2_tracking_subscriber_side;

/**
 * @brief Definition to distinguish the memory basis of the table.
 *
 * Use @ref ELEMENT_NUMBER to initialize the table with
 * space for the specified number of elements or use @ref MEMORY_SIZE
 * to initialize the table with the given memory size.
 */
typedef enum {
    ELEMENT_NUMBER = 0,                             /*!< Initialize table with given element number */
    MEMORY_SIZE                                     /*!< Initialize table with given memory size */
} PACE2_tracking_mem_basis;

/**
 * @brief Definition to distinguish midstream detection modes.
 */
typedef enum {
    DEFAULT_MIDSTREAM_MODE = 0,                     /*!< Default mode: Midstream traffic checked. */
    EXTENDED_MIDSTREAM_MODE,                        /*!< Extended mode: enable additional midstream patterns. */
    SKIP_MIDSTREAM_MODE                             /*!< Skip mode: bypass all midstream traffic. */
} PACE2_midstream_mode;

/**
 * @brief Definition to classification event creation modes.
 */
typedef enum {
    PACE2_DEFAULT_EVERY_PACKET_MODE = 0,            /*!< Default mode: classification event is generated for every packet. */
    PACE2_ON_CLASSIFICATION_CHANGE_MODE,            /*!< Classification change: whenever the classification result changes. */
    PACE2_ON_CLASSIFICATION_FINISHED_MODE,          /*!< Only create the event when no further classification changes are expected. */
    PACE2_ON_FLOW_DROPPED_MODE                      /*!< Create the event when the flow is removed from internal flow tables. */
} PACE2_classification_event_mode;

/**
 * @brief Definition of the maximum size of a hash table.
 */
typedef struct {
    PACE2_tracking_mem_basis mem_basis;             /*!< Basis type on which the memory for table will be
                                                         reserved, default: @ref MEMORY_SIZE. */
    u64 element_number;                             /*!< Number of elements for which table will be
                                                         initialized. Must be set if @ref mem_basis
                                                         is @ref ELEMENT_NUMBER. Value will be set automatically
                                                         if @ref mem_basis is @ref MEMORY_SIZE. */
    u64 memory_size;                                /*!< Memory size for initialization of the table,
                                                         default: 4 MiByte. Value will be set automatically if @ref mem_basis is
                                                         @ref ELEMENT_NUMBER. */
} PACE2_table_max_size;

/**
 * @brief Definition of generic tracking configuration parameters.
 */
typedef struct {
    PACE2_tracking_type type;                       /*!< Tracking type mechanism, default: @ref INTERNAL. */
    PACE2_table_max_size max_size;                  /*!< The maximum size of the hash table which is given by an element number or maximum memory size */

    /** Timeout of a hash table element in milliseconds. Default: 600 000 ms. */
    PACE2_timestamp timeout;

    u32 user_data_size;                             /*!< User-data size of a hash table element, default: 0 Byte. */
    u16 custom_key_size;                            /*!< only used with internal tracking. If not 0 the flow key generation is expected to
                                                         be done by the user and passed to the engine. Default 0 (off) */
    u8 enable_tracking_statistics;                  /*!< Activates the statistical tracking of transferred bytes available through the events PACE2_FLOW_PROCESS_EVENT
                                                         and PACE2_SUBSCRIBER_PROCESS_EVENT. This feature requires additional memory per flow/subscriber and
                                                         lowers the performance. default: 0 (disabled). The feature is only available for internal tracking. */
    u8 tracking_statistics_per_packet;              /*!< If @ref enable_tracking_statistics is enabled and this is set to "1", then the PACE2_FLOW_PROCESS_EVENT
                                                         or PACE2_SUBSCRIBER_PROCESS_EVENT will be thrown for every packet. Otherwise only at the end of a
                                                         flow/subscriber. default: 1 (enabled). The feature is only available for internal tracking. */
    u8 enable_tunnel_tracking;                      /*!< automatically use both IPs in tunnel traffic for internal tracking, default disabled */
    u8 pht_usage_notification_enabled;              /*!< If set to "1", @ref PACE2_PHT_USAGE_NOTIFICATION_EVENT events
                                                         are thrown if element or memory thresholds are reached.
                                                         This feature is only available for internal tracking. */
    PACE2_table_max_size pht_usage_notification_threshold; /*!< Amount of used elements or memory in the hash table
                                                                which will trigger a
                                                                @ref PACE2_PHT_USAGE_NOTIFICATION_EVENT if reached.
                                                                The default threshold is the maximum hash table memory
                                                                size / 2. */
} PACE2_tracking_generic_params;

/**
 * @brief Structure definition of tracking configuration.
 */
struct PACE2_config_tracking {
    struct flow_params {
        PACE2_tracking_generic_params generic;      /*!< Generic flow tracking parameters. */

        struct flow_tracking_specific_params {
            PACE2_timestamp timeout_queue_timeouts[PACE2_INT_FLOW_DEFAULT_TIMEOUT_QUEUES];
                                                    /*!< Timeout values for the timeout queues. By default three additional queues are
                                                         configured with the timeouts 5 seconds, 15 minutes and 60 minutes. */

            /** Caches the subscriber on the flow. This feature is only available for internal tracking.
                When active extra 40 bytes for per flow and 2 byte per subscriber are used. Default disabled. */
            PACE2_boolean cache_subscriber;
        } specific;                                 /*!< Specific flow tracking parameters. */
    } flow;                                         /*!< Flow tracking configuration. */

    struct subscriber_params {
        PACE2_tracking_generic_params generic;      /*!< Generic subscriber tracking parameters. */
        struct sub_tracking_specific_params {
            PACE2_tracking_subscriber_side side;    /*!< Tracked subscriber directions, default: ALL. */
        } specific;                                 /*!< Specific subscriber tracking parameters. */
    } subscriber;                                   /*!< Subscriber tracking configuration. */
};

/**
 * @brief Definition of Session ID Tracking configuration parameters.
 */
typedef struct {
    u32 memory;                                     /*!< The size of the internal memory buffer for session IDs in bytes,
                                                         can be 0 (if 0, the number of elements is used to determine
                                                         the memory size), default: 1048576.*/
    u32 number_of_elements;                         /*!< The number of session IDs that should be stored. Can be 0. Will only be used if
                                                         the memory argument is zero, default: 0 */
    u16 key_reduce_factor;                          /*!< The factor will be used to reduce the
	                                                     session ID and related key information to reduce memory usage
	                                                     while accepting a larger false positive ratio. Can 1, 2, or 4
	                                                     (1 means no reduction, use complete session ID), default: 1. */
    u8 enabled;                                     /*!< Flag to indicate the enabled Session ID Tracking, default: @ref IPQ_FALSE. */
} PACE2_sit_params;

/**
 * @brief Definition of IP defragmentation configuration parameters.
 */
typedef struct {
    u8 enabled;                                     /*!< Flag to indicate the enabled IP defragmentation, default: @ref IPQ_FALSE. */
    u64 memory_size;                                /*!< Maximum memory size used by the IP defragmentation, default: 1 MiByte.
                                                         Regard that the memory size is used by each thread */
    u8 enabled_memory_growth;                       /*!< Flag to indicate an enabled growing of the consumed memory, default: @ref IPQ_FALSE.
                                                         Requires a set function pointer of @ref PACE2_config_general::pace2_realloc. */

    /** Timeout of each fragment in milliseconds. Default: 1000 ms. It is recommended to keep this as low as possible
     *  (maximum 1000 milliseconds). A value of 0 ms will mark every packet as too old. */
    PACE2_timestamp timeout;

    enum ipoqe_defrag_checksum_enum checksum;       /*!< Checksum handling of the IP defragmentation engine, default
                                                         IPOQUE_DEFRAG_CHECKSUM_IGNORE_WRONG_IP_CHECKSUM. */
} PACE2_s1_defrag_params;

/**
 * @brief Structure definition of packet preparing configuration parameters.
 */
struct PACE2_config_s1_preparing {
    u32 max_framing_depth;                          /*!< Maximum framing depth, default: 10. */
    u32 max_decaps_level;                           /*!< Maximum decapsulation level, default: 0. */
    PACE2_s1_defrag_params defrag;                  /*!< IP defragmentation specific configuration parameters. */
    PACE2_boolean esp_decapsulation_enabled;        /*!< Flag to enable L4 decapsulation of unencrytped ESP headers,
                                                         default: @ref IPQ_TRUE. */
};

/**
 * @brief Structure definition of packet reordering configuration parameters.
 */
struct PACE2_config_s2_reordering {
    u8 enabled;                                     /*!< Flag to indicate the enabled packet reordering, default: @ref IPQ_FALSE. */
    u64 packet_buffer_size;                         /*!< Maximum packet buffer size used by the packet reordering, default: 16 MiByte. */
    enum ipoque_paro_core_mode mode;                /*!< Memory management operation mode, default: @ref IPOQUE_PARO_PACKET_DYNAMIC.*/
    PACE2_timestamp packet_timeout;                 /*!< Timeout of each packet in milliseconds. Default: 5000 ms. */
};

/**
 * @brief Structure definition for one custom defined classification
 */
struct PACE2_config_s3_cdc {
    pace2_cdc_callback_t callback;                   /*!< Callback for the custom defined classification */
    void *callback_userptr;                          /*!< User pointer provided to the custom classification */
    u32 flow_area_size;                              /*!< bytes provided to the custom classification on the flow */
    u32 subscriber_area_size;                        /*!< bytes provided to the custom classification on the subscriber */
};

/**
 * @brief Structure definition of SECA configuration parameters.
 */
struct PACE2_config_seca {
    u8 enabled;
    PACE2_table_max_size memory;                                /*!< Main hash table memory. */
    PACE2_timestamp timeout;                                    /*!< Global hash table timeout. Default: 1h */
};

/**
 * @brief Structure definition of DECA ip allowlist configuration parameters.
 */
struct PACE2_config_deca_allowlist {
    PACE2_ip_address ip;                                        /*!< Allowed DNS server. */
    PACE2_ip_address mask;                                      /*!< Subnet mask to allow DNS server. */
};

enum PACE2_deca_mode {
    DECA_EXCLUSIVE = 0,     /*!< DECA result will be provided as application detection in the classification
                                 result event and stops further matching */
    DECA_AND_PATTERN_MATCH  /*!< DECA result will be provided as application detection in the classification.
                                 The regular pattern matching will continue and might provide a different result */
};

/**
 * @brief Structure definition of DECA configuration parameters.
 */
struct PACE2_config_deca {
    u8 enabled;                                                 /*!< Flag to indicate the enabled deca engine, default: @ref IPQ_FALSE. */
    const struct PACE2_config_deca_allowlist * allowlist;       /*!< Pointer to an array of allowlisted IP entries. */
    u8 allowlist_count;                                         /*!< Amount of allowlisted IPs. */
    PACE2_table_max_size memory;                                /*!< Main hash table memory. */
    PACE2_timestamp timeout;                                    /*!< Global hash table timeout.
                                                                     The DNS timeout value will
                                                                     be used if not set. */
};

struct PACE2_config_s3_pa_tags {
    u8 enabled;                                     /*!< Flag to indicate the enabled protocol and application grouping, default: @ref IPQ_FALSE. */
    const char **tag_strings;                       /*!< Array of strings connected to protocol and application tag implementation, default NULL. */
    u16 number_of_tags;                             /*!< Size of the @ref tag_strings array. */
};

/**
 * @brief Structure definition of packet classification configuration parameters.
 */
struct PACE2_config_s3_classification {
    u8 enabled;                                     /*!< Flag to indicate the enabled packet classification, default: @ref IPQ_TRUE. */

    struct {
        PACE2_protocols_mask protocols;
        PACE2_boolean enable_all_applications;      /*!< Enabled by default. If disabled no application will be reported. */
    } enabled_classifications;                      /*!< Struct defining classification of which protocols shall be activated, default: all. */
    u8 csi_enabled;                                 /*!< Flag to indicate the enabled client-server indication, default: @ref IPQ_FALSE. */
    u8 nat_enabled;                                 /*!< Flag to indicate the enabled NAT detection, default: @ref IPQ_FALSE. */
    u8 os_enabled;                                  /*!< Flag to indicate the enabled operating system detection, default: @ref IPQ_FALSE. */
    u8 asym_detection_enabled;                      /*!< Flag to indicate the enabled asymmetric detection mode, default: @ref IPQ_FALSE. */
    /** OpenVPN detection mode, default: @ref AGGRESSIVE_CLASSIFICATION.
     *  @arg Use @ref AGGRESSIVE_CLASSIFICATION mode to detect all known OpenVPN handshakes.
     *       This might cause a higher false positive rate.
     *  @arg Use  @ref SAFE_CLASSIFICATION mode to minimize false positive rate. */
    PACE2_classification_mode openvpn_mode;
    /** Skype detection mode, default @ref SAFE_CLASSIFICATION.
     *  @arg Use @ref AGGRESSIVE_CLASSIFICATION mode to detect all Skype flows. This might cause a higher false positive rate.
     *  @arg Use @ref SAFE_CLASSIFICATION mode to minimize false positive rate. */
    PACE2_classification_mode skype_mode;
    /** Radius detection mode, default: @ref SAFE_CLASSIFICATION.
     *  @arg Use @ref AGGRESSIVE_CLASSIFICATION mode to detect radius flows on all ports, which causes higher computational costs.
     *  @arg Use @ref SAFE_CLASSIFICATION mode to detection radius flows on standard ports only. */
    PACE2_classification_mode radius_mode;
    PACE2_sit_params sit;                           /*!< SSL Session ID Tracking specific configuration parameters. */
    PACE2_DEPRECATED("cdn_caching_enabled is deprecated, the cdn caching feature is not used anymore.")
    u8 cdn_caching_enabled;                         /*!< Flag to indicate the enabled content delivery network caching
                                                         algorithm, default: @ref IPQ_FALSE.
                                                         @deprecated This setting was deprecated
                                                                     because the CDN caching feature is not used anymore. */
    PACE2_DEPRECATED(
        "RTP performance is deprecated here and is replaced by "
        "PACE2_config_s3_classification.dissector_config.rtp_performance_enabled.")
    u8 rtp_performance_enabled;                     /*!< Flag to indicate the enabled rtp performance measurements, default: @ref IPQ_FALSE. */

    struct PACE2_config_s3_cdc *cdc_conf_array;     /*!< Pointer to an array of custom defined classifications configuration parameters. */
    u32 cdc_number;                                 /*!< Number of custom defined classification. */

    u32 unclassified_after_mseconds;                /*!< Timeout in ms after which a flow get classified as protocol unclassified, default: 0 (off)  */
    u32 unclassified_after_packet_count;            /*!< Packet counter after which a flow get classified as protocol unclassified, default: 0 (off)  */

    /**  Packet counter after which a flow gets classified as application unclassified, default: 0 (off)
     *   The value must be higher or equal to @ref unclassified_after_packet_count. Otherwise the PACE 2 initialization
     *   will fail.
     */
    u32 application_unclassified_after_packet_count;

    /**  Timeout in ms after which a flow gets classified as application unclassified, default: 0 (off)
     *   The value must be higher or equal to @ref unclassified_after_mseconds. Otherwise the PACE 2 initialization
     *   will fail.
     */
    u32 application_unclassified_after_mseconds;

    struct dissector_config metadata_dissector_config;
                                                    /*!< Configuration structure to enable different dissector metadata. */
    struct PACE2_config_seca seca;                  /*!< SECA configuration parameters. */
    struct PACE2_config_deca deca;                  /*!< DECA configuration parameters. */

    struct PACE2_config_s3_pa_tags pa_tags;         /*!< Structure of configuration parameters for protocol and application grouping. */

    PACE2_midstream_mode midstream_mode;            /*!< Midstream traffic detection mode, default: @ref DEFAULT_MIDSTREAM_MODE. */
    PACE2_classification_event_mode classification_mode;
                                                    /*!< classification event creation mode default: @ref PACE2_DEFAULT_EVERY_PACKET_MODE */

    PACE2_DEPRECATED("Deprecated due to an extended midstream mode configuration, please use midstream_mode instead.")
    u8 midstream_enabled;                           /*!< Flag to indicate the enabled midstream detection mode, default: @ref IPQ_FALSE.
                                                         @deprecated Deprecated due to an extended midstream mode configuration.
                                                                     Instead, use @ref PACE2_config_s3_classification::midstream_mode. */
    u8 aggressive_enabled;                          /*!< Flag to indicate the enabled aggressive detection mode, default: @ref IPQ_FALSE. */

    struct {
        PACE2_boolean detect_encryption;                /*!< Flag to indicate if encrypted flows should be classified as well, default: @ref IPQ_TRUE. */
        PACE2_boolean use_only_safe_encrypted_pattern;  /*!< Flag to indicate if only safe encrypted patterns should be used, default: @ref IPQ_FALSE. */
    } bittorrent_config;

    struct {
        PACE2_boolean detect_encryption;                /*!< Flag to indicate if encrypted flows should be classified as well, default: @ref IPQ_TRUE. */
        PACE2_boolean use_only_safe_encrypted_pattern;  /*!< Flag to indicate if only safe encrypted patterns should be used, default: @ref IPQ_TRUE. */
    } edonkey_config;

    PACE2_boolean enable_new_protocols_after_du;    /*!< Flag indicating if newly added protocols should be used after DU, default: @ref IPQ_TRUE. */

    PACE2_boolean enable_cpi_with_disabled_events;  /*!< Flag to indicate if cpi functions are enabled for disabled
                                                         events, default: @ref IPQ_FALSE. */
    u8 enable_deprecated_applications;              /*!< Flag to indicate if the deprecation mode is enabled
                                                         default: @ref IPQ_FALSE. */
    u8 enable_new_tags_after_du;                    /*!< Flag indicating if newly changed tags should be used after DU,
                                                         default: @ref IPQ_FALSE. */
    u8 enable_fastpath_event;                       /*!< Flag which enables the @ref PACE2_FASTPATH_EVENT via event
                                                         queue instead of using the pace2_fastpath_get_status
                                                         function. The event will be generated per packet in stage3.
                                                         default: @ref IPQ_FALSE. */
    u8 enable_events_for_retransmissions;           /*!< Flag which enables the event creation for retransmitted
                                                         packets. This involves these events
                                                         @ref PACE2_CLASSIFICATION_RESULT event
                                                         @ref PACE2_BASIC_HTTP_REQUEST_EVENT
                                                         @ref PACE2_BASIC_HTTP_RESPONSE_EVENT
                                                         @ref PACE2_BASIC_QUIC_CLIENT_HELLO_EVENT
                                                         @ref PACE2_BASIC_SSL_SERVER_HELLO_EVENT
                                                         @ref PACE2_BASIC_SSL_CLIENT_HELLO_EVENT
                                                         @ref PACE2_OS_EVENT
                                                         @ref PACE2_BASIC_DNS_EVENT
                                                         @ref PACE2_BASIC_HTTP2_EVENT
                                                         @ref PACE2_BASIC_MDNS_EVENT
                                                         @ref PACE2_BASIC_NETBIOS_EVENT
                                                         @ref PACE2_BASIC_SIP_EVENT
                                                         @ref PACE2_BASIC_SSL_SESSION_TICKET_EVENT
                                                         @ref PACE2_BASIC_SSL_DNS_ALT_NAMES_EVENT .
                                                         default: @ref IPQ_TRUE. */
    u8 enable_name_resolution_events_for_every_packet; /*!< Flag which enables the event creation for every DNS packet.
                                                            @ref PACE2_BASIC_DNS_EVENT
                                                            @ref PACE2_BASIC_MDNS_EVENT
                                                            @ref PACE2_BASIC_NETBIOS_EVENT
                                                            default: @ref IPQ_FALSE. */
    enum PACE2_deca_mode deca_mode;                     /*!< DECA operation mode,
                                                             default: @ref DECA_EXCLUSIVE. */
    u8 enable_community_id;                             /*!< Flag which enables calculation of community id,
                                                             default: @ref IPQ_FALSE. */
    u8 enable_rst_pkt_handling;                         /*!< Flag to indicate if RST packet handling is enabled/disabled,
                                                             default: @ref IPQ_TRUE. */
    u8 enable_browser_classification;                   /*!< Flag which enables browser classification,
                                                             default: @ref IPQ_FALSE. */
};

/**
 * @brief Definition for one custom defined decoder.
 */
struct PACE2_config_s4_cdd {
    pace2_cdd_callback_t callback;                  /*!< Callback for the custom defined decoder to process PACE2 events. */
    void *callback_userptr;                         /*!< Pointer to user-specified data provided to the custom decoder. */
    u32 number_of_events;                           /*!< Number of different events this CDD will throw. */
    /** Array with number_of_events entries. Each entry reflects the size of the data for the specific event generated by the CDD. */
    u32 * event_size;
};

/**
 * @brief Enumeration to define the processing method for stage 4. See also @ref PACE2_config_s4_decoding.processing_method
 */
enum PACE2_config_s4_processing_method {
    PACE2_PROCESS_PACKETS,                          /*!< Signals the usage of pace2_s4_process_packet function for the stages 4 */
    PACE2_PROCESS_STREAM,                           /*!< Signals the usage of pace2_s4_process_stream function for the stages 4 */
};

/**
 * @brief Configurations for the event correlation feature.
 */
struct PACE2_config_event_correlation {
    u8 enable_event_correlation;                    /*!< Enables the event correlation feature. */
    PACE2_event_bitmask active_event_types;         /*!< Bitmask to state the enabled class correlations. */
};

/**
 * @brief Configuration for different hash tables.
 */
typedef struct pace2_config_hashtable {
    PACE2_table_max_size memory; /*!< Definition of the maximum size of a hash table given in kilobytes or number of
                                      elements. */
    PACE2_timestamp timeout;     /*!< Timeout of the hash table given in seconds. */
} PACE2_config_hashtable;

#ifndef PACE2_DISABLE_DECODER
/**
 * @brief Structure definition of packet decoding configuration parameters.
 */
struct PACE2_config_s4_decoding {
    u8 enabled;                                     /*!< Flag to indicate the enabled packet decoding, default: @ref IPQ_FALSE. */

    /**
     * Signals, which processing method will be used during runtime.
     * PACE2_PROCESS_STREAM means that stage 4 is called with the pace2_s4_process_stream
     * PACE2_PROCESS_PACKETS means that stage 4 is called with the pace2_s4_process_packet
     */
    enum PACE2_config_s4_processing_method processing_method;

    PACE2_protocols_mask active_advanced_decoders;      /*!< Struct defining decoding of which protocols shall be activated, default: all. */
    IPOQUE_CLASS_DECODER_BITMASK active_class_decoders; /*!< Struct defining which class decoders shall be activated, default: all. */
    IPOQUE_POSTPROC_PROTOCOL_BITMASK active_postproc_decoders; /*!< Struct defining which postproc decoders shall be activated, default: all. */
    /** Maximum buffer size for HTTP1/2 and FTP payload decompression.
     * The default value is 1 MiB */
    u32 http_gzip_buffer_size;

    /* checksumming */
    enum ipd_ip_checksumming ip_checksum;               /*!< activates the IP checksumming used by the TCP reassembly engine. Default:  @ref IPD_IP_CHECKSUM_IGNORE. */
    enum ipd_tcp_checksumming tcp_checksum;             /*!< activates the TCP checksumming used by the TCP reassembly engine. Default: @ref IPD_TCP_CHECKSUM_IGNORE. */
    u8 PACE2_DEPRECATED("do not use anymore") enable_ip_or_tcp_checksumming; /*!< Activates the IPv4 OR TCP checksumming
                                                                                  used by the TCP reassembly engine
                                                                                  @deprecated Do not use. */

    /* tcp reassembly values */
    /** Defines how long the decoder should wait for missing or bad TCP packets in milliseconds. Default: 8000. */
    u32 tcp_out_of_order_timeout;

    PACE2_DEPRECATED("Deprecated, use asymmetric_timeout in the general section instead.")
    u32 tcp_asymmetric_timeout;                         /*!< a flow is marked as asymmetric if, after the first packet, there comes no packet
                                                             in the opposite direction for this period of time;
                                                             0 disabled asymmetry detection (this is the default)
                                                             @deprecated Instead, use @ref PACE2_config_general::asymmetric_timeout. */
    u32 tcp_max_skew;                                   /*!< longest time an acknowledgment may precede the actual data that is acked
                                                             (in timestamp ticks); 0 disables message ordering (this is the default) */
    u32 tcp_max_window_size;                            /*!< Maximum memory used by the tcp reassembly stream.
                                                             The default is 16 MiB per stream direction */
    u8 tcp_hole_memory_fill_value;                      /*!< fill value for holes to mark them as bad data, recommended to keep it as default (which is 0) */

    /* dummy decoder */
    /** maximum amount of data the dummy decoder will allow to buffer for unknown connections. The default value is 150 Kbytes. */
    u64 dummy_decoder_max_buffer_size;

    /* http */
    /**
     * The maximum amount of data that is buffered for one direction when the decoder is waiting for data from the other direction.
     * When the limit is reached the decoder switches to asymmetric decoding trying to decode the direction with data available.
     * The default value is 100 Kbytes. */
    u32 http_max_direction_buffer_size;
    /** Defines how much HTTP payload data is buffered for each transaction before a
     *  @ref PACE2_CLASS_HTTP_REQUEST_DATA_TRANSFER / @ref PACE2_CLASS_HTTP_RESPONSE_DATA_TRANSFER event is issued.
     *  If set to zero the payload data will be buffered until the request/response is complete or the time
     *  limitation described below is reached. The default value is 256 KiB. */
    u64 http_class_payload_size_max;
    /** Defines how long HTTP payload data is buffered for each transaction before a
     *  @ref PACE2_CLASS_HTTP_REQUEST_DATA_TRANSFER / @ref PACE2_CLASS_HTTP_RESPONSE_DATA_TRANSFER event is issued.
     *  If set to zero the payload data will be buffered until the request/response is complete or the payload
     *  limitation described above is reached. The default value is 10 seconds. */
    u32 http_class_payload_timeout;
    /** Defines the maximum header length of a HTTP packet. Default value is 10 Kbytes. */
    u32 http_max_header_length;

    /* mail */
    /** Defines how much mail body payload data is buffered for each transaction before a
     *  @ref PACE2_CLASS_MAIL_SESSION_BODY_TRANSFER event is issued.
     *  If set to zero the payload data will be buffered until the mail body is complete or the time
     *  limitation described below is reached. The default value is 1 MiB. */
    u64 mail_class_body_size_max;
    /** Defines how long mail body payload data is buffered for each transaction before a
     *  @ref PACE2_CLASS_MAIL_SESSION_BODY_TRANSFER event is issued.
     *  If set to zero the payload data will be buffered until the mail body is complete or the payload
     *  limitation described above is reached. The default value: 900 seconds. */
    u32 mail_class_body_timeout;
    /** Enable transcoding of message headers encoded according to the "MIME Message Header Extensions for
     *  Non-ASCII Text" (RFC 2047) to UTF-8. Default: enabled. */
    u8 mail_class_enable_utf8_headers;
    /** Output attachments found in MIME messages (transported over IMAP, POP3, SMTP) through
     * ATTACHMENT_INFO/ATTACHMENT_CONTENT events, instead of through BODY_TRANSFER events, and base64-
     * decode attachment contents. Default: enabled*/
    u8 mail_class_enable_mime_attachments;
    /** Also generate @ref PACE2_CLASS_MAIL_ATTACHMENT_CONTENT_EXT events combining
     * @ref PACE2_CLASS_MAIL_ATTACHMENT_CONTENT with the according @ref PACE2_CLASS_MAIL_ATTACHMENT_INFO
     * event. Default: disabled.*/
    u8 mail_class_enable_extended_attachment_content_events;


    /* chat */
    /** Defines how long chat file transfer data is buffered for each file transfer before a
     *  @ref PACE2_CLASS_CHAT_CHATROOM_TRANSFER event with a segment of file transfer data is issued.
     *  If set to zero the file transfer data will be buffered until the file transfer is complete
     *  or the data size limitation described below is reached. The default value is 0. */
    u64 chat_class_filetransfer_dump_timeout;
    /** Defines how much chat file transfer data is buffered for each file transfer before a
     *  @ref PACE2_CLASS_CHAT_CHATROOM_TRANSFER event with a segment of file transfer data is issued.
     *  If set to zero the file transfer data will be buffered until the the file transfer is complete or the time
     *  limitation described above is reached. The default value is 0. */
    u64 chat_class_filetransfer_dump_max_size;

    /* rtp reorder buffer */
    /** Number of slots in the RTP reorder pool (needs to be >1 to enable RTP packet reordering). The default value is 0. */
    u64 rtp_buffer_global_packet_limit;
    /** Maximum number of RTP packets that will be buffered in one direction of one flow. The default value is 20. */
    u64 rtp_buffer_flow_packet_limit;
    /** Size in bytes of one slot in the RTP reorder pool. The default value is 1000 bytes. */
    u64 rtp_buffer_max_packet_size;

    /* dtmf */
    /** Enable decoding of DTMF signals in RTP flows. The default value is 0 (disabled). */
    u8 rtp_decode_dtmf_signals;
    /** Force decoding of in-band DTMF audio signals in RTP flows 0 (disabled). */
    u8 rtp_force_decode_dtmf_audio;
    /** integration point for an external AMR decoder */
    struct ipd_amr_decoder amr_decoder;
    /** integration point for an external AMR-WB decoder */
    struct ipd_amr_decoder amrwb_decoder;

    /* FTP */
    /** Defines how much file transfer data is buffered for each file transfer before a
     *  @ref PACE2_ADVANCED_FTP_DATA event with a segment of file transfer data is issued.
     *  If set to zero the file transfer data will be buffered until the file transfer is complete or the time
     *  limitation described below is reached. The default value is 1 MiB. */
    u64 ftp_class_data_dump_size_max;
    /** Defines how long file transfer data is buffered for each file transfer before a
     *  @ref PACE2_ADVANCED_FTP_DATA event with a segment of file transfer data is issued.
     *  If set to zero the file transfer data will be buffered until the file transfer is complete
     *  or the data size limitation described above is reached. The default value is 900 seconds. */
    u32 ftp_class_data_dump_timeout;

    /* BitTorrent decoder */
    /** Maximum BitTorrent message length. If the length field of a message is larger than this, the
     * respective direction of the flow is set into the invalid state. 'piece' messages are exempt from this
     * limit as their contents are not buffered until a message is complete. The largest variable-length
     * messages that need to be processed are 'bitfield' messages at
     * <tt>1&nbsp;+&nbsp;number_of_pieces_in_torrent/8</tt> bytes (rounded up). The default value is 131080 bytes. */
    u32 bittorrent_max_message_length;

    /* tls handling */
    u8 tls_enforce_processing;                      /*!< When off, encrypted tls data are ignored. Off by default. */

    u32 cdd_number;                                 /*!< Number of custom defined decoder. */
    struct PACE2_config_s4_cdd * cdd_conf_array;    /*!< Pointer to an array of custom defined decoder configurations. */

    /* MPEG-2 TS */
    /**
     * Enable aggressive decoding for MPEG-2 TS over HTTP (run for all content types
     * in the HTTP response). Default: @ref IPQ_FALSE.
     */
    PACE2_boolean mpeg_enable_aggresive_mode_over_http;

    /* lost TCP segment processing */
    /** Skip lost TCP segments and decode further received data. Default: @ref IPQ_FALSE.
     * If enabled, the decoding stage will throw events of type @ref PACE2_ADVANCED_TCP_EVENT with the event specific types
     * @ref PACE2_ADVANCED_TCP_LOST_SEGMENT and @ref PACE2_ADVANCED_TCP_SKIPPED_SEGMENT.
     */
    u8 tcp_skip_lost_segment;

    /** Skipping of lost TCP segments requires additional information for processing. These information are stored per 'hole'.
     * This variable defines the global number of available hole information slots. The available slots are shared by all TCP based decoder.
     * Default: 100.
     */
    u64 tcp_number_of_holes;

    /* NNTP */
    u8 nntp_decode_yencoding;         /*!< Determines if the yencoded data will be decoded or not. Default is 1 (enabled). */


    /* VoIP class */
    /** The size of the tracking table used for storing information about the VoIP sessions.
     * Default Size is 8 KiB. Each Session entry takes 432 Bytes.  */
    u64 voip_dialogs_hash_table_size;
    /** The VoIP dialog tracking table (stores information about a VoIP session) timeout in seconds.
     * If no H323/SIP/Jabber packets of the same flow arrive within this timeout, the Dialog will be removed from the table.
     * Default Timeout is 60 sec. */
    u32 voip_dialogs_hash_table_timeout;

    struct PACE2_config_scada {
        u8 originator_address_size;                     /*!< Size of SCADA originator address - can be different in each SCADA network.   */
        u8 asdu_address_size;                           /*!< Size of SCADA the ASDU address - can be different in each SCADA network. */
        u8 information_object_address_size;             /*!< Size of SCADA the information object address - can be different in each SCADA network. */
    } scada_config;

    /* RTP */
    /** Enable decoding of RTP without any correlated control flow like SIP, H.323 or similar. Default: @ref IPQ_TRUE. */
    u8 rtp_process_uncorrelated_flows;

    /* GTP */
    /** Disable the default behaviour of the decoder, to stop processing GTP packets when a invalid packet was seen.
     *  The default: @ref IPQ_FALSE, this is the consistent behaviour of all decoders that a invalid dataset will stop
     *  decoding for that specific flow. */
    u8 gtp_process_invalid_packet;

    /** Defines if postprocessing is enabled for asymmetric flows. Default: @ref IPQ_FALSE. */
    u8 http_enable_asymmetric_postprocessing;

    /* SIP/RTP */
    /** Disable correlation of RTP with any other control flow like SIP/H.323 or similar.
     * This means that additional information from the control flow will not be taken into account,
     * for example SIP BYE. This also means that all RTP flows are treated as uncorrelated.
     * The option @ref PACE2_config_s4_decoding::rtp_process_uncorrelated_flows is needed to
     * process uncorrelated flows.
     *
     * Default: 0.
     */
    u8 rtp_disable_correlation;

    /* SSL class */
    /** Defines how much SSL payload data is buffered for each transaction before a
     *  PACE2_CLASS_SSL_REQUEST_ENCRYPTED_DATA / PACE2_CLASS_SSL_RESPONSE_ENCRYPTED_DATA event is issued.
     *  If set to zero the payload data will be buffered until the request/response is complete or the time
     *  limitation described below is reached. The default value is 256 KiB.
     */
    u64 ssl_class_payload_size_max;

    /** Defines how long SSL payload data is buffered for each transaction before a
     *  PACE2_CLASS_SSL_REQUEST_ENCRYPTED_DATA / PACE2_CLASS_SSL_RESPONSE_ENCRYPTED_DATA event is issued.
     *  If set to zero the payload data will be buffered until the request/response is complete or the payload
     *  limitation described above is reached. The default value is 10 seconds.
     */
    u32 ssl_class_payload_timeout;

    /* Generic class */
    /**
     * Enables the correlation of flows to sessions in class decoders. Default: @ref IPQ_FALSE.
     */
    u8 enable_correlation_of_flows_to_sessions_in_class_decoders;

    /* FTP class */
    /**
     * Enables the reporting of FTP commands in the class decoder. Default: @ref IPQ_FALSE.
     */
    u8 ftp_class_enable_reporting_ftp_commands;

    /* HTTP */
    /**
     * Maximum line length in the HTTP header, default: 8192.
     * This config option is used for midstream / DIF scenarios to find the flow direction if not known yet.
     * Modification of this setting will impact the number of bytes buffered by the HTTP decoder to search for
     * a new header line. In a midstream / DIF scenario, the first HTTP header line might not be provided to the
     * customer as a part of any event if it is longer than this setting.
     */
    u16 http_max_header_line_len;

    /* RTP */
    /**
     * The starting size given in kilobytes or number of elements of the tracking table used for storing information
     * about the RTP correlations.
     * The RTP correlation tracking table (stores information to correlate an RTP stream to a control flow) timeout
     * in seconds. If no H323/SIP/Jabber packets of the same session arrive within this timeout, the RTP information
     * will be removed from the table.
     * Default size is 8 KiB.
     * Default timeout is 900 s. */
    PACE2_config_hashtable rtp_connection_hash_table;

    /* H323 */
    /**
     * The starting size given in kilobytes or number of elements of the tracking table used to store ports that are
     * announced by an H.225 call signalization. Additionally the timeout value given in seconds after which an entry is
     * removed from the table.
     * Default size is 8 KiB.
     * Default timeout is 900 s. */
    PACE2_config_hashtable h323_connection_hash_table;

    /* TCPR */
    /** Maximum memory used by the tcp reassembly per flow.
     * The limit applies to the sum of memory usage of both directions.
     * Default is 0. When the value is 0 only the
     * tcp_max_window_size applies per direction. */
    u32 tcp_max_flow_buffer_size;

    /* SCTPR */
    /** Skip lost SCTP segments and decode further received data. Default: @ref IPQ_FALSE. */
    u8 sctp_skip_lost_segment;

    /**
     * Skipping of lost SCTP segments requires additional information for processing. These information are stored per
     * 'hole'. This variable defines the global number of available hole information slots. The available slots are
     * shared by all SCTP based decoder. Default: 100.
     */
    u64 sctp_number_of_holes;

    /**
     * Lost SCTP segments generate holes but the size of these holes is unknown.
     * A smaller hole size consumes less memory. Bigger holes can be an indicator for the resynchronization decoder
     * strategy after such holes, as the complete context might be lost.
     * Minimum 1 byte, must not be 0.
     * Default: 1 byte.
     */
    u32 sctp_hole_size;

    /** Fill value for holes to mark them as bad data, recommended to keep it as default (which is 0) */
    u8 sctp_hole_memory_fill_value;

    /** Defines how long the decoder should wait for missing SCTP packets in milliseconds. Default: 8000 ms. */
    u32 sctp_out_of_order_timeout;

    /** Maximum memory used by the SCTP reassembly stream per direction in bytes. Default: 8 MiB. */
    u64 sctp_stream_direction_memory_limit;

    /**
     * Maximum memory used by the SCTP reassembly stream (sum of both directions) in bytes.
     * This value must be at least the size of @ref sctp_stream_direction_memory_limit.
     * Default: 16 MiB.
     */
    u64 sctp_stream_memory_limit;

    /** Defines the maximum header line length of a HTTP/2 packet. Default value is 8192. */
    u32 http2_max_header_line_len;

    /** Enable provision of HTTP header information decoded, from each packet, in an HTTP transaction class event;
     *  It helps in the case where the header info is split in multiple packets to not wait for the last part of the
     *  request or response to get in to see the info prior gotten. */
    u8 http_enable_intermediate_header_information_result;

    /** Enables all supported UTF8 conversions (in all decoders). Default: disabled.
     *  For historical compatibility reasons, the UTF-8 conversion is also enabled if
     *  SSL or MAIL class decoders are activated. */
    u8 enable_utf8_conversions;

    /** Allow decoding of non-RFC-conform header lines. Default value is IPQ_FALSE. */
    u8 http_allow_nonconforming_content;
};
#endif
/**
 * @brief Structure definition of dynamic upgrade configuration parameters.
 */
struct PACE2_config_dynamic_upgrade {
    u8 enabled;                                     /*!< Flag to indicate the enabled dynamic upgrade, default: @ref IPQ_FALSE. */
    u32 reserve_bytes;                              /*!< Amount of bytes to reserve for library upgrades. A value of 0 defaults to 10% of the corresponding structure. */
    pace2_dlopen_t pace2_dlopen;                    /*!< Function pointer to a POSIX compatible dlopen function, default: NULL. */
    pace2_dlsym_t pace2_dlsym;                      /*!< Function pointer to a POSIX compatible dlsym function, default NULL. */
    pace2_dlclose_t pace2_dlclose;                  /*!< Function pointer to a POSIX compatible dlclose function, default NULL. */
    /** Maximum lifetime of flows and subscribers. Should be the higher value of either the flow timeout or subscriber
     *  timeout. Is set automatically when internal tracking is used for both. Has to be set in cases of external tracking. */
    PACE2_timestamp release_memory_age;
};

/**
 * @brief Structure definition of general configuration parameters.
 */
struct PACE2_config_general {
    u32 number_of_threads;                                      /*!< Number of running threads, default: 1. The used thread ID's have to be
                                                                     in between 0 and @ref number_of_threads - 1.*/
    PACE2_table_max_size max_number_of_events;                  /*!< Maximum number of maximum size events which can be allocated inside
                                                                     the event queue. If smaller events are allocated then more will fit,
                                                                     as long as they do not require a larger total amount of memory.
                                                                     Value is per thread, default: 100 elements. */
    struct global_epol {
        PACE2_epol_structure policy;                            /*!< Global event policy which indicates the enabled events,
                                                                     default: all events are enabled. */
        PACE2_epol_operation_mode policy_op_mode;               /*!< Operation mode of the event policy, default @ref PACE2_PROGRESSIVE. */
    } event;                                                    /*!< Structure to bind all policy variables. */

    pace2_malloc_t pace2_alloc;                                 /*!< Function pointer to a memory allocation function, default: NULL.
                                                                     Must be set by the application. */
    pace2_free_t pace2_free;                                    /*!< Function pointer to a memory free function, default: NULL.
                                                                     Must be set by the application. */
    pace2_realloc_t pace2_realloc;                              /*!< Function pointer to a memory realloc function, default: NULL.
                                                                     Must be set by the application. */
    void *memory_callback_user_ptr;                             /*!< Pointer to allocation function user data: will be passed as the
                                                                     @c user_ptr argument to the the allocation functions defined above;
                                                                     default: NULL. */

    pace2_logging_callback logging_callback; /*!< Callback function pointer to handle logging messages from PACE2, default:
                                                  NULL. */

    void * logging_callback_user_ptr; /*!< User definable pointer, which is available inside the logging_callback
                                           callback, default: NULL. */
    /** Initialize processing threads of PACE 2 module discretely, default: @ref IPQ_FALSE.
     *  Setting this option to @ref IPQ_TRUE is necessary in the case that the processing threads
     *  are distributed over different CPUs in a multi-socket system to ensure that memory allocations
     *  are done on the CPU of the assigned processing thread. Otherwise the CPU bus will be burdened and
     *  the performance could decrease. If the option is set to @ref IPQ_TRUE each thread has to call
     *  the function @ref pace2_init_thread after setting the CPU affinity to initialize the PACE 2 thread instance.
     */
    PACE2_boolean discrete_thread_init;

    struct PACE2_config_event_correlation event_correlation;    /*!< Structure of the event correlation. */

    u32 pace2_instance_id;                                      /*!< Current PACE 2 instance ID, used for ID generation.
                                                                     Default: 0 */
    u8 instance_id_storage_len;                                 /*!< PACE 2 instance ID length in bit, used for ID
                                                                     generation.
                                                                     Default: 0, valid range [0..32] */
    u8 thread_id_storage_len;                                   /*!< Thread ID length in bit, used for ID generation.
                                                                     Default: 8, valid range [0..8] */
    PACE2_generate_specific_id_callbacks id_callbacks;          /*!< Structure holding callbacks for ID generation. */

    PACE2_frame_buffer_callbacks frame_buffer_callbacks;        /*!< Structure holding callbacks for the frame buffer. */

    /** Array of pointers for processing states per thread, used for monitoring. Set NULL to disable monitoring,
     *  otherwise the array size must match the number of threads. This is an array of pointers to avoid cache problems
     *  in a multi threading environment. It is advised to allocate the memory for the state of each thread in the scope
     *  of that specific thread, i.e. NUMA-aware. */
    volatile u32 ** processing_states;

    /** Input/output timestamp resolution (in ticks per second), default: 1000 (milliseconds resolution).
     *  This is the unit to be used when passing timestamps to PACE 2 functions, e.g. in @ref pace2_s1_process_packet.
     *  Provided value must be a multiple of 1000, ranging from 1000 (milliseconds resolution) to 1000000000
     *  (nanoseconds resolution). This setting does not influence the internal processing behavior or the units of any
     *  other configuration setting. */
    u32 io_timestamp_resolution;

    u32 asymmetric_timeout;                                     /*!< A flow is marked as asymmetric if, after the first
                                                                     packet, there comes no packet in the opposite
                                                                     direction for this period of time.
                                                                     The value 0 disables this feature.
                                                                     Default: 0 second. */
    u32 asymmetric_packet_limit;                                /*!< A flow is marked as asymmetric if no packet, with
                                                                     payload, in both directions is received after this
                                                                     number of packets.
                                                                     The value 0 disables this feature.
                                                                     Default: 10. */

    /** Enable tracking of memory usage for each memory scope. @ref IPQ_FALSE.
     *  The counters for memory usage can be accessed using the profiling interface.
     *  Not recommended in productive use, as it may negatively affect the performance. */
    u8 memory_enable_scope_tracking;

    /** Enable the scope ID when calling the free wrapper. @ref IPQ_FALSE.
     *  This is usually only needed if different malloc/free wrapper implementations per scope ID shall be used.
     *  Enabling this feature will require about 8 bytes more per malloc. When disabling this feature (the default) the
     *  free wrapper will always called with PACE2_SCOPE_UNDEFINED. */
    u8 memory_enable_scope_free;
};

#ifdef IPOQUE_ENABLE_LICENSING
struct PACE2_config_licensing
{
    const char *license_file;
};
#endif

/**
 * @brief The config struct for the IPC feature is part of the @ref PACE2_global_config.
 */
typedef struct pace2_ipc_config
{
    /** Enabled or disabled the IPC feature. Default: @ref IPQ_FALSE */
    PACE2_boolean                   enabled;

    /** The maximum size of each IPC message. Default: 1400 */
    u16                             max_message_size;

    /** Number of bytes the internal message buffer can store. Default: 1048576 */
    u32                             message_buffer_size;

    /** User pointer to pass to @ref PACE2_ipc_config.send_blob_callback. */
    void *                          send_blob_callback_user_ptr;

    /** Callback to inform user application about IPC blob to send. */
    pace2_ipc_send_blob_callback_fn send_blob_callback;
} PACE2_ipc_config;

/**
 * @brief Sub-structure of PACE2_global_config for all PLUG-IN related settings.
 */
typedef struct pace2_config_plugins
{
    /** List of additional PLUG-IN configurations to load. */
    PACE2_plugin_configuration *            configurations;

    /** Count of registered PLUG-INs. */
    u32                                     configurations_len;

    /** Array of pointers to PLUG-IN processing states, used for monitoring. Set NULL to disable monitoring, otherwise
      * the array size must match the number of threads. Each thread needs its own pointer to a processing state. It is
      * advised to allocate the memory for the state of each thread in the scope of that specific thread,
      * i.e. NUMA-aware.*/
    volatile u32 **                         processing_states;
} PACE2_config_plugins;

typedef struct pace2_class_config_ml {
    u32 class_id;                     /*!< The class id specifies which ML class the configuration applies to. */

    u32 packet_limit;                 /*!< This config option, if set to any value other than 0, specifies the maximum
                                           amount of stage3 packets that will be processed by the machine learning
                                           class for any particular flow. I.e. after that amount of packets no
                                           machine learning event will be triggered anymore this class and for the
                                           current flow. */

    u32 reserved[16];
} PACE2_class_config_ml;

/**
 * @brief Sub-structure of PACE2_global_config for all machine learning related settings.
 */
typedef struct pace2_config_ml
{
    PACE2_boolean enable_module;      /*!< Setting this option to IPQ_TRUE will enable the whole machine learning
                                           module. */
    PACE2_boolean enable_all_classes; /*!< Setting this option to IPQ_TRUE will enable all supported machine learning
                                           classes. */
    u32 * enabled_class_ids;          /*!< This list of IDs defines which individual machine learning learning classes
                                           should be enabled. Which ID defines which specific class can be derived from
                                           @ref pace2_ml_get_class_info. This has a direct influence on how much
                                           memory/flow memory is used. This option is shadowed by enabling
                                           @ref PACE2_config_ml::enable_all_classes. */
    u32 enabled_class_ids_len;        /*!< Defines the number of classes that are enabled via
                                           @ref PACE2_config_ml::enabled_class_ids. */

    u32 packet_limit;                 /*!< This config option, if set to any value other than 0, specifies the maximum
                                           amount of stage3 packets that will be processed by the machine learning
                                           module for any particular flow. I.e. after that amount of packets no
                                           machine learning event will be triggered anymore for the current flow. */

    struct pace2_class_config_ml * class_configs; /*!< This list provides options for configuring the individual classes
                                                       of the machine learning module. The list has to be valid at least
                                                       until @ref pace2_init has finished. */
    u32 class_configs_len;  /*!< Defines the length of the list provided via @ref PACE2_config_ml::class_configs. */

    u32 reserved[13];
} PACE2_config_ml;

/**
 * @brief Sub-structure of PACE2_global_config for all multi consumer handling related settings.
 */
typedef struct pace2_config_mch
{
    /** The number of bytes that the MCH will buffer before data will be removed. This configuration
     *  will help to avoid data removal for late matching event producer conditions. Default is 4096 bytes.*/
    u64 start_buffer_size;

    /** = 1 if the tracking of lost segments for inactive flows is skipped, = 0 otherwise. Default is 0.*/
    u8 skip_lost_segment_tracking_for_inactive_flows;
} PACE2_config_mch;

/**
 * @brief Structure definition of PACE 2 configuration parameters.
 */
struct PACE2_global_config
{
    struct PACE2_config_general             general;                            /*!< General configuration parameters. */
    struct PACE2_config_tracking            tracking;                           /*!< Tracking specific configuration parameters. */
    struct PACE2_config_dynamic_upgrade     dynamic_upgrade;                    /*!< Dynamic upgrade specific configuration parameters. */
    struct PACE2_config_s1_preparing        s1_preparing;                       /*!< Stage 1 specific configuration parameters. */
    struct PACE2_config_s2_reordering       s2_reordering;                      /*!< Stage 2 specific configuration parameters. */
    struct PACE2_config_s3_classification   s3_classification;                  /*!< Stage 3 specific configuration parameters. */
#ifndef PACE2_DISABLE_DECODER
    struct PACE2_config_s4_decoding         s4_decoding;                        /*!< Stage 4 specific configuration parameters. */
#endif
#ifdef IPOQUE_ENABLE_LICENSING
    struct PACE2_config_licensing           licensing;                          /*!< licensing configuration parameters. */
#endif

    struct pace2_ipc_config                 ipc;                                /*!< IPC configuration parameters. */
    struct pace2_config_plugins             plugin;                             /*!< PLUG-IN specific configurations. */

    struct pace2_config_ml                  ml;                                 /*!< Machine Learning (ML) specific
                                                                                     configurations. */

    struct pace2_config_mch                 mch;                                /*!< MCH specific parameters. */
};

/* Public Prototypes ************************************************************/

/**
 * @brief Initialize a configuration structure with default values.
 *
 * The function initializes a configuration structure with default values.
 * The default values can be found in the descriptions of the configuration parameters.
 *
 * @param[in]   config  Constant pointer to a configuration structure.
 */
void pace2_init_default_config(struct PACE2_global_config * const config);

/**
 * @brief Initialize license configuration structure with the given license file.
 *
 * The function initializes a license configuration structure with the given license file.
 *
 * @param[in]   config  Constant pointer to a configuration structure.
 * @param[in]   license_file path to the license file
 */
void pace2_set_license_config(struct PACE2_global_config * const config,
                              const char *license_file);

/**
 * @brief Update the configuration
 *
 * The function make a dynamic reconfiguration of PACE 2 possible during run
 * time. Currently only the reconfiguration of the enabled protocols and
 * applications is supported. Before using this function all packet processing
 * threads should be stopped to avoid configuration inconsistencies.
 *
 * @param[in]   pace2_module                                            Constant pointer to the PACE 2 module
 * @param[in]   config                                                  Constant pointer to a configuration structure.
 * @retval      PACE2_SUCCESS                                           @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_PARAM_MISSING                                     @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval      PACE2_THREAD_ID_INVALID                                 @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval      PACE2_FAILURE                                           @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval      PACE2_CDD_CALLBACK_NOT_SET                              @copydoc PACE2_return_state::PACE2_CDD_CALLBACK_NOT_SET
 * @retval      PACE2_CDD_CONFIG_ARRAY_NOT_SET                          @copydoc PACE2_return_state::PACE2_CDD_CONFIG_ARRAY_NOT_SET
 * @retval      PACE2_CDD_DECODER_DISABLED                              @copydoc PACE2_return_state::PACE2_CDD_DECODER_DISABLED
 * @retval      PACE2_CDD_EVENT_SIZE_NOT_SET                            @copydoc PACE2_return_state::PACE2_CDD_EVENT_SIZE_NOT_SET
 * @retval      PACE2_DYNAMIC_UPGRADE_RELEASE_MEMORY_AGE_TOO_SMALL      @copydoc PACE2_return_state::PACE2_DYNAMIC_UPGRADE_RELEASE_MEMORY_AGE_TOO_SMALL
 * @retval      PACE2_INVALID_DYNAMIC_UPGRADE_DLCLOSE_FUNCTION_PTR      @copydoc PACE2_return_state::PACE2_INVALID_DYNAMIC_UPGRADE_DLCLOSE_FUNCTION_PTR
 * @retval      PACE2_INVALID_DYNAMIC_UPGRADE_DLOPEN_FUNCTION_PTR       @copydoc PACE2_return_state::PACE2_INVALID_DYNAMIC_UPGRADE_DLOPEN_FUNCTION_PTR
 * @retval      PACE2_INVALID_DYNAMIC_UPGRADE_DLSYM_FUNCTION_PTR        @copydoc PACE2_return_state::PACE2_INVALID_DYNAMIC_UPGRADE_DLSYM_FUNCTION_PTR
 * @retval      PACE2_INVALID_GENERAL_ALLOC_FUNCTION_PTR                @copydoc PACE2_return_state::PACE2_INVALID_GENERAL_ALLOC_FUNCTION_PTR
 * @retval      PACE2_INVALID_GENERAL_FREE_FUNCTION_PTR                 @copydoc PACE2_return_state::PACE2_INVALID_GENERAL_FREE_FUNCTION_PTR
 * @retval      PACE2_INVALID_GENERAL_REALLOC_FUNCTION_PTR              @copydoc PACE2_return_state::PACE2_INVALID_GENERAL_REALLOC_FUNCTION_PTR
 * @retval      PACE2_NO_DYNAMIC_UPGRADE                                @copydoc PACE2_return_state::PACE2_NO_DYNAMIC_UPGRADE
 * @retval      PACE2_NO_FLOW_MEMORY_AGE_SET                            @copydoc PACE2_return_state::PACE2_NO_FLOW_MEMORY_AGE_SET
 * @retval      PACE2_NO_SUBSCRIBER_MEMORY_AGE_SET                      @copydoc PACE2_return_state::PACE2_NO_SUBSCRIBER_MEMORY_AGE_SET
 * @retval      PACE2_S2_REORDERING_PACKET_BUFFER_SIZE_TOO_SMALL        @copydoc PACE2_return_state::PACE2_S2_REORDERING_PACKET_BUFFER_SIZE_TOO_SMALL
 * @retval      PACE2_TRACKING_FLOW_TIMEOUT_TOO_SMALL                   @copydoc PACE2_return_state::PACE2_TRACKING_FLOW_TIMEOUT_TOO_SMALL
 * @retval      PACE2_STAGE3_ERROR_SET_BITTORRENT_CONFIG                @copydoc PACE2_return_state::PACE2_STAGE3_ERROR_SET_BITTORRENT_CONFIG
 */
PACE2_return_state pace2_reconfigure(PACE2_module * const pace2_module,
                                     struct PACE2_global_config const * const config);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_CONFIG_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
