/** @file      pace2_config_dict.h
  * @brief     PACE 2 configuration dictionary include file.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_CONFIG_DICT_H
#define PACE2_CONFIG_DICT_H

/* Includes *********************************************************************/
#include "pace2_basic_def.h"
#include "pace2_event_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Types *****************************************************************/

/**
 * @brief Handle to PACE2 configuration dictionary.
 */
typedef struct pace2_config_dict * PACE2_config_dict;


/**
 * @brief Defines the possible values which type of value the @ref PACE2_config_entry can contain.
 */
enum PACE2_config_entry_type {
    PACE2_ENTRY_TYPE_POINTER,   /*!< Entry type is a Pointer. */
    PACE2_ENTRY_TYPE_U8,        /*!< Entry type is a u8. */
    PACE2_ENTRY_TYPE_U16,       /*!< Entry type is a u16. */
    PACE2_ENTRY_TYPE_INT,       /*!< Entry type is a int. */
    PACE2_ENTRY_TYPE_U32,       /*!< Entry type is a u32. */
    PACE2_ENTRY_TYPE_U64,       /*!< Entry type is a u64. */
    PACE2_ENTRY_TYPE_ENUM,      /*!< Entry type is an enumeration. */
    PACE2_ENTRY_TYPE_BOOLEAN,   /*!< Entry type is a @ref PACE2_boolean. */
    PACE2_ENTRY_TYPE_STRING     /*!< Entry type is a null-terminated string. */
};

/**
 * @brief Defines the possible values for a @ref PACE2_config_entry
 */
union PACE2_config_entry_value {
    void * entry_value_pointer; /*!< Entry value as Pointer.
                                     When used in @ref pace2_config_dict_set, the user has to ensure that the pointer
                                     stays valid as long as the configuration dictionary is used. */
    u8 entry_value_u8;          /*!< Entry value as u8. */
    u16 entry_value_u16;        /*!< Entry value as u16. */
    int entry_value_int;        /*!< Entry value as int. */
    u32 entry_value_u32;        /*!< Entry value as u32. */
    u64 entry_value_u64;        /*!< Entry value as u64. */
    int entry_value_enum;       /*!< Entry value as enumeration.
                                     The possible enum values are the same as used for the configuration
                                     of the @ref PACE2_global_config. */
    PACE2_boolean entry_value_boolean; /*!< Entry type as @ref PACE2_boolean. */
    char const * entry_value_string; /*!< Entry as a null-terminated string.
                                     When used in @ref pace2_config_dict_set, the user must ensure the pointer stays
                                     valid as long as the configuration dictionary is used. */
};

/**
 * @brief Type which holds all information for a configuration entry.
 */
struct PACE2_config_entry {
    char const * name;                      /*!< Null-terminated name of the configuration entry. */
    union PACE2_config_entry_value value;   /*!< Value of the configuration entry. */
    enum PACE2_config_entry_type type;      /*!< Value type of the configuration entry. */
};

/**
 * @brief Type which holds the name and the type of the value for a configuration entry.
 */
struct PACE2_config_key {
    char const * name;                      /*!< Null-terminated name of the configuration key. */
    enum PACE2_config_entry_type type;      /*!< Value type of the configuration key. */
};

/**
 * @brief Defines the possible values which protocol bitmasks exists and can be manipulated in the PACE2 configuration.
 *
 * The value of the enumeration can be used in @ref pace2_config_dict_set_protocol_bitmask to set the
 * according bitmask.
 */
enum PACE2_config_protocol_bitmask {
    PACE2_CLASSIFICATION_BITMASK,       /*!< Classification bitmask */
    PACE2_DECODING_BITMASK,             /*!< Decoding bitmask */
};

/**
 * @brief Defines the possible values which event bitmasks exists and can be manipulated in the PACE2 configuration.
 *
 * The value of the enumeration can be used in @ref pace2_config_dict_set_event_bitmask to set the
 * according bitmask.
 */
enum PACE2_config_event_bitmask {
    PACE2_EVENT_POLICY_BITMASK,         /*!< Event policy bitmask */
    PACE2_EVENT_CORRELATION_BITMASK,    /*!< Event correlation bitmask */
};

/* Public Prototypes ************************************************************/

/**
 * @brief Initialize a PACE2 global configuration with default values.
 *
 * This function will mainly replace @ref pace2_init_default_config. It is basically a wrapper of this function but
 * creates a new @ref PACE2_global_config instance and allocates memory for it.

 * This ensures the use of an up-to-date data struct. For example if @ref PACE2_global_config gets bigger,
 * @ref pace2_config_dict_create_default ensures the use of the correct new data struct.
 * The default values can be found in the descriptions of the configuration parameters.
 *
 * @note The arguments malloc_func, realloc_func and free_func are only used by the configuration
 * dictionary, they are not passed to PACE2. The PACE2 malloc function (for example) must be set via
 * @ref pace2_config_dict_set with the @ref PACE2_config_entry::name "name" "general.pace2_alloc".
 *
 * @param[in]   malloc_func       Pointer to the memory allocation function.
 * @param[in]   realloc_func      Pointer to the memory reallocation function.
 * @param[in]   free_func         Pointer to the memory free function.
 * @param[in]   user_ptr          Pointer to user defined data.
 *
 * @returns     The new created @ref PACE2_config_dict instance or NULL on error.
 */
PACE2_config_dict pace2_config_dict_create_default(pace2_malloc_t malloc_func,
                                                   pace2_realloc_t realloc_func,
                                                   pace2_free_t free_func,
                                                   void * user_ptr);

/**
 * @brief free a PACE2 configuration dictionary.
 *
 * A created @ref PACE2_config_dict instance has to be freed after usage.
 * This function will take care of it.
 *
 * @param[in]   config          Pointer which hold the config.
 */
void pace2_free_config_dict(PACE2_config_dict config);

/**
 * @brief Set a PACE2 configuration dictionary value.
 *
 * The function is used to set a value in the configuration dictionary.
 *
 * @param[in]   config                              Pointer which hold the config.
 * @param[in]   entry                               Name, data type and value of the config parameter.
 * @retval      PACE2_SUCCESS                       @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_PARAM_MISSING                 @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval      PACE2_CONFIG_DICT_TYPE_MISMATCH     @copydoc PACE2_return_state::PACE2_CONFIG_DICT_TYPE_MISMATCH
 * @retval      PACE2_CONFIG_DICT_UNKOWN_ENTRY      @copydoc PACE2_return_state::PACE2_CONFIG_DICT_UNKOWN_ENTRY
 * @retval      PACE2_FEATURE_NOT_AVAILABLE         @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 * @retval      PACE2_OUT_OF_MEMORY                 @copydoc PACE2_return_state::PACE2_OUT_OF_MEMORY
 * @retval      PACE2_FAILURE                       @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_config_dict_set(PACE2_config_dict config, struct PACE2_config_entry const * const entry);

/**
 * @brief Set PACE2 protocol bitmask.
 *
 * The function is used to set all protocols in the bitmask.
 * It has to be specified which bitmask you want to set.
 *
 * @param[in]   config          Pointer which hold the config.
 * @param[in]   mask            Enumration value which protocol bitmask to set.
 * @param[in]   enable          If @ref IPQ_TRUE all protocols will be enabled.
 *                              If @ref IPQ_FALSE all protocols will be disabled.
 * @retval      PACE2_SUCCESS               @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_PARAM_MISSING         @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval      PACE2_FEATURE_NOT_AVAILABLE @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_config_dict_set_protocol_bitmask(PACE2_config_dict config,
                                                          enum PACE2_config_protocol_bitmask mask,
                                                          PACE2_boolean const enable);

/**
 * @brief Set PACE2 class decoder bitmask.
 *
 * The function is used set all class decoder in the bitmask.
 *
 * @param[in]   config          Pointer to the configuration dictionary.
 * @param[in]   enable          If @ref IPQ_TRUE all class decoder will be enabled.
 *                              If @ref IPQ_FALSE all class decoder will be disabled.
 * @retval      PACE2_SUCCESS               @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_PARAM_MISSING         @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval      PACE2_FEATURE_NOT_AVAILABLE @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_config_dict_set_class_decoder_bitmask(PACE2_config_dict config,
                                                               PACE2_boolean const enable);

/**
 * @brief Set PACE2 postproc decoder bitmask.
 *
 * The function is used set all postproc decoder in the bitmask.
 *
 * @param[in]   config          Pointer to the configuration dictionary.
 * @param[in]   enable          If @ref IPQ_TRUE all postproc decoder will be enabled.
 *                              If @ref IPQ_FALSE all postproc decoder will be disabled.
 * @retval      PACE2_SUCCESS               @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_PARAM_MISSING         @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval      PACE2_FEATURE_NOT_AVAILABLE @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_config_dict_set_postproc_decoder_bitmask(PACE2_config_dict config,
                                                                  PACE2_boolean const enable);

/**
 * @brief Set PACE2 event type bitmask.
 *
 * The function is used set all event types in the bitmask.
 *
 * @param[in]   config          Pointer to the configuration dictionary.
 * @param[in]   mask            Enumration value which event bitmask to set.
 * @param[in]   enable          If @ref IPQ_TRUE all events will be enabled.
 *                              If @ref IPQ_FALSE all events will be disabled.
 * @retval      PACE2_SUCCESS       @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_PARAM_MISSING @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 */
PACE2_return_state pace2_config_dict_set_event_bitmask(PACE2_config_dict config,
                                                       enum PACE2_config_event_bitmask mask,
                                                       PACE2_boolean const enable);

/**
 * @brief Set the config event policy for a group of events.
 *
 * @param[in]   config          Pointer to the configuration dictionary.
 * @param[in]   event_group     Event group for which the operation has to be performed.
 * @param[in]   mask            Enumration value which event bitmask to set.
 * @param[in]   enable_group    If @ref IPQ_TRUE the events of the given group will be enabled.
 *                              If @ref IPQ_FALSE the events of the given group will be disabled.
 * @retval      PACE2_SUCCESS       @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_PARAM_MISSING @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval      PACE2_FAILURE       @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_config_dict_set_event_group(PACE2_config_dict config,
                                                     PACE2_event_groups const event_group,
                                                     enum PACE2_config_event_bitmask mask,
                                                     PACE2_boolean const enable_group);

/**
 * @brief Get a specific PACE2 configuration entry value.
 *
 * The function is used to retrieve a value from the configuration.
 * The member @ref PACE2_config_entry::name has to bet set and the function will fill in
 * @ref PACE2_config_entry::type and @ref PACE2_config_entry::value.
 *
 * @param[in]       config                              Pointer to the configuration dictionary.
 * @param[in,out]   entry                               Name, data type and value of the config parameter.
 *                                                      The name property must be set and the function will
 *                                                      complete the type and value.
 * @retval      PACE2_SUCCESS                       @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_PARAM_MISSING                 @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval      PACE2_CONFIG_DICT_UNKOWN_ENTRY      @copydoc PACE2_return_state::PACE2_CONFIG_DICT_UNKOWN_ENTRY
 * @retval      PACE2_FEATURE_NOT_AVAILABLE         @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 * @retval      PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS    @copydoc PACE2_return_state::PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS
 */
PACE2_return_state pace2_config_dict_get(PACE2_config_dict config, struct PACE2_config_entry * const entry);

/**
 * @brief Get all available PACE2 config keys.
 *
 * The function returns all config keys as list. They are structured in the following way:
 * [{"name_of_config_option", type},...,{NULL, PACE2_ENTRY_TYPE_POINTER}]
 * Like stated above this list is delimited by the tuple {NULL, PACE2_ENTRY_TYPE_POINTER}.
 *
 * @param[out] keys   Pointer to the first @ref PACE2_config_key. This pointer must not be freed.
 * @param[out] length Optional pointer to a length field
 *
 * @retval      PACE2_SUCCESS                       @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_PARAM_MISSING                 @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 */
PACE2_return_state pace2_config_dict_get_available_keys(struct PACE2_config_key const ** keys, u32 * length);

/**
 * @brief Get all PACE2 config entries.
 *
 * The function returns all config entries as list. They are structured in the following way:
 * [{"name_of_config_option", value, type},...,{NULL, 0, PACE2_ENTRY_TYPE_POINTER}]
 * Like stated above this list is delimited by the tuple {NULL, 0, PACE2_ENTRY_TYPE_POINTER}.
 * The function allocates memory which has to be freed with @ref pace2_free_config_dict_entries.
 * @note This function can be very performance intensive.
 *
 * @param[in]   config            Pointer which holds the config.
 *
 * @returns     Pointer to first @ref PACE2_config_entry or NULL on error.
 */
struct PACE2_config_entry * pace2_config_dict_get_all_entries(PACE2_config_dict config);

/**
 * @brief Free memory allocated by @ref pace2_config_dict_get_all_entries.
 *
 * @param[in]   config          Pointer which hold the config.
 * @param[in]   entries         Pointer to the entries to free.
 */
void pace2_free_config_dict_entries(PACE2_config_dict config,
                                    struct PACE2_config_entry * const entries);

/**
 * @brief Get a @ref PACE2_global_config from a configuration dictionary.
 *
 * @param[in] config Pointer which hold the config dict.
 *
 * @return A pointer to a @ref PACE2_global_config, this can be passed @ref pace2_init.
 */
struct PACE2_global_config * pace2_config_dict_get_global_config(PACE2_config_dict config);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_CONFIG_DICT_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
