/** @file      pace2_cpi_classification.h
  * @brief     PACE 2 component programming interface of the classification.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_CPI_CLASSIFICATION_FUNC_H
#define PACE2_CPI_CLASSIFICATION_FUNC_H

#include "pace2_basic_def.h"
#include "pace2_events.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Types *****************************************************************/

/**
 * @brief Return state of component programming interface functions.
 */
typedef enum {
    PACE2_CPI_SUCCESS = 0,              /**< Function returned successfully, returned data is valid. */
    PACE2_CPI_PARAM_MISSING,            /**< Function parameter is missing. */
    PACE2_CPI_NOT_AVAILABLE,            /**< The requested data are not available. */
    PACE2_CPI_FAILURE,                  /*!< Processing failed. */
    PACE2_CPI_CALL_AGAIN,               /*!< Indicates that the function needs to be called again. */
    PACE2_CPI_NO_DATA_CALL_AGAIN        /*!< Indicates that the function needs to be called again and no data is available. */
} PACE2_CPI_return_state;

/* Public Prototypes ************************************************************/

/**
 * @brief Get the HTTP/2 basic event data of the last processed packet.
 *
 * This function returns the HTTP/2 basic event metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *          The @ref PACE2_basic_HTTP2_event must be enabled via event policy.
 *          Otherwise the function returns with @ref PACE2_CPI_NOT_AVAILABLE.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  event_data      Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_http2_event_data(PACE2_module * const pace2_module,
                                                      int const thread_ID,
                                                      PACE2_basic_HTTP2_event ** event_data);

/**
 * @brief Get the QUIC client hello data of the last processed packet.
 *
 * This function returns the  QUIC client hello metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *          The @ref PACE2_basic_QUIC_client_hello_event must be enabled via event policy.
 *          Otherwise the function returns with @ref PACE2_CPI_NOT_AVAILABLE.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  quic_data       Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_quic_client_hello_data(PACE2_module * const pace2_module,
                                                            int const thread_ID,
                                                            PACE2_basic_QUIC_client_hello_event ** quic_data);

/**
 * @brief Get the SSL server hello data of the last processed packet.
 *
 * This function returns the SSL server hello metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *          The @ref PACE2_SSL_server_hello_event must be enabled via event policy.
 *          Otherwise the function returns with @ref PACE2_CPI_NOT_AVAILABLE.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  ssl_data        Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_ssl_server_hello_data(PACE2_module * const pace2_module,
                                                           int const thread_ID,
                                                           PACE2_SSL_server_hello_event ** ssl_data);

/**
 * @brief Get the SSL client hello data of the last processed packet.
 *
 * This function returns the SSL client hello metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *          The @ref PACE2_SSL_client_hello_event must be enabled via event policy.
 *          Otherwise the function returns with @ref PACE2_CPI_NOT_AVAILABLE.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  ssl_data        Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_ssl_client_hello_data(PACE2_module * const pace2_module,
                                                           int const thread_ID,
                                                           PACE2_SSL_client_hello_event ** ssl_data);

/**
 * @brief Get the SSL session ticket data of the last processed packet.
 *
 * This function returns the SSL session ticket metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *          The @ref PACE2_SSL_session_ticket_event must be enabled via event policy.
 *          Otherwise the function returns with @ref PACE2_CPI_NOT_AVAILABLE.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  ssl_data        Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_ssl_session_ticket_data(PACE2_module * const pace2_module,
                                                             int const thread_ID,
                                                             PACE2_SSL_session_ticket_event ** ssl_data);

/**
 * @brief Get the SSL DNS alt names of the last processed packet.
 *
 * This function returns the SSL DNS alt names metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *          The @ref PACE2_SSL_dns_alt_names_event must be enabled via event policy.
 *          Otherwise the function returns with @ref PACE2_CPI_NOT_AVAILABLE.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  ssl_data        Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_ssl_dns_alt_names_data(PACE2_module * const pace2_module,
                                                            int const thread_ID,
                                                            PACE2_SSL_dns_alt_names_event ** ssl_data);

/**
 * @brief Get the SIP data of the last processed packet.
 *
 * This function returns the SIP metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  sip_data        Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_sip_data(PACE2_module * const pace2_module,
                                              int const thread_ID,
                                              PACE2_basic_SIP_event ** sip_data);

/**
 * @brief Get the browser classification data of the last processed packet.
 *
 * This function returns the browser classification metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *
 * @param[in]   pace2_module                  Pointer to the PACE 2 module.
 * @param[in]   thread_ID                     Identifier of a thread.
 * @param[out]  browser_classification_data   Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_browser_classification_data(PACE2_module * const pace2_module,
                                                                 int const thread_ID,
                                                                 PACE2_browser_classification_event ** browser_classification_data);

/**
 * @brief Get the DHCPv4 data of the last processed packet.
 *
 * This function returns the DHCPv4 metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  dhcp_data       Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_dhcp_data(PACE2_module * const pace2_module,
                                               int const thread_ID,
                                               PACE2_basic_DHCP_event ** dhcp_data);

/**
 * @brief Get the DNS data of the last processed packet.
 *
 * This function returns the DNS metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  dns_data       Pointer to a variable, which is used to store the pointer to the data.
 *
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_dns_data(PACE2_module * const pace2_module,
                                              int const thread_ID,
                                              PACE2_basic_DNS_event ** dns_data);

/**
 * @brief Get the HTTP request data of the last processed packet.
 *
 * This function returns the HTTP request metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  http_request_data   Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE           PACE2_OS_event @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_http_request_data(PACE2_module * const pace2_module,
                                                       int const thread_ID,
                                                       PACE2_basic_HTTP_request_event ** http_request_data);

/**
 * @brief Get the HTTP response data of the last processed packet.
 *
 * This function returns the HTTP response metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  http_response_data  Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_http_response_data(PACE2_module * const pace2_module,
                                                        int const thread_ID,
                                                        PACE2_basic_HTTP_response_event ** http_response_data);

/**
 * @brief Get the OS detection data of the last processed packet.
 *
 * This function returns the OS detection metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  os_detection_data   Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_os_detection_data(PACE2_module * const pace2_module,
                                                       int const thread_ID,
                                                       PACE2_OS_event ** os_detection_data);

/**
 * @brief Get the client-server indication packet direction data of the last processed packet.
 *
 * This function returns the packet direction metadata for the last packet given to the classification stage.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  packet_direction_data   Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * The function returns this value in different cases:
 *  @arg    CSI module is disabled via configuration.
 *  @arg    The fastpath of the detection  was used, requested data is not available.
 *  @arg    The client-server indication is not available.
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_packet_direction_data(PACE2_module * const pace2_module,
                                                           int const thread_ID,
                                                           PACE2_csi_packet_dir_event ** packet_direction_data);

/**
 * @brief Get the out-of-order statistics of the last TCP flow process by the classification stage.
 *
 * This function returns the out-of-order and retransmission counter values of the last TCP flow
 * processed by the detection.
 *
 * @note    The data of the structure is valid as long as no further packet processing of stage 1 to 5 is done.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  tcp_ooo_data    Pointer to a variable, which is used to store the pointer to the data.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */
PACE2_CPI_return_state pace2_cpi_get_tcp_out_of_order_data(PACE2_module * const pace2_module,
                                                           int const thread_ID,
                                                           PACE2_TCP_event ** tcp_ooo_data);


/**
 * @brief Reset the given metadata dissector state.
 *
 * This function reset the meta dissector internal state.
 * It can be called anytime however not all dissectors allow a reset.
 *
 * @note this function is usable with internal and external flow tracking.
 * For internal tracking the flow pointer should be NULL in order to indicate to
 * use the internal flow pointer.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[in]   flow            pointer to the flow to reset metadata for, can be NULL for internal tracking
 * @param[in]   type            dissector type for the specific flow data should get reset for
 * @param[out]  status          status of the dissector, the pointer must not be NULL
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */

PACE2_CPI_return_state pace2_cpi_reset_metadata_dissector(PACE2_module * const pace2_module,
                                                          int const thread_ID,
                                                          void * const flow,
                                                          enum dissector_metadata_type const type,
                                                          enum metadata_status *status);

/**
 * @brief Get current metadata dissector result
 *
 * This function returns the current metadata dissector result.
 * The corresponding metadata dissector which needs to be enabled.
 *
 * @note this function is usable with internal and external flow tracking.
 * For internal flow tracking the flow pointer can be NULL.
 * This function can also be used with @ref pace2_pht_foreach_get_next_element
 * with external tracking.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[in]   flow            pointer to the flow to reset metadata for, can be NULL for internal tracking
 * @param[in]   type            dissector type to query event for
 * @param[out]  metadata_event  event to the metadata type that got queried
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */

PACE2_CPI_return_state pace2_cpi_get_dissector_metadata(PACE2_module * const pace2_module,
                                                        int const thread_ID,
                                                        void * const flow,
                                                        enum dissector_metadata_type const type,
                                                        PACE2_dissector_metadata_event *metadata_event);

/**
 * @brief Reset the metadata dissector state for all flows.
 *
 * This function reset the meta dissector internal state.
 * It can be called anytime however not all dissectors allow a reset.
 *
 * @note this function is only usable with internal flow tracking.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[in]   type            dissector type for the specific flow data should get reset for
 * @param[out]  status          status of the dissector, the pointer must not be NULL
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */

PACE2_CPI_return_state pace2_cpi_reset_all_metadata_dissector_int_tracking(PACE2_module * const pace2_module,
                                                                           int const thread_ID,
                                                                           enum dissector_metadata_type const type,
                                                                           enum metadata_status *status);


/**
 * @brief Prepare internal flow table to retrieve dissector metadata for each flow.
 *
 * @note this function is only usable with internal flow tracking.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */

PACE2_CPI_return_state pace2_cpi_init_foreach_int_metadata_dissector(PACE2_module * const pace2_module,
                                                                     int const thread_ID);


typedef struct {
    u8 * key;                         /*!< pointer to the lookup key for the flow */
    u64 flow_id;                      /*!< PACE2 flow ID */
    PACE2_timestamp last_updated_ts;  /*!< last time the flow got updated */
} PACE2_dissector_foreach_flowinfo;

/**
 * @brief Get the next metadata dissector result
 *
 * @note This function is only usable with internal flow tracking.
 * It needs to get called after @ref pace2_cpi_init_foreach_int_metadata_dissector was called.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[in]   type            dissector type to query event for
 * @param[out]  metadata_event  event to the metadata type that got queried
 * @param[out]  flow_info       contains information about the flow see @ref PACE2_dissector_foreach_flowinfo,
 *                              can be NULL
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 *                                      No further flows to iterate.
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 * @retval PACE2_CPI_CALL_AGAIN         @copydoc PACE2_CPI_return_state::PACE2_CPI_CALL_AGAIN
 * @retval PACE2_CPI_NO_DATA_CALL_AGAIN @copydoc PACE2_CPI_return_state::PACE2_CPI_NO_DATA_CALL_AGAIN
 *  If the data is unavailable but further flows can get iterated, the function will return
 *  @ref PACE2_CPI_NO_DATA_CALL_AGAIN, this can be the case e.g. when querying TCP dissectors on UDP flows.
 */
PACE2_CPI_return_state pace2_cpi_foreach_int_get_next_dissector_metadata(PACE2_module * const pace2_module,
                                                                         int const thread_ID,
                                                                         enum dissector_metadata_type const type,
                                                                         PACE2_dissector_metadata_event *metadata_event,
                                                                         PACE2_dissector_foreach_flowinfo *flow_info);

/**
 * @brief Get the start timestamp of a flow
 *
 * This function returns timestamp of the first packet of a flow.
 * A possible use case for this CPI function is the calculation of the flow
 * duration during the flow drop event.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[in]   flow            pointer to the flow
 *
 * @return: Timestamp of the first element on success, else 0.
 */
PACE2_timestamp pace2_cpi_get_connection_start_time(PACE2_module * const pace2_module,
                                                    int const thread_ID,
                                                    const void * const flow);

/**
 * @brief Get an array of key/value pairs from an HTTP request line
 *
 * This function returns a structure containing an array of key/value pairs
 * found in the HTTP request. Additionally an preferred key may be given to be
 * returned in an easily accessible struct. If @p stop_on_prefered_key is set parsing
 * will be stopped once the desired key was found, this may be used to improve
 * performance if other keys are not of interest.
 *
 * @param[in]   pace2_module            Pointer to the PACE 2 module.
 * @param[in]   thread_ID               Identifier of a thread.
 * @param[out]  kv_event                Pointer to event which is used to store parsed data.
 * @param[in]   search_key              Pointer to string of the desired key, may be NULL.
 * @param[in]   search_key_len          Length of the @p search_key string without terminating bytes.
 * @param[in]   stop_on_search_key      If set, stops parsing once the desired @p search_key was found.
 *
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 * @retval PACE2_CPI_FAILURE            @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 */

PACE2_CPI_return_state pace2_cpi_parse_http_request_key_value_pairs(PACE2_module * const pace2_module,
                                                                    int const thread_ID,
                                                                    PACE2_basic_HTTP_parsed_request_event * const kv_event,
                                                                    char  const * const search_key,
                                                                    u32 const search_key_len,
                                                                    u8 const stop_on_search_key);

/**
 * @brief Checks if packet payload is crypted.
 *
 * Be aware that the calculation is not mathematically exact but a rough estimation for better results.
 *
 * @param[in]   pace2_module            Pointer to the PACE 2 module.
 * @param[in]   thread_ID               The ID of the thread currently processing.
 * @param[in]   tolerance               The tolerance is the aloud byte frequency distribution
 *                                      of the packet payload. This parameter is in percentage.
 *                                      The value should be between 1-200. The higher the value,
 *                                      the bigger is the chance that the function returns
 *                                      @ref PACE2_CPI_return_state::PACE2_CPI_SUCCESS.
 *
 * @retval    PACE2_CPI_SUCCESS         @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 *                                      The payload of the packet is crypted.
 * @retval    PACE2_CPI_PARAM_MISSING   @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 *                                      @arg The tolerance value is 0 or > 200.
 *                                      @arg The thread_ID is invalid.
 *                                      @arg Pointer to PACE 2 module is NULL.
 * @retval    PACE2_CPI_FAILURE         @copydoc PACE2_CPI_return_state::PACE2_CPI_FAILURE
 *                                      @arg The payload of the packet is not crypted.

 */
PACE2_CPI_return_state pace2_cpi_packet_payload_is_crypted(PACE2_module * const pace2_module,
                                                           int thread_ID,
                                                           u16 const tolerance);

/** @brief Structure of internal information of pht */
struct PACE2_cpi_internal_pht_stats_info {
    u64 dropped_elements_full;          /*!< Number of elements which dropped because the hashtable was full. */
    u64 dropped_elements_timeout;       /*!< Number of elements which dropped because of timeout limit. */
    u64 inserted_elements;              /*!< Number of new inserted elements. */
    u64 current_element_count;          /*!< Number of currently used elements. */
    u64 number_of_allocated_elements;   /*!< Number of allocated elements. */
    u64 max_number_of_elements_used;    /*!< Largest number of elements used. */
    u64 failed_insert_elements;         /*!< Number of elements dropped elements because the insert failed */
};

/** @brief Enumeration type of internal PHT components */
enum PACE2_cpi_internal_pht_stats_info_component {
    PACE2_INTERNAL_PHT_DECA = 0,
    PACE2_INTERNAL_PHT_SECA,

    PACE2_INTERNAL_PHT_COMPONENT_COUNT,
};

/** @brief Get the number of implemented internal pht components in the currently active lib
 *
 * The function supports the dynamic upgrade feature. If the library is upgraded
 * to a newer version the function implicitly calls it's upgrade variant.
 *
 * @param[in]   pace2_module         Constant pointer to the PACE 2 module
 *                                   Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in]   thread_ID            The ID of the thread currently processing.
 *                                   Only used in case of dynamic upgrade otherwise it can be 0.
 * @retval      u16                  The number of supported pht components.
 */
u16 pace2_cpi_get_number_of_internal_pht_components(PACE2_module * const pace2_module, int const thread_ID);

/**
 * @brief Get internal information of PHT
 *
 * This function returns the internal information of a used pht.
 *
 * @param[in]   pace2_module            Pointer to the PACE 2 module.
 * @param[in]   thread_ID               Identifier of a thread.
 * @param[out]  pht_info                Pointer to pht information which is used to store data.
 * @param[in]   component_id            Identifier which pht should be used.
 *
 * @retval PACE2_CPI_SUCCESS            @copydoc PACE2_CPI_return_state::PACE2_CPI_SUCCESS
 * @retval PACE2_CPI_PARAM_MISSING      @copydoc PACE2_CPI_return_state::PACE2_CPI_PARAM_MISSING
 * @retval PACE2_CPI_NOT_AVAILABLE      @copydoc PACE2_CPI_return_state::PACE2_CPI_NOT_AVAILABLE
 */
PACE2_CPI_return_state pace2_cpi_get_internal_pht_component_usage_information(
    PACE2_module * const pace2_module,
    int thread_ID,
    struct PACE2_cpi_internal_pht_stats_info * pht_info,
    enum PACE2_cpi_internal_pht_stats_info_component component_id);
#ifdef __cplusplus
}
#endif
#endif // PACE2_CPI_CLASSIFICATION_FUNC_H

/*********************************************************************************
 ** EOF
 ********************************************************************************/
