/** @file      pace2_cpi_preparing.h
  * @brief     PACE 2 component programming interface of the packet preparation
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_CPI_PREPARING_H
#define PACE2_CPI_PREPARING_H

/* Includes *********************************************************************/
#include "pace2_basic_def.h"
#include "pace2_s1_preparing.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Prototypes ************************************************************/

/** @brief Query the pointer to the layer 4 of the packet.
 *
 * @param[in]   l3                  pointer to the layer 3 data
 * @param[in]   l3_len              length of the layer 3 data
 * @param[out]  l4_return           filled with the pointer the layer 4 data if
 *                                  return value == PACE2_SUCCESS, unmodified
 *                                  otherwise
 * @param[out]  l4_len_return       filled with the length of the layer 4 data
 *                                  if return value == PACE2_SUCCESS,
 *                                  unmodified otherwise
 * @param[out]  l4_protocol_return  filled with the protocol of the layer 4
 *                                  data if return value == PACE2_SUCCESS,
 *                                  unmodified otherwise
 * @retval      PACE2_SUCCESS       @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_FAILURE       @copydoc PACE2_return_state::PACE2_FAILURE
 *                                  @arg the L4 protocol header check failed
 *                                       (invalid header)
 *                                  @arg this IP packet is part of a fragmented
 *                                       packet (packet fragment)
 */
PACE2_return_state pace2_get_l4(u8 const * l3, u16 l3_len, u8 const ** l4_return,
                                u16 * l4_len_return, u8 * l4_protocol_return);

/**
 * @brief Query the pointer to the layer 7 of the packet.
 *
 * @param[in]   l3                  pointer to the layer 3 data, may be NULL
 * @param[in]   l3_len              length of the layer 3 data, may be 0
 * @param[in]   l4                  pointer to the layer 4 data
 * @param[in]   l4_len              length of the layer 4 data
 * @param[in]   l4_protocol         protocol of the layer 4 data
 * @param[out]  l7_return           filled with the pointer the layer 7 data if
 *                                  return value == PACE2_SUCCESS, undefined
 *                                  otherwise
 * @param[out]  l7_len_return       filled with the length of the layer 7 data
 *                                  if return value == PACE2_SUCCESS, undefined
 *                                  otherwise
 * @retval      PACE2_SUCCESS       @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_FAILURE       @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_get_l7(u8 const * l3, u16 l3_len, u8 const * l4,
                                u16 l4_len, u8 l4_protocol,
                                u8 const ** l7_return, u16 * l7_len_return);

/**
 * @brief Get the unique key of a flow without providing a packet descriptor
 *
 * @param[in]   l3                  pointer to the layer 3 data, may be NULL
 * @param[in]   l3_len              length of the layer 3 data, may be 0
 * @param[in]   l4                  pointer to the layer 4 data
 * @param[in]   l4_len              length of the layer 4 data
 * @param[in]   l4_protocol         protocol of the layer 4 data
 * @param[out]  key_return          filled with the unique key if the return
 *                                  value == PACE2_SUCCESS, undefined otherwise
 * @param[out]  dir_return          filled with a direction flag (0 or 1), can
 *                                  be NULL
 *
 * @retval      PACE2_SUCCESS       @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_FAILURE       @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_build_flow_key_without_packet_descriptor(u8 const * l3, u16 l3_len,
                                                                  u8 const * l4, u16 l4_len,
                                                                  u8 l4_protocol,
                                                                  struct ipoque_unique_flow_ipv4_and_6_struct * key_return,
                                                                  u8 * dir_return);

/** @brief Try to decapsulate one tunnel layer
 *
 * @param[in]   l3                  pointer to the layer 3 data
 * @param[in]   l3_len              length of the layer 3 data
 * @param[out]  l3_return           filled with the pointer the layer 3 data if
 *                                  return value == PACE2_SUCCESS, unmodified
 *                                  otherwise
 * @param[out]  l3_len_return       filled with the length of the layer 3 data
 *                                  if return value == PACE2_SUCCESS,
 *                                  unmodified otherwise
 *
 * @retval      PACE2_SUCCESS       @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_FAILURE       @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_decapsulate(u8 const * l3, u16 l3_len,
                                     u8 const ** l3_return,
                                     u16 * l3_len_return);

/**
 * @brief Performs packet preparation of a single packet frame.
 *
 * This function initializes the packet descriptor of a corresponding packet.
 * The packet framing will be generated by this function and no IP
 * defragmentation will be performed. This function does neither work with
 * internal tracking nor stage2.
 *
 * After this function is called and the corresponding flow is obtained externally,
 * @ref pace2_cpi_flow_update2 must be called.
 *
 * @param[in]   pace2_module                Pointer to the PACE 2 module.
 * @param[in]   thread_ID                   The ID of the thread currently processing.
 * @param[in]   timestamp                   Packet timestamp value. The unit is millisecond by default, can be changed
 *                                          up to nanoseconds by using the config setting
 *                                          @ref PACE2_config_general::io_timestamp_resolution.
 * @param[in]   l3_frame_data_ptr           Pointer to the OSI layer 3 frame data.
 * @param[in]   l3_frame_data_len           Length of OSI layer 3 frame data.
 * @param[in,out] pd                        Pointer to the pointer of a packet descriptor structure,
 *                                          the difference with the packet descriptor used in stage 1
 *                                          is that here the pointer will be filled internally and points
 *                                          to an internal structure afterwards. However, it takes the reference
 *                                          to an actual packet descriptor instantiation in the stage 1.
 * @retval PACE2_S1_SUCCESS                 @copydoc PACE2_s1_return_state::PACE2_S1_SUCCESS
 * @retval PACE2_S1_PARAM_MISSING           @copydoc PACE2_s1_return_state::PACE2_S1_PARAM_MISSING
 * @retval PACE2_S1_HEADER_CHECK_FAILED     @copydoc PACE2_s1_return_state::PACE2_S1_HEADER_CHECK_FAILED
 * @retval PACE2_S1_TRUNCATED_PACKET        @copydoc PACE2_s1_return_state::PACE2_S1_TRUNCATED_PACKET
 * @retval PACE2_S1_FAILURE                 @copydoc PACE2_s1_return_state::PACE2_S1_FAILURE
 *
 */
PACE2_s1_return_state pace2_cpi_get_packet_descriptor(PACE2_module * const pace2_module,
                                                      int thread_ID,
                                                      const PACE2_timestamp timestamp,
                                                      void const * l3_frame_data_ptr,
                                                      u16 const l3_frame_data_len,
                                                      PACE2_packet_descriptor ** pd);

/**
 * @brief Updates flows after obtaining them externally
 *
 * @note The use of this function is required in combination with @ref pace2_cpi_get_packet_descriptor.
 *
 * This function does the following:
 * - Checks if the flow was reused
 *   If this is the case, the function updates the flow ID in the packet descriptor.
 * - Updates the state of TCP flows
 * - Adjusts the flow user data pointer in the packet descriptor
 * - Handles flow_id calculation for new and reused flows
 *
 * @param[in]     pace2_module          Pointer to the PACE 2 module.
 * @param[in]     thread_ID             The ID of the thread currently processing.
 * @param[in,out] pd                    Pointer to the packet descriptor structure.
 *
 * @retval        PACE2_SUCCESS         @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval        PACE2_PARAM_MISSING   @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval        PACE2_FAILURE         @copydoc PACE2_return_state::PACE2_FAILURE
 *
 */
PACE2_return_state pace2_cpi_flow_update2(PACE2_module * const pace2_module,
                                          int thread_ID,
                                          PACE2_packet_descriptor * const pd);

#ifdef __cplusplus
}
#endif
#endif // PACE2_CPI_PREPARING_H

/*********************************************************************************
 ** EOF
 ********************************************************************************/
