/** @file      pace2_event_bitmask.h
  * @brief     PACE 2 event bitmask definitions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_EVENT_BITMASK_H_
#define PACE2_EVENT_BITMASK_H_

#include "pace2_basic_def.h"

#ifdef __cplusplus
extern "C" {
#endif

#define PACE2_EVENTS_BITMASK_LENGTH 2

/**
 * @brief Structure contains the bitmask representation of PACE 2 events.
 */
typedef struct {
    u64 bitmask[PACE2_EVENTS_BITMASK_LENGTH]; /**!< Array to hold the event bitmask. */
} PACE2_event_bitmask;

/**
 * @brief Resets the value of the bitmask to zero.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_event_bitmask "PACE2_event_bitmask".
 */
#define PACE2_EVENT_BITMASK_RESET(a)                                                                                   \
    do {                                                                                                               \
        ((a).bitmask[0]) = 0;                                                                                          \
        ((a).bitmask[1]) = 0;                                                                                          \
    } while (0)

/**
 * @brief Adds a given bitmask b to the bitmask a. Uses a logical or to merge the bitmasks.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_event_bitmask "PACE2_event_bitmask".
 * @param[in]   b   Bitmask of type @ref PACE2_event_bitmask "PACE2_event_bitmask".
 */
#define PACE2_EVENT_BITMASK_ADD(a, b)                                                                                  \
    do {                                                                                                               \
        (a).bitmask[0] |= (b).bitmask[0];                                                                              \
        (a).bitmask[1] |= (b).bitmask[1];                                                                              \
    } while (0)

/**
 * @brief Combines a given bitmask a with the bitmask b into bitmask c. Uses a logical or to merge the bitmasks.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_event_bitmask "PACE2_event_bitmask".
 * @param[in]   b   Bitmask of type @ref PACE2_event_bitmask "PACE2_event_bitmask".
 * @param[in]   c   Bitmask of type @ref PACE2_event_bitmask "PACE2_event_bitmask".
 */
#define PACE2_EVENT_BITMASK_COMBINE(a, b, c)                                                                           \
    do {                                                                                                               \
        (c).bitmask[0] = (a).bitmask[0] | (b).bitmask[0];                                                              \
        (c).bitmask[1] = (a).bitmask[1] | (b).bitmask[1];                                                              \
    } while (0)

/**
 * @brief Deletes a given bitmask b from the bitmask a.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_event_bitmask "PACE2_event_bitmask".
 * @param[in]   b   Bitmask of type @ref PACE2_event_bitmask "PACE2_event_bitmask".
 */
#define PACE2_EVENT_BITMASK_DEL(a, b)                                                                                  \
    do {                                                                                                               \
        (a).bitmask[0] &= ~((b).bitmask[0]);                                                                           \
        (a).bitmask[1] &= ~((b).bitmask[1]);                                                                           \
    } while (0)

/**
 * @brief Checks if both bitmasks are identical.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_event_bitmask "PACE2_event_bitmask".
 * @param[in]   b   Bitmask of type @ref PACE2_event_bitmask "PACE2_event_bitmask".
 */
#define PACE2_EVENT_BITMASK_EQUAL(a,b) (((a).bitmask[0] == (b).bitmask[0]) && ((a).bitmask[1] == (b).bitmask[1]))

/**
 * @brief Sets all the fields for the existing events to 1 and other fields to 0.
 *
 * @param[in]   a   Bitmask of type @ref PACE2_event_bitmask "PACE2_event_bitmask".
 */
#define PACE2_EVENT_BITMASK_SET_EXISTING_EVENTS(a)                                                                     \
    do {                                                                                                               \
        u16 number_of_events = PACE2_NUMBER_OF_EVENTS;                                                                 \
        u8 bitmask_index = 0;                                                                                          \
                                                                                                                       \
        while (number_of_events > 64) {                                                                                \
            (a).bitmask[bitmask_index++] = 0xFFFFFFFFFFFFFFFFUL;                                                       \
            number_of_events -= 64;                                                                                    \
        }                                                                                                              \
                                                                                                                       \
        (a).bitmask[bitmask_index++] = (u64)(0xFFFFFFFFFFFFFFFFUL >> (64 - number_of_events));                         \
                                                                                                                       \
        while (bitmask_index <= (PACE2_EVENTS_BITMASK_LENGTH - 1)) {                                                   \
            (a).bitmask[bitmask_index++] = 0UL;                                                                        \
        }                                                                                                              \
    } while (0)

#ifdef __cplusplus
}
#endif
#endif /* PACE2_EVENT_BITMASK_H_ */
