/** @file      pace2_event_types.h
  * @brief     Global PACE 2 event definitions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_EVENT_TYPES_H
#define PACE2_EVENT_TYPES_H

/* Includes *********************************************************************/
#include "pace2_basic_def.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Event Groups *****************************************************************/

/**
 * @brief Event group definitions for events.
 */
typedef enum {
    PACE2_NO_GROUP = 0,                         /*!< A group for all events, which are not auto-generated and
                                                     not part of the group @ref PACE2_POLLING_PROTOCOL_META_DATA_GROUP */
    PACE2_START_GROUP,                          /*!< Group for subscriber and flow started events. */
    PACE2_DROP_GROUP,                           /*!< Group for subscriber and flow dropped events. */
    PACE2_CLASSIFICATION_RESULT_GROUP,          /*!< Group for classification result events. */
    PACE2_ENGINE_META_DATA_GROUP,               /*!< Group for engine metadata events like OS detection metadata. */
    PACE2_BASIC_PROTOCOL_META_DATA_GROUP,       /*!< Group for basic protocol metadata events. */
    PACE2_ADVANCED_PROTOCOL_META_DATA_GROUP,    /*!< Group for advanced protocol metadata events. */
    PACE2_CLASS_PROTOCOL_META_DATA_GROUP,       /*!< Group for class protocol metadata events. */
    PACE2_POLLING_PROTOCOL_META_DATA_GROUP,     /*!< Group for protocol metadata, which can be get by calling
                                                     the dedicated functions. The following events are assigned to
                                                     this group:
                                                     @arg @ref PACE2_BASIC_DHCP_EVENT
                                                     @arg @ref PACE2_TCP_EVENT
                                                     @arg @ref PACE2_BASIC_RTP_EVENT
                                                     @arg @ref PACE2_BASIC_RTCP_EVENT */
    PACE2_CUSTOM_DEFINED_EVENT_GROUP,           /*!< Group for custom defined events like @ref PACE2_CDD_EVENT .*/

    PACE2_NUMBER_OF_EVENT_GROUPS                /*!< Number of implemented event groups. */
} PACE2_event_groups;

/* Event Types ******************************************************************/

/**
 * @brief Global event type definitions.
 *
 * As a rule of thumb, events whose enum value starts with "PACE2_BASIC_" are only generated
 * during stage 3 processing. Events whose enum value starts with "PACE2_ADVANCED_" and "PACE2_CLASS_"
 * are only generated during stage 4 processing, whereas "PACE2_ADVANCED_" events are low level
 * events directly generated from the payload being decoded. "PACE2_CLASS_" events are
 * high level event generated by aggregating several low level events and by combining the
 * information, those low level events provide.
 * All other events may be thrown by any stage and are mostly status events.
 */
typedef enum { /* Event Types */
    PACE2_NO_EVENT = 0,                         /*!< No real event. */
    PACE2_EVENT_RES_1,
    PACE2_FLOW_STARTED_EVENT,                   /*!< Flow entry start event. */
    PACE2_FLOW_DROPPED_EVENT,                   /*!< Flow entry drop event. */
    PACE2_SUBSCRIBER_STARTED_EVENT,             /*!< Subscriber entry start event. */
    PACE2_SUBSCRIBER_DROPPED_EVENT,             /*!< Subscriber entry drop event. */
    PACE2_BASIC_HTTP_REQUEST_EVENT,             /*!< Basic event for a HTTP request packet. */
    PACE2_BASIC_HTTP_RESPONSE_EVENT,            /*!< Basic event for a HTTP response packet. */
    PACE2_ADVANCED_HTTP_EVENT,                  /*!< Advanced HTTP event. */
    PACE2_CLASS_HTTP_EVENT,                     /*!< HTTP class event. */
    PACE2_BASIC_CITRIX_EVENT,                   /*!< Meta data event of the citrix protocol. */
    PACE2_BASIC_DHCP_EVENT,                     /*!< Basic event for a DHCP packet. */
    PACE2_ADVANCED_DHCP_EVENT,                  /*!< Advanced DHCP event. */
    PACE2_OS_EVENT,                             /*!< Meta data event of the OS detection. */
    PACE2_NAT_EVENT,                            /*!< Meta data event of the NAT detection. */
    PACE2_BASIC_SIP_EVENT,                      /*!< Basic event for a SIP packet. */
    PACE2_ADVANCED_SIP_EVENT,                   /*!< Advanced SIP event. */
    PACE2_CLASS_VOIP_EVENT,                     /*!< VOIP class event. */
    PACE2_BASIC_SSL_CLIENT_HELLO_EVENT,         /*!< Basic event for a SSL client hello packet. */
    PACE2_BASIC_SSL_SERVER_HELLO_EVENT,         /*!< Basic event for a SSL server hello packet. */
    PACE2_BASIC_SIT_EVENT,                      /*!< Meta data event for the SSL session ID tracker status. */
    PACE2_CSI_PACKET_DIR_EVENT,                 /*!< Meta data event for the packet direction, detected by the client-server indication. */
    PACE2_CSI_HOST_TYPE_EVENT,                  /*!< Meta data event for the host type, detected by the client-server indication. */
    PACE2_BASIC_RTP_EVENT,                      /*!< RTP performance report event. The event is not generated automatically,
                                                     it has to be generated manually. For details refer to the integration manual. */
    PACE2_BASIC_RTCP_EVENT,                     /*!< RTCP report event. The event is not generated automatically,
                                                     it has to be generated manually. For details refer to the integration manual. */
    PACE2_TCP_EVENT,                            /*!< TCP status information event. */
    PACE2_CLASSIFICATION_STATUS,                /*!< Classification status event. */
    PACE2_FASTPATH_EVENT,                       /*!< Fastpath status event. */
    PACE2_EVENT_RES_28,
    PACE2_ADVANCED_DIAMETER_EVENT,              /*!< Advanced Diameter event. */
    PACE2_ADVANCED_DNS_EVENT,                   /*!< Advanced DNS event. */
    PACE2_ADVANCED_FTP_EVENT,                   /*!< Advanced FTP event. */
    PACE2_ADVANCED_RADIUS_EVENT,                /*!< Advanced radius event. */
    PACE2_ADVANCED_ICMPV6_EVENT,                /*!< Advanced ICMPv6 event. */
    PACE2_ADVANCED_POP3_EVENT,                  /*!< Advanced POP3 event. */
    PACE2_ADVANCED_IMAP_EVENT,                  /*!< Advanced IMAP event. */
    PACE2_ADVANCED_SMTP_EVENT,                  /*!< Advanced SMTP event. */
    PACE2_ADVANCED_GTP_EVENT,                   /*!< Advanced GTP event. */
    PACE2_ADVANCED_WAP_WSP_EVENT,               /*!< Advanced WAP and WSP event. */
    PACE2_ADVANCED_UNENCRYPTED_JABBER_EVENT,    /*!< Advanced unencrypted Jabber event. */
    PACE2_ADVANCED_PPTP_EVENT,                  /*!< Advanced PPTP event. */
    PACE2_CLASS_MAIL_EVENT,                     /*!< class Mail event. */
    PACE2_CLASSIFICATION_RESULT,                /*!< Classification result event. */
    PACE2_CDC_RESULT,                           /*!< Custom Defined Classification result event. */
    PACE2_LICENSE_EXCEEDED_EVENT,               /*!< License exceeded event. */
    PACE2_ADVANCED_IRC_EVENT,                   /*!< Advanced IRC event. */
    PACE2_FLOW_INFO_EVENT,                      /*!< Flow entry info event. */
    PACE2_FLOW_PROCESS_EVENT,                   /*!< Flow entry process event. Only available if the @ref PACE2_tracking_generic_params::enable_tracking_statistics
                                                     is enabled in the flow tracking part of the PACE2 config. */
    PACE2_SUBSCRIBER_PROCESS_EVENT,             /*!< Subscriber entry process event. Only available if the @ref PACE2_tracking_generic_params::enable_tracking_statistics
                                                     is enabled in the subscriber tracking part of the PACE2 config. */
    PACE2_CDD_EVENT,                            /*!< Custom defined decoder metadata event. */
    PACE2_ADVANCED_H323_EVENT,                  /*!< Advanced H323 event. */
    PACE2_ADVANCED_SOCKS_EVENT,                 /*!< Advanced SOCKS event. */
    PACE2_TCP_STARTED_EVENT,                    /*!< TCP start event. */
    PACE2_TCP_CLOSED_EVENT,                     /*!< TCP finished event. */
    PACE2_ADVANCED_HI5_EVENT,                   /*!< Advanced hi5 event. */
    PACE2_BASIC_SSL_DNS_ALT_NAMES_EVENT,        /*!< Basic event for a SSL DNS Alt Names. */
    PACE2_BASIC_DISSECTOR_METADATA_EVENT,       /*!< Basic dissector metadata event. */
    PACE2_ADVANCED_TELNET_EVENT,                /*!< Advanced Telnet event. */
    PACE2_ADVANCED_MMSE_EVENT,                  /*!< Advanced MMSE event. */
    PACE2_ADVANCED_L2TP_EVENT,                  /*!< Advanced L2TP event. */
    PACE2_ADVANCED_YAHOO_MSGR_EVENT,            /*!< Advanced Yahoo Messenger event. */
    PACE2_ADVANCED_RTP_EVENT,                   /*!< Advanced RTP event. */
    PACE2_ADVANCED_NNTP_EVENT,                  /*!< Advanced NNTP event. */
    PACE2_ADVANCED_FLUTE_EVENT,                 /*!< Advanced FLUTE event. */
    PACE2_ADVANCED_OSCAR_EVENT,                 /*!< Advanced Oscar event. */
    PACE2_BASIC_SSL_SESSION_TICKET_EVENT,       /*!< Basic event for a SSL session ticket packet. */
    PACE2_ADVANCED_TCP_EVENT,                   /*!< Advanced TCP event. */
    PACE2_ADVANCED_ICQ_WEBCHAT_EVENT,           /*!< Advanced ICQ webchat event. */
    PACE2_ADVANCED_MPEG_EVENT,                  /*!< Advanced MPEG2 event. */
    PACE2_ADVANCED_RTSP_EVENT,                  /*!< Advanced RTSP event. */
    PACE2_ADVANCED_BITTORRENT_EVENT,            /*!< Advanced BitTorrent event. */
    PACE2_CLASS_GENERIC_EVENT,                  /*!< Generic class event. */
    PACE2_CLASS_FTP_EVENT,                      /*!< Class FTP event. */
    PACE2_BASIC_HTTP_PARSED_REQUEST_EVENT,      /*!< HTTP parsed request event. */
    PACE2_CLASS_CHAT_EVENT,                     /*!< Class Chat event. */
    PACE2_ADVANCED_EDONKEY_EVENT,               /*!< Advanced eDonkey event. */
    PACE2_ADVANCED_SSL_EVENT,                   /*!< Advanced SSL event. */
    PACE2_CLASS_SSL_EVENT,                      /*!< Class SSL event. */
    PACE2_ADVANCED_OCSP_EVENT,                  /*!< Advanced ocsp event. */
    PACE2_ADVANCED_JAVASCRIPT_EVENT,            /*!< Advanced JavaScript event. */
    PACE2_ADVANCED_SCADA_EVENT,                 /*!< Advanced SCADA event. */
    PACE2_EVENTS_DROPPED_EVENT,                 /*!< Event to report dropped events. */
    PACE2_ADVANCED_Q931_EVENT,                  /*!< Advanced Q.931 event. */
    PACE2_ADVANCED_SMB_EVENT,                   /*!< Advanced SMB event. */
    PACE2_ADVANCED_MODBUS_EVENT,                /*!< Advanced Modbus event. */
    PACE2_BASIC_ZERO_SNI_EVENT,                 /*!< Basic event for a ZERO SNI packet. */
    PACE2_BASIC_QUIC_CLIENT_HELLO_EVENT,        /*!< Basic event for a QUIC client hello packet. */
    PACE2_PLUGIN_EVENT,                         /*!< PLUG-IN event. */
    PACE2_ADVANCED_SSH_EVENT,                   /*!< Advanced SSH event. */
    PACE2_BASIC_DNS_EVENT,                      /*!< basic DNS event. */
    PACE2_ADVANCED_MQTT_EVENT,                  /*!< Advanced MQTT event. */
    PACE2_BASIC_HTTP2_EVENT,                    /*!< Basic event for a HTTP2 HEADERS frame. */
    PACE2_ADVANCED_HTTP2_EVENT,                 /*!< Advanced event for a HTTP2 HEADERS frame. */
    PACE2_BASIC_NETBIOS_EVENT,                  /*!< basic NetBIOS event. */
    PACE2_BASIC_MDNS_EVENT,                     /*!< basic mDNS event. */
    PACE2_PHT_USAGE_NOTIFICATION_EVENT,         /*!< PHT usage notification event. */
    PACE2_ML_CLASSIFICATION_EVENT,              /*!< Machine learning classification event. */
    PACE2_ML_DEREGISTER_EVENT,                  /*!< Machine learning deregister event. */
    PACE2_BROWSER_CLASSIFICATION_EVENT,         /*!< Browser classification event. */

    PACE2_NUMBER_OF_EVENTS                      /*!< Number of implemented event types. */
} PACE2_event_type;

typedef union pace2_event PACE2_event;

typedef PACE2_DEPRECATED("Event is marked as deprecated and should not be used anymore.")
    PACE2_event_type PACE2_deprecated_events;
#define PACE2_DETECTION_RESULT ((PACE2_deprecated_events)PACE2_EVENT_RES_1)

/*
 * Macro to create an enum of verbose invalid messages. This macro should be passed as an argument to the X-Macro list
 * of verbose messages of any decoder.
 *
 * @param   The enum value
 * @param   The enum name
 * @param   The enum message string
 */
#define PACE2_ADVANCED_INVALID_VERBOSE_TO_ENUM(index, name, msg) name = index,

/*
 * Macro to generate a list of invalid messages. This macro should be passed as an argument to the X-Macro list
 * of verbose messages of any decoder.
 *
 * @param   The enum value
 * @param   The enum name
 * @param   The enum message string
 */
#define PACE2_ADVANCED_INVALID_VERBOSE_TO_STRING(index, name, msg) msg,

#ifdef __cplusplus
}
#endif
#endif /* PACE2_EVENT_TYPES_H */
