/** @file      pace2_frame_buffer.h
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */
#ifndef PACE2_FRAME_BUFFER_H
#define PACE2_FRAME_BUFFER_H

#include "pace2_basic_def.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Frame buffer function return codes.
 */
typedef enum {
    PACE2_FRAME_BUFFER_SUCCESS = 0,             /*!< Return value when frame buffer operation is successful. */
    PACE2_FRAME_BUFFER_NONE,                    /*!< Return value when the requested value is not available. */
    PACE2_FRAME_BUFFER_INVALID_ARGUMENT,        /*!< Return value when passed arguments are not valid. */
    PACE2_FRAME_BUFFER_INTERNAL_ERROR           /*!< Return value when an internal error is detected. */
} PACE2_frame_buffer_return_state;

/**
 * @brief Frame buffer context forward declaration. Has to be defined by Frame buffer provider.
 */
typedef struct PACE2_frame_buffer_context_struct PACE2_frame_buffer_context;

/**
 * @brief Frame buffer iterator context forward declaration. Has to be defined by Frame buffer Iterator provider.
 */
typedef struct PACE2_frame_buffer_iterator_context_struct PACE2_frame_buffer_iterator_context;


/********************************
 **** Frame buffer section   ****
 ********************************/

/**
 * @brief Iterator start position.
 */
typedef enum {
    PACE2_FRAME_BUFFER_ITERATOR_BEGIN,  /*!< Frame buffer iterator position at the beginning. */
    PACE2_FRAME_BUFFER_ITERATOR_END     /*!< Frame buffer iterator position at the end. */
} PACE2_frame_buffer_iterator_initial_position;

/**
 * @brief Frame buffer statistics.
 */
typedef struct {
    u32 missing_packets_number;  /*!< Number of missing packets in Frame buffer from flow beginning. */
    u32 missing_packets_size;    /*!< Number bytes in the missing packets in Frame buffer from flow beginning. */
} PACE2_frame_buffer_statistics;

/**
 * @brief Get an iterator over particular frame buffer.
 *
 * @note This function provides an iterator object. Iterator object is valid till the next call to this function.
 *
 * @param[in]  context         pointer to the frame buffer object.
 * @param[in]  thread_ID       the ID of the thread currently processing.
 * @param[out] iterator_object address of the pointer to iterator object to be filled. Can be filled by NULL if the
 *                             flow has no frame list.
 * @param[in]  init_position   initial start position of iterator object, either ITERATOR_BEGIN or ITERATOR_END.
 *
 * @retval  PACE2_FRAME_BUFFER_SUCCESS          @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_SUCCESS
 * @retval  PACE2_FRAME_BUFFER_NONE             @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_NONE
 * @retval  PACE2_FRAME_BUFFER_INVALID_ARGUMENT @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_INVALID_ARGUMENT
 * @retval  PACE2_FRAME_BUFFER_INTERNAL_ERROR   @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_INTERNAL_ERROR
 */
typedef PACE2_frame_buffer_return_state
            (*pace2_frame_buffer_get_iterator_fn)(PACE2_frame_buffer_context * context,
                                                  int thread_ID,
                                                  PACE2_frame_buffer_iterator_context ** iterator_object,
                                                  PACE2_frame_buffer_iterator_initial_position init_position);

/**
 * @brief Get statistics of particular frame buffer.
 *
 * @param[in]   context     pointer to frame buffer object.
 * @param[in]   thread_ID   the ID of the thread currently processing.
 * @param[out]  statistics  pointer to the frame buffer statistics structure to be filled.
 *
 * @retval  PACE2_FRAME_BUFFER_SUCCESS          @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_SUCCESS
 * @retval  PACE2_FRAME_BUFFER_NONE             @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_NONE
 * @retval  PACE2_FRAME_BUFFER_INVALID_ARGUMENT @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_INVALID_ARGUMENT
 * @retval  PACE2_FRAME_BUFFER_INTERNAL_ERROR   @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_INTERNAL_ERROR
 */
typedef PACE2_frame_buffer_return_state
            (*PACE2_frame_buffer_get_statistics_fn)(PACE2_frame_buffer_context const * const context,
                                                    int thread_ID,
                                                    PACE2_frame_buffer_statistics * const statistics);

/**
 * @brief Store private data in the *currently processing* frame of flow.
 *
 * @note The *currently processing* frame is not part of the frame buffer, but the frame that is passed to PACE 2 for
 *       processing. The Frame buffer knows the *currently processing* frame and will store private data in it even
 *       without passing the *currently processing* frame to this function explicitly.
 *
 * @param[in] context       pointer to frame buffer object.
 * @param[in] thread_ID     the ID of the thread currently processing.
 * @param[in] private_data  private data to be stored in currently processing frame.
 *
 * @retval  PACE2_FRAME_BUFFER_SUCCESS          @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_SUCCESS
 * @retval  PACE2_FRAME_BUFFER_NONE             @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_NONE
 * @retval  PACE2_FRAME_BUFFER_INVALID_ARGUMENT @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_INVALID_ARGUMENT
 * @retval  PACE2_FRAME_BUFFER_INTERNAL_ERROR   @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_INTERNAL_ERROR
 */
typedef PACE2_frame_buffer_return_state
            (*PACE2_frame_buffer_store_cur_frame_private_data_fn)(PACE2_frame_buffer_context * const context,
                                                                  int thread_ID,
                                                                  u64 private_data);

/*******************************************
 **** Frame buffer Iterator section *****
 ******************************************/

/**
 * @brief Frame buffer element that contains the packet data.
 */
typedef struct {
    u32 packet_number;  /*!< Sequential number of this packet in a flow. */
    u32 packet_size;    /*!< Packet size (ip header + ip payload). */

    /** Packet timestamp. The unit is defined by the config setting
     * @ref PACE2_config_general::io_timestamp_resolution. */
    u64 timestamp;

    u8 * ip_hdr;        /*!< Pointer to innermost IP header. */
    u64 * private_data; /*!< Pointer to PACE private data (8 bytes) that is kept in this frame. */
} PACE2_frame_buffer_element;

/**
 * @brief Places the iterators current position to the next frame in frame buffer.
 *        If there are no more next elements then PACE2_FRAME_BUFFER_NONE is returned.
 *
 * @param context pointer to frame buffer iterator object.
 *
 * @retval  PACE2_FRAME_BUFFER_SUCCESS          @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_SUCCESS
 * @retval  PACE2_FRAME_BUFFER_NONE             @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_NONE
 * @retval  PACE2_FRAME_BUFFER_INVALID_ARGUMENT @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_INVALID_ARGUMENT
 * @retval  PACE2_FRAME_BUFFER_INTERNAL_ERROR   @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_INTERNAL_ERROR
 */
typedef PACE2_frame_buffer_return_state
            (*PACE2_frame_buffer_iter_pos_next_fn)(PACE2_frame_buffer_iterator_context * context);

/**
 * @brief Places the iterator's current position to the previous frame in frame buffer.
 *        If there are no more previous elements then PACE2_FRAME_BUFFER_NONE is returned.
 *
 * @param context pointer to frame list iterator object.
 *
 * @retval  PACE2_FRAME_BUFFER_SUCCESS          @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_SUCCESS
 * @retval  PACE2_FRAME_BUFFER_NONE             @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_NONE
 * @retval  PACE2_FRAME_BUFFER_INVALID_ARGUMENT @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_INVALID_ARGUMENT
 * @retval  PACE2_FRAME_BUFFER_INTERNAL_ERROR   @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_INTERNAL_ERROR
 */
typedef PACE2_frame_buffer_return_state
            (*PACE2_frame_buffer_iter_pos_prev_fn)(PACE2_frame_buffer_iterator_context * context);

/**
 * @brief Fills the provided packet_object with this current frame buffer element data.
 *        If there are no more next elements then PACE2_FRAME_BUFFER_NONE is returned and packet_object is not filled
 *        (is kept as is).
 *
 * @param  context        pointer to frame list iterator object.
 * @param  packet_object  pointer to the packet structure to be filled.
 *
 * @retval  PACE2_FRAME_BUFFER_SUCCESS          @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_SUCCESS
 * @retval  PACE2_FRAME_BUFFER_NONE             @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_NONE
 * @retval  PACE2_FRAME_BUFFER_INVALID_ARGUMENT @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_INVALID_ARGUMENT
 * @retval  PACE2_FRAME_BUFFER_INTERNAL_ERROR   @copydoc PACE2_frame_buffer_return_state::PACE2_FRAME_BUFFER_INTERNAL_ERROR
 */
typedef PACE2_frame_buffer_return_state (*PACE2_frame_buffer_iter_get_frame_fn)
            (PACE2_frame_buffer_iterator_context const * const context,
             PACE2_frame_buffer_element * packet_object);

/**
 * @brief Frame buffer call backs.
 */
typedef struct {
    /** Frame buffer context. */
    PACE2_frame_buffer_context * pace2_frame_buffer_context_struct;

    /** Callback to get the frame buffer iterator. */
    pace2_frame_buffer_get_iterator_fn pace2_frame_buffer_get_iterator;

    /** Callback to advance frame buffer iterator. */
    PACE2_frame_buffer_iter_pos_next_fn pace2_frame_buffer_get_next_iterator;

    /** Callback to regress frame buffer iterator. */
    PACE2_frame_buffer_iter_pos_prev_fn pace2_frame_buffer_get_prev_iterator;

    /** Callback to get the frame buffer statistics. */
    PACE2_frame_buffer_get_statistics_fn pace2_frame_buffer_get_statistics;

    /** Callback to store private data on a frame. */
    PACE2_frame_buffer_store_cur_frame_private_data_fn pace2_frame_buffer_store_data;

    /** Callback to retrieve the frame for the current frame buffer iterator position. */
    PACE2_frame_buffer_iter_get_frame_fn pace2_frame_buffer_get_frame;

} PACE2_frame_buffer_callbacks;

#ifdef __cplusplus
}
#endif
#endif // PACE2_FRAME_BUFFER_H
