/** @file      pace2_generic_func.h
  * @brief     PACE 2 generic function definitions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_GENERIC_FUNC_H
#define PACE2_GENERIC_FUNC_H

#include "pace2_basic_def.h"
#include "pace2_classification_results.h"
#include "pace2_event_types.h"
#include "pace2_pht_core.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Prototypes ************************************************************/

/**
 * @brief Initialize a single processing thread.
 *
 * A call of this function is necessary only in the case that the processing threads
 * are distributed over different CPUs in a multi-socket system to ensure that memory allocations
 * are done on the CPU of the assigned processing thread. See @ref PACE2_config_general::discrete_thread_init
 * for detailed information.
 *
 * A precedent call of @ref pace2_init is necessary to get an initialized PACE 2 module.
 *
 * If the return value is other than @ref PACE2_return_state::PACE2_SUCCESS all initialization
 * and allocations are reverted.
 *
 * @param[in]   pace2_module            Pointer to the initialized pace 2 module.
 * @param[in]   thread_ID               ID of the thread, which should be initialized.
 * @retval  PACE2_SUCCESS               @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_OUT_OF_MEMORY         @copydoc PACE2_return_state::PACE2_OUT_OF_MEMORY
 * @retval  PACE2_PARAM_MISSING         @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_THREAD_ID_INVALID     @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval  PACE2_THREAD_INIT_DISABLED  @copydoc PACE2_return_state::PACE2_THREAD_INIT_DISABLED
 * @retval  PACE2_FAILURE               @copydoc PACE2_return_state::PACE2_FAILURE
 *                                      The initialization of necessary components failed.
 */
PACE2_return_state pace2_init_thread(PACE2_module * const pace2_module, int thread_ID);

/**
 * @brief Initialize a PACE 2 module.
 *
 * @param[in]   config          Pointer to the module configuration structure.
 * @param[out]  pace2_module    Pointer-pointer to the module structure
 * @retval  PACE2_SUCCESS                                       @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_TRACKING_FLOW_TIMEOUT_TOO_SMALL               @copydoc PACE2_return_state::PACE2_TRACKING_FLOW_TIMEOUT_TOO_SMALL
 * @retval  PACE2_INVALID_CONFIGURATION_POINTER                 @copydoc PACE2_return_state::PACE2_INVALID_CONFIGURATION_POINTER
 * @retval  PACE2_INVALID_MODULE_POINTER                        @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval  PACE2_S2_REORDERING_PACKET_BUFFER_SIZE_TOO_SMALL    @copydoc PACE2_return_state::PACE2_S2_REORDERING_PACKET_BUFFER_SIZE_TOO_SMALL
 * @retval  PACE2_NO_DYNAMIC_UPGRADE                            @copydoc PACE2_return_state::PACE2_NO_DYNAMIC_UPGRADE
 * @retval  PACE2_INVALID_GENERAL_ALLOC_FUNCTION_PTR            @copydoc PACE2_return_state::PACE2_INVALID_GENERAL_ALLOC_FUNCTION_PTR
 * @retval  PACE2_INVALID_GENERAL_FREE_FUNCTION_PTR             @copydoc PACE2_return_state::PACE2_INVALID_GENERAL_FREE_FUNCTION_PTR
 * @retval  PACE2_INVALID_GENERAL_REALLOC_FUNCTION_PTR          @copydoc PACE2_return_state::PACE2_INVALID_GENERAL_REALLOC_FUNCTION_PTR
 * @retval  PACE2_INVALID_DYNAMIC_UPGRADE_DLOPEN_FUNCTION_PTR   @copydoc PACE2_return_state::PACE2_INVALID_DYNAMIC_UPGRADE_DLOPEN_FUNCTION_PTR
 * @retval  PACE2_INVALID_DYNAMIC_UPGRADE_DLSYM_FUNCTION_PTR    @copydoc PACE2_return_state::PACE2_INVALID_DYNAMIC_UPGRADE_DLSYM_FUNCTION_PTR
 * @retval  PACE2_INVALID_DYNAMIC_UPGRADE_DLCLOSE_FUNCTION_PTR  @copydoc PACE2_return_state::PACE2_INVALID_DYNAMIC_UPGRADE_DLCLOSE_FUNCTION_PTR
 * @retval  PACE2_DYNAMIC_UPGRADE_RELEASE_MEMORY_AGE_TOO_SMALL  @copydoc PACE2_return_state::PACE2_DYNAMIC_UPGRADE_RELEASE_MEMORY_AGE_TOO_SMALL
 * @retval  PACE2_NO_FLOW_MEMORY_AGE_SET                        @copydoc PACE2_return_state::PACE2_NO_FLOW_MEMORY_AGE_SET *
 * @retval  PACE2_PARAM_MISSING                                 @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_TIMEOUT_REACHED                               @copydoc PACE2_return_state::PACE2_TIMEOUT_REACHED
 * @retval  PACE2_POINTER_NULL                                  @copydoc PACE2_return_state::PACE2_POINTER_NULL
 * @retval  PACE2_OUT_OF_MEMORY                                 @copydoc PACE2_return_state::PACE2_OUT_OF_MEMORY
 * @retval  PACE2_NO_SUBSCRIBER_MEMORY_AGE_SET                  @copydoc PACE2_return_state::PACE2_NO_SUBSCRIBER_MEMORY_AGE_SET
 * @retval  PACE2_EVENT_DECLINED                                @copydoc PACE2_return_state::PACE2_EVENT_DECLINED
 * @retval  PACE2_EVENT_QUEUE_EMPTY                             @copydoc PACE2_return_state::PACE2_EVENT_QUEUE_EMPTY
 * @retval  PACE2_CDD_INVALID_FUNCTION_CALL                     @copydoc PACE2_return_state::PACE2_CDD_INVALID_FUNCTION_CALL
 * @retval  PACE2_CDD_DECODER_DISABLED                          @copydoc PACE2_return_state::PACE2_CDD_DECODER_DISABLED
 * @retval  PACE2_CDD_CONFIG_ARRAY_NOT_SET                      @copydoc PACE2_return_state::PACE2_CDD_CONFIG_ARRAY_NOT_SET
 * @retval  PACE2_CDD_CALLBACK_NOT_SET                          @copydoc PACE2_return_state::PACE2_CDD_CALLBACK_NOT_SET
 * @retval  PACE2_CDD_EVENT_SIZE_NOT_SET                        @copydoc PACE2_return_state::PACE2_CDD_EVENT_SIZE_NOT_SET
 * @retval  PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS                    @copydoc PACE2_return_state::PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS
 * @retval  PACE2_INVALID_TAG                                   @copydoc PACE2_return_state::PACE2_INVALID_TAG
 * @retval  PACE2_THREAD_INIT_INCOMPLETE                        @copydoc PACE2_return_state::PACE2_THREAD_INIT_INCOMPLETE
 * @retval  PACE2_THREAD_INIT_DISABLED                          @copydoc PACE2_return_state::PACE2_THREAD_INIT_DISABLED
 * @retval  PACE2_THREAD_ID_INVALID                             @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval  PACE2_DYNAMIC_UPGRADE_FAILURE                       @copydoc PACE2_return_state::PACE2_DYNAMIC_UPGRADE_FAILURE
 * @retval  PACE2_STAGE1_FAILURE                                @copydoc PACE2_return_state::PACE2_STAGE1_FAILURE
 * @retval  PACE2_STAGE2_FAILURE                                @copydoc PACE2_return_state::PACE2_STAGE2_FAILURE
 * @retval  PACE2_STAGE4_FAILURE                                @copydoc PACE2_return_state::PACE2_STAGE4_FAILURE
 * @retval  PACE2_STAGE3_ERROR_CSI_DISABLED                     @copydoc PACE2_return_state::PACE2_STAGE3_ERROR_CSI_DISABLED
 * @retval  PACE2_STAGE3_ERROR_OS_DISABLED                      @copydoc PACE2_return_state::PACE2_STAGE3_ERROR_OS_DISABLED
 * @retval  PACE2_STAGE3_ERROR_ASYM_DET_DISABLED                @copydoc PACE2_return_state::PACE2_STAGE3_ERROR_ASYM_DET_DISABLED
 * @retval  PACE2_STAGE3_ERROR_PARAM_MISSING                    @copydoc PACE2_return_state::PACE2_STAGE3_ERROR_PARAM_MISSING
 * @retval  PACE2_STAGE3_ERROR_MEM_FAILURE                      @copydoc PACE2_return_state::PACE2_STAGE3_ERROR_MEM_FAILURE
 * @retval  PACE2_STAGE3_ERROR_CC_REGIST_FAILURE                @copydoc PACE2_return_state::PACE2_STAGE3_ERROR_CC_REGIST_FAILURE
 * @retval  PACE2_STAGE3_ERROR_DISSECTOR_MISSING_DEPENDENCY     @copydoc PACE2_return_state::PACE2_STAGE3_ERROR_DISSECTOR_MISSING_DEPENDENCY
 * @retval  PACE2_STAGE3_ERROR_DISSECTOR_DISABLED               @copydoc PACE2_return_state::PACE2_STAGE3_ERROR_DISSECTOR_DISABLED
 * @retval  PACE2_STAGE3_ERROR_MIDSTREAM_DISABLED               @copydoc PACE2_return_state::PACE2_STAGE3_ERROR_MIDSTREAM_DISABLED
 * @retval  PACE2_STAGE3_ERROR_AGGRESSIVE_DISABLED              @copydoc PACE2_return_state::PACE2_STAGE3_ERROR_AGGRESSIVE_DISABLED
 * @retval  PACE2_STAGE3_ERROR_METADATA_DISSECTOR_FAILURE       @copydoc PACE2_return_state::PACE2_STAGE3_ERROR_METADATA_DISSECTOR_FAILURE
 * @retval  PACE2_STAGE3_ERROR_SET_NAT_DETECTION_MODE_ERROR     @copydoc PACE2_return_state::PACE2_STAGE3_ERROR_SET_NAT_DETECTION_MODE_ERROR
 * @retval  PACE2_FAILURE                                       @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_init(struct PACE2_global_config * const config, PACE2_module ** const pace2_module);

/**
 * @brief Returns a PACE2_return_state as more detailed string representation
 *
 * @param[in] return_state The PACE2_return_state code
 * @returns A string representation of return_state
 */
const char* pace2_return_state_str(PACE2_return_state const return_state);

/**
 * @brief Shut down the PACE 2 module.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module, which should be shut down.
 * @retval  PACE2_SUCCESS       @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_PARAM_MISSING @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_FAILURE       Pace 2 module was not freed completely.
 */
PACE2_return_state pace2_exit_module(PACE2_module * pace2_module);

/**
 * @brief Flush all buffered packets for a final processing.
 *
 * This function flushes all buffered packets of the engine.
 * After calling this function, the stored packets are released and should
 * be processed by stage 3 to 5. To get the released packet a call of
 * @ref pace2_s2_get_next_packet is necessary.
 *
 * This function can be used to do a separate call of flushing packets and
 * flushing the internal tracking tables. This is useful to iterate over all
 * PHT entries after processing all packets. In all other cases, the function
 * @ref pace2_flush_engine can be used to do a combined call of functions
 * @ref pace2_flush_packets and @ref pace2_flush_tables.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module, which should be shut down.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @retval  PACE2_SUCCESS                   @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_PARAM_MISSING             @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_INVALID_MODULE_POINTER    @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval  PACE2_THREAD_ID_INVALID         @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval  PACE2_THREAD_INIT_INCOMPLETE    @copydoc PACE2_return_state::PACE2_THREAD_INIT_INCOMPLETE
 */
PACE2_return_state pace2_flush_packets(PACE2_module * const pace2_module, int const thread_ID);

/**
 * @brief Flush all polling hash tables entries.
 *
 * This function clears all polling hash tables entries of the engine. Since
 * the decoder allocates additional memory per flow/subscriber, this function
 * needs to be called, even when using @ref EXTERNAL tracking.
 * After calling this function, a call of @ref pace2_s5_handle_timeout
 * is necessary to perform the timeout handling.
 *
 * This function can be used to do a separate call of flushing packets and
 * flushing the internal tracking tables. This is useful to iterate over all
 * PHT entries after processing all packets. In all other cases, the function
 * @ref pace2_flush_engine can be used to do a combined call of the functions
 * @ref pace2_flush_packets and @ref pace2_flush_tables.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module, which should be shut down.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @retval  PACE2_SUCCESS                   @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_INVALID_MODULE_POINTER    @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval  PACE2_THREAD_ID_INVALID         @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval  PACE2_THREAD_INIT_INCOMPLETE    @copydoc PACE2_return_state::PACE2_THREAD_INIT_INCOMPLETE
 */
PACE2_return_state pace2_flush_tables(PACE2_module * const pace2_module, int const thread_ID);

/**
 * @brief Flush PACE 2 engine.
 *
 * The function flushes the PACE 2 engine and clears all polling hash tables in cases
 * the @ref INTERNAL tracking is used for flow or subscriber tracking. Additional
 * memory, allocated by the decoder, is cleared as well. Latter is also necessary,
 * when @ref EXTERNAL tracking is used.
 *
 * This function calls @ref pace2_flush_packets and @ref pace2_flush_tables implicitly.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @retval  PACE2_SUCCESS                   @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_INVALID_MODULE_POINTER    @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval  PACE2_THREAD_ID_INVALID         @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval  PACE2_THREAD_INIT_INCOMPLETE    @copydoc PACE2_return_state::PACE2_THREAD_INIT_INCOMPLETE
 */
PACE2_return_state pace2_flush_engine(PACE2_module * const pace2_module, int thread_ID);

/**
 * @brief Free flow from PACE 2 engine.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @param[in]   flow            Pointer to flow data,
 * @retval  PACE2_SUCCESS                   @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_PARAM_MISSING             @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_INVALID_MODULE_POINTER    @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval  PACE2_THREAD_ID_INVALID         @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval  PACE2_THREAD_INIT_INCOMPLETE    @copydoc PACE2_return_state::PACE2_THREAD_INIT_INCOMPLETE
 */
PACE2_return_state pace2_release_flow(PACE2_module * const pace2_module, int thread_ID, void *flow);

/**
 * @brief Add a stream descriptor to the last released flow.
 *
 * For a released flow (done by @ref pace2_release_flow) this function associates a PACE 2 stream descriptor
 * with the last released flow. All stream data which was not processed yet is handed over to the engine
 * for a final processing. The function has to be called after the external TCP reassembly processed all data and
 * before @ref pace2_s5_handle_timeout was called because the data consumption
 * will take place in @ref pace2_s5_handle_timeout.
 * It is necessary to call @ref pace2_s5_handle_timeout after this function.
 * A repeated function call of @ref pace2_add_stream_descriptor_for_released_flow will otherwise result in an error.
 *
 * @param[in]   pace2_module        Pointer to the PACE 2 module.
 * @param[in]   thread_ID           The ID of the thread currently processing.
 * @param[in]   stream_descriptor   Pointer to PACE 2 stream descriptor.
 *
 * @retval  PACE2_SUCCESS                   @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_PARAM_MISSING             @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_FORBIDDEN                 @copydoc PACE2_return_state::PACE2_FORBIDDEN
 * @retval  PACE2_THREAD_ID_INVALID         @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval  PACE2_THREAD_INIT_INCOMPLETE    @copydoc PACE2_return_state::PACE2_THREAD_INIT_INCOMPLETE
 * @retval  PACE2_FEATURE_NOT_AVAILABLE     @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state
    pace2_add_stream_descriptor_for_released_flow(PACE2_module * const pace2_module,
                                                  int thread_ID,
                                                  PACE2_stream_descriptor const * const stream_descriptor);

/**
 * @brief Get next pending event from the event queue.
 *
 * The function returns the next pending event from the event queue.
 * The received event is not longer part of the event queue after the function returned.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @returns     Pointer to the global event structure on success, else NULL.
 */
PACE2_event *pace2_get_next_event(PACE2_module * const pace2_module, int thread_ID);

/**
 * @brief Get total number of pending event still available in the queue.
 *
 * The function returns the number of pending events in the queue.
 * It returns 0 if there are no more events in the queue.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @returns     Number of available events in the event queue.
 */
u32 pace2_get_number_of_pending_events(PACE2_module * const pace2_module, int thread_ID);

/**
 * @brief Get n number of pending events in a dynamic allocated memory block.
 *
 * The function returns an array of PACE2_event pointers containing pointers to pending events.
 * The num_events parameter specifies the number of events requested while received_events represents the
 * actual number of events present in the array. The actual received events (*received_events) can be less
 * than the number of requested events (num_events). The returned list of events must be freed using
 * @ref pace2_free_event_list
 * This function returns null in case of failure.
 * For performance reasons use the alternative function @ref pace2_get_next_n_events_static
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @param[in]   num_events      The number of pending events which should be returned.
 * @param[out]  received_events The actual number of events which are present in the list.
 * @returns     Pointer to array of PACE2_event pointers if successful, else NULL.
 */
PACE2_event ** pace2_get_next_n_events(PACE2_module * const pace2_module,
                                       int thread_ID,
                                       u32 num_events,
                                       u32 * received_events);

/**
 * @brief Get n number of pending events in a static provided memory block.
 *
 * The function fills a caller provided static array (event_array) of PACE2_event pointers containing
 * pointers to pending events. The num_events parameter specifies the number of events requested while
 * received_events represents the actual number of events present in the array. The actual received
 * events (*received_events) can be less than the number of requested events (num_events).
 * This function returns PACE2_SUCCESS in case of success. If the thread is not initialized or invalid,
 * PACE2_THREAD_ID_INVALID will be returned. If event_array or pace2_module is NULL, then
 * PACE2_POINTER_NULL will be returned.
 * The amount of num_events have to be the same like available space for events in the event_array.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @param[in]   num_events      The number of pending events which should be returned.
 * @param[out]  received_events The actual number of events which are present in the list.
 * @param[out]  event_array     The array, which is filled with the requested events.
 * @returns     PACE2_return_state with the signal if the request for the events was successful.
 */
PACE2_return_state pace2_get_next_n_events_static(PACE2_module * const pace2_module,
                                                  int const thread_ID,
                                                  const u32 num_events,
                                                  u32 * const received_events,
                                                  PACE2_event ** const event_array);

/**
 * @brief Free the list of events after use.
 *
 * The function frees the memory allocated for the list of pointers to the @ref pace2_event.
 * This function does not return anything.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @param[in]   event_list      Pointer to the list of pointers to PACE2_event structures.
 */
void pace2_free_event_list(PACE2_module * const pace2_module, int thread_ID, PACE2_event ** event_list);

/**
 * @brief Get the amount of memory in bytes required per flow.
 *
 * Returns the amount of memory in bytes required per flow including the size
 * of user defined data. The returned value is the static flow memory size.
 * Depending on the configured features, a dynamic memory part can be allocated
 * during runtime per flow. See @ref pace2_get_dynamic_flow_memory_size.
 *
 * @note Has to be called after initialization of the given PACE 2 module.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @retval      0               Pointer to the PACE 2 module is NULL or thread initialization has not been done.
 * @retval      >0              Amount of memory per flow.
 */
u32 pace2_get_flow_memory_size(PACE2_module const * const pace2_module,
                               int const thread_ID);

/**
 * @brief Get the dynamically allocated amount of memory in bytes
 * required per flow.
 *
 * Returns the dynamically allocated amount of memory in bytes required per
 * flow. The size differs depending on the configuration of PACE2.
 * The following features allocate memory during runtime:
 * @arg stage 2: packet reordering
 * @arg stage 4: packet decoding
 *
 * @note Has to be called after initialization of the given PACE 2 module.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @retval      0               Pointer to the PACE 2 module is NULL or thread initialization has not been done.
 * @retval      >0              Amount of memory.
 */
u32 pace2_get_dynamic_flow_memory_size(PACE2_module const * const pace2_module,
                                       int const thread_ID);


/**
 * @brief Get the amount of memory in bytes required per subscriber.
 *
 * Returns the amount of memory in bytes required per subscriber including the size of user data.
 * Has to be called after initialization of the given PACE 2 module.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @retval      0               Pointer to the PACE 2 module is NULL or thread initialization has not been done.
 * @retval      >0              Amount of memory per subscriber.
 */
u32 pace2_get_subscriber_memory_size(PACE2_module const * const pace2_module, int const thread_ID);

/**
 * @brief Retrieve the subscriber ID for the provided subscriber pointer.
 *
 * In case of external tracking the @ref PACE2_FLOW_INFO_EVENT is also available.
 * To allow a correlation of the subscriber ID's this function can be used to retrieve the subscriber ID of a
 * provided subscriber pointer. When using external tracking, the subscriber ID is set by PACE 2 in the first call
 * to pace2_s3_process_packet when the ID is not yet set.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @param[in]   subscriber_ptr  Pointer to the subscriber, either from the packet descriptor or
 *                              obtainable during subscriber removal.
 * @returns                     ID of the subscriber
 */
u64 pace2_get_subscriber_id(PACE2_module const * const pace2_module,
                            int const thread_ID,
                            void const * const subscriber_ptr);

/**
 * @brief Get the unique key of a flow.
 *
 * @param[in]   packet_descriptor Pointer to the packet descriptor of a flow packet.
 * @param[out]  key               Pointer to the created unique flow key.
 * @param[out]  dir               Pointer to the direction flag. Can be NULL;
 * @param[in]   start_stack       Start frame of the stack to get the layer 3 or 4 frame.
 *                                Has to be < 0, if all frames should be checked.
 * @retval  PACE2_SUCCESS         @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_PARAM_MISSING   @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_FAILURE         @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_build_flow_key(PACE2_packet_descriptor const * const packet_descriptor,
                                        ipoque_unique_flow_ipv4_and_6_struct_t * const key,
                                        u8 * const dir,
                                        int const start_stack);

/**
 * @brief Get the unique key of the subscriber of a packet.
 *
 * The function builds a unique subscriber key based on a given packet descriptor.
 *
 * @param[in]   packet_descriptor Pointer to the packet descriptor.
 * @param[out]  key               Pointer to the created unique subscriber key.
 * @param[in]   start_stack       Start frame of the stack to get the layer 3 or 4 frame.
 *                                Has to be 0, if all frames should be checked.
 * @param[in]   use_dst           Flag, which indicates that the key should be build on the basis
 *                                of the destination subscriber of the packet.
 * @retval  PACE2_SUCCESS         @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_PARAM_MISSING   @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_FAILURE         @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_build_subscriber_key(PACE2_packet_descriptor const * const packet_descriptor,
                                              PACE2_subscriber_key * const key,
                                              int const start_stack,
                                              PACE2_boolean const use_dst);

/**
 * @brief Enumeration for the possible return values of a foreach callback.
 */
typedef enum pace2_foreach_callback_retval_enum {
    PACE2_FOREACH_NEXT,             /*!< Signals, that the foreach callback wants the next entry, if one more is available. */
    PACE2_FOREACH_BREAK,            /*!< Signals, that the foreach loop can be stopped. */
    PACE2_FOREACH_DELETE_AND_NEXT,  /*!< Signals, that the foreach callback wants to delete the given entry and the next entry is needed, if one more is available. */
} PACE2_foreach_callback_return_state;

/**
 * @brief Type definition for a callback which can be called from the pace2_foreach_flow function.
 *
 * @param[in] pace              The global PACE2 module
 * @param[in] thread_ID         The ID of the thread for which this function is called.
 * @param[in] key               The key data of this flow.
 * @param[in] flow_data         The stored user data for this flow.
 * @param[in] last_timestamp    The timestamp at which the entry has been accessed the last time.
 * @param[in] userptr           The user pointer which was given to pace2_foreach_flow.
 *
 * @retval  PACE2_FOREACH_NEXT              @copydoc pace2_foreach_callback_retval_enum::PACE2_FOREACH_NEXT
 * @retval  PACE2_FOREACH_BREAK             @copydoc pace2_foreach_callback_retval_enum::PACE2_FOREACH_BREAK
 * @retval  PACE2_FOREACH_DELETE_AND_NEXT   @copydoc pace2_foreach_callback_retval_enum::PACE2_FOREACH_DELETE_AND_NEXT
 */
typedef
enum pace2_foreach_callback_retval_enum (*pace2_foreach_flow_callback)(
        PACE2_module *pace,
        const int thread_ID,
        const ipoque_unique_flow_ipv4_and_6_struct_t * const key,
        void *flow_data,
        const PACE2_timestamp last_timestamp,
        void *userptr);

/**
 * @brief Run the given callback for every internal tracked flow.
 *
 * The function calls "callback" for every tracked flow in the internal hash tables. This
 * function is only available if internal flow tracking is enabled.
 *
 * @param[in] pace              The global PACE2 module
 * @param[in] thread_ID         The ID of the thread for which this function is called.
 * @param[in] callback          The callback function which should be called for every entry (tracked flow).
 * @param[in] userptr           A user pointer which is given to the callback function.
 *
 * @retval  PACE2_SUCCESS                   @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_PARAM_MISSING             @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_FAILURE                   @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval  PACE2_THREAD_ID_INVALID         @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval  PACE2_THREAD_INIT_INCOMPLETE    @copydoc PACE2_return_state::PACE2_THREAD_INIT_INCOMPLETE
 */
PACE2_return_state pace2_foreach_flow(PACE2_module *pace,
                                      const int thread_ID,
                                      pace2_foreach_flow_callback callback,
                                      void *userptr);

/**
 * @brief Type definition for a callback which can be called from the pace2_foreach_subscriber function.
 *
 * @param[in] pace              The global PACE2 module
 * @param[in] thread_ID         The ID of the thread for which this function is called.
 * @param[in] key               The key data of this flow.
 * @param[in] subscriber_data   The stored user data for this subscriber.
 * @param[in] last_timestamp    The timestamp at which the entry has been accessed the last time.
 * @param[in] userptr           The user pointer which was given to pace2_foreach_subscriber.
 *
 * @retval  PACE2_FOREACH_NEXT              @copydoc pace2_foreach_callback_retval_enum::PACE2_FOREACH_NEXT
 * @retval  PACE2_FOREACH_BREAK             @copydoc pace2_foreach_callback_retval_enum::PACE2_FOREACH_BREAK
 * @retval  PACE2_FOREACH_DELETE_AND_NEXT   @copydoc pace2_foreach_callback_retval_enum::PACE2_FOREACH_DELETE_AND_NEXT
 */
typedef
enum pace2_foreach_callback_retval_enum (*pace2_foreach_subscriber_callback)(
        PACE2_module *pace,
        const int thread_ID,
        const PACE2_subscriber_key * const key,
        void *subscriber_data,
        const PACE2_timestamp last_timestamp,
        void *userptr);

/**
 * @brief Run the given callback for every internal tracked subscriber.
 *
 * The function calls "callback" for every tracked subscriber in the internal hash tables. This
 * function is only available, if internal subscriber tracking is enabled.
 *
 * @param[in] pace              The global PACE2 module
 * @param[in] thread_ID         The ID of the thread for which this function is called.
 * @param[in] callback          The callback function which should be called for every entry (tracked subscriber).
 * @param[in] userptr           A user pointer which is given to the callback function.
 *
 * @retval  PACE2_SUCCESS                   @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_PARAM_MISSING             @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_FAILURE                   @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval  PACE2_THREAD_ID_INVALID         @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval  PACE2_THREAD_INIT_INCOMPLETE    @copydoc PACE2_return_state::PACE2_THREAD_INIT_INCOMPLETE
 */
PACE2_return_state pace2_foreach_subscriber(PACE2_module *pace,
                                            const int thread_ID,
                                            pace2_foreach_subscriber_callback callback,
                                            void *userptr);

/**
 * @brief Drop a dedicated subscriber, given by key.
 *
 * The function drops the given subscriber, if it exists in the subscriber tracking. This will lead
 * to an additional subscriber drop event (@ref PACE2_SUBSCRIBER_DROPPED_EVENT) and can result in
 * additional events, based on cleanup behavior.
 *
 * Remark: Still existing flows won't be dropped.
 *
 * Can only be used for internal subscriber tracking (see @ref PACE2_tracking_type::INTERNAL).
 *
 * @param[in] pace2_module                  The global PACE2 module
 * @param[in] thread_ID                     The ID of the thread for which this function is called.
 * @param[in] key                           Pointer to the subscriber key, that should be dropped.
 *
 * @returns                                 The result status of this function call.
 * @retval  PACE2_SUCCESS                   @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_INVALID_MODULE_POINTER    @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval  PACE2_PARAM_MISSING             @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_THREAD_ID_INVALID         @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval  PACE2_FORBIDDEN                 Usage of this function is only possible with internal subscriber
 *                                          tracking. @copydoc PACE2_return_state::PACE2_FORBIDDEN
 * @retval  PACE2_INVALID                   The subscriber could not be found with the given
 *                                          key. @copydoc PACE2_return_state::PACE2_INVALID
 * @retval  PACE2_FAILURE                   The deletion of the subscriber
 *                                          failed. @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_drop_subscriber(PACE2_module * const pace2_module,
                                         const int thread_ID,
                                         PACE2_subscriber_key const * const key);

/**
 * @brief Free subscriber from engine and throw @ref PACE2_SUBSCRIBER_DROPPED_EVENT.
 *
 * @details This function works only for external subscriber tracking where the memory
 *          for the subscriber is managed by the customer. This function should only
 *          be called after stage 5 for the current packet and before stage 1 for the
 *          next packet. Additionally it has to be made sure, that all flows related to
 *          the subscriber, have been release with a call to 'pace2_release_flow',
 *          if external flow tracking is used. This behavior can easily be guaranteed
 *          by assigning the flows a lower timeout value than the subscribers.
 *          (@see pace2_integration_example_ext_tracking.c)
 *          After this call the memory for the subscriber is invalid but not set to zero.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @param[in]   subscriber      Pointer to subscriber data
 * @retval  PACE2_SUCCESS                   @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_PARAM_MISSING             @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_INVALID_MODULE_POINTER    @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval  PACE2_THREAD_ID_INVALID         @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval  PACE2_THREAD_INIT_INCOMPLETE    @copydoc PACE2_return_state::PACE2_THREAD_INIT_INCOMPLETE
 */
PACE2_return_state pace2_release_subscriber(PACE2_module * const pace2_module, int thread_ID, void *subscriber);

/**
 * @brief Generate a classifier token.
 *
 * This function returns the address of a @ref PACE2_classifier_token to be later used by stage 4 processing
 * with a different PACE 2 instance.
 *
 * Please note that the content of the returned @ref PACE2_classifier_token is only valid until the next call to this
 * function with the same pace2_module and thread_ID given.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @param[out]  token           Pointer variable to return the token address.
 *
 * @retval  PACE2_SUCCESS                   @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_PARAM_MISSING             @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_INVALID_MODULE_POINTER    @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval  PACE2_THREAD_ID_INVALID         @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval  PACE2_THREAD_INIT_INCOMPLETE    @copydoc PACE2_return_state::PACE2_THREAD_INIT_INCOMPLETE
 */
PACE2_return_state pace2_generate_classifier_token(PACE2_module * const pace2_module,
                                                   int thread_ID,
                                                   PACE2_classifier_token const **token);

/**
 * @brief Prepares a classifier token from a protocol stack and an application.
 *
 * @param[in]   pace2_module        Pointer to the PACE 2 module.
 * @param[in]   thread_ID           The ID of the thread currently processing.
 * @param[in]   protocol_stack      Pointer to the protocol stack of the current stream.
 * @param[in]   application         The current applications.
 * @param[out]  token               Pointer to the address of the classifier token to prepare.
 *
 * @retval  PACE2_SUCCESS                   @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_FAILURE                   @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval  PACE2_PARAM_MISSING             @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 */
PACE2_return_state pace2_prepare_classifier_token(PACE2_module * const pace2_module,
                                                  int const thread_ID,
                                                  PACE2_protocol_stack const * const protocol_stack,
                                                  PACE2_applications const application,
                                                  PACE2_classifier_token const **token);

/**
 * @brief Returns the license information for the given thread.
 *
 * This function returns the pointer to a @ref ipoque_pace_license_information_t to get the current license status
 *
 * Please note that the content of the returned @ref ipoque_pace_license_information_t is only valid until the next call of this
 * function with the same pace2_module and thread_ID given.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   license_info    Pointer to a license information structure
 * @param[in]   thread_ID       The ID of the thread currently processing.
 * @retval  PACE2_SUCCESS                   @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_PARAM_MISSING             @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_INVALID_MODULE_POINTER    @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval  PACE2_FAILURE                   @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval  PACE2_THREAD_ID_INVALID         @copydoc PACE2_return_state::PACE2_THREAD_ID_INVALID
 * @retval  PACE2_THREAD_INIT_INCOMPLETE    @copydoc PACE2_return_state::PACE2_THREAD_INIT_INCOMPLETE
 */
PACE2_return_state pace2_get_license_information( PACE2_module * const pace2_module,
                                                  const ipoque_pace_license_information_t **license_info,
                                                  int thread_ID );

/**
 * @brief Set the timeout queue for the last inserted flow operation.
 *
 * This function is only usable with internal tracking
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   queue           ID of the timeout queue to be active for the current flow.
 * @param[in]   thread_id       The ID of the thread currently processing.
 * @retval    PACE2_PHT_SUCCESS         @copydoc PACE2_pht_return_state::PACE2_PHT_SUCCESS
 * @retval    PACE2_PHT_PARAM_MISSING   @copydoc PACE2_pht_return_state::PACE2_PHT_PARAM_MISSING
 * @retval    PACE2_PHT_FAILURE         @copydoc PACE2_pht_return_state::PACE2_PHT_FAILURE
 *                                      @arg Pointer to the PACE 2 module is NULL.
 *                                      @arg External tracking is used.
 *                                      @arg Timeout queue is not configured.
 */
PACE2_pht_return_state pace2_set_internal_flow_timeout_queue( PACE2_module * const pace2_module,
                                                              u16 queue, int thread_id );

/**
 * @brief Get flow ID of the current packet descriptor.
 *
 * In case this function is used after pace2_s2_process_packet and stage 2 is disabled,
 * the flow ID will be 0 for non-IP, UDP and TCP packets.
 * In case this function is used after pace2_s2_process_packet and stage 2 is enabled,
 * the flow ID will be 0 for non-IP and UDP packets.
 * In case this function is used after pace2_s1_process_packet,
 * the flow ID will be 0 for all packets.
 *
 * @param[in]   pace2_module          Pointer to the PACE 2 module.
 * @param[in]   thread_ID             The ID of the thread currently processing.
 * @param[in]   packet_descriptor     Pointer to the packet descriptor.
 * @param[out]  flow_id               Pointer to the unique flow ID.
 * @retval      PACE2_SUCCESS         @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_POINTER_NULL    @copydoc PACE2_return_state::PACE2_POINTER_NULL
 */
PACE2_return_state pace2_get_flow_id(PACE2_module * const pace2_module, int thread_ID,
                                     PACE2_packet_descriptor * packet_descriptor,
                                     u64 * flow_id);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_GENERIC_FUNC_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
