/** @file      pace2_ipc.h
  * @brief     PACE 2 IPC include file.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_IPC_H
#define PACE2_IPC_H

#include "pace2_basic_def.h"

#ifdef __cplusplus
extern "C" {
#endif

/** IPC Application ***************************************************************************************************/
/* This part covers all the types and functionality necessary for using IPC as an application. ************************/
/**********************************************************************************************************************/

/** enum to identify the data stored in @ref PACE2_ipc_meta_data */
typedef enum pace2_ipc_meta_data_type
{
    PACE2_IPC_META_DATA_TYPE_NONE = 0,

    PACE2_IPC_META_DATA_TYPE_IPV4,
    PACE2_IPC_META_DATA_TYPE_IPV4_PORT,

    PACE2_IPC_META_DATA_TYPE_IPV6,
    PACE2_IPC_META_DATA_TYPE_IPV6_PORT,
} PACE2_ipc_meta_data_type;

/** struct to pack the IPv4 and it's port together */
typedef struct pace2_ipc_meta_data_ipv4_port
{
    struct in_addr              ip;             /*!< IPv4 address */
    u16                         port;           /*!< port for @ref ip */
} PACE2_ipc_meta_data_ipv4_port;

/** struct to pack the IPv6 and it's port together */
typedef struct pace2_ipc_meta_data_ipv6_port
{
    struct in6_addr             ip;             /*!< IPv6 address */
    u16                         port;           /*!< port for @ref ip */
} PACE2_ipc_meta_data_ipv6_port;

/** struct to store IPC filter meta data in */
typedef struct pace2_ipc_meta_data
{
    PACE2_ipc_meta_data_type            type;           /*!< Type to specify the data stored in this union */
    union {
        struct in_addr                  ipv4;           /*!< IPv4 (type PACE2_IPC_META_DATA_TYPE_IPV4) */
        PACE2_ipc_meta_data_ipv4_port   ipv4_port;      /*!< IPv4 and it's matching port (type PACE2_IPC_META_DATA_TYPE_IPV4_PORT) */
        struct in6_addr                 ipv6;           /*!< IPv6 (type PACE2_IPC_META_DATA_TYPE_IPV6) */
        PACE2_ipc_meta_data_ipv6_port   ipv6_port;      /*!< IPv6 and it's matching port (type PACE2_IPC_META_DATA_TYPE_IPV6_PORT) */
    } data;
} PACE2_ipc_meta_data;

/** Callback to handle IPC blob (from IPC to user application) */
typedef void (*pace2_ipc_send_blob_callback_fn)(PACE2_module * const pace2_module,
                                                int thread_id,
                                                PACE2_ipc_meta_data const * const meta_data,
                                                void const * data,
                                                u32 len,
                                                void * user_ptr);

/**
 * @brief Enqueues a received IPC blob to be processed by the registered IPC clients.
 *
 * This function is part of the IPC application interface.
 *
 * @param[in] pace2_module The PACE2 module.
 * @param[in] blob         Pointer to the IPC message data.
 * @param[in] blob_size    Number of bytes stored in blob.
 * @retval                 PACE2_SUCCESS on success
 * @retval                 PACE2_FAILURE if PACE2 failed to process the message
 * @retval                 PACE2_OUT_OF_MEMORY if the message buffer is out of memory
 */
PACE2_return_state pace2_ipc_receive_blob(PACE2_module * const pace2_module,
                                          void const * const blob,
                                          u32 const blob_size);

/**
 * @brief Enqueues a received IPC blob to be processed by the registered IPC clients.
 *
 * This function is part of the IPC application interface.
 *
 * @param[in] pace2_module The PACE2 module.
 * @param[in] timestamp    Message receipt timestamp. If @p timestamp is older than the time internally tracked by
 *                         PACE2, the more recent time will be used.
 *                         The unit is defined by the config setting
 *                         @ref PACE2_config_general::io_timestamp_resolution.
 * @param[in] blob         Pointer to the IPC message data.
 * @param[in] blob_size    Number of bytes stored in blob.
 * @retval                 PACE2_SUCCESS on success
 * @retval                 PACE2_FAILURE if PACE2 failed to process the message
 * @retval                 PACE2_OUT_OF_MEMORY if the message buffer is out of memory
 *
 * @attention The @p timestamp argument has to be in sync with the upcoming network packet and must not be newer than
 * that. It is crucial to use the same time source for IPC as for the call to @ref pace2_s1_process_packet.
 */
PACE2_return_state pace2_ipc_receive_blob_ts(PACE2_module * const pace2_module,
                                             PACE2_timestamp const timestamp,
                                             void const * const blob,
                                             u32 const blob_size);



/** IPC Client ********************************************************************************************************/
/* This part covers all the types and functionality necessary for using IPC as a CDD/PLUG/... (IPC Client). ***********/
/**********************************************************************************************************************/

/**
 * @brief Unique ID to identify the different IPC clients (this needs to be extended if a new IPC client is implemented
 *        inside PACE2)
 */
typedef enum pace2_ipc_client_id
{
    PACE2_IPC_CLIENT_ID_UNKNOWN = 0,            /*!< Client ID is not known */

    PACE2_IPC_CLIENT_ID_USER_DEFINED = 10000,   /*!< Start client ID for user defined IPC clients (CDC, CDD, ...) */

} PACE2_ipc_client_id;

/**
 * @brief Callback to handle IPC data (from IPC to IPC client).
 *
 * @param[in] pace2_module  PACE2 detection module.
 * @param[in] thread_id     ID of the thread that executed the callback.
 * @param[in] timestamp     Message receipt timestamp. The unit is defined by the config setting
 *                          @ref PACE2_config_general::io_timestamp_resolution.
 * @param[in] data          Received data.
 * @param[in] len           Number of bytes stored in data.
 * @param[in] user_ptr      User pointer set with @ref pace2_ipc_register_callback_ts.
 */
typedef void (*pace2_ipc_data_ts_callback_fn)(PACE2_module * const pace2_module,
                                              int thread_id,
                                              PACE2_timestamp const timestamp,
                                              void const * data,
                                              u32 len,
                                              void * user_ptr);

/**
 * @brief Register a new IPC client callback in the IPC system.
 *
 * A client can for example be a PLUG, CDD or any other internal classifications or decoder.
 * This function is part of the IPC client interface.
 *
 * @param[in] pace2_module          The PACE2 module.
 * @param[in] own_client_id         Own ID of the IPC client to register callback for.
 * @param[in] requested_client_id   ID of the IPC client to receive callback data from.
 * @param[in] data_callback         Callback to call if new data was received.
 * @param[in] user_ptr              User pointer to pass to callback.
 *
 * @retval                          PACE2_SUCCESS on success
 * @retval                          PACE2_POINTER_NULL on PACE2 module is NULL
 * @retval                          PACE2_PARAM_MISSING on invalid or missing parameter
 * @retval                          PACE2_OUT_OF_MEMORY on out of memory
 */
PACE2_return_state pace2_ipc_register_callback_ts(PACE2_module * const pace2_module,
                                                  u32 own_client_id,
                                                  u32 requested_client_id,
                                                  pace2_ipc_data_ts_callback_fn data_callback,
                                                  void * user_ptr);

/**
 * @brief Unregister a IPC client callback in the IPC system.
 *
 * A client can for example be a PLUG, CDD or any other internal classifications or decoder.
 * This function is part of the IPC client interface.
 *
 * @param[in] pace2_module          The PACE2 module.
 * @param[in] own_client_id         ID of the client for which a callback has to be deleted.
 * @param[in] requested_client_id   ID for which the callback has to be deleted.
 *                                  If UINT32_MAX as value is passed, all callbacks will be deleted.
 *
 * @retval                          PACE2_SUCCESS on success
 * @retval                          PACE2_POINTER_NULL on PACE2 module is NULL
 * @retval                          PACE2_FAILURE if no callback matching the own_client_id and requested_client_id
 *                                                could be found and therefore nothing was unregistered.
 */
PACE2_return_state pace2_ipc_unregister_callback(PACE2_module * const pace2_module,
                                                 u32 const own_client_id,
                                                 u32 const requested_client_id);

/**
 * @brief Reserves memory for IPC data.
 *
 * This method will reserve memory from the internal IPC message queue. A pointer to the reserved memory is passed in
 * the @p data parameter. After the call to @ref pace2_ipc_data_reserve the IPC client can write the shared data to
 * the reserved memory. The write process on the reserved memory must be finished with a call to
 * @ref pace2_ipc_data_push or @ref pace2_ipc_data_discard, otherwise the IPC system will block if the internal
 * message buffer is exceeded!
 *
 * This function is part of the IPC client interface.
 *
 * @param[in]  pace2_module     The PACE2 module.
 * @param[in]  thread_id        ID of the thread that wants to send the data.
 * @param[in]  producer_id      Own ID of the IPC client.
 * @param[in]  meta_data        Meta data for filtering this IPC data/message, can be NULL.
 * @param[in]  size             Number of bytes to reserve for the new IPC data. The maximum is configurable and
 *                              can be queried with @ref pace2_ipc_get_max_data_size.
 * @param[out] data             Pointer to the reserved data, the data is not initialized.
 * @retval                      PACE2_SUCCESS on success
 * @retval                      PACE2_OUT_OF_MEMORY if the message buffer is out of memory
 * @retval                      PACE2_INVALID if @p size exceeds the maximum configured size or IPC is disabled.
 * @retval                      PACE2_POINTER_NULL if one of the required pointer parameter is NULL.
 */
PACE2_return_state pace2_ipc_data_reserve(PACE2_module * const pace2_module,
                                          int thread_id,
                                          u32 producer_id,
                                          PACE2_ipc_meta_data const * const meta_data,
                                          u32 size,
                                          void * * const data);

/**
 * @brief Finish the IPC message that was reserved with @ref pace2_ipc_data_reserve before and
 *        send it to the other IPC clients.
 *
 * This function is part of the IPC client interface.
 *
 * @param[in] pace2_module  PACE2_module
 * @param[in] thread_id     ID of the thread that wants to send the data.
 * @param[in] data          Pointer to reserved data returned by @ref pace2_ipc_data_reserve.
 * @retval                  PACE2_SUCCESS on success.
 * @retval                  PACE2_POINTER_NULL if pace2_module or data is NULL.
 * @retval                  PACE2_THREAD_ID_INVALID if called from wrong thread.
 * @retval                  PACE2_INVALID if IPC is disabled.
 */
PACE2_return_state pace2_ipc_data_push(PACE2_module * const pace2_module,
                                       int thread_id,
                                       void * const data);

/**
 * @brief Discard the IPC message that was reserved with @ref pace2_ipc_data_reserve before. Discarded data is not
 *        available to clients any longer. The data will be overwritten automatically.
 *
 * This function is part of the IPC client interface.
 *
 * @param[in] pace2_module  PACE2_module
 * @param[in] thread_id     ID of the thread that wants to discard the data.
 * @param[in] data          Pointer to reserved data returned by @ref pace2_ipc_data_reserve.
 * @retval                  PACE2_SUCCESS on success.
 * @retval                  PACE2_POINTER_NULL if pace2_module or data is NULL.
 * @retval                  PACE2_THREAD_ID_INVALID if called from wrong thread.
 * @retval                  PACE2_INVALID if IPC is disabled.
 */
PACE2_return_state pace2_ipc_data_discard(PACE2_module * const pace2_module,
                                          int thread_id,
                                          void * const data);

/**
 * @brief Returns the maximum number of bytes which can be reserved by @ref pace2_ipc_data_reserve.
 *
 * @param[in]  pace2_module         PACE2_module
 * @param[in]  thread_id            ID of the thread.
 * @param[out] max_ipc_data_size    Pointer to store the maximum number of bytes in.
 * @retval                          PACE2_SUCCESS on success.
 * @retval                          PACE2_POINTER_NULL if pace2_module or data is NULL.
 * @retval                          PACE2_THREAD_ID_INVALID if called from wrong thread.
 * @retval                          PACE2_INVALID if IPC is disabled.
 */
PACE2_return_state pace2_ipc_get_max_data_size(PACE2_module * const pace2_module,
                                               int thread_id,
                                               u32 * max_ipc_data_size);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_IPC_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
