/** @file      pace2_ml.h
  * @brief     PACE 2 ML API.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_ML_H
#define PACE2_ML_H

#include <stddef.h>
#include "pace2_config.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Helper macro to check whether a specific ML class is set within a ML class bitfield.
 * E.g: PACE2_ML_CLASS_BITFIELD(bitfield, PACE2_ML_CLASS_ID_APPLE_SERVICES).
 */
#define PACE2_ML_CLASS_BITFIELD_CONTAINS(bitfield, class_id) \
    (((bitfield) >> (class_id) & 1) == 1)

/**
 * @brief Reconfigures the ML module.
 *
 * Use this function to enable, disable and reconfigure the ML module during the runtime of @productname.
 *
 * @note To use this function, the ML module has to be enabled during the initialization of @productname via the
 *       configuration option `ml.enable_module = IPQ_TRUE`.
 *       Otherwise, this function will return @ref PACE2_FEATURE_DISABLED.
 *       This means that the ML module can only by disabled and later re-enabled via this
 *       function if the ML module has initially been enabled during startup.
 *
 * @param[in] pace2_module                PACE2 module struct.
 * @param[in] thread_ID                   ID of the current thread.
 * @param[out] ml_config                  Address of an ML config instance.
 *
 * @retval PACE2_SUCCESS                  @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING            @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                        The parameter @p number_of_classes must not be @em NULL.
 * @retval PACE2_INVALID_MODULE_POINTER   @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval PACE2_FAILURE                  @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_POINTER_NULL             @copydoc PACE2_return_state::PACE2_POINTER_NULL
 * @retval PACE2_FEATURE_DISABLED         @copydoc PACE2_return_state::PACE2_FEATURE_DISABLED
 * @retval PACE2_FEATURE_NOT_AVAILABLE    @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_reconfigure(PACE2_module * pace2_module,
                                        int thread_ID,
                                        PACE2_config_ml const * ml_config);

/**
 * @brief Retrieves the currently active configuration of the ML module.
 *
 * Use this function to get the currently active configuration of the ML module.
 *
 * @note As with @ref pace2_ml_reconfigure, this function can also only be used when the ML module has be enabled
 *       during the initialization of @productname. Otherwise it will return @ref PACE2_FEATURE_DISABLED.
 *
 * @param[in] pace2_module                PACE2 module struct.
 * @param[in] thread_ID                   ID of the current thread.
 * @param[out] ml_config                  Address of an ML config instance that is to be filled.
 *
 * @retval PACE2_SUCCESS                  @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING            @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                        The parameter @p number_of_classes must not be @em NULL.
 * @retval PACE2_INVALID_MODULE_POINTER   @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval PACE2_POINTER_NULL             @copydoc PACE2_return_state::PACE2_POINTER_NULL
 * @retval PACE2_FEATURE_DISABLED         @copydoc PACE2_return_state::PACE2_FEATURE_DISABLED
 * @retval PACE2_FEATURE_NOT_AVAILABLE    @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_config(PACE2_module * pace2_module,
                                       int thread_ID,
                                       PACE2_config_ml * ml_config);

/**
 * @brief Gets the minimum number of packets that the ML module requires to determine a specific class,
 *        by class index.
 *
 * Can be used to iterate over all ML classes.
 *
 * @param[in] pace2_module                PACE2 module struct.
 *                                        Only used for dynamic upgrades. Otherwise, set to NULL.
 * @param[in] thread_ID                   ID of the current thread.
 *                                        Only used for dynamic upgrades. Otherwise, set to 0.
 * @param[in] ml_class_index              ML class index.
 * @param[out] min_number_of_packets      Minimum number of packets required for classification.
 *
 * @retval PACE2_SUCCESS                  @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING            @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                        The parameter @p number_of_classes must not be @em NULL.
 * @retval PACE2_INVALID_MODULE_POINTER   @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval PACE2_POINTER_NULL             @copydoc PACE2_return_state::PACE2_POINTER_NULL
 * @retval PACE2_FAILURE                  @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE    @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_min_number_of_packets_by_index(PACE2_module const * pace2_module,
                                                               int thread_ID,
                                                               u32 ml_class_index,
                                                               u32 * min_number_of_packets);

/**
 * @brief Gets the minimum number of packets that the ML module requires to determine a specific class,
 *        by class ID.
 *
 * @note You cannot use this function to iterate over all classes,
         because the class ID is not guaranteed to be a linearly increasing number.
 *       To iterate over all ML classes, use @ref pace2_ml_get_min_number_of_packets_by_index.
 *
 * @param[in] pace2_module                PACE2 module struct.
 *                                        Only used for dynamic upgrades. Otherwise, set to NULL.
 * @param[in] thread_ID                   ID of the current thread.
 *                                        Only used for dynamic upgrades. Otherwise, set to 0.
 * @param[in] ml_class_id                 ML class ID.
 * @param[out] min_number_of_packets      Minimum number of packets required for classification.
 *
 * @retval PACE2_SUCCESS                  @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING            @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                        The parameter @p number_of_classes must not be @em NULL.
 * @retval PACE2_INVALID_MODULE_POINTER   @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval PACE2_POINTER_NULL             @copydoc PACE2_return_state::PACE2_POINTER_NULL
 * @retval PACE2_FAILURE                  @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE    @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_min_number_of_packets(PACE2_module const * pace2_module,
                                                      int thread_ID,
                                                      u32 ml_class_id,
                                                      u32 * min_number_of_packets);

/**
 * @brief Retrieves a list of supported PACE2 applications by the ML module for a specific class by its index.
 *
 * Can be used to iterate over all ML classes.
 *
 * @note The list of application IDs (`application_id_list`) that this function returns
 *       is valid until a dynamic upgrade.
 *
 * @param[in] pace2_module                PACE2 module struct.
 *                                        Only used for dynamic upgrades. Otherwise, set to NULL.
 * @param[in] thread_ID                   ID of the current thread.
 *                                        Only used for dynamic upgrades. Otherwise, set to 0.
 * @param[in] ml_class_index              ML class index.
 * @param[out] application_id_list        List of supported applications.
 * @param[out] application_id_list_len    Length of the list of supported applications.
 *
 * @retval PACE2_SUCCESS                  @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING            @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                        The parameter @p number_of_classes must not be @em NULL.
 * @retval PACE2_INVALID_MODULE_POINTER   @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval PACE2_POINTER_NULL             @copydoc PACE2_return_state::PACE2_POINTER_NULL
 * @retval PACE2_FAILURE                  @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE    @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_supported_applications_by_index(PACE2_module const * pace2_module,
                                                                int thread_ID,
                                                                u32 ml_class_index,
                                                                PACE2_applications const ** application_id_list,
                                                                size_t * application_id_list_len);

/**
 * @brief Retrieves a list of supported PACE2 applications by the ML module for a specific class by its ID.
 *
 * As the class ID must not be a linearly increasing number this function can not be used to iterate over all classes.
 * @see pace2_ml_get_supported_applications_by_index.
 *
 * @note The application_id_list retrieved from this function is valid until a dynamic upgrade.
 *
 * @param[in] pace2_module                PACE2 module struct.
 *                                        Only used in case of dynamic upgrade otherwise it can be NULL.
 * @param[in] thread_ID                   ID of the current thread.
 *                                        Only used in case of dynamic upgrade otherwise it can be 0.
 * @param[in] ml_class_id                 ML class ID.
 * @param[out] application_id_list        List of supported applications.
 * @param[out] application_id_list_len    Length of the list of supported applications.
 *
 * @retval PACE2_SUCCESS                  @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING            @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                        The parameter @p number_of_classes must not be @em NULL.
 * @retval PACE2_INVALID_MODULE_POINTER   @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval PACE2_POINTER_NULL             @copydoc PACE2_return_state::PACE2_POINTER_NULL
 * @retval PACE2_FAILURE                  @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE    @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_supported_applications(PACE2_module const * pace2_module,
                                                       int thread_ID,
                                                       u32 ml_class_id,
                                                       PACE2_applications const ** application_id_list,
                                                       size_t * application_id_list_len);

/**
 * @brief Gets the attributes of an application classification supported by the ML module.
 *
 * @param[in]   pace2_module            Constant pointer to the PACE 2 module.
 *                                      Only used for dynamic upgrades. Otherwise, set to NULL.
 * @param[in]   thread_ID               ID of the current thread.
 *                                      Only used for dynamic upgrades. Otherwise, set to 0.
 * @param[in]   application_id          The application ID.
 * @param[out]  application_attributes  Pointer to bitmask of type @ref PACE2_application_attributes_bitmask
 *
 * @retval      PACE2_SUCCESS           @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval      PACE2_PARAM_MISSING     @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval      PACE2_FAILURE           @copydoc PACE2_return_state::PACE2_FAILURE
 *                                      The given application ID does not exist.
 */
PACE2_return_state pace2_ml_get_application_attributes_of_application(
    PACE2_module const * pace2_module,
    int thread_ID,
    PACE2_applications application_id,
    PACE2_application_attributes_mask * application_attributes);

/**
 * @brief Retrieves a bitfield of ML classes that are currently active for a specific flow.
 * 
 * @note The parameter class_bitfield only contains a valid bitfield of active classes if the function returned with
 *       @ref PACE2_return_state::PACE2_SUCCESS.
 *
 * @param[in] pace2_module                  Constant pointer to the PACE 2 module.
 * @param[in] thread_ID                     ID of the current thread.
 * @param[in] flow_data                     Pointer to the flow data.
 * @param[out] class_bitfield               32 bit bitfield storing the which classes are currently active.
 *                                          In order to check which classes are active within this bitfield, the helper
 *                                          macro @see PACE2_ML_CLASS_BITFIELD_CONTAINS should be used.
 * @retval PACE2_SUCCESS                    @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING              @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                          The parameter @p pace2_module must not be @em NULL.
 *                                          The parameter @p flow_data must not be @em NULL.
 *                                          The parameter @p class_bitfield must not be @em NULL.
 * @retval PACE2_FAILURE                    @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_DISABLED           @copydoc PACE2_return_state::PACE2_FEATURE_DISABLED
 * @retval PACE2_FEATURE_NOT_AVAILABLE      @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_active_classes(PACE2_module const * pace2_module,
                                               int thread_ID,
                                               PACE2_flow_data * flow_data,
                                               u32 * class_bitfield);

/**
 * @brief Enum defining available application lists.
 */
enum PACE2_ml_application_lists {
    /**
     * Application classifications that can trigger a ML module to classify other applications.
     */
    PACE2_ML_APPLICATION_LIST_APP_TO_APP_INPUT = 0,
    /**
     * Applications that can be classified by a ML module based on the classification of other applications.
     */
    PACE2_ML_APPLICATION_LIST_APP_TO_APP_OUTPUT = 1,
    /**
     * Application classifications that can trigger a ML module to classify application attributes.
     */
    PACE2_ML_APPLICATION_LIST_APP_TO_ATTRIBUTE_INPUT = 2,

    /**
     * Number of valid ml PACE2_ml_application_lists.
     */
    PACE2_ML_APPLICATION_LISTS_COUNT = 3,
};

/**
 * @brief Retrieves the number of entries in a specified application list.
 *
 * @note The length of an application lists includes its deactivated entries.
 *
 * @param[in] pace2_module                  Pointer to the PACE 2 module.
 * @param[in] thread_ID                     ID of the current thread.
 * @param[in] app_list                      The @ref PACE2_ml_application_lists to query.
 * @param[out] result_len                   Pointer where the retrieved length will be written to.
 *
 * @retval PACE2_SUCCESS                    @copydoc PACE2_return_state::PACE2_SUCCESS
 *                                          The retrieved length was written to result_len.
 * @retval PACE2_FAILURE                    @copydoc PACE2_return_state::PACE2_FAILURE
 *                                          The parameter @p app_list must be a valid application list.
 * @retval PACE2_PARAM_MISSING              @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                          The parameter @p pace2_module must not be @em NULL.
 *                                          The parameter @p thread_ID must be a valid thread_ID.
 *                                          The parameter @p result_len must not be @em NULL.
 * @retval PACE2_POINTER_NULL               @copydoc PACE2_return_state::PACE2_POINTER_NULL
 * @retval PACE2_FEATURE_DISABLED           @copydoc PACE2_return_state::PACE2_FEATURE_DISABLED
 */
PACE2_return_state pace2_ml_application_lists_get_len(PACE2_module const * pace2_module,
                                                      int thread_ID,
                                                      enum PACE2_ml_application_lists const app_list,
                                                      u32 * result_len);

/**
 * @brief Retrieves an application from a specified index and application list.
 *
 * @note Reading from an index that is either out-of-bounds or for an application that was deactivated (for the specified
 *       application list) is still considered successful.
 *       Then the result_app will be set to PACE2_APPLICATION_UNKNOWN.
 *
 * @param[in] pace2_module                  Pointer to the PACE 2 module.
 * @param[in] thread_ID                     ID of the current thread.
 * @param[in] app_list                      The @ref PACE2_ml_application_lists relevant to the inquiry.
 * @param[in] index                         The index from where to retrieve the application from the list.
 * @param[out] result_app                   Pointer where the retrieved app will be written to.
 *
 * @retval PACE2_SUCCESS                    @copydoc PACE2_return_state::PACE2_SUCCESS
 *                                          The retrieved app or PACE2_APPLICATION_UNKNOWN was written to result_app.
 * @retval PACE2_FAILURE                    @copydoc PACE2_return_state::PACE2_FAILURE
 *                                          The parameter @p app_list must be a valid application list.
 * @retval PACE2_PARAM_MISSING              @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                          The parameter @p pace2_module must not be @em NULL.
 *                                          The parameter @p thread_ID must be a valid thread_ID.
 *                                          The parameter @p result_app must not be @em NULL.
 * @retval PACE2_POINTER_NULL               @copydoc PACE2_return_state::PACE2_POINTER_NULL
 * @retval PACE2_FEATURE_DISABLED           @copydoc PACE2_return_state::PACE2_FEATURE_DISABLED
 */
PACE2_return_state pace2_ml_application_lists_get_app(PACE2_module const * pace2_module,
                                                      int thread_ID,
                                                      enum PACE2_ml_application_lists const app_list,
                                                      u32 const index,
                                                      PACE2_applications * result_app);

/**
 * @brief Retrieves the index of a specified application and application list.
 *
 * @note Trying to retrieve an application that is not part of the list or is deactivated is still considered
 *       successful.
 *       Then the result_index will be set to 0xFFFFFFFF.
 *
 * @param[in] pace2_module                  Pointer to the PACE 2 module.
 * @param[in] thread_ID                     ID of the current thread.
 * @param[in] app_list                      The @ref PACE2_ml_application_lists relevant to the inquiry.
 * @param[in] app                           The application of which the index is to be retrieved.
 * @param[out] result_index                 Pointer where the retrieved index will be written to.
 *
 * @retval PACE2_SUCCESS                    @copydoc PACE2_return_state::PACE2_SUCCESS
 *                                          The retrieved index or 0xFFFFFFFF was written to result_index.
 * @retval PACE2_FAILURE                    @copydoc PACE2_return_state::PACE2_FAILURE
 *                                          The parameter @p app_list must be a valid application list.
 * @retval PACE2_PARAM_MISSING              @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                          The parameter @p pace2_module must not be @em NULL.
 *                                          The parameter @p thread_ID must be a valid thread_ID.
 *                                          The parameter @p result_index must not be @em NULL.
 * @retval PACE2_POINTER_NULL               @copydoc PACE2_return_state::PACE2_POINTER_NULL
 * @retval PACE2_FEATURE_DISABLED           @copydoc PACE2_return_state::PACE2_FEATURE_DISABLED
 */
PACE2_return_state pace2_ml_application_lists_get_index(PACE2_module const * pace2_module,
                                                        int thread_ID,
                                                        enum PACE2_ml_application_lists const app_list,
                                                        PACE2_applications const app,
                                                        u32 * result_index);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_ML_CLASS_H */
