/** @file      pace2_ml_class.h
  * @brief     PACE 2 ML event class API.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_ML_CLASS_H
#define PACE2_ML_CLASS_H

#include <stddef.h>
#include "pace2_basic_def.h"
#include "ipq_config.h"

#ifdef __cplusplus
extern "C" {
#endif

/** @file */

/**
 * @defgroup CLASS_IDS IDs of all supported classes for machine learning classifications.
 *
 * These defines describe all classes that are currently supported. The defines can be used to statically enable
 * individual classes via the configuration. They can also be used to query a detailed definition of each class
 * via @ref pace2_ml_get_class_info.
 *
 * @{
 */
#define PACE2_ML_CLASS_ID_APPLE_SERVICES 0
#define PACE2_ML_CLASS_ID_STREAMING_DOWNLOADING 1
#define PACE2_ML_CLASS_ID_GOOGLE_SERVICES 2

#define PACE2_ML_CLASS_COUNT PACE2_ML_CLASS_COUNT_INT
/** @} */

/**
 * @defgroup CATEGORY_IDS IDs of all supported categories for machine learning classifications.
 *
 * These defines describe all categories that are currently supported.They can be used to query a detailed definition of
 * each category via @ref pace2_ml_get_category_info.
 *
 * @{
 */
#define PACE2_ML_CATEGORY_ID_APPLICATION 0
#define PACE2_ML_CATEGORY_ID_APPLICATION_ATTRIBUTE 1
/** @} */

/**
 * @brief ML event category information.
 *
 * This structure defines categories for machine learning based classifications. Each class is part of one category and
 * as such will contain explicitly subclasses that fall into this category. A category can be anything from applications
 * and attributes to other properties of network traffic that can be classified with machine learning techniques.
 */
typedef struct pace2_ml_category_info {

    u32 category_id;           /*!< Unique category ID. */
    char const * name;         /*!< Internal subclass name (only valid C names). */
    char const * display_name; /*!< User-friendly name of this category. */
    char const * description;  /*!< Full-text description of this category. */

} PACE2_ml_category_info;

/**
 * @brief ML event class information.
 *
 * This structure defines the classes of machine learning classifications. A class groups a set of subclasses
 * under one machine learning model.
 */
typedef struct pace2_ml_class_info {

    u32 class_id;              /*!< Unique class ID. */
    u32 category_id;           /*!< Information about the category this class belongs to. */
    char const * name;         /*!< Internal class name (only valid C names). */
    char const * display_name; /*!< User-friendly name of this class. */
    char const * description;  /*!< Full-text description of this class. */

    u32 version; /*!< Version of the class. This will be incremented with each new model for that class. */

} PACE2_ml_class_info;

/**
 * @brief ML event subclass information.
 *
 * This structure defines the subclasses of machine learning classifications. A subclass is part of a class.
 * Subclasses describe a single classification instance and represent a unique property of a flow,
 * such as an application or attribute.
 */
typedef struct pace2_ml_subclass_info {

    u32 subclass_id;           /*!< Unique subclass ID. */
    char const * name;         /*!< Internal subclass name (only valid C names). */
    char const * display_name; /*!< User-friendly name of this subclass. */
    char const * description;  /*!< Full-text description of this subclass. */

} PACE2_ml_subclass_info;

/**
 * @brief Gets the number of supported ML categories.
 *
 * @param[in] pace2_module                PACE2 module struct.
 * @param[in] thread_ID                   ID of the current thread.
 * @param[out] number_of_categories       Address to store the number of ML categories.
 *
 * @retval PACE2_SUCCESS                  @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING            @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                        (The parameter @p number_of_categories must not be @em NULL.)
 * @retval PACE2_FAILURE                  @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE    @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_number_of_categories(PACE2_module const * pace2_module,
                                                     int thread_ID,
                                                     u32 * number_of_categories);

/**
 * @brief Gets the number of supported ML classes.
 *
 * @param[in] pace2_module                PACE2 module struct.
 * @param[in] thread_ID                   ID of the current thread.
 * @param[out] number_of_classes          Address to store the number of ML classes.
 *
 * @retval PACE2_SUCCESS                  @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING            @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                        (The parameter @p number_of_classes must not be @em NULL.)
 * @retval PACE2_FEATURE_NOT_AVAILABLE    @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_number_of_classes(PACE2_module const * pace2_module,
                                                  int thread_ID,
                                                  u32 * number_of_classes);

/**
 * @brief Gets the number of supported ML subclasses for a specific class.
 *
 * As the class ID must not be a linearly increasing number this function can not be used to iterate over all classes.
 * @see pace2_ml_get_number_of_subclasses_by_index.
 *
 * @param[in] pace2_module                PACE2 module struct.
 * @param[in] thread_ID                   ID of the current thread.
 * @param[in] ml_class_id                 ML class ID.
 * @param[out] number_of_subclasses       Address to store the number of ML subclasses.
 *
 * @retval PACE2_SUCCESS                  @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING            @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                        (The parameter @p number_of_subclasses must not be @em NULL.)
 * @retval PACE2_FAILURE                  @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE    @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_number_of_subclasses(PACE2_module const * pace2_module,
                                                     int thread_ID,
                                                     u32 ml_class_id,
                                                     u32 * number_of_subclasses);

/**
 * @brief Gets the ML category information by the category ID.
 *
 * Can be used to access a single ML category information for a category ID.
 * As the category ID must not be a linearly increasing number this function can not be used to iterate over all
 * categories. @see pace2_ml_get_category_info_by_index.
 *
 * @param[in] pace2_module                   PACE2 module struct.
 * @param[in] thread_ID                      ID of the current thread.
 * @param[in] ml_category_id                 ML category ID.
 * @param[out] ml_category_info              Address to store the address of the ML category information. The content of
 *                                           the ml_category_info instance must not be changed. Valid until either a
 *                                           dynamic-upgrade occurs or PACE2 is de-initialized via
 *                                           @ref pace2_exit_module.
 *
 * @retval PACE2_SUCCESS                     @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING               @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                           (The parameter @p ml_category_info must not be @em NULL.)
 * @retval PACE2_FAILURE                     @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE       @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_category_info(PACE2_module const * pace2_module,
                                              int thread_ID,
                                              u32 ml_category_id,
                                              PACE2_ml_category_info const ** ml_category_info);

/**
 * @brief Gets the ML class information by the class ID.
 *
 * Can be used to access a single ML class information for a class ID.
 * As the class ID must not be a linearly increasing number this function can not be used to iterate over all classes.
 * @see pace2_ml_get_class_info_by_index.
 *
 * @param[in] pace2_module                   PACE2 module struct.
 * @param[in] thread_ID                      ID of the current thread.
 * @param[in] ml_class_id                    ML class ID.
 * @param[out] ml_class_info                 Address to store the address of the ML class information. The content of
 *                                           the ml_class_info instance must not be changed. Valid until either a
 *                                           dynamic-upgrade occurs or PACE2 is de-initialized via
 *                                           @ref pace2_exit_module.
 *
 * @retval PACE2_SUCCESS                     @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING               @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                           (The parameter @p ml_class_info must not be @em NULL.)
 * @retval PACE2_FAILURE                     @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE       @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_class_info(PACE2_module const * pace2_module,
                                           int thread_ID,
                                           u32 ml_class_id,
                                           PACE2_ml_class_info const ** ml_class_info);

/**
 * @brief Gets the ML subclass information for a specific class by the class ID and subclass ID.
 *
 * Can be used to access a single ML subclass information for a class ID and subclass ID.
 * As the class ID and the subclass ID must not be a linearly increasing number this function can not be used to iterate
 * over all classes and subclasses. @see pace2_ml_get_subclass_info_by_index.
 *
 * @param[in] pace2_module                   PACE2 module struct.
 * @param[in] thread_ID                      ID of the current thread.
 * @param[in] ml_class_id                    ML class ID.
 * @param[in] ml_subclass_id                 ML subclass ID.
 * @param[out] ml_subclass_info              Address to store the address of the ML subclass information. The content of
 *                                           the ml_subclass_info instance must not be changed. Valid until either a
 *                                           dynamic-upgrade occurs or PACE2 is de-initialized via
 *                                           @ref pace2_exit_module.
 *
 * @retval PACE2_SUCCESS                     @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING               @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                           (The parameter @p ml_subclass_info must not be @em NULL.)
 * @retval PACE2_FAILURE                     @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE       @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_subclass_info(PACE2_module const * pace2_module,
                                              int thread_ID,
                                              u32 ml_class_id,
                                              u32 ml_subclass_id,
                                              PACE2_ml_subclass_info const ** ml_subclass_info);

/**
 * @brief Gets the ML class and subclass information for a specific class by the class ID and subclass ID.
 *
 * Can be used to access a single ML class and subclass information for a class ID and subclass ID.
 * As the category ID, class ID and subclass ID must not be a linearly increasing number this function can not be used
 * to iterate over all categories, classes and subclasses.
 *
 * @param[in] pace2_module                   PACE2 module struct.
 * @param[in] thread_ID                      ID of the current thread.
 * @param[in] ml_category_id                 ML category ID.
 * @param[in] ml_class_id                    ML class ID.
 * @param[in] ml_subclass_id                 ML subclass ID.
 * @param[out] ml_category_info              Address to store the address of the ML category information. The content of
 *                                           the ml_category_info instance must not be changed. Valid until either a
 *                                           dynamic-upgrade occurs or PACE2 is de-initialized via
 *                                           @ref pace2_exit_module.
 * @param[out] ml_class_info                 Address to store the address of the ML class information. The content of
 *                                           the ml_class_info instance must not be changed. Valid until either a
 *                                           dynamic-upgrade occurs or PACE2 is de-initialized via
 *                                           @ref pace2_exit_module.
 * @param[out] ml_subclass_info              Address to store the address of the ML subclass information. The content of
 *                                           the ml_subclass_info instance must not be changed. Valid until either a
 *                                           dynamic-upgrade occurs or PACE2 is de-initialized via
 *                                           @ref pace2_exit_module.
 *
 * @retval PACE2_SUCCESS                     @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING               @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                           (The parameters @p ml_category_info, @p ml_class_info
 *                                           and @p ml_subclass_info must not be @em NULL.)
 * @retval PACE2_FAILURE                     @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE       @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_category_class_subclass_info(PACE2_module const * pace2_module,
                                                             int thread_ID,
                                                             u32 ml_category_id,
                                                             u32 ml_class_id,
                                                             u32 ml_subclass_id,
                                                             PACE2_ml_category_info const ** ml_category_info,
                                                             PACE2_ml_class_info const ** ml_class_info,
                                                             PACE2_ml_subclass_info const ** ml_subclass_info);

/**
 * @brief Gets the number of supported ML subclasses for a specific class by a linearly increasing index.
 *
 * Can be used to iterate over all ML classes.
 *
 * @param[in] pace2_module                  PACE2 module struct.
 * @param[in] thread_ID                     ID of the current thread.
 * @param[in] ml_class_index                ML class index. This index must be smaller than the number of classes
 *                                          obtained from @ref pace2_ml_get_number_of_classes.
 * @param[out] number_of_subclasses         Address to store the number of ML subclasses.
 *
 * @retval PACE2_SUCCESS                    @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING              @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                          (The parameter @p number_of_subclasses must not be @em NULL.)
 * @retval PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS @copydoc PACE2_return_state::PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS
 * @retval PACE2_FAILURE                    @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE      @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_number_of_subclasses_by_index(PACE2_module const * pace2_module,
                                                              int thread_ID,
                                                              u32 ml_class_index,
                                                              u32 * number_of_subclasses);

/**
 * @brief Gets the ML category information by a linearly increasing index.
 *
 * Can be used to iterate over all ML categories.
 *
 * @param[in] pace2_module                   PACE2 module struct.
 * @param[in] thread_ID                      ID of the current thread.
 * @param[in] ml_category_index              ML category index. This index must be smaller than the number of categories
 *                                           obtained from @ref pace2_ml_get_number_of_categories.
 * @param[out] ml_category_info              Address to store the address of the ML category information. The content of
 *                                           the ml_category_info instance must not be changed. Valid until either a
 *                                           dynamic-upgrade occurs or PACE2 is de-initialized via
 *                                           @ref pace2_exit_module.
 *
 * @retval PACE2_SUCCESS                     @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING               @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                           (The parameter @p ml_category_info must not be @em NULL.)
 * @retval PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS  @copydoc PACE2_return_state::PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS
 * @retval PACE2_FAILURE                     @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE       @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_category_info_by_index(PACE2_module const * pace2_module,
                                                       int thread_ID,
                                                       u32 ml_category_index,
                                                       PACE2_ml_category_info const ** ml_category_info);

/**
 * @brief Gets the ML class information by a linearly increasing index.
 *
 * Can be used to iterate over all ML classes.
 *
 * @param[in] pace2_module                   PACE2 module struct.
 * @param[in] thread_ID                      ID of the current thread.
 * @param[in] ml_class_index                 ML class index. This index must be smaller than the number of classes
 *                                           obtained from @ref pace2_ml_get_number_of_classes.
 * @param[out] ml_class_info                 Address to store the address of the ML class information. The content of
 *                                           the ml_class_info instance must not be changed. Valid until either a
 *                                           dynamic-upgrade occurs or PACE2 is de-initialized via
 *                                           @ref pace2_exit_module.
 *
 * @retval PACE2_SUCCESS                     @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING               @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                           (The parameter @p ml_class_info must not be @em NULL.)
 * @retval PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS  @copydoc PACE2_return_state::PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS
 * @retval PACE2_FAILURE                     @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE       @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_class_info_by_index(PACE2_module const * pace2_module,
                                                    int thread_ID,
                                                    u32 ml_class_index,
                                                    PACE2_ml_class_info const ** ml_class_info);

/**
 * @brief Gets the ML subclass information by a linearly increasing index for a specific ML class ID.
 *
 * Can be used to iterate over all ML subclasses.
 *
 * @param[in] pace2_module                   PACE2 module struct.
 * @param[in] thread_ID                      ID of the current thread.
 * @param[in] ml_class_id                    ML class ID.
 * @param[in] ml_subclass_index              ML subclass index. This index must be smaller than the number of subclasses
 *                                           obtained from @ref pace2_ml_get_number_of_subclasses.
 * @param[out] ml_subclass_info              Address to store the address of the ML subclass information. The content of
 *                                           the ml_subclass_info instance must not be changed. Valid until either a
 *                                           dynamic-upgrade occurs or PACE2 is de-initialized via
 *                                           @ref pace2_exit_module.
 *
 * @retval PACE2_SUCCESS                     @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING               @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                           (The parameter @p ml_subclass_info must not be @em NULL.)
 * @retval PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS  @copydoc PACE2_return_state::PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS
 * @retval PACE2_FAILURE                     @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE       @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_subclass_info_by_index(PACE2_module const * pace2_module,
                                                       int thread_ID,
                                                       u32 ml_class_id,
                                                       u32 ml_subclass_index,
                                                       PACE2_ml_subclass_info const ** ml_subclass_info);

/**
 * @brief Gets the ML class and subclass information by a linearly increasing index for both the ML class and the ML
 *        subclass.
 *
 * Can be used to iterate over all ML classes and ML subclasses.
 *
 * @param[in] pace2_module                   PACE2 module struct.
 * @param[in] thread_ID                      ID of the current thread.
 * @param[in] ml_class_index                 ML class index. This index must be smaller than the number of classes
 *                                           obtained from @ref pace2_ml_get_number_of_classes.
 * @param[in] ml_subclass_index              ML subclass index. This index must be smaller than the number of subclasses
 *                                           obtained from @ref pace2_ml_get_number_of_subclasses.
 * @param[out] ml_class_info                 Address to store the address of the ML class information. The content of
 *                                           the ml_class_info instance must not be changed. Valid until either a
 *                                           dynamic-upgrade occurs or PACE2 is de-initialized via
 *                                           @ref pace2_exit_module.
 * @param[out] ml_subclass_info              Address to store the address of the ML subclass information. The content of
 *                                           the ml_subclass_info instance must not be changed. Valid until either a
 *                                           dynamic-upgrade occurs or PACE2 is de-initialized via
 *                                           @ref pace2_exit_module.
 *
 * @retval PACE2_SUCCESS                     @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_PARAM_MISSING               @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *                                           (The parameters @p ml_class_info and @p ml_subclass_info
 *                                           must not be @em NULL.)
 *                                           ml_class_info parameter must not be NULL.
 * @retval PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS  @copydoc PACE2_return_state::PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS
 * @retval PACE2_FAILURE                     @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval PACE2_FEATURE_NOT_AVAILABLE       @copydoc PACE2_return_state::PACE2_FEATURE_NOT_AVAILABLE
 */
PACE2_return_state pace2_ml_get_class_subclass_info_by_index(PACE2_module const * pace2_module,
                                                             int thread_ID,
                                                             u32 ml_class_index,
                                                             u32 ml_subclass_index,
                                                             PACE2_ml_class_info const ** ml_class_info,
                                                             PACE2_ml_subclass_info const ** ml_subclass_info);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_ML_CLASS_H */
