/** @file      pace2_packet_stack.h
  * @brief     PACE 2 packet stack type definitions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_PACKET_STACK_H
#define PACE2_PACKET_STACK_H

/* Includes *********************************************************************/
#ifdef __KERNEL__
#include <linux/in.h>
#elif !defined(_WIN32) && !(defined(__OCTEON__) && !defined(__linux__))
#include <netinet/in.h>
#endif // __KERNEL__

#include "pace2_protocols.h"
#include "ipq_basic_def.h"

#ifdef _WIN32
#include <winsock2.h>
#include <in6addr.h>
#endif /* _WIN32 */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Definition of a packet frame type.
 */
typedef enum {
    IPv4,  /*!< IPv4 frame. */
    IPv6,  /*!< IPv6 frame. */
    TCP,   /*!< TCP frame. */
    UDP,   /*!< UDP frame. */
    L2,    /*!< Generic layer 2 frame. */
    L3,    /*!< Generic layer 3 frame. */
    L4,    /*!< Generic layer 4 frame. */
    L7,    /*!< Generic layer 7 frame. */
    OTHER, /*!< Any other frame type. */
    SCTP,  /*!< SCTP frame. */
} PACE2_packet_frame_type;

/**
 * @brief Definition of a frame descriptor.
 */
typedef struct pace2_packet_frame_descriptor {
    union tt {
        const struct iphdr * ipv4;   /*!< Pointer to IPv4 frame data. */
        const struct ip6_hdr * ipv6; /*!< Pointer to IPv6 frame data. */
        const struct sctphdr * sctp; /*!< Pointer to SCTP frame data. */
        const struct tcphdr * tcp;   /*!< Pointer to TCP frame data. */
        const struct udphdr * udp;   /*!< Pointer to UDP frame data. */
        const void * l2_data;        /*!< Pointer to OSI layer 2 frame data. */
        const void * l3_data;        /*!< Pointer to OSI layer 3 frame data. */
        const void * l4_data;        /*!< Pointer to OSI layer 4 frame data. */
        const void * l7_data;        /*!< Pointer to OSI layer 7 frame data. */
        const void * other_data;     /*!< Pointer to any other frame data. */
    } frame_data;                    /*!< Named union which contains the data pointer of the frame type. */
    union {
        struct {
            u16 frame_length;                    /*!< Frame data length. */
            u16 frame_layer_protocol;            /*!< The layer 4 protocol type. */
            PACE2_packet_frame_type type;        /*!< Frame type. */
            enum pace2_protocols pace2_protocol; /*!< The PACE 2 protocol enum type. */
        };
        u64 padding[4]; /*!< Reserved bytes for future extensions. */
    };
} PACE2_packet_frame_descriptor;

/**
 * @brief Definition of a frame stack, which contains the frame descriptors.
 */
typedef struct pace2_packet_stack {
    PACE2_packet_frame_descriptor * stack; /*!< Pointer to the frame stack. */
    u32 stack_size;                        /*!< Frame stack size. */
    u8 inner_ip_index;                     /*!< index of the most inner IP layer */
    u8 outer_ip_index;                     /*!< index of the most outer IP layer */
} PACE2_packet_stack;

#ifdef __cplusplus
}
#endif
#endif /* PACE2_PACKET_STACK_H */
