/** @file      pace2_pht.h
  * @brief     Polling hash table related API functions
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_PHT_H
#define PACE2_PHT_H

/* Includes *********************************************************************/
#include "pace2_basic_def.h"
#include "pace2_pht_core.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Size of one flow or subscriber dump.
 */
#define PACE2_PHT_DUMP_DATA_SIZE 128

/**
 * @brief Create a PACE2 hash table.
 *
 * Memory allocations issued by the created hash table will use PACE2_SCOPE_EXTERNAL.
 *
 * @param[in] config    pointer to the configuration structure
 * @param[in] thread_id hash table thread identifier
 * @retval    PACE2_pht pointer to the hash table
 * @retval    NULL      hash table initialization failed
 */
PACE2_pht * pace2_pht_create(PACE2_pht_config const * config, int thread_id);

/**
 * @brief Decapsulate the IPv4 or IPv6 header and do an insert operation on a
 *        hash table with the 5 tuple.
 *
 * The maximum tunnel decapsulation level is used from the PACE2 configuration.
 *
 * @param[in] module          pointer to the detection module
 * @param[in] pht             pointer to the hash table
 * @param[in] thread_id       hash table thread identifier
 * @param[in] ip_ptr          pointer to the ip header
 * @param[in] ipsize          length of the ip packet
 * @param[out] new_connection set to != 0 for a new element
 * @retval                    pointer to the flow structure
 * @retval                    NULL on failure
 */
void * pace2_pht_get_current_flow_v4_v6_decapsulate(
    PACE2_module * module, u16 thread_id, PACE2_pht * pht, void const * ip_ptr, u16 ipsize, u8 * new_connection);

/**
 * @brief Dump the flow information in the given buffer.
 *
 * @param[in]  module          pointer to the detection module
 * @param[in]  thread_id       hash table thread identifier
 * @param[in]  flow            a void pointer to the flow
 * @param[out] buffer          the buffer where to dump in,
 *                             needs the minimum size of
 *                             @ref PACE2_PHT_DUMP_DATA_SIZE
 */
void pace2_pht_dump_flow_data(PACE2_module * module, int thread_id, void const * flow, u8 * buffer);

/**
 * @brief Load the flow information from the given flow dump.
 *
 * @param[in]  module          pointer to the detection module
 * @param[in]  thread_id       hash table thread identifier
 * @param[in]  flow            a void pointer to the flow
 * @param[in]  flow_dump       the dump from which the flow
 *                             information are loaded
 */
void pace2_pht_load_flow_dump(PACE2_module * module, int thread_id, void * flow, u8 const * flow_dump);

/**
 * @brief Dump the subscriber information in the given buffer.
 *
 * @param[in]  module          pointer to the detection module
 * @param[in]  thread_id       hash table thread identifier
 * @param[in]  id              a void pointer to the subscriber
 * @param[out] buffer          the buffer where to dump in,
 *                             needs the minimum size of
 *                             @ref PACE2_PHT_DUMP_DATA_SIZE
 */
void pace2_pht_dump_id_data(PACE2_module * module, int thread_id, void * id, u8 * buffer);

/**
 * @brief Load the subscriber information from the given id dump.
 *
 * @param[in]  module          pointer to the detection module
 * @param[in]  thread_id       hash table thread identifier
 * @param[in]  id              a void pointer to the subscriber
 * @param[in]  id_dump         the dump from which the subscriber
 *                             information are loaded
 */
void pace2_pht_load_id_dump(PACE2_module * module, int thread_id, void * id, u8 const * id_dump);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_PHT_H */
