/** @file      pace2_pht_core.h
  * @brief     Polling hash table core API
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_PHT_CORE_H
#define PACE2_PHT_CORE_H

/* Includes *********************************************************************/
#include "pace2_basic_def.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Helper macros to use PHT with fixed millisecond resolution, to avoid currently existing accuracy problems, when
 * using higher resolutions. */
#define PACE2_PHT_RESOLUTION_MILLI 1000
#define PACE2_PHT_TIME_TO_MILLI(_time_, _time_resolution_) \
    ((_time_) / ((_time_resolution_) / PACE2_PHT_RESOLUTION_MILLI))


/* Public Types *****************************************************************/

typedef struct pace2_pht PACE2_pht;
typedef struct pace2_pht_raw_element PACE2_pht_raw_element;

/**
 * @brief Hash table resize mode
 */
typedef enum PACE2_pht_resize_mode {
    PACE2_PHT_RESIZE_DISABLED, /*!< Disabled hash table resize. Old elements
                                    will automatically be removed. */
    PACE2_PHT_RESIZE_ENABLED   /*!< The hash table will resize automatically
                                    if it is full. */
} PACE2_pht_resize_mode;

/**
 * @brief Return state of pht PACE2 functions.
 */
typedef enum {
    PACE2_PHT_SUCCESS = 0,              /*!< Processing was successful, no more elements available. */
    PACE2_PHT_PARAM_MISSING,            /*!< Function parameter is missing. */
    PACE2_PHT_OUT_OF_MEMORY,            /*!< Processing failed, because no more memory available. */
    PACE2_PHT_FAILURE,                  /*!< Processing failed. */
    PACE2_PHT_MORE_ELEMENTS             /*!< requested element was available. */
} PACE2_pht_return_state;

/**
 * @brief The reason for a hash table element removal
 */
typedef enum pace2_pht_removal_reason {
    PHT_NONE,     /*!< Default value. The item was removed without a reason. This should not happen.*/
    PHT_TIMEOUT,  /*!< the item was removed because of a timeout */
    PHT_OVERFLOW, /*!< the item was removed because the hash table was full */
    PHT_DELETED,  /*!< the item was manually removed from the hash table */
    PHT_REUSED,   /*!< The item is reused and was not removed from the hash table.
                       The value is used for reused TCP flows as part of PACE2 only. */

    PTH_NUMBER_OF_REASONS   /*!< Number of removal reasons and not a reason itself. */
} PACE2_pht_removal_reason;

/**
 * @brief Hash table configuration structure
 */
typedef struct PACE2_pht_config {
    u64 memory_size;                        /*!< Amount of memory to allocate in bytes. If dynamic resizing is
                                                 enabled this will be the initial amount. */

    u32 unique_key_size;                    /*!< Size of the hash table key in bytes. Must be multiple of 4. */

    u32 user_buffer_size;                   /*!< Size of a single element to be stored.
                                                 May be 0 if only the existence of an entry is of interest. */

    PACE2_timestamp timeout;                /*!< Element timeout. Has to be specified in the same unit which wants to be
                                                 used for the input/output timestamps in all other pace2_pht_*
                                                 functions. It's not yet recommended to use higher resolutions than
                                                 1000 (ms), because due to internal optimizations, overflows (incorrect
                                                 timeouts) may happen otherwise. */

    pace2_malloc_t ipq_malloc;              /*!< Pointer to the memory allocation function. */

    pace2_free_t ipq_free;                  /*!< Pointer to the memory deallocation function. */

    void * allocation_userptr;              /*!< User pointer that will be given to the memory functions */

    PACE2_pht_resize_mode resize_mode;      /*!< Allow dynamic resizing or not */

    u64 max_memory_size;                    /*!< Maximum amount of memory in bytes.
                                                 Set an upper limit for resizing hash tables.
                                                 Has no effect if resizing is disabled. 0 means: no limit. */
    u16 timeout_queues;                     /*!< Number of registered timeout queues. */
} PACE2_pht_config;

/* Public Prototypes ************************************************************/

/**
 * @brief Apply a default configuration structure to the hash table.
 *
 * @param[in] config    pointer to the configuration structure
 */
void pace2_pht_init_default_config(PACE2_pht_config * config);

/**
 * @brief Get a hash table element and it's timestamp value.
 *
 * @param[in]  pht             pointer to the hash table
 * @param[in]  unique_key      pointer to the key
 * @param[out] last_access_ts  A pointer to a variable to store the timestamp value of the last call for this element.
 *                             The value is only valid for successful (not NULL) return.
 *
 * @returns                    The pointer to the hash table data.
 * @retval     NULL            at failed lookup
 */
void * pace2_pht_lookup_ts(PACE2_pht * pht, u8 const * unique_key, PACE2_timestamp * last_access_ts);

/**
 * @brief Get a hash table element.
 *
 * @param[in] pht             pointer to the hash table
 * @param[in] unique_key      pointer to the key
 *
 * @returns                   The pointer to the hash table data.
 * @retval    NULL            failed lookup
 */
void * pace2_pht_lookup(PACE2_pht * pht, u8 const * unique_key);

/**
 * @brief Update the timestamp for an element retrieved by a lookup call
 *
 * @param[in] pht                       pointer to the hash table
 * @param[in] user_buffer               pointer to the user data
 *
 * @retval    PACE2_PHT_SUCCESS         @copydoc PACE2_pht_return_state::PACE2_PHT_SUCCESS
 * @retval    PACE2_PHT_PARAM_MISSING   @copydoc PACE2_pht_return_state::PACE2_PHT_PARAM_MISSING
 * @retval    PACE2_PHT_FAILURE         @copydoc PACE2_pht_return_state::PACE2_PHT_FAILURE
 */
PACE2_pht_return_state pace2_pht_update_ts_of_element(PACE2_pht * pht, void * user_buffer);

/**
 * @brief Insert/update an element into the hash table.
 *
 * If the hashtable is full this operation will fail and return NULL.
 * To avoid this the necessary memory must be reserved prior to the
 * insert with pace2_pht_reserve_elements followed by pace2_pht_get_next_element_to_remove.
 *
 * In case of an insert of an already inserted element its timestamp
 * will get updated.
 *
 * @param[in] pht             pointer to the hash table
 * @param[in] unique_key      pointer to the lookup key
 * @param[out] new_entry      pointer to an u8, will be set to != 0 when the
 *                            insert has created a new entry in the hash table.
 *                            Ignored, if NULL.
 *
 * @returns                   The pointer to the hash table data.
 * @retval     NULL           on failure
 */
void * pace2_pht_insert(PACE2_pht * pht, u8 const * unique_key, u8 * new_entry);

/**
 * @brief Remove one element from the hash table.
 *
 * @param[in] pht                       pointer to the hash table
 * @param[in] unique_key                pointer to the lookup key
 *
 * @retval    PACE2_PHT_SUCCESS         @copydoc PACE2_pht_return_state::PACE2_PHT_SUCCESS
 * @retval    PACE2_PHT_PARAM_MISSING   @copydoc PACE2_pht_return_state::PACE2_PHT_PARAM_MISSING
 */
PACE2_pht_return_state pace2_pht_delete(PACE2_pht * pht, u8 const * unique_key);

/**
 * @brief Free all memory which has been allocated for a hash table.
 *
 * @param[in] pht                       pointer to the hash table
 *
 * @retval    PACE2_PHT_SUCCESS         @copydoc PACE2_pht_return_state::PACE2_PHT_SUCCESS
 * @retval    PACE2_PHT_PARAM_MISSING   @copydoc PACE2_pht_return_state::PACE2_PHT_PARAM_MISSING
 */
PACE2_pht_return_state pace2_pht_destroy(PACE2_pht * pht);

/**
 * @brief Set the timestamp on the hash table. Should be done before any
 *        insertion or lookup operation.
 *
 * @param[in] pht                       pointer to the hash table
 * @param[in] timestamp                 the timestamp in ticks
 *
 * @retval    PACE2_PHT_SUCCESS         @copydoc PACE2_pht_return_state::PACE2_PHT_SUCCESS
 * @retval    PACE2_PHT_PARAM_MISSING   @copydoc PACE2_pht_return_state::PACE2_PHT_PARAM_MISSING
 */
PACE2_pht_return_state pace2_pht_set_timestamp(PACE2_pht * pht, PACE2_timestamp timestamp);

/**
 * @brief Set or change the timeout of a timeout queue.
 *
 * @param[in] pht                       pointer to the hash table
 * @param[in] timeout                   the current timestamp of the hash table can be
 *                                      @ref IPQ_TOH_TIMEOUT_INFINITE for infinite
 *                                      timeouts
 * @param[in] queue                     The event queue for the timeout
 *
 * @retval    PACE2_PHT_SUCCESS         @copydoc PACE2_pht_return_state::PACE2_PHT_SUCCESS
 * @retval    PACE2_PHT_PARAM_MISSING   @copydoc PACE2_pht_return_state::PACE2_PHT_PARAM_MISSING
 * @retval    PACE2_PHT_FAILURE         @copydoc PACE2_pht_return_state::PACE2_PHT_FAILURE
 */
PACE2_pht_return_state pace2_pht_set_timeout(PACE2_pht * pht, PACE2_timestamp timeout, u16 queue);

/**
 * @brief Set the timeout queue for the last insert operation.
 *
 * @param[in] pht                       pointer to the hash table
 * @param[in] queue                     the new timeout queue for the element
 *
 * @retval    PACE2_PHT_SUCCESS         @copydoc PACE2_pht_return_state::PACE2_PHT_SUCCESS
 * @retval    PACE2_PHT_PARAM_MISSING   @copydoc PACE2_pht_return_state::PACE2_PHT_PARAM_MISSING
 * @retval    PACE2_PHT_FAILURE         @copydoc PACE2_pht_return_state::PACE2_PHT_FAILURE
 */
PACE2_pht_return_state pace2_pht_set_timeout_queue(PACE2_pht * pht, u16 queue);

/**
 * @brief Delete all entries from the hash table.
 *
 * @param[in] pht             pointer to the hash table
 */
void pace2_pht_clear(PACE2_pht * pht);

/**
 * @brief Retrieve the current timeout for a specific queue.
 *
 * @param[in] pht             pointer to the hash table
 * @param[in] queue           the timeout queue
 *
 * @returns                   the timeout value
 */
PACE2_timestamp pace2_pht_get_timeout(PACE2_pht * pht, u16 queue);

/**
 * @brief Initialize a foreach operation on the hash table. Should be done
 *        before calling @ref pace2_pht_foreach_get_next_element
 *
 * @param[in] pht                       pointer to the hash table
 *
 * @retval    PACE2_PHT_SUCCESS         @copydoc PACE2_pht_return_state::PACE2_PHT_SUCCESS
 * @retval    PACE2_PHT_PARAM_MISSING   @copydoc PACE2_pht_return_state::PACE2_PHT_PARAM_MISSING
 */
PACE2_pht_return_state pace2_pht_foreach_init(PACE2_pht * pht);

/**
 * @brief Get the next hash table element.
 *
 * @param[in] pht                           pointer to the hash table
 * @param[out] unique_key                   pointer to the hash table element
 *                                          The value is only valid on @p PACE2_PHT_MORE_ELEMENTS return.
 * @param[out] user_buffer                  pointer to the user pointer
 *                                          The value is only valid on @p PACE2_PHT_MORE_ELEMENTS return.
 * @param[out] last_timestamp               timestamp of the current element
 *                                          The value is only valid on @p PACE2_PHT_MORE_ELEMENTS return.
 *
 * @retval     PACE2_PHT_SUCCESS            the foreach operation is done
 * @retval     PACE2_PHT_PARAM_MISSING      @copydoc PACE2_pht_return_state::PACE2_PHT_PARAM_MISSING
 * @retval     PACE2_PHT_MORE_ELEMENTS      @copydoc PACE2_pht_return_state::PACE2_PHT_MORE_ELEMENTS
 */
PACE2_pht_return_state pace2_pht_foreach_get_next_element(
    PACE2_pht * pht, u8 ** unique_key, void ** user_buffer, PACE2_timestamp * last_timestamp);

/**
 * @brief Get the size of a hash table.
 *
 * @param[in] elements         number of elements
 * @param[in] unique_key_size  size of the key element
 * @param[in] user_buffer_size size of the data per key element
 *
 * @note This computation assumes exactly 1 timeout queue.
 *
 * @returns                    the memory size in bytes
 */
unsigned long pace2_pht_calculate_hash_size(u32 elements, u32 unique_key_size, u32 user_buffer_size);

/**
 * @brief Get the size of a hash table.
 *
 * @param[in] elements         number of elements
 * @param[in] unique_key_size  size of the key element
 * @param[in] user_buffer_size size of the data per key element
 * @param[in] timeout_queues   number of timeout queues
 *
 * @returns                    the memory size in bytes
 */
unsigned long pace2_pht_calculate_hash_size_multiple_timeout_queues(u32 elements,
                                                                    u32 unique_key_size,
                                                                    u32 user_buffer_size,
                                                                    u16 timeout_queues);

/**
 * @brief Get the current active timeout queue.
 *
 * @param[in] pht   pointer to the hash table
 *
 * @returns         the queue of the last insert operation
 */
u16 pace2_pht_get_timeout_queue(PACE2_pht * pht);

/**
 * @brief Get the maximum number of usable elements in the hash table.
 *
 * @param[in] pht   pointer to the hash table
 *
 * @returns         number of hash table elements
 */
u32 pace2_pht_number_of_elements(PACE2_pht * pht);

/**
 * @brief Get the number of used elements in the hash table.
 *
 * @param[in] pht   pointer to the hash table
 *
 * @returns         number of used hash table elements
 */
u32 pace2_pht_used_elements(PACE2_pht * pht);

/**
 * @brief Get the maximum number of elements in use at the same time,
 *        ever seen since the initialization of the hash table.
 *
 * @param[in] pht   pointer to the hash table
 *
 * @returns         number of maximum used elements
 */
u32 pace2_pht_maximum_number_of_used_elements(PACE2_pht * pht);

/**
 * @brief Get the raw insert pointer from the last insert operation.
 *
 * This can be used to retrieve the last accessed unique and the last accessed
 * user data pointer.
 *
 * @param[in] pht                   pointer to the hash table
 *
 * @returns                         Pointer to the raw insert element.
 * @retval    NULL                  on failure
 */
PACE2_pht_raw_element * pace2_pht_get_last_inserted_raw_element(PACE2_pht * pht);

/**
 * @brief Get the size of the key element as initialized from
 *        @ref pace2_pht_create
 *
 * @param[in] pht   pointer to the hash table
 *
 * @returns         the key element size
 */
u32 pace2_pht_get_unique_key_size(PACE2_pht * pht);

/**
 * @brief Get the size of the user data as initialized from
 *        @ref pace2_pht_create
 *
 * @param[in] pht   pointer to the hash table
 *
 * @returns         the user data size per element
 */
u32 pace2_pht_get_user_buffer_size(PACE2_pht * pht);

/**
 * @brief Get the timestamp of the oldest entry in the hash table.
 *
 * @param[in] pht   pointer to the hash table
 *
 * @returns         timestamp of the oldest entry or 0 if there is none
 */
PACE2_timestamp pace2_pht_get_timestamp_of_oldest_entry(PACE2_pht * pht);

/**
 * @brief Get the pointer to the unique key for the given user data pointer.
 *
 * @param[in] pht          pointer to the hash table
 * @param[in] user_buffer  pointer to the user data, must be a valid element
 *
 * @returns                The pointer to the unique key
 * @retval    NULL         on failure
 */
void const * pace2_pht_get_unique_key_from_user_buffer(PACE2_pht * pht, void const * user_buffer);


/**
 * @brief Get the last timestamp of the hash table entry.
 *
 * @param[in] pht   pointer to the hash table
 * @param[in] user_buffer  pointer to the user data, must be a valid element
 *
 * @returns         timestamp of the last update or 0 if there is none
 */
PACE2_timestamp pace2_pht_get_last_timestamp_from_user_buffer(PACE2_pht * pht, void const * user_buffer);

/**
 * @brief Reserve elements in the hash table.
 *
 * It is possible to check whether enough space is available for new elements
 * so that the next insert is guaranteed to succeed without removing an element.
 * This is useful when processing needs to be done before removing the elements.
 * If an element is overwritten by the insert operation the memory contents of the
 * previous element are no longer accessible. By reserving the required amount of
 * elements before the insert it is still possible to access the memory of the
 * elements that will be removed.
 * If the pht config variable @ref PACE2_pht_config.max_memory_size was set (to a value greater 0),
 * the value has to be high enough to guarantee a reservation.
 *
 * This function adds the number of elements to be reserved to an internal reservation list.
 * The function @ref pace2_pht_get_next_element_to_remove has to be called to
 * remove elements in the case this functions returns with the value @ref PACE2_PHT_OUT_OF_MEMORY
 * to have enough space for new elements.
 *
 * @param[in] pht                       pointer to the hash table
 * @param[in] elements                  number of elements to be reserved
 *
 * @retval    PACE2_PHT_SUCCESS         enough free element space is available
 * @retval    PACE2_PHT_PARAM_MISSING   @copydoc PACE2_pht_return_state::PACE2_PHT_PARAM_MISSING
 * @retval    PACE2_PHT_OUT_OF_MEMORY   @copydoc PACE2_pht_return_state::PACE2_PHT_OUT_OF_MEMORY
 */
PACE2_pht_return_state pace2_pht_reserve_elements(PACE2_pht * pht, u32 elements);

/**
 * @brief Get the next elements which will be removed.
 *
 * Useful when iterating over removed elements is needed.
 *
 * @param[in]  pht              pointer to the hash table
 * @param[out] call_again       1 if more elements will be removed
 *                              0 otherwise (optional)
 * @param[out] removal_reason   The reason for the removal (optional)
 *
 * @returns                     Pointer to the user data of the next removed element.
 * @retval     NULL             If no elements are left or an error occurred.
 */
void * pace2_pht_get_next_element_to_remove(
    PACE2_pht * pht, u8 * call_again, PACE2_pht_removal_reason * removal_reason);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_PHT_CORE_H */
