/** @file      pace2_pif.h
  * @brief     PACE 2 Profiling Information Framework (PIF) public API.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_PIF_H
#define PACE2_PIF_H

#include "pace2_basic_def.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef struct pace2_pif_component_info {
    char const * name;        /**< Name of the component. */
    char const * description; /**< Description of the component (optional). */
} PACE2_pif_component_info;

typedef struct pace2_pif_counter_info {
    u64 const * counter;      /**< Pointer to registered counter. */
    char const * name;        /**< Name of the counter. */
    char const * description; /**< Description of the counter (optional). */
} PACE2_pif_counter_info;

typedef struct pace2_pif_component PACE2_pif_component;

#define PACE2_PIF_COMPONENT_NO_HANDLE ((PACE2_pif_component *)0)
#define PACE2_PIF_COMPONENT_ROOT_HANDLE ((PACE2_pif_component *)-1)

typedef struct pace2_pif_counter PACE2_pif_counter;

#define PACE2_PIF_COUNTER_NO_HANDLE ((PACE2_pif_counter *)0)
#define PACE2_PIF_COUNTER_ROOT_HANDLE ((PACE2_pif_counter *)-1)

/**
 *  @brief  Returns the first child component of a registered component.
 *
 *  @param[in]  pace2_module       Pointer to the PACE 2 module.
 *  @param[in]  component_handle   Handle of the registered (parent) component.
 *                                 Can be PACE2_PIF_COMPONENT_ROOT_HANDLE but NOT PACE2_PIF_COMPONENT_NO_HANDLE.
 *
 *  @return  Handle of the child component.
 *  @retval  PACE2_PIF_COMPONENT_NO_HANDLE  if no child component found or argument is invalid.
 */
PACE2_pif_component * pace2_pif_component_get_first_child(PACE2_module * const pace2_module,
                                                          PACE2_pif_component * component_handle);

/**
 *  @brief  Returns the component at the same level which follows a given component.
 *
 *  @param[in]  pace2_module       Pointer to the PACE 2 module.
 *  @param[in]  component_handle   Handle of the registered component. Has to be a valid component handle.
 *
 *  @return  ID of the next component.
 *  @retval  PACE2_PIF_COMPONENT_NO_HANDLE  if no component follows or on invalid arguments.
 */
PACE2_pif_component * pace2_pif_component_get_next_sibling(PACE2_module * const pace2_module,
                                                           PACE2_pif_component * component_handle);

/**
 *  @brief  Returns information about a registered component.
 *
 *  @param[in]   pace2_module       Pointer to the PACE 2 module.
 *  @param[in]   component_handle   Component handle to find information for.
 *                                  PACE2_PIF_COMPONENT_NO_HANDLE is not valid here.
 *  @param[out]  info               Pointer to storage for information.
 *
 *  @return Error status to signal success or failure.
 *  @retval PACE2_SUCCESS   @copydoc PACE2_return_state::PACE2_SUCCESS
 *  @retval PACE2_FAILURE   @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_pif_component_get_info(PACE2_module * const pace2_module,
                                                PACE2_pif_component * component_handle,
                                                PACE2_pif_component_info * const info);

/**
 *  @brief  Returns next counter.
 *
 *  @param[in] pace2_module       Pointer to the PACE 2 module.
 *  @param[in] component_handle   Component handle for which a counter has to be returned. If the input value for
 *                                @p counter_handle is valid @p component_handle is ignored.
 *  @param[in] counter_handle     Handle of current counter.
 *                                Use PACE2_PIF_COUNTER_ROOT_HANDLE to start at first counter for a given valid
 *                                component handle.
 *
 *  @return  Handle of next counter.
 *  @retval  PACE2_PIF_COUNTER_NO_HANDLE  if no counter follows.
 */
PACE2_pif_counter * pace2_pif_counter_get_next(PACE2_module * const pace2_module,
                                               PACE2_pif_component * component_handle,
                                               PACE2_pif_counter * counter_handle);

/**
 *  @brief  Returns information about a registered counter.
 *
 *  This function will fail with @ref PACE2_FORBIDDEN if dynamic upgrade is enabled in global config.
 *
 *
 *  @param[in]   pace2_module      Pointer to the PACE 2 module.
 *  @param[in]   counter_handle    Counter handle to find information for.
 *  @param[out]  info              Pointer to storage for information.
 *
 *  @return Error status to signal success or failure.
 *  @retval PACE2_SUCCESS                   @copydoc PACE2_return_state::PACE2_SUCCESS
 *  @retval PACE2_FAILURE                   @copydoc PACE2_return_state::PACE2_FAILURE
 *  @retval PACE2_INVALID_MODULE_POINTER    @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 *  @retval PACE2_PARAM_MISSING             @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 *  @retval PACE2_FORBIDDEN                 @copydoc PACE2_return_state::PACE2_FORBIDDEN
 *
 */
PACE2_return_state pace2_pif_counter_get_info(PACE2_module * const pace2_module,
                                              PACE2_pif_counter * counter_handle,
                                              PACE2_pif_counter_info * const info);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_PIF_H */
