/** @file      pace2_plugins.h
  * @brief     PACE 2 PLUG-INs
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_PLUGINS_H
#define PACE2_PLUGINS_H

/* Includes *********************************************************************/
#ifndef PACE_KERNEL_LIB
#include <stdint.h>
#endif

#include "pace2_basic_def.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Types *****************************************************************/

/**
 * The prefix of the PLUG-INs that are written by IPOQUE GMBH
 */
#define PACE2_PLUGIN_ID_PREFIX_RS 0x00

/**
 * @brief The observer flow limit types which indicates how to interpret the value.
 *
 * @details In the future this enumeration type may be extended.
 *
 */
typedef enum pace2_observer_flow_limit_type {
    PACE2_OBSERVER_FLOW_LIMIT_DISABLED = 0, /*!< Feature is disabled and no limit is applied. */
    PACE2_OBSERVER_FLOW_LIMIT_PACKET_COUNT, /*!< Feature is enabled and packet count is limited per flow. */
} PACE2_observer_flow_limit_type;

/**
 * @brief The observer flow limit including type and value.
 */
typedef struct pace2_observer_flow_limit {
    u64 value;                              /*!< The value of the limit. */
    PACE2_observer_flow_limit_type type;    /*!< The type of the limit. */
} PACE2_observer_flow_limit;

/**
 * @brief Definition of the configuration of the PLUG-INs
 *
 * This configuration struct is needed to load PLUG-INs in PACE 2. See
 * @ref PACE2_config_plugins and @ref PACE2_global_config::plugin.
 */
typedef struct pace2_plugin_configuration
{
    char const * plugin_file;        /*!< The path to the PLUG-IN file. */
    PACE2_plugin_config_type plugin_config_type; /*!< Data type of the configuration data (struct, text or blob). */
    u32    plugin_config_len;        /*!< Length of plugin_config, if type is text or blob. */
    void * plugin_config;            /*!< The pointer to a PLUG-IN specific configuration for the PLUG-IN. This pointer
                                          must be valid for the whole time the corresponding PLUG-IN uses the
                                          configuration structure. Usually this is until the threads are initialized,
                                          but each PLUG-IN could handle this differently. */
    pace2_malloc_t  memory_malloc;   /*!< Function pointer to a memory allocation function, default: pace2_alloc. */
    pace2_free_t    memory_free;     /*!< Function pointer to a memory free function, default: pace2_free. */
    pace2_realloc_t memory_realloc;  /*!< Function pointer to a memory free function, default: pace2_realloc. */
    void *          memory_user_ptr; /*!< PLUG-IN specific user definable pointer. Passed to all memory callback
                                          functions. */
    pace2_logging_callback logging_callback; /*!< Callback function pointer to log messages from
                                                  the PLUG-IN, default: NULL. */
    void * logging_user_ptr;         /*!< PLUG-IN specific user definable pointer, passed to all logging callbacks. */
    PACE2_observer_flow_limit flow_limit;    /*!< The flow limit configuration. Memset this struct to '0' to
                                                  disable all limits for future configuration changes. */

    /* Reserved for future extensions. Do not use! */
    uintptr_t reserved[8];

} PACE2_plugin_configuration;

/* Public Prototypes ************************************************************/

/**
 * A collection of PLUG-IN related basic information.
 */
typedef struct pace2_plugin_info {
    /** The PLUG-IN ID including a prefix and the actual ID. */
    PACE2_plugin_id_struct plugin_id;
    /** PLUG-IN specific version number. */
    u32 plugin_version;
    /** Number of different events this PLUG-IN will throw. */
    u32 number_of_events;
    /** Count in bytes of registered memory per flow. */
    u32 registered_flow_bytes;
    /** The path to the PLUG-IN file. */
    char const * plugin_file;
} PACE2_plugin_info;

/**
 * @brief Denotes the type of functionality a PLUG-IN extension entity provides.
 */
typedef enum {
    PACE2_TRAFFIC_IDENTIFIER,            /*!< Traffic Identifier. */
    PACE2_EVENT_PRODUCER,                /*!< Event Producer. */
    PACE2_NORMALIZATION_FUNCTION,        /*!< Normalization Function. */
} PACE2_extension_entity_type;

/**
 * @brief A mapping between a extension entity ID and its type.
 */
typedef struct pace2_extension_entity_info {
    u32 entity_id;                              /* Unique ID of the extension entity. */
    PACE2_extension_entity_type entity_type;    /* Corresponding type of this extension entity. */
} PACE2_extension_entity_info;

/**
 * @brief Returns the count of registered PLUG-INs.
 *
 * @param[in]  pace2_module       Pointer to the PACE 2 module.
 * @param[out] number_of_plugins  Address to store the number of registered PLUG-IN.
 *
 * @retval PACE2_SUCCESS          @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_FAILURE          @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_announcement_get_number_of_plugins(PACE2_module * pace2_module, u32 * number_of_plugins);

/**
 * @brief Gets the PLUG-IN information of a registered PLUG-IN by index.
 *
 * @param[in]  pace2_module     Pointer to the PACE 2 module.
 * @param[in]  plugin_index     Index of the queried PLUG-IN.
 * @param[out] plugin_info      PLUG-IN information of the queried PLUG-IN. See @ref PACE2_plugin_info
 *
 * @retval PACE2_SUCCESS        @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_FAILURE        @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_announcement_get_plugin_info_by_index(PACE2_module * pace2_module, u32 plugin_index,
                                                               PACE2_plugin_info * plugin_info);

/**
 * @brief Gets the number of extension entities provided by a specific PLUG-IN.
 *
 * @param[in] pace2_module                    PACE2 module struct.
 * @param[in] plugin_id                       The PLUG-IN ID struct including a prefix and the actual ID.
 * @param[out] number_of_entities             Address to store the number of extension entities.
 *
 * @retval PACE2_SUCCESS                      @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_INVALID_MODULE_POINTER       @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval PACE2_PARAM_MISSING                plugin_id parameter must not be NULL.
 * @retval PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS   The PLUG-IN with the given PLUG-IN ID could not be found.
 */
PACE2_return_state pace2_announcement_get_number_of_extension_entities_by_plugin_id(PACE2_module const * pace2_module,
                                                                                    PACE2_plugin_id_struct plugin_id,
                                                                                    u64 * number_of_entities);

/**
 * @brief Gets the number of extension entities provided by a specific PLUG-IN by linearly increasing index.
 *
 * @param[in] pace2_module                    PACE2 module struct.
 * @param[in] plugin_id                       The PLUG-IN ID struct including a prefix and the actual ID.
 * @param[in] entity_index                    The index of the queried entity.
 * @param[out] extension_entity               Address to store the type and ID of the returned extension entity.
 *
 * @retval PACE2_SUCCESS                      @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_INVALID_MODULE_POINTER       @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval PACE2_PARAM_MISSING                plugin_id and extension_entity parameter must not be NULL.
 * @retval PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS   @copydoc PACE2_return_state::PACE2_INDEX_OUTSIDE_ARRAY_BOUNDS
 */
PACE2_return_state pace2_announcement_get_extension_entity_info_by_index(
        PACE2_module const * pace2_module,
        PACE2_plugin_id_struct plugin_id,
        u64 entity_index,
        PACE2_extension_entity_info * extension_entity);

/**
 * @brief Reloads PLUG-INs during runtime.
 *
 * @details This feature is not supported on all architectures.
 *          In case of unsupported architecture PACE2_FEATURE_NOT_AVAILABLE is returned.
 *          This function is only successfully if the reloaded plugin registered the same
 *          observers like the already running one. This function call is only valid if all
 *          created PACE2 threads for the given PACE2 module are inactive. This means especially
 *          that the whole PACE2 module has no pace2 stage call state.
 *
 * @param[in,out]  pace2_module          Pointer to the module structure
 * @param[in]      plugin_config         Pointer to the module plugin configuration structure.
 *
 * @retval  PACE2_SUCCESS                               @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_PARAM_MISSING                         @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval  PACE2_PLUGIN_DLWRAPPER_NOT_SET              @copydoc PACE2_return_state::PACE2_PLUGIN_DLWRAPPER_NOT_SET
 * @retval  PACE2_PLUGIN_DLOPEN_ERROR                   @copydoc PACE2_return_state::PACE2_PLUGIN_DLOPEN_ERROR
 * @retval  PACE2_PLUGIN_REGISTER_FUNCTION_NOT_FOUND    @copydoc PACE2_return_state::PACE2_PLUGIN_REGISTER_FUNCTION_NOT_FOUND
 * @retval  PACE2_PLUGIN_INCOMPATIBLE_ABI_VERSION       @copydoc PACE2_return_state::PACE2_PLUGIN_INCOMPATIBLE_ABI_VERSION
 * @retval  PACE2_PLUGIN_REGISTER_FAILED                @copydoc PACE2_return_state::PACE2_PLUGIN_REGISTER_FAILED
 * @retval  PACE2_PLUGIN_FORBIDDEN_PLUGIN_ID            @copydoc PACE2_return_state::PACE2_PLUGIN_FORBIDDEN_PLUGIN_ID
 * @retval  PACE2_OUT_OF_MEMORY                         @copydoc PACE2_return_state::PACE2_OUT_OF_MEMORY
 * @retval  PACE2_INVALID                               @copydoc PACE2_return_state::PACE2_INVALID
 * @retval  PACE2_FAILURE                               @copydoc PACE2_return_state::PACE2_FAILURE
 * @retval  PACE2_FORBIDDEN                             The number of registered observers differs.
 *                                                      The PLUG-IN can not be reloaded because of the used feature set.
 * @retval  PACE2_FEATURE_NOT_AVAILABLE                 PACE2 does not support the feature.
 */
PACE2_return_state pace2_reload_plugin(PACE2_module * pace2_module,
                                       PACE2_plugin_configuration const * plugin_config);

/**
 * @brief Stops a PLUG-IN from further data processing.
 *
 * @details After calling this function, all callback functions that PLUG-IN had registered won't get called until the
 *          PLUG-IN is resumed again. Except, the on_flow_cleanup callback will still be invoked to avoid memory leaks.
 *          This function can be called at any time after PACE 2 initialization. Calling this function on an already
 *          suspended PLUG-IN will not show any additional effect.
 *
 * @param[in] pace2_module                    Pointer to the PACE 2 module.
 * @param[in] plugin_id                       The PLUG-IN ID struct including a prefix and the actual ID.
 *
 * @retval  PACE2_SUCCESS                     @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_INVALID_MODULE_POINTER      @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval  PACE2_INVALID                     A PLUG-IN using the provided PLUG-IN id was not found.
 * @retval  PACE2_FEATURE_NOT_AVAILABLE       PACE2 does not support the feature.
 */
PACE2_return_state pace2_suspend_plugin(PACE2_module * pace2_module,
                                        PACE2_plugin_id_struct plugin_id);

/**
 * @brief Resume a previously suspended PLUG-IN.
 *
 * @details After calling this function a previously suspended PLUG-IN resumes its data processing, meaning callbacks
 *          are now active again. When this function is called on a not suspended PLUG-IN it won't have any effect on
 *          the PLUG-IN. Note that depending on the type of callback functions the PLUG-IN had registered, callbacks
 *          may not get invoked for flows that had already entered stage 3 or stage 4. This function can be called at
 *          any time after PACE 2 initialization.
 *
 * @param[in] pace2_module                    Pointer to the PACE 2 module.
 * @param[in] plugin_id                       The PLUG-IN ID struct including a prefix and the actual ID.
 *
 * @retval  PACE2_SUCCESS                     @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval  PACE2_INVALID_MODULE_POINTER      @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval  PACE2_INVALID                     A PLUG-IN using the provided PLUG-IN id was not found.
 * @retval  PACE2_FEATURE_NOT_AVAILABLE       PACE2 does not support the feature.
 */
PACE2_return_state pace2_resume_plugin(PACE2_module * pace2_module,
                                       PACE2_plugin_id_struct const plugin_id);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_PLUGINS_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
