/** @file      pace2_s1_preparing.h
  * @brief     PACE 2 preparing stage definitions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_S1_PREPARING_H
#define PACE2_S1_PREPARING_H

#ifdef __KERNEL__
#include <linux/in.h>
#elif !defined(_WIN32) && !(defined(__OCTEON__) && !defined(__linux__))
#include <netinet/in.h>
#endif // __KERNEL__

#ifdef _WIN32
#include <winsock2.h>
#include <in6addr.h>
#endif /* _WIN32 */

#include "pace2_basic_def.h"
#include "pace2_protocols.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Types *****************************************************************/

/**
 * @brief Supported frame type of stage 1 processing.
 */
typedef enum {
    PACE2_S1_L2 = 0,                    /*!< Generic layer 2 frame. */
    PACE2_S1_L3                         /*!< Generic layer 3 frame. */
} PACE2_s1_input_frame_type;

/**
 * @brief Return state of stage 1 processing function.
 */
typedef enum {
    PACE2_S1_SUCCESS = 0,               /*!< Processing was successful. Includes support for TSO packets. */
    PACE2_S1_PARAM_MISSING,             /*!< Function parameter is missing. */
    PACE2_S1_UNSUPPORTED_FRAME,         /*!< The given frame type is not supported. */
    PACE2_S1_DEFRAG_WRONG_CHECKSUM,     /*!< Check of the checksum failed during defragmentation.
                                             The value will be returned only, if the @ref
                                             PACE2_s1_defrag_params::checksum "checksum handling" of the
                                             defragmentation is configured as
                                             IPOQUE_DEFRAG_CHECKSUM_DISCARD_WRONG_IP_CHECKSUM.  */
    PACE2_S1_DEFRAG_ONGOING,            /*!< Defragmentation is still running.
                                             No further processing should be done for this packet. */
    PACE2_S1_HEADER_CHECK_FAILED,       /*!< A length check of any packet stack header failed. This also happens
                                             for all non-first IP fragments in case IP defragmentation is disabled or not supported.
                                             No further processing should be done for this packet. */
    PACE2_S1_FAILURE,                   /*!< Processing failed. */
    PACE2_S1_TRUNCATED_PACKET,          /*!< Processed a truncated packet. This can be an IP packet with length smaller than IP header
                                             total length field or a first IP fragment in case IP defragmentation is disabled or not supported.
                                             In any case the packet stack is properly filled and processing can be continued for this packet.
                                             If a truncated packet does not contain enough information such that processing could be continued,
                                             PACE2_S1_HEADER_CHECK_FAILED is returned instead. */
} PACE2_s1_return_state;


/**
 * @brief   Decapsulated protocol types, which are stored in pace2_osi_layer_2_info_data structure
 *          (in pace2_decaps_protocol_enum next variable).
 */
typedef enum pace2_decaps_protocol_enum {
    PACE2_NEXT_UNKNOWN = 0,
    PACE2_NEXT_L2_ETHERNET,
    PACE2_NEXT_L3_IPv4,
    PACE2_NEXT_L3_IPv6,
    PACE2_NEXT_L4_TCP,
    PACE2_NEXT_L4_UDP,
    PACE2_NEXT_L4_OTHER,
    PACE2_NEXT_L7,
    PACE2_NEXT_L4_SCTP,
} pace2_decaps_protocol_enum_t;

/**
 * @brief   Maximum number of VLAN tags in packet, supported by decapsulation function.
 *
 * @deprecated This macro will be removed. Please use @ref PACE2_MAX_L2_TAGS.
 */
#define PACE2_MAX_VLAN_LABELS     5

/**
 * @brief   Maximum number of MPLS tags in packet, supported by decapsulation function.
 *
 * @deprecated This macro will be removed. Please use @ref PACE2_MAX_L2_TAGS.
 */
#define PACE2_MAX_MPLS_LABELS     5

/**
 * @brief   Maximum number of Layer 2 tags in packet, supported by decapsulation function.
 */
#define PACE2_MAX_L2_TAGS   10

/**
 * @brief   Maximum number of processed protocols in a packet, which is calculated as follows:
 *          - one protocol for each l2 tag
 *          - PACE2_PROTOCOLS_ETHERNET before handling the l2 tags, as we are handling ethernet types
 *          - possibly one l3 protocol after handling the l2 tags
 */
#define PACE2_MAX_PROCESSED_PROTOCOLS   PACE2_MAX_L2_TAGS + 2

/**
 * @brief   Ethernet Address Length
 */
#define PACE2_ETH_ADD_LEN   6

/**
 * @brief   Structure contains VLAN information
 */
typedef struct pace2_vlan_tag {
    u8 priority_code_point;
    u8 drop_eligible_indicator;
    u16 identifier;
} pace2_vlan_tag_t;

/**
 * @brief   Structure contains MPLS information
 */
typedef struct pace2_mpls_tag {
    u32 label;
    u8 traffic_class;
    u8 bottom_of_stack;
    u8 time_to_live;
} pace2_mpls_tag_t;

/**
 * @brief   Structure contains IEEE 802.1ah information
 */
typedef struct pace2_ieee8021ah_tag {
    u32 i_tag;
} pace2_ieee8021ah_tag_t;

/**
 * @brief   Structure contains information about a L2 tag
 */
typedef struct pace2_l2_tag {

    /** Source MAC address, only valid if ether_type == IPQ_ETH_RESERVED == 0xFFFF or
     * ether_type == ETH_P_8021AH == 0x88E7 */
    u8 src_mac_address[PACE2_ETH_ADD_LEN];

    /** Destination MAC address, only valid if ether_type == IPQ_ETH_RESERVED == 0xFFFF or
     * ether_type == ETH_P_8021AH == 0x88E7 */
    u8 dst_mac_address[PACE2_ETH_ADD_LEN];

    /** Specifies whether the union contains
     * VLAN data (ether_type == ETH_P_8021Q == 0x8100, ether_type == ETH_P_8021AD == 0x88A8),
     * MPLS data (ether_type == ETH_P_MPLS_UC == 0x8847) or
     * IEEE 802.1ah data (ether_type == ETH_P_8021AH == 0x00E7) or is
     * Ethernet (ether_type == IPQ_ETH_RESERVED == 0xFFFF).*/
    u16 ether_type;

    union {
        pace2_vlan_tag_t vlan;
        pace2_mpls_tag_t mpls;
        pace2_ieee8021ah_tag_t ieee8021ah;
    };
} pace2_l2_tag_t;

/**
 * @brief   Structure contains L2 information about the header of currently processed packet.
 */
typedef struct pace2_osi_layer_2_info_data {
    u16 offset;             /*!< Header offset to skip currently supported ethertypes and tags in the packet header */
    u16 length;                             /*!< Complete packet length minus offset */
    enum pace2_decaps_protocol_enum next;   /*!< Stores the next header type to be processed during decapsulation. */

    u8 tag_count;                           /*!< Number of L2 tags */
    pace2_l2_tag_t tags[PACE2_MAX_L2_TAGS]; /*!< L2 tag values from lowest to highest nesting */

    u8 processed_protocol_count; /*!< Number of protocols processed by pace2_detection_get_l2_data_int */

    /**
     * ATTENTION: Within the next DU break release, this member will be replaced by:
     * l2_processed_protocol_t processed_protocols[PACE2_MAX_PROCESSED_PROTOCOLS];
     * l2_processed_protocol_t is a new struct that holds both offset and name of a processed layer 2 protocol
     */
    PACE2_protocols processed_protocols[PACE2_MAX_PROCESSED_PROTOCOLS]; /*!< Processed layer 2 protocols */
} pace2_osi_layer_2_info_data_t;


/* Public Prototypes ************************************************************/

/**
 * @brief Performs stage 1 processing of a single packet frame.
 *
 * The function initializes the packet descriptor of a corresponding packet.
 * The packet framing will be generated by this function and if configured,
 * the IP defragmentation will be performed.
 *
 * @param[in]   pace2_module        Pointer to the PACE 2 module.
 * @param[in]   thread_ID           The ID of the thread currently processing.
 * @param[in]   timestamp           Packet timestamp value. The unit is defined by the config setting
 *                                  @ref PACE2_config_general::io_timestamp_resolution.
 * @param[in]   frame_data_ptr      Pointer to the frame payload data.
 * @param[in]   frame_data_len      Length of the frame payload data.
 * @param[in]   frame_type          Type of the frame.
 * @param[in,out] packet_descriptor Pointer to a packet descriptor structure, which will be used for data storage.
 * @param[in]   packet_user_data    Pointer to user defined packet data. This is optional.
 * @param[in]   packet_user_data_len Length of the user defined packet data.
 * @retval PACE2_S1_SUCCESS         @copydoc PACE2_s1_return_state::PACE2_S1_SUCCESS
 * @retval PACE2_S1_PARAM_MISSING   @copydoc PACE2_s1_return_state::PACE2_S1_PARAM_MISSING
 * @retval PACE2_S1_UNSUPPORTED_FRAME       @copydoc PACE2_s1_return_state::PACE2_S1_UNSUPPORTED_FRAME
 * @retval PACE2_S1_DEFRAG_WRONG_CHECKSUM   @copydoc PACE2_s1_return_state::PACE2_S1_DEFRAG_WRONG_CHECKSUM
 * @retval PACE2_S1_DEFRAG_ONGOING  @copydoc PACE2_s1_return_state::PACE2_S1_DEFRAG_ONGOING
 * @retval PACE2_S1_HEADER_CHECK_FAILED     @copydoc PACE2_s1_return_state::PACE2_S1_HEADER_CHECK_FAILED
 * @retval PACE2_S1_TRUNCATED_PACKET        @copydoc PACE2_s1_return_state::PACE2_S1_TRUNCATED_PACKET
 * @retval PACE2_S1_FAILURE         @copydoc PACE2_s1_return_state::PACE2_S1_FAILURE
 *                                  The given thread ID is invalid or the thread instance is not initialized.
 */
PACE2_s1_return_state pace2_s1_process_packet(PACE2_module * const pace2_module,
                                              int thread_ID,
                                              PACE2_timestamp timestamp,
                                              const void *frame_data_ptr,
                                              u16 frame_data_len,
                                              PACE2_s1_input_frame_type frame_type,
                                              PACE2_packet_descriptor *packet_descriptor,
                                              const void *packet_user_data,
                                              u16 packet_user_data_len);

/**
 * @brief   Function detects correct offset to skip L2 header block of the packet.
 *
 * @param[in]       frame_data_ptr                  Pointer to the current data packet
 * @param[in]       current_accessible_data_length  Accessible L2 data length (L2 header + L2 payload).
 *                                                  @p frame_data_ptr is only accessed and examined up to this length.
 * @param[in]       current_complete_data_length    Complete L2 data length (L2 header + L2 payload).
 *                                                  Used to determine the correct L3 packet length even if not all data
 *                                                  is accessible as can be specified by @p current_accesible_data_length.
 * @param[in]       type                            Type of the packet's protocol (currently only ethernet supported).
 * @param[in,out]   l2data                          Pointer to the structure containing info about offset and
 *                                                  packet length without header block. The result of the function's work saved here.
 *
 * @return          [0]: l2data and decap_data are valid.
 *                  [< 0]: they are NOT valid:
 *                          [-1]: incorrect packet length (unexpected reach of the packet's end);
 *                          [-3]: l2data length is smaller than ethernet header (unsupported);
 *                          [-4]: unsupported protocol type detected;
 *                          [-6]: packet type is not ethernet.
 *                  [1]: requires more data (current_accessible_data_length too small).
 */
int pace2_detection_get_l2_data(const u8 * frame_data_ptr,
                                const u16 current_accessible_data_length,
                                const u16 current_complete_data_length,
                                enum pace2_decaps_protocol_enum type,
                                struct pace2_osi_layer_2_info_data * l2data);

/**
 * @brief   Function detects correct offset to skip L2 header block of the packet starting
 *          from IP header
 *
 * @param[in]       frame_data_ptr                  Pointer to the current IP header
 * @param[in]       current_accessible_data_length  Accessible L2 data length (L2 header + L2 payload).
 *                                                  @p frame_data_ptr is only accessed and examined up to this length.
 * @param[in]       current_complete_data_length    Complete L2 data length (L2 header + L2 payload).
 *                                                  Used to determine the correct L3 packet length even if not all data
 *                                                  is accessible as can be specified by @p current_accesible_data_length.
 * @param[in]       type                            Type of the packet's protocol (currently only ethernet supported).
 * @param[in,out]   l2data                          Pointer to the structure containing info about offset and
 *                                                  packet length without header block. The result of the function's work saved here.
 *
 * @return          [0]: l2data and decap_data are valid.
 *                  [< 0]: they are NOT valid:
 *                          [-1]: incorrect packet length (unexpected reach of the packet's end);
 *                          [-3]: l2data length is smaller than ethernet header (unsupported);
 *                          [-4]: unsupported protocol type detected;
 *                          [-6]: packet type is not ethernet.
 *                  [1]: requires more data (current_accessible_data_length too small).
 */
int pace2_detection_get_inner_l2_data_from_ip(const u8 * frame_data_ptr,
                                              const u16 current_accessible_data_length,
                                              const u16 current_complete_data_length,
                                              enum pace2_decaps_protocol_enum type,
                                              struct pace2_osi_layer_2_info_data * l2data );

/**
 *  @brief  Profiling of defragmentation engine
 *
 *  @param[in]      pace2                    Pointer to the PACE 2 module, not NULL.
 *  @param[in]      thread_ID                The ID of the thread currently processing.
 *  @param[in,out]  used_elements            number of currently used elements, not NULL.
 *  @param[in,out]  allocated_elements       number of currently allocated elements, not NULL.
 *
 *  @retval         PACE2_S1_SUCCESS        @copydoc PACE2_s1_return_state::PACE2_S1_SUCCESS
 *  @retval         PACE2_S1_PARAM_MISSING  @copydoc PACE2_s1_return_state::PACE2_S1_PARAM_MISSING
 *  @retval         PACE2_S1_FAILURE        @copydoc PACE2_s1_return_state::PACE2_S1_FAILURE
 *                                          The given thread ID is invalid or the thread
 *                                          instance is not initialized.
 */
PACE2_s1_return_state pace2_detection_defrag_status(PACE2_module * pace2,
                                                     u32 thread_ID,
                                                     u64 * used_elements,
                                                     u64 * allocated_elements);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_S1_PREPARING_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
