/** @file      pace2_s2_reordering.h
  * @brief     PACE 2 reordering definitions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_S2_REORDERING_H
#define PACE2_S2_REORDERING_H

#include "pace2_basic_def.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Types *****************************************************************/

/**
 * @brief Return state of stage 2 processing functions.
 */
typedef enum {
    PACE2_S2_SUCCESS = 0,                                    /*!< Processing was successful. */
    PACE2_S2_FAILURE,                                        /*!< Processing failed. Generic error during PARO . No further processing should be done for this packet. */
    PACE2_S2_FAILURE_PARO_FLOW_LOOKUP_FAILED,                /*!< Processing failed. Flow lookup failed. No further processing should be done for this packet. */
    PACE2_S2_FAILURE_PARO_INSERT_PACKET_INVALID_PACKET_LEN,  /*!< Processing failed. Packet length invalid. No further processing should be done for this packet. */
    PACE2_S2_FAILURE_FRAMING_STACK_SIZE,                     /*!< Processing failed. The framing stack is to small. No further processing should be done for this packet. */
    PACE2_S2_FAILURE_INITIALIZE,                             /*!< Processing failed. Not initialized variables. */
} PACE2_s2_return_state;

/* Public Prototypes ************************************************************/

/**
 * @brief Performs stage 2 processing of a single packet.
 *
 * TCP Packets might optionally be reordered by the packet reordering engine.
 * The engine will buffer out-of-order packets until the missing packets arrive or a
 * specific timeout occurs.
 *
 * @param[in]   pace2_module        Pointer to the PACE 2 module.
 * @param[in]   thread_ID           The ID of the thread currently processing.
 * @param[in]   packet_descriptor   The packet descriptor given by pace2_stage2_get_next_packet.
 * @retval PACE2_S2_SUCCESS         @copydoc PACE2_s2_return_state::PACE2_S2_SUCCESS
 * @retval PACE2_S2_FAILURE         @copydoc PACE2_s2_return_state::PACE2_S2_FAILURE
 *                                  The given thread ID is invalid or the thread instance is not initialized.
 */
PACE2_s2_return_state pace2_s2_process_packet(PACE2_module * const pace2_module,
                                              int thread_ID,
                                              PACE2_packet_descriptor *packet_descriptor );

/**
 * @brief Update timestamp of the packet reordering engine.
 *
 * The function updates the timestamp of the packet reordering engine.
 * If @ref INTERNAL flow tracking is used, the timestamp of the polling hash table
 * will be updated too.
 *
 * @param[in]   pace2_module        Pointer to the PACE 2 module.
 * @param[in]   thread_ID           The ID of the thread currently processing.
 * @param[in]   timestamp           Timestamp value.
 * @returns     0, if processing was successful.
 * @returns     1, if the given thread ID is invalid or the thread instance is not initialized.
 */
int pace2_s2_update_timestamp(PACE2_module * const pace2_module, int thread_ID, PACE2_timestamp timestamp);

/**
 * @brief Get next packet for processing from the reordering engine.
 *
 * The function returns the next reordered packet for further processing.
 *
 * @param[in]   pace2_module        Pointer to the PACE 2 module.
 * @param[in]   thread_ID           The ID of the thread currently processing.
 * @returns     Packet descriptor pointer to next packet for further processing on success, else NULL.
 */
PACE2_packet_descriptor * pace2_s2_get_next_packet(PACE2_module * const pace2_module, int thread_ID);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_S2_REORDERING_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
