/** @file      pace2_s4_decoding.h
  * @brief     PACE 2 decoding definitions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_S4_DECODING_H
#define PACE2_S4_DECODING_H

#include "pace2_basic_def.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Types *****************************************************************/

/**
 * @brief Return state of stage 4 processing functions.
 */
typedef enum {
    PACE2_S4_SUCCESS = 0,           /*!< Processing was successful. */
    PACE2_S4_PARAM_MISSING,         /*!< Function parameter is missing. */
    PACE2_S4_INPUT_MISSING,         /*!< Input data is missing. */
    PACE2_S4_OUT_OF_MEMORY,         /*!< Memory allocation failed. */
    PACE2_S4_NOT_INITIALIZED,       /*!< Stage 4 is not initialized. */
    PACE2_S4_FAILURE                /*!< Processing failed. */
} PACE2_s4_return_state;

/* Public Prototypes ************************************************************/

/**
 * @brief Performs stage 4 processing of a single packet.
 *
 * The function performs packet based protocol decoding. It provides advanced and class metadata.
 * The stage 4 part of @ref PACE2_global_config has to be configured for packet based decoding.
 *
 * @param[in]   pace2_module        Pointer to the PACE 2 module.
 * @param[in]   thread_ID           The ID of the thread currently processing.
 * @param[in]   packet_descriptor   The packet descriptor given by @ref pace2_s2_get_next_packet.
 * @param[in]   classifier_token    Contains required data for stage 4 processing.
 *                                  Can be left out (set to NULL) if stage 4 processing is done by the PACE2_module as lower stages
 *                                  and no other flow has been processed in between.
 * @param[out]  return_event_mask   The mask to be filled with event types created during the current call.
 * @retval PACE2_S4_SUCCESS         @copydoc PACE2_s4_return_state::PACE2_S4_SUCCESS
 * @retval PACE2_S4_PARAM_MISSING   @copydoc PACE2_s4_return_state::PACE2_S4_PARAM_MISSING
 *                                  The given pace2_module, thread_id or stream_descriptor is missing or invalid.
 * @retval PACE2_S4_OUT_OF_MEMORY   @copydoc PACE2_s4_return_state::PACE2_S4_OUT_OF_MEMORY
 *                                  @arg Memory allocation for stage 4 flow data failed. This value will be returned
 *                                  only in cases the @ref EXTERNAL flow tracking is used.
 * @retval PACE2_S4_NOT_INITIALIZED @copydoc PACE2_s4_return_state::PACE2_S4_NOT_INITIALIZED
 *                                  @arg Stage 4 is not configured for packet based decoding.
 * @retval PACE2_S4_FAILURE         @copydoc PACE2_s4_return_state::PACE2_S4_FAILURE
 *                                  Decoding of the given packet failed.
 */
PACE2_s4_return_state pace2_s4_process_packet(PACE2_module * const pace2_module,
                                              int thread_ID,
                                              PACE2_packet_descriptor *packet_descriptor,
                                              const PACE2_classifier_token *classifier_token,
                                              PACE2_bitmask *return_event_mask);

/**
 * @brief Performs stage 4 processing on a data stream.
 *
 * The function performs stream based protocol decoding. It provides advanced and class metadata.
 * The stage 4 part of @ref PACE2_global_config has to be configured for stream based decoding.
 *
 * After the function returned successfully, the number of bytes to be removed from the input data is stored in
 * @ref PACE2_stream_descriptor::stream_remove. That number of bytes has to be removed from the input data before
 * calling the function again.
 *
 * @param[in]     pace2_module        Pointer to the PACE 2 module.
 * @param[in]     thread_ID           The ID of the thread currently processing.
 * @param[in,out] stream_descriptor   The stream descriptor is prefilled by @ref pace2_s4_prepare_stream.
 *                                    Afterwards, the user is required to add the stream data separately.
 * @param[in]     classifier_token    Contains required data for stage 4 processing.
 *                                    Can be left out (set to NULL) if stage 4 processing is done by the PACE2_module
 *                                    as lower stages and no other flow has been processed in between.
 * @param[out]    return_event_mask   The mask to be filled with event types created during the current call.
 *                                    See @ref PACE2_event_groups.
 *
 * @retval PACE2_S4_SUCCESS         @copydoc PACE2_s4_return_state::PACE2_S4_SUCCESS
 *                                  Processing was successful.
 *                                  In this case the number of bytes to be removed from the input data is stored in
 *                                  @ref PACE2_stream_descriptor::stream_remove. That number of bytes has to be removed
 *                                  from the input data before calling the function again.
 * @retval PACE2_S4_PARAM_MISSING   @copydoc PACE2_s4_return_state::PACE2_S4_PARAM_MISSING
*                                   The given pace2_module, thread_id or stream_descriptor is missing or invalid.
 * @retval PACE2_S4_INPUT_MISSING   @copydoc PACE2_s4_return_state::PACE2_S4_INPUT_MISSING
 *                                  @arg No flow data for stage 4 available.
 * @retval PACE2_S4_NOT_INITIALIZED @copydoc PACE2_s4_return_state::PACE2_S4_NOT_INITIALIZED
 *                                  @arg Stage 4 is not configured for stream based decoding.
 * @retval PACE2_S4_FAILURE         @copydoc PACE2_s4_return_state::PACE2_S4_FAILURE
 *                                  Decoding of the input data failed. Discard all input data.
 *
 * @note currently works only using external tracking
 */
PACE2_s4_return_state pace2_s4_process_stream(PACE2_module * const pace2_module,
                                              int thread_ID,
                                              PACE2_stream_descriptor *stream_descriptor,
                                              const PACE2_classifier_token *classifier_token,
                                              PACE2_bitmask *return_event_mask);

/**
 * @brief Prepares a stream descriptor for use in stage 4 from an already existing packet descriptor.
 *
 * @param[in]   pace2_module        Pointer to the PACE 2 module.
 * @param[in]   thread_ID           The ID of the thread currently processing.
 * @param[in]   packet_descriptor   Pointer to an already existing packet descriptor.
 * @param[out]  stream_descriptor   Pointer to the stream descriptor to prepare.
 * @retval PACE2_S4_SUCCESS         @copydoc PACE2_s4_return_state::PACE2_S4_SUCCESS
 * @retval PACE2_S4_PARAM_MISSING   @copydoc PACE2_s4_return_state::PACE2_S4_PARAM_MISSING
 * @retval PACE2_S4_FAILURE         @copydoc PACE2_s4_return_state::PACE2_S4_FAILURE
 *                                  The given thread ID is invalid or the thread instance is not initialized.
 */
PACE2_s4_return_state pace2_s4_prepare_stream(PACE2_module * const pace2_module,
                                              int const thread_ID,
                                              PACE2_packet_descriptor const * const packet_descriptor,
                                              PACE2_stream_descriptor * const stream_descriptor);

/**
 * @brief Obtains all correlated flows for the given session.
 *
 * @param[in]   pace2_module        Pointer to the PACE 2 module.
 * @param[in]   thread_ID           The ID of the thread currently processing.
 * @param[in]   session_id          Session for which the correlated flows are provided.
 * @param[out]  flow_id_count       Number of the correlated flows (to be filled by this function).
 * @param[out]  flow_id_array       Array of the correlated flows (to be filled by this function).
 *
 * @retval PACE2_S4_SUCCESS         @copydoc PACE2_s4_return_state::PACE2_S4_SUCCESS
 * @retval PACE2_S4_PARAM_MISSING   @copydoc PACE2_s4_return_state::PACE2_S4_PARAM_MISSING
 * @retval PACE2_S4_FAILURE         @copydoc PACE2_s4_return_state::PACE2_S4_FAILURE
 *                                  The given thread ID is invalid or the thread instance is not initialized.
 */
PACE2_s4_return_state pace2_s4_get_flow_ids_for_session(PACE2_module * const pace2_module,
                                                        int const thread_ID,
                                                        u64 const session_id,
                                                        u8 * const flow_id_count,
                                                        u64 ** const flow_id_array);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_S4_DECODING_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
