/** @file      pace2_serialization.h
  * @brief     Event serialization related API
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_SERIALIZATION_H
#define PACE2_SERIALIZATION_H

/* Includes ***********************************************************************************************************/
#include "pace2_basic_def.h"
#include "pace2_event_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Types *******************************************************************************************************/

/* @brief Defines all available serialization types */
typedef enum pace2_serialization_type {
    PACE2_SERIALIZATION_JSON, /*!< JSON event serialization format. */
} PACE2_serialization_type;

struct pace2_serialization_parameters_json {
    u8 disable_buffers;       /*!< Disable the output of buffer pointers to save memory, affects structure members with the
                                   name 'ptr', 'data_ptr', 'buffer' and 'oid_part'. */
    u8 event_name_as_value;   /*!< Add the event name as value in the struct rather than as key of the struct. */
    u8 short_event_name;      /*!< Use short event name without PACE2_ prefix. */
    u8 flat_output;           /*!< Output as flat json format */
};

/* Public Prototypes **************************************************************************************************/

/**
 * @brief Serialize an event to a given buffer for a maximum given size.
 *
 * @param[in]       pace2_module  Pointer to the pace 2 module.
 * @param[in]       thread_ID     The unique thread identifier.
 * @param[in]       event         Pointer to the event, which should be serialized.
 * @param[out]      buffer        Pointer to a valid byte buffer where the event will be written in.
 * @param[out]      output_len    The actual length of the serialized data.
 * @param[in]       type          The chosen serialization type.
 * @param[in]       parameter     Additional parameter for the serialization, can be NULL.
 *
 * @returns The function returns @ref PACE2_SUCCESS on success,
 *                               @ref PACE2_OUT_OF_MEMORY when the buffer length is too small,
 *                               @ref PACE2_EVENT_QUEUE_EMPTY when the internal event queue is empty,
 *                               @ref PACE2_FAILURE otherwise.
 */
PACE2_return_state pace2_event_serialize(PACE2_module * const pace2_module, int const thread_ID,
                                         PACE2_event const * const event, PACE2_byte_buffer * buffer, u32 * output_len,
                                         PACE2_serialization_type const type, void * const parameter);

/**
 * @brief Serialize the next event to a given buffer for a maximum given size.
 *
 * @param[in]       pace2_module  Pointer to the pace 2 module.
 * @param[in]       thread_ID     The unique thread identifier.
 * @param[out]      buffer        Pointer to a valid byte buffer where the event will be written in.
 * @param[out]      output_len    The actual length of the serialized data.
 * @param[in]       type          The chosen serialization type.
 * @param[in]       parameter     Additional parameter for the serialization, can be NULL.
 *
 * @returns The function returns @ref PACE2_SUCCESS          on success,
 *                               @ref PACE2_OUT_OF_MEMORY    when the buffer length is too small,
 *                               @ref PACE2_FAILURE          otherwise.
 */
PACE2_return_state pace2_get_next_event_serialized(PACE2_module * const pace2_module, int const thread_ID,
                                                   PACE2_byte_buffer * const buffer, u32 * output_len,
                                                   PACE2_serialization_type const type, void * const parameter);

/**
 * @brief Serialize the next n events to a given buffer for a maximum given size.
 *
 * @param[in]       pace2_module    Pointer to the pace 2 module.
 * @param[in]       thread_ID       The unique thread identifier.
 * @param[in]       num_events      The number of events which should be serialized.
 * @param[out]      received_events The number of received events in the byte buffer.
 * @param[out]      buffer          Pointer to a valid byte buffer where the events will be written in.
 * @param[out]      output_len      The actual length of the serialized data.
 * @param[in]       type            The chosen serialization type.
 * @param[in]       parameter       Additional parameter for the serialization, can be NULL.
 *
 * @returns The function returns @ref PACE2_SUCCESS          on success,
 *                               @ref PACE2_OUT_OF_MEMORY    when the buffer length is too small,
 *                               @ref PACE2_FAILURE          otherwise.
 */
PACE2_return_state pace2_get_next_n_events_serialized(PACE2_module * const pace2_module, int const thread_ID,
                                                      u32 const num_events, u32 * const received_events,
                                                      PACE2_byte_buffer * const buffer, u32 * output_len,
                                                      PACE2_serialization_type const type, void * const parameter);

/**
 * @brief Serialize the next n events to a given array of buffers for a maximum given size.
 *
 * @param[in]       pace2_module    Pointer to the pace 2 module.
 * @param[in]       thread_ID       The unique thread identifier.
 * @param[in]       num_events      The number of events which should be serialized.
 * @param[out]      received_events The number of received events in the byte buffer.
 * @param[out]      buffer_array    A byte buffer array where the events will be written in.
 * @param[out]      output_lengths  Array of length for the byte buffer_array.
 * @param[in]       type            The chosen serialization type.
 * @param[in]       parameter       Additional parameter for the serialization, can be NULL.
 *
 * @returns The function returns @ref PACE2_SUCCESS          on success,
 *                               @ref PACE2_OUT_OF_MEMORY    when the buffer length is too small,
 *                               @ref PACE2_FAILURE          otherwise.
 */
PACE2_return_state pace2_get_next_n_events_serialized_array(PACE2_module * const pace2_module, int const thread_ID,
                                                            u32 const num_events, u32 * received_events,
                                                            PACE2_byte_buffer buffer_array[], u32 output_lengths[],
                                                            PACE2_serialization_type const type,
                                                            void * const parameter);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_SERIALIZATION_H */
/* EOF ****************************************************************************************************************/
