/** @file      pace2_status_classification.h
  * @brief     PACE 2 module wrapper for classification status functions.
  * @copyright Copyright (C) 2014 by ipoque GmbH, a Rohde & Schwarz company. All rights reserved.
  *            See also the DISCLAIMER file that is part of your delivery. */

#ifndef PACE2_STATUS_CLASSIFICATION_H
#define PACE2_STATUS_CLASSIFICATION_H

/* Includes *********************************************************************/
#include "pace2_basic_def.h"
#include "event_types/classification_events.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Public Types *****************************************************************/

/**
 * @brief Return state of classification status functions.
 */
typedef enum {
    PACE2_CLASS_SUCCESS = 0,            /**< Function returned successfully, returned data is valid. */
    PACE2_CLASS_PARAM_MISSING,          /**< Function parameter is missing. */
    PACE2_CLASS_STAT_UNAVAILABLE,       /**< Status is not available. */
    PACE2_CLASS_UNKNOWN_ERROR,          /**< An unknown error occurred */
    PACE2_CLASS_FAILURE,                /**< The given thread ID is invalid or the thread instance is not initialized. */
} PACE2_class_return_state;

/* @brief Defines all available serialization types */
typedef enum {
    PACE2_CLASS_FINISHED_ALL_PROTOCOLS_EXCLUDED = 0,    /**< All protocols did not match and not detected was set. */
    PACE2_CLASS_FINISHED_PACKET_LIMIT,                  /**< The packet limit was reached and not detected was set. */
    PACE2_CLASS_FINISHED_TIMEOUT,                       /**< The Flow timeout limit was reached and not detected was set. */
    PACE2_CLASS_FINISHED_TCP_FIN_RST,                   /**< The TCP fin/rst occurred and not detected was set. */
} PACE2_class_not_detected_reason;

/* Public Prototypes ************************************************************/

/**
 * @brief Get the detections API version number.
 *
 * This function returns the API version number of the main library.
 * The API version will be increased if the API has been changed.
 *
 * @param[out]  class_event     Pointer to a variable, which is used to store the received data.
 *                              This variable is a event structure of type @ref PACE2_classification_status_event.
 * @retval PACE2_CLASS_SUCCESS          @copydoc PACE2_class_return_state::PACE2_CLASS_SUCCESS
 * @retval PACE2_CLASS_PARAM_MISSING    @copydoc PACE2_class_return_state::PACE2_CLASS_PARAM_MISSING
 */
PACE2_class_return_state pace2_class_get_api_version(PACE2_classification_status_event * const class_event);

/**
 * @brief Get the detections release version number.
 *
 * This function returns the release version number of the library. If dynamic upgrade is enabled,
 * it's the version of the active library, otherwise it's the version of the main library.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  class_event     Pointer to a variable, which is used to store the received data.
 *                              This variable is a event structure of type @ref PACE2_classification_status_event.
 * @retval PACE2_CLASS_SUCCESS          @copydoc PACE2_class_return_state::PACE2_CLASS_SUCCESS
 * @retval PACE2_CLASS_PARAM_MISSING    @copydoc PACE2_class_return_state::PACE2_CLASS_PARAM_MISSING
 * @retval PACE2_CLASS_UNKNOWN_ERROR    @copydoc PACE2_class_return_state::PACE2_CLASS_UNKNOWN_ERROR
 */
PACE2_class_return_state pace2_class_get_version(PACE2_module * const pace2_module,
                                                 int const thread_ID,
                                                 PACE2_classification_status_event * const class_event);

/**
 * @brief Get the classifications license information.
 *
 * This function returns the license information of the classification.
 *
 * @param[in]   pace2_module    Pointer to the PACE 2 module.
 * @param[in]   thread_ID       Identifier of a thread.
 * @param[out]  class_event     Pointer to a variable, which is used to store the received data.
 *                              This variable is a event structure of type @ref PACE2_classification_status_event.
 * @retval PACE2_CLASS_SUCCESS          @copydoc PACE2_class_return_state::PACE2_CLASS_SUCCESS
 * @retval PACE2_CLASS_PARAM_MISSING    @copydoc PACE2_class_return_state::PACE2_CLASS_PARAM_MISSING
 * @retval PACE2_CLASS_STAT_UNAVAILABLE @copydoc PACE2_class_return_state::PACE2_CLASS_STAT_UNAVAILABLE
 *                                      @arg Licensing is disabled.
 * @retval PACE2_CLASS_UNKNOWN_ERROR    @copydoc PACE2_class_return_state::PACE2_CLASS_UNKNOWN_ERROR
 * @retval PACE2_CLASS_FAILURE          @copydoc PACE2_class_return_state::PACE2_CLASS_FAILURE
 */
PACE2_class_return_state pace2_class_get_license(PACE2_module * const pace2_module,
                                                 int const thread_ID,
                                                 PACE2_classification_status_event * const class_event);

/**
 * @brief Get the classifications reason for setting not detected.
 *
 * @param[in]   pace2_module            Pointer to the PACE 2 module.
 * @param[in]   thread_ID               If thread_ID >= 0, The ID of the current thread.
 * @param[in]   flow                    Pointer to the PACE 2 flow.
 * @param[out]  reason                  Enum contains reason for not detected.
 * @retval PACE2_CLASS_SUCCESS          @copydoc PACE2_class_return_state::PACE2_CLASS_SUCCESS
 * @retval PACE2_CLASS_PARAM_MISSING    @copydoc PACE2_class_return_state::PACE2_CLASS_PARAM_MISSING
 * @retval PACE2_CLASS_UNKNOWN_ERROR    @copydoc PACE2_class_return_state::PACE2_CLASS_UNKNOWN_ERROR
 * @retval PACE2_CLASS_STAT_UNAVAILABLE @copydoc PACE2_class_return_state::PACE2_CLASS_STAT_UNAVAILABLE
 * @retval PACE2_CLASS_FAILURE          @copydoc PACE2_class_return_state::PACE2_CLASS_FAILURE
 */
PACE2_class_return_state pace2_class_get_not_detected_reason_from_flow(PACE2_module * const pace2_module,
                                                                       int const thread_ID,
                                                                       PACE2_flow_data * const flow,
                                                                       PACE2_class_not_detected_reason * reason);


#ifdef __cplusplus
}
#endif
#endif /* PACE2_STATUS_CLASSIFICATION_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
