/**********************************************************************************************************************/
/**
 **
 ** @file       pace2_version.h
 ** @brief      PACE 2 version include file.
 **
 ** IPOQUE GMBH RESERVES ALL RIGHTS IN THE PROGRAM AS DELIVERED.
 ** THE PROGRAM OR ANY PORTION THEREOF MAY NOT BE USED OR REPRODUCED IN ANY FORM WHATSOEVER WITHOUT THE WRITTEN
 ** CONSENT OF IPOQUE GMBH EXCEPT AS PROVIDED BY LICENSE.
 ** THIS FILE IS PART OF R&S PACE 2 AND IS PROVIDED BY IPOQUE GMBH "AS IS" AND ANY EXPRESS OR
 ** IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 ** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL IPOQUE GMBH BE LIABLE FOR ANY DIRECT,
 ** INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES.
 **
 ** COPYRIGHT (C) 2014 - 2022 BY IPOQUE GMBH
 **/
/**********************************************************************************************************************/

#ifndef PACE2_VERSION_H
#define PACE2_VERSION_H

#include "pace2_basic_def.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Old PACE2 Version defines */
/** @deprecated Deprecated. Instead, use @ref PACE2_VERSION. */
#define IPOQUE_PACE_VERSION "23.06.30"

/** @deprecated Deprecated. Do not use. */
#define IPOQUE_PACE_VERSION_NUMBER 2300060030

/** @deprecated Deprecated. Do not use. */
#define IPOQUE_PACE_DYNAMIC_UPGRADE_COMPATIBLE_VERSION_NUMBER 2300020024

#define IPOQUE_PACE_API_VERSION 442

/**
 * PACE2 product name as string.
 */
#define PACE2_PRODUCT_NAME "PACE2"

/**
 * PACE2 version as string.
 */
#define PACE2_VERSION "3.0.18"

/**
 * PACE2 major version as number.
 */
#define PACE2_VERSION_MAJOR 3

/**
 * PACE2 minor version as number.
 */
#define PACE2_VERSION_MINOR 0

/**
 * PACE2 patch version as number.
 */
#define PACE2_VERSION_PATCH 18

/**
 * @brief Describes the version of a given currently running PACE 2 module.
 *
 */
struct PACE2_version {
    /**
     * @brief The major PACE 2 version. When comparing to another version of PACE 2, a difference place usually means
     *        that, depending on the used feature set, the integration code has to be adapted and/or recompiled upon
     *        an intended update.
     */
    u16 major_version;

    /**
     * @brief The minor PACE 2 version. When comparing to another version of PACE 2, a difference in this place means
     *        that dynamic upgrade to this version is not possible. As such, the integrating application would need to
     *        be restarted.
     */
    u16 minor_version;

    /** @brief The patch level of the PACE 2 version. */
    u16 patch_version;

    /** @brief A string representation "MAJOR.MINOR.PATCH". */
    char version_string[32];

    /**
     * @brief Same as @ref version_string, but a build ID may be appended, which can be used e.g. to identify nightly
     *        and snapshot builds.
     */
    char build_string[64];
};

/**
 * @brief Get version information about the PACE 2 module.
 *
 * This function returns the release version information of the library. If dynamic upgrade is enabled,
 * it's the version of the active library, otherwise it's the version of the main library.
 *
 * @param[in]   pace2_module      Pointer to the PACE 2 module.
 * @param[in]   thread_ID         The thread ID of the thread, from which this callback was called.
 * @param[out]  version           Pointer to a variable, which is used to store the version.
 *                                This variable is structure of type @ref PACE2_version.
 * @retval PACE2_SUCCESS                    @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_INVALID_MODULE_POINTER     @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval PACE2_PARAM_MISSING              @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval PACE2_FAILURE                    @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_get_version(PACE2_module * const pace2_module,
                                     const int thread_ID,
                                     struct PACE2_version * version);

/**
 * @brief Get version information about the initially loaded PACE 2 module.
 *
 * This function returns the release version information of the main library, i.e. the one initially loaded, in case of
 * enabled dynamic upgrade. Otherwise, it's the same as returned by @ref pace2_get_version .
 *
 * @param[in]   pace2_module      Pointer to the PACE 2 module.
 * @param[in]   thread_ID         The thread ID of the thread, from which this callback was called.
 * @param[out]  version           Pointer to a variable, which is used to store the version.
 *                                This variable is structure of type @ref PACE2_version.
 * @retval PACE2_SUCCESS                    @copydoc PACE2_return_state::PACE2_SUCCESS
 * @retval PACE2_INVALID_MODULE_POINTER     @copydoc PACE2_return_state::PACE2_INVALID_MODULE_POINTER
 * @retval PACE2_PARAM_MISSING              @copydoc PACE2_return_state::PACE2_PARAM_MISSING
 * @retval PACE2_FAILURE                    @copydoc PACE2_return_state::PACE2_FAILURE
 */
PACE2_return_state pace2_get_base_version(PACE2_module * const pace2_module,
                                          const int thread_ID,
                                          struct PACE2_version * version);

#ifdef __cplusplus
}
#endif
#endif /* PACE2_VERSION_H */

/*********************************************************************************
 ** EOF
 ********************************************************************************/
