#!/bin/sh
# This script is executed as part of the hotplug event with
# HOTPLUG_TYPE=iface, triggered by various scripts when an interface
# is configured (ACTION=ifup) or deconfigured (ACTION=ifdown).  The
# interface is available as INTERFACE, the real device as DEVICE.

. /lib/zone/zone_api.sh
. /lib/functions.sh
. /lib/nat/nat_core.sh

if [[ -z "$INTERFACE" ]]; then
	[ -z "$DEVICE" ] && exit 0
	INTERFACE=$(zone_get_iface_bydev $DEVICE)
fi

[ -z "$INTERFACE" ] && exit 0 
[ "$DEVICE" == "lo" ] && exit 0

mkdir -p /tmp/.nat
local HOTP_LOG="/tmp/.nat/hotplug.log"
if [[ "`ls -l $HOTP_LOG | awk '{print $5}'`" -gt 50000 ]]; then
	echo -n "" > $HOTP_LOG
fi

HP_LOG()
{
	echo -e "`date +%Y%m%d-%H:%M:%S` $@" >> $HOTP_LOG
}
HP_LOG "$INTERFACE $ACTION"

{
	[ "$INTERFACE" == "lan" ] && {
		nat_napt_iface_list_hotplug $ACTION $DEVICE
	}
}

{
	flock -x 50
	if [[ ! -f "$NAT_READY" ]]; then
		HP_LOG "NAT is not ready, ingore this event"
		exit 0
	fi
	flock -u 50
} 50<>/tmp/nat.lock

local ruleNum=0
local needSave=0

get_rule_num()
{
	config_get interface $1 interface
	config_get enable $1 enable
	
	if [ -n "${interface}" -a "${enable}" == "on" ];then
		let ruleNum++
	fi
}

# delete one nat rule when wan iface is not static
modifyOnenatRule()
{
	local cfg="$1"
	local iface="$2"
	local zone
	local interface

	zone=$(uci_get nat "$cfg" interface)
	interface=$(zone_get_effect_ifaces "$zone")
	if [[ "$interface" == "$iface" ]]; then
		uci delete nat.$cfg
		needSave=1
	fi
}

# modify multinat sys rule when lan ip changed
modifyNatSysRule()
{
	local cfg="$1"
	local subnet="$2"
	local mask="$3"
	local sysRuleFlag
	local oldIpaddr
	local oldMask

	oldIpaddr=$(uci_get nat "$cfg" ipaddr)
	oldMask=$(uci_get nat "$cfg" mask)
	sysRuleFlag=$(uci_get nat "$cfg" sysrule)

	if [[ "$oldIpaddr" != "$subnet" || "$oldMask" != "$mask" ]]; then	
		if [[ -n "$sysRuleFlag" && "$sysRuleFlag" = "1" ]]; then
			uci set nat.$1.mask=${mask}
			uci set nat.$1.ipaddr=${subnet}
			needSave=1
		fi
	fi
}

lanIfHotPlugEvent()
{
	local lanIf="lan"
	if [[ -n "${INTERFACE}" && "${INTERFACE}" == "${lanIf}" ]]; then
		HP_LOG "LAN interface $ACTION, modify sys napt rules, interface: ${INTERFACE}"
		local ipaddr
		local subnet
		local netmask
		local prefix

		config_load nat
		config_foreach get_rule_num rule_vs
		config_foreach get_rule_num rule_dmz
		config_foreach get_rule_num rule_onenat
		config_foreach get_rule_num rule_pt
		
		ipaddr=$(nat_napt_lan_ipaddr)
		netmask=$(nat_napt_lan_netmask)
		if [[ -z "$ipaddr" || -z "$netmask" ]]; then
			HP_LOG "lan subnet is null error"
			return
		else
			HP_LOG "ipaddr: $ipaddr, netmask: $netmask"
			prefix=$(ipcalc -p $ipaddr $netmask)
			prefix=${prefix#*=}
			subnet=$(ipcalc -n $ipaddr $netmask)
			subnet=${subnet#*=}
			HP_LOG "prefix: $prefix, subnet: $subnet"
		fi

		config_load nat
		config_foreach modifyNatSysRule rule_napt $subnet $prefix

		if [[ "$needSave" -eq 1 ]]; then
			uci commit nat
			/lib/cfg_save/config.sh save
		fi

		HP_LOG "restart nat"
		if [[ "$ruleNum" -gt 0 ]]; then
			/etc/init.d/nat restart
		else
			/etc/init.d/nat restart napt
		fi
	fi
}

wanIfHotPlugEvent()
{
	local lanIf="lan"
	if [[ -n "${INTERFACE}" && "${INTERFACE}" != "${lanIf}" ]]; then
		HP_LOG "WAN interface $ACTION, modify one nat rules if needed, interface: ${INTERFACE}"

		config_load nat
		config_foreach get_rule_num rule_vs
		config_foreach get_rule_num rule_dmz
		config_foreach get_rule_num rule_onenat
		config_foreach get_rule_num rule_pt

		local proto=`uci get network.${INTERFACE}.proto`
		
		if [[ "$proto" != "static" ]]; then
			config_load nat
			config_foreach modifyOnenatRule rule_onenat ${INTERFACE}
		fi

		if [[ "$needSave" -eq 1 ]]; then
			uci commit nat
			/lib/cfg_save/config.sh save
		fi

		HP_LOG "restart nat"
		if [[ "$ruleNum" -gt 0 ]]; then
			/etc/init.d/nat restart
		else
			/etc/init.d/nat restart napt
		fi
	fi
}

case "$ACTION" in
	ifup)
		lanIfHotPlugEvent
		wanIfHotPlugEvent
	;;
	ifdown)
		lanIfHotPlugEvent
		wanIfHotPlugEvent
	;;
	#ifupdate)
	#	lanIfHotPlugEvent
	#	wanIfHotPlugEvent
	#;;
esac
