#!/bin/sh
# add/delete route rule, add/delete nat rule
#
#${ACTION}
#${IFNAME}
#${PPPD_USERNAME}
#${PPPD_PID}
#${PPPD_TYPE}
#${PPPD_CONFIGPATH}
#${IPREMOTE}
#${IPLOCAL}
#

. /lib/functions.sh
. /lib/firewall/fw.sh
. /lib/firewall/core_tpfirewall.sh
. /lib/pppox/pppox-default-variables.sh
. /lib/zone/zone_api.sh

#VPN相关策略路由条目都在table id为25的路由表中，其中内核arp.c中有相关修改，若是需要修改优先级，请同时修改arp.c的相关代码
VPN_RULE_NUM=25
VPN_FWMARK=0x20000
VPN_FORWARD_TABLE=VPN_FORWARD_MARK
VPN_ROUTING_TABLE=VPN_ROUTING_MARK

do_internet_vpn_up()
{
	local CHAIN=zone_vpn_${IFNAME}
	if ! fw list 4 f | grep "\-N ${CHAIN}"
	then
		fw add 4 f ${CHAIN}
		fw add 4 f forward ${CHAIN}
	else
		fw flush 4 f ${CHAIN}
		fw del 4 f forward ${CHAIN}
		fw del 4 f ${CHAIN}
	fi

	local wanmode=`uci get network.global.wanmode`
	local wan_numbers="`uci get network.global.wan_numbers`"
        local WAN_LIST=""

	[ -z "${wan_numbers}" ] && wan_numbers="$(seq 1 $wanmode)"

        for i in ${wan_numbers}; do
            WAN_LIST="WAN$i $WAN_LIST"
        done

	for wan in $WAN_LIST; do
		device_list=`zone_get_wan_devices $wan`
		for wan_device in $device_list; do
			fw add 4 f ${CHAIN} DROP $ { "-s ${IPREMOTE} -i ${IFNAME} -o $wan_device" }
		done
	done

	fw_event_manage_dev "${IFNAME}:up"		# Fix Bug 195146, let the packets from host to vpn go out.
}

do_internet_vpn_down()
{
	local CHAIN=zone_vpn_${IFNAME}
	if fw list 4 f | grep "\-N ${CHAIN}"
	then
		fw flush 4 f ${CHAIN}
		fw del 4 f forward ${CHAIN}
		fw del 4 f ${CHAIN}
	fi

	fw_event_manage_dev "${IFNAME}:down"	# Fix Bug 195146, let the packets from host to vpn go out.
}

fix_mss()
{
	[ $mss_offset == "0" ] && {
		return
	}

	local chain=${IFNAME}_MSSFIX

	[ $1 == "up" ] && {
		iface=$(zone_get_effect_ifaces "${zone}")
		[ -n "$iface" ] && {
			local cur_mtu=$(uci get network.$iface.mtu)
			local cur_mru=$(uci get network.$iface.mru)
			local cur_mss=1460

			cur_mtu=${cur_mtu:=$cur_mru}
			cur_mtu=${cur_mtu:=1500}
			cur_mss=$(($cur_mtu-40))
			cur_mss=$(($cur_mss-$mss_offset))		

			local rule="-p tcp -m tcp --tcp-flags SYN,RST SYN"
			local target_size="TCPMSS --set-mss ${cur_mss}"
			local target_mtu="TCPMSS --clamp-mss-to-pmtu"

			iptables -w -t mangle -C zone_MSSFIX -i $IFNAME -j ${chain} >/dev/null 2>&1
			if [[ "$?" -ne 0 ]]; then
				fw add 4 m ${chain}
				fw add 4 m zone_MSSFIX ${chain} ^ { -o ${IFNAME} }
				fw add 4 m zone_MSSFIX ${chain} ^ { -i ${IFNAME} }
				fw add 4 m ${chain} "$target_size" $ { "$rule" }
				fw add 4 m ${chain} "$target_mtu" $ { "$rule" }
			fi
		}
	}

	[ $1 == "down" ] && {
		fw flush 4 m ${chain}
		fw del 4 m zone_MSSFIX ${chain} ^ { -o ${IFNAME} }
		fw del 4 m zone_MSSFIX ${chain} ^ { -i ${IFNAME} }
		fw del 4 m ${chain}
	}
}

get_param()
{
	config_get netmode $1 "netmode"
	config_get remotesubnet $1 "remotesubnet"
	config_get local_network $1 "local_network"
	config_get workmode $1 "workmode"
	config_get zone $1 "outif"
	config_get en $1 "enable"
}

route_table_init()
{
	ret=`ip rule list | grep $VPN_RULE_NUM:`
	if [ -z "$ret" ]; then
		iptables -w -t mangle -F ${VPN_FORWARD_TABLE}
		iptables -w -t mangle -F ${VPN_ROUTING_TABLE}
		ip rule add pref $VPN_RULE_NUM fwmark ${VPN_FWMARK}/${VPN_FWMARK} table $VPN_RULE_NUM
	fi
}

vnete_route_table_init()
{
	ret=`vnete ip rule list | grep $VPN_RULE_NUM:`
	if [ -z "$ret" ]; then
		vnete ip rule add pref $VPN_RULE_NUM table $VPN_RULE_NUM
	fi
}
#1: local_network
#2: remotesubnet
route_table_mangle()
{
	local lnets="${2}"
	local rnets="${3}"
	local label="${4}"

	if [ "${1}" == "add" ]; then
		if [ -z "$lnets" ]; then
			lnet="0.0.0.0/0"
			for rnet in ${rnets}; do
				iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -i br-lan -s ${lnet} -d ${rnet} -m comment --comment "if=${label}" -j MARK --or-mark ${VPN_FWMARK}
				iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -i br-lan -s ${lnet} -d ${rnet} -m comment --comment "if=${label}" -j CONNMARK --or-mark ${VPN_FWMARK}
				iptables -w -t mangle -A ${VPN_ROUTING_TABLE} -i "${IFNAME}" -s ${rnet} -m set --match-set $(zone_allwan_group) dst -m comment --comment "if=${label},vnet=${lnet}" -j CONNMARK --or-mark ${VPN_FWMARK}
				iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -m set --match-set $(zone_allwan_group) src -m conntrack --ctdir REPLY -m connmark --mark ${VPN_FWMARK}/${VPN_FWMARK} -m comment --comment "if=${label},vnet=${lnet}" -j MARK --or-mark ${VPN_FWMARK}
			done
            iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -i br-lan -s ${lnet} -d ${IPREMOTE} -m comment --comment "if=${label}" -j MARK --or-mark ${VPN_FWMARK}
			iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -i br-lan -s ${lnet} -d ${IPREMOTE} -m comment --comment "if=${label}" -j CONNMARK --or-mark ${VPN_FWMARK}
		else
			local net_type=`name_or_network ${lnets}`
			case "${net_type}" in
				"name")
					for lnet in ${lnets}; do
						local netmask=$(uci get vnetwork."${lnet}".netmask)
						local ipaddr=$(uci get vnetwork."${lnet}".ipaddr)
						local network=$(ipcalc -n ${ipaddr} ${netmask} | cut -d '=' -f2)
						local prefix=$(ipcalc -p ${ipaddr} ${netmask} | cut -d '=' -f2)
						local cidr="${network}/${prefix}" 
						
						for rnet in ${rnets}; do
							iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -i br-lan -s ${cidr} -d ${rnet} -m comment --comment "if=${label},vnet=${lnet}" -j MARK --or-mark ${VPN_FWMARK}
							iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -i br-lan -s ${cidr} -d ${rnet} -m comment --comment "if=${label},vnet=${lnet}" -j CONNMARK --or-mark ${VPN_FWMARK}
							iptables -w -t mangle -A ${VPN_ROUTING_TABLE} -i "${IFNAME}" -s ${rnet} -m set --match-set $(zone_allwan_group) dst -m comment --comment "if=${label},vnet=${lnet}" -j CONNMARK --or-mark ${VPN_FWMARK}
							iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -m set --match-set $(zone_allwan_group) src -m conntrack --ctdir REPLY -m connmark --mark ${VPN_FWMARK}/${VPN_FWMARK} -m comment --comment "if=${label},vnet=${lnet}" -j MARK --or-mark ${VPN_FWMARK}
						done
                        iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -i br-lan -s ${cidr} -d ${IPREMOTE} -m comment --comment "if=${label}" -j MARK --or-mark ${VPN_FWMARK}
				        iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -i br-lan -s ${cidr} -d ${IPREMOTE} -m comment --comment "if=${label}" -j CONNMARK --or-mark ${VPN_FWMARK}
					done
					;;
				"network")
					for lnet in ${lnets}; do
						local cidr=${lnet}
						for rnet in ${rnets}; do
							iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -i br-lan -s ${cidr} -d ${rnet} -m comment --comment "if=${label},vnet=${lnet}" -j MARK --or-mark ${VPN_FWMARK}
							iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -i br-lan -s ${cidr} -d ${rnet} -m comment --comment "if=${label},vnet=${lnet}" -j CONNMARK --or-mark ${VPN_FWMARK}
							iptables -w -t mangle -A ${VPN_ROUTING_TABLE} -i "${IFNAME}" -s ${rnet} -m set --match-set $(zone_allwan_group) dst -m comment --comment "if=${label},vnet=${lnet}" -j CONNMARK --or-mark ${VPN_FWMARK}
							iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -m set --match-set $(zone_allwan_group) src -m conntrack --ctdir REPLY -m connmark --mark ${VPN_FWMARK}/${VPN_FWMARK} -m comment --comment "if=${label},vnet=${lnet}" -j MARK --or-mark ${VPN_FWMARK}
						done
                        iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -i br-lan -s ${lnet} -d ${IPREMOTE} -m comment --comment "if=${label}" -j MARK --or-mark ${VPN_FWMARK}
			            iptables -w -t mangle -A ${VPN_FORWARD_TABLE} -i br-lan -s ${lnet} -d ${IPREMOTE} -m comment --comment "if=${label}" -j CONNMARK --or-mark ${VPN_FWMARK}
					done
					;;
			esac
		fi
		iptables -w -t mangle -A ${VPN_ROUTING_TABLE} -i "${IFNAME}" -s ${IPREMOTE} -m set --match-set $(zone_allwan_group) dst -m comment --comment "if=${label},vnet=${IPREMOTE}" -j CONNMARK --or-mark ${VPN_FWMARK}
	else
		tables="${VPN_FORWARD_TABLE} ${VPN_ROUTING_TABLE}"
		for tbl in $tables
		do
			nums=$(iptables -w -t mangle -nL ${tbl} --line-numbers | grep "\b${label}\b" | cut -d ' ' -f 1)
			while [ -n "${nums}" ]; do
				for num in $nums;
				do
					iptables -w -t mangle -D ${tbl} ${num}
					break; # just delete the first once.
				done
				nums=$(iptables -w -t mangle -nL ${tbl} --line-numbers | grep "\b${label}\b" | cut -d ' ' -f 1)
			done
		done
	fi
}

route_table_add()
{
	ip route add "$@" table $VPN_RULE_NUM
}

route_table_del()
{
	ip route del "$@" table $VPN_RULE_NUM
}

# sdnc5.6 add by local network
name_or_network() 
{
    network=$1
    # eg. network="NETWORK1 NETWORK2" or "192.168.0.1/24 192.168.1.1/22"
    [ `expr "$network" : 'NETWORK'` -ne 0 ] && echo "name" && return 
    echo "network"
}

do_ifup()
{
	local hostnet=

	route_table_init
	vnete_route_table_init

	route_table_del ${IPREMOTE} dev ${IFNAME}
	case "${PPPD_TYPE}" in
		"server")
			route_table_add ${IPREMOTE} metric ${metric} dev ${IFNAME} # Fix Bug 195146, prior to Static Route

#			vnete sysctl -w /proc/sys/net/ipv4/conf/all/proxy_arp=1
			vnete ip route add ${IPREMOTE} via 169.254.11.21 table $VPN_RULE_NUM

			if [ "${netmode}" = "client2lan" -o "${netmode}" = "pc2lan" ]; then
#				pppd will do this
#				route add -host ${IPREMOTE} dev ${IFNAME}
				echo ""
			else
				#route add ${hostnet} ${remotesubnet} gw ${IPREMOTE} metric ${metric} dev ${IFNAME}
				for subnet in ${remotesubnet}; do
					hostnet="-net"
					[ "${subnet#*/}" = "32" ] && { hostnet="-host"; 
						subnet=${subnet%/*}
					}
					#echo "route add ${hostnet} ${subnet} gw ${IPREMOTE} metric ${metric} dev ${IFNAME}" > /dev/console
					route_table_add ${subnet} via ${IPREMOTE} metric ${metric} dev ${IFNAME}
				done
			fi
			;;

		"client")
			[ "${lns}" != "${IPREMOTE}" ] && {
				route_table_add ${IPREMOTE} metric ${metric} dev ${IFNAME}
			}
			#route add ${hostnet} ${remotesubnet} gw ${IPREMOTE} metric ${metric} dev ${IFNAME}
			#[ "$?" != "0" ] && route add ${hostnet} ${remotesubnet} metric ${metric} dev ${IFNAME}
			for subnet in ${remotesubnet}; do
				hostnet="-net"
				[ "${subnet#*/}" = "32" ] && { hostnet="-host"; 
					subnet=${subnet%/*}
				}
				#echo "route add ${hostnet} ${subnet} gw ${IPREMOTE} metric ${metric} dev ${IFNAME}" > /dev/console
				route_table_add ${subnet} via ${IPREMOTE} metric ${metric} dev ${IFNAME}
				[ "$?" != "0" ] && route_table_add ${subnet} metric ${metric} dev ${IFNAME}
			done

			;;
	esac
#	add interface into ZV_WANx ipset item
	[ -n "$zone" -a -n "$IFNAME" ] && {
		zone_add_vgroup "${IFNAME}" "${zone}"
	}
	
	[ ${workmode} = "nat" ] && {
		iptables -w -t nat -C postrouting_rule_multinat -o ${IFNAME} -j MASQUERADE 2>/dev/null
		[ "$?" != "0" ] && fw add 4 nat postrouting_rule_multinat MASQUERADE  $ { -o ${IFNAME} }
	}

	unset UCI_CONFIG_DIR
	do_internet_vpn_up
	#echo "local $local_network" > /dev/console
	if [ -z "$local_network" ]; then
		#fw add 4 f ppp_forward_rule ACCEPT { -i br-lan -o ${IFNAME} -m comment --comment "vnet="}
		iptables -w -A ppp_forward_rule -i br-lan -o ${IFNAME} -m comment --comment "vnet=" -j ACCEPT 
	else
		local net_type=`name_or_network ${local_network}`
		case "${net_type}" in
			"name")
				for lnet in ${local_network}; do
					local netmask=$(uci get vnetwork."${lnet}".netmask)
					local ipaddr=$(uci get vnetwork."${lnet}".ipaddr)
					local network=$(ipcalc -n ${ipaddr} ${netmask} | cut -d '=' -f2)
					local prefix=$(ipcalc -p ${ipaddr} ${netmask} | cut -d '=' -f2)
					local cidr="${network}/${prefix}" 
					#fw add 4 f ppp_forward_rule ACCEPT { -i br-lan -o ${IFNAME} -s ${cidr} -m comment --comment "vnet=${lnet}"} > /dev/console
					iptables -w -A ppp_forward_rule -i br-lan -o ${IFNAME} -s ${cidr} -m comment --comment "vnet=${lnet}" -j ACCEPT
					iptables -w -I ppp_forward_rule -i ${IFNAME} -o br-lan -d ${cidr} -m comment --comment "receive_local_subnet_${lnet}_and_drop_other_vlan" -j ACCEPT
					iptables -w -A ppp_forward_rule -i ${IFNAME} -o br-lan -d 0.0.0.0/0 -m comment --comment "drop_other_vlan_except_${lnet}" -j DROP
				done
				;;
			"network")
				for lnet in ${local_network}; do
					iptables -w -A ppp_forward_rule -i br-lan -o ${IFNAME} -s ${lnet} -m comment --comment "custom_net=${lnet}" -j ACCEPT
					iptables -w -I ppp_forward_rule -i ${IFNAME} -o br-lan -d ${lnet} -m comment --comment "receive_custom_net_${lnet}_and_drop_other_vlan" -j ACCEPT
					iptables -w -A ppp_forward_rule -i ${IFNAME} -o br-lan -d 0.0.0.0/0 -m comment --comment "drop_other_vlan_except_${lnet}" -j DROP
				done
				;;
		esac
	fi

	route_table_mangle add "${local_network}" "${remotesubnet}" "${IFNAME}"

	#fix mss for tcp pkt in vpn tunnel
	fix_mss "up"

	case "${PPPD_TYPE}" in
		"server")
			echo "";;
		"client")
			sleep 1
			for rnet in ${remotesubnet}; do
				local r_netmask=$(ipcalc -m ${rnet} | sed 's/NETMASK=//g')
				local r_ipaddr=$(ipcalc -n ${rnet} | sed 's/NETWORK=//g')
				if [ -z "$local_network" ]; then
					#echo "conntrack -D --dst ${r_ipaddr} --mask-dst ${r_netmask}" > /dev/console
					if [ $rnet = "0.0.0.0/0" ]; then
						[ -x /usr/sbin/conntrack ] && conntrack -D -m 0x10000 2>/dev/console
					else
						[ -x /usr/sbin/conntrack ] && conntrack -D --dst ${r_ipaddr} --mask-dst ${r_netmask} 2>/dev/console
						[ -x /usr/sbin/conntrack ] && conntrack -D --src ${r_ipaddr} --mask-src ${r_netmask} 2>/dev/console
					fi
				else
					local net_type=`name_or_network ${local_network}`
					case "${net_type}" in
						"name")
							for lnet in ${local_network}; do
								local l_netmask=$(uci get vnetwork."${lnet}".netmask)
								local l_ipaddr=$(uci get vnetwork."${lnet}".ipaddr)
								#echo "conntrack -D --src ${l_ipaddr} --mask-src ${l_netmask} --dst ${r_ipaddr} --mask-dst ${r_netmask}" > /dev/console
								#echo "conntrack -D --src ${r_ipaddr} --mask-src ${r_netmask} --dst ${l_ipaddr} --mask-dst ${l_netmask}" > /dev/console
								if [ $rnet = "0.0.0.0/0" ]; then
									[ -x /usr/sbin/conntrack ] && conntrack -D -m 0x10000 2>/dev/console
								else
									[ -x /usr/sbin/conntrack ] && conntrack -D --src ${l_ipaddr} --mask-src ${l_netmask} --dst ${r_ipaddr} --mask-dst ${r_netmask} 2>/dev/console
									[ -x /usr/sbin/conntrack ] && conntrack -D --src ${r_ipaddr} --mask-src ${r_netmask} --dst ${l_ipaddr} --mask-dst ${l_netmask} 2>/dev/console
								fi
							done
							;;
						"network")
							for lnet in ${local_network}; do
								local l_netmask=$(ipcalc -m "${lnet}" | cut -d "=" -f2)
								local l_ipaddr=$(echo "${lnet}" | cut -d "/" -f1)
								if [ $rnet = "0.0.0.0/0" ]; then
									[ -x /usr/sbin/conntrack ] && conntrack -D -m 0x10000 2>/dev/console
								else
									[ -x /usr/sbin/conntrack ] && conntrack -D --src ${l_ipaddr} --mask-src ${l_netmask} --dst ${r_ipaddr} --mask-dst ${r_netmask} 2>/dev/console
									[ -x /usr/sbin/conntrack ] && conntrack -D --src ${r_ipaddr} --mask-src ${r_netmask} --dst ${l_ipaddr} --mask-dst ${l_netmask} 2>/dev/console
								fi
							done
							;;
					esac
				fi
			done
			[ -x /usr/sbin/xfwdm ] && xfwdm -F
			[ -e "/etc/init.d/hwnat" ] && hwnat -y
			[ -e /sys/kernel/debug/ecm/ecm_db/defunct_all ] && echo 1 > /sys/kernel/debug/ecm/ecm_db/defunct_all
			;;
	esac

}

do_ifdown()
{
	local hostnet=

	route_table_init
	vnete_route_table_init

	case "${PPPD_TYPE}" in
		"server")
			route_table_del ${IPREMOTE} dev ${IFNAME}

			vnete ip route del ${IPREMOTE} via 169.254.11.21 table $VPN_RULE_NUM

			if [ "${netmode}" = "client2lan" -o "${netmode}" = "pc2lan" ]; then
#				pppd will do this
#				route del -host ${IPREMOTE} dev ${IFNAME}
				echo ""
			else
				#route del ${hostnet} ${remotesubnet} dev ${IFNAME}
				#echo "remotesubnet $remotesubnet" > /dev/console
				for subnet in ${remotesubnet}; do
					hostnet="-net"
					[ "${subnet#*/}" = "32" ] && { hostnet="-host"; 
						subnet=${subnet%/*}
					}
					#echo "route del ${hostnet} ${subnet} dev ${IFNAME}" > /dev/console
					route_table_del ${subnet} dev ${IFNAME}
				done
			fi
			;;

		"client")
			fw del 4 nat postrouting_rule_multinat MASQUERADE  $ { -o ${IFNAME} }
			#route del ${hostnet} ${remotesubnet} dev ${IFNAME}
			for subnet in ${remotesubnet}; do
				hostnet="-net"
				[ "${subnet#*/}" = "32" ] && { hostnet="-host"; 
					subnet=${subnet%/*}
				}
				#echo "route del ${hostnet} ${subnet} dev ${IFNAME}" > /dev/console
				route_table_del ${subnet} dev ${IFNAME}
			done
			route_table_del ${IPREMOTE} dev ${IFNAME}

			[ ${vvtype} == "pptp" ] && [ ${en} == "on" ] && [ -f ${path}/ppp ] && [ -f ${path}/mgr ] && [ -f ${path}/options.pptp ] && {
				ppp_id=`cat ${path}/ppp`
				mgr_id=`cat ${path}/mgr`
				sleep 2
				cnt=`ps | grep "pppd" | grep -E "$ppp_id | $mgr_id" | wc -l`
				if [ $cnt -lt 2 ] ; then
				# restart new pppd
					lusername=`uci -c ${path} get myconfig.@pnc[0].username`
					lpassword=`uci -c ${path} get myconfig.@pnc[0].password`
					lpns=`uci -c ${path} get myconfig.@pnc[0].pns`
					lworkmode=`uci -c ${path} get myconfig.@pnc[0].workmode`
					lremotesubnet=`uci -c ${path} get myconfig.@pnc[0].remotesubnet`
					loutif=`uci -c ${path} get myconfig.@pnc[0].outif`
					local_network=`uci -c ${path} get myconfig.@pnc[0].local_network`

					local lmppe=1
					if cat ${path}/options.pptp | grep "#mppe" ; then
						lmppe=0
					fi

					/lib/pptp/pptp-client-update.sh ${PPPD_CONFIGPATH} on on "$lusername" "$lpassword" $lpns $lworkmode "$lremotesubnet" $loutif $lmppe "${local_network}"
				else
					echo "ppp_id=$ppp_id is exist" > /dev/console
				fi
			}
;;

	esac
#               delete interface into ZV_WANx ipset item
	[ -n "$zone" -a -n "$IFNAME" ] && {
		zone_delete_vgroup "${IFNAME}" "${zone}"
	}	
	#[ ${workmode} = "nat" ] && fw del 4 nat POSTROUTING MASQUERADE  $ { -o ${IFNAME} }	

	do_internet_vpn_down
	
	#echo "local_network down ${local_network}" > /dev/console
	#if [ -z "$local_network" ]; then
	#	fw del 4 f ppp_forward_rule ACCEPT { -i br-lan -o ${IFNAME} }
	#else
	#	for local in ${local_network}; do
	#		fw del 4 f ppp_forward_rule ACCEPT { -i br-lan -o ${IFNAME} -s ${local} }
	#	done
	#fi
	local localnet=$(iptables -w -nvL ppp_forward_rule | grep ${IFNAME} | awk '{print $8}')
	#for local in ${localnet}; do
		#echo "fw del $-i br-lan -o ${IFNAME} -s ${local}" > /dev/console
	#	fw del 4 f ppp_forward_rule ACCEPT { -i br-lan -o ${IFNAME} -s ${local} }
	#done
	
	nums=$(iptables -w -nvL ppp_forward_rule --line-numbers | grep "\b${IFNAME}\b" | cut -d ' ' -f 1)
	while [ -n "${nums}" ]; do
		for num in $nums;
		do
			iptables -w -D ppp_forward_rule ${num}
			break; # just delete the first once.
		done
		nums=$(iptables -w -nvL ppp_forward_rule --line-numbers | grep "\b${IFNAME}\b" | cut -d ' ' -f 1)
	done

	route_table_mangle del "${localnet}" "${remotesubnet}" "${IFNAME}"

	fix_mss "down"

}
echo "pppox if ${ACTION}" > /dev/console
#PPPD_CONFIGPATH like "/tmp/l2tp/server/server0", note: no "config"
path=${PPPD_CONFIGPATH}/config
UCI_CONFIG_DIR=${path}
pidpath=/tmp/ppp/pid

metric=0
mss_offset=40
vvtype=l2tp
opt_name=lns
if echo ${PPPD_CONFIGPATH}|grep -q "^/tmp/l2tp"; then
	metric=`uci get system_params.metric.l2tp 2>/dev/null`
	vvtype=l2tp
	mss_offset=40
	vpt_name=lns
elif echo ${PPPD_CONFIGPATH}|grep -q "^/tmp/pptp"; then
	metric=`uci get system_params.metric.pptp 2>/dev/null`
	vvtype=pptp
	mss_offset=56
	opt_name=pns
elif echo ${PPPD_CONFIGPATH}|grep -q "^/tmp/pppoe"; then
	metric=`uci get system_params.metric.pppoe 2>/dev/null`
	vvtype=pppoe
	mss_offset=0
fi
metric=${metric:-0}

if [ "${PPPD_TYPE}" = "server" ]; then
#workmode for server must be route
	pppuser=`asciimaker -a "${PPPD_USERNAME}"`
	netmode=`uci -c ${path} get ${pppox_if_config}.${pppuser}.netmode 2>/dev/null`
	remotesubnet=`uci -c ${path} get ${pppox_if_config}.${pppuser}.remotesubnet 2>/dev/null`
	local_network=`uci -c ${path} get ${pppox_if_config}.${pppuser}.local_network 2>/dev/null`
	if [ -z $local_network ]; then
		server_name=`echo "${PPPD_CONFIGPATH}" | cut -d '/' -f 5 2>/dev/null`
		server_type=`echo "${PPPD_CONFIGPATH}" | cut -d '/' -f 3 2>/dev/null`
		loop=0
		section_name="pns"
		if [ ${server_type} = "l2tp" ]; then
			section_name="lns"
		fi
		result=$(uci get ${server_type}-server.@${section_name}[$loop].server 2>/dev/null)
		while [ $result != &{server_name} ];do
			loop=$(($loop+1))
		done
		local_network=`uci get ${server_type}-server.@${section_name}[$loop].local_network`
	fi

	zone=`uci -c ${path} get ${pppox_configname}.@${opt_name}[0].bindif 2>/dev/null`
	workmode=route
	if [ "${netmode}" = "client2lan" -o "${netmode}" = "pc2lan" ]; then
		remotesubnet="0.0.0.0/0"
	fi
	#echo "uci -c ${path} get ${pppox_if_config}.${pppuser} out ${remotesubnet} ${local_network}" > /dev/console
elif [ "${PPPD_TYPE}" = "client" ]; then
	configname=${pppox_configname}
	
	config_load ${configname}
	config_foreach get_param
	config_foreach config_clear

#	netmode=`uci -c ${path} get ${configname}.@${vvtype}[-1].netmode 2>/dev/null`
#	remotesubnet=`uci -c ${path} get ${configname}.@${vvtype}[-1].remotesubnet 2>/dev/null`
#	workmode=`uci -c ${path} get ${configname}.@${vvtype}[-1].workmode 2>/dev/null`

	lns=`cat ${path}/serverip`
	netmode=lan2lan	
else
	return 1
fi

if [ "${ACTION}" = "ifup" ]; then
	uci -c ${pidpath} set ${PPPD_PID}.@info[0].uptime=$(date +%s)
	uci -c ${pidpath} commit ${PPPD_PID}
	
	#echo "uci -c ${pidpath} set ${PPPD_PID}.@info[0].uptime = $(date +%s)" > /dev/console
fi

#hostnet="-net"
#[ "${remotesubnet#*/}" = "32" ] && { hostnet="-host"; 
#		remotesubnet=${remotesubnet%/*}
#	}
case "${ACTION}" in
	"ifup")
		do_ifup;;
	"ifdown")
		do_ifdown;;
	*)
		return 1;;
esac
if [ "${PPPD_TYPE}" = "client" ]; then
	[ -x /etc/init.d/dnsproxy ] && /etc/init.d/dnsproxy restart
fi
