local sys  = require "luci.sys"
local json = require 'luci.json'
local utl = require "luci.util"
local dbg  = require 'luci.tools.debug'
local error_print  = require 'luci.tools.error_print'
local form = require "luci.tools.form"
local uci  = require "luci.model.uci"
local json_m = require "luci.json"
local timesetting = require 'luci.controller.admin.timesetting'

local MINUTES = 60
local TIMEZONE_MIN = -12
local TIMEZONE_NUM_MAX = 1500
local TIMEZONE_NUM_MIN = 0
local INVALID_IP_BEGINNING = 255
local INVALID_TIMEZONE_VALUE = -1
local SMALLMONTH_DAY_MAX = 30
local FEBRUARY_NUM = 2
local SMALL_FEB_DAY_MAX = 28
local BIG_FEB_DAY_MAX = 29

local uci_r = uci.cursor()

local TimeZone = {
    0,60,120,180,240,300,360,420,450,480,510,540,600,660,720,780,840,900,930,960,990,1020,1050,1065,1080,1110,1140,1200,1260,1290,1320,1380,1440,1500
}

local SmallMonth = {4,6,9,11}

function set_system_time_manual(time)
    local http_form = {}
    local sets = {}
    local cur_time = os.time()
    sets.date = os.date("%m/%d/%Y", cur_time)
    sets.time = os.date("%X", cur_time)
    local current_time = sets.date.." "..sets.time
    
    local entry = string.split(time, '-')
    local date = entry[1]
    local time = entry[2]
    local month, day, year = string.match(date, "(%d+)/(%d+)/(%d+)")

    for _, month_num in ipairs(SmallMonth) do
        if month_num == tonumber(month) and tonumber(day)>SMALLMONTH_DAY_MAX then
            dbg.print_tty("Error: Invalid parameter "..time..".")
            return "FAILED"
        end
    end

    if tonumber(month) == FEBRUARY_NUM then
        if((is_leap_year(tonumber(year))) and tonumber(day)>BIG_FEB_DAY_MAX) or ((not is_leap_year(tonumber(year))) and tonumber(day)>SMALL_FEB_DAY_MAX) then
            dbg.print_tty("Error: Invalid parameter "..time..".")
            return "FAILED"
        end
    end

    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] = 
        {
                ["current_time"] = current_time,
                ["type"] = "manual",
                ["date"] = date,
                ["time"] = time,
        }
    })
    http_form.form = "settings"
    ret = timesetting.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        return "FAILED"
    end

	return "SUCCESS"
end

function set_system_time_ntp1(timezone, ntpserver1)
    local http_form = {}
    local current_time = get_current_time()

    local timezone_num = get_timezone_num(timezone)
    if(timezone_num==INVALID_TIMEZONE_VALUE) then
        return "FAILED"
    end

    if(not ip_beginning_check(ntpserver1)) then
        return "FAILED"
    end

    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] = 
        {
                ["current_time"] = current_time,
                ["type"] = "auto",
                ["timezone"] = tostring(timezone_num),
                ["ntp_svr1"] = tostring(ntpserver1),
                ["ntp_svr2"] = ""
        }
    })
    
    http_form.form = "settings"
    ret = timesetting.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        return "FAILED"
    end

	return "SUCCESS"
end

function set_system_time_ntp2(timezone, ntpserver1, ntpserver2)
    local http_form = {}
    local current_time = get_current_time()

    local timezone_num = get_timezone_num(timezone)
    if(timezone_num==INVALID_TIMEZONE_VALUE) then
        return "FAILED"
    end

    if(not ip_beginning_check(ntpserver1)) then
        return "FAILED"
    elseif(not ip_beginning_check(ntpserver2)) then
        return "FAILED"
    end

    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] = 
        {
                ["current_time"] = current_time,
                ["type"] = "auto",
                ["timezone"] = tostring(timezone_num),
                ["ntp_svr1"] = tostring(ntpserver1),
                ["ntp_svr2"] = tostring(ntpserver2)
        }
    })
    http_form.form = "settings"
    ret = timesetting.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        return "FAILED"
    end

	return "SUCCESS"
end

function show_time_settings()
    local sets = {}
    local cur_time = os.time()
    sets.date = os.date("%m/%d/%Y", cur_time)
    sets.time = os.date("%X", cur_time)
    sets.type = uci_r:get("system", "ntp", "type")

    local sname = uci_r:get_first("system", "system", nil, nil)
    local cur_tz = uci_r:get("system", sname, "timezone")
        
    dbg.print_tty(sets.time.."("..cur_tz..")"..sets.date)
    dbg.print_tty("Time source is "..sets.type)

    return "SUCCESS"

end

function show_time_ntp_settings()
    local sets = {}
    sets.type = uci_r:get("system", "ntp", "type")

    local ntp_table = uci_r:get("system", "ntp", "user_server")
    sets.ntp_svr1 = ntp_table and ntp_table[1] or ""
    sets.ntp_svr2 = ntp_table and ntp_table[2] or ""

    if sets.ntp_svr1 == "" and sets.ntp_svr2 == "" or sets.type == "manual" then
        dbg.print_tty("Error: NTP is not enabled.")
         
    else
        local sname = uci_r:get_first("system", "system", nil, nil)
        local cur_tz = uci_r:get("system", sname, "timezone")

        dbg.print_tty("Time zone: "..cur_tz)   
        dbg.print_tty("Prefered NTP server: "..sets.ntp_svr1)
        dbg.print_tty("Backup NTP server: "..sets.ntp_svr2)
    end

    return "SUCCESS"
    
end

function get_current_time()
    local cur_time = os.time()
    local current_time = os.date("%m/%d/%Y", cur_time).." "..os.date("%X", cur_time)
    return current_time
end

function get_timezone_num(timezone)

    local symbol = string.find(timezone,"-")
    if(symbol~=nil) then
        symbol = "-"
    else
        symbol = string.find(timezone,"+")
        if(symbol~=nil) then
            symbol = "+"
        else
            dbg.print_tty("Error: Invalid parameter "..timezone..".")
            return INVALID_TIMEZONE_VALUE
        end
    end
    local time_zone, zone_part = string.match(timezone,"(%d+):(%d+)")
    local time_zone_num = tonumber(time_zone)
    local zone_part_num = tonumber(zone_part)
    local timezone_num = 0

    if symbol=="-" then
        timezone_num = (-time_zone_num-TIMEZONE_MIN)*MINUTES - zone_part_num
    elseif symbol=="+" then
        timezone_num = (time_zone_num-TIMEZONE_MIN)*MINUTES + zone_part_num        
    end
     
    if(not parse_timezone(timezone_num)) then
        dbg.print_tty("Error: Invalid parameter "..timezone..".") 
    	return INVALID_TIMEZONE_VALUE
    end

    return timezone_num
end

function parse_timezone(zone)  
    for _, time_num in ipairs(TimeZone) do
        if time_num == zone then
            return true
        end
    end
    return false
end

function ip_beginning_check(ntpserver)
    local server = string.split(ntpserver, '.')
    if(tonumber(server[1])==INVALID_IP_BEGINNING) then
        dbg.print_tty("Error: The address cannot start with "..INVALID_IP_BEGINNING..".")
        return false
    end
    return true
end

function is_leap_year(year)
    if(year%4==0 and year%100~=0) or year%400==0 then
        return true
    end
    return false
end

local result = "SUCCESS"
if arg[1] == "system-time manual" then
    local time = arg[2]
    result = set_system_time_manual(time)

elseif arg[1] == "system-time ntp timezone ntp-server1" then
    local timezone = arg[2]
    local ntpserver1 = arg[3]
    result = set_system_time_ntp1(timezone, ntpserver1)

elseif arg[1] == "system-time ntp timezone ntp-server1 ntp-server2" then
    local timezone = arg[2]
    local ntpserver1 = arg[3]
    local ntpserver2 = arg[4]
    result = set_system_time_ntp2(timezone, ntpserver1, ntpserver2)

elseif arg[1] == "show system-time" then
    result = show_time_settings()

elseif arg[1] == "show system-time ntp" then
    result = show_time_ntp_settings()
end

io.stdout:write(result)