local sys  = require "luci.sys"
local json = require "luci.json"
local interfaces = require "luci.model.interfaces"
local userconfig = require "luci.model.userconfig"
local lan  = require "luci.controller.admin.lan"
local IPSEC = require "luci.controller.admin.ipsec"
local IPPOOL = require "luci.controller.admin.ippool"
local utl = require "luci.util"
local dbg  = require "luci.tools.debug"
local err = require "luci.tools.error"
local form = require "luci.tools.form"
local uci  = require "luci.model.uci"
local ippoolref = require "luci.model.reference"
local error_print = require "luci.tools.error_print"

local uci_r = uci.cursor()
local form = form.Form(uci_r, {"vlanid"})

local CLI_CFG_PATH = "/tmp/cli_config"
local CFG_PATH = "/etc/config"

local enc_algs = {  ["--"] = "--", 
                    ["des"] = "1",
                    ["3des"] = "2",
                    ["aes128"] = "3",
                    ["aes192"] = "4",
                    ["aes256"] = "5"}

local auth_algs = { ["--"] = "--", 
                    ["md5"] = "1",
                    ["sha1"] = "2",
                    ["sha256"] = "3",
                    ["sha384"] = "4",
                    ["sha512"] = "5"}
--default params
local DEFAULT_LIFETIME = 28800
local DEFAULT_TUNNEL_MODE = "tunnel"
local DEFAULT_EXCHANGE_MODE = "main"
local DEFAULT_CONNECTION_TYPE = "initiator"
local DEFAULT_PFS = "none"
local DEFAULT_DPD_ENABLE = "enable"
local DEFAULT_DPD_INTERVAL = "10"

--ret val
local IPSEC_FAIL_STATUS = "FAIL"
local IPSEC_SUCCESS_STAUS = "SUCCESS"

--policy,transform_set,map may have the same name, so add prefix section name to discriminate them
local IPSEC_POLICY_SECTION_PRE = "pol_"
local IPSEC_TRANSFORM_SECTION_PRE = "tset_"
local IPSEC_MAP_SECTION_PRE = "map_"

--set config directory to "/tmp/cli_config"
uci_r:set_confdir(CLI_CFG_PATH)

--interface file
local interface_config_file_name = CLI_CFG_PATH.."/interface_form.txt"

local filename = CLI_CFG_PATH .."/ipsec_cli"
local file = io.open(filename,"r")

if file == nil then
    sys.exec("touch " .. filename)
else
    file:close()
end

function check_ike_policy_complete(pol_name)
    pol_name = IPSEC_POLICY_SECTION_PRE .. pol_name
    local policy = uci_r:get_all("ipsec_cli", pol_name)

    if (policy == nil or policy.ike_version == nil or policy.encrypt == nil or policy.auth == nil or policy.dh_group == nil
        or policy.lifetime == nil or policy.dpd_enable == nil or policy.key == nil
        or (policy.ike_version == "ikev1" and policy.exchange_mode == nil)) then
        return false
    elseif (policy.dpd_enable and policy.dpd_interval == nil) then
        dbg.print_tty("Error: Not set dpd_interval.")
        return false
    else
        return true
    end
end

function check_transform_set_complete(tset_name)
    tset_name = IPSEC_TRANSFORM_SECTION_PRE .. tset_name
    local tset = uci_r:get_all("ipsec_cli", tset_name)

    if (tset == nil or tset.encrypt == nil or tset.auth == nil or tset.lifetime == nil or tset.pfs == nil or tset.mode == nil) then
        return false
    else
        return true
    end
end

function delete_map(map)
    --dbg.print_tty("delete_map")
    if not uci_r:delete("ipsec_cli", IPSEC_POLICY_SECTION_PRE .. map.policy) then
        return false
    end
    if not uci_r:delete("ipsec_cli", IPSEC_TRANSFORM_SECTION_PRE .. map.transform_set) then
        return false
    end
    if not uci_r:delete("ipsec_cli", IPSEC_MAP_SECTION_PRE .. map.name) then
        return false
    end
    return true
end
    
function check_map_complete(map)
    if (map.remote_peer == nil or map.local_network == nil or map.remote_network == nil or map.connection_type == nil) then
        dbg.print_tty("Error: Map not complete.")
        return false
    end

    if(map.policy == nil or (map.policy and not check_ike_policy_complete(map.policy))) then
        dbg.print_tty("Error: Policy not complete.")
        return false
    end

    if(map.transform_set == nil or (map.transform_set and not check_transform_set_complete(map.transform_set))) then
        dbg.print_tty("Error: Transform set not complete.")
        return false
    end

    return true
end

function get_section_from_def_path(section_type, section_name)
    local find_section = {}
    local ret = 0
    uci_r:set_confdir(CFG_PATH)
    uci_r:foreach("ipsec", section_type, function (section)
        --dbg.print_tty(section.name)
        if section.name == section_name then
            find_section = section
            ret = 1
            return find_section
        end
    end)
    uci_r:set_confdir(CLI_CFG_PATH)
    return ret, find_section
end

function get_option_from_def_path(section_type, section_name, option_name)
    local find_option
    local find_list = {}
    local ret = false
    uci_r:set_confdir(CFG_PATH)
    uci_r:foreach("ipsec", section_type, function (section)
        --dbg.print_tty(section.name)
        if section.name == section_name then
            if type(section[option_name]) == "table" then
                find_list = section[option_name]
                find_option = find_list[1]
                ret = true
                return ret
            else
                find_option = section[option_name]
            end
            ret = true
            return ret
        end
    end)
    uci_r:set_confdir(CLI_CFG_PATH)
    return ret, find_option
end
--@retval 
--@0 not found
--@1 found in /tmp/cli_config/ipsec
--@2 found in /etc/config/ipsec
function check_exist(pre_name, section_type, section_name)
    local ret = 0
    local section
    local cli_name = pre_name .. section_name

    --find in /tmp/cli_config/ipsec
    section = uci_r:get_all("ipsec_cli", cli_name)
    if section then
        --dbg.print_tty("find in /tmp/cli_config/ipsec")
        return 1
    end
    --find in /etc/config/ipsec
    ret, section = get_section_from_def_path(section_type, section_name)
    if ret ~= 0 and section then
        --dbg.print_tty("find in /etc/config/ipsec")
        return 2
    end
    return ret
end

function refresh_ike_proposal(section_name)
    local section = uci_r:get_all("ipsec_cli", section_name)
    local proposal
    if not section then
        return false
    end
    if section.encrypt and section.auth and section.dh_group then
        proposal = section.auth .. "-" .. section.encrypt .. "-" .. section.dh_group
        uci_r:set("ipsec_cli", section_name, "ike_proposal", proposal)
    end
    return true
end

function convert_uci_to_json_for_add(map)
    --dbg.print_tty("begin: \r\n")
	--- Add ipsec Rule
    local ret
    local http_form = {}
    pol_name = IPSEC_POLICY_SECTION_PRE .. map.policy
    tset_name = IPSEC_TRANSFORM_SECTION_PRE .. map.transform_set
    local policy = {}
    local tset = {}
    if tonumber(map.def_path_pol) == 1 then 
        policy = uci_r:get_all("ipsec_cli", pol_name)
    else
        ret, policy = get_section_from_def_path("ike_stage1", map.policy)
    end
    if tonumber(map.def_path_tset) == 1 then
        tset = uci_r:get_all("ipsec_cli", tset_name)
    else
        ret, tset = get_section_from_def_path("ike_stage2", map.transform_set) 
    end
    local http_form = {}
	http_form.data = json.encode(
	{
		["method"] = "add",
		["params"] =
		{
			["index"] = 1,
			["old"] = "add",
			["new"] =
			{
				["name"] = map.name, -- 前后端命名规则差异，不使用前端名称
				["pmode"] = "LAN-to-LAN",
				["remote_peer"] = map.remote_peer,
				["remote_host"] = map.remote_peer,
				--- change
				["local_binding"] = map.local_binding,
				["local_network"] = map.local_network,
				["local_vid"] = 1,
				["remote_network"] = map.remote_network,
				["negotiation_strategy"] = "",
				["ike_version"] = policy.ike_version,
				["psk"] = policy.key,
				["status"] = "on",
                ["ike_name"] = policy.name,
				["ike_proposal_1"] = policy.ike_proposal,
				["ike_proposal_2"] = "---",
				["ike_proposal_3"] = "---",
				["ike_proposal_4"] = "---",
				["exchange_mode"] = policy.exchange_mode,
				["connection_type"] = map.connection_type,
				["config_mode"] = "disable",
				["local_id_type"] = "IP_ADDRESS",
				["remote_id_type"] = "IP_ADDRESS",
				["ikelifetime"] = tostring(policy.lifetime),
				["dpd_enable"] = policy.dpd_enable,
				["dpd_interval"] = tostring(policy.dpd_interval),
				["mode"] = tset.mode,
                ["ph2_name"] = tset.name,
				["ph2_proposal_1"] = tset.ike_stage2_proposal,
				["ph2_proposal_2"] = "---",
				["ph2_proposal_3"] = "---",
				["ph2_proposal_4"] = "---",
				["pfs"] = tset.pfs,
				["lifetime"] = tset.lifetime,
                ["from_cli"] = "1"
			},
			["key"] = "add"
		}
	})
	http_form.form = "conn"
	
	--dbg.dumptable(http_form)
    return http_form
end

function convert_uci_to_json_for_modify(map, section_type, section_name, option_name, option_val)
    uci_r:set_confdir(CFG_PATH)
    local policy = {}
    local tset = {}
    local new_map = map
    local new_policy = {}
    local new_tset = {}
    local http_form = {}
    local key = ""
    local new_key = ""

    --get policy and transform from default config path
    uci_r:foreach("ipsec", "ike_stage1",
        function(section)
            if section.name == map.policy then
                policy = section
                new_policy = section
                return
            end
        end
    )
    uci_r:foreach("ipsec", "ike_stage2",
        function(section)
            if section.name == map.transform_set then
                tset = section
                new_tset = section
                return 
            end
        end
    )
    --get psk
    uci_r:foreach("ipsec_secrets", "authentication", 
        function(section)
            if section.name == map.name then
                key = section.psk
                new_key = key
            end
        end
    )

    if section_type == "connection" then
        if option_name == "ike_stage1" then
            option_name = "policy"
        end
        if option_name == "ike_stage2" then
            option_name = "transform_set"
        end
        if new_map[option_name] == option_val then
            return 0, http_form
        else
            new_map[option_name] = option_val
        end
    elseif section_type == "ike_stage1" then
        if option_name == "key" then
            if new_key == option_val then 
                return 0, http_form
            else
                new_key = option_val
            end
        else
            if option_name == "ike_proposal" then
                if new_policy.ike_proposal[1] == option_val then
                    return 0, http_form
                else
                    new_policy.ike_proposal[1] = option_val
                end
            elseif new_policy[option_name] == option_val then
                return 0, http_form
            else
                new_policy[option_name] = option_val
            end
        end
    elseif section_type == "ike_stage2" then
        if option_name == "ike_stage2_proposal" then
            if new_tset.ike_stage2_proposal[1] == option_val then
                return 0, http_form
            else
                new_tset.ike_stage2_proposal[1] = option_val
            end
        elseif new_tset[option_name] == option_val then
            return 0, http_form
        else
            new_tset[option_name] = option_val
        end
    end

    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] =
        {
            ["index"] = map.index,
            ["old"] = 
            {
                ["name"] = map.name, -- 前后端命名规则差异，不使用前端名称
                ["pmode"] = "LAN-to-LAN",
                ["remote_peer"] = map.remote_peer,
                ["remote_host"] = map.remote_peer,
                --- change
                ["local_binding"] = map.local_binding,
                ["local_network"] = map.local_network,
                ["local_vid"] = 1,
                ["remote_network"] = map.remote_network,
                ["negotiation_strategy"] = "",
                ["ike_version"] = policy.ike_version,
                ["psk"] = key,
                ["status"] = "on",
                ["ike_name"] = policy.name,
                ["ike_proposal_1"] = policy.ike_proposal[1],
                ["ike_proposal_2"] = "---",
                ["ike_proposal_3"] = "---",
                ["ike_proposal_4"] = "---",
                ["exchange_mode"] = policy.exchange_mode,
                ["connection_type"] = map.connection_type,
                ["config_mode"] = "disable",
                ["local_id_type"] = "IP_ADDRESS",
                ["remote_id_type"] = "IP_ADDRESS",
                ["ikelifetime"] = tostring(policy.lifetime),
                ["dpd_enable"] = policy.dpd_enable,
                ["dpd_interval"] = tostring(policy.dpd_interval),
                ["mode"] = tset.mode,
                ["ph2_name"] = tset.name,
                ["ph2_proposal_1"] = tset.ike_stage2_proposal[1],
                ["ph2_proposal_2"] = "---",
                ["ph2_proposal_3"] = "---",
                ["ph2_proposal_4"] = "---",
                ["pfs"] = tset.pfs,
                ["lifetime"] = tset.lifetime,
            },
            ["new"] =
            {
                ["name"] = new_map.name, -- 前后端命名规则差异，不使用前端名称
                ["pmode"] = "LAN-to-LAN",
                ["remote_peer"] = new_map.remote_peer,
                ["remote_host"] = new_map.remote_peer,
                --- change
                ["local_binding"] = new_map.local_binding,
                ["local_network"] = new_map.local_network,
                ["local_vid"] = 1,
                ["remote_network"] = new_map.remote_network,
                ["negotiation_strategy"] = "",
                ["ike_version"] = new_policy.ike_version,
                ["psk"] = new_key,
                ["status"] = "on",
                ["ike_name"] = new_policy.name,
                ["ike_proposal_1"] = new_policy.ike_proposal[1],
                ["ike_proposal_2"] = "---",
                ["ike_proposal_3"] = "---",
                ["ike_proposal_4"] = "---",
                ["exchange_mode"] = new_policy.exchange_mode,
                ["connection_type"] = new_map.connection_type,
                ["config_mode"] = "disable",
                ["local_id_type"] = "IP_ADDRESS",
                ["remote_id_type"] = "IP_ADDRESS",
                ["ikelifetime"] = tostring(new_policy.lifetime),
                ["dpd_enable"] = new_policy.dpd_enable,
                ["dpd_interval"] = tostring(new_policy.dpd_interval),
                ["mode"] = new_tset.mode,
                ["ph2_name"] = new_tset.name,
                ["ph2_proposal_1"] = new_tset.ike_stage2_proposal[1],
                ["ph2_proposal_2"] = "---",
                ["ph2_proposal_3"] = "---",
                ["ph2_proposal_4"] = "---",
                ["pfs"] = new_tset.pfs,
                ["lifetime"] = new_tset.lifetime,
            },
            ["key"] = "key-2"
        }
    })
    http_form.form = "conn"

    --dbg.dumptable(http_form)
    return 1, http_form
end

function modify_used_config(section_type, section_name, option_name, option_val)
    local maps = {}
    local http_form
    local ret
    local index = 0
    uci_r:set_confdir(CFG_PATH)
    --Only extract required params in CLI
    uci_r:foreach("ipsec","connection",
        function(section)
            if section[section_type] == section_name then
                --dbg.print_tty("connection matched")
                maps[#maps + 1] = {}
                maps[#maps].name = section.name
                maps[#maps].policy = section.ike_stage1
                maps[#maps].transform_set = section.ike_stage2
                maps[#maps].remote_peer = section.remote_peer
                maps[#maps].remote_peer_type = section.remote_peer_type
                maps[#maps].local_network = section.local_network
                maps[#maps].remote_network = section.remote_network
                maps[#maps].local_binding = section.local_binding
                maps[#maps].connection_type = section.connection_type
                maps[#maps].index = index
            end
            index = index + 1
        end
    )

    if section_type == "name" then
        section_type = "connection"
    end
    for _, map in pairs(maps) do
        --dbg.print_tty("start to convert")
        ret, http_form = convert_uci_to_json_for_modify(map, section_type, section_name, option_name, option_val)
        if ret == 0 then
            return IPSEC_SUCCESS_STAUS
        end
        ret = IPSEC.dispatch(http_form)
        if (not ret) or (ret.error_code ~= "0") then
            --result = err_2_sdnerr(ret.error_code)
            --return false,result
            dbg.print_tty(error_print.check_error(ret))
            return IPSEC_FAIL_STATUS
        end
        return IPSEC_SUCCESS_STAUS
    end
end

function convert_uci_to_json_for_del(section_name)
    local http_form = {}
    ret, errcode = IPSEC.get_all_ipsec_conns()
	if not ret then
		--dbg.print_tty(errcode)
		return IPSEC_FAIL_STATUS
	end
	ruleNum = table.getn(ret)
	for i=ruleNum,1,-1
	do
		if (ret[i].name == section_name) then
		
			index = i-1
			
			http_form.data = json.encode(
			{
				["method"] = "delete",
				["params"] = 
				{
					["key"] = "unused",
					["index"] = tostring(index)
				}		
			})
			http_form.form = "conn"
			
			--dbg.print_tty("IPSEC: \r\n")
			--dbg.dumptable(http_form)
			rv = IPSEC.dispatch(http_form)
			if (not rv) or (rv.error_code ~= "0") then
				--dbg.print_tty(rv.error_code)
				return IPSEC_FAIL_STATUS
            end
		end
	end
    return IPSEC_SUCCESS_STAUS
end

function show_all_transform_info()
    local result = IPSEC_SUCCESS_STAUS
    local cnt = 0
    uci_r:foreach("ipsec_cli", "ike_stage2", function (section)
        dbg.print_tty("---------------------------------------------------------------------")
        dbg.print_tty("transform set name:        " .. (section.name or ""))
        dbg.print_tty("encryption:                " .. (section.encrypt or ""))
        dbg.print_tty("hash:                      " .. (section.auth or ""))
        dbg.print_tty("lifetime:                  " .. (section.lifetime or ""))
        dbg.print_tty("pfs:                       " .. (section.pfs or ""))
        dbg.print_tty("encapsulation-mode:        " .. (section.mode or ""))
        cnt = cnt + 1
    end)

    uci_r:set_confdir(CFG_PATH)
    uci_r:foreach("ipsec", "ike_stage2", function (section)
	local ike_stage2_proposals = ""
        for _, v in pairs(section.ike_stage2_proposal) do
            if v then
                ike_stage2_proposals = v .. " " .. ike_stage2_proposals
            end
        end
        dbg.print_tty("---------------------------------------------------------------------")
        dbg.print_tty("transform set name:        " .. (section.name or ""))
        dbg.print_tty("hash-enc:                  " .. (ike_stage2_proposals or ""))
        dbg.print_tty("lifetime:                  " .. (section.lifetime or ""))
        dbg.print_tty("pfs:                       " .. (section.pfs or ""))
        dbg.print_tty("encapsulation-mode:        " .. (section.mode or ""))
        cnt = cnt + 1
    end)
    uci_r:set_confdir(CLI_CFG_PATH)
    if cnt ~= 0 then
        dbg.print_tty("---------------------------------------------------------------------")
    end
    return result
end

function show_transform_info(section_name)
    local result = IPSEC_SUCCESS_STAUS
    local cnt = 0

    if not section_name then 
        return IPSEC_FAIL_STATUS
    end

    uci_r:foreach("ipsec_cli", "ike_stage2", function (section)
        if section.name == section_name then
            dbg.print_tty("---------------------------------------------------------------------")
            dbg.print_tty("transform set name:        " .. (section.name or ""))
            dbg.print_tty("encryption:                " .. (section.encrypt or ""))
            dbg.print_tty("hash:                      " .. (section.auth or ""))
            dbg.print_tty("hash-enc:                  " .. (section.auth or "()") .. "-" .. (section.encrypt or "()"))
            dbg.print_tty("lifetime:                  " .. (section.lifetime or ""))
            dbg.print_tty("pfs:                       " .. (section.pfs or ""))
            dbg.print_tty("encapsulation-mode:        " .. (section.mode or ""))
            cnt = cnt + 1
        end
    end)

    uci_r:set_confdir(CFG_PATH)
    uci_r:foreach("ipsec", "ike_stage2", function (section)
        if section.name and section.name == section_name then
            local ike_stage2_proposals = ""
            for _, v in pairs(section.ike_stage2_proposal) do
                if v then
                    ike_stage2_proposals = v .. " " .. ike_stage2_proposals
                end
            end
            dbg.print_tty("---------------------------------------------------------------------")
            dbg.print_tty("transform set name:        " .. (section.name or ""))
            dbg.print_tty("hash-enc:                  " .. (ike_stage2_proposals))
            dbg.print_tty("lifetime:                  " .. (section.lifetime or ""))
            dbg.print_tty("pfs:                       " .. (section.pfs or ""))
            dbg.print_tty("encapsulation-mode:        " .. (section.mode or ""))
            cnt = cnt + 1
        end
    end)
    uci_r:set_confdir(CLI_CFG_PATH)

    if cnt ~= 0 then
        dbg.print_tty("---------------------------------------------------------------------")
    end
    return result
end

function show_all_policy_info(ike_version)
    local result = IPSEC_SUCCESS_STAUS
    local cnt = 0

    uci_r:foreach("ipsec_cli", "ike_stage1", function (section)
        if section.ike_version == ike_version then
            dbg.print_tty("---------------------------------------------------------------------")
            dbg.print_tty("ike policy name:           " .. (section.name or ""))
            dbg.print_tty("ike_version:               " .. (section.ike_version or ""))
            dbg.print_tty("encryption:                " .. (section.encrypt or ""))
            dbg.print_tty("hash:                      " .. (section.auth or ""))
            dbg.print_tty("group:                     " .. (section.dh_group or ""))
            dbg.print_tty("hash-enc-dh:               " .. (section.auth or "()") .. "-" .. (section.encrypt or "()") .. "-" .. (section.dh_group or ""))
            dbg.print_tty("lifetime:                  " .. (section.lifetime or ""))
            dbg.print_tty("dpd_enable:                " .. (section.dpd_enable or ""))
            dbg.print_tty("dpd_interval:              " .. (section.dpd_interval or ""))
            dbg.print_tty("exchange_mode:             " .. (section.exchange_mode or ""))
            cnt = cnt + 1
        end
    end)

    uci_r:set_confdir(CFG_PATH)
    uci_r:foreach("ipsec", "ike_stage1", function (section)
        if section.ike_version == ike_version then
            local ike_proposals = ""
            dbg.print_tty("---------------------------------------------------------------------")
            dbg.print_tty("ike policy name:           " .. (section.name or ""))
            dbg.print_tty("ike_version:               " .. (section.ike_version ))
            --dbg.print_tty("encryption:                " .. (section.encrypt or ""))
            --dbg.print_tty("hash:                      " .. (section.auth or ""))
            --dbg.print_tty("group:                     " .. (section.dh_group or ""))
            for _, v in pairs(section.ike_proposal) do
                if v then
                    ike_proposals = v .. " " .. ike_proposals
                end
            end
            dbg.print_tty("hash-enc-dh:               " .. ike_proposals)
            dbg.print_tty("lifetime:                  " .. (section.lifetime or ""))
            dbg.print_tty("dpd_enable:                " .. (section.dpd_enable or ""))
            dbg.print_tty("dpd_interval:              " .. (section.dpd_interval or ""))
            dbg.print_tty("exchange_mode:             " .. (section.exchange_mode or ""))
            cnt = cnt + 1
        end
    end)
    uci_r:set_confdir(CLI_CFG_PATH)
    
    if cnt ~= 0 then
        dbg.print_tty("---------------------------------------------------------------------")
    end
    return result
end

function show_policy_info(section_name, ike_version)
    local result = IPSEC_SUCCESS_STAUS
    local cnt = 0
    
    if not section_name then 
        return IPSEC_FAIL_STATUS
    end

    uci_r:foreach("ipsec_cli", "ike_stage1", function (section)
        if section.name == section_name and section.ike_version == ike_version then
            dbg.print_tty("---------------------------------------------------------------------")
            dbg.print_tty("ike policy name:           " .. (section.name or ""))
            dbg.print_tty("ike_version:               " .. (section.ike_version ))
            dbg.print_tty("encryption:                " .. (section.encrypt or ""))
            dbg.print_tty("hash:                      " .. (section.auth or ""))
            dbg.print_tty("group:                     " .. (section.dh_group or ""))
            dbg.print_tty("hash-enc-dh:               " .. (section.auth or "()") .. "-" .. (section.encrypt or "()") .. "-" .. (section.dh_group or ""))
            dbg.print_tty("lifetime:                  " .. (section.lifetime or ""))
            dbg.print_tty("dpd_enable:                " .. (section.dpd_enable or ""))
            dbg.print_tty("dpd_interval:              " .. (section.dpd_interval or ""))
            dbg.print_tty("exchange_mode:             " .. (section.exchange_mode or ""))
            cnt = cnt + 1
        end
    end)

    uci_r:set_confdir(CFG_PATH)
    uci_r:foreach("ipsec", "ike_stage1", function (section)
        if section.name == section_name and section.ike_version == ike_version then
            local ike_proposals = ""
            dbg.print_tty("---------------------------------------------------------------------")
            dbg.print_tty("ike policy name:           " .. (section.name or ""))
            dbg.print_tty("ike_version:               " .. (section.ike_version ))
            --dbg.print_tty("encryption:                " .. (section.encrypt or ""))
            --dbg.print_tty("hash:                      " .. (section.auth or ""))
            --dbg.print_tty("group:                     " .. (section.dh_group or ""))
            for _, v in pairs(section.ike_proposal) do
                if v then
                    ike_proposals = v .. " " .. ike_proposals
                end
            end
            dbg.print_tty("hash-enc-dh:               " .. ike_proposals)
            dbg.print_tty("lifetime:                  " .. (section.lifetime or ""))
            dbg.print_tty("dpd_enable:                " .. (section.dpd_enable or ""))
            dbg.print_tty("dpd_interval:              " .. (section.dpd_interval or ""))
            dbg.print_tty("exchange_mode:             " .. (section.exchange_mode or ""))
            cnt = cnt + 1
        end
    end)
    uci_r:set_confdir(CLI_CFG_PATH)

    if cnt ~= 0 then
        dbg.print_tty("---------------------------------------------------------------------")
    end
    return result
end

function show_all_map_info()
    local result = IPSEC_SUCCESS_STAUS
    local cnt = 0
    uci_r:foreach("ipsec_cli", "map", function (section)
            dbg.print_tty("---------------------------------------------------------------------")
            dbg.print_tty("crypto map name:           " .. (section.name or ""))
            dbg.print_tty("policy:                    " .. (section.policy or ""))
            dbg.print_tty("transform_set:             " .. (section.transform_set or ""))
            dbg.print_tty("peer:                      " .. (section.remote_peer or ""))
            dbg.print_tty("localsubnet:               " .. (section.local_network or ""))
            dbg.print_tty("remotesubnet:              " .. (section.remote_network or ""))
            dbg.print_tty("initial-mode:              " .. (section.connection_type or ""))
            cnt = cnt + 1
    end)

    uci_r:set_confdir(CFG_PATH)
    uci_r:foreach("ipsec", "connection", function (section)
            dbg.print_tty("---------------------------------------------------------------------")
            dbg.print_tty("crypto map name:           " .. (section.name or ""))
            dbg.print_tty("policy:                    " .. (section.ike_stage1 or ""))
            dbg.print_tty("transform_set:             " .. (section.ike_stage2 or ""))
            dbg.print_tty("peer:                      " .. (section.remote_peer or ""))
            dbg.print_tty("localsubnet:               " .. (section.local_network or ""))
            dbg.print_tty("remotesubnet:              " .. (section.remote_network or ""))
            dbg.print_tty("initial-mode:              " .. (section.connection_type or ""))
            cnt = cnt + 1
    end)
    uci_r:set_confdir(CLI_CFG_PATH)

    if cnt ~= 0 then
        dbg.print_tty("---------------------------------------------------------------------")
    end
    return result
end

function show_map_info(section_name)
    local result = IPSEC_SUCCESS_STAUS
    local cnt = 0
    uci_r:foreach("ipsec_cli", "map", function (section)
        if section.name == section_name then
            dbg.print_tty("---------------------------------------------------------------------")
            dbg.print_tty("crypto map name:           " .. (section.name or ""))
            dbg.print_tty("policy:                    " .. (section.policy or ""))
            dbg.print_tty("transform_set:             " .. (section.transform_set or ""))
            dbg.print_tty("peer:                      " .. (section.remote_peer or ""))
            dbg.print_tty("localsubnet:               " .. (section.local_network or ""))
            dbg.print_tty("remotesubnet:              " .. (section.remote_network or ""))
            dbg.print_tty("initial-mode:              " .. (section.connection_type or ""))
            cnt = cnt + 1
        end
    end)

    uci_r:set_confdir(CFG_PATH)
    uci_r:foreach("ipsec", "connection", function (section)
        if section.name == section_name then
            dbg.print_tty("---------------------------------------------------------------------")
            dbg.print_tty("crypto map name:           " .. (section.name or ""))
            dbg.print_tty("policy:                    " .. (section.ike_stage1 or ""))
            dbg.print_tty("transform_set:             " .. (section.ike_stage2 or ""))
            dbg.print_tty("peer:                      " .. (section.remote_peer or ""))
            dbg.print_tty("localsubnet:               " .. (section.local_network or ""))
            dbg.print_tty("remotesubnet:              " .. (section.remote_network or ""))
            dbg.print_tty("initial-mode:              " .. (section.connection_type or ""))
            cnt = cnt + 1
        end
    uci_r:set_confdir(CLI_CFG_PATH)
    end)
    if cnt ~= 0 then
        dbg.print_tty("---------------------------------------------------------------------")
    end
    return IPSEC_SUCCESS_STAUS
end

function show_crypto_ipsec_sa(map_name)
    local ret = {}
    local cnt = 0
    --ipsec tun
	local http_form = {}
	http_form.data = json.encode(
	{
		["method"] = "get",
		["params"] = 
		{
		}
	})
	http_form.form = "ipsec_sa"
	
	--dbg.dumptable(http_form)
	ret = IPSEC.dispatch(http_form)
    if ret ~= nil then
        local result = ret.result
        for i, v in pairs(result) do
            if v.name == map_name then
                dbg.print_tty("---------------------------------------------------------------------")
                dbg.print_tty("ipsec sa name:             " .. v.name)
                dbg.print_tty("ipsec sa direction:        " .. v.direction)
                dbg.print_tty("ipsec sa tunnel:          " .. v.tunnel)
                dbg.print_tty("data flow:                " .. v.flows)
                dbg.print_tty("spi:                       " .. v.spi)
                dbg.print_tty("encapsule:                 " .. v.encapsule)
                dbg.print_tty("encryption alogrithm:      " .. v.encapsule_alg)
                dbg.print_tty("hash alogrithm:            " .. v.esp_auth_alg)
                cnt = cnt + 1
            end
        end
    end
    if cnt ~= 0 then
        dbg.print_tty("---------------------------------------------------------------------")
    end
    return IPSEC_SUCCESS_STAUS
end
function set_ipsec_cfg(arg)
    local result = IPSEC_SUCCESS_STAUS
    local cmd = arg[1]
    local section_name
    if arg[2] then
        section_name = arg[2]
    end
    
    if not cmd then
        return IPSEC_FAIL_STATUS
    end

    --crypto ikev1 policy <pol_name>
    if cmd == "crypto ikev1 policy" or cmd == "crypto ikev2 policy" then
        local name = arg[3]
        section_name = IPSEC_POLICY_SECTION_PRE .. section_name
        --检验参数正确性
        local ret = check_exist(IPSEC_POLICY_SECTION_PRE, "ike_stage1", name)
        --dbg.print_tty(ret)
        if ret == 0 then
            uci_r:set("ipsec_cli", section_name, "ike_stage1")
            uci_r:set("ipsec_cli", section_name, "name", name)
            if cmd == "crypto ikev1 policy" then
                uci_r:set("ipsec_cli", section_name, "ike_version", "ikev1")
            else
                uci_r:set("ipsec_cli", section_name, "ike_version", "ikev2")
            end
            --set default value fisrt
            uci_r:set("ipsec_cli", section_name, "lifetime", DEFAULT_LIFETIME)
            uci_r:set("ipsec_cli", section_name, "exchange_mode", DEFAULT_EXCHANGE_MODE)
            uci_r:set("ipsec_cli", section_name, "dpd_enable", DEFAULT_DPD_ENABLE)
            uci_r:set("ipsec_cli", section_name, "dpd_interval", DEFAULT_DPD_INTERVAL)
        else
            --dbg.print_tty(name .. "has been created!!!")
        end
        return result
    end

    -- dpd enable
    if cmd == "dpd enable" then
        name = section_name
        section_name = IPSEC_POLICY_SECTION_PRE .. section_name

        local ret = check_exist(IPSEC_POLICY_SECTION_PRE, "ike_stage1", name)
        --dbg.print_tty(ret)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:set("ipsec_cli", section_name, "dpd_enable", "enable")
        elseif ret == 2 then
            --dbg.print_tty("start to modify config")
            result = modify_used_config("ike_stage1", name, "dpd_enable", "enable")
        end
        return result
    end

    -- dpd disable
    if cmd == "dpd disable" then
        name = section_name
        section_name = IPSEC_POLICY_SECTION_PRE .. section_name

        local ret = check_exist(IPSEC_POLICY_SECTION_PRE, "ike_stage1", name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:set("ipsec_cli", section_name, "dpd_enable", "disable")
        elseif ret == 2 then
            --dbg.print_tty("start to modify config")
            result = modify_used_config("ike_stage1", name, "dpd_enable", "disable")
        end
        return result
    end

    -- dpd-interval <time_out>
    if cmd == "dpd-interval" then
        local val = arg[3]
        name = section_name
        section_name = IPSEC_POLICY_SECTION_PRE .. section_name

        local ret = check_exist(IPSEC_POLICY_SECTION_PRE, "ike_stage1", name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            local dpd_enable = uci_r:get("ipsec_cli", section_name, "dpd_enable")
            if dpd_enable ~= nil and dpd_enable == "enable" then
                uci_r:set("ipsec_cli", section_name, "dpd_interval", val)
            else
                result = IPSEC_FAIL_STATUS
            end
            uci_r:set("ipsec_cli", section_name, "dpd_enable", "enable")
        elseif ret == 2 then
            --dbg.print_tty("start to modify config")
            result = modify_used_config("ike_stage1", name, "dpd_interval", val)
        end
        --[[
        local dpd_enable = uci_r:get("ipsec_cli", section_name, "dpd_enable")
        if dpd_enable ~= nil and dpd_enable == "enable" then
            uci_r:set("ipsec_cli", section_name, "dpd_interval", val)
        else
            result = IPSEC_FAIL_STATUS
        end
        ]]--
        return result
    end

    --encryption <alg_name>
    if cmd == "encryption" then
        local cli_name = IPSEC_POLICY_SECTION_PRE .. section_name
        local val = arg[3]
        local ret
        local res

        local ret = check_exist(IPSEC_POLICY_SECTION_PRE, "ike_stage1", section_name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            if val == "des" then
                local hash_val = uci_r:get("ipsec_cli", cli_name, "auth")
                --dbg.print_tty("hash: " .. hash_val)
                if hash_val ~= nil and (hash_val == "sha256" or hash_val == "sha384" or hash_val == "sha512") then
                    dbg.print_tty("Error: Invalid proposal.")
                    return IPSEC_FAIL_STATUS
                end
            end
            uci_r:set("ipsec_cli", cli_name, "encrypt", val)
            uci_r:commit("ipsec_cli")
            res = refresh_ike_proposal(cli_name)
            if not res then
                --dbg.print_tty("refresh_ike_proposal failed")
            end
        elseif ret == 2 then
            --dbg.print_tty("start to modify config")
            local ike_proposal, encryption, hash, dh_group
            res, ike_proposal = get_option_from_def_path("ike_stage1", section_name, "ike_proposal")
            if not res then
                return IPSEC_FAIL_STATUS
            end
            --dbg.print_tty("ike_proposal " .. ike_proposal)
            hash, encryption, dh_group = ike_proposal:match("(%w+)%-(%w+)%-(%w+)")
            if encryption == val then 
                return result
            end
            if val == "des" then
                if  hash ~= nil and (hash == "sha256" or hash == "sha384" or hash == "sha512") then
                    dbg.print_tty("Error: Invalid proposal.")
                    return IPSEC_FAIL_STATUS
                end
            end
            ike_proposal = hash .. "-" .. val .. "-" .. dh_group
            --dbg.print_tty("ike_proposal " .. ike_proposal)
            result = modify_used_config("ike_stage1", section_name, "ike_proposal", ike_proposal)
        end
        return result
    end

    --hash <alg_name>
    if cmd == "hash" then
        local cli_name = IPSEC_POLICY_SECTION_PRE .. section_name
        local val = arg[3]
        local ret
        local res

        ret = check_exist(IPSEC_POLICY_SECTION_PRE, "ike_stage1", section_name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            if val == "sha256" or val == "sha384" or val == "sha512" then
                local enc_val = uci_r:get("ipsec_cli", cli_name, "encrypt")
                local gr_val = uci_r:get("ipsec_cli", cli_name, "dh_group")
                --dbg.print_tty("encryption: " .. enc_val)
                --dbg.print_tty("group: " .. gr_val)
                if enc_val ~= nil and (enc_val == "des") then
                    dbg.print_tty("Error: Invalid proposal.")
                    return IPSEC_FAIL_STATUS
                elseif gr_val ~= nil and (gr_val == "modp768" or gr_val == "modp1024") then
                    dbg.print_tty("Error: Invalid proposal.")
                    return IPSEC_FAIL_STATUS
                end
            end
            uci_r:set("ipsec_cli", cli_name, "auth", val)
            uci_r:commit("ipsec_cli")
            res = refresh_ike_proposal(cli_name)
            if not res then
                --dbg.print_tty("refresh_ike_proposal failed")
            end
        elseif ret == 2 then
            --dbg.print_tty("start to modify config")
            local ike_proposal, encryption, hash, dh_group
            res, ike_proposal = get_option_from_def_path("ike_stage1", section_name, "ike_proposal")
            if not res then
                return IPSEC_FAIL_STATUS
            end
            --dbg.print_tty("ike_proposal " .. ike_proposal)
            hash, encryption, dh_group = ike_proposal:match("(%w+)%-(%w+)%-(%w+)")
            if hash == val then 
                return result
            end
            if val == "sha256" or val == "sha384" or val == "sha512" then
                if encryption == "des" then
                    dbg.print_tty("Error: Invalid proposal.")
                    return IPSEC_FAIL_STATUS
                end
                if dh_group == "modp768" or dh_group == "modp1024" then
                    dbg.print_tty("Error: Invalid proposal.")
                    return IPSEC_FAIL_STATUS
                end
            end            
            ike_proposal = val .. "-" .. encryption .. "-" .. dh_group
            --dbg.print_tty("ike_proposal " .. ike_proposal)
            result = modify_used_config("ike_stage1", section_name, "ike_proposal", ike_proposal)
        end
        return result
    end

    --group <group_num>
    if cmd == "group" then
        local dh_group = arg[3]
        local ret
        local res
        local cli_name = IPSEC_POLICY_SECTION_PRE .. section_name

        --dh1
        if (tonumber(dh_group) == 1) then
            dh_group = "modp768"
        --dh2
        elseif (tonumber(dh_group) == 2) then
            dh_group = "modp1024"
        --dh5
        elseif (tonumber(dh_group) == 5) then
            dh_group = "modp1536"
        --dh14
        elseif (tonumber(dh_group) == 14) then
            dh_group = "modp2048"
        --dh15
        elseif (tonumber(dh_group) == 15) then
            dh_group = "modp3072"
        --dh16
        elseif (tonumber(dh_group) == 16) then
            dh_group = "modp4096"
        --dh19
        elseif (tonumber(dh_group) == 19) then
            dh_group = "ecp256"
        --dh20
        elseif (tonumber(dh_group) == 20) then
            dh_group = "ecp384"
        --dh21
        elseif (tonumber(dh_group) == 21) then
            dh_group = "ecp521"
        --dh25
        elseif (tonumber(dh_group) == 25) then
            dh_group = "ecp192"
        --dh26
        elseif (tonumber(dh_group) == 26) then
            dh_group = "ecp224"
        else
            return IPSEC_FAIL_STATUS
        end

       ret = check_exist(IPSEC_POLICY_SECTION_PRE, "ike_stage1", section_name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            if dh_group == "modp768" or dh_group == "modp1024" then
                local hash_val = uci_r:get("ipsec_cli", cli_name, "auth")
                --dbg.print_tty("hash: " .. hash_val)
                if hash_val ~= nil and (hash_val == "sha256" or hash_val == "sha384" or hash_val == "sha512") then
                    dbg.print_tty("Error: Invalid proposal.")
                    return IPSEC_FAIL_STATUS
                end
            end
            uci_r:set("ipsec_cli", cli_name, "dh_group", dh_group)
            uci_r:commit("ipsec_cli")
            res = refresh_ike_proposal(cli_name)
            if not res then
                --dbg.print_tty("refresh_ike_proposal failed")
            end
        elseif ret == 2 then
            --dbg.print_tty("start to modify config")
            local ike_proposal, encryption, hash, dh_group1
            res, ike_proposal = get_option_from_def_path("ike_stage1", section_name, "ike_proposal")
            if not res then
                return IPSEC_FAIL_STATUS
            end
            --dbg.print_tty("ike_proposal " .. ike_proposal)
            hash, encryption, dh_group1 = ike_proposal:match("(%w+)%-(%w+)%-(%w+)")
            if dh_group1 == dh_group then 
                return result
            end
            if dh_group == "modp768" or dh_group == "modp1024" then
                if hash ~= nil and (hash == "sha256" or hash == "sha384" or hash == "sha512") then
                    dbg.print_tty("Error: Invalid proposal.")
                    return IPSEC_FAIL_STATUS
                end
            end
            ike_proposal = hash .. "-" .. encryption .. "-" .. dh_group  
            --dbg.print_tty("ike_proposal " .. ike_proposal)
            result = modify_used_config("ike_stage1", section_name, "ike_proposal", ike_proposal)
        end
        return result
    end

    --crypto isakmp key <key>
    if cmd == "crypto isakmp key" then
        local cli_name = IPSEC_POLICY_SECTION_PRE .. section_name
        local val = arg[3]
        local ret
        
        ret = check_exist(IPSEC_POLICY_SECTION_PRE, "ike_stage1", section_name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:set("ipsec_cli", cli_name, "key", val)
        elseif ret == 2 then
            --dbg.print_tty("start to modify config")
            result = modify_used_config("ike_stage1", section_name, "key", val)
        end
        return result
    end

    --ike-lifetime <time>
    if cmd == "ike-lifetime" then
        local cli_name = IPSEC_POLICY_SECTION_PRE .. section_name
        local val = arg[3]
        local ret

        ret = check_exist(IPSEC_POLICY_SECTION_PRE, "ike_stage1", section_name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:set("ipsec_cli", cli_name, "lifetime", val)
        elseif ret == 2 then
            --dbg.print_tty("start to modify config")
            result = modify_used_config("ike_stage1", section_name, "lifetime", val)
        end
        return result
    end
    --exchange-mode <main | aggressive>
    if cmd == "exchange-mode" then
        local cli_name = IPSEC_POLICY_SECTION_PRE .. section_name
        local val = arg[3]
        local ret
        
        ret = check_exist(IPSEC_POLICY_SECTION_PRE, "ike_stage1", section_name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:set("ipsec_cli", cli_name, "exchange_mode", val)
        elseif ret == 2 then
            --dbg.print_tty("start to modify config")
            result = modify_used_config("ike_stage1", section_name, "exchange_mode", val)
        end
        return result
    end
    
    --crypto ipsec transform-set <transform_name>
    if cmd == "crypto ipsec transform-set" then
        local ret = 0
        local algs;
        local cli_name = IPSEC_TRANSFORM_SECTION_PRE .. section_name
        local ike_stage2_proposal
        local enc = arg[4]
        local hash = arg[5]

        if not enc or not hash then
            return IPSEC_FAIL_STATUS
        end
        
        _enc_mode1, _enc_alg = enc:match("(%w+)%-(%w+)")
        _enc_mode2, _hash_alg = hash:match("(%w+)%-(%w+)")
        if not _enc_mode1 or not _enc_mode2 then
            dbg.print_tty("Error: Not find encapsulation mode from the input.")
            return IPSEC_FAIL_STATUS
        end
        if _enc_mode1 ~= "esp" and _enc_mode1  ~= "ah" then
            dbg.print_tty("Error: Wrong encapsulation mode.")
            return IPSEC_FAIL_STATUS
        end
        if _enc_mode1 ~=_enc_mode2 then
            dbg.print_tty("Error: The encapsulation mode of the encrption and the hash alogrithm is not the same.")
            return IPSEC_FAIL_STATUS
        end
        if enc_algs[_enc_alg] == nil then
            dbg.print_tty("Error: The entry does not exist.")
            return IPSEC_FAIL_STATUS
        end
        if auth_algs[_hash_alg] == nil then
            dbg.print_tty("Error: The entry does not exist.")
            return IPSEC_FAIL_STATUS
        end
        
        if _enc_mode1 == "esp" then
            ike_stage2_proposal = _enc_mode1 .. "-" ..  _hash_alg .. "-" .. _enc_alg
        elseif _enc_mode1 == "ah" then
            ike_stage2_proposal = _enc_mode1 .. "-" ..  _hash_alg
        end
        
        ret = check_exist(IPSEC_TRANSFORM_SECTION_PRE, "ike_stage2", section_name)
        if ret == 0 then 
            uci_r:set("ipsec_cli", cli_name, "ike_stage2")
            uci_r:set("ipsec_cli", cli_name, "name", section_name)
            uci_r:set("ipsec_cli", cli_name, "encapsulation_mode", _enc_mode1)
            uci_r:set("ipsec_cli", cli_name, "encrypt", _enc_alg)
            uci_r:set("ipsec_cli", cli_name, "auth", _hash_alg)
            uci_r:set("ipsec_cli", cli_name, "ike_stage2_proposal", ike_stage2_proposal)
            --set default value
            uci_r:set("ipsec_cli", cli_name, "mode", DEFAULT_TUNNEL_MODE)
            uci_r:set("ipsec_cli", cli_name, "pfs", DEFAULT_PFS)
            uci_r:set("ipsec_cli", cli_name, "lifetime", DEFAULT_LIFETIME)
        elseif ret == 1 then
            uci_r:set("ipsec_cli", cli_name, "encrypt", _enc_alg)
            uci_r:set("ipsec_cli", cli_name, "auth", _hash_alg)
            uci_r:set("ipsec_cli", cli_name, "ike_stage2_proposal", ike_stage2_proposal)
            --set default value
            uci_r:set("ipsec_cli", cli_name, "mode", DEFAULT_TUNNEL_MODE)
            uci_r:set("ipsec_cli", cli_name, "pfs", DEFAULT_PFS)
            uci_r:set("ipsec_cli", cli_name, "lifetime", DEFAULT_LIFETIME)
        elseif ret == 2 then
            --need to modify later
            result = modify_used_config("ike_stage2", section_name, "ike_stage2_proposal", ike_stage2_proposal)
        end
        return result
    end

    --lifetime <time>
    if cmd == "lifetime" then
        local ret
        local val = arg[3]
        local cli_name = IPSEC_TRANSFORM_SECTION_PRE .. section_name

        ret = check_exist(IPSEC_TRANSFORM_SECTION_PRE, "ike_stage2", section_name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:set("ipsec_cli", cli_name, "lifetime", val)
        elseif ret == 2 then
            result = modify_used_config("ike_stage2", section_name, "lifetime", val)
        end
        return result
    end

    --encapsulation-mode <tunnel | transport>
    if cmd == "encapsulation-mode" then
        local ret
        local val = arg[3]
        local cli_name = IPSEC_TRANSFORM_SECTION_PRE .. section_name

        ret = check_exist(IPSEC_TRANSFORM_SECTION_PRE, "ike_stage2", section_name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:set("ipsec_cli", cli_name, "mode", val)
        elseif ret == 2 then
            result = modify_used_config("ike_stage2", section_name, "mode", val)
        end
        return result
    end

    --pfs <none | 1 | 2 | 5 | 14 | 15>
    if cmd == "pfs" then
        local ret
        local val = arg[3]
        local cli_name = IPSEC_TRANSFORM_SECTION_PRE .. section_name

        if val ~= "none" then
            if (tonumber(val) == 1) then
                val = "modp768"
            elseif (tonumber(val) == 2) then
                val = "modp1024"
            elseif (tonumber(val) == 5) then
                val = "modp1536"
            elseif (tonumber(val) == 14) then
                val = "modp2048"
            elseif (tonumber(val) == 15) then
                val = "modp3072"
            end
        end

        ret = check_exist(IPSEC_TRANSFORM_SECTION_PRE, "ike_stage2", section_name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:set("ipsec_cli", cli_name, "pfs", val)
        elseif ret == 2 then
            result = modify_used_config("ike_stage2", section_name, "pfs", val)
        end
        return result
    end

    --show transform-set <set_name>
    if cmd == "show transform-set" then
        return show_transform_info(section_name)
    end
    --show all transform-set
    if cmd == "show all transform-set" then
        return show_all_transform_info()
    end
    --show ikev1 policy [pol_name]
    if cmd == "show ikev1 policy" then
        return show_policy_info(section_name, "ikev1") 
    end

    --show all ikev1 policy
    if cmd == "show all ikev1 policy" then
        return show_all_policy_info("ikev1")
    end

    --show ikev2 policy [pol_name]
    if cmd == "show ikev2 policy" then
        return show_policy_info(section_name, "ikev2")
    end

    --show crypto ipsec sa [map_name]
    if cmd == "show crypto ipsec sa" then
        return show_crypto_ipsec_sa(section_name)
    end

    --show all ikev2 policy
    if cmd == "show all ikev2 policy" then
        return show_all_policy_info("ikev2")
    end

    --show map [map_name]
    if cmd == "show crypto map" then
        return show_map_info(section_name)
    end

    --show all map
    if cmd == "show all crypto map" then
        --dbg.print_tty("show all crypto map")
        return show_all_map_info()
    end

    --negotiation-mode <initiator | responder>
    if cmd == "negotiation-mode" then
        local ret
        local val = arg[3]
        local cli_name = IPSEC_MAP_SECTION_PRE .. section_name

        ret = check_exist(IPSEC_MAP_SECTION_PRE, "connection", section_name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:set("ipsec_cli", cli_name, "connection_type", val)
        elseif ret == 2 then
            result = modify_used_config("name", section_name, "connection_type", val)
        end
        return result
    end

    --set peer <ip_address>
    if cmd == "set peer" then
        local ret
        local val = arg[4]
        local cli_name = IPSEC_MAP_SECTION_PRE .. section_name

        ret = check_exist(IPSEC_MAP_SECTION_PRE, "connection", section_name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            if arg[3] == "ADDR" then
                uci_r:set("ipsec_cli", cli_name, "remote_peer_type", "ip_address")
            elseif arg[3] == "NAME" then
                uci_r:set("ipsec_cli", cli_name, "remote_peer_type", "domain_name")
            else
                return IPSEC_FAIL_STATUS
            end
            uci_r:set("ipsec_cli", cli_name, "remote_peer", val)
        elseif ret == 2 then
            if arg[3] == "ADDR" then
                result = modify_used_config("name", section_name, "remote_peer_type", "ip_address")
            elseif arg[3] == "NAME" then
                result = modify_used_config("name", section_name, "remote_peer_type", "domain_name")
            else
                return IPSEC_FAIL_STATUS
            end
            result = modify_used_config("name", section_name, "remote_peer", val)
        end
        return result
    end

    if cmd == "crypto map" then
        local find_map = 0
        local find_pol = 0
        local val = arg[4]
        local cli_name = IPSEC_MAP_SECTION_PRE .. section_name

        find_map = check_exist(IPSEC_MAP_SECTION_PRE, "connection", section_name)
        find_pol = check_exist(IPSEC_POLICY_SECTION_PRE, "ike_stage1", val)
        if find_pol == 0 then
            dbg.print_tty(string.format("Error: No such policy named %s.", val))
            return IPSEC_FAIL_STATUS
        end
        if find_map == 0 then
            uci_r:set("ipsec_cli", cli_name, "map")
            uci_r:set("ipsec_cli", cli_name, "name", section_name)
            uci_r:set("ipsec_cli", cli_name, "policy", val)
            uci_r:set("ipsec_cli", cli_name, "def_path_pol", find_pol)
            uci_r:set("ipsec_cli", cli_name, "connection_type", DEFAULT_CONNECTION_TYPE)
        elseif find_map == 1 then
            uci_r:set("ipsec_cli", cli_name, "policy", val)
            uci_r:set("ipsec_cli", cli_name, "def_path_pol", find_pol)
        elseif find_map == 2 then
            result = modify_used_config("name", section_name, "ike_stage1", val)
        end
        return result
    end

    --set transfrom-set <tset>
    if cmd == "set transform-set" then
        local find_map = 0
        local find_tset = 0
        local val = arg[3]
        local cli_name = IPSEC_MAP_SECTION_PRE .. section_name

        find_map = check_exist(IPSEC_MAP_SECTION_PRE, "connection", section_name)
        find_tset = check_exist(IPSEC_TRANSFORM_SECTION_PRE, "ike_stage2", val)
        if find_tset == 0 then
            dbg.print_tty(string.format("Error: No such transfrom-set named %s.", val))
            return IPSEC_FAIL_STATUS
        end
        if find_map == 0 then
            return IPSEC_FAIL_STATUS
        elseif find_map == 1 then
            uci_r:set("ipsec_cli", cli_name, "transform_set", val)
            uci_r:set("ipsec_cli", cli_name, "def_path_tset", find_tset)
        elseif find_map == 2 then
            result = modify_used_config("name", section_name, "ike_stage2", val)
        end
        return result
    end

    --localsubnet <ip_address>/<mask>
    --localsubnet <ip_address> <ip_mask>
    if cmd == "localsubnet" then
        local ret = 0
        local val = arg[3]
        local cli_name = IPSEC_MAP_SECTION_PRE .. section_name

        ret = check_exist(IPSEC_MAP_SECTION_PRE, "connection", section_name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:set("ipsec_cli", cli_name, "local_network", val)
        elseif ret == 2 then
            result = modify_used_config("name", section_name, "local_network", val)
        end
        return result
    end

    --remotesubnet <ip_address>/<mask>
    --remotesubnet <ip_address> <ip_mask>
    if cmd == "remotesubnet" then
        local ret = 0
        local val = arg[3]
        local cli_name = IPSEC_MAP_SECTION_PRE .. section_name

        ret = check_exist(IPSEC_MAP_SECTION_PRE, "connection", section_name)
        if ret == 0 then
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:set("ipsec_cli", cli_name, "remote_network", val)
        elseif ret == 2 then
            result = modify_used_config("name", section_name, "remote_network", val)
        end
        return result
    end

    --ipsec map <map_name>
    if cmd == "ipsec map" then
        local http_form
        local ret
        local local_binding
        local vlan_str
        local vlan_info = {}
        local f = io.open(interface_config_file_name, "rb")
        if f then
            vlan_str = f:read("*l")
            --dbg.print_tty(vlan_str)

            vlan_info = json.decode(vlan_str)
            local vlan
            if vlan_info and vlan_info.vlan then
                vlan = vlan_info.vlan
            else
                return IPSEC_FAIL_STATUS
            end
            uci_r:set_confdir(CFG_PATH)
            uci_r:foreach("zone", "zone", function (section)
                if section.dev_eth then
                    if_vlan = string.gsub(section.dev_eth,"[^%.]*%.","")
                    if if_vlan == vlan then
                        local_binding = section[".name"]
                    end
                end
            end)
            uci_r:set_confdir(CLI_CFG_PATH)
            f:close()
        else
            return IPSEC_FAIL_STATUS
        end
        section_name = IPSEC_MAP_SECTION_PRE .. section_name
        --just generate map form ipsec_cli, later realize "modify" operation
        local map = uci_r:get_all("ipsec_cli", section_name)
        if map == nil then
            return IPSEC_FAIL_STATUS
        end
        uci_r:set("ipsec_cli", section_name, "local_binding", local_binding)
        uci_r:commit("ipsec_cli")
        if not check_map_complete(map) then
            return IPSEC_FAIL_STATUS
        end
        map = uci_r:get_all("ipsec_cli", section_name)
        http_form = convert_uci_to_json_for_add(map)
        ret = IPSEC.dispatch(http_form)
        if (not ret) or (ret.error_code ~= "0") then
            --result = err_2_sdnerr(ret.error_code)
            --return false,result
            dbg.print_tty(error_print.check_error(ret))
            return IPSEC_FAIL_STATUS
        end
        uci_r:set_confdir(CLI_CFG_PATH)
        --dbg.print_tty(uci_r:get_confdir())
        if not delete_map(map) then
            return IPSEC_FAIL_STATUS
        end
        return result
    end

    --no crypto ikev1 policy <pol_name>
    if cmd == "no crypto ikev1 policy" then
        local ret

        ret = check_exist(IPSEC_POLICY_SECTION_PRE, "ike_stage1", section_name)
        if ret == 0 then
            dbg.print_tty("Error: No such policy created yet.")
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:delete("ipsec_cli", IPSEC_POLICY_SECTION_PRE .. section_name)
            return result
        elseif ret == 2 then
            dbg.print_tty("Error: The policy is used by other map, please delete the using map first.")
        else
            --dbg.print_tty("unknown error")
            return IPSEC_FAIL_STATUS
        end
    end

    --no crypto ikev2 policy <pol_name>
    if cmd == "no crypto ikev2 policy" then
        local ret

        ret = check_exist(IPSEC_POLICY_SECTION_PRE, "ike_stage1", section_name)
        if ret == 0 then
            dbg.print_tty("Error: No such policy created yet.")
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:delete("ipsec_cli", IPSEC_POLICY_SECTION_PRE .. section_name)
            return result
        elseif ret == 2 then
            dbg.print_tty("Error: The policy is used by other map, please delete the using map first.")
        else
            --dbg.print_tty("unknown error")
            return IPSEC_FAIL_STATUS
        end
    end

    --no crypto tranform_set <tset_name>
    if cmd == "no crypto transform-set" then
        local ret

        ret = check_exist(IPSEC_TRANSFORM_SECTION_PRE, "ike_stage2", section_name)
        if ret == 0 then
            dbg.print_tty("Error: No such transform set created yet.")
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:delete("ipsec_cli", IPSEC_TRANSFORM_SECTION_PRE .. section_name)
            return result
        elseif ret == 2 then
            dbg.print_tty("Error: The transform set is used by other map, please delete the using map first.")
        else
            --dbg.print_tty("unknown error")
            return IPSEC_FAIL_STATUS
        end
    end

    --no crypto map <map_name>
    if cmd == "no crypto map" then
        local ret

        ret = check_exist(IPSEC_MAP_SECTION_PRE, "connection", section_name)
        if ret == 0 then
            dbg.print_tty("Error: No such map created yet.")
            return IPSEC_FAIL_STATUS
        elseif ret == 1 then
            uci_r:delete("ipsec_cli", IPSEC_MAP_SECTION_PRE .. section_name)
            return result
        elseif ret == 2 then
            uci_r:set_confdir(CFG_PATH)
            result = convert_uci_to_json_for_del(section_name)
            uci_r:set_confdir(CLI_CFG_PATH)
            return result
        else
            --dbg.print_tty("unknown error")
            return IPSEC_FAIL_STATUS
        end
    end
    
    return IPSEC_FAIL_STATUS
end



--脚本总入口
local result = IPSEC_SUCCESS_STAUS
result = set_ipsec_cfg(arg)
if result == IPSEC_SUCCESS_STAUS then
    uci_r:commit("ipsec_cli")
end
io.stdout:write(result)
