--[[
Copyright(c) 2008-2014 Shenzhen TP-LINK Technologies Co.Ltd.

File    :  rip.lua
Details :  command parsing of rip model in cli_server
Author  :  Dong Yuqing <dongyuqing@tp-link.com.hk>
Version :  1.0.0
Date    :  5 Sep, 2022
]]--

local sys  = require "luci.sys"
local json = require 'luci.json'
local utl = require "luci.util"
local dbg  = require ('luci.tools.debug')
local form = require "luci.tools.form"
local uci  = require "luci.model.uci"
local rip = require 'luci.controller.admin.rip'
local error_print = require "luci.tools.error_print"

local uci_r = uci.cursor()
local CLI_CFG_PATH = "/tmp/cli_config"
local CFG_PATH = "/etc/config"

--[[local filename = CLI_CFG_PATH.."/cli_rip"
local file = io.open(filename,"r")
if file == nil then
    sys.exec("touch " .. filename)
else
    file:close()
end]]--

function set_rip_enabled(enabled)
    local result  = "SUCCESS"
    local http_form = {}
    local enable_on_off
    if (enabled == "enable") then
        --若rip进程已开启则直接返回
        local rip_status = uci_r:get("rip", "global", "status")
        if rip_status == "on" then
            return result
        end

        enable_on_off = "on"
    elseif (enabled == "disable") then
        --若rip进程已关闭则直接返回
        local rip_status = uci_r:get("rip", "global", "status")
        if rip_status == "off" then
            return result
        end

        enable_on_off = "off"
    else
        dbg.print_tty("Error: Only support the parameter of 'enable' or 'disable'.")
        return "FAILED"
    end

    http_form.form = "enable_cfg"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] = 
        {   
            ["enable"] = enable_on_off,
        }
    })

    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        return "FAILED"
    end

    return result

end

local function set_rip_version(version)
    local result = "SUCCESS"
    if (string.find(version, "ripv1") == nil and string.find(version, "ripv2") == nil and string.find(version, "default") == nil) then 
        dbg.print_tty("Error: Only support the parameter of 'ripv1' or 'ripv2' or 'default'.")
        return "FAILED"
    end
    local http_form = {}
    http_form.form = "config"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] = 
        {   
            ["rip_version"] = version,
        }
    })

    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end

    return result
end


local function set_rip_distance(distance)
    local result = "SUCCESS"
    local http_form = {}

    http_form.form = "config"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] = 
        {   
            ["rip_distance"] = tostring(distance),
        }
    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end

    return result
end

local function set_auto_summary_enabled(auto_summary_enabled)
    local result = "SUCCESS"
    local http_form = {}

    http_form.form = "config"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] = 
        {   
            ["auto_summary"] = auto_summary_enabled,
        }
    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end

    return result
end

local function set_default_metric(default_metric)
    local result = "SUCCESS"
    local http_form = {}

    http_form.form = "config"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] = 
        {   
            ["default_metric"] = default_metric,
        }
    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end

    return result
end

local function set_redistribute_static(redistribute_static_enabled, redistribute_static_metric)
    local result = "SUCCESS"
    local http_form = {}

    http_form.form = "config"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] = 
        {   
            ["redistribute_static"] = redistribute_static_enabled,
            ["redistribute_static_metric"] = redistribute_static_metric,
        }
    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end

    return result
end

local function set_redistribute_ospf(redistribute_ospf_enabled, redistribute_ospf_metric)
    local result = "SUCCESS"
    local http_form = {}

    http_form.form = "config"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] = 
        {   
            ["redistribute_ospf"] = redistribute_ospf_enabled,
            ["redistribute_ospf_metric"] = redistribute_ospf_metric,
        }
    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end

    return result
end

local function set_update_timer(update_timer_value)
    local result = "SUCCESS"
    local http_form = {}

    http_form.form = "config"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] = 
        {   
            ["update_timer"] = update_timer_value,
        }
    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end

    return result
end

local function set_timeout_timer(timeout_timer_value)
    local result = "SUCCESS"
    local http_form = {}

    http_form.form = "config"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] = 
        {   
            ["timeout_timer"] = timeout_timer_value,
        }
    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end

    return result
end

local function set_garbage_timer(garbage_timer_value)
    local result = "SUCCESS"
    local http_form = {}

    http_form.form = "config"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["params"] = 
        {   
            ["garbage_timer"] = garbage_timer_value,
        }
    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end

    return result
end

local function mask_to_masklen(mask)
    local IP_ADDR_BITS = 32
    local MAX_UINT_32 = 4294967295

    local b1, b2, b3, b4 = mask:match("^(%d+)%.(%d+)%.(%d+)%.(%d+)$")
    local ip_num = bit.lshift(b1, 24) + bit.lshift(b2, 16) + bit.lshift(b3, 8) + b4
    local highest_bit = bit.rshift(ip_num, 31);

    if (highest_bit == 0) then
        ip_num = bit.bxor(ip_num, MAX_UINT_32)
    end
    local wild_mask_ip_num = bit.bxor(ip_num, MAX_UINT_32)

    local cnt_zero = 0
    while cnt_zero <= IP_ADDR_BITS do
        if (bit.band(ip_num, 1) == 0) then
            cnt_zero = cnt_zero + 1;
        else
            break
        end
        ip_num = bit.rshift(ip_num, 1);
    end

    local cnt_one = 0
    while cnt_one <= IP_ADDR_BITS do
        if (bit.band(ip_num, 1) == 1) then
            cnt_one = cnt_one + 1;
        else
            break
        end
        ip_num = bit.rshift(ip_num, 1);
    end

    dbg("mask_to_masklen", cnt_zero, cnt_one, wild_mask_ip_num)
    if (cnt_zero + cnt_one ~= IP_ADDR_BITS) then
        return nil
    end
    return cnt_one, wild_mask_ip_num
end

local function set_network(method, ip_address, mask)
    local result = "SUCCESS"
    local http_form = {}
    if string.sub(ip_address,1,3) == "255" then
        dbg.print_tty("Error: The address cannot start with 255.")
        return "FAILED"
    end
    http_form.form = "network"
    if (method == "add") then
        if mask_to_masklen(mask) == nil then
            dbg.print_tty("Error: Invalid mask format.")
            return "FAILED"
        end
        http_form.data = json.encode(
        {
            ["method"] = method,
            ["params"] = 
            {   
                ["new"] = {["ipaddress"] = ip_address,
                            ["mask"] = mask},
            }
        })   
    elseif (method == "delete") then
        local index = 0
        local delete_index = -1
        uci_r:set_confdir(CFG_PATH)
        uci_r:foreach("rip", "network",
            function(section)
                if (section.ipaddress == ip_address) then
                    delete_index = index
                end
                index = index + 1
            end
        )
        if (delete_index >= 0) then
            http_form.data = json.encode(
            {
                ["method"] = method,
                ["params"] = 
                {   
                    ["key"] = ip_address,
                    ["index"] = tostring(delete_index),
                }
            })
        else
            dbg.print_tty("Error: Network does not exit.")
            return "FAILED"
        end
    end
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    elseif ret.result.ipaddress ~= nil and ret.result.id == nil then
        dbg.print_tty("Error: Network " .. ret.result.ipaddress .. " already exists.")
        result = "FAILED"
    end

    return result
end

local function set_receive_version(receive_version, ifname)
    local result = "SUCCESS"
    if (string.find(receive_version, "ripv1") == nil and string.find(receive_version, "ripv2") == nil and string.find(receive_version, "both") == nil) then 
        dbg.print_tty("Error: Only support the parameter of 'ripv1' or 'ripv2' or 'both'.")
        return "FAILED"
    end
    local http_form = {}
    http_form.form = "interface_cfg"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["config_source"] = "cli",
        ["params"] = 
        {   
            ["old"] = {["interface"] = ifname},
            ["new"] = {["receive_version"] = receive_version,},
        }
    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end

    return result
end

local function set_send_version(send_version, ifname)
    local result = "SUCCESS"
    if (string.find(send_version, "ripv1") == nil and string.find(send_version, "ripv2") == nil) then 
        dbg.print_tty("Error: Only support the parameter of 'ripv1' or 'ripv2'.")
        return "FAILED"
    end
    local http_form = {}
    http_form.form = "interface_cfg"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["config_source"] = "cli",
        ["params"] = 
        {   
            ["old"] = {["interface"] = ifname},
            ["new"] = {["send_version"] = send_version,},
        }
    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end

    return result
end

local function set_split_horizon_mode(split_horizon_mode, ifname)
    local result = "SUCCESS"
    local http_form = {}

    http_form.form = "interface_cfg"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["config_source"] = "cli",
        ["params"] = 
        {   
            ["old"] = {["interface"] = ifname},
            ["new"] = {["split_horizon_mode"] = split_horizon_mode,},
        }
    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end

    return result
end

local function set_authentication_mode(mode, key, key_id, ifname)
    local result = "SUCCESS"
    local http_form = {}

    http_form.form = "interface_cfg"
    http_form.data = json.encode(
    {
        ["method"] = "set",
        ["config_source"] = "cli",
        ["params"] = 
        {   
            ["old"] = {["interface"] = ifname},
            ["new"] = 
            {
                ["authen_mode"] = mode,
                ["key"] = key,
                ["key_id"] = key_id,
            },
        }
    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end

    return result
end

local function show_ip_rip()
    local result = "SUCCESS"
    local http_form = {}

    http_form.form = "enable_cfg"
    http_form.data = json.encode(
    {
        ["method"] = "get"
    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end
    local result_enable_cfg = ret.result
    local enable_on_off = result_enable_cfg.enable
    local rip_enable = ""
    if (enable_on_off == "on") then
        rip_enable = "enabled"
    elseif (enable_on_off == "off") then
        rip_enable = "disabled"
    end

    http_form.form = "config"
    http_form.data = json.encode(
    {
        ["method"] = "get"
    })
    ret = rip.dispatch(http_form)
    local result_config = ret.result

    http_form.form = "network"
    http_form.data = json.encode(
    {
        ["method"] = "get",
        ["params"] = {}

    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end
    local result_network = ret.result

    http_form.form = "interface_cfg"
    http_form.data = json.encode(
    {
        ["method"] = "get",
        ["params"] = {}

    })
    ret = rip.dispatch(http_form)
    if (not ret) or (ret.error_code ~= "0") then
        dbg.print_tty(error_print.check_error(ret))
        result = "FAILED"
    end
    local result_interface_cfg = ret.result
    dbg.print_tty("---------------------------------------------------------------------")
    dbg.print_tty("RIP Enable")
    dbg.print_tty("RIP Protocol:" .. "                  " .. rip_enable)
    dbg.print_tty("---------------------------------------------------------------------")
    dbg.print_tty("Global Config")
    dbg.print_tty("RIP Version:" .. "                   " .. result_config.rip_version)
    dbg.print_tty("RIP Distance:" .. "                  " .. result_config.rip_distance)
    dbg.print_tty("Auto Summary:" .. "                  " .. result_config.auto_summary)
    dbg.print_tty("Update Timer:" .. "                  " .. result_config.update_timer)
    dbg.print_tty("Timeout Timer:" .. "                 " .. result_config.timeout_timer)
    dbg.print_tty("Garbage Timer:" .. "                 " .. result_config.garbage_timer)
    dbg.print_tty("---------------------------------------------------------------------")
    dbg.print_tty(string.format("%-18s", "RIP Network List") .. string.format("%-18s", "Mask"))
    for _,v in ipairs(result_network) do
        dbg.print_tty(string.format("%-18s", v.ipaddress) .. string.format("%-18s", v.mask))
    end
    dbg.print_tty("---------------------------------------------------------------------------------------------------------------------------------------------------------")
    dbg.print_tty("Interface Config")
    dbg.print_tty(string.format("%-15s", "Name") .. string.format("%-15s", "Interface") .. string.format("%-18s", "IP Adress") .. string.format("%-15s", "Status")
                 .. string.format("%-15s", "Send Version") .. string.format("%-18s", "Receive Version") .. string.format("%-22s", "Split horizon mode") .. string.format("%-15s", "Authen Mode")
                  .. string.format("%-15s", "Key ID") .. string.format("%-16s", "Key"))
    for _,v in ipairs(result_interface_cfg) do
        dbg.print_tty(string.format("%-15s", v.interface) .. string.format("%-15s", v.interface_eth)  .. string.format("%-18s", v.ipaddress) .. string.format("%-15s", v.status)
                     .. string.format("%-15s", v.send_version)  .. string.format("%-18s", v.receive_version) .. string.format("%-22s", v.split_horizon_mode) .. string.format("%-15s", v.authen_mode)
                      .. string.format("%-15s", (v.key_id or "--")) .. string.format("%-16s", (v.key or "--")))
    end
    dbg.print_tty("---------------------------------------------------------------------------------------------------------------------------------------------------------")
    return result
end

local result = "SUCCESS"
if arg[1] == "router rip" then
    local enabled = "enable"
    result = set_rip_enabled(enabled)
elseif arg[1] == "no router rip" then
    local enabled = "disable"
    result = set_rip_enabled(enabled)

elseif arg[1] == "version" then 
    local version = arg[2]
    result = set_rip_version(version)

elseif arg[1] == "no version" then 
    local version = "default"
    result = set_rip_version(version)

elseif arg[1] == "distance" then
    local distance = arg[2]
    result = set_rip_distance(distance)

elseif arg[1] == "no distance" then
    local distance = "120"
    result = set_rip_distance(distance)

elseif arg[1] == "auto-summary" then
    local auto_summary_enabled = "enable"
    result = set_auto_summary_enabled(auto_summary_enabled)

elseif arg[1] == "no auto-summary" then
    local auto_summary_enabled = "disable"
    result = set_auto_summary_enabled(auto_summary_enabled) 

elseif arg[1] == "default-metric" then
    local default_metric = arg[2]
    result = set_default_metric(default_metric)

elseif arg[1] == "no default-metric" then
    local default_metric = "1"
    result = set_default_metric(default_metric)

elseif arg[1] == "redistribute static metric" then
    local redistribute_static_metric = arg[2]
    local redistribute_static_enabled = "enable"
    result = set_redistribute_static(redistribute_static_enabled, redistribute_static_metric)

elseif arg[1] == "no redistribute static" then
    uci_r:set_confdir(CFG_PATH)
    local redistribute_static_metric = uci_r:get("rip", "config", "redistribute_static_metric")
    local redistribute_static_enabled = "disable"
    result = set_redistribute_static(redistribute_static_enabled, redistribute_static_metric)

elseif arg[1] == "redistribute ospf metric" then
    local redistribute_ospf_metric = arg[2]
    local redistribute_ospf_enabled = "enable"
    result = set_redistribute_ospf(redistribute_ospf_enabled, redistribute_ospf_metric)

elseif arg[1] == "no redistribute ospf" then
    uci_r:set_confdir(CFG_PATH)
    local redistribute_ospf_metric = uci_r:get("rip", "config", "redistribute_ospf_metric")
    local redistribute_ospf_enabled = "disable"
    result = set_redistribute_ospf(redistribute_ospf_enabled, redistribute_ospf_metric)

elseif arg[1] == "timers basic update" then
    local update_timer_value = arg[2]
    result = set_update_timer(update_timer_value)

elseif arg[1] == "timers basic timeout" then
    local timeout_timer_value = arg[2]
    result = set_timeout_timer(timeout_timer_value)

elseif arg[1] == "timers basic garbage-collect" then
    local garbage_timer_value = arg[2]
    result = set_garbage_timer(garbage_timer_value)

elseif arg[1] == "no timers basic" then
    local update_timer_value = "30"
    result = set_update_timer(update_timer_value)
    local timeout_timer_value = "180"
    result = set_timeout_timer(timeout_timer_value)
    local garbage_timer_value = "240"
    result = set_garbage_timer(garbage_timer_value)

elseif arg[1] == "network" then
    local ip_address = arg[2]
    local mask = arg[3]
    local method = "add"
    result = set_network(method, ip_address, mask)

elseif arg[1] == "no network" then
    local ip_address = arg[2]
    local method = "delete"
    result = set_network(method, ip_address)

elseif arg[1] == "ip rip receive version" then
    local receive_version = arg[2]
    local ifname = arg[3]
    result = set_receive_version(receive_version, ifname)

elseif arg[1] == "no ip rip receive version" then
    local ifname = arg[2]
    local rip_version = uci_r:get("rip", "config", "rip_version")
    if (rip_version == "default") then 
        rip_version = "both"
    end
    result = set_receive_version(rip_version, ifname)   

elseif arg[1] == "ip rip send version" then
    local send_version = arg[2]
    local ifname = arg[3]
    result = set_send_version(send_version, ifname)

elseif arg[1] == "no ip rip send version" then
    local ifname = arg[2]
    local rip_version = uci_r:get("rip", "config", "rip_version")    
    if (rip_version == "default") then
        rip_version = "ripv2" 
    end   
    result = set_send_version(rip_version, ifname)

elseif arg[1] == "ip rip poison-reverse" then
    local split_horizon_mode = "poison_reverse"
    local ifname = arg[2]
    result = set_split_horizon_mode(split_horizon_mode, ifname)

elseif arg[1] == "no ip rip poison-reverse" then
    local split_horizon_mode = "none"
    local ifname = arg[2]
    result = set_split_horizon_mode(split_horizon_mode, ifname)

elseif arg[1] == "ip rip split-horizon" then
    local split_horizon_mode = "split_horizon"
    local ifname = arg[2]
    result = set_split_horizon_mode(split_horizon_mode, ifname)

elseif arg[1] == "no ip rip split-horizon" then
    local split_horizon_mode = "none"
    local ifname = arg[2]
    result = set_split_horizon_mode(split_horizon_mode, ifname)

elseif arg[1] == "ip rip authentication-mode simple" then
    local mode = "simple"
    local key_id = nil
    local key = arg[2]
    local ifname = arg[3]
    result = set_authentication_mode(mode, key, key_id, ifname)

elseif arg[1] == "ip rip authentication-mode md5" then
    local mode = "md5"
    local key_id = arg[2]
    local key = arg[3]
    local ifname = arg[4]
    result = set_authentication_mode(mode, key, key_id, ifname)

elseif arg[1] == "no ip rip authentication-mode" then
    local mode = "none"
    local key_id = nil
    local key = ""
    local ifname = arg[2]
    result = set_authentication_mode(mode, key, key_id, ifname)

elseif arg[1] == "show ip rip" then
    result = show_ip_rip()
end
io.stdout:write(result)