/*
 * sfe_ipv4.c
 *	Shortcut forwarding engine - IPv4 edition.
 *
 * Copyright (c) 2013-2016 The Linux Foundation. All rights reserved.
 * Permission to use, copy, modify, and/or distribute this software for
 * any purpose with or without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all copies.
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <linux/module.h>
#include <linux/sysfs.h>
#include <linux/skbuff.h>
#include <linux/icmp.h>
#include <linux/version.h>
#include <net/tcp.h>
#include <linux/etherdevice.h>
#include <net/netfilter/nf_conntrack.h>
#include <linux/if_pppox.h>
#include <linux/ppp_defs.h>
#include <linux/if_vlan.h>
#include "sfe.h"
#include "sfe_cm.h"
#include "sfe_backport.h"
#if  defined(CONFIG_RA_HW_NAT) || defined(CONFIG_RA_HW_NAT_MODULE)
#include "net/ra_nat.h"
#endif
#define TPLINK_DEBUG(s, ...)  //printk(KERN_ERR""s, ##__VA_ARGS__)

int sfe_debug;
module_param(sfe_debug, int, 0644);
EXPORT_SYMBOL(sfe_debug);
#define SFE_IPSTAT
#ifdef SFE_IPSTAT
#include "sfe_ipstat.h"
#endif

#if defined(CONFIG_XFRM) && defined(SFE_SUPPORT_IPSEC)
#include <crypto/aead.h>
#include <crypto/authenc.h>
#include <net/xfrm.h>
#include <net/esp.h>
#include <linux/scatterlist.h>
#define SFE_ACCELERATE_IPSEC (1)
#else
#define SFE_ACCELERATE_IPSEC (0)
#endif

/* L2 proto types */
#define	SFE_PROTO_DEF 0x01
#define	SFE_PROTO_PPPOE 0x02
#define	SFE_PROTO_PPTP 0x04
#define	SFE_PROTO_L2TP 0x08
#define	SFE_PROTO_IPSEC 0x10

/*
 * By default Linux IP header and transport layer header structures are
 * unpacked, assuming that such headers should be 32-bit aligned.
 * Unfortunately some wireless adaptors can't cope with this requirement and
 * some CPUs can't handle misaligned accesses.  For those platforms we
 * define SFE_IPV4_UNALIGNED_IP_HEADER and mark the structures as packed.
 * When we do this the compiler will generate slightly worse code than for the
 * aligned case (on most platforms) but will be much quicker than fixing
 * things up in an unaligned trap handler.
 */
#define SFE_IPV4_UNALIGNED_IP_HEADER 1
#if SFE_IPV4_UNALIGNED_IP_HEADER
#define SFE_IPV4_UNALIGNED_STRUCT __attribute__((packed))
#else
#define SFE_IPV4_UNALIGNED_STRUCT
#endif
int counter = 0;
/*
 * An Ethernet header, but with an optional "packed" attribute to
 * help with performance on some platforms (see the definition of
 * SFE_IPV4_UNALIGNED_STRUCT)
 */
struct sfe_ipv4_eth_hdr {
	__be16 h_dest[ETH_ALEN / 2];
	__be16 h_source[ETH_ALEN / 2];
	__be16 h_proto;
} SFE_IPV4_UNALIGNED_STRUCT;

#define SFE_IPV4_DSCP_MASK 0x3
#define SFE_IPV4_DSCP_SHIFT 2

/*
 * An IPv4 header, but with an optional "packed" attribute to
 * help with performance on some platforms (see the definition of
 * SFE_IPV4_UNALIGNED_STRUCT)
 */
struct sfe_ipv4_ip_hdr {
#if defined(__LITTLE_ENDIAN_BITFIELD)
	__u8 ihl:4,
	     version:4;
#elif defined (__BIG_ENDIAN_BITFIELD)
	__u8 version:4,
	     ihl:4;
#else
#error	"Please fix <asm/byteorder.h>"
#endif
	__u8 tos;
	__be16 tot_len;
	__be16 id;
	__be16 frag_off;
	__u8 ttl;
	__u8 protocol;
	__sum16 check;
	__be32 saddr;
	__be32 daddr;

	/*
	 * The options start here.
	 */
} SFE_IPV4_UNALIGNED_STRUCT;

/*
 * A UDP header, but with an optional "packed" attribute to
 * help with performance on some platforms (see the definition of
 * SFE_IPV4_UNALIGNED_STRUCT)
 */
struct sfe_ipv4_udp_hdr {
	__be16 source;
	__be16 dest;
	__be16 len;
	__sum16 check;
} SFE_IPV4_UNALIGNED_STRUCT;

/*
 * A TCP header, but with an optional "packed" attribute to
 * help with performance on some platforms (see the definition of
 * SFE_IPV4_UNALIGNED_STRUCT)
 */
struct sfe_ipv4_tcp_hdr {
	__be16 source;
	__be16 dest;
	__be32 seq;
	__be32 ack_seq;
#if defined(__LITTLE_ENDIAN_BITFIELD)
	__u16 res1:4,
	      doff:4,
	      fin:1,
	      syn:1,
	      rst:1,
	      psh:1,
	      ack:1,
	      urg:1,
	      ece:1,
	      cwr:1;
#elif defined(__BIG_ENDIAN_BITFIELD)
	__u16 doff:4,
	      res1:4,
	      cwr:1,
	      ece:1,
	      urg:1,
	      ack:1,
	      psh:1,
	      rst:1,
	      syn:1,
	      fin:1;
#else
#error	"Adjust your <asm/byteorder.h> defines"
#endif
	__be16 window;
	__sum16	check;
	__be16 urg_ptr;
} SFE_IPV4_UNALIGNED_STRUCT;

#ifdef TPLINK_PPP_TUNNEL
struct sfe_ipv4_pptp_gre_hdr {
	__u8  flags;
	__u8  ver;
	__be16 protocol;
	__be16 payload_len;
	__be16 call_id;
	__be32 seq;
	__be32 ack;
} __packed;

struct sfe_ipv4_ppp_hdr {
	__be16 addr_ctrl;
	__be16 protocol;
};

struct sfe_ipv4_pppoe_ppp_hdr {
	__be16 protocol;
};

struct sfe_ipv4_l2tpv2_hdr {
	__be16 flags_ver;
	__be16 payload_len;
	__be16 tunnel_id;
	__be16 session_id;
	__be16 ns;
	__be16 nr;
	__be16 off_size;
	__be16 off_pad;
};

struct sfe_ipv4_pppoe_hdr {
#if defined(__LITTLE_ENDIAN_BITFIELD)
	__u8 type : 4;
	__u8 ver : 4;
#elif defined(__BIG_ENDIAN_BITFIELD)
	__u8 ver : 4;
	__u8 type : 4;
#else
#error	"Please fix <asm/byteorder.h>"
#endif
	__u8 code;
	__be16 sid;
	__be16 length;
	struct pppoe_tag tag[0];
} __attribute__((packed));


#define PPTP_GRE_PROTO  0x880B
#define PPTP_GRE_VER    0x1

#define PPTP_GRE_FLAG_C	0x80
#define PPTP_GRE_FLAG_R	0x40
#define PPTP_GRE_FLAG_K	0x20
#define PPTP_GRE_FLAG_S	0x10
#define PPTP_GRE_FLAG_A	0x80

#define PPTP_GRE_IS_C(f) ((f)&PPTP_GRE_FLAG_C)
#define PPTP_GRE_IS_R(f) ((f)&PPTP_GRE_FLAG_R)
#define PPTP_GRE_IS_K(f) ((f)&PPTP_GRE_FLAG_K)
#define PPTP_GRE_IS_S(f) ((f)&PPTP_GRE_FLAG_S)
#define PPTP_GRE_IS_A(f) ((f)&PPTP_GRE_FLAG_A)

/* L2TP header constants */
#define L2TP_HDRFLAG_T	   0x8000
#define L2TP_HDRFLAG_L	   0x4000
#define L2TP_HDRFLAG_S	   0x0800
#define L2TP_HDRFLAG_O	   0x0200
#define L2TP_HDRFLAG_P	   0x0100

#define L2TP_HDR_VER_MASK  0x000F
#define L2TP_HDR_VER_2	   0x0002
#define L2TP_HDR_VER_3	   0x0003

/* PPPOE */
#define PPPOE_VERSION		0x01
#define PPPOE_TYPE			0x01
#define PPPOE_SESS_CODE	0x00
#endif

/*
 * Specifies the lower bound on ACK numbers carried in the TCP header
 */
#define SFE_IPV4_TCP_MAX_ACK_WINDOW 65520

/*
 * IPv4 TCP connection match additional data.
 */
struct sfe_ipv4_tcp_connection_match {
	uint8_t win_scale;		/* Window scale */
	uint32_t max_win;		/* Maximum window size seen */
	uint32_t end;			/* Sequence number of the next byte to send (seq + segment length) */
	uint32_t max_end;		/* Sequence number of the last byte to ack */
};

/*
 * Bit flags for IPv4 connection matching entry.
 */
#define SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_SRC (1<<0)
					/* Perform source translation */
#define SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_DEST (1<<1)
					/* Perform destination translation */
#define SFE_IPV4_CONNECTION_MATCH_FLAG_NO_SEQ_CHECK (1<<2)
					/* Ignore TCP sequence numbers */
#define SFE_IPV4_CONNECTION_MATCH_FLAG_WRITE_FAST_ETH_HDR (1<<3)
					/* Fast Ethernet header write */
#define SFE_IPV4_CONNECTION_MATCH_FLAG_WRITE_L2_HDR (1<<4)
					/* Fast Ethernet header write */
#define SFE_IPV4_CONNECTION_MATCH_FLAG_PRIORITY_REMARK (1<<5)
					/* remark priority of SKB */
#define SFE_IPV4_CONNECTION_MATCH_FLAG_DSCP_REMARK (1<<6)
					/* remark DSCP of packet */

/*
 * IPv4 connection matching structure.
 */
struct sfe_ipv4_connection_match {
	/*
	 * References to other objects.
	 */
	struct sfe_ipv4_connection_match *next;
					/* Next connection match entry in a list */
	struct sfe_ipv4_connection_match *prev;
					/* Previous connection match entry in a list */
	struct sfe_ipv4_connection *connection;
					/* Pointer to our connection */
	struct sfe_ipv4_connection_match *counter_match;
#ifdef TPLINK_LAN_WAN_NETSTACK
	struct sfe_ipv4_connection_match *partner_match;
#endif
					/* Pointer to the connection match in the "counter" direction to this one */
	struct sfe_ipv4_connection_match *active_next;
					/* Pointer to the next connection in the active list */
	struct sfe_ipv4_connection_match *active_prev;
					/* Pointer to the previous connection in the active list */
	bool active;			/* Flag to indicate if we're on the active list */

	/*
	 * Characteristics that identify flows that match this rule.
	 */
	struct net_device *match_dev;	/* Network device */
	uint8_t match_protocol;		/* Protocol */
	__be32 match_src_ip;		/* Source IP address */
	__be32 match_dest_ip;		/* Destination IP address */
	__be16 match_src_port;		/* Source port/connection ident */
	__be16 match_dest_port;		/* Destination port/connection ident */

	/*
	 * Control the operations of the match.
	 */
	uint32_t flags;			/* Bit flags */
#ifdef CONFIG_NF_FLOW_COOKIE
	uint32_t flow_cookie;		/* used flow cookie, for debug */
#endif
#ifdef CONFIG_XFRM
    uint16_t set_accel;              /* set flow_accel or not */ 
	uint16_t flow_accel;             /* The flow accelerated or not */
    struct xfrm_state *x;
#endif

	/*
	 * Connection state that we track once we match.
	 */
	union {				/* Protocol-specific state */
		struct sfe_ipv4_tcp_connection_match tcp;
	} protocol_state;
	uint32_t rx_packet_count;	/* Number of packets RX'd */
	uint32_t rx_byte_count;		/* Number of bytes RX'd */

	/*
	 * Packet translation information.
	 */
	__be32 xlate_src_ip;		/* Address after source translation */
	__be16 xlate_src_port;	/* Port/connection ident after source translation */
	uint16_t xlate_src_csum_adjustment;
					/* Transport layer checksum adjustment after source translation */
	uint16_t xlate_src_partial_csum_adjustment;
					/* Transport layer pseudo header checksum adjustment after source translation */

	__be32 xlate_dest_ip;		/* Address after destination translation */
	__be16 xlate_dest_port;	/* Port/connection ident after destination translation */
	uint16_t xlate_dest_csum_adjustment;
					/* Transport layer checksum adjustment after destination translation */
	uint16_t xlate_dest_partial_csum_adjustment;
					/* Transport layer pseudo header checksum adjustment after destination translation */

	/*
	 * QoS information
	 */
	uint32_t priority;
	uint32_t dscp;

	/*
	 * Packet transmit information.
	 */
	struct net_device *xmit_dev;	/* Network device on which to transmit */
	unsigned short int xmit_dev_mtu;
					/* Interface MTU */
	uint16_t xmit_dest_mac[ETH_ALEN / 2];
					/* Destination MAC address to use when forwarding */
	uint16_t xmit_src_mac[ETH_ALEN / 2];
					/* Source MAC address to use when forwarding */

	/*
	 * Summary stats.
	 */
	uint64_t rx_packet_count64;	/* Number of packets RX'd */
	uint64_t rx_byte_count64;	/* Number of bytes RX'd */
};

/*
 * Per-connection data structure.
 */
struct sfe_ipv4_connection {
	struct sfe_ipv4_connection *next;
					/* Pointer to the next entry in a hash chain */
	struct sfe_ipv4_connection *prev;
					/* Pointer to the previous entry in a hash chain */
	int protocol;			/* IP protocol number */
	__be32 src_ip;			/* Source IP address */
	__be32 src_ip_xlate;		/* NAT-translated source IP address */
	__be32 dest_ip;			/* Destination IP address */
	__be32 dest_ip_xlate;		/* NAT-translated destination IP address */
	__be16 src_port;		/* Source port */
	__be16 src_port_xlate;		/* NAT-translated source port */
	__be16 dest_port;		/* Destination port */
	__be16 dest_port_xlate;		/* NAT-translated destination port */
	struct sfe_ipv4_connection_match *original_match;
					/* Original direction matching structure */
	struct net_device *original_dev;
					/* Original direction source device */
	struct sfe_ipv4_connection_match *reply_match;
					/* Reply direction matching structure */
	struct net_device *reply_dev;	/* Reply direction source device */
	uint64_t last_sync_jiffies;	/* Jiffies count for the last sync */
	struct sfe_ipv4_connection *all_connections_next;
					/* Pointer to the next entry in the list of all connections */
	struct sfe_ipv4_connection *all_connections_prev;
					/* Pointer to the previous entry in the list of all connections */
	uint32_t mark;			/* mark for outgoing packet */
#if defined(CONFIG_NF_CONNTRACK_EXTMARK)
	uint32_t extmark;		/* extmark for outgoing packet */
#endif

	uint32_t debug_read_seq;	/* sequence number for debug dump */
	struct nf_conntrack	*nfct;

#ifdef TPLINK_FRAGMENT_OUTPUT
	struct dst_entry *dst;
	struct dst_entry *reply_dst;
#endif

	int sfe_proto;
	unsigned long slowpath_jiffies; /* sfe heartbeat timer for HWNAT. */
};

/*
 * IPv4 connections and hash table size information.
 */
#define SFE_IPV4_CONNECTION_HASH_SHIFT 16
#define SFE_IPV4_CONNECTION_HASH_SIZE (1 << SFE_IPV4_CONNECTION_HASH_SHIFT)
#define SFE_IPV4_CONNECTION_HASH_MASK (SFE_IPV4_CONNECTION_HASH_SIZE - 1)

#ifdef CONFIG_NF_FLOW_COOKIE
#define SFE_FLOW_COOKIE_SIZE 2048
#define SFE_FLOW_COOKIE_MASK 0x7ff

struct sfe_flow_cookie_entry {
	struct sfe_ipv4_connection_match *match;
	unsigned long last_clean_time;
};
#endif
#if defined(CONFIG_NET_MEDIATEK_HNAT) || defined(CONFIG_NET_MEDIATEK_HNAT_MODULE)
// copy from "drivers/net/ethernet/mediatek/mtk_hnat/nf_hnat_mtk.h"
#define UN_HIT 0x0D /* FOE Un-hit */
#define HIT_UNBIND 0x0E /* FOE Hit unbind */
#define HIT_UNBIND_RATE_REACH 0x0F
#define SLOWPATH_TIMER 	(1*HZ)
#define HWNAT_SLOW_PROTO (SFE_PROTO_L2TP | SFE_PROTO_PPTP | SFE_PROTO_IPSEC)

struct hnat_desc {
	u32 entry : 14;
	u32 crsn : 5;
	u32 sport : 3;
	u32 rev : 1;
	u32 alg : 1;
	u32 iface : 4;
	u32 resv : 4;
	u32 magic_tag_protect : 16;
	u32 wdmaid : 8;
	u32 rxid : 2;
	u32 wcid : 8;
	u32 bssid : 6;
} __packed;

#define skb_hnat_reason(skb) (((struct hnat_desc *)(skb->head))->crsn)
#define skb_hnat_alg(skb) (((struct hnat_desc *)(skb->head))->alg)

int hwnat_taken_skb(struct sk_buff *skb, struct sfe_ipv4_ip_hdr *iph,
	struct sfe_ipv4_connection_match *cm, struct sfe_ipv4_connection_match *partner_cm)
{
	unsigned int frag_off = ntohs(iph->frag_off);

	if ((frag_off & IP_OFFSET) || (frag_off & IP_MF)) {
		/* fragment skb */
		return 0;
	}

	if ((cm && (cm->connection->sfe_proto & HWNAT_SLOW_PROTO)) ||
		(partner_cm && (partner_cm->connection->sfe_proto & HWNAT_SLOW_PROTO))) {
		/* slow path proto types */
		skb_hnat_alg(skb) = 1;
		return 0;
	}

	switch (skb_hnat_reason(skb)) {
	case HIT_UNBIND_RATE_REACH:
		/*
		  HWNAT report these reasons because of:
		    <1> can ACCELERATE, but no entry in table. SKB should be pass up to IPSTACK for create HWNAT entry.
		    <2> can't ACCELERATE (L2TP/PPTP/IPSEC). SKB should be forward by SFE directly.
		 */
		if (jiffies >= (cm->connection->slowpath_jiffies + SLOWPATH_TIMER)) {
			DEBUG_TRACE("slowpath reason:%x\n", skb_hnat_reason(skb));
			cm->connection->slowpath_jiffies = jiffies;
			return 1;
		}
	default:
		return 0;
	}
}
#else
int hwnat_taken_skb(struct sk_buff *skb, struct sfe_ipv4_ip_hdr *iph,
	struct sfe_ipv4_connection_match *cm, struct sfe_ipv4_connection_match *partner_cm)
{
	return 0;
}
#endif
enum sfe_ipv4_exception_events {
	SFE_IPV4_EXCEPTION_EVENT_UDP_HEADER_INCOMPLETE,
	SFE_IPV4_EXCEPTION_EVENT_UDP_NO_CONNECTION,
	SFE_IPV4_EXCEPTION_EVENT_UDP_HWNAT_CONTROL,
	SFE_IPV4_EXCEPTION_EVENT_UDP_IP_OPTIONS_OR_INITIAL_FRAGMENT,
	SFE_IPV4_EXCEPTION_EVENT_UDP_SMALL_TTL,
	SFE_IPV4_EXCEPTION_EVENT_UDP_NEEDS_FRAGMENTATION,
	SFE_IPV4_EXCEPTION_EVENT_TCP_HEADER_INCOMPLETE,
	SFE_IPV4_EXCEPTION_EVENT_TCP_NO_CONNECTION_SLOW_FLAGS,
	SFE_IPV4_EXCEPTION_EVENT_TCP_NO_CONNECTION_FAST_FLAGS,
	SFE_IPV4_EXCEPTION_EVENT_TCP_IP_OPTIONS_OR_INITIAL_FRAGMENT,
	SFE_IPV4_EXCEPTION_EVENT_TCP_SMALL_TTL,
	SFE_IPV4_EXCEPTION_EVENT_TCP_NEEDS_FRAGMENTATION,
	SFE_IPV4_EXCEPTION_EVENT_TCP_FLAGS,
	SFE_IPV4_EXCEPTION_EVENT_TCP_SEQ_EXCEEDS_RIGHT_EDGE,
	SFE_IPV4_EXCEPTION_EVENT_TCP_SMALL_DATA_OFFS,
	SFE_IPV4_EXCEPTION_EVENT_TCP_BAD_SACK,
	SFE_IPV4_EXCEPTION_EVENT_TCP_BIG_DATA_OFFS,
	SFE_IPV4_EXCEPTION_EVENT_TCP_SEQ_BEFORE_LEFT_EDGE,
	SFE_IPV4_EXCEPTION_EVENT_TCP_ACK_EXCEEDS_RIGHT_EDGE,
	SFE_IPV4_EXCEPTION_EVENT_TCP_ACK_BEFORE_LEFT_EDGE,
	SFE_IPV4_EXCEPTION_EVENT_ICMP_HEADER_INCOMPLETE,
	SFE_IPV4_EXCEPTION_EVENT_ICMP_UNHANDLED_TYPE,
	SFE_IPV4_EXCEPTION_EVENT_ICMP_IPV4_HEADER_INCOMPLETE,
	SFE_IPV4_EXCEPTION_EVENT_ICMP_IPV4_NON_V4,
	SFE_IPV4_EXCEPTION_EVENT_ICMP_IPV4_IP_OPTIONS_INCOMPLETE,
	SFE_IPV4_EXCEPTION_EVENT_ICMP_IPV4_UDP_HEADER_INCOMPLETE,
	SFE_IPV4_EXCEPTION_EVENT_ICMP_IPV4_TCP_HEADER_INCOMPLETE,
	SFE_IPV4_EXCEPTION_EVENT_ICMP_IPV4_UNHANDLED_PROTOCOL,
	SFE_IPV4_EXCEPTION_EVENT_ICMP_NO_CONNECTION,
	SFE_IPV4_EXCEPTION_EVENT_ICMP_FLUSHED_CONNECTION,
	SFE_IPV4_EXCEPTION_EVENT_HEADER_INCOMPLETE,
	SFE_IPV4_EXCEPTION_EVENT_BAD_TOTAL_LENGTH,
	SFE_IPV4_EXCEPTION_EVENT_NON_V4,
	SFE_IPV4_EXCEPTION_EVENT_NON_INITIAL_FRAGMENT,
	SFE_IPV4_EXCEPTION_EVENT_DATAGRAM_INCOMPLETE,
	SFE_IPV4_EXCEPTION_EVENT_IP_OPTIONS_INCOMPLETE,
	SFE_IPV4_EXCEPTION_EVENT_UNHANDLED_PROTOCOL,
	SFE_IPV4_EXCEPTION_EVENT_BANDWIDTH_CONTROL,
	SFE_IPV4_EXCEPTION_EVENT_PPTP_GRE_UNHANDLED,
	SFE_IPV4_EXCEPTION_EVENT_PPTP_NO_TUNNEL,
	SFE_IPV4_EXCEPTION_EVENT_PPTP_HEADER_INVALID,
	SFE_IPV4_EXCEPTION_EVENT_L2TP_TUNNEL_UNHANDLED,
	SFE_IPV4_EXCEPTION_EVENT_L2TP_NO_TUNNEL,
	SFE_IPV4_EXCEPTION_EVENT_L2TP_HEADER_INVALID,
	SFE_IPV4_EXCEPTION_EVENT_PPPOE_TUNNEL_UNHANDLED,
	SFE_IPV4_EXCEPTION_EVENT_PPPOE_NO_TUNNEL,
	SFE_IPV4_EXCEPTION_EVENT_PPPOE_HEADER_INVALID,
	SFE_IPV4_EXCEPTION_EVENT_LAST
};

static char *sfe_ipv4_exception_events_string[SFE_IPV4_EXCEPTION_EVENT_LAST] = {
	"UDP_HEADER_INCOMPLETE",
	"UDP_NO_CONNECTION",
	"UDP_HWNAT_CONTROL",
	"UDP_IP_OPTIONS_OR_INITIAL_FRAGMENT",
	"UDP_SMALL_TTL",
	"UDP_NEEDS_FRAGMENTATION",
	"TCP_HEADER_INCOMPLETE",
	"TCP_NO_CONNECTION_SLOW_FLAGS",
	"TCP_NO_CONNECTION_FAST_FLAGS",
	"TCP_IP_OPTIONS_OR_INITIAL_FRAGMENT",
	"TCP_SMALL_TTL",
	"TCP_NEEDS_FRAGMENTATION",
	"TCP_FLAGS",
	"TCP_SEQ_EXCEEDS_RIGHT_EDGE",
	"TCP_SMALL_DATA_OFFS",
	"TCP_BAD_SACK",
	"TCP_BIG_DATA_OFFS",
	"TCP_SEQ_BEFORE_LEFT_EDGE",
	"TCP_ACK_EXCEEDS_RIGHT_EDGE",
	"TCP_ACK_BEFORE_LEFT_EDGE",
	"ICMP_HEADER_INCOMPLETE",
	"ICMP_UNHANDLED_TYPE",
	"ICMP_IPV4_HEADER_INCOMPLETE",
	"ICMP_IPV4_NON_V4",
	"ICMP_IPV4_IP_OPTIONS_INCOMPLETE",
	"ICMP_IPV4_UDP_HEADER_INCOMPLETE",
	"ICMP_IPV4_TCP_HEADER_INCOMPLETE",
	"ICMP_IPV4_UNHANDLED_PROTOCOL",
	"ICMP_NO_CONNECTION",
	"ICMP_FLUSHED_CONNECTION",
	"HEADER_INCOMPLETE",
	"BAD_TOTAL_LENGTH",
	"NON_V4",
	"NON_INITIAL_FRAGMENT",
	"DATAGRAM_INCOMPLETE",
	"IP_OPTIONS_INCOMPLETE",
	"UNHANDLED_PROTOCOL",
	"BANDWIDTH_CONTROL",
	"PPTP_GRE_UNHANDLED",
	"PPTP_NO_TUNNEL",
	"PPTP_HEADER_INVALID",
	"L2TP_TUNNEL_UNHANDLED",
	"L2TP_NO_TUNNEL",
	"L2TP_HEADER_INVALID",
	"PPPOE_TUNNEL_UNHANDLED",
	"PPPOE_NO_TUNNEL",
	"PPPOE_HEADER_INVALID"
};

/*
 * Per-module structure.
 */
struct sfe_ipv4 {
	spinlock_t lock;		/* Lock for SMP correctness */
	struct sfe_ipv4_connection_match *active_head;
					/* Head of the list of recently active connections */
	struct sfe_ipv4_connection_match *active_tail;
					/* Tail of the list of recently active connections */
	struct sfe_ipv4_connection *all_connections_head;
					/* Head of the list of all connections */
	struct sfe_ipv4_connection *all_connections_tail;
					/* Tail of the list of all connections */
	unsigned int num_connections;	/* Number of connections */
	struct timer_list timer;	/* Timer used for periodic sync ops */
	sfe_sync_rule_callback_t __rcu sync_rule_callback;
					/* Callback function registered by a connection manager for stats syncing */
	struct sfe_ipv4_connection *conn_hash[SFE_IPV4_CONNECTION_HASH_SIZE];
					/* Connection hash table */
	struct sfe_ipv4_connection_match *conn_match_hash[SFE_IPV4_CONNECTION_HASH_SIZE];
					/* Connection match hash table */
#ifdef CONFIG_NF_FLOW_COOKIE
	struct sfe_flow_cookie_entry sfe_flow_cookie_table[SFE_FLOW_COOKIE_SIZE];
					/* flow cookie table*/
	flow_cookie_set_func_t flow_cookie_set_func;
					/* function used to configure flow cookie in hardware*/
	int flow_cookie_enable;
					/* Enable/disable flow cookie at runtime */
#endif

	/*
	 * Statistics.
	 */
	uint32_t connection_create_requests;
					/* Number of IPv4 connection create requests */
	uint32_t connection_create_collisions;
					/* Number of IPv4 connection create requests that collided with existing hash table entries */
	uint32_t connection_destroy_requests;
					/* Number of IPv4 connection destroy requests */
	uint32_t connection_destroy_misses;
					/* Number of IPv4 connection destroy requests that missed our hash table */
	uint32_t connection_match_hash_hits;
					/* Number of IPv4 connection match hash hits */
	uint32_t connection_match_hash_reorders;
					/* Number of IPv4 connection match hash reorders */
	uint32_t connection_flushes;	/* Number of IPv4 connection flushes */
	uint32_t packets_forwarded;	/* Number of IPv4 packets forwarded */
	uint32_t packets_not_forwarded;	/* Number of IPv4 packets not forwarded */
	uint32_t exception_events[SFE_IPV4_EXCEPTION_EVENT_LAST];

	/*
	 * Summary tatistics.
	 */
	uint64_t connection_create_requests64;
					/* Number of IPv4 connection create requests */
	uint64_t connection_create_collisions64;
					/* Number of IPv4 connection create requests that collided with existing hash table entries */
	uint64_t connection_destroy_requests64;
					/* Number of IPv4 connection destroy requests */
	uint64_t connection_destroy_misses64;
					/* Number of IPv4 connection destroy requests that missed our hash table */
	uint64_t connection_match_hash_hits64;
					/* Number of IPv4 connection match hash hits */
	uint64_t connection_match_hash_reorders64;
					/* Number of IPv4 connection match hash reorders */
	uint64_t connection_flushes64;	/* Number of IPv4 connection flushes */
	uint64_t packets_forwarded64;	/* Number of IPv4 packets forwarded */
	uint64_t packets_not_forwarded64;
					/* Number of IPv4 packets not forwarded */
	uint64_t exception_events64[SFE_IPV4_EXCEPTION_EVENT_LAST];

	/*
	 * Control state.
	 */
	struct kobject *sys_sfe_ipv4;	/* sysfs linkage */
	int debug_dev;			/* Major number of the debug char device */
	uint32_t debug_read_seq;	/* sequence number for debug dump */
};

/*
 * Enumeration of the XML output.
 */
enum sfe_ipv4_debug_xml_states {
	SFE_IPV4_DEBUG_XML_STATE_START,
	SFE_IPV4_DEBUG_XML_STATE_CONNECTIONS_START,
	SFE_IPV4_DEBUG_XML_STATE_CONNECTIONS_CONNECTION,
	SFE_IPV4_DEBUG_XML_STATE_CONNECTIONS_END,
	SFE_IPV4_DEBUG_XML_STATE_EXCEPTIONS_START,
	SFE_IPV4_DEBUG_XML_STATE_EXCEPTIONS_EXCEPTION,
	SFE_IPV4_DEBUG_XML_STATE_EXCEPTIONS_END,
	SFE_IPV4_DEBUG_XML_STATE_STATS,
#ifdef TPLINK_PPP_TUNNEL_DUMP
	SFE_IPV4_DEBUG_XML_STATE_TUNNEL_INFO_START,
	SFE_IPV4_DEBUG_XML_STATE_TUNNEL_INFO_INFOMATION,
	SFE_IPV4_DEBUG_XML_STATE_TUNNEL_INFO_END,
#endif
#ifdef TPLINK_XFRM_DST_DUMP
	SFE_IPV4_DEBUG_XML_STATE_XFRM_DST_START,
	SFE_IPV4_DEBUG_XML_STATE_XFRM_DST_INFOMATION,
	SFE_IPV4_DEBUG_XML_STATE_XFRM_DST_END,
#endif
	SFE_IPV4_DEBUG_XML_STATE_END,
	SFE_IPV4_DEBUG_XML_STATE_DONE
};

static int sfe_ipv4_dev_queue_xmit(struct sk_buff *skb);
static int sfe_ipv4_fragment(struct sk_buff *skb, int l2_header);

/*
 * XML write state.
 */
struct sfe_ipv4_debug_xml_write_state {
	enum sfe_ipv4_debug_xml_states state;
					/* XML output file state machine state */
	int iter_exception;		/* Next exception iterator */
};

#ifdef TPLINK_PPP_TUNNEL
static int sfe_ipv4_recv_core(struct net_device *dev, struct sk_buff *skb, int ppp_header_level, int sfe_proto);
#else
static int sfe_ipv4_recv_core(struct net_device *dev, struct sk_buff *skb, int sfe_proto);
#endif

typedef bool (*sfe_ipv4_debug_xml_write_method_t)(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
						  int *total_read, struct sfe_ipv4_debug_xml_write_state *ws);

struct sfe_ipv4 __si;

extern int sfe_ipstat_switch;

#ifdef CONFIG_NF_FASTPATH_NAT_HOOK
extern int (*fastpath_nat_forward_hook)(struct sk_buff *skb, struct nf_conn *ct, enum ip_conntrack_dir dir);
extern int (*qos_fastpath_nat_forward_hook)(struct sk_buff *skb, struct nf_conn *ct, enum ip_conntrack_dir dir);
extern int (*fastpath_nat_forward_on)(void);
extern int (*qos_fastpath_nat_forward_on)(void);
static int fastpath_nat_enabled;
static int qos_fastpath_nat_enabled;
#endif

#ifdef TPLINK_LAN_WAN_NETSTACK
char lan_inner_name[64] = "veth1";     /* default */
char wan_inner_name[64] = "br-lan";    /* default */
module_param_string(lan_inner_name, lan_inner_name, 63, 0);
module_param_string(wan_inner_name, wan_inner_name, 63, 0);

struct net_device *lan_inner_dev;
struct net_device *wan_inner_dev;

void sfe_ipv4_destroy_inner_dev(struct net_device *dev)
{
	struct sfe_ipv4 *si = &__si;

	if (dev) {
		spin_lock_bh(&si->lock);
		if (!strcmp(dev->name, lan_inner_name)) {
			RCU_INIT_POINTER(lan_inner_dev, NULL);
		}
		if (!strcmp(dev->name, wan_inner_name)) {
			RCU_INIT_POINTER(wan_inner_dev, NULL);
		}
		spin_unlock_bh(&si->lock);
	}
}

void sfe_ipv4_create_inner_dev(struct net_device *dev)
{
	struct sfe_ipv4 *si = &__si;

	if (dev) {
		spin_lock_bh(&si->lock);
		if (!strcmp(dev->name, lan_inner_name)) {
			RCU_INIT_POINTER(lan_inner_dev, dev);
		}
		if (!strcmp(dev->name, wan_inner_name)) {
			RCU_INIT_POINTER(wan_inner_dev, dev);
		}
		spin_unlock_bh(&si->lock);
	}
}

#endif

/*
 * HWNAT flags of mtk.
 */
static void hw_nat_slow_path(struct sk_buff *skb)
{
#if  defined(CONFIG_RA_HW_NAT) || defined(CONFIG_RA_HW_NAT_MODULE)
	if( IS_SPACE_AVAILABLED(skb) &&
		((FOE_MAGIC_TAG(skb) == FOE_MAGIC_PCI) ||
		 (FOE_MAGIC_TAG(skb) == FOE_MAGIC_WLAN) ||
		 (FOE_MAGIC_TAG(skb) == FOE_MAGIC_GE))){
		FOE_ALG(skb)=1;
	}
#endif
}

/*
 * sfe_ipv4_gen_ip_csum()
 *	Generate the IP checksum for an IPv4 header.
 *
 * Note that this function assumes that we have only 20 bytes of IP header.
 */
static inline uint16_t sfe_ipv4_gen_ip_csum(struct sfe_ipv4_ip_hdr *iph)
{
	uint32_t sum;
	uint16_t *i = (uint16_t *)iph;

	iph->check = 0;

	/*
	 * Generate the sum.
	 */
	sum = i[0] + i[1] + i[2] + i[3] + i[4] + i[5] + i[6] + i[7] + i[8] + i[9];

	/*
	 * Fold it to ones-complement form.
	 */
	sum = (sum & 0xffff) + (sum >> 16);
	sum = (sum & 0xffff) + (sum >> 16);

	return (uint16_t)sum ^ 0xffff;
}

/*
 * sfe_ipv4_get_connection_match_hash()
 *	Generate the hash used in connection match lookups.
 */
static inline unsigned int sfe_ipv4_get_connection_match_hash(struct net_device *dev, uint8_t protocol,
							      __be32 src_ip, __be16 src_port,
							      __be32 dest_ip, __be16 dest_port)
{
	size_t dev_addr = (size_t)dev_net(dev);

	uint32_t hash = ((uint32_t)dev_addr) ^ ntohl(src_ip ^ dest_ip) ^ protocol ^ ntohs(src_port ^ dest_port);
	return ((hash >> SFE_IPV4_CONNECTION_HASH_SHIFT) ^ hash) & SFE_IPV4_CONNECTION_HASH_MASK;
}

/*
 * sfe_ipv4_find_sfe_ipv4_connection_match()
 *	Get the IPv4 flow match info that corresponds to a particular 5-tuple.
 *
 * On entry we must be holding the lock that protects the hash table.
 */
static struct sfe_ipv4_connection_match *
sfe_ipv4_find_sfe_ipv4_connection_match(struct sfe_ipv4 *si, struct net_device *dev, uint8_t protocol,
					__be32 src_ip, __be16 src_port,
					__be32 dest_ip, __be16 dest_port) __attribute__((always_inline));
static struct sfe_ipv4_connection_match *
sfe_ipv4_find_sfe_ipv4_connection_match(struct sfe_ipv4 *si, struct net_device *dev, uint8_t protocol,
					__be32 src_ip, __be16 src_port,
					__be32 dest_ip, __be16 dest_port)
{
	struct sfe_ipv4_connection_match *cm = NULL;
	struct sfe_ipv4_connection_match *head = NULL;
	unsigned int conn_match_idx = 0;

	conn_match_idx = sfe_ipv4_get_connection_match_hash(dev, protocol, src_ip, src_port, dest_ip, dest_port);
	
	//DEBUG_TRACE("Find connection: conn_match_hash %d , dev:%s, pro:%d, %pI4:%u To: %pI4:%u\n",
	//    conn_match_idx, dev->name, protocol, &src_ip, src_port, &dest_ip, dest_port);
	/*if(strcmp(dev->name,"lo"))
	{
		DEBUG_TRACE("conn_match_hash %d info - dev: %s ,src_ip:%pI4,src_port:%u, dest_ip:%pI4,dest_port:%u\n",\
	   conn_match_idx,dev->name,&src_ip,src_port,&dest_ip,dest_port);
	}*/
	cm = si->conn_match_hash[conn_match_idx];

	/*while(xhead != NULL)
	{
	   DEBUG_TRACE("conn_match_hash %d info - dev: %s ,src_ip:%pI4,src_port:%u, dest_ip:%pI4,dest_port:%u\n",\
	   conn_match_idx,xhead->match_dev->name,&xhead->match_src_ip,xhead->match_src_port,&xhead->match_dest_ip,xhead->match_dest_port);
	   xhead = xhead->next;
	} */
	/*
	 * If we don't have anything in this chain then bale.
	 */
	if (unlikely(!cm)) {
		return cm;
	}

	/*
	 * Hopefully the first entry is the one we want.
	 */
	if (likely(cm->match_src_port == src_port)
	    && likely(cm->match_dest_port == dest_port)
	    && likely(cm->match_src_ip == src_ip)
	    && likely(cm->match_dest_ip == dest_ip)
	    && likely(cm->match_protocol == protocol)
	    && likely(dev_net(cm->match_dev) == dev_net(dev))) {
		si->connection_match_hash_hits++;
		return cm;
	}

	/*
	 * We may or may not have a matching entry but if we do then we want to
	 * move that entry to the top of the hash chain when we get to it.  We
	 * presume that this will be reused again very quickly.
	 */
	head = cm;
	do {
		cm = cm->next;
	} while (cm && (cm->match_src_port != src_port
		 || cm->match_dest_port != dest_port
		 || cm->match_src_ip != src_ip
		 || cm->match_dest_ip != dest_ip
		 || cm->match_protocol != protocol
		 || dev_net(cm->match_dev) != dev_net(dev)));

	/*
	 * Not found then we're done.
	 */
	if (unlikely(!cm)) {
		return cm;
	}

	/*
	 * We found a match so move it.
	 */
	if (cm->next) {
		cm->next->prev = cm->prev;
	}
	cm->prev->next = cm->next;
	cm->prev = NULL;
	cm->next = head;
	head->prev = cm;
	si->conn_match_hash[conn_match_idx] = cm;
	si->connection_match_hash_reorders++;

	return cm;
}

/*
 * sfe_ipv4_connection_match_update_summary_stats()
 *	Update the summary stats for a connection match entry.
 */
static inline void sfe_ipv4_connection_match_update_summary_stats(struct sfe_ipv4_connection_match *cm)
{
	cm->rx_packet_count64 += cm->rx_packet_count;
	cm->rx_packet_count = 0;
	cm->rx_byte_count64 += cm->rx_byte_count;
	cm->rx_byte_count = 0;
}

/*
 * sfe_ipv4_connection_match_compute_translations()
 *	Compute port and address translations for a connection match entry.
 */
static void sfe_ipv4_connection_match_compute_translations(struct sfe_ipv4_connection_match *cm)
{
	/*
	 * Before we insert the entry look to see if this is tagged as doing address
	 * translations.  If it is then work out the adjustment that we need to apply
	 * to the transport checksum.
	 */
	if (cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_SRC) {
		/*
		 * Precompute an incremental checksum adjustment so we can
		 * edit packets in this stream very quickly.  The algorithm is from RFC1624.
		 */
		uint16_t src_ip_hi = cm->match_src_ip >> 16;
		uint16_t src_ip_lo = cm->match_src_ip & 0xffff;
		uint32_t xlate_src_ip = ~cm->xlate_src_ip;
		uint16_t xlate_src_ip_hi = xlate_src_ip >> 16;
		uint16_t xlate_src_ip_lo = xlate_src_ip & 0xffff;
		uint16_t xlate_src_port = ~cm->xlate_src_port;
		uint32_t adj;

		/*
		 * When we compute this fold it down to a 16-bit offset
		 * as that way we can avoid having to do a double
		 * folding of the twos-complement result because the
		 * addition of 2 16-bit values cannot cause a double
		 * wrap-around!
		 */
		adj = src_ip_hi + src_ip_lo + cm->match_src_port
		      + xlate_src_ip_hi + xlate_src_ip_lo + xlate_src_port;
		adj = (adj & 0xffff) + (adj >> 16);
		adj = (adj & 0xffff) + (adj >> 16);
		cm->xlate_src_csum_adjustment = (uint16_t)adj;

	}

	if (cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_DEST) {
		/*
		 * Precompute an incremental checksum adjustment so we can
		 * edit packets in this stream very quickly.  The algorithm is from RFC1624.
		 */
		uint16_t dest_ip_hi = cm->match_dest_ip >> 16;
		uint16_t dest_ip_lo = cm->match_dest_ip & 0xffff;
		uint32_t xlate_dest_ip = ~cm->xlate_dest_ip;
		uint16_t xlate_dest_ip_hi = xlate_dest_ip >> 16;
		uint16_t xlate_dest_ip_lo = xlate_dest_ip & 0xffff;
		uint16_t xlate_dest_port = ~cm->xlate_dest_port;
		uint32_t adj;

		/*
		 * When we compute this fold it down to a 16-bit offset
		 * as that way we can avoid having to do a double
		 * folding of the twos-complement result because the
		 * addition of 2 16-bit values cannot cause a double
		 * wrap-around!
		 */
		adj = dest_ip_hi + dest_ip_lo + cm->match_dest_port
		      + xlate_dest_ip_hi + xlate_dest_ip_lo + xlate_dest_port;
		adj = (adj & 0xffff) + (adj >> 16);
		adj = (adj & 0xffff) + (adj >> 16);
		cm->xlate_dest_csum_adjustment = (uint16_t)adj;
	}

	if (cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_SRC) {
		uint32_t adj = ~cm->match_src_ip + cm->xlate_src_ip;
		if (adj < cm->xlate_src_ip) {
			adj++;
		}

		adj = (adj & 0xffff) + (adj >> 16);
		adj = (adj & 0xffff) + (adj >> 16);
		cm->xlate_src_partial_csum_adjustment = (uint16_t)adj;
	}

	if (cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_DEST) {
		uint32_t adj = ~cm->match_dest_ip + cm->xlate_dest_ip;
		if (adj < cm->xlate_dest_ip) {
			adj++;
		}

		adj = (adj & 0xffff) + (adj >> 16);
		adj = (adj & 0xffff) + (adj >> 16);
		cm->xlate_dest_partial_csum_adjustment = (uint16_t)adj;
	}

}

/*
 * sfe_ipv4_update_summary_stats()
 *	Update the summary stats.
 */
static void sfe_ipv4_update_summary_stats(struct sfe_ipv4 *si)
{
	int i;

	si->connection_create_requests64 += si->connection_create_requests;
	si->connection_create_requests = 0;
	si->connection_create_collisions64 += si->connection_create_collisions;
	si->connection_create_collisions = 0;
	si->connection_destroy_requests64 += si->connection_destroy_requests;
	si->connection_destroy_requests = 0;
	si->connection_destroy_misses64 += si->connection_destroy_misses;
	si->connection_destroy_misses = 0;
	si->connection_match_hash_hits64 += si->connection_match_hash_hits;
	si->connection_match_hash_hits = 0;
	si->connection_match_hash_reorders64 += si->connection_match_hash_reorders;
	si->connection_match_hash_reorders = 0;
	si->connection_flushes64 += si->connection_flushes;
	si->connection_flushes = 0;
	si->packets_forwarded64 += si->packets_forwarded;
	si->packets_forwarded = 0;
	si->packets_not_forwarded64 += si->packets_not_forwarded;
	si->packets_not_forwarded = 0;

	for (i = 0; i < SFE_IPV4_EXCEPTION_EVENT_LAST; i++) {
		si->exception_events64[i] += si->exception_events[i];
		si->exception_events[i] = 0;
	}
}

/*
 * sfe_ipv4_insert_sfe_ipv4_connection_match()
 *	Insert a connection match into the hash.
 *
 * On entry we must be holding the lock that protects the hash table.
 */
static inline void sfe_ipv4_insert_sfe_ipv4_connection_match(struct sfe_ipv4 *si, struct sfe_ipv4_connection_match *cm)
{
	struct sfe_ipv4_connection_match **hash_head = NULL;
	struct sfe_ipv4_connection_match *prev_head = NULL;
	struct sfe_ipv4_connection_match *head = NULL;
	unsigned int conn_match_idx
		= sfe_ipv4_get_connection_match_hash(cm->match_dev, cm->match_protocol,
						     cm->match_src_ip, cm->match_src_port,
						     cm->match_dest_ip, cm->match_dest_port);
	hash_head = &si->conn_match_hash[conn_match_idx];

	prev_head = *hash_head;
	cm->prev = NULL;
	if (prev_head) {
		prev_head->prev = cm;
	}

	cm->next = prev_head;
	*hash_head = cm;

	head = *hash_head;
	
	while(head != NULL)
	{
		DEBUG_TRACE("conn_match_hash %d info - dev: %s ,src_ip:%pI4,src_port:%u, dest_ip:%pI4,dest_port:%u\n",\
	   conn_match_idx,head->match_dev->name,&head->match_src_ip,head->match_src_port,&head->match_dest_ip,head->match_dest_port);
	   head = head->next;
	} 

#ifdef CONFIG_NF_FLOW_COOKIE
	if (!si->flow_cookie_enable)
		return;

	/*
	 * Configure hardware to put a flow cookie in packet of this flow,
	 * then we can accelerate the lookup process when we received this packet.
	 */
	for (conn_match_idx = 1; conn_match_idx < SFE_FLOW_COOKIE_SIZE; conn_match_idx++) {
		struct sfe_flow_cookie_entry *entry = &si->sfe_flow_cookie_table[conn_match_idx];

		if ((NULL == entry->match) && time_is_before_jiffies(entry->last_clean_time + HZ)) {
			flow_cookie_set_func_t func;

			rcu_read_lock();
			func = rcu_dereference(si->flow_cookie_set_func);
			if (func) {
				if (!func(cm->match_protocol, cm->match_src_ip, cm->match_src_port,
					 cm->match_dest_ip, cm->match_dest_port, conn_match_idx)) {
					entry->match = cm;
					cm->flow_cookie = conn_match_idx;
				}
			}
			rcu_read_unlock();

			break;
		}
	}
#endif
}

/*
 * sfe_ipv4_remove_sfe_ipv4_connection_match()
 *	Remove a connection match object from the hash.
 *
 * On entry we must be holding the lock that protects the hash table.
 */
static inline void sfe_ipv4_remove_sfe_ipv4_connection_match(struct sfe_ipv4 *si, struct sfe_ipv4_connection_match *cm)
{
#ifdef CONFIG_NF_FLOW_COOKIE
	if (si->flow_cookie_enable) {
		/*
		 * Tell hardware that we no longer need a flow cookie in packet of this flow
		 */
		unsigned int conn_match_idx;

		for (conn_match_idx = 1; conn_match_idx < SFE_FLOW_COOKIE_SIZE; conn_match_idx++) {
			struct sfe_flow_cookie_entry *entry = &si->sfe_flow_cookie_table[conn_match_idx];

			if (cm == entry->match) {
				flow_cookie_set_func_t func;

				rcu_read_lock();
				func = rcu_dereference(si->flow_cookie_set_func);
				if (func) {
					func(cm->match_protocol, cm->match_src_ip, cm->match_src_port,
					     cm->match_dest_ip, cm->match_dest_port, 0);
				}
				rcu_read_unlock();

				cm->flow_cookie = 0;
				entry->match = NULL;
				entry->last_clean_time = jiffies;
				break;
			}
		}
	}
#endif

	/*
	 * Unlink the connection match entry from the hash.
	 */
	if (cm->prev) {
		cm->prev->next = cm->next;
	} else {
		unsigned int conn_match_idx
			= sfe_ipv4_get_connection_match_hash(cm->match_dev, cm->match_protocol,
							     cm->match_src_ip, cm->match_src_port,
							     cm->match_dest_ip, cm->match_dest_port);
		si->conn_match_hash[conn_match_idx] = cm->next;
	}

	if (cm->next) {
		cm->next->prev = cm->prev;
	}

	//DEBUG_TRACE("Remove connection: dev:%s, pro:%d, %pI4:%u To: %pI4:%u\n",
	//    cm->match_dev->name, cm->match_protocol, &cm->match_src_ip, cm->match_src_port, &cm->match_dest_ip, cm->match_dest_port);

	/*
	 * If the connection match entry is in the active list remove it.
	 */
	if (cm->active) {
		if (likely(cm->active_prev)) {
			cm->active_prev->active_next = cm->active_next;
		} else {
			si->active_head = cm->active_next;
		}

		if (likely(cm->active_next)) {
			cm->active_next->active_prev = cm->active_prev;
		} else {
			si->active_tail = cm->active_prev;
		}
	}
}

/*
 * sfe_ipv4_get_connection_hash()
 *	Generate the hash used in connection lookups.
 */
static inline unsigned int sfe_ipv4_get_connection_hash(struct net *this_net, uint8_t protocol, __be32 src_ip, __be16 src_port,
							__be32 dest_ip, __be16 dest_port)
{
	uint32_t hash = ntohl(src_ip ^ dest_ip) ^ protocol ^ ntohs(src_port ^ dest_port) ^(uint32_t)this_net;
	return ((hash >> SFE_IPV4_CONNECTION_HASH_SHIFT) ^ hash) & SFE_IPV4_CONNECTION_HASH_MASK;
}

/*
 * sfe_ipv4_find_sfe_ipv4_connection()
 *	Get the IPv4 connection info that corresponds to a particular 5-tuple.
 *
 * On entry we must be holding the lock that protects the hash table.
 */
static inline struct sfe_ipv4_connection *sfe_ipv4_find_sfe_ipv4_connection(struct sfe_ipv4 *si, struct net *this_net, uint32_t protocol,
									    __be32 src_ip, __be16 src_port,
									    __be32 dest_ip, __be16 dest_port)
{
	struct sfe_ipv4_connection *c;
	unsigned int conn_idx = sfe_ipv4_get_connection_hash(this_net, protocol, src_ip, src_port, dest_ip, dest_port);
	c = si->conn_hash[conn_idx];

	/*
	 * If we don't have anything in this chain then bale.
	 */
	if (unlikely(!c)) {
		return c;
	}

	/*
	 * Hopefully the first entry is the one we want.
	 */
	if (likely(dev_net(c->original_dev) == this_net)
	    && likely(c->src_port == src_port)
	    && likely(c->dest_port == dest_port)
	    && likely(c->src_ip == src_ip)
	    && likely(c->dest_ip == dest_ip)
	    && likely(c->protocol == protocol)) {
		return c;
	}

	/*
	 * We may or may not have a matching entry but if we do then we want to
	 * move that entry to the top of the hash chain when we get to it.  We
	 * presume that this will be reused again very quickly.
	 */
	do {
		c = c->next;
	} while (c && ((dev_net(c->original_dev) != this_net)
		 || c->src_port != src_port
		 || c->dest_port != dest_port
		 || c->src_ip != src_ip
		 || c->dest_ip != dest_ip
		 || c->protocol != protocol));

	/*
	 * Will need connection entry for next create/destroy metadata,
	 * So no need to re-order entry for these requests
	 */
	return c;
}

/*
 * sfe_ipv4_mark_rule()
 *	Updates the mark for a current offloaded connection
 *
 * Will take hash lock upon entry
 */
void sfe_ipv4_mark_rule(struct sfe_connection_mark *mark)
{
	struct sfe_ipv4 *si = &__si;
	struct sfe_ipv4_connection *c;

	spin_lock_bh(&si->lock);
	c = sfe_ipv4_find_sfe_ipv4_connection(si, mark->net, mark->protocol,
					      mark->src_ip.ip, mark->src_port,
					      mark->dest_ip.ip, mark->dest_port);
	if (c) {
		DEBUG_TRACE("Matching connection found for mark, "
			    "setting from %08x to %08x\n",
			    c->mark, mark->mark);
#if defined(CONFIG_NF_CONNTRACK_EXTMARK)
		DEBUG_TRACE("Matching connection found for extmark, "
			    "setting from %08x to %08x\n",
			    c->extmark, mark->extmark);
		WARN_ON((0 != c->extmark) && (0 == mark->extmark));
#endif
		WARN_ON((0 != c->mark) && (0 == mark->mark));
		c->mark = mark->mark;
#if defined(CONFIG_NF_CONNTRACK_EXTMARK)
		c->extmark = mark->extmark;
#endif
	}
	spin_unlock_bh(&si->lock);
}

int dev_tuntype(const struct net_device *dev)
{
	return (dev && 
	          (dev->flags & IFF_POINTOPOINT) &&
	          (dev->rtnl_link_ops) &&
	          (dev->rtnl_link_ops->kind) &&
	          (!strcmp(dev->rtnl_link_ops->kind, "tun"))
	         );
}

/*
 * sfe_ipv4_insert_sfe_ipv4_connection()
 *	Insert a connection into the hash.
 *
 * On entry we must be holding the lock that protects the hash table.
 */
static void sfe_ipv4_insert_sfe_ipv4_connection(struct sfe_ipv4 *si, struct sfe_ipv4_connection *c)
{
	struct sfe_ipv4_connection **hash_head = NULL;
	struct sfe_ipv4_connection *prev_head = NULL;
	unsigned int conn_idx;
    DEBUG_TRACE("*sfe_ipv4_insert_sfe_ipv4_connection $$$$$$$ ");
	/*
	 * Insert entry into the connection hash.
	 */
	conn_idx = sfe_ipv4_get_connection_hash(dev_net(c->original_dev), c->protocol, c->src_ip, c->src_port,
						c->dest_ip, c->dest_port);
	DEBUG_TRACE("conn_idx is %u",conn_idx);

	hash_head = &si->conn_hash[conn_idx];
	prev_head = *hash_head;
	
	
	c->prev = NULL;
	if (prev_head) {
		prev_head->prev = c;
	}

	c->next = prev_head;
	*hash_head = c;

	/*
	 * Insert entry into the "all connections" list.
	 */
	if (si->all_connections_tail) {
		c->all_connections_prev = si->all_connections_tail;
		si->all_connections_tail->all_connections_next = c;
	} else {
		c->all_connections_prev = NULL;
		si->all_connections_head = c;
	}

	si->all_connections_tail = c;
	c->all_connections_next = NULL;
	si->num_connections++;

	/*
	 * Insert the connection match objects too.
	 */
	sfe_ipv4_insert_sfe_ipv4_connection_match(si, c->original_match);
	sfe_ipv4_insert_sfe_ipv4_connection_match(si, c->reply_match);
}

/*
 * sfe_ipv4_remove_sfe_ipv4_connection()
 *	Remove a sfe_ipv4_connection object from the hash.
 *
 * On entry we must be holding the lock that protects the hash table.
 */
static void sfe_ipv4_remove_sfe_ipv4_connection(struct sfe_ipv4 *si, struct sfe_ipv4_connection *c)
{
	/*
	 * Remove the connection match objects.
	 */
	sfe_ipv4_remove_sfe_ipv4_connection_match(si, c->reply_match);
	sfe_ipv4_remove_sfe_ipv4_connection_match(si, c->original_match);

#ifdef TPLINK_LAN_WAN_NETSTACK
	if (c->original_match->partner_match) {
		c->original_match->partner_match->counter_match->partner_match = NULL;
		c->original_match->partner_match = NULL;
	}
	if (c->reply_match->partner_match) {
		c->reply_match->partner_match->counter_match->partner_match = NULL;
		c->reply_match->partner_match = NULL;
	}
#endif
	/*
	 * Unlink the connection.
	 */
	if (c->prev) {
		c->prev->next = c->next;
	} else {
		unsigned int conn_idx = sfe_ipv4_get_connection_hash(dev_net(c->original_dev), c->protocol, c->src_ip, c->src_port,
								     c->dest_ip, c->dest_port);
		si->conn_hash[conn_idx] = c->next;
	}

	if (c->next) {
		c->next->prev = c->prev;
	}

	/*
	 * Unlink connection from all_connections list
	 */
	if (c->all_connections_prev) {
		c->all_connections_prev->all_connections_next = c->all_connections_next;
	} else {
		si->all_connections_head = c->all_connections_next;
	}

	if (c->all_connections_next) {
		c->all_connections_next->all_connections_prev = c->all_connections_prev;
	} else {
		si->all_connections_tail = c->all_connections_prev;
	}

	si->num_connections--;
}

/*
 * sfe_ipv4_sync_sfe_ipv4_connection()
 *	Sync a connection.
 *
 * On entry to this function we expect that the lock for the connection is either
 * already held or isn't required.
 */
static void sfe_ipv4_gen_sync_sfe_ipv4_connection(struct sfe_ipv4 *si, struct sfe_ipv4_connection *c,
						  struct sfe_connection_sync *sis, sfe_sync_reason_t reason,
						  uint64_t now_jiffies)
{
	struct sfe_ipv4_connection_match *original_cm;
	struct sfe_ipv4_connection_match *reply_cm;

	/*
	 * Fill in the update message.
	 */
	sis->is_v6 = 0;
	sis->protocol = c->protocol;
	sis->src_ip.ip = c->src_ip;
	sis->src_ip_xlate.ip = c->src_ip_xlate;
	sis->dest_ip.ip = c->dest_ip;
	sis->dest_ip_xlate.ip = c->dest_ip_xlate;
	sis->src_port = c->src_port;
	sis->src_port_xlate = c->src_port_xlate;
	sis->dest_port = c->dest_port;
	sis->dest_port_xlate = c->dest_port_xlate;
	original_cm = c->original_match;
	reply_cm = c->reply_match;
	sis->src_td_max_window = original_cm->protocol_state.tcp.max_win;
	sis->src_td_end = original_cm->protocol_state.tcp.end;
	sis->src_td_max_end = original_cm->protocol_state.tcp.max_end;
	sis->dest_td_max_window = reply_cm->protocol_state.tcp.max_win;
	sis->dest_td_end = reply_cm->protocol_state.tcp.end;
	sis->dest_td_max_end = reply_cm->protocol_state.tcp.max_end;

	sis->src_new_packet_count = original_cm->rx_packet_count;
	sis->src_new_byte_count = original_cm->rx_byte_count;
	sis->dest_new_packet_count = reply_cm->rx_packet_count;
	sis->dest_new_byte_count = reply_cm->rx_byte_count;

	sfe_ipv4_connection_match_update_summary_stats(original_cm);
	sfe_ipv4_connection_match_update_summary_stats(reply_cm);

	sis->src_dev = original_cm->match_dev;
	sis->src_packet_count = original_cm->rx_packet_count64;
	sis->src_byte_count = original_cm->rx_byte_count64;

	sis->dest_dev = reply_cm->match_dev;
	sis->dest_packet_count = reply_cm->rx_packet_count64;
	sis->dest_byte_count = reply_cm->rx_byte_count64;

	sis->reason = reason;

	/*
	 * Get the time increment since our last sync.
	 */
	sis->delta_jiffies = now_jiffies - c->last_sync_jiffies;
	c->last_sync_jiffies = now_jiffies;
}

/*
 * sfe_ipv4_flush_sfe_ipv4_connection()
 *	Flush a connection and free all associated resources.
 *
 * We need to be called with bottom halves disabled locally as we need to acquire
 * the connection hash lock and release it again.  In general we're actually called
 * from within a BH and so we're fine, but we're also called when connections are
 * torn down.
 */
static void sfe_ipv4_flush_sfe_ipv4_connection(struct sfe_ipv4 *si, struct sfe_ipv4_connection *c, sfe_sync_reason_t reason)
{
	struct sfe_connection_sync sis;
	uint64_t now_jiffies;
	sfe_sync_rule_callback_t sync_rule_callback;

	rcu_read_lock();
	spin_lock_bh(&si->lock);
	si->connection_flushes++;
	sync_rule_callback = rcu_dereference(si->sync_rule_callback);
	spin_unlock_bh(&si->lock);

	if (sync_rule_callback) {
		/*
		 * Generate a sync message and then sync.
		 */
		now_jiffies = get_jiffies_64();
		sfe_ipv4_gen_sync_sfe_ipv4_connection(si, c, &sis, reason, now_jiffies);
		sync_rule_callback(&sis);
	}

	rcu_read_unlock();

	/*
	 * Release our hold of the source and dest devices and free the memory
	 * for our connection objects.
	 */
	dev_put(c->original_dev);
	dev_put(c->reply_dev);

#if SFE_ACCELERATE_IPSEC
    if ((c->original_match) && (c->original_match->set_accel) && (c->original_match->flow_accel) && (c->original_match->x))
    {
        xfrm_state_put(c->original_match->x);
    }
    if ((c->reply_match) && (c->reply_match->set_accel) && (c->reply_match->flow_accel) && (c->reply_match->x))
    {
        xfrm_state_put(c->reply_match->x);
    }
#endif

	kfree(c->original_match);
	kfree(c->reply_match);
#ifdef TPLINK_FRAGMENT_OUTPUT
	if (c->dst)
		dst_release(c->dst);
	if (c->reply_dst)
		dst_release(c->reply_dst);
#endif
	nf_conntrack_put(c->nfct);
	kfree(c);
}

#ifdef SFE_IPSTAT
static inline void sfe_ipstat_hook(const struct sfe_ipv4_connection_match *cm, const struct sk_buff *skb)
{
	struct sfe_ipv4_stat_info sisi;
	struct nf_conn *ct = (struct nf_conn *)cm->connection->nfct;
	if ((0 == sfe_ipstat_switch) || (cm->rx_packet_count64 > 1)) {
		return;
	}

	sisi.protocol = cm->match_protocol;
	sisi.in = cm->match_dev;
	sisi.out = cm->xmit_dev;
	sisi.src_ip = cm->match_src_ip;
	sisi.dst_ip = cm->xlate_dest_ip;
#ifdef SFE_CONNMARK_STAT
	if(ct)
	{
		sisi.connmark = ct->mark;
	}
#endif
	memcpy(sisi.dst_mac, cm->xmit_dest_mac, ETH_ALEN);
	sfe_ipstat_in(&sisi, skb);
}
#endif

#ifdef TPLINK_FRAGMENT
#define FRAG_CMLIST_SLOT 64
#define FRAG_CMLIST_COUNT 16
#define FRAG_CMLIST_HASH(id)	(id%FRAG_CMLIST_SLOT)

struct frag_node {
	uint16_t id;
	uint8_t proto;
	uint32_t src;
	uint32_t dst;
	uint16_t sport;
	uint16_t dport;
};
struct frag_node FRAG_CMLIST[FRAG_CMLIST_SLOT][FRAG_CMLIST_COUNT];
static uint32_t FRAG_INDEX[FRAG_CMLIST_SLOT];


/* todo: two flows have same id. */
static inline struct frag_node *frag_cmlist_find(uint16_t id, uint8_t proto, uint32_t src, uint32_t dst)
{
	int i = 0, j = 0;
	uint32_t slot = FRAG_CMLIST_HASH(id);
	struct frag_node *nd = NULL;

	j = (FRAG_INDEX[slot] + FRAG_CMLIST_COUNT - 1) % FRAG_CMLIST_COUNT;
	for (i = 0; i < FRAG_CMLIST_COUNT; i++) {
		nd = &FRAG_CMLIST[slot][j];
		if ((nd->id == id)
		  &&(nd->proto == proto)
		  &&(nd->src == src)
		  &&(nd->dst == dst)) {
			return nd;
		}
		j = (j + FRAG_CMLIST_COUNT - 1) % FRAG_CMLIST_COUNT;
	}
	return NULL;
}

static inline void frag_cmlist_insert(uint16_t id, uint8_t proto, uint32_t src, uint32_t dst, uint16_t sport, uint16_t dport)
{
	uint32_t slot = FRAG_CMLIST_HASH(id);
	struct frag_node *nd = NULL;

	nd = &FRAG_CMLIST[slot][FRAG_INDEX[slot] % FRAG_CMLIST_COUNT];
	nd->id = id;
	nd->proto = proto;
	nd->src = src;
	nd->dst = dst;
	nd->sport = sport;
	nd->dport = dport;
	FRAG_INDEX[slot] = (FRAG_INDEX[slot]+1) % FRAG_CMLIST_COUNT;
}
#endif

#ifdef TPLINK_PPP_TUNNEL

#define TUNNEL_MAX 8
static struct tunnel_info *find_tunnel_info_origin(struct net_device *ppp_dev);

enum tunnel_type {
	TUNNEL_PPPOE,
	TUNNEL_PPTP,
	TUNNEL_L2TP
};

struct tunnel_info {
	union {
		struct {
			uint16_t call_id;
			uint32_t saddr;
		} pptp;
		struct {
			uint32_t tunnel_id;
			uint32_t session_id;
			uint16_t sport;
			uint16_t dport;
#if SFE_ACCELERATE_IPSEC
			int   	 encap; 
			struct xfrm_state *x_in;
			struct xfrm_state *x_out;
#endif
		} l2tp;
		struct {
			uint16_t sid;
			uint8_t  source[ETH_ALEN];
		} pppoe;
	} info;
	enum tunnel_type type; /* 0: pppoe, 1: pptp, 2: l2tp */

	uint32_t magic_number;
	uint32_t saddr;
	uint32_t daddr;
	uint16_t dest_mac[ETH_ALEN / 2];
	uint16_t src_mac[ETH_ALEN / 2];
	struct net_device *tunnel_dev;
	struct net_device *ppp_dev; /* key */

	uint32_t debug_read_seq;

	struct sk_buff *(*add_header)(struct tunnel_info *ti, struct sk_buff *skb, uint16_t protocol);
}TUNNEL_LIST[TUNNEL_MAX];
#endif

#if SFE_ACCELERATE_IPSEC

#define XFRM_DST_MAX 512
DEFINE_RWLOCK(xfrm_dst_rwlock);

struct sfe_esp_xfrm_dst {
	struct xfrm_state *x;
	struct net_device *dev;
	uint32_t src_ip;
	uint32_t dst_ip;
	uint16_t dest_mac[ETH_ALEN / 2];
	uint16_t src_mac[ETH_ALEN / 2];
	uint32_t debug_read_seq;
}SFE_XFRM_DST[XFRM_DST_MAX];

static struct sfe_esp_xfrm_dst * sfe_get_xfrm_dst_byip(uint32_t sip, uint32_t dip)
{
	int i = 0;
	read_lock_bh(&xfrm_dst_rwlock);
	for (i = 0; i < XFRM_DST_MAX; i++) {
		if (SFE_XFRM_DST[i].x && 
			SFE_XFRM_DST[i].src_ip == sip &&
			SFE_XFRM_DST[i].dst_ip == dip) {
			read_unlock_bh(&xfrm_dst_rwlock);
			return &SFE_XFRM_DST[i];
		}		
	}

	read_unlock_bh(&xfrm_dst_rwlock);
	return NULL;
}

static struct sfe_esp_xfrm_dst * sfe_get_xfrm_dst_byxfrm(struct xfrm_state *x)
{
	int i = 0;

	read_lock_bh(&xfrm_dst_rwlock);
	for (i = 0; i < XFRM_DST_MAX; i++) {
		if (SFE_XFRM_DST[i].x && 
			SFE_XFRM_DST[i].x == x &&
			SFE_XFRM_DST[i].src_ip &&
			SFE_XFRM_DST[i].dst_ip) {
			read_unlock_bh(&xfrm_dst_rwlock);
			return &SFE_XFRM_DST[i];
		}			
	}

	read_unlock_bh(&xfrm_dst_rwlock);
	return NULL;
}

static int sfe_init_xfrm_dst(struct xfrm_state *x, struct net_device *dev, uint8_t *dest_mac, uint32_t sip, uint32_t dip)
{
	int i = 0;
	struct sfe_esp_xfrm_dst *dst = NULL;
	DEBUG_INFO("sfe_init_xfrm_dst");
	write_lock_bh(&xfrm_dst_rwlock);
	for (i = 0; i < XFRM_DST_MAX; i++) {
		if (!SFE_XFRM_DST[i].dev) {
			DEBUG_TRACE("find a empty xfrm_dst struct(%d)\n", i);
			dst = &(SFE_XFRM_DST[i]);
			dst->x = x;
			dst->dev = dev;
			dst->src_ip = sip;
			dst->dst_ip = dip;
			dst->debug_read_seq = 0;
			memcpy(dst->dest_mac, dest_mac, ETH_ALEN);
			memcpy(dst->src_mac, dev->dev_addr, ETH_ALEN);
			write_unlock_bh(&xfrm_dst_rwlock);
			return 0;
		}
	}
	write_unlock_bh(&xfrm_dst_rwlock);
	return -1;
}

static void sfe_destroy_xfrm_dst_byxfrm(struct xfrm_state *x)
{
	int i = 0;
	write_lock_bh(&xfrm_dst_rwlock);
	for (i = 0; i < XFRM_DST_MAX; i++) {
		if (SFE_XFRM_DST[i].x == x) {
			SFE_XFRM_DST[i].x = NULL;
			SFE_XFRM_DST[i].dev = NULL;
			memset(&SFE_XFRM_DST[i], 0, sizeof(struct sfe_esp_xfrm_dst));
			DEBUG_TRACE("xfrm_dst cleaned\n");
			continue;
		}
	}
	write_unlock_bh(&xfrm_dst_rwlock);
	return;
}

static void sfe_destroy_xfrm_dst_bydev(struct ney_device *dev)
{
	int i = 0;
	write_lock_bh(&xfrm_dst_rwlock);
	for (i = 0; i < XFRM_DST_MAX; i++) {
		if (!dev)
		{
			SFE_XFRM_DST[i].x = NULL;
			SFE_XFRM_DST[i].dev = NULL;
			memset(&SFE_XFRM_DST[i], 0, sizeof(struct sfe_esp_xfrm_dst));
			DEBUG_TRACE("ALL: xfrm_dst cleaned\n");
			continue;
		}
		if (SFE_XFRM_DST[i].dev == dev) {
			SFE_XFRM_DST[i].x = NULL;
			SFE_XFRM_DST[i].dev = NULL;
			memset(&SFE_XFRM_DST[i], 0, sizeof(struct sfe_esp_xfrm_dst));
			DEBUG_TRACE("xfrm_dst cleaned\n");
			continue;
		}
	}
	write_unlock_bh(&xfrm_dst_rwlock);
	return;
}

static bool sfe_xfrm_state_equal(struct xfrm_state *x1, struct xfrm_state *x2)
{
	return (x1->props.family == x2->props.family &&
			x1->props.reqid == x2->props.reqid &&
			x1->id.proto == x2->id.proto &&
			x1->props.mode == x2->props.mode &&
			xfrm_state_addr_check(x1, &x2->id.daddr, &x2->props.saddr, x2->props.family)
			);
}

static void sfe_update_xfrm_dst(struct xfrm_state *x)
{
	int i = 0;

	if (!x)
		return;
	write_lock_bh(&xfrm_dst_rwlock);
	for (i = 0; i < XFRM_DST_MAX; i++) {
		if (SFE_XFRM_DST[i].x && sfe_xfrm_state_equal(SFE_XFRM_DST[i].x, x)) {
			SFE_XFRM_DST[i].x = x;
			DEBUG_TRACE("update xfrm_dst !!!");
		}
		continue;
	}
	write_unlock_bh(&xfrm_dst_rwlock);
	return;
}

#define ESP_SKB_CB(__skb) ((struct esp_skb_cb *)&((__skb)->cb[0]))

struct esp_skb_cb {
	struct xfrm_skb_cb xfrm;
	void *tmp;
};

struct esp_output_extra {
	__be32 seqhi;
	u32 esphoff;
};

static bool find_dst_dev(struct net *net, __be32 dip, struct net_device **dev, uint8_t *mac_addr, __be32 sip)
{
    bool ret = false;
	struct neighbour *neigh = NULL;
	struct rtable *rt = NULL;
	struct dst_entry *dst = NULL;
	struct net_device *mac_dev = NULL;

	/*
	 * Look up the rtable entry for the IP address then get the hardware
	 * address from its neighbour structure.  This means this work when the
	 * neighbours are routers too.
	 */
    rt = ip_route_output(net, dip, sip, 0, 0);
	if (unlikely(IS_ERR(rt))) 
    {
        return false;
	}

	dst = (struct dst_entry *)rt;

	rcu_read_lock();
	neigh = dst_neigh_lookup(dst, &dip);
	if (unlikely(!neigh)) {
		goto release;
	}

	if (unlikely(!(neigh->nud_state & NUD_VALID))) {
		goto release;
	}

	mac_dev = neigh->dev;
	if (!mac_dev) {
		goto release;
	}

	memcpy(mac_addr, neigh->ha, (size_t)mac_dev->addr_len);

	dev_hold(mac_dev);
	
	*dev = mac_dev;
	ret = true;

release:
    rcu_read_unlock();
    if (neigh)
    {
        neigh_release(neigh);
    }
    if (dst)
    {
        dst_release(dst);
    }

	return ret;
}

/*
 * It is similar with function esp_alloc_tmp() in esp4.c
 */
static void *sfe_esp_alloc_tmp(struct crypto_aead *aead, int nfrags, int extralen)
{
	unsigned int len;

	len = extralen;

	len += crypto_aead_ivsize(aead);

	if (len) {
		len += crypto_aead_alignmask(aead) &
		       ~(crypto_tfm_ctx_alignment() - 1);
		len = ALIGN(len, crypto_tfm_ctx_alignment());
	}

	len += sizeof(struct aead_request) + crypto_aead_reqsize(aead);
	len = ALIGN(len, __alignof__(struct scatterlist));

	len += sizeof(struct scatterlist) * nfrags;

	return kmalloc(len, GFP_ATOMIC);
}

/* almost is the same with functin xfrm4_mode_tunnel_output in xfrm4_mode_tunnel.c */
static void sfe_add_outer_ip_hdr(struct xfrm_state *x, struct sk_buff *skb)
{
	struct iphdr *top_iph = NULL;
	int flags = 0;

	skb_set_network_header(skb, -x->props.header_len);
	skb->mac_header = skb->network_header +
			  offsetof(struct iphdr, protocol);
	skb->transport_header = skb->network_header + sizeof(*top_iph);
	top_iph = ip_hdr(skb);

	top_iph->ihl = 5;
	top_iph->version = 4;

	top_iph->protocol = IPPROTO_IPIP;

	/* DS disclosing depends on XFRM_SA_XFLAG_DONT_ENCAP_DSCP */
	if (x->props.extra_flags & XFRM_SA_XFLAG_DONT_ENCAP_DSCP)
		top_iph->tos = 0;
	else
		top_iph->tos = XFRM_MODE_SKB_CB(skb)->tos;
	top_iph->tos = INET_ECN_encapsulate(top_iph->tos,
					    XFRM_MODE_SKB_CB(skb)->tos);

	flags = x->props.flags;
	if (flags & XFRM_STATE_NOECN)
		IP_ECN_clear(top_iph);

	top_iph->frag_off = (flags & XFRM_STATE_NOPMTUDISC) ?
		0 : (XFRM_MODE_SKB_CB(skb)->frag_off & htons(IP_DF));

    top_iph->ttl = 64;

	top_iph->saddr = x->props.saddr.a4;
	top_iph->daddr = x->id.daddr.a4;
	ip_select_ident(xs_net(x), skb, NULL);

	return;
}

static void sfe_esp_dec_done(struct crypto_async_request *base, int err)
{
	struct sk_buff *skb = base->data;
	struct iphdr *iph;
	struct xfrm_state *x = xfrm_input_state(skb);
	struct crypto_aead *aead = x->data;
	int alen = crypto_aead_authsize(aead);
	int hlen = sizeof(struct ip_esp_hdr) + crypto_aead_ivsize(aead);
	int elen = skb->len - hlen;
	int ihl;
	u8 nexthdr[2];
	int padlen;
	uint8_t dest_mac[ETH_ALEN];
    struct net_device *dst_dev = NULL;

	kfree(ESP_SKB_CB(skb)->tmp);
	if (unlikely(err))
	{
		goto drop_it;
	}
	skb_hnat_alg(skb) = 1;
	if (skb_copy_bits(skb, skb->len-alen-2, nexthdr, 2))
	{
		BUG();
	}

    /* check pad */
	padlen = nexthdr[0];
	if (padlen + 2 + alen >= elen)
	{
		goto drop_it;
	}

	iph = ip_hdr(skb);
	ihl = iph->ihl * 4;

	if (x->encap) {
		struct xfrm_encap_tmpl *encap = x->encap;
		struct udphdr *uh = (void *)(skb_network_header(skb) + ihl);

		/*
		 * 1) if the NAT-T peer's IP or port changed then
		 *    advertize the change to the keying daemon.
		 *    This is an inbound SA, so just compare
		 *    SRC ports.
		 */
		if (iph->saddr != x->props.saddr.a4 ||
		    uh->source != encap->encap_sport) {
			xfrm_address_t ipaddr;

			ipaddr.a4 = iph->saddr;
			km_new_mapping(x, &ipaddr, uh->source);

			/* XXX: perhaps add an extra
			 * policy check here, to see
			 * if we should allow or
			 * reject a packet from a
			 * different source
			 * address/port.
			 */
		}

		/*
		 * 2) ignore UDP/TCP checksums in case
		 *    of NAT-T in Transport Mode, or
		 *    perform other post-processing fixes
		 *    as per draft-ietf-ipsec-udp-encaps-06,
		 *    section 3.1.2
		 */
		if (x->props.mode == XFRM_MODE_TRANSPORT)
			skb->ip_summed = CHECKSUM_UNNECESSARY;
	}

    pskb_trim(skb, skb->len - alen - padlen - 2);

    __skb_pull(skb, hlen);
    
    if (!(x->outer_mode->flags & XFRM_MODE_FLAG_TUNNEL))
	{
		skb_set_transport_header(skb, -ihl);

        if (x->inner_mode)
        {
            x->inner_mode->input(x, skb);
        }

        iph = ip_hdr(skb);
        iph->protocol = nexthdr[1];
        __skb_push(skb, skb->data - skb_network_header(skb));
        iph->tot_len = htons(skb->len);
        ip_send_check(iph);
	}

    skb->protocol = htons(ETH_P_IP);
    skb_reset_network_header(skb);
    skb_reset_mac_len(skb);

    if (!sfe_ipv4_recv_core(NULL, skb, 0, SFE_PROTO_IPSEC))
    {
		skb->mark |= x->mark.v;
        DEBUG_INFO("maybe can not be accelerated, go slow path:%d dev:%s\n", skb->len, skb->dev);
		skb_dst_drop(skb);
		netif_rx(skb);
    }

    return;

drop_it:
 	DEBUG_TRACE("sfe_esp_dec_done drop_it\n");
    kfree_skb(skb);
    return;
}

static int sfe_esp_dec_done2(struct sk_buff *skb)
{
	struct iphdr *iph;
	struct xfrm_state *x = xfrm_input_state(skb);
	struct crypto_aead *aead = x->data;
	int alen = crypto_aead_authsize(aead);
	int hlen = sizeof(struct ip_esp_hdr) + crypto_aead_ivsize(aead);
	int elen = skb->len - hlen;
	int ihl;
	u8 nexthdr[2];
	int padlen;

	kfree(ESP_SKB_CB(skb)->tmp);
	skb_hnat_alg(skb) = 1;
	if (skb_copy_bits(skb, skb->len-alen-2, nexthdr, 2))
		BUG();

	padlen = nexthdr[0];
	if (padlen + 2 + alen >= elen)
		return -1;

	/* ... check padding bits here. Silly. :-) */

	iph = ip_hdr(skb);
	ihl = iph->ihl * 4;

	if (x->encap) {
		struct xfrm_encap_tmpl *encap = x->encap;
		struct udphdr *uh = (void *)(skb_network_header(skb) + ihl);

		/*
		 * 1) if the NAT-T peer's IP or port changed then
		 *    advertize the change to the keying daemon.
		 *    This is an inbound SA, so just compare
		 *    SRC ports.
		 */
		if (iph->saddr != x->props.saddr.a4 ||
		    uh->source != encap->encap_sport) {
			xfrm_address_t ipaddr;

			ipaddr.a4 = iph->saddr;
			km_new_mapping(x, &ipaddr, uh->source);

			/* XXX: perhaps add an extra
			 * policy check here, to see
			 * if we should allow or
			 * reject a packet from a
			 * different source
			 * address/port.
			 */
		}

		/*
		 * 2) ignore UDP/TCP checksums in case
		 *    of NAT-T in Transport Mode, or
		 *    perform other post-processing fixes
		 *    as per draft-ietf-ipsec-udp-encaps-06,
		 *    section 3.1.2
		 */
		if (x->props.mode == XFRM_MODE_TRANSPORT)
			skb->ip_summed = CHECKSUM_UNNECESSARY;
	}

	pskb_trim(skb, skb->len - alen - padlen - 2);
	__skb_pull(skb, hlen);
	if (!(x->outer_mode->flags & XFRM_MODE_FLAG_TUNNEL))
	{
		skb_set_transport_header(skb, -ihl);

        if (x->inner_mode)
        {
            x->inner_mode->input(x, skb);
        }

        iph = ip_hdr(skb);
        iph->protocol = nexthdr[1];
        __skb_push(skb, skb->data - skb_network_header(skb));
        iph->tot_len = htons(skb->len);
        ip_send_check(iph);
	}
	else
		skb_reset_transport_header(skb);

    skb->protocol = htons(ETH_P_IP);
    skb_reset_network_header(skb);
    skb_reset_mac_len(skb);

    return 0;
}

/*
 * sfe_ipv4_recv_esp()
 * Handle esp packet receives and forwarding.
 * Returns 1 if the packet is forwarded or 0 if it isn't.
 */
static int sfe_ipv4_recv_esp(struct sfe_ipv4 *si, struct sk_buff *skb, struct sfe_ipv4_ip_hdr *iph, bool espInUdp, int sfe_proto)
{
	struct crypto_aead *aead = NULL;
	struct aead_request *req;
	struct sk_buff *trailer;
	struct scatterlist *sg = NULL;
    struct xfrm_state *x = NULL;
	struct ip_esp_hdr *esph = NULL;
    struct sec_path *sp = NULL;
	u8 *iv = NULL;
    int ivlen = 0;
	int elen = 0;
	int nfrags = 0;
	void *tmp = NULL;
    int ret = 0;

    if (espInUdp)
    {
        if (unlikely(!pskb_may_pull(skb, (sizeof(struct ip_esp_hdr) + sizeof(struct sfe_ipv4_udp_hdr) + (iph->ihl<<2))))) 
        {
            DEBUG_TRACE("packet too short for IPSec-in-UDP header\n");
            return 0;
        }
        esph = (struct ip_esp_hdr *)(skb->data + (iph->ihl<<2) + sizeof(struct sfe_ipv4_udp_hdr));
    }
    else
    {
        esph = (struct ip_esp_hdr *)(skb->data + (iph->ihl<<2));
    }

	x = xfrm_state_lookup(dev_net(skb->dev), skb->mark, (const xfrm_address_t *)&iph->daddr, esph->spi, IPPROTO_ESP, AF_INET);
    if (x == NULL) 
    {
        /* maybe l2tp-over-ipsec, look up xfrm_state by spi. */
		x = xfrm_state_lookup_byspi(dev_net(skb->dev), esph->spi, AF_INET);
		if (x == NULL)
			return 0;
    }
    if ((x->data == NULL) || (x->km.state != XFRM_STATE_VALID))
    {
        xfrm_state_put(x);
        return 0;
    }

    if (skb->dev && 
        (sfe_proto > SFE_PROTO_DEF && sfe_proto < SFE_PROTO_IPSEC ) && 
        (skb->dev->flags & IFF_POINTOPOINT) && 
        (!dev_tuntype(skb->dev))) {
            ppp_dev_add_stats(skb->dev, skb, skb->len, 1);
        }

	skb_hnat_alg(skb) = 1;
	/* Allocate new secpath or COW existing one. */
	if (!skb->sp || atomic_read(&skb->sp->refcnt) != 1)
    {
		sp = secpath_dup(skb->sp);
		if (!sp) {
            xfrm_state_put(x);
            kfree_skb(skb);
			DEBUG_INFO("xfrm_state_lookup_byspi secpath_dup\n");
            return 1;
		}
		if (skb->sp)
			secpath_put(skb->sp);
		skb->sp = sp;
	}

    /* in xfrm_state_lookup(), xfrm_state_hold() is called, so pointer x can be saved in xvec */
    skb->sp->xvec[skb->sp->len++] = x;

    spin_lock(&x->lock);
    if ((x->repl->check(x, skb, esph->seq_no)) || (xfrm_state_check_expire(x))) 
    {
        spin_unlock(&x->lock);
        goto drop_it;
    }
    
	spin_unlock(&x->lock);

	__skb_pull(skb, (iph->ihl<<2));
    if (espInUdp)
    {
        __skb_pull(skb, sizeof(struct sfe_ipv4_udp_hdr));
    }

    aead = x->data;
    ivlen = crypto_aead_ivsize(aead);
	elen = skb->len - sizeof(*esph) - ivlen;
	if (elen <= 0)
    {
        goto drop_it;
    }

	if (!pskb_may_pull(skb, sizeof(*esph) + ivlen))
    {
        goto drop_it;
    }

	if (0 > (nfrags = skb_cow_data(skb, 0, &trailer)))
    {
        goto drop_it;
    }

	tmp = sfe_esp_alloc_tmp(aead, nfrags, 0);
	if (!tmp)
	{
        goto drop_it;
	}

	ESP_SKB_CB(skb)->tmp = tmp;

    if (crypto_aead_ivsize(aead))
    {
        iv = PTR_ALIGN((u8 *)tmp, crypto_aead_alignmask(aead) + 1);
    }
    else
    {
        iv = tmp;
    }

    req = (void *)PTR_ALIGN(iv + crypto_aead_ivsize(aead), crypto_tfm_ctx_alignment());
    aead_request_set_tfm(req, aead);

    sg = (void *)ALIGN((unsigned long)(req + 1) + crypto_aead_reqsize(aead), __alignof__(struct scatterlist));

	skb->ip_summed = CHECKSUM_NONE;

	aead_request_set_callback(req, 0, sfe_esp_dec_done, skb);

	sg_init_table(sg, nfrags);
	if (unlikely(skb_to_sgvec(skb, sg, 0, skb->len) < 0))
	{
        goto drop_it;
	}

	aead_request_set_crypt(req, sg, sg, elen + ivlen, iv);
	aead_request_set_ad(req, sizeof(*esph));

    ret = crypto_aead_decrypt(req);
    switch (ret)
    {
    case -EINPROGRESS:
        return 1;

    case 0:
        /* not hard crypto, maybe soft sync decrypt */
        if (!((NULL != aead->base.__crt_alg) && (NULL != strstr(aead->base.__crt_alg->cra_driver_name ,"eip93")))) 
        {
            if (0 == sfe_esp_dec_done2(skb))
            {
                return sfe_ipv4_recv_core(NULL, skb, 0, SFE_PROTO_IPSEC);
            }
        }

        break;

    default:
        break;
    }

drop_it:
    if (tmp != NULL)
    {
        kfree(tmp);
        ESP_SKB_CB(skb)->tmp = NULL;
    }

    kfree_skb(skb);
    return 1;
}

static int sfe_esp_ppp_out(struct sk_buff *skb, struct net_device *dst_dev, int flag, uint8_t *dest_mac)
{

#if (LINUX_VERSION_CODE > KERNEL_VERSION(3, 9, 11))
		skb->vlan_proto = 0;
#else
		if (vlan_tx_tag_present(skb))
			skb->protocol = 0;
#endif 
		skb->vlan_tci = 0;

		skb->dev = dst_dev;

		//DEBUG_TRACE("encap success, start to send it to pppdev:%s <sip:%pI4 dip:%pI4> skblen:%d mtu:%d\n",
					//dst_dev->name, skb->len, skb->dev->mtu, &iph->saddr, &iph->daddr);
		if (flag)
			dev_put(dst_dev);
			
		//skb->fast_forwarded = 1;
#ifdef TPLINK_FRAGMENT_OUTPUT
		if (skb->len > skb->dev->mtu) 
			return sfe_ipv4_fragment(skb, false);
#endif
		return sfe_ipv4_dev_queue_xmit(skb);
}

static int sfe_esp_eth_out(struct sk_buff *skb, struct net_device *dst_dev, int flag, uint8_t *dest_mac)
{
	struct sfe_ipv4_eth_hdr *eth = NULL;

	eth = (struct sfe_ipv4_eth_hdr *)__skb_push(skb, ETH_HLEN);
	eth->h_proto = htons(ETH_P_IP);

	memcpy(eth->h_dest, dest_mac, ETH_ALEN);
	memcpy(eth->h_source, dst_dev->dev_addr, ETH_ALEN);

#if (LINUX_VERSION_CODE > KERNEL_VERSION(3, 9, 11))
	skb->vlan_proto = 0;
#else
	if (vlan_tx_tag_present(skb))
		skb->protocol = 0;
#endif 
	skb->vlan_tci = 0;
	skb->dev = dst_dev;

	//DEBUG_TRACE("encap success, start to send it to ethdev:%s <sip:%pI4 dip:%pI4>\n", 
				//dst_dev->name, &iph->saddr, &iph->daddr);
	if (flag)
		dev_put(dst_dev);

	skb->fast_forwarded = 1;
	return dev_queue_xmit(skb);
}

static void sfe_esp_enc_done2(struct sk_buff *skb)
{
	struct iphdr *iph = NULL;
	uint8_t dest_mac[ETH_ALEN];
    struct net_device *dst_dev = NULL;
    struct sfe_ipv4_eth_hdr *eth = NULL;
	struct sfe_esp_xfrm_dst *dst = NULL;
	int flag = 0;

	kfree(ESP_SKB_CB(skb)->tmp);
    ESP_SKB_CB(skb)->tmp = NULL;
	skb_hnat_alg(skb) = 1;
	iph = (struct iphdr *)(skb->data);
	dst = sfe_get_xfrm_dst_byip(iph->saddr, iph->daddr);
	if (!dst)
	{
		if ((wan_inner_dev) && find_dst_dev(dev_net(wan_inner_dev), iph->daddr, &dst_dev, dest_mac, iph->saddr))
		{
			flag = 1;
		}
		else
		{
			DEBUG_TRACE("No found xfrm_dst_cache, free it <sip:%pI4 dip:%pI4>\n", &iph->saddr, &iph->daddr);
			kfree_skb(skb);
			return;
		}
	}
	else
	{
		DEBUG_TRACE("found xfrm_dst_cache, accelerate it <sip:%pI4 dip:%pI4>\n", &iph->saddr, &iph->daddr);
		dst_dev = dst->dev;
		memcpy(dest_mac, dst->dest_mac, ETH_ALEN);
	}

	iph->tot_len = htons(skb->len);
	ip_send_check(iph);

#ifdef TPLINK_PPP_TUNNEL        
	if (dst_dev->flags & IFF_POINTOPOINT)
	{
		return sfe_esp_ppp_out(skb, dst_dev, flag, dest_mac);
	}
#endif

	return sfe_esp_eth_out(skb, dst_dev, flag, dest_mac);
}

static void sfe_esp_enc_done(struct crypto_async_request *base, int err)
{
    sfe_esp_enc_done2((struct sk_buff *)base->data);
    return;
}

static int sfe_ipv4_encap_esp_one(struct xfrm_state *x, struct sk_buff *skb, struct sfe_ipv4_ip_hdr *iph)
{
	struct ip_esp_hdr *esph;
	struct crypto_aead *aead;
	struct aead_request *req;
	struct scatterlist *sg;
	struct sk_buff *trailer;
	void *tmp = NULL;
	u8 *iv;
	u8 *tail;
	u8 proto;
	int blksize;
	int clen;
	int alen;
	int plen;
	int tailen;
	int ivlen;
	int nfrags;
	int assoclen;
	__be64 seqno;
	int err = 0;
    int max_headroom = 0;
	max_headroom = x->props.header_len + sizeof(struct sfe_ipv4_eth_hdr);
	if (skb_headroom(skb) < max_headroom) {
		DEBUG_WARN("====skb_headroom(skb): %u, max_headroom:%u skb->len %d skb_headlen(skb) %d\n", skb_headroom(skb), max_headroom,skb->len);
		if(skb_cow_head(skb, max_headroom))
		{
			DEBUG_ERROR( "skb_cow_head wrong ====skb_headroom(skb): %u, max_headroom:%u skb->len %d skb_headlen(skb) %d\n", skb_headroom(skb), max_headroom,skb->len);
			goto drop_it;
		}
	}
    if (x->outer_mode->flags & XFRM_MODE_FLAG_TUNNEL)
	{
        XFRM_MODE_SKB_CB(skb)->protocol = iph->protocol;
        XFRM_MODE_SKB_CB(skb)->ihl = sizeof(*iph);
        XFRM_MODE_SKB_CB(skb)->id = iph->id;
        XFRM_MODE_SKB_CB(skb)->frag_off = iph->frag_off;
        XFRM_MODE_SKB_CB(skb)->tos = iph->tos;
        XFRM_MODE_SKB_CB(skb)->ttl = iph->ttl;
        XFRM_MODE_SKB_CB(skb)->optlen = iph->ihl * 4 - sizeof(*iph);
        XFRM_MODE_SKB_CB(skb)->flow_lbl[0] = 0;
        XFRM_MODE_SKB_CB(skb)->flow_lbl[1] = 0;
        XFRM_MODE_SKB_CB(skb)->flow_lbl[2] = 0;
        
        skb->protocol = htons(ETH_P_IP);
        
        sfe_add_outer_ip_hdr(x, skb);
	}
    else
    {
        x->outer_mode->output(x, skb);
    }
    spin_lock_bh(&x->lock);
    if (unlikely(x->km.state != XFRM_STATE_VALID)) 
    {
        spin_unlock_bh(&x->lock);
        goto drop_it;
    }
    
    if (xfrm_state_check_expire(x)) 
    {
        spin_unlock_bh(&x->lock);
        goto drop_it;
    }
    
    if (x->repl->overflow(x, skb))
    {
        spin_unlock_bh(&x->lock);
        goto drop_it;
    }
    
    x->curlft.bytes += skb->len;
    x->curlft.packets++;
    spin_unlock_bh(&x->lock);
    /* try to enc */
    aead = x->data;
    alen = crypto_aead_authsize(aead);
    ivlen = crypto_aead_ivsize(aead);

    blksize = ALIGN(crypto_aead_blocksize(aead), 4);
    clen = ALIGN(skb->len + 2, blksize);
    plen = clen - skb->len;
	tailen = plen + alen;
    err = skb_cow_data(skb, plen + alen, &trailer);
    if (err < 0)
    {
        goto drop_it;
    }
    nfrags = err;
    assoclen = sizeof(*esph);
    tmp = sfe_esp_alloc_tmp(aead, nfrags, 0);
    if (!tmp) 
    {
        goto drop_it;
    }
    
    if (crypto_aead_ivsize(aead))
    {
        iv = PTR_ALIGN((u8 *)tmp, crypto_aead_alignmask(aead) + 1);
    }
    else
    {
        iv = tmp;
    }
    
	req = (void *)PTR_ALIGN(iv + crypto_aead_ivsize(aead), crypto_tfm_ctx_alignment());
	aead_request_set_tfm(req, aead);

    sg = (void *)ALIGN((unsigned long)(req + 1) + crypto_aead_reqsize(aead), __alignof__(struct scatterlist));

	esph = ip_esp_hdr(skb);
	proto = *skb_mac_header(skb);
    *skb_mac_header(skb) = IPPROTO_ESP;
	/* this is non-NULL only with UDP Encapsulation */
	if (x->encap) {
		struct xfrm_encap_tmpl *encap = x->encap;
		struct udphdr *uh;
		__be32 *udpdata32;
		__be16 sport, dport;
		int encap_type;

		spin_lock_bh(&x->lock);
		sport = encap->encap_sport;
		dport = encap->encap_dport;
		encap_type = encap->encap_type;
		spin_unlock_bh(&x->lock);

		uh = (struct udphdr *)esph;
		uh->source = sport;
		uh->dest = dport;
		uh->len = htons(skb->len + tailen - skb_transport_offset(skb));
		uh->check = 0;

		switch (encap_type) {
		default:
		case UDP_ENCAP_ESPINUDP:
			esph = (struct ip_esp_hdr *)(uh + 1);
			break;
		case UDP_ENCAP_ESPINUDP_NON_IKE:
			udpdata32 = (__be32 *)(uh + 1);
			udpdata32[0] = udpdata32[1] = 0;
			esph = (struct ip_esp_hdr *)(udpdata32 + 2);
			break;
		}

		*skb_mac_header(skb) = IPPROTO_UDP;
	}
    /* Fill padding... */
    tail = skb_tail_pointer(trailer);
    do {
        int i;
        for (i = 0; i < plen - 2; i++)
            tail[i] = i + 1;
    } while (0);
    tail[plen - 2] = plen - 2;
    tail[plen - 1] = proto;
    pskb_put(skb, trailer, clen - skb->len + alen);
    
    skb_push(skb, -skb_network_offset(skb));
    esph->seq_no = htonl(XFRM_SKB_CB(skb)->seq.output.low);

    aead_request_set_callback(req, 0, sfe_esp_enc_done, skb);

    esph->spi = x->id.spi;

    sg_init_table(sg, nfrags);
    err = skb_to_sgvec(skb, sg,
                   (unsigned char *)esph - skb->data,
                   assoclen + ivlen + clen + alen);
    if (unlikely(err < 0))
    {
        goto free_mem;
    }

    aead_request_set_crypt(req, sg, sg, ivlen + clen, iv);
    aead_request_set_ad(req, assoclen);

    seqno = cpu_to_be64(XFRM_SKB_CB(skb)->seq.output.low +
                ((u64)XFRM_SKB_CB(skb)->seq.output.hi << 32));

    memset(iv, 0, ivlen);
    memcpy(iv + ivlen - min(ivlen, 8), (u8 *)&seqno + 8 - min(ivlen, 8),
           min(ivlen, 8));

    ESP_SKB_CB(skb)->tmp = tmp;
    err = crypto_aead_encrypt(req);

    switch (err)
    {
    case -EINPROGRESS:
        return 1;

    case 0:
        /* not hard crypto, maybe soft sync encrypt */
        if (!((NULL != aead->base.__crt_alg) && (NULL != strstr(aead->base.__crt_alg->cra_driver_name ,"eip93")))) 
        {
            sfe_esp_enc_done2(skb);
            return 1;
        }

        break;

    default:
        break;
    }

free_mem:
    kfree(tmp);
    ESP_SKB_CB(skb)->tmp = NULL;

drop_it:
    kfree_skb(skb);
    return 1;
}

static int sfe_ipv4_encap_esp_gso(struct xfrm_state *x, struct sk_buff *skb)
{
    int err = -1;
	struct sk_buff *segs = NULL;
    struct sfe_ipv4_ip_hdr *iph = NULL;
    struct sk_buff *nskb = NULL;

	BUILD_BUG_ON(sizeof(*IPCB(skb)) > SKB_SGO_CB_OFFSET);
	BUILD_BUG_ON(sizeof(*IP6CB(skb)) > SKB_SGO_CB_OFFSET);
	segs = skb_gso_segment(skb, 0);
	kfree_skb(skb);
	if (IS_ERR(segs) || (segs == NULL))
	{
        return 0;
	}

	do 
    {
		nskb = segs->next;
		segs->next = NULL;

        iph = (struct sfe_ipv4_ip_hdr *)segs->data;
		err = sfe_ipv4_encap_esp_one(x, segs, iph);

		if (err != 1) 
        {
			kfree_skb_list(nskb);
			return 1;
		}

		segs = nskb;
	} while (segs);

	return 1;
}

static int sfe_ipv4_encap_esp(struct xfrm_state *x, struct sk_buff *skb, struct sfe_ipv4_ip_hdr *iph)
{
	uint8_t dest_mac[ETH_ALEN];
    struct net_device *dst_dev = NULL;
	struct sfe_esp_xfrm_dst *dst = NULL;
	int err = -1;
	int ret = 0;

	dst = sfe_get_xfrm_dst_byxfrm(x);
	if (!dst) 
	{
		if ((wan_inner_dev) && find_dst_dev(dev_net(wan_inner_dev), x->id.daddr.a4, &dst_dev, dest_mac, x->props.saddr.a4))
		{
#ifdef TPLINK_PPP_TUNNEL
			if (dst_dev->flags & IFF_POINTOPOINT && 
				!find_tunnel_info_origin(dst_dev))
#else
			if (dst_dev->flags & IFF_POINTOPOINT)
#endif
			{
				dev_put(dst_dev);
				DEBUG_INFO("no found ppp tunnel: %s \n", dst_dev->name);
				return 0;
			}
			/*init xfrm_dst*/
			err = sfe_init_xfrm_dst(x, dst_dev, dest_mac, x->props.saddr.a4, x->id.daddr.a4);
			DEBUG_INFO("init xfrm_dst: saddr:%pI4 daddr:%pI4 dst_dev:%s!!!\n", &x->props.saddr.a4, &x->id.daddr.a4, dst_dev->name);
			dev_put(dst_dev);
			if (err)
			{
				DEBUG_WARN("XFRM_DST is FULL!!!\n");
			}
		}
		else {
			DEBUG_TRACE("no found dst_dev\n");
			return 0;
		}
	}

    if (skb_is_gso(skb))
    {
        return sfe_ipv4_encap_esp_gso(x, skb);
    }
    else
    {
		if (skb->ip_summed == CHECKSUM_PARTIAL) {
			err = skb_checksum_help(skb);
			if (err) {
				kfree_skb(skb);
				return 1;
			}
		}
        return sfe_ipv4_encap_esp_one(x, skb, iph);
    }
}

#endif

#ifdef TPLINK_PPP_TUNNEL
extern int (* fastpath_lookup_chan)(uint16_t call_id, uint32_t s_addr);
extern int (* fastpath_update_chan)(uint16_t call_id, uint32_t s_addr, uint8_t *ghdr);
extern int (* fastpath_add_pptp_header)(uint16_t call_id, uint32_t s_addr, struct sk_buff *skb);
extern int (* fastpath_lookup_session)(struct net *net, uint32_t tunnel_id, uint32_t session_id, uint16_t hdrflags);
extern int (* fastpath_update_session)(struct net *net, uint32_t tunnel_id, uint32_t session_id, uint16_t hdrflags);
extern int (* fastpath_add_l2tp_header)(struct net *net, uint32_t tunnel_id, uint32_t session_id, struct sk_buff *skb);
extern int (* fastpath_lookup_item)(struct net_device *dev, uint16_t sid, unsigned char source[]);
extern int (* fastpath_add_pppoe_header)(struct net_device *dev, uint16_t sid, unsigned char source[], struct sk_buff *skb);
#if SFE_ACCELERATE_IPSEC
extern void* (* fastpath_lookup_l2tp_xfrm)(struct net *net, u32 tunnel_id);
#endif

#define TPLINK_RESERVE_PPP_LEN 64 /* need change driver rx skb reserve length */
#define MAX_PPP_HEADER 1
#define MAX_ESP_HEADER 2
#define PPP_LCP_CONFIG_ACK 0x2
#define PPP_LCP_ECHO_REQUEST 0x9
#define PPP_LCP_ECHO_REPLY 0xa
#define MIN_SIZE_OF_LCP	8
#define PPP_LCP_MAGIC_NUMBER 0x5

struct tunnel_key {
	union {
		struct {
			uint32_t tunnel_id;
			uint32_t session_id;
		} l2tp;
		struct {
			uint16_t sid;
			uint8_t  *source;
		} pppoe;
		struct {
			uint16_t call_id;
			uint32_t saddr;
		} pptp;
	};
	struct net_device *tdev;
};

#define GETCHAR(c, cp) { \
	(c) = *(cp)++; \
}
#define PUTCHAR(c, cp) { \
	*(cp)++ = (uint8_t) (c); \
}

#define GETSHORT(s, cp) { \
	(s) = *(cp)++ << 8; \
	(s) |= *(cp)++; \
}
#define PUTSHORT(s, cp) { \
	*(cp)++ = (uint8_t) ((s) >> 8); \
	*(cp)++ = (uint8_t) (s); \
}

#define GETLONG(l, cp) { \
	(l) = *(cp)++ << 8; \
	(l) |= *(cp)++; (l) <<= 8; \
	(l) |= *(cp)++; (l) <<= 8; \
	(l) |= *(cp)++; \
}
#define PUTLONG(l, cp) { \
	*(cp)++ = (uint8_t) ((l) >> 24); \
	*(cp)++ = (uint8_t) ((l) >> 16); \
	*(cp)++ = (uint8_t) ((l) >> 8); \
	*(cp)++ = (uint8_t) (l); \
}

DEFINE_RWLOCK(tunnel_rwlock);

static struct sk_buff *sfe_ipv4_add_pppoe_header(struct tunnel_info *ti, struct sk_buff *skb, uint16_t protocol);
static struct sk_buff *sfe_ipv4_add_pptp_header(struct tunnel_info *ti, struct sk_buff *skb, uint16_t protocol);
static struct sk_buff *sfe_ipv4_add_l2tp_header(struct tunnel_info *ti, struct sk_buff *skb, uint16_t protocol);

#if SFE_ACCELERATE_IPSEC
static void update_tunnel_info_by_xfrm_state(struct xfrm_state *x)
{
	int i = 0;
	struct tunnel_info *ti = NULL;
	struct xfrm_state *conn_x = NULL;

	write_lock_bh(&tunnel_rwlock);
	for (i = 0; i < TUNNEL_MAX; i++) {
		ti = &TUNNEL_LIST[i];
		if (ti && ti->type == TUNNEL_L2TP && ti->info.l2tp.encap == 1) {
			if (ti->info.l2tp.x_in) {
				conn_x = ti->info.l2tp.x_in;
				if (x->props.family == conn_x->props.family &&
					x->props.reqid == conn_x->props.reqid &&
					x->id.proto == conn_x->id.proto &&
					x->props.mode == conn_x->props.mode &&
					xfrm_state_addr_check(x, &conn_x->id.daddr, &conn_x->props.saddr, conn_x->props.family)) {
					xfrm_state_put(conn_x);
					xfrm_state_hold(x);
					ti->info.l2tp.x_in = x;
				}
			}
			if (ti->info.l2tp.x_out) {
				conn_x = ti->info.l2tp.x_out;
				if (x->props.family == conn_x->props.family &&
					x->props.reqid == conn_x->props.reqid &&
					x->id.proto == conn_x->id.proto &&
					x->props.mode == conn_x->props.mode &&
					xfrm_state_addr_check(x, &conn_x->id.daddr, &conn_x->props.saddr, conn_x->props.family)) {
					xfrm_state_put(conn_x);
					xfrm_state_hold(x);
					ti->info.l2tp.x_out = x;
				}
			}
		}
	}
	write_unlock_bh(&tunnel_rwlock);
}

static void destroy_tunnel_info_by_xfrm_state(struct xfrm_state *x)
{
	int i = 0;
	struct tunnel_info *ti = NULL;
	write_lock_bh(&tunnel_rwlock);
	for (i = 0; i < TUNNEL_MAX; i++) {
		ti = &TUNNEL_LIST[i];
		if (ti && ti->type == TUNNEL_L2TP && ti->info.l2tp.encap == 1) {
			if (ti->info.l2tp.x_in == x) {
				xfrm_state_put(x);
				xfrm_state_put(ti->info.l2tp.x_out);
				ti->info.l2tp.x_in = NULL;
				ti->tunnel_dev = NULL;
				ti->ppp_dev = NULL;
				memset(ti, 0, sizeof(struct tunnel_info));
			}
			if (ti->info.l2tp.x_out == x) {
				xfrm_state_put(x);
				xfrm_state_put(ti->info.l2tp.x_out);
				ti->info.l2tp.x_out = NULL;
				ti->tunnel_dev = NULL;
				ti->ppp_dev = NULL;
				memset(ti, 0, sizeof(struct tunnel_info));
			}
		}
	}
	write_unlock_bh(&tunnel_rwlock);
}

static void destroy_l2tp_xfrm_tunnel_info(struct tunnel_info *ti)
{
	if (ti->type == TUNNEL_L2TP && ti->info.l2tp.encap == 1) {
		if (ti->info.l2tp.x_in) {
			xfrm_state_put(ti->info.l2tp.x_in);
			ti->info.l2tp.x_in = NULL;
		}
		if (ti->info.l2tp.x_out) {
			xfrm_state_put(ti->info.l2tp.x_out);
			ti->info.l2tp.x_out = NULL;
		}
	} 
}
#endif

static void destroy_tunnel_info(struct net_device *dev)
{
	int i = 0;
	write_lock_bh(&tunnel_rwlock);
	for (i = 0; i < TUNNEL_MAX; i++) {
		if (!dev) {
#if SFE_ACCELERATE_IPSEC
			destroy_l2tp_xfrm_tunnel_info(&TUNNEL_LIST[i]);
#endif
			TUNNEL_LIST[i].tunnel_dev = NULL;
			TUNNEL_LIST[i].ppp_dev = NULL;
			memset(&TUNNEL_LIST[i], 0, sizeof(struct tunnel_info));
			continue;
		}
		if (TUNNEL_LIST[i].tunnel_dev && (TUNNEL_LIST[i].tunnel_dev == dev)) {
#if SFE_ACCELERATE_IPSEC
			destroy_l2tp_xfrm_tunnel_info(&TUNNEL_LIST[i]);
#endif
			TUNNEL_LIST[i].tunnel_dev = NULL;
			TUNNEL_LIST[i].ppp_dev = NULL;
			memset(&TUNNEL_LIST[i], 0, sizeof(struct tunnel_info));
		}
		if (TUNNEL_LIST[i].ppp_dev && (TUNNEL_LIST[i].ppp_dev == dev)) {
#if SFE_ACCELERATE_IPSEC
			destroy_l2tp_xfrm_tunnel_info(&TUNNEL_LIST[i]);
#endif
			TUNNEL_LIST[i].tunnel_dev = NULL;
			TUNNEL_LIST[i].ppp_dev = NULL;
			memset(&TUNNEL_LIST[i], 0, sizeof(struct tunnel_info));
		}
	}
	write_unlock_bh(&tunnel_rwlock);
}

static struct tunnel_info *find_tunnel_info_origin(struct net_device *ppp_dev)
{
	int i = 0;

	for (i = 0; i < TUNNEL_MAX; i++) {
		if (TUNNEL_LIST[i].tunnel_dev && (TUNNEL_LIST[i].ppp_dev == ppp_dev)) {
			return &TUNNEL_LIST[i];
		}
	}
	return NULL;
}

static struct tunnel_info * find_tunnel_info_reply(enum tunnel_type type, struct tunnel_key *pkey)
{
	struct tunnel_info *ti = NULL;
	int i = 0;

	if (!pkey) {
		return NULL;
	}

	for (i = 0; i < TUNNEL_MAX; i++) {
		if (TUNNEL_LIST[i].tunnel_dev) {
			TPLINK_DEBUG("iterate a tunnel struct(%d)\n");
			ti = &(TUNNEL_LIST[i]);
			if (type != ti->type) {
				continue;
			}
			switch (type) {
				case TUNNEL_PPPOE:
					{
						TPLINK_DEBUG("get sid:%u,%02x:%02x:%02x:%02x:%02x:%02x, \n", pkey->pppoe.sid,
							pkey->pppoe.source[0],pkey->pppoe.source[1],pkey->pppoe.source[2],
							pkey->pppoe.source[3],pkey->pppoe.source[4],pkey->pppoe.source[5]);
						if ((ti->tunnel_dev == pkey->tdev) && (pkey->pppoe.sid == ti->info.pppoe.sid) && 
							(!memcmp(ti->info.pppoe.source, pkey->pppoe.source, ETH_ALEN))) {
							TPLINK_DEBUG("==>get session_id:%u,\n", pkey->pppoe.sid);
							goto out;
						}
					}
					break;
				case TUNNEL_PPTP:
					{
						TPLINK_DEBUG("get call_id:%u, saddr:%x\n", pkey->pptp.call_id, ntohl(pkey->pptp.saddr));
						if ((ti->tunnel_dev == pkey->tdev) && (pkey->pptp.call_id == ti->info.pptp.call_id) &&
						     (pkey->pptp.saddr == ti->info.pptp.saddr)) {
							TPLINK_DEBUG("==>get call_id:%u, saddr:%x\n", pkey->pptp.call_id, ntohl(pkey->pptp.saddr));
							goto out;
						}
					}
					break;
				case TUNNEL_L2TP:
					{
						uint32_t tunnel_id = pkey->l2tp.tunnel_id;
						uint32_t session_id = pkey->l2tp.session_id;
						uint32_t b_tunnel_id = ti->info.l2tp.tunnel_id;
						uint32_t b_session_id = ti->info.l2tp.session_id;

						TPLINK_DEBUG("get session_id:%u,%u, \n", tunnel_id, session_id);
						if ((ti->tunnel_dev == pkey->tdev) && (tunnel_id == b_tunnel_id) && (session_id == b_session_id)) {
							TPLINK_DEBUG("==>get session_id:%u,%u, \n", b_tunnel_id, b_session_id);
							goto out;
						}
					}
					break;
				default:
					break;
			}
		} 
	}
	ti = NULL;

out:
	return ti;
}

static struct tunnel_info * init_tunnel_info(enum tunnel_type type, struct tunnel_key *pkey, struct sk_buff *skb, int len)
{
	struct tunnel_info *ti = NULL;
	struct net_device *tdev = skb->dev;
	struct sfe_ipv4_eth_hdr      *eth = NULL;
	struct sfe_ipv4_ip_hdr	     *iph = NULL;
	struct sfe_ipv4_pptp_gre_hdr *ghdr = NULL;
	struct sfe_ipv4_udp_hdr *uhdr = NULL;
	struct sfe_ipv4_l2tpv2_hdr *lhdr   = NULL;
	struct sfe_ipv4_pppoe_hdr *poe_hdr = NULL;
	uint32_t tunnel_id = 0, session_id = 0;
	int ppp_ifindex = 0;
	struct net_device *ppp_dev = NULL;
	int i = 0;
#if SFE_ACCELERATE_IPSEC
	struct xfrm_state *x_in = NULL;
	struct xfrm_state *x_out = NULL;
	int encap = 0;
#endif

	if (!pkey) {
		return NULL;
	}

	eth = (struct sfe_ipv4_eth_hdr *)eth_hdr(skb);
	iph = (struct sfe_ipv4_ip_hdr *)ip_hdr(skb);
	poe_hdr = (struct sfe_ipv4_pppoe_hdr *)(skb->data);

	switch (type) {
		case TUNNEL_PPPOE:
			{
				typeof(fastpath_lookup_item) lookup_item = rcu_dereference(fastpath_lookup_item);
				uint16_t sid;

				sid = poe_hdr->sid;
				TPLINK_DEBUG("find sid:%u, source:%u\n", ntohs(sid), (unsigned char *)eth->h_source);
				if (!lookup_item || !(ppp_ifindex = lookup_item(tdev, sid, (unsigned char *)eth->h_source))) {
					TPLINK_DEBUG("no chain\n");
					//skb_dump(skb, 64, 8, "nochain: %s,%u,%pM",tdev->name, ntohs(sid), eth->h_source);
					ti = NULL;
					goto out;
				}
				pkey->pppoe.sid = sid;
				pkey->pppoe.source = (unsigned char *)eth->h_source;
				pkey->tdev = tdev;
			}
			break;
		case TUNNEL_PPTP:
			{
				typeof(fastpath_lookup_chan) lookup_chain = rcu_dereference(fastpath_lookup_chan);
				uint16_t call_id;

				ghdr = (struct sfe_ipv4_pptp_gre_hdr *)(skb->data + len);
				call_id = ntohs(ghdr->call_id);
				TPLINK_DEBUG("find call_id:%u, iph->saddr:%u\n", call_id, iph->saddr);
				if (!lookup_chain || !(ppp_ifindex = lookup_chain(call_id, iph->saddr))){
					TPLINK_DEBUG("no chain\n");
					ti = NULL;
					goto out;
				}
				pkey->pptp.call_id = call_id;
				pkey->pptp.saddr   = iph->saddr;
				pkey->tdev = tdev;
			}
			break;
		case TUNNEL_L2TP:
			{
				typeof(fastpath_lookup_session) lookup_session = rcu_dereference(fastpath_lookup_session);
#if SFE_ACCELERATE_IPSEC
				typeof(fastpath_lookup_l2tp_xfrm) lookup_xfrm = rcu_dereference(fastpath_lookup_l2tp_xfrm);
#endif
				unsigned char *ptr;
				uint16_t hdrflags;
				
				/* SKB & TUNNEL CHECK */
				uhdr = (struct sfe_ipv4_udp_hdr *)(skb->data + len);
				lhdr = (struct sfe_ipv4_l2tpv2_hdr *)(skb->data + len + sizeof(*uhdr));
				hdrflags = ntohs(lhdr->flags_ver) & (~L2TP_HDR_VER_MASK);
				ptr = (unsigned char *)&lhdr->payload_len;
				
				/* If length is present, skip it */
				if (hdrflags & L2TP_HDRFLAG_L)
					ptr += 2;
				
				/* Extract tunnel and session ID */
				tunnel_id = ntohs(*(__be16 *) ptr);
				ptr += 2;
				session_id = ntohs(*(__be16 *) ptr);

				/* get TUNNEL & SESSION */
				DEBUG_TRACE("find tunnel:%u, session:%u, hdrflags:%u\n", tunnel_id, session_id, hdrflags);
				if (!lookup_session || !(ppp_ifindex = lookup_session(dev_net(tdev), tunnel_id, session_id, hdrflags))) {
					DEBUG_TRACE("no chain\n");
					ti = NULL;
					goto out;
				}
				pkey->l2tp.tunnel_id  = tunnel_id;
				pkey->l2tp.session_id = session_id;
				pkey->tdev = tdev;
#if SFE_ACCELERATE_IPSEC
				if (lookup_xfrm)
					x_out = (struct xfrm_state*)lookup_xfrm(dev_net(tdev), tunnel_id);
				if (x_out) {
					xfrm_state_hold(x_out);
					DEBUG_TRACE("dir:out find xfrm_state(spi:0x%x) of l2tp tunnel\n",ntohl(x_out->id.spi));
					x_in = xfrm_state_lookup_byaddr(dev_net(skb->dev), x_out->mark.v, 
						(const xfrm_address_t *)&iph->daddr, (const xfrm_address_t *)&iph->saddr, IPPROTO_ESP, AF_INET);
					if (x_in) {
						DEBUG_TRACE("dir:in find xfrm_state(spi:0x%x) of l2tp tunnel\n",ntohl(x_in->id.spi));
						encap = 1;
					} else {
						xfrm_state_put(x_out);
						x_out = NULL;
					}
				} else {
					DEBUG_TRACE("not found xfrm_state, maybe not encap type\n");
				}
#endif
			}
			break;
		default:
			goto out;
	}

	read_lock_bh(&tunnel_rwlock);
	ti = find_tunnel_info_reply(type, pkey);
	read_unlock_bh(&tunnel_rwlock);
	if (ti) {
		goto out;
	}
	write_lock_bh(&tunnel_rwlock);
	/* find again, maybe someone create it! */
	ti = find_tunnel_info_reply(type, pkey);
	if (ti) {
#if SFE_ACCELERATE_IPSEC
		if (x_in)
			xfrm_state_put(x_in);
		if (x_out)
			xfrm_state_put(x_out);
#endif
		write_unlock_bh(&tunnel_rwlock);
		goto out;
	}
	ppp_dev = dev_get_by_index(dev_net(tdev), ppp_ifindex);
	if (!ppp_dev) {
#if SFE_ACCELERATE_IPSEC
		if (x_in)
			xfrm_state_put(x_in);
		if (x_out)
			xfrm_state_put(x_out);
#endif
		write_unlock_bh(&tunnel_rwlock);
		goto out;
	}
	for (i = 0; i < TUNNEL_MAX; i++) {
		if (!TUNNEL_LIST[i].tunnel_dev) {
			DEBUG_TRACE("find a empty tunnel struct(%d)\n", i);
			ti = &(TUNNEL_LIST[i]);
			ti->tunnel_dev = tdev;
			ti->ppp_dev = ppp_dev;
			ti->type	   = type;
			ti->src_mac[0] = eth->h_dest[0];
			ti->src_mac[1] = eth->h_dest[1];
			ti->src_mac[2] = eth->h_dest[2];
			ti->dest_mac[0] = eth->h_source[0];
			ti->dest_mac[1] = eth->h_source[1];
			ti->dest_mac[2] = eth->h_source[2];
			ti->magic_number = 0;
			ti->debug_read_seq = 0;
			switch (type) {
				case TUNNEL_PPPOE:
					if (poe_hdr && eth){
						TPLINK_DEBUG("create sid:%u\n", poe_hdr->sid);
						ti->info.pppoe.sid = poe_hdr->sid;
						memcpy(ti->info.pppoe.source, (unsigned char *)eth->h_source, ETH_ALEN);
						ti->add_header = sfe_ipv4_add_pppoe_header;
					}
					break;
				case TUNNEL_PPTP:
					if(ghdr && iph){
						ti->saddr	   = iph->daddr;
						ti->daddr	   = iph->saddr;
						TPLINK_DEBUG("create call_id:%u, iph->saddr:%u\n", ntohs(ghdr->call_id), iph->saddr);
						ti->info.pptp.call_id = ntohs(ghdr->call_id);
						ti->info.pptp.saddr   = iph->saddr;
						ti->add_header = sfe_ipv4_add_pptp_header;
					}
					break;
				case TUNNEL_L2TP:
					if(lhdr && uhdr){
						ti->saddr	   = iph->daddr;
						ti->daddr	   = iph->saddr;
						DEBUG_TRACE("create tunnel_id:%u, session_id:%u\n", tunnel_id, session_id);
						ti->info.l2tp.tunnel_id = tunnel_id;
						ti->info.l2tp.session_id = session_id;
						ti->info.l2tp.sport = uhdr->dest;
						ti->info.l2tp.dport = uhdr->source;
						ti->add_header = sfe_ipv4_add_l2tp_header;
#if SFE_ACCELERATE_IPSEC
						ti->info.l2tp.encap = encap;
						ti->info.l2tp.x_in = encap ? x_in : NULL;
						ti->info.l2tp.x_out = encap ? x_out : NULL;
#endif
					}
					break;
				default:
					break;
			}
			break;
		}
	}
	dev_put(ppp_dev);
	write_unlock_bh(&tunnel_rwlock);

out:
	return ti;
}

static void update_tunnel_info(enum tunnel_type type, struct tunnel_key *pkey, unsigned char *phdr, unsigned char *uphdr)
{
	struct tunnel_info *ti = NULL;

	if (!pkey) {
		return;
	}

	read_lock_bh(&tunnel_rwlock);
	ti = find_tunnel_info_reply(type, pkey);
	if (!ti || !ti->tunnel_dev) {
		read_unlock_bh(&tunnel_rwlock);
		return;
	}

	switch (type) {
		case TUNNEL_PPPOE:
			break;
		case TUNNEL_PPTP:
			{
				typeof(fastpath_update_chan) update_chan = rcu_dereference(fastpath_update_chan);
				struct sfe_ipv4_pptp_gre_hdr *ghdr = (struct sfe_ipv4_pptp_gre_hdr *)phdr;
				struct sfe_ipv4_ip_hdr *iphdr = (struct sfe_ipv4_ip_hdr *)uphdr;
	
				if (update_chan) {
					update_chan(ntohs(ghdr->call_id), iphdr->saddr, phdr);
				}
			}
			break;
		case TUNNEL_L2TP:
			{
				typeof(fastpath_update_session) update_session = rcu_dereference(fastpath_update_session);
				struct sfe_ipv4_l2tpv2_hdr *lhdr = (struct sfe_ipv4_l2tpv2_hdr *)phdr;
				uint32_t tunnel_id, session_id;
				uint16_t hdrflags;
				
				tunnel_id = ti->info.l2tp.tunnel_id;
				session_id = ti->info.l2tp.session_id;
				hdrflags = ntohs(lhdr->flags_ver) & (~L2TP_HDR_VER_MASK);
				if (update_session) {
					update_session(dev_net(ti->tunnel_dev), tunnel_id, session_id, hdrflags);
				}
			}
			break;
	}
	read_unlock_bh(&tunnel_rwlock);
}

static int sfe_ipv4_dev_queue_xmit_protocol(struct sk_buff *skb, uint16_t protocol)
{
	int ret = 0;
	struct net_device *odev = NULL;
	int ppp_header_level = 0;

#if SFE_ACCELERATE_IPSEC
    struct iphdr *iph = ip_hdr(skb);
    struct xfrm_state *x = NULL;
#endif

restart:
	if ((!skb_is_gso(skb)) && skb->dev && (skb->dev->flags & IFF_POINTOPOINT)) {
		struct tunnel_info *ti = NULL;

		if ((skb->ip_summed == CHECKSUM_PARTIAL) && (skb_checksum_help(skb) < 0)) {
			goto out;
		}

		odev = skb->dev;
		read_lock_bh(&tunnel_rwlock);
		ti = find_tunnel_info_origin(skb->dev);
		if (ti) {
			if(!dev_tuntype(skb->dev))
			{	
				ppp_dev_add_stats(skb->dev, skb, skb->len, 0);
			}
			TPLINK_DEBUG("=================add tunnel header for skb(%p)\n", skb);
			skb = ti->add_header(ti, skb, protocol);
			if(!skb)
			{
				read_unlock_bh(&tunnel_rwlock);
				return 0;
			}
#if SFE_ACCELERATE_IPSEC
			if (ti->type == TUNNEL_L2TP && ti->info.l2tp.encap == 1) {
				x = ti->info.l2tp.x_out;
				if ((x->data != NULL) && (x->km.state == XFRM_STATE_VALID)) {
                    __skb_pull(skb, ETH_HLEN);
                    skb_reset_mac_header(skb);
                    sfe_ipv4_encap_esp(x, skb, iph);
                    read_unlock_bh(&tunnel_rwlock);
                    return 0;
                }
			}
#endif
#if 0 //#if SFE_ACCELERATE_IPSEC process it later 
            if (NULL != (x = xfrm_state_lookup_byskb(skb, 0)))
            {
                if ((x->data != NULL) && (x->km.state == XFRM_STATE_VALID)) 
                {
                    __skb_pull(skb, ETH_HLEN);
                    skb_reset_mac_header(skb);
                    sfe_ipv4_encap_esp(x, skb, iph);
                    xfrm_state_put(x);
                    read_unlock_bh(&tunnel_rwlock);
                    return 0;
                }
                else
                {
                    xfrm_state_put(x);
                }
            }
#endif
        }
		read_unlock_bh(&tunnel_rwlock);

		if ((odev != skb->dev) && ((++ppp_header_level) < MAX_PPP_HEADER)) {
			goto restart;
		}
	}

out:
	TPLINK_DEBUG("=================send skb(%p)\n", skb);
#ifdef TPLINK_ACCELERATE_NAT
	if (skb->shortcut_data) {
		sfe_hard_xmit_func_t xmit = (sfe_hard_xmit_func_t)skb->shortcut_data;
		ret = xmit(skb);
		skb->shortcut_data = 0;
	}
	else {
		ret = dev_queue_xmit(skb);
	}
#else
	ret = dev_queue_xmit(skb);
#endif
	return ret;
}

static int sfe_ipv4_dev_queue_xmit(struct sk_buff *skb)
{
	return sfe_ipv4_dev_queue_xmit_protocol(skb, htons(PPP_IP));
}

static int sfe_ipv4_send_ppp_echoreply(struct sk_buff *skb, struct net_device *output, uint32_t magic)
{
	uint8_t *lcp_header = (uint8_t *)(skb->data);

	PUTCHAR(PPP_LCP_ECHO_REPLY, lcp_header);	/* code */
	lcp_header++; /* skip id */
	PUTSHORT(MIN_SIZE_OF_LCP, lcp_header); /* length */
	PUTLONG(magic, lcp_header); /* magic */
	skb->dev = output;
	skb->len = MIN_SIZE_OF_LCP;
	sfe_ipv4_dev_queue_xmit_protocol(skb, htons(PPP_LCP));
	return 1;
}

static int sfe_ipv4_recv_ppp_lcpack(struct sk_buff *skb, int pull_len, uint32_t *magic)
{
	uint8_t *lcp_header = (uint8_t *)(skb->data + pull_len);
	uint16_t length = 0;
	uint8_t dtype = 0, dlength = 0;

	lcp_header++;	/* skip code */
	lcp_header++;	/* skip identify */
	GETSHORT(length, lcp_header);	/* length */

	if (!pskb_may_pull(skb, pull_len + length)) {
		return 0;
	}

	length -= 4;	/* data length */
	while (length >= 2) {
		GETCHAR(dtype, lcp_header);	 /* option type */
		GETCHAR(dlength, lcp_header);/* option len */

		if (dtype != PPP_LCP_MAGIC_NUMBER) {
			length -= dlength;
			continue;
		}
	
		if (length >= dlength) {
			GETLONG(*magic, lcp_header);
			return 1;
		}
		length -= dlength;
	}
	return 0;
}

static int sfe_ipv4_recv_pptp(struct sfe_ipv4 *si, struct sk_buff *skb, struct net_device *dev,
			     unsigned int len, struct sfe_ipv4_ip_hdr *iph, unsigned int ihl, bool flush_on_find, int ppp_header_level)
{
	struct sfe_ipv4_pptp_gre_hdr *ghdr = NULL;
	struct sfe_ipv4_ppp_hdr *phdr = NULL;
	struct tunnel_key tkey;
	int ret = 0;
	int pull_len = 0;
	if (!pskb_may_pull(skb, ihl + sizeof(struct sfe_ipv4_pptp_gre_hdr) + sizeof(struct sfe_ipv4_ppp_hdr))) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_HEADER_INCOMPLETE]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("len: gre is too short for header of size: 12\n");
		return 0;
	}
	
	ghdr = (struct sfe_ipv4_pptp_gre_hdr *)(skb->data + ihl);
	if (ntohs(ghdr->protocol) != PPTP_GRE_PROTO || /* PPTP-GRE protocol for PPTP */
		PPTP_GRE_IS_C(ghdr->flags) || 	       /* flag C should be clear */
		PPTP_GRE_IS_R(ghdr->flags) || 	       /* flag R should be clear */
		!PPTP_GRE_IS_K(ghdr->flags) ||	       /* flag K should be set */
		!PPTP_GRE_IS_S(ghdr->flags) ||
		(ghdr->flags&0xF) != 0) {		       /* routing and recursion ctrl = 0 */
		/* if invalid, discard this packet */
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_PPTP_HEADER_INVALID]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("gre header is invalid\n");
		return 0;
	}
	if (PPTP_GRE_IS_A(ghdr->ver)) {
		phdr = (struct sfe_ipv4_ppp_hdr *)(skb->data + ihl + sizeof(struct sfe_ipv4_pptp_gre_hdr));
		pull_len = ihl + sizeof(struct sfe_ipv4_pptp_gre_hdr) + sizeof(struct sfe_ipv4_ppp_hdr);
	} else {
		phdr = (struct sfe_ipv4_ppp_hdr *)(skb->data + ihl + sizeof(struct sfe_ipv4_pptp_gre_hdr) - sizeof(ghdr->ack));
		pull_len = ihl + sizeof(struct sfe_ipv4_pptp_gre_hdr) + sizeof(struct sfe_ipv4_ppp_hdr) - sizeof(ghdr->ack);
	}

	memset(&tkey, 0, sizeof(tkey));
	if (ntohs(phdr->protocol) == PPP_IP) {
		struct tunnel_info *ti = NULL;
		struct sfe_ipv4_pptp_gre_hdr bhdr = {0};
		struct sfe_ipv4_ip_hdr uphdr = {0};
		struct net_device *odev = skb->dev;
		
		ti = init_tunnel_info(TUNNEL_PPTP, &tkey, skb, (iph->ihl << 2));
		if (ti) {
			skb_pull_rcsum(skb, pull_len);
			skb->protocol = htons(ETH_P_IP);
			skb->dev = ti->ppp_dev;
			skb_reset_network_header(skb);
			skb_reset_mac_len(skb);
			memcpy(&bhdr, ghdr, sizeof(bhdr));
			memcpy(&uphdr, iph, sizeof(uphdr));
			ret = sfe_ipv4_recv_core(NULL, skb, ++ppp_header_level, SFE_PROTO_PPTP);
			if (!ret) {
				DEBUG_TRACE("fail to recv a tunnel pkt\n");
				__skb_push(skb, pull_len);
				skb->protocol = htons(ETH_P_IP);
				skb->dev = odev;
				skb_reset_network_header(skb);
				skb_reset_mac_len(skb);
				return 0;
			}
			DEBUG_TRACE("succ to recv a tunnel pkt & update tunnel info: %u\n", pull_len);
			update_tunnel_info(TUNNEL_PPTP, &tkey, (unsigned char *)&bhdr, (unsigned char *)&uphdr);
			return ret;
		}
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_PPTP_NO_TUNNEL]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);
	} else if (ntohs(phdr->protocol) == PPP_LCP) {
		uint8_t lcp_code = (*(uint8_t *)(skb->data + pull_len));
		if (pskb_may_pull(skb, pull_len + MIN_SIZE_OF_LCP) && (lcp_code == PPP_LCP_ECHO_REQUEST)) {	
			struct tunnel_info *ti = NULL;
			struct net_device *ppp_dev = NULL;
			uint32_t magic_number = 0;

			ti = init_tunnel_info(TUNNEL_PPTP, &tkey, skb, (iph->ihl << 2));
			if (ti) {
				read_lock_bh(&tunnel_rwlock);
				ti = find_tunnel_info_reply(TUNNEL_PPTP, &tkey);
				if (ti && ti->ppp_dev && ti->magic_number) {
					ppp_dev = ti->ppp_dev;
					magic_number = ti->magic_number;
					read_unlock_bh(&tunnel_rwlock);

					skb_pull_rcsum(skb, pull_len);
					ret = sfe_ipv4_send_ppp_echoreply(skb, ppp_dev, magic_number);
					return ret;
				}
				read_unlock_bh(&tunnel_rwlock);
			}
		} else if (pskb_may_pull(skb, pull_len + MIN_SIZE_OF_LCP) && (lcp_code == PPP_LCP_CONFIG_ACK)) {
			uint32_t magic_number = 0;

			ret = sfe_ipv4_recv_ppp_lcpack(skb, pull_len, &magic_number);
			if (ret) {
				DEBUG_TRACE("MAGIC_NUMBER:%x\n", magic_number);
			}
			return 0;
		}
	} else {
		DEBUG_TRACE("protocol:%u, ver:%u, pull:%u\n", ntohs(phdr->protocol), ghdr->ver, pull_len);
		//skb_dump(skb, __FUNCTION__, __LINE__);
	}
	spin_lock_bh(&si->lock);
	si->exception_events[SFE_IPV4_EXCEPTION_EVENT_PPTP_GRE_UNHANDLED]++;
	si->packets_not_forwarded++;
	spin_unlock_bh(&si->lock);
	DEBUG_TRACE("pptp gre unhandled!\n");

	return 0;
}

static struct sk_buff *sfe_ipv4_add_pptp_header(struct tunnel_info *ti, struct sk_buff *skb, uint16_t protocol)
{
	typeof(fastpath_add_pptp_header) add_header = rcu_dereference(fastpath_add_pptp_header);
	struct sfe_ipv4_eth_hdr      *eth = NULL;
    struct sfe_ipv4_ip_hdr	     *iph = NULL;
    struct sfe_ipv4_pptp_gre_hdr *greh = NULL;	
    struct sfe_ipv4_ppp_hdr	     *ppph = NULL;
    int32_t		   header_len = 0;
	int max_headroom = 0;
	int ret = 0;
	if (!add_header) {
		return skb;
	}

	max_headroom = LL_RESERVED_SPACE(ti->tunnel_dev) + sizeof(*eth) + sizeof(*iph) + sizeof(*greh) + sizeof(*ppph);
	if (skb_headroom(skb) < max_headroom) {
		DEBUG_WARN("====skb_headroom(skb): %u, max_headroom:%u\n", skb_headroom(skb), max_headroom);
		if(skb_cow_head(skb, max_headroom))
		{
			DEBUG_ERROR("skb_cow_head error ,free skb, max_headroom:%u skb->len %d skb_headroom(skb) %d\n", max_headroom,skb->len,skb_headroom(skb));
			kfree_skb(skb);
			return NULL;
		}
	}
    /* add PPP */
	header_len = sizeof(*ppph);
	skb_push(skb, header_len);
    ppph = (struct sfe_ipv4_ppp_hdr *)(skb->data);
    ppph->addr_ctrl = htons(0xFF03);
    ppph->protocol = protocol;

    /* add gre header */
    ret = add_header(ti->info.pptp.call_id, ti->info.pptp.saddr, skb);
    if (!ret) {
    	skb_pull(skb, sizeof(*ppph));
    	return skb;
    }

    /* add IP header */
	skb_reset_transport_header(skb);
	skb_push(skb, sizeof(*iph));
	skb_reset_network_header(skb);
    iph = (struct sfe_ipv4_ip_hdr*)(skb->data);    
    memset(iph, 0, sizeof(struct sfe_ipv4_ip_hdr));
    iph->version	= 4;
    iph->ihl		= 5;
    iph->ttl		= 64;
	iph->frag_off	= 0;
    iph->protocol	= IPPROTO_GRE;
	iph->tos        = 0;
    iph->saddr		= ti->saddr;
    iph->daddr		= ti->daddr;
    /* TODO: To enable checksum offload in the furure */		
    iph->tot_len	= htons(skb->len);

	skb_dst_drop(skb);

	//nf_reset(skb);

	skb->ip_summed	=	CHECKSUM_NONE;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 0, 0))
	ip_select_ident(dev_net(ti->tunnel_dev), skb, NULL);
#elif (LINUX_VERSION_CODE > KERNEL_VERSION(3, 10, 49))
    ip_select_ident(skb, NULL);
#else
	{
		struct dst_entry fake_dst;
		fake_dst.dev = ti->tunnel_dev;
		ip_select_ident(skb, &fake_dst, NULL);
	}
#endif
    ip_send_check((struct iphdr *)iph);

    /* add eth hdr */
	skb_push(skb, sizeof(*eth));
	eth =  (struct sfe_ipv4_eth_hdr*)(skb->data);
	eth->h_proto = htons(ETH_P_IP);;
	eth->h_dest[0] = ti->dest_mac[0];
	eth->h_dest[1] = ti->dest_mac[1];
	eth->h_dest[2] = ti->dest_mac[2];
	eth->h_source[0] = ti->src_mac[0];
	eth->h_source[1] = ti->src_mac[1];
	eth->h_source[2] = ti->src_mac[2];
	skb_reset_mac_header(skb);

	skb->dev = ti->tunnel_dev;
    return skb;
}

static int sfe_ipv4_recv_l2tp(struct sfe_ipv4 *si, struct sk_buff *skb, struct net_device *dev,
			     unsigned int len, struct sfe_ipv4_ip_hdr *iph, unsigned int ihl, bool flush_on_find, int ppp_header_level)
{
	struct sfe_ipv4_udp_hdr *uhdr = NULL;
	struct sfe_ipv4_ppp_hdr *phdr = NULL;
	struct sfe_ipv4_l2tpv2_hdr *lhdr = NULL;
	struct tunnel_key tkey;
	uint16_t hdrflags = 0;
	uint16_t version = 0;

	int length = 0;
	int ret = 0;
	int pull_len = 0;
	
	if (!iph || (ntohs(iph->frag_off) & IP_OFFSET)) {
		DEBUG_TRACE("non-first fragment!\n");
		return 0;
	}

	if (!pskb_may_pull(skb, ihl + sizeof(*uhdr))) {
		DEBUG_TRACE("len: udp packet is too short for header of l2tp tunnel,%s, %u, %u, %u, %u\n", skb->dev->name, skb_headroom(skb), skb->len, skb->data_len, ihl + sizeof(*uhdr) + sizeof(*lhdr) + sizeof(*phdr));
		return 0;
	}

	uhdr = (struct sfe_ipv4_udp_hdr *)(skb->data + ihl);
	if ((uhdr->dest != htons(1701)) && (uhdr->source != htons(1701))) {
		DEBUG_TRACE("not l2tp udp packet, (%s, %u,%u)\n", skb->dev->name, ntohs(uhdr->source), ntohs(uhdr->dest));
		return 0;
	}

	if (!pskb_may_pull(skb, ihl + sizeof(*uhdr) + sizeof(*lhdr) + sizeof(*phdr))) {
		DEBUG_TRACE("len: udp packet is too short for header of l2tp tunnel, %u, %u, %u, %u\n", skb_headroom(skb), skb->len, skb->data_len, ihl + sizeof(*uhdr) + sizeof(*lhdr) + sizeof(*phdr));
		return 0;
	}

	lhdr = (struct sfe_ipv4_l2tpv2_hdr *)(skb->data + ihl + sizeof(*uhdr));
	hdrflags = ntohs(lhdr->flags_ver) & (~L2TP_HDR_VER_MASK);
	version = ntohs(lhdr->flags_ver) & L2TP_HDR_VER_MASK;
	if ((version != L2TP_HDR_VER_2) ||
		(hdrflags & L2TP_HDRFLAG_T)){
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_L2TP_HEADER_INVALID]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);
		
		DEBUG_TRACE("l2tp header is invalid: [%pI4:%d===>%pI4:%d] ver:%x hdrflags:%x\n", 
				&iph->saddr, ntohs(uhdr->source), &iph->daddr, ntohs(uhdr->dest), version, hdrflags);
		return 0;
	}

	length = sizeof(*lhdr);
	if (!(hdrflags & L2TP_HDRFLAG_L)) {
		length -=2;
	}
	if (!(hdrflags & L2TP_HDRFLAG_S)) {
		length -= 4;
	}
	if (!(hdrflags & L2TP_HDRFLAG_O)) {
		length -= 4;
	}
	phdr = (struct sfe_ipv4_ppp_hdr *)(skb->data + ihl + sizeof(*uhdr) + length);
	pull_len = ihl + sizeof(*uhdr) + length + sizeof(struct sfe_ipv4_ppp_hdr);

	memset(&tkey, 0, sizeof(tkey));
	if (ntohs(phdr->protocol) ==  PPP_IP) {
		struct tunnel_info *ti = NULL;
		struct sfe_ipv4_l2tpv2_hdr bhdr = {0};
		struct net_device *odev = skb->dev;

		DEBUG_TRACE("recv ppp ip pkt (%u)!\n", pull_len);	
		ti = init_tunnel_info(TUNNEL_L2TP, &tkey, skb, (iph->ihl << 2));
		if (ti) {
			DEBUG_TRACE("get a tunnel info\n");
			skb_pull_rcsum(skb, pull_len);
			skb->protocol = htons(ETH_P_IP);
			skb->dev = ti->ppp_dev;
			skb_reset_network_header(skb);
			skb_reset_mac_len(skb);
			memcpy(&bhdr, lhdr, sizeof(*lhdr));
			ret = sfe_ipv4_recv_core(NULL, skb, ++ppp_header_level, SFE_PROTO_L2TP);
			if (!ret) {
				DEBUG_TRACE("fail to recv a tunnel pkt\n");
				__skb_push(skb, pull_len);
				skb->dev = odev;
				skb->protocol = htons(ETH_P_IP);
				skb_reset_network_header(skb);
				skb_reset_mac_len(skb);
				return 0;
			}
			DEBUG_TRACE("succ to recv a tunnel pkt & update tunnel info: %u\n", pull_len);
			update_tunnel_info(TUNNEL_L2TP, &tkey, (unsigned char *)&bhdr, NULL);
			return ret;
		}
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_L2TP_NO_TUNNEL]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);
	} else if (ntohs(phdr->protocol) == PPP_LCP) {
		uint8_t lcp_code = (*(uint8_t *)(skb->data + pull_len));
		if (pskb_may_pull(skb, pull_len + MIN_SIZE_OF_LCP) && (lcp_code == PPP_LCP_ECHO_REQUEST)) {	
			struct tunnel_info *ti = NULL;
			struct net_device *ppp_dev = NULL;
			uint32_t magic_number = 0;

			ti = init_tunnel_info(TUNNEL_L2TP, &tkey, skb, (iph->ihl << 2));
			if (ti) {
				read_lock_bh(&tunnel_rwlock);
				ti = find_tunnel_info_reply(TUNNEL_L2TP, &tkey);
				if (ti && ti->ppp_dev && ti->magic_number) {
					ppp_dev = ti->ppp_dev;
					magic_number = ti->magic_number;
					read_unlock_bh(&tunnel_rwlock);

					skb_pull_rcsum(skb, pull_len);
					ret = sfe_ipv4_send_ppp_echoreply(skb, ppp_dev, magic_number);
					return ret;
				}
				read_unlock_bh(&tunnel_rwlock);
			}
		} else if (pskb_may_pull(skb, pull_len + MIN_SIZE_OF_LCP) && (lcp_code == PPP_LCP_CONFIG_ACK)) {
			uint32_t magic_number = 0;

			ret = sfe_ipv4_recv_ppp_lcpack(skb, pull_len, &magic_number);
			if (ret) {
				DEBUG_TRACE("MAGIC_NUMBER:%x\n", magic_number);
			}
			return 0;
		}
	} else {
		DEBUG_TRACE("protocol:%u, ver:%u, pull:%u\n", ntohs(phdr->protocol), version, pull_len);
		//skb_dump(skb, __FUNCTION__, __LINE__);
	}
	spin_lock_bh(&si->lock);
	si->exception_events[SFE_IPV4_EXCEPTION_EVENT_L2TP_TUNNEL_UNHANDLED]++;
	si->packets_not_forwarded++;
	spin_unlock_bh(&si->lock);
	DEBUG_TRACE("l2tp unhandled!\n");

	return 0;
}

static struct sk_buff *sfe_ipv4_add_l2tp_header(struct tunnel_info *ti, struct sk_buff *skb, uint16_t protocol)
{
	typeof(fastpath_add_l2tp_header) add_header = rcu_dereference(fastpath_add_l2tp_header);
	struct sfe_ipv4_eth_hdr      *eth = NULL;
    struct sfe_ipv4_ip_hdr	     *iph = NULL;
    struct sfe_ipv4_udp_hdr      *udph = NULL;	
    struct sfe_ipv4_ppp_hdr	     *ppph = NULL;
	int max_headroom = 0;
	int ret = 0;

	if (!add_header) {
		return skb;
	}

	max_headroom = LL_RESERVED_SPACE(ti->tunnel_dev) + sizeof(*eth) + sizeof(*iph) + sizeof(*udph) + sizeof(struct sfe_ipv4_l2tpv2_hdr) + sizeof(*ppph);
	if (skb_headroom(skb) < max_headroom) {
		DEBUG_WARN("====skb_headroom(skb): %u, max_headroom:%u\n", skb_headroom(skb), max_headroom);
		if(skb_cow_head(skb, max_headroom))
		{
			DEBUG_ERROR("skb_cow_head error ,free skb, max_headroom:%u skb->len %d skb_headroom(skb) %d\n", max_headroom,skb->len,skb_headroom(skb));
			kfree_skb(skb);
			return NULL;
		}
	}

    /* add PPP */
	skb_push(skb, sizeof(*ppph));
    ppph = (struct sfe_ipv4_ppp_hdr *)(skb->data);
    ppph->addr_ctrl = htons(0xFF03);
    ppph->protocol = protocol;

    /* add l2tp header */
    ret = add_header(dev_net(ti->tunnel_dev), ti->info.l2tp.tunnel_id,  ti->info.l2tp.session_id, skb);
    if (!ret) {
    	skb_pull(skb, sizeof(*ppph));
    	return skb;
    }

	/* add udp header */
	skb_push(skb, sizeof(*udph));
	skb_reset_transport_header(skb);
	udph = (struct sfe_ipv4_udp_hdr *)(skb->data);
	udph->source = ti->info.l2tp.sport;
	udph->dest = ti->info.l2tp.dport;
	udph->len = htons(skb->len);
	udph->check = 0;

    /* add IP header */
	skb_push(skb, sizeof(*iph));
	skb_reset_network_header(skb);
    iph = (struct sfe_ipv4_ip_hdr*)(skb->data);    
    memset(iph, 0, sizeof(struct sfe_ipv4_ip_hdr));
    iph->version	= 4;
    iph->ihl		= 5;
    iph->ttl		= 64;
	iph->frag_off	= 0;
    iph->protocol	= IPPROTO_UDP;
	iph->tos        = 0;
    iph->saddr		= ti->saddr;
    iph->daddr		= ti->daddr;
    /* TODO: To enable checksum offload in the furure */		
    iph->tot_len	= htons(skb->len);

	skb_dst_drop(skb);

	//nf_reset(skb);

	skb->ip_summed	=	CHECKSUM_NONE;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 0, 0))
		ip_select_ident(dev_net(ti->tunnel_dev), skb, NULL);
#elif (LINUX_VERSION_CODE > KERNEL_VERSION(3, 10, 49))
		ip_select_ident(skb, NULL);
#else
		{
			struct dst_entry fake_dst;
			fake_dst.dev = ti->tunnel_dev;
			ip_select_ident(skb, &fake_dst, NULL);
		}
#endif
    ip_send_check((struct iphdr *)iph);

    /* add eth hdr */
	skb_push(skb, sizeof(*eth));
	eth =  (struct sfe_ipv4_eth_hdr*)(skb->data);
	eth->h_proto = htons(ETH_P_IP);;
	eth->h_dest[0] = ti->dest_mac[0];
	eth->h_dest[1] = ti->dest_mac[1];
	eth->h_dest[2] = ti->dest_mac[2];
	eth->h_source[0] = ti->src_mac[0];
	eth->h_source[1] = ti->src_mac[1];
	eth->h_source[2] = ti->src_mac[2];
	skb_reset_mac_header(skb);

	skb->dev = ti->tunnel_dev;

    return skb;
}

int sfe_ipv4_recv_pppoe(struct net_device *dev, struct sk_buff *skb)
{
	struct sfe_ipv4 *si = &__si;
	struct sfe_ipv4_pppoe_hdr *poe_hdr = NULL;
	struct sfe_ipv4_pppoe_ppp_hdr *ppp_hdr = NULL;
	struct tunnel_key tkey;
	int pull_len = 0;
	int ret = 0;

	/*
	 * Check that we have space for an PPPoE header here.
	 */
	pull_len = sizeof(struct sfe_ipv4_pppoe_hdr) + sizeof(struct sfe_ipv4_pppoe_ppp_hdr);
	if (unlikely(!pskb_may_pull(skb, pull_len))) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_HEADER_INCOMPLETE]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("len: %u is too short\n", pull_len);
		return 0;
	}

	poe_hdr = (struct sfe_ipv4_pppoe_hdr *)skb->data;
	if ((poe_hdr->ver != PPPOE_VERSION) ||
	     (poe_hdr->type != PPPOE_TYPE) ||
	     (poe_hdr->code != PPPOE_SESS_CODE) ||
	     (ntohs(poe_hdr->sid) == 0XFFFF)) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_PPPOE_HEADER_INVALID]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);
		
		DEBUG_TRACE("pppoe header is invalid\n");
		return 0;
	}
	ppp_hdr = (struct sfe_ipv4_pppoe_ppp_hdr *)(skb->data + sizeof(*poe_hdr));

	memset(&tkey, 0, sizeof(tkey));
	if (ntohs(ppp_hdr->protocol) ==  PPP_IP) {
		struct tunnel_info *ti = NULL;
		struct net_device *odev = skb->dev;

		TPLINK_DEBUG("recv ppp ip pkt (%u)!\n", pull_len);	
		ti = init_tunnel_info(TUNNEL_PPPOE, &tkey, skb, 0);
		if (ti) {
			TPLINK_DEBUG("get a tunnel info\n");
			skb_pull_rcsum(skb, pull_len);
			skb->protocol = htons(ETH_P_IP);
			skb_reset_network_header(skb);
			skb_reset_mac_len(skb);
			skb->dev = ti->ppp_dev;
			ret = sfe_ipv4_recv_core(NULL, skb, 1, SFE_PROTO_PPPOE);
			if (!ret) {
				TPLINK_DEBUG("fail to recv a tunnel pkt\n");
				__skb_push(skb, pull_len);
				skb->dev = odev;
				skb->protocol = htons(ETH_P_PPP_SES);
				skb_reset_network_header(skb);
				skb_reset_mac_len(skb);
				return 0;
			}
			TPLINK_DEBUG("succ to recv a tunnel pkt & update tunnel info: %u\n", pull_len);
			return ret;
		}
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_PPPOE_NO_TUNNEL]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);
	} else if (ntohs(ppp_hdr->protocol) == PPP_LCP) {
		uint8_t lcp_code = (*(uint8_t *)(skb->data + pull_len));
		if (pskb_may_pull(skb, pull_len + MIN_SIZE_OF_LCP) && (lcp_code == PPP_LCP_ECHO_REQUEST)) {	
			struct tunnel_info *ti = NULL;
			struct net_device *ppp_dev = NULL;
			uint32_t magic_number = 0;

			ti = init_tunnel_info(TUNNEL_PPPOE, &tkey, skb, 0);
			if (ti) {
				read_lock_bh(&tunnel_rwlock);
				ti = find_tunnel_info_reply(TUNNEL_PPPOE, &tkey);
				if (ti && ti->ppp_dev && ti->magic_number) {
					ppp_dev = ti->ppp_dev;
					magic_number = ti->magic_number;
					read_unlock_bh(&tunnel_rwlock);

					skb_pull_rcsum(skb, pull_len);
					ret = sfe_ipv4_send_ppp_echoreply(skb, ppp_dev, magic_number);
					return ret;
				}
				read_unlock_bh(&tunnel_rwlock);
			}
		} else if (pskb_may_pull(skb, pull_len + MIN_SIZE_OF_LCP) && (lcp_code == PPP_LCP_CONFIG_ACK)) {
			uint32_t magic_number = 0;

			ret = sfe_ipv4_recv_ppp_lcpack(skb, pull_len, &magic_number);
			if (ret) {
				TPLINK_DEBUG("MAGIC_NUMBER:%x\n", magic_number);
			}
			return 0;
		}
	} else {
		TPLINK_DEBUG("protocol:%u, sid:%u, pull:%u\n", ntohs(ppp_hdr->protocol), ntohs(poe_hdr->sid), pull_len);
		//skb_dump(skb, __FUNCTION__, __LINE__);
	}
	spin_lock_bh(&si->lock);
	si->exception_events[SFE_IPV4_EXCEPTION_EVENT_PPPOE_TUNNEL_UNHANDLED]++;
	si->packets_not_forwarded++;
	spin_unlock_bh(&si->lock);
	DEBUG_TRACE("pppoe unhandled!\n");

	return 0;
}

static struct sk_buff *sfe_ipv4_add_pppoe_header(struct tunnel_info *ti, struct sk_buff *skb, uint16_t protocol)
{
	typeof(fastpath_add_pppoe_header) add_header = rcu_dereference(fastpath_add_pppoe_header);
	struct sfe_ipv4_eth_hdr 	 *eth = NULL;
	struct sfe_ipv4_pppoe_hdr		 *poeh = NULL;
	struct sfe_ipv4_pppoe_ppp_hdr 	 *ppph = NULL;
	int max_headroom = 0;
	int ret = 0;

	if (!add_header) {
		return skb;
	}

	max_headroom = LL_RESERVED_SPACE(ti->tunnel_dev) + sizeof(*eth) + sizeof(*poeh) + sizeof(*ppph);
	if (skb_headroom(skb) < max_headroom) {
		DEBUG_WARN("====skb_headroom(skb): %u, max_headroom:%u\n", skb_headroom(skb), max_headroom);
		if(skb_cow_head(skb, max_headroom))
		{
			DEBUG_ERROR("skb_cow_head error, free skb, max_headroom:%u skb->len %d skb_headroom(skb) %d\n", max_headroom,skb->len,skb_headroom(skb));
			kfree_skb(skb);
			return NULL;
		}
	}

	/* add PPP */
	skb_push(skb, sizeof(*ppph));
	ppph = (struct sfe_ipv4_pppoe_ppp_hdr *)(skb->data);
	ppph->protocol = protocol;

	/* add pppoe header */
	ret = add_header(ti->tunnel_dev, ti->info.pppoe.sid, ti->info.pppoe.source, skb);
	if (!ret) {
		skb_pull(skb, sizeof(*ppph));
		return skb;
	}
	skb_reset_network_header(skb);

	/* add eth hdr */
	skb_push(skb, sizeof(*eth));
	eth =  (struct sfe_ipv4_eth_hdr*)(skb->data);
	eth->h_proto = htons(ETH_P_PPP_SES);
	eth->h_dest[0] = ti->dest_mac[0];
	eth->h_dest[1] = ti->dest_mac[1];
	eth->h_dest[2] = ti->dest_mac[2];
	eth->h_source[0] = ti->src_mac[0];
	eth->h_source[1] = ti->src_mac[1];
	eth->h_source[2] = ti->src_mac[2];
	skb_reset_mac_header(skb);

	skb->dev = ti->tunnel_dev;

	return skb;
}
#endif

#ifdef TPLINK_FRAGMENT_OUTPUT
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 0, 0))
static int sfe_ipv4_fragment_output(struct net *net, struct sock *sk, struct sk_buff *skb)
#else
static int sfe_ipv4_fragment_output(struct sk_buff *skb)
#endif
{
	struct sfe_ipv4_eth_hdr *eth, *eth1;

	if (skb->shortcut_data) {
		eth1 = (struct sfe_ipv4_eth_hdr *)skb->shortcut_data;
		eth  = (struct sfe_ipv4_eth_hdr *)__skb_push(skb, ETH_HLEN);
		eth->h_proto = eth1->h_proto;
		eth->h_dest[0] = eth1->h_dest[0];
		eth->h_dest[1] = eth1->h_dest[1];
		eth->h_dest[2] = eth1->h_dest[2];
		eth->h_source[0] = eth1->h_source[0];
		eth->h_source[1] = eth1->h_source[1];
		eth->h_source[2] = eth1->h_source[2];
		skb->shortcut_data = 0;
	}
	TPLINK_DEBUG("head_room(%s)(%p):%u\n", skb->dev->name, skb, skb_headroom(skb));
#ifdef TPLINK_PPP_TUNNEL
	return sfe_ipv4_dev_queue_xmit(skb);
#else
	return dev_queue_xmit(skb);
#endif
}

static int sfe_ipv4_fragment(struct sk_buff *skb, int l2_header)
{
	struct sfe_ipv4_eth_hdr xmit_eth;
	struct rtable *rt = skb_rtable(skb);
	struct net_device *frag_dev = skb->dev;
	uint32_t old_head_room = 0;

	if (unlikely(!rt)) {
		kfree_skb(skb);
		return -1;
	}

	skb->shortcut_data = 0;
	if (l2_header) {
		struct sfe_ipv4_eth_hdr *eth = (struct sfe_ipv4_eth_hdr *)skb->data;
		xmit_eth.h_proto = eth->h_proto;
		xmit_eth.h_dest[0] = eth->h_dest[0];
		xmit_eth.h_dest[1] = eth->h_dest[1];
		xmit_eth.h_dest[2] = eth->h_dest[2];
		xmit_eth.h_source[0] = eth->h_source[0];
		xmit_eth.h_source[1] = eth->h_source[1];
		xmit_eth.h_source[2] = eth->h_source[2];
		skb->shortcut_data = (unsigned long)&xmit_eth;

		__skb_pull(skb, ETH_HLEN);
	}
	TPLINK_DEBUG("pre frag: head_room(%s)(%p):%u\n", skb->dev->name, skb, skb_headroom(skb));
	if (frag_dev && (frag_dev->flags & IFF_POINTOPOINT)) {
		TPLINK_DEBUG("expand new skb headroom1\n");
		old_head_room = frag_dev->needed_headroom;
		frag_dev->needed_headroom = frag_dev->needed_headroom + TPLINK_RESERVE_PPP_LEN;
#ifdef TPLINK_PPP_TUNNEL
		{
			struct tunnel_info *ti = NULL;
			read_lock_bh(&tunnel_rwlock);
			ti = find_tunnel_info_origin(frag_dev);
			if (ti) {
				TPLINK_DEBUG("expand new skb headroom2\n");
				frag_dev->needed_headroom += LL_RESERVED_SPACE(ti->tunnel_dev); /* change headroom for add ppp header */
			}
			read_unlock_bh(&tunnel_rwlock);
		}
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 0, 0))
		ip_do_fragment(dev_net(frag_dev), NULL, skb, sfe_ipv4_fragment_output);
#else
		ip_fragment(skb, sfe_ipv4_fragment_output);
#endif
		frag_dev->needed_headroom = old_head_room; /* revert headroom for add ppp header */
	}
	else {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 0, 0))
		ip_do_fragment(dev_net(frag_dev), NULL, skb, sfe_ipv4_fragment_output);
#else
		ip_fragment(skb, sfe_ipv4_fragment_output);
#endif
	}
	TPLINK_DEBUG("after frag: %p\n", skb);
	return 0;
}
#endif
//void sfe_skb_ct_set(struct sk_buff *skb, uint16_t l3protocol, uint32_t saddr, uint32_t daddr, uint8_t l4protocol, uint16_t sport, uint16_t dport)
//{
//	struct nf_conntrack_tuple_hash *h;
//    struct nf_conntrack_tuple tuple;
//	struct nf_conn *ct = NULL;
//
//	if (unlikely(skb->nfct)) {
//		return;
//	}
//	
//	/*
//	 * Create a tuple so as to be able to look up a connection
//	 */
//    memset(&tuple, 0, sizeof(tuple));
//    tuple.src.u3.ip = saddr;
//    tuple.dst.u3.ip = daddr;
//    tuple.dst.protonum = l4protocol;
//    tuple.src.l3num = l3protocol;
//    tuple.dst.dir = IP_CT_DIR_ORIGINAL;
//    tuple.src.u.all = sport;
//    tuple.dst.u.all = dport;
//
//	h = nf_conntrack_find_get(&init_net, NF_CT_DEFAULT_ZONE, &tuple);
//	if (unlikely(h)) {
//		ct = nf_ct_tuplehash_to_ctrack(h);
//		skb->nfct = ct;
//	}
//}


#ifdef TPLINK_FRAGMENT_OUTPUT
static inline void sfe_skb_dst_set(struct sk_buff *skb, struct sfe_ipv4_ip_hdr *iph, struct sfe_ipv4_connection *connection)
{
	__be32 src_ip = iph->saddr;
	__be32 dest_ip = iph->daddr;

	if (likely(connection->dst) &&
		((src_ip == connection->src_ip && dest_ip == connection->dest_ip) || 
		(src_ip == connection->src_ip_xlate && dest_ip == connection->dest_ip_xlate))) {
		dst_hold(connection->dst);
		skb_dst_set(skb, connection->dst);
	}
	else if (likely(connection->reply_dst) &&
		((src_ip == connection->dest_ip && dest_ip == connection->src_ip) || 
		(src_ip == connection->dest_ip_xlate && dest_ip == connection->src_ip_xlate))) {
		dst_hold(connection->reply_dst);
		skb_dst_set(skb, connection->reply_dst);
	}
	else
		return;
}
#endif

/*
 * sfe_ipv4_recv_udp()
 *	Handle UDP packet receives and forwarding.
 */
static int sfe_ipv4_recv_udp(struct sfe_ipv4 *si, struct sk_buff *skb, struct net_device *dev,
			     unsigned int len, struct sfe_ipv4_ip_hdr *iph, unsigned int ihl, bool flush_on_find, int sfe_proto)
{
	struct sfe_ipv4_udp_hdr *udph = NULL;
	__be32 src_ip;
	__be32 dest_ip;
	__be16 src_port;
	__be16 dest_port;
	struct sfe_ipv4_connection_match *cm;
	uint8_t ttl;
	struct net_device *xmit_dev;
	struct net_device *recv_dev;
	struct net_device *match_dev;
	uint16_t udsum = 0;
#ifdef CONFIG_NF_FASTPATH_NAT_HOOK
	typeof(fastpath_nat_forward_hook) fastpath_nat_hook;
	typeof(qos_fastpath_nat_forward_hook) qos_fastpath_nat_hook;
#endif
#ifdef TPLINK_LAN_WAN_NETSTACK
	struct sfe_ipv4_connection_match *run_cm = NULL;
	struct sfe_ipv4_connection_match *partner_cm = NULL;
#endif
#ifdef TPLINK_ACCELERATE_NAT
	int accelerate = !!(skb->shortcut_data);
#else
	int accelerate = 0;
#endif
//	int k = 0;

#ifdef TPLINK_FRAGMENT_OUTPUT
	int need_frag = false;
	int l2_header = false;
#endif

#if SFE_ACCELERATE_IPSEC
    int ret = 0;
    struct xfrm_state *x = NULL;
#endif

#ifdef TPLINK_FRAGMENT
	unsigned int frag_off = ntohs(iph->frag_off);

	if (likely(!(frag_off & IP_OFFSET))) {
#endif
	/*
	 * Is our packet too short to contain a valid UDP header?
	 */
	if (unlikely(!pskb_may_pull(skb, (sizeof(struct sfe_ipv4_udp_hdr) + ihl)))) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_UDP_HEADER_INCOMPLETE]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("packet too short for UDP header\n");
		return 0;
	}
//	TPLINK_DEBUG("cc======three==%lx=%u=%u=%u=\n", (unsigned long)iph, iph->ihl, skb->len, skb->data_len);
//	TPLINK_DEBUG("cc======\n");	
//	for (k = 0; k < (iph->ihl * 4); k++) {
//		TPLINK_DEBUG("%02x ", ((char *)iph)[k]);
//	}
//	TPLINK_DEBUG("acc======\n");
	/*
	 * Read the IP address and port information.  Read the IP header data first
	 * because we've almost certainly got that in the cache.  We may not yet have
	 * the UDP header cached though so allow more time for any prefetching.
	 */
	src_ip = iph->saddr;
	dest_ip = iph->daddr;

	udph = (struct sfe_ipv4_udp_hdr *)(skb->data + ihl);
	src_port = udph->source;
	dest_port = udph->dest;
	udsum = udph->check;
	//DEBUG_TRACE("Receive from: %pI4:%u To: %pI4:%u\n",
	//    &iph->saddr, src_port, &iph->daddr, dest_port);
	spin_lock_bh(&si->lock);

#ifdef TPLINK_FRAGMENT
		if (unlikely(frag_off & IP_MF)) {	
			struct frag_node *nd = NULL;
			nd = frag_cmlist_find(iph->id, IPPROTO_UDP, iph->saddr, iph->daddr);
			if (unlikely(nd && nd->sport == 0 && nd->dport == 0)){
				spin_unlock_bh(&si->lock);
				return 0;
			}
			else{
				frag_cmlist_insert(iph->id, IPPROTO_UDP, src_ip, dest_ip, src_port, dest_port);
			}
			
		}
	} else {
		struct frag_node *nd = NULL;

		udph = NULL;
		spin_lock_bh(&si->lock);
		nd = frag_cmlist_find(iph->id, IPPROTO_UDP, iph->saddr, iph->daddr);
		if (unlikely(!nd || (nd && nd->sport == 0 && nd->dport == 0))) {
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_UDP_NO_CONNECTION]++;
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_NON_INITIAL_FRAGMENT]++;
			si->packets_not_forwarded++;
			frag_cmlist_insert(iph->id, IPPROTO_UDP, iph->saddr, iph->daddr, 0, 0);
			spin_unlock_bh(&si->lock);
			DEBUG_TRACE("no connection found\n");
			return 0;
		}

		src_ip = iph->saddr;
		dest_ip = iph->daddr;

		src_port = nd->sport;
		dest_port = nd->dport;
	}
#endif

#if SFE_ACCELERATE_IPSEC

    if ((htons(4500) == src_port) || (htons(4500) == dest_port))
    {
        if (!(frag_off & (IP_OFFSET|IP_MF)))
        {
            if (0 != sfe_ipv4_recv_esp(si, skb, iph, true,sfe_proto))
            {
                spin_unlock_bh(&si->lock);
                return 1;
            }
        }
        else
        {
            spin_unlock_bh(&si->lock);
            return 0;
        }
    }
#endif

	/*
	 * Look for a connection match.
	 */
#ifdef CONFIG_NF_FLOW_COOKIE
	cm = si->sfe_flow_cookie_table[skb->flow_cookie & SFE_FLOW_COOKIE_MASK].match;
	if (unlikely(!cm)) {
		cm = sfe_ipv4_find_sfe_ipv4_connection_match(si, skb->dev, IPPROTO_UDP, src_ip, src_port, dest_ip, dest_port);
	}
#else
	cm = sfe_ipv4_find_sfe_ipv4_connection_match(si, skb->dev, IPPROTO_UDP, src_ip, src_port, dest_ip, dest_port);
#endif
	if (unlikely(!cm)) {
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_UDP_NO_CONNECTION]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("no connection found [%pI4:%d===>%pI4:%d]\n", &src_ip, ntohs(src_port), &dest_ip, ntohs(dest_port));
		return 0;
	}

	if (cm->connection->sfe_proto <= SFE_PROTO_DEF) {
		cm->connection->sfe_proto = sfe_proto;
	}
#ifdef TPLINK_LAN_WAN_NETSTACK
	run_cm = cm;
	if (!cm->partner_match) {
		__be32 p_src_ip = cm->xlate_src_ip;
		__be32 p_dest_ip = cm->xlate_dest_ip;
		__be16 p_src_port = cm->xlate_src_port;
		__be16 p_dest_port = cm->xlate_dest_port;

		partner_cm = NULL;
		if (lan_inner_dev && wan_inner_dev && (cm->xmit_dev == lan_inner_dev)) {
			partner_cm = sfe_ipv4_find_sfe_ipv4_connection_match(si, wan_inner_dev, IPPROTO_UDP, p_src_ip, p_src_port, p_dest_ip, p_dest_port);
		} else if (wan_inner_dev && lan_inner_dev && (cm->xmit_dev == wan_inner_dev)) {
			partner_cm = sfe_ipv4_find_sfe_ipv4_connection_match(si, lan_inner_dev, IPPROTO_UDP, p_src_ip, p_src_port, p_dest_ip, p_dest_port);
		}
		if (partner_cm) {
			cm->partner_match = partner_cm;
			partner_cm->counter_match->partner_match = cm->counter_match;
		}
	} else {
		partner_cm = cm->partner_match;
	}
#endif

	/* if hwnat need this skb, we don't forward it. */
	if (hwnat_taken_skb(skb, iph, cm, partner_cm)) {
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_UDP_HWNAT_CONTROL]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("hwnat need this skb\n");
		return 0;
	}

	/*
	 * If our packet has beern marked as "flush on find" we can't actually
	 * forward it in the fast path, but now that we've found an associated
	 * connection we can flush that out before we process the packet.
	 */
	if (unlikely(flush_on_find)) {
		struct sfe_ipv4_connection *c = cm->connection;
		sfe_ipv4_remove_sfe_ipv4_connection(si, c);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_UDP_IP_OPTIONS_OR_INITIAL_FRAGMENT]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("flush on find\n");
		sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
		return 0;
	}

	/*
	 * Does our TTL allow forwarding?
	 */
	ttl = iph->ttl;
	if (unlikely(ttl < 2)) {
		struct sfe_ipv4_connection *c = cm->connection;
		sfe_ipv4_remove_sfe_ipv4_connection(si, c);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_UDP_SMALL_TTL]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("ttl too low\n");
		sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
		return 0;
	}

	/*
	 * If our packet is larger than the MTU of the transmit interface then
	 * we can't forward it easily.
	 */

#ifndef TPLINK_LAN_WAN_NETSTACK
#ifndef TPLINK_FRAGMENT_OUTPUT
	if (unlikely(len > cm->xmit_dev_mtu)) {
#else
	if (unlikely((len > cm->xmit_dev_mtu) && (!cm->connection->dst || (frag_off & IP_DF)))) {
#endif
		struct sfe_ipv4_connection *c = cm->connection;
		sfe_ipv4_remove_sfe_ipv4_connection(si, c);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_UDP_NEEDS_FRAGMENTATION]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("larger than mtu\n");
		sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
		return 0;
	}
#else /* TPLINK_LAN_WAN_NETSTACK */
#ifndef TPLINK_FRAGMENT_OUTPUT
	if (unlikely(partner_cm && (len > partner_cm->xmit_dev_mtu))) {
#else
	if (unlikely(partner_cm && (len > partner_cm->xmit_dev_mtu) && (!partner_cm->connection->dst || (frag_off & IP_DF)))) {
#endif
		partner_cm = NULL;
	}
#ifndef TPLINK_FRAGMENT_OUTPUT
	if (unlikely(!partner_cm && (len > cm->xmit_dev_mtu))) {
#else
	if (unlikely(!partner_cm && (len > cm->xmit_dev_mtu) && (!cm->connection->dst || (frag_off & IP_DF)))) {
#endif
		struct sfe_ipv4_connection *c = cm->connection;
		sfe_ipv4_remove_sfe_ipv4_connection(si, c);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_UDP_NEEDS_FRAGMENTATION]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("larger than mtu\n");
		sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
		return 0;
	}
#endif /* TPLINK_LAN_WAN_NETSTACK */

#ifdef TPLINK_LAN_WAN_NETSTACK
	cm = (dev_net(run_cm->match_dev) == &init_net) ? run_cm : (partner_cm ? partner_cm : NULL);	
	if (cm) {	// start of cm = (dev_net(run_cm->match_dev) == &init_net) ? run_cm : (partner_cm ? partner_cm : NULL);	
#endif

#ifdef CONFIG_NF_FASTPATH_NAT_HOOK
	fastpath_nat_hook = rcu_dereference(fastpath_nat_forward_hook);
	if (fastpath_nat_enabled && fastpath_nat_hook && cm->connection->nfct) {
		struct nf_conn *ct = (struct nf_conn *)cm->connection->nfct;
		enum ip_conntrack_dir dir = (cm->connection->original_match == cm) ? IP_CT_DIR_ORIGINAL : IP_CT_DIR_REPLY;
		switch (fastpath_nat_hook(skb, ct, dir)) {
			case 0:
				break;
			case -1:
				si->exception_events[SFE_IPV4_EXCEPTION_EVENT_BANDWIDTH_CONTROL]++;
				si->packets_not_forwarded++;
				spin_unlock_bh(&si->lock);
				return 0;
			default:
				si->exception_events[SFE_IPV4_EXCEPTION_EVENT_BANDWIDTH_CONTROL]++;
				si->packets_not_forwarded++;
				spin_unlock_bh(&si->lock);
				kfree_skb(skb);
				return 1;
		}
	}

	qos_fastpath_nat_hook = rcu_dereference(qos_fastpath_nat_forward_hook);
        if (qos_fastpath_nat_enabled && qos_fastpath_nat_hook && cm->connection->nfct) {
                struct nf_conn *ct = (struct nf_conn *)cm->connection->nfct;
                enum ip_conntrack_dir dir = (cm->connection->original_match == cm) ? IP_CT_DIR_ORIGINAL : IP_CT_DIR_REPLY;
                switch (qos_fastpath_nat_hook(skb, ct, dir)) {
                        case 0:
                                break;
                        case -1:
                                si->exception_events[SFE_IPV4_EXCEPTION_EVENT_BANDWIDTH_CONTROL]++;
                                si->packets_not_forwarded++;
                                spin_unlock_bh(&si->lock);
                                return 0;
                        default:
                                si->exception_events[SFE_IPV4_EXCEPTION_EVENT_BANDWIDTH_CONTROL]++;
                                si->packets_not_forwarded++;
                                spin_unlock_bh(&si->lock);
                                kfree_skb(skb);
                                return 1;
                }
        }

#endif

#ifdef SFE_IPSTAT
	sfe_ipstat_hook(cm, skb);
#endif

#ifdef TPLINK_LAN_WAN_NETSTACK
	} // end of cm = (dev_net(run_cm->match_dev) == &init_net) ? run_cm : (partner_cm ? partner_cm : NULL);	
#endif

	/*
	 * From this point on we're good to modify the packet.
	 */

	if (!accelerate) {
	/*
	 * Update DSCP
	 */
#ifdef TPLINK_LAN_WAN_NETSTACK
	/* TODO: twice xlate src */
	cm = (partner_cm && (partner_cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_DSCP_REMARK)) ? partner_cm : run_cm;
#endif
	if (unlikely(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_DSCP_REMARK)) {
		iph->tos = (iph->tos & SFE_IPV4_DSCP_MASK) | cm->dscp;
	}
	}

	/*
	 * Decrement our TTL.
	 */
	if ((skb->dev != lan_inner_dev) && (partner_cm || (run_cm->xmit_dev != lan_inner_dev))) {
		iph->ttl = ttl - 1;
	}


	/*
	 * Do we have to perform translations of the source address/port?
	 */
#ifdef TPLINK_LAN_WAN_NETSTACK
	/* TODO: twice xlate src */
	cm = (partner_cm && (partner_cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_SRC)) ? partner_cm : run_cm;
#endif
	if (unlikely(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_SRC)) {
		uint16_t udp_csum;

		iph->saddr = cm->xlate_src_ip;
		if (likely(udph)) {
			udph->source = cm->xlate_src_port;

			/*
			 * Do we have a non-zero UDP checksum?  If we do then we need
			 * to update it.
			 */
			udp_csum = udph->check;
			if (likely(udp_csum)) {
				uint32_t sum;

				if (unlikely(skb->ip_summed == CHECKSUM_PARTIAL)) {
					sum = udp_csum + cm->xlate_src_partial_csum_adjustment;
				} else {
					sum = udp_csum + cm->xlate_src_csum_adjustment;
				}

				sum = (sum & 0xffff) + (sum >> 16);
				udph->check = (uint16_t)sum;
			}
		}
	}

	/*
	 * Do we have to perform translations of the destination address/port?
	 */
#ifdef TPLINK_LAN_WAN_NETSTACK
	/* TODO: twice xlate dest */
	cm = (partner_cm && (partner_cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_DEST)) ? partner_cm : run_cm;
#endif
	if (unlikely(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_DEST)) {
		uint16_t udp_csum;

		iph->daddr = cm->xlate_dest_ip;
		if (likely(udph)) {
			udph->dest = cm->xlate_dest_port;

			/*
			 * Do we have a non-zero UDP checksum?  If we do then we need
			 * to update it.
			 */
			udp_csum = udph->check;
			if (likely(udp_csum)) {
				uint32_t sum;

				if (unlikely(skb->ip_summed == CHECKSUM_PARTIAL)) {
					sum = udp_csum + cm->xlate_dest_partial_csum_adjustment;
				} else {
					sum = udp_csum + cm->xlate_dest_csum_adjustment;
				}

				sum = (sum & 0xffff) + (sum >> 16);
				udph->check = (uint16_t)sum;
			}
		}
	}

	/*
	 * Replace the IP checksum.
	 */
	iph->check = sfe_ipv4_gen_ip_csum(iph);

	/*
	 * Update traffic stats.
	 */
	run_cm->rx_packet_count++;
	run_cm->rx_byte_count += len;
	if (partner_cm) {
		partner_cm->rx_packet_count++;
		partner_cm->rx_byte_count += len;
	}

	/*
	 * If we're not already on the active list then insert ourselves at the tail
	 * of the current list.
	 */
	if (unlikely(!cm->active)) {
		cm->active = true;
		cm->active_prev = si->active_tail;
		if (likely(si->active_tail)) {
			si->active_tail->active_next = cm;
		} else {
			si->active_head = cm;
		}
		si->active_tail = cm;
	}

#if SFE_ACCELERATE_IPSEC
    if (unlikely(!cm->set_accel))
    {
        if (NULL != (x = xfrm_state_lookup_byskb(skb, 0)))
        {
            if ((x->data != NULL) && (x->km.state == XFRM_STATE_VALID)) 
            {
                cm->flow_accel = 1;
                cm->x = x; /* xfrm_state_lookup_byskb has hold x, so no need xfrm_state_put here. */
            }
        }
        cm->set_accel = 1;
    }
    if ((cm->set_accel) && (cm->flow_accel) && (cm->x))
    {
        /* now only support esp accelerate */
        if ((cm->x->id.proto == IPPROTO_ESP) && (!(frag_off & IP_OFFSET|IP_MF)))
        {
            ret = sfe_ipv4_encap_esp(cm->x, skb, iph);
        }
        if (ret == 0)
        {
            if (unlikely(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_DEST)) {
            iph->daddr = dest_ip;
            if (likely(udph)) {
                udph->dest = dest_port;
                udph->check = udsum;
            }
            iph->check = sfe_ipv4_gen_ip_csum(iph);
            }
        }

        spin_unlock_bh(&si->lock);
        return ret;
    }
#endif


#ifdef TPLINK_LAN_WAN_NETSTACK
	cm = partner_cm?partner_cm:run_cm;
	match_dev = run_cm->match_dev;
#else
	match_dev = cm->match_dev;
#endif
	recv_dev = skb->dev;
	xmit_dev = cm->xmit_dev;
	skb->dev = xmit_dev;

	/*
	 * Check to see if we need to write a header.
	 */
	if (likely(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_WRITE_L2_HDR)) {
		if (unlikely(!(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_WRITE_FAST_ETH_HDR))) {
			dev_hard_header(skb, xmit_dev, ETH_P_IP,
					cm->xmit_dest_mac, cm->xmit_src_mac, len);
		} else {
			/*
			 * For the simple case we write this really fast.
			 */
			struct sfe_ipv4_eth_hdr *eth = (struct sfe_ipv4_eth_hdr *)__skb_push(skb, ETH_HLEN);
			eth->h_proto = htons(ETH_P_IP);
			eth->h_dest[0] = cm->xmit_dest_mac[0];
			eth->h_dest[1] = cm->xmit_dest_mac[1];
			eth->h_dest[2] = cm->xmit_dest_mac[2];
			eth->h_source[0] = cm->xmit_src_mac[0];
			eth->h_source[1] = cm->xmit_src_mac[1];
			eth->h_source[2] = cm->xmit_src_mac[2];
		}
#ifdef TPLINK_FRAGMENT_OUTPUT
		l2_header = true;
#endif
	}

	if (!accelerate) {
	/*
	 * Remove vlan info for packets fast-forwarding. Or packets would be blocked by vlan filter.
	 */
	skb->vlan_proto = 0;
	skb->vlan_tci = 0;

	/*
	 * Update priority of skb.
	 */
#ifdef TPLINK_LAN_WAN_NETSTACK
	cm = (partner_cm && (partner_cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_PRIORITY_REMARK)) ? partner_cm : run_cm;
#endif
	if (unlikely(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_PRIORITY_REMARK)) {
		skb->priority = cm->priority;
	}

	/*
	 * Mark outgoing packet.
	 */
#ifdef TPLINK_LAN_WAN_NETSTACK
	cm = partner_cm?partner_cm:run_cm;
#endif
	skb->mark = ((struct nf_conn *)run_cm->connection->nfct)->mark;
	skb->mark |= partner_cm?((struct nf_conn *)partner_cm->connection->nfct)->mark:0;
	if (skb->mark) {
		DEBUG_TRACE("SKB MARK is NON ZERO %x\n", skb->mark);
	}

	if ((SKB_NFCT_GET(skb) == NULL) && atomic_inc_not_zero(&cm->connection->nfct->use)) {
		SKB_NFCT_SET(skb, (struct nf_conn *)cm->connection->nfct, 
		    (cm->connection->original_match == cm) ? IP_CT_ESTABLISHED : IP_CT_ESTABLISHED_REPLY);
	}
	}

#if defined(CONFIG_NF_CONNTRACK_EXTMARK)
	skb->extmark = ((struct nf_conn *)run_cm->connection->nfct)->extmark;
	skb->extmark |= partner_cm?((struct nf_conn *)partner_cm->connection->nfct)->extmark:0;
	if (skb->extmark) {
		DEBUG_TRACE("SKB EXTMARK is NON ZERO %x\n", skb->extmark);
	}
#endif

	si->packets_forwarded++;

#ifdef TPLINK_FRAGMENT_OUTPUT
	//printk("len:%d, xmit_dev_name:%s, cm->xmit_dev_mtu:%u\n", len, cm->xmit_dev->name, cm->xmit_dev_mtu);
	if (unlikely(len > cm->xmit_dev_mtu)) {
		need_frag = true;
		sfe_skb_dst_set(skb, iph, cm->connection);
	}
#endif

	spin_unlock_bh(&si->lock);

	if (!accelerate) {
	/*
	 * We're going to check for GSO flags when we transmit the packet so
	 * start fetching the necessary cache line now.
	 */
	prefetch(skb_shinfo(skb));
	}
	/*
	 * Mark that this packet has been fast forwarded.
	 */
	skb->fast_forwarded = 1;

	/*
	 * Send the packet on its way.
	 */
	//if (!dev){
	//	printk(KERN_ERR"a == iph->id:%x, iph->tot_len:%u, skb->len:%u, skb->data_len:%u, %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x \n", 
	//	iph->id, ntohs(iph->tot_len), skb->len, skb->data_len,
	//	skb->data[0],skb->data[1],skb->data[2],skb->data[3],skb->data[4],skb->data[5],skb->data[6],skb->data[7],skb->data[8],skb->data[9],skb->data[10],skb->data[11],skb->data[12],skb->data[13],skb->data[14],skb->data[15],skb->data[16],skb->data[17],skb->data[18],skb->data[19],skb->data[20],skb->data[21],skb->data[22],skb->data[23],skb->data[24],skb->data[25],skb->data[26],skb->data[27],skb->data[28],skb->data[29],skb->data[30],skb->data[31],skb->data[32],skb->data[33],skb->data[34],skb->data[35],skb->data[36],skb->data[37],skb->data[38],skb->data[39],skb->data[40],skb->data[41],skb->data[42],skb->data[43],skb->data[44],skb->data[45],skb->data[46],skb->data[47],skb->data[48],skb->data[49]);
		//printk(KERN_ERR"\n");
	//printk(KERN_ERR"xmit:%s\n",skb->dev->name);
	//}

	if (!dev || (skb->dev && (skb->dev->flags & IFF_POINTOPOINT))) {
		hw_nat_slow_path(skb);
	}
#ifdef TPLINK_PPP_TUNNEL
	if (match_dev && (sfe_proto > SFE_PROTO_DEF && sfe_proto < SFE_PROTO_IPSEC ) && (match_dev->flags & IFF_POINTOPOINT) && (!dev_tuntype(match_dev))) {
		ppp_dev_add_stats(match_dev, skb, len, 1);
	}
#endif

#ifdef TPLINK_FRAGMENT_OUTPUT
	if (need_frag)
		sfe_ipv4_fragment(skb, l2_header);
	else
#endif
#ifdef TPLINK_PPP_TUNNEL
	sfe_ipv4_dev_queue_xmit(skb);
#else
#ifdef TPLINK_ACCELERATE_NAT
	if (skb->shortcut_data) {
		sfe_hard_xmit_func_t xmit = (sfe_hard_xmit_func_t)skb->shortcut_data;
		xmit(skb);
		skb->shortcut_data = 0;
	}
	else {
		dev_queue_xmit(skb);
	}
#else
	dev_queue_xmit(skb);
#endif
#endif
	return 1;
}

/*
 * sfe_ipv4_process_tcp_option_sack()
 *	Parse TCP SACK option and update ack according
 */
static bool sfe_ipv4_process_tcp_option_sack(const struct sfe_ipv4_tcp_hdr *th, const uint32_t data_offs,
					     uint32_t *ack) __attribute__((always_inline));
static bool sfe_ipv4_process_tcp_option_sack(const struct sfe_ipv4_tcp_hdr *th, const uint32_t data_offs,
					     uint32_t *ack)
{
	uint32_t length = sizeof(struct sfe_ipv4_tcp_hdr);
	uint8_t *ptr = (uint8_t *)th + length;

	/*
	 * If option is TIMESTAMP discard it.
	 */
	if (likely(data_offs == length + TCPOLEN_TIMESTAMP + 1 + 1)
	    && likely(ptr[0] == TCPOPT_NOP)
	    && likely(ptr[1] == TCPOPT_NOP)
	    && likely(ptr[2] == TCPOPT_TIMESTAMP)
	    && likely(ptr[3] == TCPOLEN_TIMESTAMP)) {
		return true;
	}

	/*
	 * TCP options. Parse SACK option.
	 */
	while (length < data_offs) {
		uint8_t size;
		uint8_t kind;

		ptr = (uint8_t *)th + length;
		kind = *ptr;

		/*
		 * NOP, for padding
		 * Not in the switch because to fast escape and to not calculate size
		 */
		if (kind == TCPOPT_NOP) {
			length++;
			continue;
		}

		if (kind == TCPOPT_SACK) {
			uint32_t sack = 0;
			uint8_t re = 1 + 1;

			size = *(ptr + 1);
			if ((size < (1 + 1 + TCPOLEN_SACK_PERBLOCK))
			    || ((size - (1 + 1)) % (TCPOLEN_SACK_PERBLOCK))
			    || (size > (data_offs - length))) {
				return false;
			}

			re += 4;
			while (re < size) {
				uint32_t sack_re;
				uint8_t *sptr = ptr + re;
				sack_re = (sptr[0] << 24) | (sptr[1] << 16) | (sptr[2] << 8) | sptr[3];
				if (sack_re > sack) {
					sack = sack_re;
				}
				re += TCPOLEN_SACK_PERBLOCK;
			}
			if (sack > *ack) {
				*ack = sack;
			}
			length += size;
			continue;
		}
		if (kind == TCPOPT_EOL) {
			return true;
		}
		size = *(ptr + 1);
		if (size < 2) {
			return false;
		}
		length += size;
	}

	return true;
}

/*
 * sfe_ipv4_recv_tcp()
 *	Handle TCP packet receives and forwarding.
 */
static int sfe_ipv4_recv_tcp(struct sfe_ipv4 *si, struct sk_buff *skb, struct net_device *dev,
			     unsigned int len, struct sfe_ipv4_ip_hdr *iph, unsigned int ihl, bool flush_on_find, int sfe_proto)
{
	struct sfe_ipv4_tcp_hdr *tcph;
	__be32 src_ip;
	__be32 dest_ip;
	__be16 src_port;
	__be16 dest_port;
	struct sfe_ipv4_connection_match *cm;
	struct sfe_ipv4_connection_match *counter_cm;
	uint8_t ttl;
	uint32_t flags = 0;
	struct net_device *xmit_dev;
	struct net_device *recv_dev;
	struct net_device *match_dev;
	uint16_t tcsum = 0;
#ifdef CONFIG_NF_FASTPATH_NAT_HOOK
	typeof(fastpath_nat_forward_hook) fastpath_nat_hook;
	typeof(qos_fastpath_nat_forward_hook) qos_fastpath_nat_hook;
#endif

#ifdef TPLINK_LAN_WAN_NETSTACK
	struct sfe_ipv4_connection_match *run_cm = NULL;
	struct sfe_ipv4_connection_match *partner_cm = NULL;
	int i = 0;
#endif

#ifdef TPLINK_FRAGMENT_OUTPUT
	int need_frag = false;
	int l2_header = false;
#endif

#if SFE_ACCELERATE_IPSEC
    int ret = 0;
    struct xfrm_state *x = NULL;
#endif

#ifdef TPLINK_FRAGMENT
	unsigned int frag_off = ntohs(iph->frag_off);

	if (likely(!(frag_off & IP_OFFSET))) {
#endif
	/*
	 * Is our packet too short to contain a valid UDP header?
	 */
	if (unlikely(!pskb_may_pull(skb, (sizeof(struct sfe_ipv4_tcp_hdr) + ihl)))) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_HEADER_INCOMPLETE]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("packet too short for TCP header\n");
		return 0;
	}

	/*
	 * Read the IP address and port information.  Read the IP header data first
	 * because we've almost certainly got that in the cache.  We may not yet have
	 * the TCP header cached though so allow more time for any prefetching.
	 */
	src_ip = iph->saddr;
	dest_ip = iph->daddr;

	tcph = (struct sfe_ipv4_tcp_hdr *)(skb->data + ihl);
	src_port = tcph->source;
	dest_port = tcph->dest;
	tcsum = tcph->check;
	flags = tcp_flag_word(tcph);

	spin_lock_bh(&si->lock);
#ifdef TPLINK_FRAGMENT
		if (unlikely(frag_off & IP_MF)) {
			frag_cmlist_insert(iph->id, IPPROTO_TCP, src_ip, dest_ip, src_port, dest_port);
		}
	} else {
		struct frag_node *nd = NULL;

		tcph = NULL;
		spin_lock_bh(&si->lock);
		nd = frag_cmlist_find(iph->id, IPPROTO_TCP, iph->saddr, iph->daddr);
		if (unlikely(!nd)) {
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_NO_CONNECTION_SLOW_FLAGS]++;
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_NON_INITIAL_FRAGMENT]++;
			si->packets_not_forwarded++;
			spin_unlock_bh(&si->lock);

			DEBUG_TRACE("no fragment connection found\n");
			return 0;
		}

		src_ip = iph->saddr;
		dest_ip = iph->daddr;

		src_port = nd->sport;
		dest_port = nd->dport;
	}
#endif

	/*
	 * Look for a connection match.
	 */
#ifdef CONFIG_NF_FLOW_COOKIE
	cm = si->sfe_flow_cookie_table[skb->flow_cookie & SFE_FLOW_COOKIE_MASK].match;
	if (unlikely(!cm)) {
		cm = sfe_ipv4_find_sfe_ipv4_connection_match(si, skb->dev, IPPROTO_TCP, src_ip, src_port, dest_ip, dest_port);
	}
#else
	cm = sfe_ipv4_find_sfe_ipv4_connection_match(si, skb->dev, IPPROTO_TCP, src_ip, src_port, dest_ip, dest_port);
#endif
	if (unlikely(!cm)) {
		/*
		 * We didn't get a connection but as TCP is connection-oriented that
		 * may be because this is a non-fast connection (not running established).
		 * For diagnostic purposes we differentiate this here.
		 */
		if (likely(tcph) && likely((flags & (TCP_FLAG_SYN | TCP_FLAG_RST | TCP_FLAG_FIN | TCP_FLAG_ACK)) == TCP_FLAG_ACK)) {
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_NO_CONNECTION_FAST_FLAGS]++;
			si->packets_not_forwarded++;
			spin_unlock_bh(&si->lock);

			//DEBUG_TRACE("no connection found - fast flags,flags = %x\n",flags);
			//DEBUG_TRACE("conntrack info - dev: %s \n"
                   //"  src_ip:%pI4,src_port:%u\n"
                   //"  dest_ip:%pI4,dest_port:%u\n",
                   //dev->name,&src_ip,src_port,&dest_ip,dest_port);

			return 0;
		}
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_NO_CONNECTION_SLOW_FLAGS]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		//DEBUG_TRACE("no connection found - slow flags: 0x%x\n",
			    //flags & (TCP_FLAG_SYN | TCP_FLAG_RST | TCP_FLAG_FIN | TCP_FLAG_ACK));
		return 0;
	}

#ifdef TPLINK_LAN_WAN_NETSTACK
	run_cm = cm;
	if (!cm->partner_match) {
		__be32 p_src_ip = cm->xlate_src_ip;
		__be32 p_dest_ip = cm->xlate_dest_ip;
		__be16 p_src_port = cm->xlate_src_port;
		__be16 p_dest_port = cm->xlate_dest_port;

		partner_cm = NULL;
		if (lan_inner_dev && wan_inner_dev && (cm->xmit_dev == lan_inner_dev)) {
			partner_cm = sfe_ipv4_find_sfe_ipv4_connection_match(si, wan_inner_dev, IPPROTO_TCP, p_src_ip, p_src_port, p_dest_ip, p_dest_port);
		} else if (wan_inner_dev && lan_inner_dev && (cm->xmit_dev == wan_inner_dev)) {
			partner_cm = sfe_ipv4_find_sfe_ipv4_connection_match(si, lan_inner_dev, IPPROTO_TCP, p_src_ip, p_src_port, p_dest_ip, p_dest_port);
		}
		if (partner_cm) {
			cm->partner_match = partner_cm;
			partner_cm->counter_match->partner_match = cm->counter_match;
		}
	} else {
		partner_cm = cm->partner_match;
	}
#endif

	/*
	 * If our packet has beern marked as "flush on find" we can't actually
	 * forward it in the fast path, but now that we've found an associated
	 * connection we can flush that out before we process the packet.
	 */
	if (unlikely(flush_on_find)) {
		struct sfe_ipv4_connection *c = cm->connection;
		sfe_ipv4_remove_sfe_ipv4_connection(si, c);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_IP_OPTIONS_OR_INITIAL_FRAGMENT]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("flush on find\n");
		sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
		return 0;
	}

	/*
	 * Does our TTL allow forwarding?
	 */
	ttl = iph->ttl;
	if (unlikely(ttl < 2)) {
		struct sfe_ipv4_connection *c = cm->connection;
		sfe_ipv4_remove_sfe_ipv4_connection(si, c);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_SMALL_TTL]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("ttl too low\n");
		sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
		return 0;
	}

	/*
	 * If our packet is larger than the MTU of the transmit interface then
	 * we can't forward it easily.
	 */
#ifndef TPLINK_LAN_WAN_NETSTACK
#ifndef TPLINK_FRAGMENT_OUTPUT
	if (unlikely((len > cm->xmit_dev_mtu) && !skb_is_gso(skb))) {
#else
	if (unlikely((len > cm->xmit_dev_mtu) && !skb_is_gso(skb) && (!cm->connection->dst || (frag_off & IP_DF)))) {
#endif
		struct sfe_ipv4_connection *c = cm->connection;
		sfe_ipv4_remove_sfe_ipv4_connection(si, c);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_NEEDS_FRAGMENTATION]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("larger than mtu\n");
		sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
		return 0;
	}
#else /* TPLINK_LAN_WAN_NETSTACK */
#ifndef TPLINK_FRAGMENT_OUTPUT
	if (unlikely(partner_cm && (len > partner_cm->xmit_dev_mtu) && !skb_is_gso(skb))) {
#else
	if (unlikely(partner_cm && (len > partner_cm->xmit_dev_mtu) && !skb_is_gso(skb) && (!partner_cm->connection->dst || (frag_off & IP_DF)))) {
#endif
		partner_cm = NULL;
	}

#ifndef TPLINK_FRAGMENT_OUTPUT
	if (unlikely(!partner_cm && (len > cm->xmit_dev_mtu) && !skb_is_gso(skb))) {
#else
	if (unlikely(!partner_cm && (len > cm->xmit_dev_mtu) && !skb_is_gso(skb) && (!cm->connection->dst || (frag_off & IP_DF)))) {
#endif
		struct sfe_ipv4_connection *c = cm->connection;
		sfe_ipv4_remove_sfe_ipv4_connection(si, c);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_NEEDS_FRAGMENTATION]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("larger than mtu\n");
		sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
		return 0;
	}
#endif /* TPLINK_LAN_WAN_NETSTACK */
	/*
	 * Look at our TCP flags.  Anything missing an ACK or that has RST, SYN or FIN
	 * set is not a fast path packet.
	 */
	if (likely(tcph) && unlikely((flags & (TCP_FLAG_SYN | TCP_FLAG_RST | TCP_FLAG_FIN | TCP_FLAG_ACK)) != TCP_FLAG_ACK)) {
		struct sfe_ipv4_connection *c = cm->connection;
		sfe_ipv4_remove_sfe_ipv4_connection(si, c);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_FLAGS]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("TCP flags: 0x%x are not fast\n",
			    flags & (TCP_FLAG_SYN | TCP_FLAG_RST | TCP_FLAG_FIN | TCP_FLAG_ACK));
		sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
		return 0;
	}

	counter_cm = cm->counter_match;

	/*
	 * Are we doing sequence number checking?
	 */
#ifdef TPLINK_LAN_WAN_NETSTACK
	for (i = 0; i < 2; i++) {
		cm = ((i == 0) ? run_cm : partner_cm);
		if (!cm) {
			break;
		}
#endif

	if (likely(tcph) && likely(!(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_NO_SEQ_CHECK))) {
		uint32_t seq;
		uint32_t ack;
		uint32_t sack;
		uint32_t data_offs;
		uint32_t end;
		uint32_t left_edge;
		uint32_t scaled_win;
		uint32_t max_end;

		/*
		 * Is our sequence fully past the right hand edge of the window?
		 */
		seq = ntohl(tcph->seq);
		if (unlikely((int32_t)(seq - (cm->protocol_state.tcp.max_end + 1)) > 0)) {
			struct sfe_ipv4_connection *c = cm->connection;
#ifdef TPLINK_LAN_WAN_NETSTACK
			if (cm == partner_cm) {
				partner_cm = NULL;
				break;
			}
#endif
			sfe_ipv4_remove_sfe_ipv4_connection(si, c);
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_SEQ_EXCEEDS_RIGHT_EDGE]++;
			si->packets_not_forwarded++;
			spin_unlock_bh(&si->lock);

			DEBUG_TRACE("seq: %u exceeds right edge: %u\n",
				    seq, cm->protocol_state.tcp.max_end + 1);
			sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
			return 0;
		}

		/*
		 * Check that our TCP data offset isn't too short.
		 */
		data_offs = tcph->doff << 2;
		if (unlikely(data_offs < sizeof(struct sfe_ipv4_tcp_hdr))) {
			struct sfe_ipv4_connection *c = cm->connection;
#ifdef TPLINK_LAN_WAN_NETSTACK
			if (cm == partner_cm) {
				partner_cm = NULL;
				break;
			}
#endif

			sfe_ipv4_remove_sfe_ipv4_connection(si, c);
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_SMALL_DATA_OFFS]++;
			si->packets_not_forwarded++;
			spin_unlock_bh(&si->lock);

			DEBUG_TRACE("TCP data offset: %u, too small\n", data_offs);
			sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
			return 0;
		}

		/*
		 * Update ACK according to any SACK option.
		 */
		ack = ntohl(tcph->ack_seq);
		sack = ack;
		if (unlikely(!sfe_ipv4_process_tcp_option_sack(tcph, data_offs, &sack))) {
			struct sfe_ipv4_connection *c = cm->connection;
#ifdef TPLINK_LAN_WAN_NETSTACK
			if (cm == partner_cm) {
				partner_cm = NULL;
				break;
			}
#endif

			sfe_ipv4_remove_sfe_ipv4_connection(si, c);
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_BAD_SACK]++;
			si->packets_not_forwarded++;
			spin_unlock_bh(&si->lock);

			DEBUG_TRACE("TCP option SACK size is wrong\n");
			sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
			return 0;
		}

		/*
		 * Check that our TCP data offset isn't past the end of the packet.
		 */
		data_offs += sizeof(struct sfe_ipv4_ip_hdr);
		if (unlikely(len < data_offs)) {
			struct sfe_ipv4_connection *c = cm->connection;
#ifdef TPLINK_LAN_WAN_NETSTACK
			if (cm == partner_cm) {
				partner_cm = NULL;
				break;
			}
#endif

			sfe_ipv4_remove_sfe_ipv4_connection(si, c);
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_BIG_DATA_OFFS]++;
			si->packets_not_forwarded++;
			spin_unlock_bh(&si->lock);

			DEBUG_TRACE("TCP data offset: %u, past end of packet: %u\n",
				    data_offs, len);
			sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
			return 0;
		}

		end = seq + len - data_offs;

		/*
		 * Is our sequence fully before the left hand edge of the window?
		 */
		if (unlikely((int32_t)(end - (cm->protocol_state.tcp.end
						- cm->counter_match->protocol_state.tcp.max_win - 1)) < 0)) {
			struct sfe_ipv4_connection *c = cm->connection;
#ifdef TPLINK_LAN_WAN_NETSTACK
			if (cm == partner_cm) {
				partner_cm = NULL;
				break;
			}
#endif

			sfe_ipv4_remove_sfe_ipv4_connection(si, c);
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_SEQ_BEFORE_LEFT_EDGE]++;
			si->packets_not_forwarded++;
			spin_unlock_bh(&si->lock);

			DEBUG_TRACE("seq: %u before left edge: %u\n",
				    end, cm->protocol_state.tcp.end - cm->counter_match->protocol_state.tcp.max_win - 1);
			sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
			return 0;
		}

		/*
		 * Are we acking data that is to the right of what has been sent?
		 */
		/*counter++ ;
		if(counter%10000 == 0)
		{
			DEBUG_TRACE("ack: %u ,right edge: %u\n",
				    sack, counter_cm->protocol_state.tcp.end + 1);
		}*/
		if (unlikely((int32_t)(sack - (cm->counter_match->protocol_state.tcp.end + 1)) > 0)) {
			struct sfe_ipv4_connection *c = cm->connection;
#ifdef TPLINK_LAN_WAN_NETSTACK
			if (cm == partner_cm) {
				partner_cm = NULL;
				break;
			}
#endif

			sfe_ipv4_remove_sfe_ipv4_connection(si, c);
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_ACK_EXCEEDS_RIGHT_EDGE]++;
			si->packets_not_forwarded++;
			spin_unlock_bh(&si->lock);

			DEBUG_TRACE("ack: %u exceeds right edge: %u\n",
				    sack, cm->counter_match->protocol_state.tcp.end + 1);
			sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
			return 0;
		}
		//counter_cm->protocol_state.tcp.end = sack + 5;

		/*
		 * Is our ack too far before the left hand edge of the window?
		 */
		left_edge = cm->counter_match->protocol_state.tcp.end
			    - cm->protocol_state.tcp.max_win
			    - SFE_IPV4_TCP_MAX_ACK_WINDOW
			    - 1;
		if (unlikely((int32_t)(sack - left_edge) < 0)) {
			struct sfe_ipv4_connection *c = cm->connection;
#ifdef TPLINK_LAN_WAN_NETSTACK
			if (cm == partner_cm) {
				partner_cm = NULL;
				break;
			}
#endif

			sfe_ipv4_remove_sfe_ipv4_connection(si, c);
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_TCP_ACK_BEFORE_LEFT_EDGE]++;
			si->packets_not_forwarded++;
			spin_unlock_bh(&si->lock);

			DEBUG_TRACE("ack: %u before left edge: %u\n", sack, left_edge);
			sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
			return 0;
		}

		/*
		 * Have we just seen the largest window size yet for this connection?  If yes
		 * then we need to record the new value.
		 */
		scaled_win = ntohs(tcph->window) << cm->protocol_state.tcp.win_scale;
		scaled_win += (sack - ack);
		if (unlikely(cm->protocol_state.tcp.max_win < scaled_win)) {
			cm->protocol_state.tcp.max_win = scaled_win;
		}

		/*
		 * If our sequence and/or ack numbers have advanced then record the new state.
		 */
		if (likely((int32_t)(end - cm->protocol_state.tcp.end) >= 0)) {
			cm->protocol_state.tcp.end = end;
		}

		max_end = sack + scaled_win;
		if (likely((int32_t)(max_end - cm->counter_match->protocol_state.tcp.max_end) >= 0)) {
			cm->counter_match->protocol_state.tcp.max_end = max_end;
		}
	}

#ifdef TPLINK_LAN_WAN_NETSTACK
	}

	cm = (dev_net(run_cm->match_dev) == &init_net) ? run_cm : (partner_cm ? partner_cm : NULL);	
	if (cm) {	// start of cm = (dev_net(run_cm->match_dev) == &init_net) ? run_cm : (partner_cm ? partner_cm : NULL);	
#endif
#ifdef CONFIG_NF_FASTPATH_NAT_HOOK
	fastpath_nat_hook = rcu_dereference(fastpath_nat_forward_hook);
	if (fastpath_nat_enabled && fastpath_nat_hook && cm->connection->nfct) {
		struct nf_conn *ct = (struct nf_conn *)cm->connection->nfct;
		enum ip_conntrack_dir dir = (cm->connection->original_match == cm) ? IP_CT_DIR_ORIGINAL : IP_CT_DIR_REPLY;
		switch (fastpath_nat_hook(skb, ct, dir)) {
			case 0:
				break;
			case -1:
				si->exception_events[SFE_IPV4_EXCEPTION_EVENT_BANDWIDTH_CONTROL]++;
				si->packets_not_forwarded++;
				spin_unlock_bh(&si->lock);
				return 0;
			default:
				si->exception_events[SFE_IPV4_EXCEPTION_EVENT_BANDWIDTH_CONTROL]++;
				si->packets_not_forwarded++;
				spin_unlock_bh(&si->lock);
				kfree_skb(skb);
				return 1;
		}
	}

	qos_fastpath_nat_hook = rcu_dereference(qos_fastpath_nat_forward_hook);
        if (qos_fastpath_nat_enabled && qos_fastpath_nat_hook && cm->connection->nfct) {
                struct nf_conn *ct = (struct nf_conn *)cm->connection->nfct;
                enum ip_conntrack_dir dir = (cm->connection->original_match == cm) ? IP_CT_DIR_ORIGINAL : IP_CT_DIR_REPLY;
                switch (qos_fastpath_nat_hook(skb, ct, dir)) {
                        case 0:
                                break;
                        case -1:
                                si->exception_events[SFE_IPV4_EXCEPTION_EVENT_BANDWIDTH_CONTROL]++;
                                si->packets_not_forwarded++;
                                spin_unlock_bh(&si->lock);
                                return 0;
                        default:
                                si->exception_events[SFE_IPV4_EXCEPTION_EVENT_BANDWIDTH_CONTROL]++;
                                si->packets_not_forwarded++;
                                spin_unlock_bh(&si->lock);
                                kfree_skb(skb);
                                return 1;
                }
        }
#endif

#ifdef SFE_IPSTAT
	sfe_ipstat_hook(cm, skb);
#endif

#ifdef TPLINK_LAN_WAN_NETSTACK
	} // end of cm = (dev_net(run_cm->match_dev) == &init_net) ? run_cm : (partner_cm ? partner_cm : NULL);	
#endif
	/*
	 * From this point on we're good to modify the packet.
	 */

	/*
	 * Update DSCP
	 */
#ifdef TPLINK_LAN_WAN_NETSTACK
		/* TODO: twice xlate src */
	cm = (partner_cm && (partner_cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_DSCP_REMARK)) ? partner_cm : run_cm;
#endif
	if (unlikely(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_DSCP_REMARK)) {
		iph->tos = (iph->tos & SFE_IPV4_DSCP_MASK) | cm->dscp;
	}

	/*
	 * Decrement our TTL.
	 */
	if ((skb->dev != lan_inner_dev) && (partner_cm || (run_cm->xmit_dev != lan_inner_dev))) {
		iph->ttl = ttl - 1;
	}

	/*
	 * Do we have to perform translations of the source address/port?
	 */
#ifdef TPLINK_LAN_WAN_NETSTACK
	/* TODO: twice xlate src */
	cm = (partner_cm && (partner_cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_SRC)) ? partner_cm : run_cm;
#endif
	if (unlikely(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_SRC)) {
		uint16_t tcp_csum;
		uint32_t sum;

		iph->saddr = cm->xlate_src_ip;
		if (likely(tcph)) {
			tcph->source = cm->xlate_src_port;

			/*
			 * Do we have a non-zero UDP checksum?  If we do then we need
			 * to update it.
			 */
			tcp_csum = tcph->check;
			if (unlikely(skb->ip_summed == CHECKSUM_PARTIAL)) {
				sum = tcp_csum + cm->xlate_src_partial_csum_adjustment;
			} else {
				sum = tcp_csum + cm->xlate_src_csum_adjustment;
			}

			sum = (sum & 0xffff) + (sum >> 16);
			tcph->check = (uint16_t)sum;
		}
	}

	/*
	 * Do we have to perform translations of the destination address/port?
	 */
#ifdef TPLINK_LAN_WAN_NETSTACK
	/* TODO: twice xlate dest */
	cm = (partner_cm && (partner_cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_DEST)) ? partner_cm : run_cm;
#endif
	if (unlikely(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_DEST)) {
		uint16_t tcp_csum;
		uint32_t sum;

		iph->daddr = cm->xlate_dest_ip;
		if (likely(tcph)) {
			tcph->dest = cm->xlate_dest_port;

			/*
			 * Do we have a non-zero UDP checksum?  If we do then we need
			 * to update it.
			 */
			tcp_csum = tcph->check;
			if (unlikely(skb->ip_summed == CHECKSUM_PARTIAL)) {
				sum = tcp_csum + cm->xlate_dest_partial_csum_adjustment;
			} else {
				sum = tcp_csum + cm->xlate_dest_csum_adjustment;
			}

			sum = (sum & 0xffff) + (sum >> 16);
			tcph->check = (uint16_t)sum;
		}
	}

	/*
	 * Replace the IP checksum.
	 */
	iph->check = sfe_ipv4_gen_ip_csum(iph);

	/*
	 * Update traffic stats.
	 */
	run_cm->rx_packet_count++;
	run_cm->rx_byte_count += len;
	if (partner_cm) {
		partner_cm->rx_packet_count++;
		partner_cm->rx_byte_count += len;
	}
	/*
	 * If we're not already on the active list then insert ourselves at the tail
	 * of the current list.
	 */
	if (unlikely(!cm->active)) {
		cm->active = true;
		cm->active_prev = si->active_tail;
		if (likely(si->active_tail)) {
			si->active_tail->active_next = cm;
		} else {
			si->active_head = cm;
		}
		si->active_tail = cm;
	}

#if SFE_ACCELERATE_IPSEC
    if (unlikely(!cm->set_accel))
    {
        if (NULL != (x = xfrm_state_lookup_byskb(skb, 0)))
        {
            if ((x->data != NULL) && (x->km.state == XFRM_STATE_VALID)) 
            {
                cm->flow_accel = 1;
                cm->x = x; /* xfrm_state_lookup_byskb has hold x, so no need xfrm_state_put here. */
            }
        }
        
        cm->set_accel = 1;
    }

    if ((cm->set_accel) && (cm->flow_accel) && (cm->x))
    {
        /* now only support esp accelerate */
        if (cm->x->id.proto == IPPROTO_ESP)
        {
            ret = sfe_ipv4_encap_esp(cm->x, skb, iph);
        }
        if (ret == 0)
        {
            if (unlikely(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_DEST)) {
            iph->daddr = dest_ip;
            if (likely(tcph)) {
                tcph->dest = dest_port;
                tcph->check = tcsum;
            }
            iph->check = sfe_ipv4_gen_ip_csum(iph);
            }
        }

        spin_unlock_bh(&si->lock);
        return ret;
    }
#endif

#ifdef TPLINK_LAN_WAN_NETSTACK
	cm = partner_cm?partner_cm:run_cm;
	match_dev = run_cm->match_dev;
#else
	match_dev = cm->match_dev;
#endif
	recv_dev = skb->dev;
	xmit_dev = cm->xmit_dev;
	skb->dev = xmit_dev;

	/*
	 * Check to see if we need to write a header.
	 */
	if (likely(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_WRITE_L2_HDR)) {
		if (unlikely(!(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_WRITE_FAST_ETH_HDR))) {
			dev_hard_header(skb, xmit_dev, ETH_P_IP,
					cm->xmit_dest_mac, cm->xmit_src_mac, len);
		} else {
			/*
			 * For the simple case we write this really fast.
			 */
			struct sfe_ipv4_eth_hdr *eth = (struct sfe_ipv4_eth_hdr *)__skb_push(skb, ETH_HLEN);
			eth->h_proto = htons(ETH_P_IP);
			eth->h_dest[0] = cm->xmit_dest_mac[0];
			eth->h_dest[1] = cm->xmit_dest_mac[1];
			eth->h_dest[2] = cm->xmit_dest_mac[2];
			eth->h_source[0] = cm->xmit_src_mac[0];
			eth->h_source[1] = cm->xmit_src_mac[1];
			eth->h_source[2] = cm->xmit_src_mac[2];
		}
#ifdef TPLINK_FRAGMENT_OUTPUT
		l2_header = true;
#endif
	}

	/*
	 * Remove vlan info for packets fast-forwarding. Or packets would be blocked by vlan filter.
	 */
	skb->vlan_proto = 0;
	skb->vlan_tci = 0;

	/*
	 * Update priority of skb.
	 */
#ifdef TPLINK_LAN_WAN_NETSTACK
	cm = (partner_cm && (partner_cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_PRIORITY_REMARK)) ? partner_cm : run_cm;
#endif
	if (unlikely(cm->flags & SFE_IPV4_CONNECTION_MATCH_FLAG_PRIORITY_REMARK)) {
		skb->priority = cm->priority;
	}

	/*
	 * Mark outgoing packet
	 */
#ifdef TPLINK_LAN_WAN_NETSTACK
	cm = partner_cm?partner_cm:run_cm;
#endif
	skb->mark = ((struct nf_conn *)run_cm->connection->nfct)->mark;
	skb->mark |= partner_cm?((struct nf_conn *)partner_cm->connection->nfct)->mark:0;
	if (skb->mark) {
		//DEBUG_TRACE("SKB MARK is NON ZERO %x\n", skb->mark);
	}
#if defined(CONFIG_NF_CONNTRACK_EXTMARK)
	skb->extmark = ((struct nf_conn *)run_cm->connection->nfct)->extmark;
	skb->extmark |= partner_cm?((struct nf_conn *)partner_cm->connection->nfct)->extmark:0;
	if (skb->extmark) {
		DEBUG_TRACE("SKB EXTMARK is NON ZERO %x\n", skb->extmark);
	}
#endif

	if ((SKB_NFCT_GET(skb) == NULL) && atomic_inc_not_zero(&cm->connection->nfct->use)) {
		SKB_NFCT_SET(skb, (struct nf_conn *)cm->connection->nfct, 
		    (cm->connection->original_match == cm) ? IP_CT_ESTABLISHED : IP_CT_ESTABLISHED_REPLY);
	}

	si->packets_forwarded++;

#ifdef TPLINK_FRAGMENT_OUTPUT
	if (unlikely((len > cm->xmit_dev_mtu) && !skb_is_gso(skb))) {
		need_frag = true;
		sfe_skb_dst_set(skb, iph, cm->connection);
	}
#endif
	spin_unlock_bh(&si->lock);

	/*
	 * We're going to check for GSO flags when we transmit the packet so
	 * start fetching the necessary cache line now.
	 */
	prefetch(skb_shinfo(skb));

	/*
	 * Mark that this packet has been fast forwarded.
	 */
	skb->fast_forwarded = 1;
	// DEBUG_TRACE("fast forward packet - dev: %s \n"
	// 		"  src_ip:%pI4,src_port:%u\n"
	// 		"  dest_ip:%pI4,dest_port:%u\n",
	// 		skb->dev->name,&iph->saddr,src_port,&iph->daddr,dest_port);
	if (!dev || (skb->dev && (skb->dev->flags & IFF_POINTOPOINT))) {
		hw_nat_slow_path(skb);
	}
#ifdef TPLINK_PPP_TUNNEL
	if (match_dev && (sfe_proto > SFE_PROTO_DEF && sfe_proto < SFE_PROTO_IPSEC ) && (match_dev->flags & IFF_POINTOPOINT) && (!dev_tuntype(match_dev))) {
		ppp_dev_add_stats(match_dev, skb, len, 1);
	}
#endif

#ifdef TPLINK_FRAGMENT_OUTPUT
	if (need_frag)
		sfe_ipv4_fragment(skb, l2_header);
	else
#endif
#ifdef TPLINK_PPP_TUNNEL
	sfe_ipv4_dev_queue_xmit(skb);
#else
#ifdef TPLINK_ACCELERATE_NAT
		if (skb->shortcut_data) {
			sfe_hard_xmit_func_t xmit = (sfe_hard_xmit_func_t)skb->shortcut_data;
			xmit(skb);
			skb->shortcut_data = 0;
		}
		else {
			dev_queue_xmit(skb);
		}
#else
		dev_queue_xmit(skb);
#endif
#endif
	return 1;
}

/*
 * sfe_ipv4_recv_icmp()
 *	Handle ICMP packet receives.
 *
 * ICMP packets aren't handled as a "fast path" and always have us process them
 * through the default Linux stack.  What we do need to do is look for any errors
 * about connections we are handling in the fast path.  If we find any such
 * connections then we want to flush their state so that the ICMP error path
 * within Linux has all of the correct state should it need it.
 */
static int sfe_ipv4_recv_icmp(struct sfe_ipv4 *si, struct sk_buff *skb, struct net_device *dev,
			      unsigned int len, struct sfe_ipv4_ip_hdr *iph, unsigned int ihl)
{
	struct icmphdr *icmph;
	struct sfe_ipv4_ip_hdr *icmp_iph;
	unsigned int icmp_ihl_words;
	unsigned int icmp_ihl;
	uint32_t *icmp_trans_h;
	struct sfe_ipv4_udp_hdr *icmp_udph;
	struct sfe_ipv4_tcp_hdr *icmp_tcph;
	__be32 src_ip;
	__be32 dest_ip;
	__be16 src_port;
	__be16 dest_port;
	struct sfe_ipv4_connection_match *cm;
	struct sfe_ipv4_connection *c;
	uint32_t pull_len = sizeof(struct icmphdr) + ihl;

	/*
	 * Is our packet too short to contain a valid UDP header?
	 */
	len -= ihl;
	if (!pskb_may_pull(skb, pull_len)) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_ICMP_HEADER_INCOMPLETE]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("packet too short for ICMP header\n");
		return 0;
	}

	/*
	 * We only handle "destination unreachable" and "time exceeded" messages.
	 */
	icmph = (struct icmphdr *)(skb->data + ihl);
	if ((icmph->type != ICMP_DEST_UNREACH)
	    && (icmph->type != ICMP_TIME_EXCEEDED)) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_ICMP_UNHANDLED_TYPE]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("unhandled ICMP type: 0x%x\n", icmph->type);
		return 0;
	}

	/*
	 * Do we have the full embedded IP header?
	 */
	len -= sizeof(struct icmphdr);
	pull_len += sizeof(struct sfe_ipv4_ip_hdr);
	if (!pskb_may_pull(skb, pull_len)) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_ICMP_IPV4_HEADER_INCOMPLETE]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("Embedded IP header not complete\n");
		return 0;
	}

	/*
	 * Is our embedded IP version wrong?
	 */
	icmp_iph = (struct sfe_ipv4_ip_hdr *)(icmph + 1);
	if (unlikely(icmp_iph->version != 4)) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_ICMP_IPV4_NON_V4]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("IP version: %u\n", icmp_iph->version);
		return 0;
	}

	/*
	 * Do we have the full embedded IP header, including any options?
	 */
	icmp_ihl_words = icmp_iph->ihl;
	icmp_ihl = icmp_ihl_words << 2;
	pull_len += icmp_ihl - sizeof(struct sfe_ipv4_ip_hdr);
	if (!pskb_may_pull(skb, pull_len)) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_ICMP_IPV4_IP_OPTIONS_INCOMPLETE]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("Embedded header not large enough for IP options\n");
		return 0;
	}

	len -= icmp_ihl;
	icmp_trans_h = ((uint32_t *)icmp_iph) + icmp_ihl_words;

	/*
	 * Handle the embedded transport layer header.
	 */
	switch (icmp_iph->protocol) {
	case IPPROTO_UDP:
		/*
		 * We should have 8 bytes of UDP header - that's enough to identify
		 * the connection.
		 */
		pull_len += 8;
		if (!pskb_may_pull(skb, pull_len)) {
			spin_lock_bh(&si->lock);
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_ICMP_IPV4_UDP_HEADER_INCOMPLETE]++;
			si->packets_not_forwarded++;
			spin_unlock_bh(&si->lock);

			DEBUG_TRACE("Incomplete embedded UDP header\n");
			return 0;
		}

		icmp_udph = (struct sfe_ipv4_udp_hdr *)icmp_trans_h;
		src_port = icmp_udph->source;
		dest_port = icmp_udph->dest;
		break;

	case IPPROTO_TCP:
		/*
		 * We should have 8 bytes of TCP header - that's enough to identify
		 * the connection.
		 */
		pull_len += 8;
		if (!pskb_may_pull(skb, pull_len)) {
			spin_lock_bh(&si->lock);
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_ICMP_IPV4_TCP_HEADER_INCOMPLETE]++;
			si->packets_not_forwarded++;
			spin_unlock_bh(&si->lock);

			DEBUG_TRACE("Incomplete embedded TCP header\n");
			return 0;
		}

		icmp_tcph = (struct sfe_ipv4_tcp_hdr *)icmp_trans_h;
		src_port = icmp_tcph->source;
		dest_port = icmp_tcph->dest;
		break;

	default:
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_ICMP_IPV4_UNHANDLED_PROTOCOL]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("Unhandled embedded IP protocol: %u\n", icmp_iph->protocol);
		return 0;
	}

	src_ip = icmp_iph->saddr;
	dest_ip = icmp_iph->daddr;

	spin_lock_bh(&si->lock);

	/*
	 * Look for a connection match.  Note that we reverse the source and destination
	 * here because our embedded message contains a packet that was sent in the
	 * opposite direction to the one in which we just received it.  It will have
	 * been sent on the interface from which we received it though so that's still
	 * ok to use.
	 */
	cm = sfe_ipv4_find_sfe_ipv4_connection_match(si, skb->dev, icmp_iph->protocol, dest_ip, dest_port, src_ip, src_port);
	if (unlikely(!cm)) {
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_ICMP_NO_CONNECTION]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		//DEBUG_TRACE("no connection found\n");
		return 0;
	}

	/*
	 * We found a connection so now remove it from the connection list and flush
	 * its state.
	 */
	c = cm->connection;
	sfe_ipv4_remove_sfe_ipv4_connection(si, c);
	si->exception_events[SFE_IPV4_EXCEPTION_EVENT_ICMP_FLUSHED_CONNECTION]++;
	si->packets_not_forwarded++;
	spin_unlock_bh(&si->lock);
	DEBUG_TRACE("Destination is unreachable. Flush this rules\n");
	sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_FLUSH);
	return 0;
}

/*
 * sfe_ipv4_recv()
 *	Handle packet receives and forwaring.
 *
 * Returns 1 if the packet is forwarded or 0 if it isn't.
 */
#ifdef TPLINK_PPP_TUNNEL
static int sfe_ipv4_recv_core(struct net_device *dev, struct sk_buff *skb, int ppp_header_level, int sfe_proto)
#else
static int sfe_ipv4_recv_core(struct net_device *dev, struct sk_buff *skb, int sfe_proto)
#endif
{
	struct sfe_ipv4 *si = &__si;
	unsigned int len;
	unsigned int tot_len;
#ifndef TPLINK_FRAGMENT
	unsigned int frag_off;
#endif
	unsigned int ihl;
	bool flush_on_find = false;
	bool ip_options;
	struct sfe_ipv4_ip_hdr *iph;
	uint32_t protocol;
	int ret = 0;
	flush_on_find = false;

	/*
	 * Check that we have space for an IP header here.
	 */
	len = skb->len;
	if (unlikely(!pskb_may_pull(skb, sizeof(struct sfe_ipv4_ip_hdr)))) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_HEADER_INCOMPLETE]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("len: %u is too short\n", len);
		return 0;
	}

	/*
	 * Check that our "total length" is large enough for an IP header.
	 */
	iph = (struct sfe_ipv4_ip_hdr *)skb->data;
	tot_len = ntohs(iph->tot_len);
	if (unlikely(tot_len < sizeof(struct sfe_ipv4_ip_hdr))) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_BAD_TOTAL_LENGTH]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("tot_len: %u is too short\n", tot_len);
		return 0;
	}
	//if (!dev){
	//	int i = 0; 
	//	for (i = 0; i < 20; i++) {
	//		printk(KERN_ERR"%02x ", skb->data[i]);
	//	}
	//	printk(KERN_ERR"\n");
	//	printk(KERN_ERR"iph->tot_len:%u, skb->len:%u\n", ntohs(iph->tot_len), skb->len);
	//}

	/*
	 * Is our IP version wrong?
	 */
	if (unlikely(iph->version != 4)) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_NON_V4]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("IP version: %u\n", iph->version);
		return 0;
	}

	/*
	 * Does our datagram fit inside the skb?
	 */
	if (unlikely(tot_len > len)) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_DATAGRAM_INCOMPLETE]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("tot_len: %u, exceeds len: %u\n", tot_len, len);
		return 0;
	}

#ifndef TPLINK_FRAGMENT
	/*
	 * Do we have a non-initial fragment?
	 */
	frag_off = ntohs(iph->frag_off);
	if (unlikely(frag_off & IP_OFFSET)) {
		spin_lock_bh(&si->lock);
		si->exception_events[SFE_IPV4_EXCEPTION_EVENT_NON_INITIAL_FRAGMENT]++;
		si->packets_not_forwarded++;
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("non-initial fragment\n");
		return 0;
	}

	/*
	 * If we have a (first) fragment then mark it to cause any connection to flush.
	 */
	flush_on_find = unlikely(frag_off & IP_MF) ? true : false;
#endif
	/*
	 * Do we have any IP options?  That's definite a slow path!  If we do have IP
	 * options we need to recheck our header size.
	 */
	ihl = iph->ihl << 2;
	ip_options = unlikely(ihl != sizeof(struct sfe_ipv4_ip_hdr)) ? true : false;
	if (unlikely(ip_options)) {
		if (unlikely(len < ihl)) {
			spin_lock_bh(&si->lock);
			si->exception_events[SFE_IPV4_EXCEPTION_EVENT_IP_OPTIONS_INCOMPLETE]++;
			si->packets_not_forwarded++;
			spin_unlock_bh(&si->lock);

			DEBUG_TRACE("len: %u is too short for header of size: %u\n", len, ihl);
			return 0;
		}

		flush_on_find = true;
	}

	protocol = iph->protocol;
#if SFE_ACCELERATE_IPSEC
    if (IPPROTO_ESP == protocol) {
        if ((!((dev == NULL) && (ppp_header_level >= MAX_ESP_HEADER)) )
             && (!(ntohs(iph->frag_off) & (IP_OFFSET|IP_MF))))
        {
            return  sfe_ipv4_recv_esp(si, skb, iph, false,sfe_proto);
        }
        else
        {
			DEBUG_TRACE("not accelerate esp ,ppp_header_level %d\n",ppp_header_level);
            return 0;
        }
    }
#endif

	if (IPPROTO_UDP == protocol) {
		ret = sfe_ipv4_recv_udp(si, skb, dev, len, iph, ihl, flush_on_find, sfe_proto);
		if (ret) {
			return ret;
		}
		DEBUG_TRACE("dev:%s ppp_header_level:%d\n", dev ? dev->name : "NULL", ppp_header_level);
#ifdef TPLINK_PPP_TUNNEL
		if (dev && (ppp_header_level < MAX_PPP_HEADER) && 
            (!(dev->flags & IFF_POINTOPOINT) || skb->sp)) {
			ret = sfe_ipv4_recv_l2tp(si, skb, dev, len, iph, ihl, flush_on_find, ppp_header_level);
			if (ret) {
				return ret;
			}
		}
#endif
		return ret;

	}

	if (IPPROTO_TCP == protocol) {
		ret = sfe_ipv4_recv_tcp(si, skb, dev, len, iph, ihl, flush_on_find, sfe_proto);
		return ret;
	}

	if (IPPROTO_ICMP == protocol) {
		ret = sfe_ipv4_recv_icmp(si, skb, dev, len, iph, ihl);
		return ret;
	}
#ifdef TPLINK_PPP_TUNNEL
	if (dev && !(dev->flags & IFF_POINTOPOINT) && (ppp_header_level < MAX_PPP_HEADER) && (IPPROTO_GRE == protocol)) {
		ret = sfe_ipv4_recv_pptp(si, skb, dev, len, iph, ihl, flush_on_find, ppp_header_level);
		return ret;
	}
#endif
	spin_lock_bh(&si->lock);
	si->exception_events[SFE_IPV4_EXCEPTION_EVENT_UNHANDLED_PROTOCOL]++;
	si->packets_not_forwarded++;
	spin_unlock_bh(&si->lock);

	DEBUG_TRACE("not UDP, TCP or ICMP: %u\n", protocol);
	return 0;
}

int sfe_ipv4_recv(struct net_device *dev, struct sk_buff *skb)
{
#ifdef TPLINK_PPP_TUNNEL
	return sfe_ipv4_recv_core(dev, skb, 0, SFE_PROTO_DEF);
#else
	return sfe_ipv4_recv_core(dev, skb, SFE_PROTO_DEF);
#endif
}

static void
sfe_ipv4_update_tcp_state(struct sfe_ipv4_connection *c,
			  struct sfe_connection_create *sic)
{
	struct sfe_ipv4_connection_match *orig_cm;
	struct sfe_ipv4_connection_match *repl_cm;
	struct sfe_ipv4_tcp_connection_match *orig_tcp;
	struct sfe_ipv4_tcp_connection_match *repl_tcp;

	orig_cm = c->original_match;
	repl_cm = c->reply_match;
	orig_tcp = &orig_cm->protocol_state.tcp;
	repl_tcp = &repl_cm->protocol_state.tcp;

	/* update orig */
	if (orig_tcp->max_win < sic->src_td_max_window) {
		orig_tcp->max_win = sic->src_td_max_window;
	}
	if ((int32_t)(orig_tcp->end - sic->src_td_end) < 0) {
		orig_tcp->end = sic->src_td_end;
	}
	if ((int32_t)(orig_tcp->max_end - sic->src_td_max_end) < 0) {
		orig_tcp->max_end = sic->src_td_max_end;
	}

	/* update reply */
	if (repl_tcp->max_win < sic->dest_td_max_window) {
		repl_tcp->max_win = sic->dest_td_max_window;
	}
	if ((int32_t)(repl_tcp->end - sic->dest_td_end) < 0) {
		repl_tcp->end = sic->dest_td_end;
	}
	if ((int32_t)(repl_tcp->max_end - sic->dest_td_max_end) < 0) {
		repl_tcp->max_end = sic->dest_td_max_end;
	}

	/* update match flags */
	orig_cm->flags &= ~SFE_IPV4_CONNECTION_MATCH_FLAG_NO_SEQ_CHECK;
	repl_cm->flags &= ~SFE_IPV4_CONNECTION_MATCH_FLAG_NO_SEQ_CHECK;
	if (sic->flags & SFE_CREATE_FLAG_NO_SEQ_CHECK) {
		orig_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_NO_SEQ_CHECK;
		repl_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_NO_SEQ_CHECK;
	}
}

static void
sfe_ipv4_update_protocol_state(struct sfe_ipv4_connection *c,
			       struct sfe_connection_create *sic)
{
	switch (sic->protocol) {
	case IPPROTO_TCP:
		sfe_ipv4_update_tcp_state(c, sic);
		break;
	}
}

void sfe_ipv4_update_rule(struct sfe_connection_create *sic)
{
	struct sfe_ipv4_connection *c;
	struct sfe_ipv4 *si = &__si;

	spin_lock_bh(&si->lock);

	c = sfe_ipv4_find_sfe_ipv4_connection(si, dev_net(sic->src_dev),
					      sic->protocol,
					      sic->src_ip.ip,
					      sic->src_port,
					      sic->dest_ip.ip,
					      sic->dest_port);
	if (c != NULL) {
		sfe_ipv4_update_protocol_state(c, sic);
	}

	spin_unlock_bh(&si->lock);
}

/*
 * sfe_ipv4_create_rule()
 *	Create a forwarding rule.
 */
int sfe_ipv4_create_rule(struct sfe_connection_create *sic)
{
	struct sfe_ipv4 *si = &__si;
	struct sfe_ipv4_connection *c;
	struct sfe_ipv4_connection_match *original_cm;
	struct sfe_ipv4_connection_match *reply_cm;
	struct net_device *dest_dev;
	struct net_device *src_dev;

	dest_dev = sic->dest_dev;
	src_dev = sic->src_dev;

	if (unlikely((dest_dev->reg_state != NETREG_REGISTERED) ||
		     (src_dev->reg_state != NETREG_REGISTERED))) {
		return -EINVAL;
	}

	spin_lock_bh(&si->lock);
	si->connection_create_requests++;

	/*
	 * Check to see if there is already a flow that matches the rule we're
	 * trying to create.  If there is then we can't create a new one.
	 */
	c = sfe_ipv4_find_sfe_ipv4_connection(si, dev_net(src_dev),
					      sic->protocol,
					      sic->src_ip.ip,
					      sic->src_port,
					      sic->dest_ip.ip,
					      sic->dest_port);
	if (c != NULL) {
		si->connection_create_collisions++;

		/*
		 * If we already have the flow then it's likely that this
		 * request to create the connection rule contains more
		 * up-to-date information. Check and update accordingly.
		 */
		sfe_ipv4_update_protocol_state(c, sic);
		spin_unlock_bh(&si->lock);

		DEBUG_TRACE("connection already exists - mark: %08x, p: %d\n"
			    "  s: %s:%pM:%pI4:%u, d: %s:%pM:%pI4:%u\n",
			    sic->mark, sic->protocol,
			    sic->src_dev->name, sic->src_mac, &sic->src_ip.ip, ntohs(sic->src_port),
			    sic->dest_dev->name, sic->dest_mac, &sic->dest_ip.ip, ntohs(sic->dest_port));
		return -EADDRINUSE;
	}

	/*
	 * Allocate the various connection tracking objects.
	 */
	c = (struct sfe_ipv4_connection *)kmalloc(sizeof(struct sfe_ipv4_connection), GFP_ATOMIC);
	if (unlikely(!c)) {
		spin_unlock_bh(&si->lock);
		return -ENOMEM;
	}

	original_cm = (struct sfe_ipv4_connection_match *)kmalloc(sizeof(struct sfe_ipv4_connection_match), GFP_ATOMIC);
	if (unlikely(!original_cm)) {
		spin_unlock_bh(&si->lock);
		kfree(c);
		return -ENOMEM;
	}

	reply_cm = (struct sfe_ipv4_connection_match *)kmalloc(sizeof(struct sfe_ipv4_connection_match), GFP_ATOMIC);
	if (unlikely(!reply_cm)) {
		spin_unlock_bh(&si->lock);
		kfree(original_cm);
		kfree(c);
		return -ENOMEM;
	}

	/*
	 * Fill in the "original" direction connection matching object.
	 * Note that the transmit MAC address is "dest_mac_xlate" because
	 * we always know both ends of a connection by their translated
	 * addresses and not their public addresses.
	 */
	original_cm->match_dev = src_dev;
	original_cm->match_protocol = sic->protocol;
	original_cm->match_src_ip = sic->src_ip.ip;
	original_cm->match_src_port = sic->src_port;
	original_cm->match_dest_ip = sic->dest_ip.ip;
	original_cm->match_dest_port = sic->dest_port;
	original_cm->xlate_src_ip = sic->src_ip_xlate.ip;
	original_cm->xlate_src_port = sic->src_port_xlate;
	original_cm->xlate_dest_ip = sic->dest_ip_xlate.ip;
	original_cm->xlate_dest_port = sic->dest_port_xlate;
	original_cm->rx_packet_count = 0;
	original_cm->rx_packet_count64 = 0;
	original_cm->rx_byte_count = 0;
	original_cm->rx_byte_count64 = 0;
	original_cm->xmit_dev = dest_dev;
	original_cm->xmit_dev_mtu = sic->dest_mtu;
	memcpy(original_cm->xmit_src_mac, dest_dev->dev_addr, ETH_ALEN);
	memcpy(original_cm->xmit_dest_mac, sic->dest_mac_xlate, ETH_ALEN);
	original_cm->connection = c;
	original_cm->counter_match = reply_cm;
	original_cm->flags = 0;
	if (sic->flags & SFE_CREATE_FLAG_REMARK_PRIORITY) {
		original_cm->priority = sic->src_priority;
		original_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_PRIORITY_REMARK;
	}
	if (sic->flags & SFE_CREATE_FLAG_REMARK_DSCP) {
		original_cm->dscp = sic->src_dscp << SFE_IPV4_DSCP_SHIFT;
		original_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_DSCP_REMARK;
	}
#ifdef CONFIG_NF_FLOW_COOKIE
	original_cm->flow_cookie = 0;
#endif
#ifdef CONFIG_XFRM
	original_cm->set_accel = 0;
	original_cm->flow_accel = 0;
    original_cm->x = NULL;
#endif
	original_cm->active_next = NULL;
	original_cm->active_prev = NULL;
	original_cm->active = false;
#ifdef TPLINK_LAN_WAN_NETSTACK
	original_cm->partner_match = NULL;
#endif

	/*
	 * For PPP links we don't write an L2 header.  For everything else we do.
	 */
	if (!(dest_dev->flags & IFF_POINTOPOINT)) {
		original_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_WRITE_L2_HDR;

		/*
		 * If our dev writes Ethernet headers then we can write a really fast
		 * version.
		 */
		if (dest_dev->header_ops) {
			if (dest_dev->header_ops->create == eth_header) {
				original_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_WRITE_FAST_ETH_HDR;
			}
		}
	}

	/*
	 * Fill in the "reply" direction connection matching object.
	 */
	reply_cm->match_dev = dest_dev;
	reply_cm->match_protocol = sic->protocol;
	reply_cm->match_src_ip = sic->dest_ip_xlate.ip;
	reply_cm->match_src_port = sic->dest_port_xlate;
	reply_cm->match_dest_ip = sic->src_ip_xlate.ip;
	reply_cm->match_dest_port = sic->src_port_xlate;
	reply_cm->xlate_src_ip = sic->dest_ip.ip;
	reply_cm->xlate_src_port = sic->dest_port;
	reply_cm->xlate_dest_ip = sic->src_ip.ip;
	reply_cm->xlate_dest_port = sic->src_port;
	reply_cm->rx_packet_count = 0;
	reply_cm->rx_packet_count64 = 0;
	reply_cm->rx_byte_count = 0;
	reply_cm->rx_byte_count64 = 0;
	reply_cm->xmit_dev = src_dev;
	reply_cm->xmit_dev_mtu = sic->src_mtu;
	memcpy(reply_cm->xmit_src_mac, src_dev->dev_addr, ETH_ALEN);
	memcpy(reply_cm->xmit_dest_mac, sic->src_mac, ETH_ALEN);
	reply_cm->connection = c;
	reply_cm->counter_match = original_cm;
	reply_cm->flags = 0;
	if (sic->flags & SFE_CREATE_FLAG_REMARK_PRIORITY) {
		reply_cm->priority = sic->dest_priority;
		reply_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_PRIORITY_REMARK;
	}
	if (sic->flags & SFE_CREATE_FLAG_REMARK_DSCP) {
		reply_cm->dscp = sic->dest_dscp << SFE_IPV4_DSCP_SHIFT;
		reply_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_DSCP_REMARK;
	}
#ifdef CONFIG_NF_FLOW_COOKIE
	reply_cm->flow_cookie = 0;
#endif
#ifdef CONFIG_XFRM
	reply_cm->set_accel = 0;
	reply_cm->flow_accel = 0;
	reply_cm->x = NULL;
#endif
	reply_cm->active_next = NULL;
	reply_cm->active_prev = NULL;
	reply_cm->active = false;
#ifdef TPLINK_LAN_WAN_NETSTACK
	reply_cm->partner_match = NULL;
#endif

	/*
	 * For PPP links we don't write an L2 header.  For everything else we do.
	 */
	if (!(src_dev->flags & IFF_POINTOPOINT)) {
		reply_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_WRITE_L2_HDR;

		/*
		 * If our dev writes Ethernet headers then we can write a really fast
		 * version.
		 */
		if (src_dev->header_ops) {
			if (src_dev->header_ops->create == eth_header) {
				reply_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_WRITE_FAST_ETH_HDR;
			}
		}
	}


	if (sic->dest_ip.ip != sic->dest_ip_xlate.ip || sic->dest_port != sic->dest_port_xlate) {
		original_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_DEST;
		reply_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_SRC;
	}

	if (sic->src_ip.ip != sic->src_ip_xlate.ip || sic->src_port != sic->src_port_xlate) {
		original_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_SRC;
		reply_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_XLATE_DEST;
	}

	c->protocol = sic->protocol;
	c->src_ip = sic->src_ip.ip;
	c->src_ip_xlate = sic->src_ip_xlate.ip;
	c->src_port = sic->src_port;
	c->src_port_xlate = sic->src_port_xlate;
	c->original_dev = src_dev;
	c->original_match = original_cm;
	c->dest_ip = sic->dest_ip.ip;
	c->dest_ip_xlate = sic->dest_ip_xlate.ip;
	c->dest_port = sic->dest_port;
	c->dest_port_xlate = sic->dest_port_xlate;
	c->reply_dev = dest_dev;
	c->reply_match = reply_cm;
	c->mark = sic->mark;
#if defined(CONFIG_NF_CONNTRACK_EXTMARK)
	c->extmark = sic->extmark;
#endif
	c->nfct = sic->nfct;
	nf_conntrack_get(c->nfct);
	c->debug_read_seq = 0;
	c->last_sync_jiffies = get_jiffies_64();
#ifdef TPLINK_FRAGMENT_OUTPUT
	if (sic->dst) {
		dst_hold(sic->dst);
		c->dst = sic->dst;
	}
	if (sic->reply_dst) {
		dst_hold(sic->reply_dst);
		c->reply_dst = sic->reply_dst;
	}
#endif
	c->sfe_proto = 0;
	c->slowpath_jiffies = 0;

	/*
	 * Take hold of our source and dest devices for the duration of the connection.
	 */
	dev_hold(c->original_dev);
	dev_hold(c->reply_dev);

	/*
	 * Initialize the protocol-specific information that we track.
	 */
	switch (sic->protocol) {
	case IPPROTO_TCP:
		original_cm->protocol_state.tcp.win_scale = sic->src_td_window_scale;
		original_cm->protocol_state.tcp.max_win = sic->src_td_max_window ? sic->src_td_max_window : 1;
		original_cm->protocol_state.tcp.end = sic->src_td_end;
		original_cm->protocol_state.tcp.max_end = sic->src_td_max_end;
		reply_cm->protocol_state.tcp.win_scale = sic->dest_td_window_scale;
		reply_cm->protocol_state.tcp.max_win = sic->dest_td_max_window ? sic->dest_td_max_window : 1;
		reply_cm->protocol_state.tcp.end = sic->dest_td_end;
		reply_cm->protocol_state.tcp.max_end = sic->dest_td_max_end;
		if (sic->flags & SFE_CREATE_FLAG_NO_SEQ_CHECK) {
			original_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_NO_SEQ_CHECK;
			reply_cm->flags |= SFE_IPV4_CONNECTION_MATCH_FLAG_NO_SEQ_CHECK;
		}
		break;
	}

	sfe_ipv4_connection_match_compute_translations(original_cm);
	sfe_ipv4_connection_match_compute_translations(reply_cm);
	sfe_ipv4_insert_sfe_ipv4_connection(si, c);

	spin_unlock_bh(&si->lock);

	/*
	 * We have everything we need!
	 */
	DEBUG_INFO("new connection - mark: %08x, p: %d\n"
		   "  s: %s:%pM(%pM):%pI4(%pI4):%u(%u)\n"
		   "  d: %s:%pM(%pM):%pI4(%pI4):%u(%u)\n",
		   sic->mark, sic->protocol,
		   sic->src_dev->name, sic->src_mac, sic->src_mac_xlate,
		   &sic->src_ip.ip, &sic->src_ip_xlate.ip, ntohs(sic->src_port), ntohs(sic->src_port_xlate),
		   dest_dev->name, sic->dest_mac, sic->dest_mac_xlate,
		   &sic->dest_ip.ip, &sic->dest_ip_xlate.ip, ntohs(sic->dest_port), ntohs(sic->dest_port_xlate));

	return 0;
}

/*
 * sfe_ipv4_destroy_rule()
 *	Destroy a forwarding rule.
 */
void sfe_ipv4_destroy_rule(struct sfe_connection_destroy *sid)
{
	struct sfe_ipv4 *si = &__si;
	struct sfe_ipv4_connection *c;

	spin_lock_bh(&si->lock);
	si->connection_destroy_requests++;

	/*
	 * Check to see if we have a flow that matches the rule we're trying
	 * to destroy.  If there isn't then we can't destroy it.
	 */
	c = sfe_ipv4_find_sfe_ipv4_connection(si, sid->net, sid->protocol, sid->src_ip.ip, sid->src_port,
					      sid->dest_ip.ip, sid->dest_port);
	if (!c) {
		si->connection_destroy_misses++;
		spin_unlock_bh(&si->lock);

		//DEBUG_TRACE("connection does not exist - p: %d, s: %pI4:%u, d: %pI4:%u\n",
			    //sid->protocol, &sid->src_ip, ntohs(sid->src_port),
			    //&sid->dest_ip, ntohs(sid->dest_port));
		return;
	}

	/*
	 * Remove our connection details from the hash tables.
	 */
	sfe_ipv4_remove_sfe_ipv4_connection(si, c);
	spin_unlock_bh(&si->lock);

	sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_DESTROY);

	DEBUG_INFO("connection destroyed - p: %d, s: %pI4:%u, d: %pI4:%u\n",
		   sid->protocol, &sid->src_ip.ip, ntohs(sid->src_port),
		   &sid->dest_ip.ip, ntohs(sid->dest_port));
}

/*
 * sfe_ipv4_register_sync_rule_callback()
 *	Register a callback for rule synchronization.
 */
void sfe_ipv4_register_sync_rule_callback(sfe_sync_rule_callback_t sync_rule_callback)
{
	struct sfe_ipv4 *si = &__si;

	spin_lock_bh(&si->lock);
	rcu_assign_pointer(si->sync_rule_callback, sync_rule_callback);
	spin_unlock_bh(&si->lock);
}

/*
 * sfe_ipv4_get_debug_dev()
 */
static ssize_t sfe_ipv4_get_debug_dev(struct device *dev,
				      struct device_attribute *attr,
				      char *buf)
{
	struct sfe_ipv4 *si = &__si;
	ssize_t count;
	int num;

	spin_lock_bh(&si->lock);
	num = si->debug_dev;
	spin_unlock_bh(&si->lock);

	count = snprintf(buf, (ssize_t)PAGE_SIZE, "%d\n", num);
	return count;
}

/*
 * sysfs attributes.
 */
static const struct device_attribute sfe_ipv4_debug_dev_attr =
	__ATTR(debug_dev, S_IWUSR | S_IRUGO, sfe_ipv4_get_debug_dev, NULL);

/*
 * sfe_ipv4_destroy_all_rules_for_dev()
 *	Destroy all connections that match a particular device.
 *
 * If we pass dev as NULL then this destroys all connections.
 */
void sfe_ipv4_destroy_all_rules_for_dev(struct net_device *dev)
{
	struct sfe_ipv4 *si = &__si;
	struct sfe_ipv4_connection *c;

another_round:
	spin_lock_bh(&si->lock);

	for (c = si->all_connections_head; c; c = c->all_connections_next) {
		/*
		 * Does this connection relate to the device we are destroying?
		 */
		if (!dev
		    || (dev == c->original_dev)
		    || (dev == c->reply_dev)) {
			break;
		}
	}

	if (c) {
		sfe_ipv4_remove_sfe_ipv4_connection(si, c);
	}

	spin_unlock_bh(&si->lock);
	DEBUG_INFO("Connection destroyed all\n");
	if (c) {
		sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_DESTROY);
		goto another_round;
	}
#ifdef TPLINK_PPP_TUNNEL
	destroy_tunnel_info(dev);
#endif

#if SFE_ACCELERATE_IPSEC
	sfe_destroy_xfrm_dst_bydev(dev);
#endif
}

/*
 * sfe_ipv4_periodic_sync()
 */
static void sfe_ipv4_periodic_sync(unsigned long arg)
{
	struct sfe_ipv4 *si = (struct sfe_ipv4 *)arg;
	uint64_t now_jiffies;
	int quota;
	sfe_sync_rule_callback_t sync_rule_callback;
	typeof(fastpath_nat_forward_on) fastpath_nat_on;
	typeof(qos_fastpath_nat_forward_on) qos_fastpath_nat_on;

	fastpath_nat_on = rcu_dereference(fastpath_nat_forward_on);
        if (fastpath_nat_on) {
		fastpath_nat_enabled = fastpath_nat_on();
	}
	qos_fastpath_nat_on = rcu_dereference(qos_fastpath_nat_forward_on);
        if (qos_fastpath_nat_on) {
                qos_fastpath_nat_enabled = qos_fastpath_nat_on();
        }
	now_jiffies = get_jiffies_64();

	rcu_read_lock();
	sync_rule_callback = rcu_dereference(si->sync_rule_callback);
	if (!sync_rule_callback) {
		rcu_read_unlock();
		goto done;
	}

	spin_lock_bh(&si->lock);
	sfe_ipv4_update_summary_stats(si);

	/*
	 * Get an estimate of the number of connections to parse in this sync.
	 */
	quota = (si->num_connections + 63) / 64;

	/*
	 * Walk the "active" list and sync the connection state.
	 */
	while (quota--) {
		struct sfe_ipv4_connection_match *cm;
		struct sfe_ipv4_connection_match *counter_cm;
		struct sfe_ipv4_connection *c;
		struct sfe_connection_sync sis;

		cm = si->active_head;
		if (!cm) {
			break;
		}

		/*
		 * There's a possibility that our counter match is in the active list too.
		 * If it is then remove it.
		 */
		counter_cm = cm->counter_match;
		if (counter_cm->active) {
			counter_cm->active = false;

			/*
			 * We must have a connection preceding this counter match
			 * because that's the one that got us to this point, so we don't have
			 * to worry about removing the head of the list.
			 */
			counter_cm->active_prev->active_next = counter_cm->active_next;

			if (likely(counter_cm->active_next)) {
				counter_cm->active_next->active_prev = counter_cm->active_prev;
			} else {
				si->active_tail = counter_cm->active_prev;
			}

			counter_cm->active_next = NULL;
			counter_cm->active_prev = NULL;
		}

		/*
		 * Now remove the head of the active scan list.
		 */
		cm->active = false;
		si->active_head = cm->active_next;
		if (likely(cm->active_next)) {
			cm->active_next->active_prev = NULL;
		} else {
			si->active_tail = NULL;
		}
		cm->active_next = NULL;

		/*
		 * Sync the connection state.
		 */
		c = cm->connection;
		sfe_ipv4_gen_sync_sfe_ipv4_connection(si, c, &sis, SFE_SYNC_REASON_STATS, now_jiffies);

		/*
		 * We don't want to be holding the lock when we sync!
		 */
		spin_unlock_bh(&si->lock);
		sync_rule_callback(&sis);
		spin_lock_bh(&si->lock);
	}

	spin_unlock_bh(&si->lock);
	rcu_read_unlock();

done:
	mod_timer(&si->timer, jiffies + ((HZ + 99) / 100));
}

#define CHAR_DEV_MSG_SIZE 768

/*
 * sfe_ipv4_debug_dev_read_start()
 *	Generate part of the XML output.
 */
static bool sfe_ipv4_debug_dev_read_start(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
					  int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	int bytes_read;

	si->debug_read_seq++;

	bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "<sfe_ipv4>\n");
	if (copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	ws->state++;
	return true;
}

/*
 * sfe_ipv4_debug_dev_read_connections_start()
 *	Generate part of the XML output.
 */
static bool sfe_ipv4_debug_dev_read_connections_start(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
						      int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	int bytes_read;

	bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t<connections>\n");
	if (copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	ws->state++;
	return true;
}

/*
 * sfe_ipv4_debug_dev_read_connections_connection()
 *	Generate part of the XML output.
 */
static bool sfe_ipv4_debug_dev_read_connections_connection(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
							   int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	struct sfe_ipv4_connection *c;
	struct sfe_ipv4_connection_match *original_cm;
	struct sfe_ipv4_connection_match *reply_cm;
	int bytes_read;
	int protocol;
	struct net_device *src_dev;
	__be32 src_ip;
	__be32 src_ip_xlate;
	__be16 src_port;
	__be16 src_port_xlate;
	uint64_t src_rx_packets;
	uint64_t src_rx_bytes;
	struct net_device *dest_dev;
	__be32 dest_ip;
	__be32 dest_ip_xlate;
	__be16 dest_port;
	__be16 dest_port_xlate;
	uint64_t dest_rx_packets;
	uint64_t dest_rx_bytes;
	uint64_t last_sync_jiffies;
	uint32_t mark, src_priority, dest_priority, src_dscp, dest_dscp;
#ifdef CONFIG_NF_FLOW_COOKIE
	int src_flow_cookie, dst_flow_cookie;
#endif
#if defined(CONFIG_NF_CONNTRACK_EXTMARK)
	uint32_t extmark;
#endif

	spin_lock_bh(&si->lock);

	for (c = si->all_connections_head; c; c = c->all_connections_next) {
		if (c->debug_read_seq < si->debug_read_seq) {
			c->debug_read_seq = si->debug_read_seq;
			break;
		}
	}

	/*
	 * If there were no connections then move to the next state.
	 */
	if (!c) {
		spin_unlock_bh(&si->lock);
		ws->state++;
		return true;
	}

	original_cm = c->original_match;
	reply_cm = c->reply_match;

	protocol = c->protocol;
	src_dev = c->original_dev;
	src_ip = c->src_ip;
	src_ip_xlate = c->src_ip_xlate;
	src_port = c->src_port;
	src_port_xlate = c->src_port_xlate;
	src_priority = original_cm->priority;
	src_dscp = original_cm->dscp >> SFE_IPV4_DSCP_SHIFT;

	sfe_ipv4_connection_match_update_summary_stats(original_cm);
	sfe_ipv4_connection_match_update_summary_stats(reply_cm);

	src_rx_packets = original_cm->rx_packet_count64;
	src_rx_bytes = original_cm->rx_byte_count64;
	dest_dev = c->reply_dev;
	dest_ip = c->dest_ip;
	dest_ip_xlate = c->dest_ip_xlate;
	dest_port = c->dest_port;
	dest_port_xlate = c->dest_port_xlate;
	dest_priority = reply_cm->priority;
	dest_dscp = reply_cm->dscp >> SFE_IPV4_DSCP_SHIFT;
	dest_rx_packets = reply_cm->rx_packet_count64;
	dest_rx_bytes = reply_cm->rx_byte_count64;
	last_sync_jiffies = get_jiffies_64() - c->last_sync_jiffies;
	mark = c->mark;
#if defined(CONFIG_NF_CONNTRACK_EXTMARK)
	extmark = c->extmark;
#endif
#ifdef CONFIG_NF_FLOW_COOKIE
	src_flow_cookie = original_cm->flow_cookie;
	dst_flow_cookie = reply_cm->flow_cookie;
#endif
	spin_unlock_bh(&si->lock);

	bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t\t<connection "
				"protocol=\"%u\" "
				"src_dev=\"%s\" "
				"src_ip=\"%pI4\" src_ip_xlate=\"%pI4\" "
				"src_port=\"%u\" src_port_xlate=\"%u\" "
				"src_priority=\"%u\" src_dscp=\"%u\" "
				"src_rx_pkts=\"%llu\" src_rx_bytes=\"%llu\" "
				"dest_dev=\"%s\" "
				"dest_ip=\"%pI4\" dest_ip_xlate=\"%pI4\" "
				"dest_port=\"%u\" dest_port_xlate=\"%u\" "
				"dest_priority=\"%u\" dest_dscp=\"%u\" "
				"dest_rx_pkts=\"%llu\" dest_rx_bytes=\"%llu\" "
#ifdef CONFIG_NF_FLOW_COOKIE
				"src_flow_cookie=\"%d\" dst_flow_cookie=\"%d\" "
#endif
				"last_sync=\"%llu\" "
#ifdef TPLINK_LAN_WAN_NETSTACK
				"partner=\"%p,%p,%p,%p\" "
#endif
#if defined(CONFIG_NF_CONNTRACK_EXTMARK)
				"extmark=\"%08x\" "
#endif
				"mark=\"%08x\" />\n",
				protocol,
				src_dev->name,
				&src_ip, &src_ip_xlate,
				ntohs(src_port), ntohs(src_port_xlate),
				src_priority, src_dscp,
				src_rx_packets, src_rx_bytes,
				dest_dev->name,
				&dest_ip, &dest_ip_xlate,
				ntohs(dest_port), ntohs(dest_port_xlate),
				dest_priority, dest_dscp,
				dest_rx_packets, dest_rx_bytes,
#ifdef CONFIG_NF_FLOW_COOKIE
				src_flow_cookie, dst_flow_cookie,
#endif
				last_sync_jiffies,
#ifdef TPLINK_LAN_WAN_NETSTACK
				original_cm,
				reply_cm,
				original_cm->partner_match,
				reply_cm->partner_match,
#endif
#if defined(CONFIG_NF_CONNTRACK_EXTMARK)
				extmark,
#endif
				mark);

	if (copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	return true;
}

/*
 * sfe_ipv4_debug_dev_read_connections_end()
 *	Generate part of the XML output.
 */
static bool sfe_ipv4_debug_dev_read_connections_end(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
						    int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	int bytes_read;

	bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t</connections>\n");
	if (copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	ws->state++;
	return true;
}

/*
 * sfe_ipv4_debug_dev_read_exceptions_start()
 *	Generate part of the XML output.
 */
static bool sfe_ipv4_debug_dev_read_exceptions_start(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
						     int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	int bytes_read;

	bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t<exceptions>\n");
	if (copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	ws->state++;
	return true;
}

/*
 * sfe_ipv4_debug_dev_read_exceptions_exception()
 *	Generate part of the XML output.
 */
static bool sfe_ipv4_debug_dev_read_exceptions_exception(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
							 int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	uint64_t ct;

	spin_lock_bh(&si->lock);
	ct = si->exception_events64[ws->iter_exception];
	spin_unlock_bh(&si->lock);

	if (ct) {
		int bytes_read;

		bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE,
				      "\t\t<exception name=\"%s\" count=\"%llu\" />\n",
				      sfe_ipv4_exception_events_string[ws->iter_exception],
				      ct);
		if (copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
			return false;
		}

		*length -= bytes_read;
		*total_read += bytes_read;
	}

	ws->iter_exception++;
	if (ws->iter_exception >= SFE_IPV4_EXCEPTION_EVENT_LAST) {
		ws->iter_exception = 0;
		ws->state++;
	}

	return true;
}

/*
 * sfe_ipv4_debug_dev_read_exceptions_end()
 *	Generate part of the XML output.
 */
static bool sfe_ipv4_debug_dev_read_exceptions_end(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
						   int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	int bytes_read;

	bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t</exceptions>\n");
	if (copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	ws->state++;
	return true;
}

/*
 * sfe_ipv4_debug_dev_read_stats()
 *	Generate part of the XML output.
 */
static bool sfe_ipv4_debug_dev_read_stats(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
					  int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	int bytes_read;
	unsigned int num_connections;
	uint64_t packets_forwarded;
	uint64_t packets_not_forwarded;
	uint64_t connection_create_requests;
	uint64_t connection_create_collisions;
	uint64_t connection_destroy_requests;
	uint64_t connection_destroy_misses;
	uint64_t connection_flushes;
	uint64_t connection_match_hash_hits;
	uint64_t connection_match_hash_reorders;

	spin_lock_bh(&si->lock);
	sfe_ipv4_update_summary_stats(si);

	num_connections = si->num_connections;
	packets_forwarded = si->packets_forwarded64;
	packets_not_forwarded = si->packets_not_forwarded64;
	connection_create_requests = si->connection_create_requests64;
	connection_create_collisions = si->connection_create_collisions64;
	connection_destroy_requests = si->connection_destroy_requests64;
	connection_destroy_misses = si->connection_destroy_misses64;
	connection_flushes = si->connection_flushes64;
	connection_match_hash_hits = si->connection_match_hash_hits64;
	connection_match_hash_reorders = si->connection_match_hash_reorders64;
	spin_unlock_bh(&si->lock);

	bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t<stats "
			      "num_connections=\"%u\" "
			      "pkts_forwarded=\"%llu\" pkts_not_forwarded=\"%llu\" "
			      "create_requests=\"%llu\" create_collisions=\"%llu\" "
			      "destroy_requests=\"%llu\" destroy_misses=\"%llu\" "
			      "flushes=\"%llu\" "
			      "hash_hits=\"%llu\" hash_reorders=\"%llu\" />\n",
			      num_connections,
			      packets_forwarded,
			      packets_not_forwarded,
			      connection_create_requests,
			      connection_create_collisions,
			      connection_destroy_requests,
			      connection_destroy_misses,
			      connection_flushes,
			      connection_match_hash_hits,
			      connection_match_hash_reorders);
	if (copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	ws->state++;
	return true;
}

/*
 * sfe_ipv4_debug_dev_read_end()
 *	Generate part of the XML output.
 */
static bool sfe_ipv4_debug_dev_read_end(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
					int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	int bytes_read;

	bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "</sfe_ipv4>\n");
	if (copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	ws->state++;
	return true;
}

#ifdef TPLINK_PPP_TUNNEL_DUMP
static bool sfe_ipv4_debug_dev_read_tunnel_info_start(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
					  int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	int bytes_read;

	bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t<tunnel>\n");
	if (copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	ws->state++;
	return true;

}
static bool sfe_ipv4_debug_dev_read_tunnel_info_infomation(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
					  int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	struct tunnel_info *ti = NULL;
	int i = 0;
	int bytes_read = 0;

	struct net_device *tunnel_dev;
	struct net_device *ppp_dev;
	unsigned char *smac;
	unsigned char *dmac;
	uint32_t *saddr;
	uint32_t *daddr;
	uint16_t pppoe_sid;
	unsigned char *pppoe_source;
	uint16_t pptp_call_id;
	uint32_t *pptp_saddr;
	uint32_t l2tp_tunnel_id;
	uint32_t l2tp_session_id;
	uint16_t l2tp_sport;
	uint16_t l2tp_dport;
	uint32_t magic;
#if SFE_ACCELERATE_IPSEC
	uint32_t i_spi;
	uint32_t o_spi;
	int 	 encap;
#endif

	read_lock_bh(&tunnel_rwlock);
	for (i = 0; i < TUNNEL_MAX; i++) {
		if (TUNNEL_LIST[i].tunnel_dev) {
			if (TUNNEL_LIST[i].debug_read_seq < si->debug_read_seq) {
				ti = &TUNNEL_LIST[i];
				ti->debug_read_seq = si->debug_read_seq;
				break;
			}
		}
	}

	if (!ti) {
		read_unlock_bh(&tunnel_rwlock);
		ws->state++;
		return true;
	}
	
	tunnel_dev = ti->tunnel_dev;
	ppp_dev = ti->ppp_dev;
	smac = (unsigned char *)ti->src_mac;
	dmac = (unsigned char *)ti->dest_mac;
	magic = ti->magic_number;
	switch(ti->type) {
	case TUNNEL_PPPOE:
		pppoe_sid = ntohs(ti->info.pppoe.sid);
		pppoe_source = ti->info.pppoe.source;
		read_unlock_bh(&tunnel_rwlock);
		bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t\t<PPPoE "
				  "tunnel_dev=\"%s\" ppp_dev=\"%s\" "
				  "smac=\"%pM\" dmac=\"%pM\" "
				  "sid=\"%u\" source=\"%pM\" magic=\"%u\"/>\n",
				  tunnel_dev?tunnel_dev->name:"NULL", ppp_dev?ppp_dev->name:"NULL",
				  smac, smac,
				  pppoe_sid, pppoe_source, magic);
		break;
	case TUNNEL_PPTP:
		saddr = &ti->saddr;
		daddr = &ti->daddr;
		pptp_call_id = ti->info.pptp.call_id;
		pptp_saddr = &ti->info.pptp.saddr;
		read_unlock_bh(&tunnel_rwlock);
		bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t\t<PPTP "
				  "tunnel_dev=\"%s\" ppp_dev=\"%s\" "
				  "smac=\"%pM\" dmac=\"%pM\" "
				  "saddr=\"%pI4\" daddr=\"%pI4\" "
				  "call_id=\"%u\" src_addr=\"%pI4\" magic=\"%u\"/>\n",
				  tunnel_dev?tunnel_dev->name:"NULL", ppp_dev?ppp_dev->name:"NULL",
				  smac, dmac,
				  saddr, daddr,
				  pptp_call_id, pptp_saddr, magic);
		break;
	case TUNNEL_L2TP:
		saddr = &ti->saddr;
		daddr = &ti->daddr;
		l2tp_tunnel_id = ti->info.l2tp.tunnel_id;
		l2tp_session_id = ti->info.l2tp.session_id;
		l2tp_sport = ntohs(ti->info.l2tp.sport);
		l2tp_dport = ntohs(ti->info.l2tp.dport);
#if SFE_ACCELERATE_IPSEC
		encap = ti->info.l2tp.encap;
		i_spi = encap ? (ti->info.l2tp.x_in ? ntohl(ti->info.l2tp.x_in->id.spi) : 0) : 0;
		o_spi = encap ? (ti->info.l2tp.x_out ? ntohl(ti->info.l2tp.x_out->id.spi) : 0) : 0;
#endif
		read_unlock_bh(&tunnel_rwlock);
#ifndef SFE_ACCELERATE_IPSEC
		bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t\t<L2TP "
				  "tunnel_dev=\"%s\" ppp_dev=\"%s\" "
				  "smac=\"%pM\" dmac=\"%pM\" "
				  "saddr=\"%pI4\" daddr=\"%pI4\" "
				  "tunnel_id=\"%u\" session_id=\"%u\" sport=\"%u\" dport=\"%u\" magic=\"%u\"/>\n",
				  tunnel_dev?tunnel_dev->name:"NULL", ppp_dev?ppp_dev->name:"NULL",
				  smac, dmac,
				  saddr, daddr,
				  l2tp_tunnel_id, l2tp_session_id, l2tp_sport, l2tp_dport, magic,
				  );
#else
		bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t\t<L2TP "
				  "tunnel_dev=\"%s\" ppp_dev=\"%s\" "
				  "smac=\"%pM\" dmac=\"%pM\" "
				  "saddr=\"%pI4\" daddr=\"%pI4\" "
				  "tunnel_id=\"%u\" session_id=\"%u\" sport=\"%u\" dport=\"%u\" magic=\"%u\" "
				  "encap=\"%d\" i_spi=\"0x%x\" o_spi=\"0x%x\" />\n",
				  tunnel_dev?tunnel_dev->name:"NULL", ppp_dev?ppp_dev->name:"NULL",
				  smac, dmac,
				  saddr, daddr,
				  l2tp_tunnel_id, l2tp_session_id, l2tp_sport, l2tp_dport, magic, encap, i_spi, o_spi
				  );
#endif
		break;
	default:
		read_unlock_bh(&tunnel_rwlock);
		break;
	}

	if (bytes_read && copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	return true;
}

static bool sfe_ipv4_debug_dev_read_tunnel_info_end(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
					  int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	int bytes_read;

	bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t</tunnel>\n");
	if (copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	ws->state++;
	return true;
}
#endif

#ifdef TPLINK_PPP_TUNNEL_DUMP
static bool sfe_ipv4_debug_dev_read_xfrm_dst_start(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
					  int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	int bytes_read;

	bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t<xfrm_dst>\n");
	if (copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	ws->state++;
	return true;

}

static bool sfe_ipv4_debug_dev_read_xfrm_dst_infomation(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
					  int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	struct sfe_esp_xfrm_dst *dst = NULL;
	int i = 0;
	int bytes_read = 0;

	struct net_device *dst_dev;
	read_lock_bh(&xfrm_dst_rwlock);

	for (i = 0; i < XFRM_DST_MAX; i++) {
		if (SFE_XFRM_DST[i].dev) {
			if (SFE_XFRM_DST[i].debug_read_seq < si->debug_read_seq) {
				dst = &SFE_XFRM_DST[i];
				dst->debug_read_seq = si->debug_read_seq;
				break;
			}
		}
	}

	if (!dst) {
		read_unlock_bh(&xfrm_dst_rwlock);
		ws->state++;
		return true;
	}

	bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t\t<XFRM_DST "
			"xfrm_state=\"0x%x\" dst_dev=\"%s\" "
			"smac=\"%pM\" dmac=\"%pM\" "
			"saddr=\"%pI4\" daddr=\"%pI4\" />\n",
			dst->x ? ntohl(dst->x->id.spi) : 0, dst->dev ? dst->dev->name : "NULL",
			dst->src_mac, dst->dest_mac, &dst->src_ip, &dst->dst_ip
			);

	read_unlock_bh(&xfrm_dst_rwlock);
	if (bytes_read && copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	return true;
}

static bool sfe_ipv4_debug_dev_read_xfrm_dst_end(struct sfe_ipv4 *si, char *buffer, char *msg, size_t *length,
					  int *total_read, struct sfe_ipv4_debug_xml_write_state *ws)
{
	int bytes_read;

	bytes_read = snprintf(msg, CHAR_DEV_MSG_SIZE, "\t</xfrm_dst>\n");
	if (copy_to_user(buffer + *total_read, msg, CHAR_DEV_MSG_SIZE)) {
		return false;
	}

	*length -= bytes_read;
	*total_read += bytes_read;

	ws->state++;
	return true;
}
#endif

/*
 * Array of write functions that write various XML elements that correspond to
 * our XML output state machine.
 */
sfe_ipv4_debug_xml_write_method_t sfe_ipv4_debug_xml_write_methods[SFE_IPV4_DEBUG_XML_STATE_DONE] = {
	sfe_ipv4_debug_dev_read_start,
	sfe_ipv4_debug_dev_read_connections_start,
	sfe_ipv4_debug_dev_read_connections_connection,
	sfe_ipv4_debug_dev_read_connections_end,
	sfe_ipv4_debug_dev_read_exceptions_start,
	sfe_ipv4_debug_dev_read_exceptions_exception,
	sfe_ipv4_debug_dev_read_exceptions_end,
	sfe_ipv4_debug_dev_read_stats,
#ifdef TPLINK_PPP_TUNNEL_DUMP
	sfe_ipv4_debug_dev_read_tunnel_info_start,
	sfe_ipv4_debug_dev_read_tunnel_info_infomation,
	sfe_ipv4_debug_dev_read_tunnel_info_end,
#endif
#ifdef TPLINK_XFRM_DST_DUMP
	sfe_ipv4_debug_dev_read_xfrm_dst_start,
	sfe_ipv4_debug_dev_read_xfrm_dst_infomation,
	sfe_ipv4_debug_dev_read_xfrm_dst_end,
#endif
	sfe_ipv4_debug_dev_read_end,
};

/*
 * sfe_ipv4_debug_dev_read()
 *	Send info to userspace upon read request from user
 */
static ssize_t sfe_ipv4_debug_dev_read(struct file *filp, char *buffer, size_t length, loff_t *offset)
{
	char msg[CHAR_DEV_MSG_SIZE];
	int total_read = 0;
	struct sfe_ipv4_debug_xml_write_state *ws;
	struct sfe_ipv4 *si = &__si;

	ws = (struct sfe_ipv4_debug_xml_write_state *)filp->private_data;
	while ((ws->state != SFE_IPV4_DEBUG_XML_STATE_DONE) && (length > CHAR_DEV_MSG_SIZE)) {
		if ((sfe_ipv4_debug_xml_write_methods[ws->state])(si, buffer, msg, &length, &total_read, ws)) {
			continue;
		}
	}

	return total_read;
}

/*
 * sfe_ipv4_debug_dev_write()
 *	Write to char device resets some stats
 */
static ssize_t sfe_ipv4_debug_dev_write(struct file *filp, const char *buffer, size_t length, loff_t *offset)
{
	struct sfe_ipv4 *si = &__si;

	spin_lock_bh(&si->lock);
	sfe_ipv4_update_summary_stats(si);

	si->packets_forwarded64 = 0;
	si->packets_not_forwarded64 = 0;
	si->connection_create_requests64 = 0;
	si->connection_create_collisions64 = 0;
	si->connection_destroy_requests64 = 0;
	si->connection_destroy_misses64 = 0;
	si->connection_flushes64 = 0;
	si->connection_match_hash_hits64 = 0;
	si->connection_match_hash_reorders64 = 0;
	spin_unlock_bh(&si->lock);

	return length;
}

/*
 * sfe_ipv4_debug_dev_open()
 */
static int sfe_ipv4_debug_dev_open(struct inode *inode, struct file *file)
{
	struct sfe_ipv4_debug_xml_write_state *ws;

	ws = (struct sfe_ipv4_debug_xml_write_state *)file->private_data;
	if (!ws) {
		ws = kzalloc(sizeof(struct sfe_ipv4_debug_xml_write_state), GFP_KERNEL);
		if (!ws) {
			return -ENOMEM;
		}

		ws->state = SFE_IPV4_DEBUG_XML_STATE_START;
		file->private_data = ws;
	}

	return 0;
}

/*
 * sfe_ipv4_debug_dev_release()
 */
static int sfe_ipv4_debug_dev_release(struct inode *inode, struct file *file)
{
	struct sfe_ipv4_debug_xml_write_state *ws;

	ws = (struct sfe_ipv4_debug_xml_write_state *)file->private_data;
	if (ws) {
		/*
		 * We've finished with our output so free the write state.
		 */
		kfree(ws);
	}

	return 0;
}

/*
 * File operations used in the debug char device
 */
static struct file_operations sfe_ipv4_debug_dev_fops = {
	.read = sfe_ipv4_debug_dev_read,
	.write = sfe_ipv4_debug_dev_write,
	.open = sfe_ipv4_debug_dev_open,
	.release = sfe_ipv4_debug_dev_release
};

#ifdef CONFIG_NF_FLOW_COOKIE
/*
 * sfe_register_flow_cookie_cb
 *	register a function in SFE to let SFE use this function to configure flow cookie for a flow
 *
 * Hardware driver which support flow cookie should register a callback function in SFE. Then SFE
 * can use this function to configure flow cookie for a flow.
 * return: 0, success; !=0, fail
 */
int sfe_register_flow_cookie_cb(flow_cookie_set_func_t cb)
{
	struct sfe_ipv4 *si = &__si;

	BUG_ON(!cb);

	if (si->flow_cookie_set_func) {
		return -1;
	}

	rcu_assign_pointer(si->flow_cookie_set_func, cb);
	return 0;
}

/*
 * sfe_unregister_flow_cookie_cb
 *	unregister function which is used to configure flow cookie for a flow
 *
 * return: 0, success; !=0, fail
 */
int sfe_unregister_flow_cookie_cb(flow_cookie_set_func_t cb)
{
	struct sfe_ipv4 *si = &__si;

	RCU_INIT_POINTER(si->flow_cookie_set_func, NULL);
	return 0;
}

/*
 * sfe_ipv4_get_flow_cookie()
 */
static ssize_t sfe_ipv4_get_flow_cookie(struct device *dev,
					struct device_attribute *attr,
					char *buf)
{
	struct sfe_ipv4 *si = &__si;
	return snprintf(buf, (ssize_t)PAGE_SIZE, "%d\n", si->flow_cookie_enable);
}

/*
 * sfe_ipv4_set_flow_cookie()
 */
static ssize_t sfe_ipv4_set_flow_cookie(struct device *dev,
					struct device_attribute *attr,
					const char *buf, size_t size)
{
	struct sfe_ipv4 *si = &__si;
	strict_strtol(buf, 0, (long int *)&si->flow_cookie_enable);

	return size;
}

/*
 * sysfs attributes.
 */
static const struct device_attribute sfe_ipv4_flow_cookie_attr =
	__ATTR(flow_cookie_enable, S_IWUSR | S_IRUGO, sfe_ipv4_get_flow_cookie, sfe_ipv4_set_flow_cookie);
#endif /*CONFIG_NF_FLOW_COOKIE*/

#if SFE_ACCELERATE_IPSEC
static void sfe_ipv4_destroy_all_rules_for_xfrm(struct xfrm_state *x)
{
    struct sfe_ipv4 *si = &__si;
    struct sfe_ipv4_connection *c;

    if (x == NULL)
    {
        return;
    }

another_round:
    spin_lock_bh(&si->lock);

    for (c = si->all_connections_head; c; c = c->all_connections_next) {
        /*
         * Does this connection relate to the xfrm_state we are destroying?
         */
		if (((c->original_match) && (c->original_match->x == x))
			|| ((c->reply_match) && (c->reply_match->x == x)))
		{
			break;
		}
    }

    if (c) {
        sfe_ipv4_remove_sfe_ipv4_connection(si, c);
    }

    spin_unlock_bh(&si->lock);
    DEBUG_INFO("Connection destroyed all\n");
    if (c) {
        sfe_ipv4_flush_sfe_ipv4_connection(si, c, SFE_SYNC_REASON_DESTROY);
        goto another_round;
    }
#ifdef TPLINK_PPP_TUNNEL
    destroy_tunnel_info_by_xfrm_state(x);
#endif
}

static void sfe_ipv4_update_rules_for_xfrm(struct xfrm_state *x)
{
    struct sfe_ipv4 *si = &__si;
    struct sfe_ipv4_connection *c;
	struct xfrm_state *conn_x = NULL;
	int i = 0;
	int k = 0;
    if (x == NULL)
    {
        return;
    }
    spin_lock_bh(&si->lock);
    for (c = si->all_connections_head; c; c = c->all_connections_next) {
        /*
         * Does this connection relate to the xfrm_state we are destroying?
         */
		if ((c->original_match) && (c->original_match->x))
		{
			conn_x = c->original_match->x;
			if (sfe_xfrm_state_equal(x, conn_x))
			{
				xfrm_state_put(conn_x);
				xfrm_state_hold(x);
				c->original_match->x = x;
			}
		}
		
		if ((c->reply_match) && (c->reply_match->x))
		{
			conn_x = c->reply_match->x;
			if (sfe_xfrm_state_equal(x, conn_x))
			{
				xfrm_state_put(conn_x);
				xfrm_state_hold(x);
				c->reply_match->x = x;
			}
		}
    }
	spin_unlock_bh(&si->lock);

#ifdef TPLINK_PPP_TUNNEL
    update_tunnel_info_by_xfrm_state(x);
#endif
}


static int sfe_acquire(struct xfrm_state *x, struct xfrm_tmpl *xt, struct xfrm_policy *xp)
{
    /* do nothing */
    return 0;
}

static int sfe_xfrm_state_notify(struct xfrm_state *x, const struct km_event *c)
{    
    DEBUG_INFO(KERN_NOTICE "xfrm_user: received SA event %d\n", c->event);

	switch (c->event) 
    {
	case XFRM_MSG_EXPIRE:
		DEBUG_TRACE("do nothing with this event%d, spi:0x%x\n", c->event, ntohl(x->id.spi));
		break;
    case XFRM_MSG_DELSA:
    case XFRM_MSG_FLUSHSA:
		DEBUG_INFO("del/flush this event%d, spi:0x%x\n", c->event, ntohl(x->id.spi));
		sfe_ipv4_destroy_all_rules_for_xfrm(x);
		sfe_destroy_xfrm_dst_byxfrm(x);
		break;

	case XFRM_MSG_NEWAE:
	case XFRM_MSG_UPDSA:
	case XFRM_MSG_NEWSA:
		sfe_ipv4_update_rules_for_xfrm(x);
		sfe_update_xfrm_dst(x);
       break;

	default:
		DEBUG_INFO(KERN_NOTICE "xfrm_user: Unknown SA event %d\n", c->event);
		break;
	}

	return 0;

}

static struct xfrm_mgr sfe_mgr = {
	.id	= "sfe_xfrm_mgr",
	.notify	= sfe_xfrm_state_notify,
	.acquire = sfe_acquire,
	.compile_policy	= NULL,
	.notify_policy = NULL,
	.report	= NULL,
	.migrate= NULL,
	.new_mapping = NULL,
	.is_alive = NULL,
};
#endif

/*
 * sfe_ipv4_init()
 */
static int __init sfe_ipv4_init(void)
{
	struct sfe_ipv4 *si = &__si;
	int result = -1;

	DEBUG_INFO("SFE IPv4 init\n");

	/*
	 * Create sys/sfe_ipv4
	 */
	si->sys_sfe_ipv4 = kobject_create_and_add("sfe_ipv4", NULL);
	if (!si->sys_sfe_ipv4) {
		DEBUG_ERROR("failed to register sfe_ipv4\n");
		goto exit1;
	}

	/*
	 * Create files, one for each parameter supported by this module.
	 */
	result = sysfs_create_file(si->sys_sfe_ipv4, &sfe_ipv4_debug_dev_attr.attr);
	if (result) {
		DEBUG_ERROR("failed to register debug dev file: %d\n", result);
		goto exit2;
	}

#ifdef CONFIG_NF_FLOW_COOKIE
	result = sysfs_create_file(si->sys_sfe_ipv4, &sfe_ipv4_flow_cookie_attr.attr);
	if (result) {
		DEBUG_ERROR("failed to register flow cookie enable file: %d\n", result);
		goto exit3;
	}
#endif /* CONFIG_NF_FLOW_COOKIE */

	/*
	 * Register our debug char device.
	 */
	result = register_chrdev(0, "sfe_ipv4", &sfe_ipv4_debug_dev_fops);
	if (result < 0) {
		DEBUG_ERROR("Failed to register chrdev: %d\n", result);
		goto exit4;
	}

	si->debug_dev = result;

	/*
	 * Create a timer to handle periodic statistics.
	 */
	setup_timer(&si->timer, sfe_ipv4_periodic_sync, (unsigned long)si);
	mod_timer(&si->timer, jiffies + ((HZ + 99) / 100));

#if SFE_ACCELERATE_IPSEC
    if(0 < xfrm_register_km(&sfe_mgr))
    {
        DEBUG_ERROR("Failed to register xfrm km \n");
        goto exit4;
    }
#endif

	spin_lock_init(&si->lock);

#ifdef SFE_IPSTAT
	sfe_ipstat_init();
#endif
	return 0;

exit4:
#ifdef CONFIG_NF_FLOW_COOKIE
	sysfs_remove_file(si->sys_sfe_ipv4, &sfe_ipv4_flow_cookie_attr.attr);

exit3:
#endif /* CONFIG_NF_FLOW_COOKIE */
	sysfs_remove_file(si->sys_sfe_ipv4, &sfe_ipv4_debug_dev_attr.attr);

exit2:
	kobject_put(si->sys_sfe_ipv4);

exit1:
	return result;
}

/*
 * sfe_ipv4_exit()
 */
static void __exit sfe_ipv4_exit(void)
{
	struct sfe_ipv4 *si = &__si;

	DEBUG_INFO("SFE IPv4 exit\n");

	/*
	 * Destroy all connections.
	 */
	sfe_ipv4_destroy_all_rules_for_dev(NULL);

	del_timer_sync(&si->timer);

	unregister_chrdev(si->debug_dev, "sfe_ipv4");

#ifdef CONFIG_NF_FLOW_COOKIE
	sysfs_remove_file(si->sys_sfe_ipv4, &sfe_ipv4_flow_cookie_attr.attr);
#endif /* CONFIG_NF_FLOW_COOKIE */
	sysfs_remove_file(si->sys_sfe_ipv4, &sfe_ipv4_debug_dev_attr.attr);

	kobject_put(si->sys_sfe_ipv4);

#if SFE_ACCELERATE_IPSEC
    xfrm_unregister_km(&sfe_mgr);
#endif

#ifdef SFE_IPSTAT
	sfe_ipstat_exit();
#endif
}

module_init(sfe_ipv4_init)
module_exit(sfe_ipv4_exit)

#ifdef TPLINK_LAN_WAN_NETSTACK
EXPORT_SYMBOL(sfe_ipv4_destroy_inner_dev);
EXPORT_SYMBOL(sfe_ipv4_create_inner_dev);
EXPORT_SYMBOL(sfe_ipv4_recv_pppoe);
#endif
EXPORT_SYMBOL(sfe_ipv4_recv);
EXPORT_SYMBOL(sfe_ipv4_create_rule);
EXPORT_SYMBOL(sfe_ipv4_destroy_rule);
EXPORT_SYMBOL(sfe_ipv4_destroy_all_rules_for_dev);
EXPORT_SYMBOL(sfe_ipv4_register_sync_rule_callback);
EXPORT_SYMBOL(sfe_ipv4_mark_rule);
EXPORT_SYMBOL(sfe_ipv4_update_rule);
#ifdef CONFIG_NF_FLOW_COOKIE
EXPORT_SYMBOL(sfe_register_flow_cookie_cb);
EXPORT_SYMBOL(sfe_unregister_flow_cookie_cb);
#endif

MODULE_AUTHOR("Qualcomm Atheros Inc.");
MODULE_DESCRIPTION("Shortcut Forwarding Engine - IPv4 edition");
MODULE_LICENSE("Dual BSD/GPL");

